/*********************************************************************
 *
 * DESCRIPTION:
 *      This file contains the CreekDocumentGroup class which handles
 *      all of the documents in the Creek program. Other classes in
 *      the system will use this class to access the documents.
 *
 ********************************************************************/

package jcreek.cke;

import jcreek.cke.command.CreekCommand;
import jcreek.cke.document.CreekDocument;
import jcreek.cke.graph.CreekModel;
import jcreek.cke.document.KnowledgeModelController;
import jcreek.cke.command.UndoAction;
import jcreek.cke.command.RedoAction;
import jcreek.representation.*; import jcreek.representation.cbr.*;
// import jcreek.cke.command.SaveViewAction;
import jcreek.cke.browser.CreekUndoManager;
import jcreek.cke.browser.UndoHandler;
import jcreek.cke.conceptview.*;
import jcreek.cke.graph.VertexList;
import jcreek.cke.graph.RelationList;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.*;
import javax.swing.ProgressMonitorInputStream;
import java.awt.print.PageFormat;
import java.awt.print.PrinterJob;
import java.util.Arrays;

import jcreek.cke.graph.Graph;


/**
 * This class manages all of the CreekDocument objects in the program.
 * Use this class when you want to perform operations on a document,
 * rather than calling the document directly. For the less common
 * operations, the user can easily get at the current document and
 * communicate with it.
 * <p>
 * This is a singleton class, which means there's only one instance of
 * this class in the system. To get a reference to that single instance,
 * use the instanceOf() method.
 *
 * @author
 * @version
 */


public class CreekDocumentGroup
{

    /**
     * Reference to the Knowledge Model
     */
    protected  KnowledgeModel km;
   /**
     * Reference to the single instance of this class.
     */
    protected static CreekDocumentGroup instance;
    /**
     * Index into documents to the current document.
     */
    protected int curDocument;

    /**
     * List of documents held under this document group.
     */
    protected Vector documents;
    /**
     * Table of the documents, keyed by the associated views.
     * Used by MainWindow and others to find the document
     * associated with a given view object.
     */
    protected Hashtable viewDocs;
    /**
     * Table of the supported document types, keyed by the
     * filename extension.
     */
    protected Hashtable docTypes;
    /**
     * Table of the supported document types, keyed by the
     * filename extension.
     */
    protected Hashtable kmTypes;
    /**
     * Undo handler that handles everything to do with undo.
     * As documents are created and opened, it will be added
     * as an undoable edit listener.
     */
    protected UndoHandler undoHandler;
    /**
     * Makes supporting property change listeners trivial.
     * When the current document becomes dirty these listeners are
     * notified via the PropertyChangeEvent. The listeners must
     * implement the PropertyChangeListener interface.
     */
    protected PropertyChangeSupport propSupport;
    /**
     * Default filename extension for new documents. This is used
     * to map to the matching document class in docTypes.
     */
    protected String defaultExtensionDoc;
    /**
     * Default filename extension for new documents. This is used
     * to map to the matching document class in docTypes.
     */
    protected String defaultExtensionKM;
    /**
     * Default page format for Printing the Documents
     */
    protected PageFormat pf;

    /**
     * No-arg constructor for our class. This is protected so we can
     * control the number of instances of this class. If you wish to
     * get a reference to the single instance of this class, use the
     * instanceOf() method.
     */
    protected CreekDocumentGroup() {
        documents = new Vector();
        viewDocs = new Hashtable();
        docTypes = new Hashtable();
        kmTypes  = new Hashtable();
        undoHandler = new UndoHandler();
        propSupport = new PropertyChangeSupport( this );
    } // CreekDocumentGroup

    /**
     * Add the given Creek document object to group and set it as
     * the current document. Also registers this document group
     * as an undoable edit listener, to make sure that undoable
     * edits are registered with the undo manager.
     *
     * @param  doc  document to add to group
     */
    protected synchronized void addDocument( CreekDocument doc ) {
        if ( doc != null ) {
            viewDocs.put( doc.getView(), doc );
            documents.addElement( doc );
            // addElement always adds at the end
            curDocument = documents.size() - 1;
            doc.addUndoableEditListener( undoHandler );
            fireDocAdded(doc);
        }
    } // addDocument

    /**
     * Add a property change listener to this document group.
     *
     * @param  l  new listener to add notification list
     */
    public void addPropertyChangeListener( PropertyChangeListener l ) {
        propSupport.addPropertyChangeListener( l );
    } // addPropertyChangeListener

    /**
     * Add a property change listener to this document group.
     *
     * @param  name  name of property to listen to specifically
     * @param  l     new listener to add notification list
     */
    public void addPropertyChangeListener( String name,
                                           PropertyChangeListener l ) {
        propSupport.addPropertyChangeListener( name, l );
    } // addPropertyChangeListener

    /**
     * Closes the current document. This leaves the group without
     * a current document until the user selects one via the
     * internal windows. This does not save the document.
     *
     * @return  true if successful, false if no current document
     */
    public boolean closeDocument() {
        CreekDocument oldDoc = getCurrentDocument();
        if ( oldDoc == null ) {
            return false;
        }
        // remove all edits associated with this document
        undoHandler.removeEdits( oldDoc );
        undoHandler.update();
        // removing the document sets the curDocument to -1
        removeDocument( oldDoc );
        //curDocument = -1;

        // be sure to close document close to last, otherwise
        // many of its fields will be set null
        oldDoc.removeUndoableEditListener(undoHandler);
        oldDoc.close();
        fireDocChange( oldDoc, getCurrentDocument() );
        return true;
    } // closeDocument

    public boolean deleteContext(CreekDocument selDoc)
    {
        if(documents.contains(selDoc))
        {
            CreekDocument oldDoc = selDoc;
            if ( oldDoc == null ) {
                return false;
            }
            // remove all edits associated with this document
            undoHandler.removeEdits( oldDoc );
            undoHandler.update();
            // removing the document sets the curDocument to -1
            removeDocument( oldDoc );

            // be sure to close document close to last, otherwise
            // many of its fields will be set null
            oldDoc.close();
            km.removeContext((Context)selDoc);
            fireDocChange( selDoc, getCurrentDocument() );

        }
        else
        {
            selDoc.close();
            km.removeContext((Context)selDoc);
            fireDocChange( selDoc, getCurrentDocument() );

        }
        return true;
    }

    public boolean closeModel()
    {
        KnowledgeModel oldKM = km;
//        km.removeModelChangeListener(this); //0312 change
        km = null;
//        fireKMChanged(oldKM,null); //0312 change
        fireModelClosed(oldKM);
        return true;
    }

    /**
     * Let all the property change listeners know that the current
     * document has become dirty or become clean.
     *
     * @param  doc  document that has changed
     */
    protected void fireDirtyChange( CreekDocument doc ) {
        if ( doc != null ) {
            boolean dirty = doc.isDirty();
            if ( dirty ) {
                propSupport.firePropertyChange( "dirty", false, true );
            } else {
                propSupport.firePropertyChange( "dirty", true, false );
            }
        }
    } // fireDirtyChange

    /**
     * Let all the property change listeners know that the knowledge model
     * editable status has changed.
     *
     * @param  doc  document that has changed
     */
    protected void fireKMEditChange(boolean newEdit)
    {
      if(newEdit != kmEditable)
      {
        propSupport.firePropertyChange( "kmEditable", kmEditable, newEdit );
      }
    } // fireDirtyChange

    /**
     * Let all the property change listeners know that the current
     * document has been switched with another document, making
     * that other document the new current document.
     *
     * @param  oldDoc  previous current document
     * @param  newDoc  new current document
     */
    protected void fireDocChange( CreekDocument oldDoc,
                                  CreekDocument newDoc ) {
        propSupport.firePropertyChange( "current", oldDoc, newDoc );
    } // fireDocChange

    /**
     * Let all the property change listeners know that the current
     * document has been switched with another document, making
     * that other document the new current document.
     *
     * @param  oldDoc  previous current document
     * @param  newDoc  new current document
     */
    protected void fireDocRemoved( CreekDocument oldDoc) {
        propSupport.firePropertyChange( "removed", oldDoc,null);
    } // fireDocChange


    protected void fireDocAdded( CreekDocument oldDoc) {
        propSupport.firePropertyChange( "added", oldDoc,null);
    } // fireDocChange


    /**
     * Let all the property change listeners know that the current
     * knowledge model has been opened
     *
     * @param  newKM  new KnowledgeModel
     */
    protected void fireModelOpened (KnowledgeModel newKM ) {
        propSupport.firePropertyChange( "opened",null,newKM );
    } // fireModelOpened

    /**
     * Let all the property change listeners know that the current
     * knowledge model has been closed
     *
     * @param  oldKM  old KnowledgeModel
     */
    protected void fireModelClosed (KnowledgeModel oldKM ) {
        propSupport.firePropertyChange( "closed",oldKM,null );
    } // fireModelClosed

    /**
     * Let all the property change listeners know that the current
     * knowledge model has been opened
     *
     * @param  oldDoc  previous current document
     * @param  newDoc  new current document
     */
    protected void fireSelectionChange( CreekDocument oldDoc ) {
        propSupport.firePropertyChange( "selection", oldDoc,null);
    } // fireDocChange

    /**
     * Returns the reference to the current document in the group.
     * If there is no current document, null is returned.
     *
     * @return  reference to current document, or null
     */
    public CreekDocument getCurrentDocument() {
        return getDocument( curDocument );
    } // getCurrentDocument

    /**
     * Returns the reference to the requested document in the group.
     * If there is no such document, null is returned.
     *
     * @param  index  index of the document to be retrieved
     * @return  reference to current document, or null
     */
    public CreekDocument getDocument( int index ) {
        try {
            return (CreekDocument)documents.elementAt( index );
        } catch ( ArrayIndexOutOfBoundsException aioobe ) {
            return null;
        }
    } // getDocument

    /**
     * Retrieve the document object associated with the given view.
     *
     * @param  view  view showing the requested document
     */
    public CreekDocument getDocument( ConceptView view ) {
        return (CreekDocument)viewDocs.get( view );
    } // getDocument


    /**
     * Returns an array of the current documents in the group. If there
     * are no documents, null is returned.
     *
     * @return  array of Creek documents, or null
     */
      public synchronized CreekDocument[] getDocuments() {
          int numDocs = documents.size();
          return (CreekDocument[])documents.toArray
              ( new CreekDocument[numDocs]);
      } // getDocuments

      public Context[] getContexts()
      {
          if(km!= null)
          {
                 return km.getContexts();
          }
          else
            return new Context[0];
      }
    /**
     * Retrieve the KnowledgeModel.
     *
     */
    public KnowledgeModel getKM(  ) {
        return km;
    } // getKM


    /**
     * Retrieve a list of all the registered document types.
     * Returns the list in the form of a hashtable for the
     * convenience of looking up the types by extension.
     * Each key of the hastable is a filename extension, while
     * the value is the document class that supports files
     * with that extension.
     *
     * @return  hashtable of registered document types
     */
    public Hashtable getDocumentTypes() {
        return docTypes;
    } // getDocumentTypes

    public Hashtable getKMTypes() {
        return kmTypes;
    } // getDocumentTypes


    /**
     * Returns the reference to the single instance of this class.
     *
     * @return  reference to this singleton
     */
    public static CreekDocumentGroup instanceOf() {
        if ( instance == null ) {
            instance = new CreekDocumentGroup();
        }
        return instance;
    } // instanceOf

    /**
     * Sends out notification to any listeners if this document
     * is the current document. This is used by the documents
     * to let the document group (and its listeners) know that
     * it is dirty. It is _not_ meant to mark the document dirty.
     */
    public void markDocumentDirty( CreekDocument doc ) {
        if ( doc == getCurrentDocument() ) {
            // send out notification that current document's
            // dirty state has changed
            fireDirtyChange( doc );
        }
    } // markDocumentDirty

    /**
     * Sends out notification to any listeners if this document
     * is the current document. This is used by the documents
     * to let the document group (and its listeners) know that
     * a slection change has taken place.
     */
    public void markSelection( CreekDocument doc ) {
        if ( doc == getCurrentDocument() ) {
            // send out notification that current document's
            // dirty state has changed
            fireSelectionChange( doc );
        }
    } // markDocumentDirty

    /**
     * Creates a new document object, as well as the needed UI and
     * adds it to the group. This document will become the current
     * document. The name of the document will be of the form
     * "Untitled #" where # is a number from 1 to 32767.
     *
     * @param  view  view to associate with this document
     * @return  true if successful, false otherwise
     */

    public boolean newDocument( ConceptView view ) {
        if (km ==null)
            return false;
        // generate a unique "untitled" title
        File f = null;
        int n = 1;
        boolean exists;
        Context[] contexts = km.getContexts();
        /*
        do {
            f = new File( ".", "Untitled " + n );
            exists = f.exists();
            int i = 0;
            while ( ( !exists ) && ( i < documents.size() ) ) {
                exists = getDocument( i ).getTitle().equals( f.getName() );
                i++;
            }
            while ( ( !exists ) && ( i < contexts.length ) ) {
                exists = ((CreekDocument)contexts[i]).getTitle().equals( f.getName() );
                i++;
            }
            n++;
        } while ( exists );
        f.deleteOnExit();
        // create standard document object ^*/

        Class docType = (Class)docTypes.get( defaultExtensionDoc );
        CreekDocument doc = null;
        try {
            doc = (CreekDocument)docType.newInstance();
        } catch ( Exception e ) { e.printStackTrace(); }
//        doc.setFile( f );
        doc.setTitled(true);
        doc.setTitle("View "+contexts.length);
        doc.setView( view );
        CreekDocument oldDoc = getCurrentDocument();
        addDocument( doc );
        km.addContext(doc);
        doc.open();
        fireDirtyChange( doc );
        fireDocChange( oldDoc, doc );
        return true;
    } // newDocument

    public void newKM(KnowledgeModel km)
    {
        this.km = km;
//        km.addModelChangeListener(this);  //0312 change
//        return saveAsKM(file,dir);
        fireModelOpened(km);
//        return true;
    }

    /**
     * Add the given Document as a Creek document, creating the
     * needed UI and the document object. The new document will
     * be added to the group and will become the current document.
     *
     * @param  view  view to associate with this document
     * @param  doc     document to be opened as a Creek document
     * @return  true if successful, false if unknown file format,
     *          or corrupt file structure
     */
    public boolean openDocument( ConceptView view, CreekDocument doc ) {
        if(doc==null)
          return false;
        // add document to group
        CreekDocument oldDoc = getCurrentDocument();
        if(!isActive(doc))
        {
            doc.setView(view);
            addDocument( doc );
            doc.getModel().unselectAllVertices();
            doc.getModel().markDirty();
        }
        else
            return false;
        fireDirtyChange( doc );
        fireDocChange( oldDoc, doc );
        return true;
    } // openDocument

    public void setKM(KnowledgeModel km)
    {
      this.km = km;
      fireModelOpened(km);
    }


    public boolean isActive(CreekDocument doc)
    {
        return documents.contains(doc);
    }


    /**
     * Opens a Test KnowledgeModel as Creek document, creating the
     * needed UI and the document object. The new document will
     * be added to the group and will become the current document.
     *
     * @param  file  file name
     * @param  dir   directory path
     */
    public void openTestKM()
    {
        km = new LocalKnowledgeModel();
        Vector vector = new Vector();
        km = jcreek.examples.FruitModel.generateModel();

          /*
          Case case12 = (Case) km.getEntity("Case#12");
          javax.swing.JFrame frame = new javax.swing.JFrame(case12.getName());
          jcreek.cke.gui.CasePane cp = new jcreek.cke.gui.CasePane(case12);
          cp.setEditable(true);
          cp.setActive(true);
          frame.setContentPane(new javax.swing.JScrollPane(cp));
          frame.setSize(400,600);
          frame.show();
          */
      fireModelOpened(km);
    }



    public void openKM(File f ) throws IOException, ClassNotFoundException
    {
     /* try
      {
        km = new KnowledgeModel(f);
      }
      catch(IOException e)
      {
        e.printStackTrace();
        return false;
      }
      catch(ClassNotFoundException e)
      {
        e.printStackTrace();
        return false;
      } */
      km = new LocalKnowledgeModel(f);

      Context[] contexts = km.getContexts(CreekDocument.class);
      System.out.println("CreekDocuments: "+contexts.length+" All documents: "+km.getContexts().length);
      for(int i=0;i<contexts.length;i++)
        ((CreekDocument) contexts[i]).open();

      fireModelOpened(km);

//      return true;
    }

/*
        ObjectInputStream input = null;
        FileInputStream file = null;

        // determine file format from filename
        String fname = f.getName();
        fname = fname.toLowerCase();
        KnowledgeModel km = null;
        int dotIdx = fname.lastIndexOf( '.' );
        Class kmType = (Class)kmTypes.get( fname.substring( dotIdx + 1 ) );
        try
        {
            file = new FileInputStream(f);
            input = new ObjectInputStream(file);
        }
        catch(IOException ie)
        {
            ie.printStackTrace();
        }
        if ( kmType != null ) {
            // Found document type, instantiate that document.
            try {
                km = new KnowledgeModel(input);
            } catch ( Exception e )
            {
              e.printStackTrace();
              return false;
            }
            km.setFile( f );
            km.setTitled( true );
        } else {
            // No matching type, instantiate default document.
            kmType = (Class)kmTypes.get( defaultExtensionKM );
            try
            {
                km = new KnowledgeModel(input);
            } catch ( Exception e )
            {
              e.printStackTrace();
              return false;
            }
            km.setFile( f );
            km.setTitled( true );
        }
*/

        // add KM to CreekDocumentGroup
//        this.km = km;
//        this.km.addModelChangeListener(this); //0312 change

/*
        // cause document to read in its data for context
        Object content =  null;
        Object savedTitle= null;
        int cnt=0;
        try
        {
            cnt = input.readInt();
        }
        catch(IOException e)
        {
            e.printStackTrace();
        }
        for(int j = 0;j<cnt;j++)
        {
            try
            {
                savedTitle = input.readObject();
                content    = input.readObject();

            }
            catch(Exception e)
            {
                e.printStackTrace();
            }
            openContext(content,savedTitle);
        }


//        fireKMChanged( null,km ); //0312 change




        return true;
    } */
      /*
    public void openContext(Context c)
    {

        CreekDocument doc=null;
        File f = null;
        int n = 1;
        boolean exists;
        do {
            f = new File( ".", "Untitled " + n );
            exists = f.exists();
            int i = 0;
            while ( ( !exists ) && ( i < documents.size() ) ) {
                exists = getDocument( i ).getTitle().equals( f.getName() );
                i++;
            }
            n++;
        } while ( exists );
        f.deleteOnExit();
       // No matching type, instantiate default document.
        Class docType = (Class)docTypes.get( defaultExtensionDoc );
        try {
            doc = (CreekDocument)docType.newInstance();
        } catch ( Exception e ) {e.printStackTrace(); }
//        doc.setFile( f );
        doc.setTitled( true );
        // cause document to read in its data
        if ( !doc.revert((Graph)content,(String)savedTitle) ) {
            // clean up upon error
            closeDocument();
            return false;
        }

      if(c instanceof CreekDocument)
      {
        ((CreekDocument) c).open();
      }
    }*/

    /**
     * Registers a filename extension with a given document
     * class. This tells the document group to use that class
     * when opening files with the given extension.
     * <p>
     * The first document type registered will be treated as
     * the default type, used when creating new documents.
     *
     * @param  ext  filename extension
     * @param  doc  document class
     */
    public void registerDocument( String ext, Class doc ) {
        if ( docTypes.size() == 0 ) {
            defaultExtensionDoc = ext;
        }
        docTypes.put( ext, doc );
    } // registerDocument

    /**
     * Registers a filename extension with a given document
     * class. This tells the document group to use that class
     * when opening files with the given extension.
     * <p>
     * The first document type registered will be treated as
     * the default type, used when creating new documents.
     *
     * @param  ext  filename extension
     * @param  doc  document class
     */
    public void registerkm( String ext, Class doc ) {
        if ( kmTypes.size() == 0 ) {
            defaultExtensionKM = ext;
        }
        kmTypes.put( ext, doc );
    } // registerDocument

    /**
     * Remove the Creek document from the document group. Updates
     * the current document index data member to -1 so there is no
     * current document.
     *
     * @param  doc  document to be removed
     */
    protected synchronized void removeDocument( CreekDocument doc ) {
        if ( doc == null ) {
            return;
        }
//        if(!doc.isTitled())
//        km.removeContext(doc);
        doc.removeUndoableEditListener( undoHandler );
        viewDocs.remove( doc.getView() );
        documents.removeElement( doc );
        // We expect the MainWindow class to eventually set
        // the current document to something. Meanwhile we
        // will have a null current document.
        fireDocRemoved(doc);
        curDocument = -1;
    } // removeDocument

    /**
     * Remove a property change listener from this document group.
     *
     * @param  l  listener to remove from notification list
     */
    public void removePropertyChangeListener( PropertyChangeListener l ) {
        propSupport.removePropertyChangeListener( l );
    } // removePropertyChangeListener

    /**
     * Remove a property change listener from this document group.
     *
     * @param  name  property name to listen for specifically
     * @param  l     listener to remove from notification list
     */
    public void removePropertyChangeListener( String name,
                                              PropertyChangeListener l ) {
        propSupport.removePropertyChangeListener( name, l );
    } // removePropertyChangeListener

    /**
     * Reverts the current document from disk. If there is no current
     * document then this returns false.
     *
     * @return  return value from CreekDocument.revert()
     */
    public boolean revertDocument() {
        boolean okay = false;
        CreekDocument doc = getCurrentDocument();
        if ( doc != null ) {
            okay = doc.revert();
            if ( okay ) {
                fireDirtyChange( doc );
            }
        }
        return okay;
    } // revertDocument

/*
    public boolean saveAsDocument( String file, String dir ) {
        CreekDocument oldDoc = getCurrentDocument();
        if ( oldDoc == null ) {
            // we were called by mistake
            return true;
        }
        CreekDocument newDoc = null;
        File newFile = new File( dir, file );
        String fname = file.toLowerCase();

        // instantiate new document object for given filename
        // NOTE: we have to use the existing document's view
        // because it is attached to the internal window and
        // it's too difficult to change it without closing
        // that window and creating a new one
        int dotIdx = fname.lastIndexOf( '.' );
        Class docType = (Class)docTypes.get( fname.substring( dotIdx + 1 ) );
        if ( docType != null ) {
            // found matching type, instantiate new document
            try {
                newDoc = (CreekDocument)docType.newInstance();
            } catch ( Exception e ) {
                return false;
            }
            newDoc.setFile( newFile );
            newDoc.setTitled( true );
            newDoc.setView( oldDoc.getView() );
        } else {
            // user didn't provide a known extension, use default
            newFile = new File( dir, file + "." + defaultExtensionDoc );
            docType = (Class)docTypes.get( defaultExtensionDoc );
            try {
                newDoc = (CreekDocument)docType.newInstance();
            } catch ( Exception e ) {
                return false;
            }
            newDoc.setFile( newFile );
            newDoc.setTitled( true );
            newDoc.setView( oldDoc.getView() );
        }

        // copy existing data over to new model
        CreekModel oldData = oldDoc.getModel();
        CreekModel newData = newDoc.getModel();
        newData.replace( oldData.getGraph() );
        // we have to reattach all the edits to the new document
        undoHandler.reattachEdits( oldDoc, newDoc );
        // remove old document and add new one (defaults to current)
        removeDocument( oldDoc );
        addDocument( newDoc );

        //remove old context and add new one
        km.removeContext(oldDoc);
        km.addContext(newDoc);

        // let everyone know the documents have changed
        // NOTE: this has to come here so those who talk to
        // documents can get the correct state information
        fireDocChange( oldDoc, newDoc );
        // now it's safe to close the old document
        oldDoc.close();
        // HACK: we have to reattach the view since closing
        // the old document detached the view
        ConceptView newView = newDoc.getView();
        newView.attach();
        newView.dataChanged( newData.getEntities(),newData.getRelations() );

        // save the new current document
        boolean okay = newDoc.saveAs( newFile );
        if ( okay ) {
            fireDirtyChange( newDoc );
        }
        return okay;
    } // saveAsDocument
*/
    /**
     * Saves the current document to disk. If there is no current
     * document then this returns true.
     *
     * @return  return value from CreekDocument.save()
     */
/*    public boolean saveDocument() {
        boolean okay = true;
        CreekDocument doc = getCurrentDocument();
        if ( doc != null ) {
            okay = doc.save();
            if ( okay ) {
                fireDirtyChange( doc );
            }
        }
        return okay;
    } // saveDocument
*/


    /**
     * Saves the KnowledgeModel under a new name and possibly
     * a new file format. The file format is determined by the
     * file extension. If the new file extension is not
     * recognized, we'll just write out in the current format
     * to the given filename.
     *
     * @param  file  file name
     * @param  dir   directory path
     * @return  true if successful, false if error
     */
    public void saveAsKM( String file, String dir ) throws IOException
    {
        if ( km == null ) {
            // we were called by mistake
            return;
        }
        File newFile = new File( dir, file );
        String fname = file.toLowerCase();

        makeSubmodelFromView();
        ((LocalKnowledgeModel) km).saveAs(newFile);

/*        try
        {
          km.saveAs(newFile);
        }
        catch(IOException e)
        {
          e.printStackTrace();
          return false;
        }
        return true; */
    }
/*
        int dotIdx = fname.lastIndexOf( '.' );
        Class kmType = (Class)kmTypes.get( fname.substring( dotIdx + 1 ) );
        if ( kmType != null ) {
            km.setFile( newFile );
            km.setTitled( true );
        } else {
            // user didn't provide a known extension, use default
            newFile = new File( dir, file + "." + defaultExtensionKM );
            kmType = (Class)kmTypes.get( defaultExtensionKM );
            km.setFile( newFile );
            km.setTitled( true );
        }
        // save the new KnowledgeModel
        km.saveAs( newFile );
        if ( okay ) {
//            fireKMChanged( null,km ); //0312 change
            fireModelOpened(km);
        }
        return okay;

    } // saveAsDocument */

    /**
     * Saves the current KnowledgeModel to disk.
     *
     * @return  return value from KnowledgeModel.save()
     */
    public void saveKM() throws IOException
    {
        if ( km != null )
        {
          makeSubmodelFromView();
          ((LocalKnowledgeModel) km).save();
/*          try
          {

          }
          catch(IOException e)
          {
            e.printStackTrace();
            return false;
          }
        }
        return true; */
        }
    } // saveDocument

    // Temporary and transitional to make views into submodels in the km.
    private void makeSubmodelFromView()
    {
      CreekDocument[] docs = getDocuments();
//      System.out.println("[makeSubmodelFromView] "+docs.length+" documents.");
      for(int i=0;i<docs.length;i++)
      {
        String name = docs[i].getTitle();
//        System.out.println("[makeSubmodelFromView] Name: "+name);
        try
        {
          if(getKM().hasPartition(name))
          {
            getKM().getPartition(name).remove();
          }
          Partition partition = new Partition(getKM(), name);
          VertexList vl = docs[i].getModel().getEntities();
          for(int j=0;j<vl.size();j++)
          {
            VisualEntity ve = vl.get(j);
            partition.add(ve.getEntity());
          }
          RelationList rl = docs[i].getModel().getRelations();
          for(int j=0;j<rl.size();j++)
          {
            partition.add(rl.get(j).getRelation());
          }
        }
        catch(jcreek.util.CreekException e)
        {
          e.printStackTrace();
        }
      }
    }

    /**
     * Sets the given document to be the current one. Sends out
     * notification to any listeners that current document has
     * been switched.
     *
     * @param  newDoc  document to make current
     */
    public void setCurrentDocument( CreekDocument newDoc ) {
        CreekDocument oldDoc = getCurrentDocument();
        // if newDoc is null current document will be null
        curDocument = -1;
        if ( newDoc != null ) {
            int size = documents.size();
            for ( int i = 0; i < size; i++ ) {
                CreekDocument doc = getDocument( i );
                if ( doc == newDoc ) {
                    curDocument = i;
                    fireDocChange( oldDoc, newDoc );
                    break;
                }
            }
        }
    } // setCurrentDocument



    public boolean isDirty()
    {
      return true;
//        return km != null && ((LocalKnowledgeModel) km).isDirty();
    }//isDirty


  protected PageFormat getDefaultPageFormat ()
  {
    PrinterJob pj = PrinterJob.getPrinterJob ();
    PageFormat pf = pj.defaultPage ();
    pf = pj.validatePage (pf);
    return pf;
  }

  public PageFormat getPageFormat()
  {
      if(pf==null)
          pf = getDefaultPageFormat();
      return pf;
  }

  public void setPageFormat(PageFormat pf)
  {
      this.pf = pf;
  }

  protected boolean kmEditable = false;

  public void setKnowledgeModelEditable(boolean editable)
  {
    fireKMEditChange(editable);
    kmEditable = editable;
  }

  public boolean getKnowledgeModelEditable()
  {
    return kmEditable;
  }

} // CreekDocumentGroup


