package jcreek.cke;

import jcreek.cke.browser.Visualizer;

import java.net.URL;
import java.util.Hashtable;
import java.util.Locale;
import java.util.MissingResourceException;
import java.lang.NumberFormatException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.swing.Action;
import javax.swing.UIManager;

/**
 * This class implements the application manager. It is the one class
 * that organizes the rest of the classes in the application. It holds
 * the localized resource bundle and available actions for the program.
 * <p>
 * This class is a Singleton which means only one instance of this class
 * exists. You use the instanceOf() method to access that instance.
 */

public class MainWindow
{
    /**
     * Reference to the single instance of this class.
     */
    protected static MainWindow instance;
    /**
     * The list of main windows in this application.
     */
    protected Vector windows;
    /**
     * String and object resources for our program.
     * Used to get localized strings and objects (such as
     * images) for user interface components.
     */
    protected ResourceBundle resources;

    /**
     * Table of available actions in our program. Other classes
     * can look up actions by their action string using the
     * <code>getAction()</code> method.
     */
    protected Hashtable actions;


    /**
     * Suffix added to command string to retrieve action classes.
     */
    protected static String actionSuffix = "Action";

    /**
     * No-arg constructor for Visualizer class. It is protected since this
     * class is a Singleton. To get the single instance of this class
     * use the instanceOf() method.
     */

    protected MainWindow()
    {
        try
        {
            resources = ResourceBundle.getBundle
                ( "jcreek.cke.resources.Creek",
                  Locale.getDefault() );
        }
        catch ( MissingResourceException mre )
        {
            System.err.println( "resources/Creek.properties not found" );
            System.exit( 1 );
        }
        actions = new Hashtable();
        windows = new Vector();
    }

    /**
     * Retrieves a string from the localized resource bundle.
     * The resources are located in the appropriate
     * resources/JavaCreek.properties file.
     *
     * @param  name  key name of the string resource to find
     * @return  string resource or null if not found
     * @see #getResource
     */
    public String getResourceString( String name )
    {
        String str = null;
        try
        {
            str = resources.getString( name );
        }
        catch ( MissingResourceException mre )
        {
        }
        return str;
    } // getResourceString

    /**
     * Retrieves an object from the localized resource bundle.
     * In most cases this is an image.
     * The resources are located in the appropriate
     * resources/JavaCreek.properties file.
     *
     * @param  key  key name of the resource to find
     * @return  URL pointing to the object resource
     * @see #getResourceString
     */
    public URL getResource( String key )
    {
        String name = getResourceString( key );
        if ( name != null )
        {
            URL url = this.getClass().getResource( name );
            return url;
        }
        return null;
    } // getResource

    /**
     * Returns the reference to the single instance of this class.
     * If an instance does not exist it will be created.
     *
     * @return  the instance of this class
     */
    public static MainWindow instanceOf()
    {
        if ( instance == null )
        {
            instance = new MainWindow();
        }
        return instance;
    } // instanceOf

    /**
     * Creates, shows, and returns a new main window. This is
     * added to the list of main windows stored in this object.
     * To remove a window when finished, call
     * <code>removeWindow()</code>.
     *
     * @return  new main window
     */

    public Visualizer newWindow()
    {
        Visualizer win = new Visualizer
            ( getResourceString( "AppTitle" ) );
        windows.addElement( win );
        win.show();
        return win;
    } // newWindow

    /**
     * Closes and removes a main window. In the event that there
     * are no more main windows, the application will exit the JVM.
     *
     * @param  win  main window to close and remove
     */

    public void removeWindow( Visualizer win ) {
        win.setVisible( false );
        win.detach();
        win.dispose();
        windows.remove( win );
        if ( windows.size() == 0 ) {
            // Save user preferences.
//            AppSettings.instanceOf().commit();  got to find out about this
            System.exit( 0 );
        }
    } // removeWindow

    /**
     * Take the given string and chop it up into a series
     * of strings on whitespace boundries. This is useful
     * for trying to get an array of strings out of the
     * resource file. If input is null, returns a zero-length
     * array of String.
     *
     * @param  input  string to be split apart
     * @return  array of strings from input
     */
    public static String[] tokenize( String input )
    {
        if ( input == null )
        {
            return new String[0];
        }
        StringTokenizer t = new StringTokenizer( input );
        int size = t.countTokens();
        String cmd[] = new String[size];
        for ( int i = 0; i < size; i++ )
        {
            cmd[i] = t.nextToken();
        }
        return cmd;
    } // tokenize

    /**
     * Stores the action in the command table. It can later
     * be retrieved using <code>getAction()</code>.
     *
     * @param  cmd  action to store in command table
     * @see #getAction
     */
    public void addAction( Action cmd )
    {
        actions.put( cmd.getValue( Action.NAME ), cmd );
    } // getAction

    /**
     * Retrieves the action corresponding to the given command string.
     * This will first look in the list of actions registered with
     * Visualizer. If the action is not there, it will try to
     * instantiate the action. If successful, it adds the action
     * to the list of registered actions and returns a reference
     * to the action. Else it does nothing and returns null.
     *
     * @param  cmd  command string to find action for
     * @return action matching command string or null if not found
     * @see #addAction
     */
    public Action getAction( String cmd )
    {
        Action a = (Action)actions.get( cmd );
        if ( a == null )
        {
            String actionName = getResourceString( cmd + actionSuffix );
            if ( actionName != null )
            {
                try
                {
                    Class actionClass = Class.forName
                        ( "jcreek.cke.command." + actionName );
                    a = (Action)actionClass.newInstance();
                    addAction( a );
                }
                catch ( Exception e )
                {
                    System.err.println( "ERROR: Couldn't instantiate action: "
                                        + actionName + ": " + e );
                }
            }
        }
        return a;
    } // getAction


}
