package jcreek.cke.browser;

import jcreek.cke.document.CreekDocument;
import jcreek.cke.command.CreekCommand;
import jcreek.cke.command.UndoAction;
import jcreek.cke.command.RedoAction;
import jcreek.cke.MainWindow;

import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.*;


/**
 * This class handles all things undo-related in Creek. It
 * creates and manages the undo manager object. It also has
 * the undo and redo actions and registers them with the
 * MainWindow class.
 *
 * @author
 * @version
 */
public class UndoHandler implements UndoableEditListener {
    /**
     * Global undo manager for performing undo/redo events.
     */
    protected CreekUndoManager undo;
    /**
     * Our global undo action.
     */
    protected UndoAction undoAction;
    /**
     * Our global redo action.
     */
    protected RedoAction redoAction;

    /**
     * Creates a new UndoHandler object, to handle undoable
     * edit events. Also creates the undo/redo actions and
     * sets up the undo manager.
     */
    public UndoHandler() {
        undo = new CreekUndoManager();
        MainWindow mw = MainWindow.instanceOf();
        undoAction = (UndoAction)mw.getAction( "undo" );
        if ( undoAction == null ) {
            undoAction = new UndoAction( undo );
            mw.addAction( undoAction );
        } else {
            undoAction.setManager( undo );
        }
        redoAction = (RedoAction)mw.getAction( "redo" );
        if ( redoAction == null ) {
            redoAction = new RedoAction( undo );
            mw.addAction( redoAction );
        } else {
            redoAction.setManager( undo );
        }
    } // UndoHandler

    /**
     * Tell the edits associated with the old document that they
     * should reattach themselves to the new document.
     *
     * @param  oldDoc  previous document
     * @param  newDoc  new document
     */
    public void reattachEdits( CreekDocument oldDoc,
                               CreekDocument newDoc ) {
        undo.reattachEdits( oldDoc, newDoc );
    } // reattachEdits

    /**
     * Trim the edits associated with the given document.
     *
     * @param  doc  document being closed
     */
    public void removeEdits( CreekDocument doc ) {
        undo.trimEdits( doc );
    } // removeEdits

    /**
     * Called when a document has performed an edit. The edit
     * is added to undo manager and the undo/redo actions are
     * updated.
     *
     * @param  e  undoable edit event
     */
    public void undoableEditHappened( UndoableEditEvent e ) {
        undo.addEdit( e.getEdit() );
        update();
    } // undoableEditHappened

    /**
     * Causes the undo and redo actions to update themselves.
     */
    public void update() {
        undoAction.update();
        redoAction.update();
    } // update
} // UndoHandler
