/*********************************************************************
 *
 * DESCRIPTION:
 *      This file defines the abstract class used to define actions
 *      that perform various program operations.
 *
 ********************************************************************/

package jcreek.cke.command;

import jcreek.cke.MainWindow;
import jcreek.cke.browser.Visualizer;
import jcreek.cke.document.CreekDocument;
import jcreek.cke.graph.CreekModel;
import jcreek.cke.conceptview.ConceptView;
import jcreek.cke.CreekDocumentGroup;

import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.JOptionPane;

/**
 * Base action class which all other actions subclass.
 * Provides some utility functions needed by nearly all
 * of the actions in Creek.
 *
 * @author
 * @version
 */
public abstract class CreekAction extends AbstractAction {

    /**
     * Creates a new CreekAction command with the given
     * action command string.
     *
     * @param  name  action command string
     */
    public CreekAction( String name ) {
        super( name );
    } // CreekAction

    /**
     * Retrieves a string resource from the app's resource
     * bundle.
     *
     * @param  key  key of resource to retrieve
     * @return  string from resource bundle
     */
    public String getString( String key ) {
	MainWindow mw = MainWindow.instanceOf();
	return mw.getResourceString( key );
    } // getString

    /**
     * Finds the main window that contains the component
     * that is the source of the given action event.
     *
     * @param  e  action event to find window
     * @return  main window
     */
    public Visualizer getWindow( ActionEvent e ) {
        // get graph window that contains our invoker
        return Visualizer.getFrame( (Component)e.getSource() );
    } // getWindow

    /**
     * Displays a dialog indicating that there's no current
     * document to perform the desired operation.
     *
     * @param  win  frame to attach dialog to
     */
    public void noCurrentDocument( Frame win ) {
	JOptionPane.showMessageDialog
	    ( win,
	      getString( "noCurDocMsg" ),
	      getString( "usrErrorTitle" ),
	      JOptionPane.ERROR_MESSAGE );
    } // noCurrentDocument

    /**
     * Returns the current creek document, if any. If there's
     * no current document it returns null.
     *
     * @return  current creek document
     */
    public CreekDocument getDocument() {
        return CreekDocumentGroup.instanceOf().getCurrentDocument();
    } // getDocument

    /**
     * Returns the current creek model, if any. If there's
     * no current document it returns null.
     *
     * @return  current creek model
     */
    public CreekModel getModel() {
        CreekDocument doc = getDocument();
        if ( doc == null ) {
            return null;
        }
        return doc.getModel();
    } // getModel

    /**
     * Returns the current concept view, if any. If there's
     * no current document it returns null.
     *
     * @return  current concept view
     */
    public ConceptView getView() {
        CreekDocument doc = getDocument();
        if ( doc == null ) {
            return null;
        }
        return doc.getView();
    } // getView


} // CreekAction
