/*********************************************************************
 *
 * DESCRIPTION:
 *      This file defines the command class extended by all the
 *      command classes in Visualizer.
 *
 ********************************************************************/

package jcreek.cke.command;

import jcreek.cke.document.CreekDocument;
import javax.swing.undo.UndoableEdit;

/**
 * Defines the abstract class to be extended by all command classes
 * in Visualizer. Provides a way of executing commands and being
 * able to undo/redo them later. This class defines an additional
 * method for performing the command for the first time.
 *
 * @author
 * @version
 */
public abstract class CreekCommand implements UndoableEdit {
    /**
     * The document this command is associated with.
     */
    protected CreekDocument document;
    /**
     * Indicates if this command and be undone/redone.
     */
    protected boolean canUndoRedo;

    /**
     * Creates a new CreekCommand object associated with the
     * given document object.
     *
     * @param  doc  document we are performed on
     */
    public CreekCommand(CreekDocument doc ) {
    	this.document = doc;
    	canUndoRedo = true;
    } // CreekCommand

    /**
     * True if it is still possible to redo this operation.
     *
     * @return  true if redo-able
     */
    public boolean canRedo() {
        return canUndoRedo;
    } // canRedo

    /**
     * True if it is still possible to undo this operation.
     *
     * @return  true if undo-able
     */
    public boolean canUndo() {
        return canUndoRedo;
    } // canUndo

    /**
     * Returns the document this command was performed on.
     *
     * @return  document this command is associated with
     */
    public CreekDocument getDocument() {
	return document;
    } // getDocument

    /**
     * Do the command for the first time. The arguments
     * are given in the constructor for the particular
     * implementation of the command.
     *
     * @return  true if command executed, false if error
     */
    public abstract boolean perform();

    /**
     * Our document has been changed on us. We must save this
     * new document object reference.
     */
    public void setDocument( CreekDocument doc ) {
        this.document = doc;
    } // setDocument

    /**
     * Indicates this object has been replaced by another.
     * From now on this object can no longer undo/redo and
     * should throw exceptions if asked to do so.
     */
    public void replaced() {
        canUndoRedo = false;
    } // replaced
} //CreekCommand
