package jcreek.cke.command;

import javax.swing.undo.UndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import jcreek.cke.graph.VertexList;
import jcreek.cke.document.CreekDocument;
import jcreek.cke.graph.CreekModel;
import jcreek.cke.conceptview.VisualEntity;

/**
 * Title:        JavaCreek
 * Description:  A Case Based Reasoning approach developed in Java
 * Copyright:    Copyright (c) 1999
 * Company:      Phil Systems
 * @author Victor Dias
 * @version
 */

public class EntityMoveCommand extends CreekCommand
{

    /**
     * List of entities to be moved by the delta values.
     */
    protected VertexList entities;
    /**
     * Delta X move of the entity list.
     */
    protected int deltaX;
    /**
     * Delta Y move of the entity list.
     */
    protected int deltaY;


    /**
     * Constructor for this command. Saves the entitylist and document
     * for use with do, undo, and redo operations.
     *
     * @param  doc  document we're performed on
     * @param  nl   entity list to be moved
     * @param  dx   delta x move for entity list
     * @param  dy   delta y move for entity list
     * @param  dz   delta z move for entity list
     */

  public EntityMoveCommand(CreekDocument doc,
			      VertexList nl,
			      int dx, int dy)
  {
	super( doc );
        canUndoRedo = true;
	entities = (VertexList)nl.clone();
	deltaX = dx;
	deltaY = dy;

  }


    /**
     * Return the human-readable name for undoing this command.
     *
     * @return  human-readable name for undoable form of this command
     */
  public String getUndoPresentationName()
  {
        if ( entities.size() > 1 ) {
            return "Undo Move Entities";
        } else {
            return "Undo Move Entity";
        }
  }

    /**
     * Perform the action for the first time.
     *
     * @return  true if command executed, false if error
     */
  public boolean perform()
  {
        if ( ( getDocument() != null ) && ( entities != null ) ) {
	    // Move enttities to new position.
            CreekModel model = getDocument().getModel();
            return model.moveEntityList( entities, deltaX, deltaY);
        }
        return false;
  }


    /**
     * May be sent to inform an edit that it should no longer be used.
     */
  public void die()
  {
        canUndoRedo = false;
        entities.clear();
        entities = null;
  }

    /**
     * Return true if this command is not a side-effect of some
     * other change. Moving a entity is always significant.
     *
     * @return  true if change is a side-effect of other changes
     */
  public boolean isSignificant()
  {
        return true;
  }

    /**
     * Moves the vertex back into the model.
     */
  public void redo()
  {
        if ( ( canUndoRedo ) && ( entities != null ) ) {
            CreekModel model = getDocument().getModel();
            model.moveEntityList( entities, deltaX, deltaY);
        } else {
            throw new CannotRedoException();
        }
  }


    /**
     * Moves the moved entity back to where it was located.
     */
  public void undo()
  {
        if ( ( canUndoRedo ) && ( entities != null ) ) {
            CreekModel model = getDocument().getModel();
            model.moveEntityList( entities, -deltaX, -deltaY);
        } else {
            throw new CannotUndoException();
        }
  }

    /**
     * Try to merge anEdit with this edit.
     *
     * @param  anEdit  edit to absorb
     * @return  true if anEdit was absorbed
     */

  public boolean addEdit(UndoableEdit parm1)
  {
	// For now we don't support merging.
        return false;
  }


    /**
     * Return the human-readable name for redoing this command.
     *
     * @return  human-readable name for redoable form of this command
     */
  public String getRedoPresentationName()
  {
        if ( entities.size() > 1 ) {
            return "Redo Move Entities";
        } else {
            return "Redo Move Entity";
        }
  }


    /**
     * See if we can be the given edit, thus replacing it in the
     * undo queue.
     *
     * @param  anEdit  edit command to replace
     * @return  true if anEdit replaced
     */
  public boolean replaceEdit(UndoableEdit parm1)
  {
        // Assume that we can't for now.
        return false;
  }

    /**
     * Return the human-readable name for this action.
     *
     * @return  human-readable name for this command
     */

  public String getPresentationName()
  {
        if ( entities.size() > 1 ) {
            return "Move Entities";
        } else {
            return "Move Entity";
        }
  }
}