/*********************************************************************
 *
 * DESCRIPTION:
 *      This file defines the file open action. It also defines the
 *      filter class used to select only supported graph files.
 *
 ********************************************************************/

package jcreek.cke.command;

import jcreek.cke.browser.Visualizer;
import jcreek.cke.MainWindow;
import jcreek.cke.CreekDocumentGroup;
import jcreek.cke.conceptview.ConceptView;
import jcreek.cke.visualizer.AppSettings;

import java.awt.event.ActionEvent;
import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;
import javax.swing.DefaultBoundedRangeModel;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.awt.Cursor;
import java.awt.Component;

/**
 * Implements the open file action used to open an existing
 * creek document. This class also defines a file filter used
 * with the file selector to limit the user to choosing only
 * supported graph files.
 *
 * @author
 * @version
 */
public class OpenKnowledgeModelAction extends CreekAction implements PropertyChangeListener{

    /**
     * Creates a new OpenAction object with the default action
     * command string of "open".
     */
    public OpenKnowledgeModelAction() {
        super( "openKnowledgeModel" );

        // we start life disabled, until a document is opened
        // and made dirty (which we'll know about by being a
        // change listener for the document group
        setEnabled( true );
        // we want to be notified of changes in the documents
        CreekDocumentGroup.instanceOf().addPropertyChangeListener( this );
    } // OpenAction

    /**
     * Performs the open action. This presents the user with
     * a file selector and opens the selected file as a graph
     * document. Causes a new window to be opened.
     *
     * @param  event  action event
     */
    public void actionPerformed( ActionEvent event ) {
        // get graph window that contains our invoker
        Visualizer win = getWindow( event );//rem03/02
        CreekDocumentGroup dg = CreekDocumentGroup.instanceOf();

        JFileChooser fc = new JFileChooser(AppSettings.instanceOf().getCurrentDirectory());
        fc.setFileFilter( new CreekFileFilter( dg.getKMTypes() ) );
        if ( fc.showOpenDialog( win ) == JFileChooser.CANCEL_OPTION ) {
            // user hit cancel
            return;
        }
        File f = fc.getSelectedFile();

        AppSettings.instanceOf().setCurrentDirectory(f.getAbsolutePath());
        //User hit escape key
        if(f==null)
          return;
//        try {
//            // remember the new current directory for next time
//            System.setProperty( "user.dir",
//                                fc.getCurrentDirectory().toString() );
//        } catch ( SecurityException se ) { }
        // see if we can read from the file
        try {
            if ( !f.canRead() ) {
                JOptionPane.showMessageDialog
                    ( win,
		      getString( "cantReadFileMsg" ),
		      getString( "operFailedTitle" ),
                      JOptionPane.ERROR_MESSAGE );
                return;
            }
        } catch ( SecurityException se ) {
            JOptionPane.showMessageDialog
                ( win,
		  getString( "noFileAccessMsg" ),
		  getString( "operFailedTitle" ),
                  JOptionPane.ERROR_MESSAGE );
            return;
        }

        ((Component)win).setCursor(new Cursor(Cursor.WAIT_CURSOR));
        // have document group create a new document for this file
//        ConceptView view = win.newWindow( f.getName() );
        String error = null;
        try
        {
          dg.openKM(f);
        }
        catch(IOException e)
        {
          error = e.toString();
        }
        catch(ClassNotFoundException e)
        {
          error = e.toString();
        }
        if (error != null ) {
          ((Component)win).setCursor(Cursor.getDefaultCursor());
            // failed, close window we just created
            win.closeWindow();
            // chances are we couldn't understand the file format
            JOptionPane.showMessageDialog
                ( win,
		  getString( "couldntOpenFileMsg" )+" "+error,
		  getString( "operFailedTitle" ),
                  JOptionPane.ERROR_MESSAGE );
        }
        else
          win.setTitle(MainWindow.instanceOf().getResourceString("AppTitle")
                      + " " +dg.getKM().getTitle());
        ((Component)win).setCursor(Cursor.getDefaultCursor());

    } // actionPerformed

  public void propertyChange( PropertyChangeEvent e ) {
        if ( e.getPropertyName().equals( "opened" ) )
        {
            if(e.getNewValue()==null)
                  setEnabled(true);
            else
                  setEnabled(false);
        }
        else if(e.getPropertyName().equals( "closed" ))
            setEnabled(true);
  }

} // OpenAction
