package jcreek.cke.conceptview;

/**
 * <p>Title: ConceptMap</p>
 * <p>Description: Creates a "drawing area" for concept map like representations.
 * This class servers as a view to a GraphModel model.</p>
 * @author Victor Dias, Frode Srmo
 * @version 1.0
 */

import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.Iterator;
import jcreek.cke.graph.*;
import jcreek.cke.util.*;

public class ConceptMap extends JComponent
{
  protected GraphModel graphModel;
  protected double scale = 1.0;
  protected Rectangle bounds = new Rectangle(0,0,300,300);
  protected BufferedImage backImage = null;

  private static final int EXTRA_SPACE = 50;

  public ConceptMap(GraphModel graph)
  {
    this.graphModel = graph;
    calculateBounds();
  }

  public void calculateBounds()
  {
    for(Iterator it = graphModel.getGraph().getNodeIterator();it.hasNext();)
    {
      Node node = (Node) it.next();
      if(node.getPosition().x-EXTRA_SPACE < bounds.x)
        bounds.x = (int) Math.floor(node.getPosition().x)-EXTRA_SPACE;
      if(node.getPosition().y-EXTRA_SPACE < bounds.y)
        bounds.x = (int) Math.floor(node.getPosition().y)-EXTRA_SPACE;
      if(node.getPosition().x+EXTRA_SPACE > bounds.width+bounds.x)
        bounds.width = (int) Math.floor(node.getPosition().x-bounds.x)+EXTRA_SPACE;
      if(node.getPosition().y+EXTRA_SPACE > bounds.height+bounds.y)
        bounds.height = (int) Math.floor(node.getPosition().y-bounds.y)+EXTRA_SPACE;
    }
    setPreferredSize(new Dimension((int) Math.round((bounds.width-bounds.x)*scale),(int) Math.round((bounds.width-bounds.x)*scale)));
    backImage = null;
  }

  public Rectangle getBoundRectangle()
  {
    return bounds;
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    if(backImage == null)
      backImage = (BufferedImage) createImage(getPreferredSize().width, getPreferredSize().height);

    Rectangle clipArea = getVisibleRect();

//      firePaintBefore( (Graphics2D)g );


    Graphics2D graphics = (Graphics2D) backImage.createGraphics();
//    graphics.setFont(font_);
    graphics.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                               RenderingHints.VALUE_ANTIALIAS_ON );
    graphics.setColor(Color.white);
    graphics.setPaintMode();
    graphics.fillRect(clipArea.x, clipArea.y, clipArea.width, clipArea.height);

    graphics.setColor(Color.black);

    Rectangle virtualClipArea = new Rectangle((int) Math.round(clipArea.x*scale), (int) Math.round(clipArea.y*scale),
        (int) Math.round(clipArea.width*scale), (int) Math.round(clipArea.height*scale));
    drawObjects(graphics, clipArea);

    g.setPaintMode();
    g.drawImage(backImage, 0, 0, null);

//    firePaintAfter( (Graphics2D)g );
    graphics.dispose();
  }

  public synchronized void drawObjects(Graphics2D graphics, Rectangle clipArea)
  {
      VisualEntity v;
      VisualRelation vr;

      Graph graph = graphModel.getGraph();

      // Then draw all of the vertices.
      if ( graph != null )
      {

        for(Iterator it = graph.getNodeIterator(); it.hasNext(); )
        {
          Object node = it.next();
          if(node instanceof VisualEntity)
          {
            v = (VisualEntity) node;
            if(v.getDrawBounds(graphics).intersects(clipArea)) ;
              v.draw(graphics, null, scale);
//              System.out.println("Node "+v.creekLEntity+" pos "+v.getPosition().x+","+v.getPosition().y);
          }
        }

        // Then draw all of the relations.
        for(Iterator it = graph.getEdgeIterator(); it.hasNext(); )
        {
          Object rel = it.next();
          if(rel instanceof VisualRelation)
          {
            vr = (VisualRelation) rel;
            if((vr.getHead().getPosition().x <= clipArea.x+clipArea.width &&
               vr.getHead().getPosition().x >= clipArea.x &&
               vr.getHead().getPosition().y <= clipArea.y+clipArea.height &&
               vr.getHead().getPosition().y >= clipArea.y) ||
               (vr.getTail().getPosition().x <= clipArea.x+clipArea.width &&
               vr.getTail().getPosition().x >= clipArea.x &&
               vr.getTail().getPosition().y <= clipArea.y+clipArea.height &&
               vr.getTail().getPosition().y >= clipArea.y)) ;
              vr.draw(graphics, null, scale);
          }
        }
    }
  }

  public void centerOn(Point p)
  {
    int centerX = (int) Math.round(p.x*scale);
    int centerY = (int) Math.round(p.y*scale);

    Rectangle newViewport = new Rectangle(-getVisibleRect().width/2, centerY-getVisibleRect().height/2,
        centerX+getVisibleRect().width/2, centerY+getVisibleRect().height/2);
    scrollRectToVisible(newViewport);
    repaint();
  }

  public void setScale(double newScale)
  {
    Point center = new Point((int) Math.round((getVisibleRect().x+(getVisibleRect().width/2))/scale),
                             (int) Math.round((getVisibleRect().y+(getVisibleRect().height/2))*scale));

    setPreferredSize(new Dimension((int) Math.round((bounds.width-bounds.x)*scale),(int) Math.round((bounds.width-bounds.x)*scale)));
    this.scale = newScale;
    centerOn(center);
    backImage = null;
    revalidate();
  }
}