/*********************************************************************
 *
 * DESCRIPTION:
 *      This file contains the GraphView interface which defines the
 *      API for view-related activities, such as displaying the graph
 *      data on the screen.
 *
 ********************************************************************/

package jcreek.cke.conceptview;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.awt.print.Printable;

import jcreek.cke.graph.VertexList;
import jcreek.cke.graph.RelationList;
import jcreek.cke.util.Matrix33;
import jcreek.representation.Context;
import jcreek.cke.graph.Graph;
import jcreek.cke.util.DPoint;
import jcreek.cke.graph.layout.Layout;
import jcreek.cke.graph.CreekModel;

/**
 * Defines the ConceptView interface for all view classes to implement.
 * The view will be responsible for displaying the graph to the user,
 * typically extending one of the AWT or JFC classes to do so. This
 * class is not responsible for dealing with user input or managing
 * the graph data. All this class does is display the graph data.
 *
 * @author
 * @version
 */

public interface ConceptView extends Printable
{
    /**
     * Adds a view listener to this object's listener list.
     *
     * @param  l  listener to add to view's list
     */
    public void addViewListener( ConceptViewListener l );

    /**
     * Adds a view paint listener to this object's listener
     * list.
     *
     * @param  l  listener to add to view's list
     */
    public void addViewPaintListener( ConceptViewPaintListener l );

    /**
     * We're being attached to a document for the first time.
     */
    public void attach(CreekModel model);

    /**
     * Updates the display to match the current state of the model.
     *
     * @param  el  list of entities
     */
    public void setGraph(Graph graph);

    /**
     * The view is being shut down. This gives the view a chance
     * to remove itself from any listener lists it may be on.
     */
    public void detach();

    /**
     * Removes a view listener from this object's listener list.
     *
     * @param  l  listener to remove from view's list
     */
    public void removeViewListener( ConceptViewListener l );

    /**
     * Removes a view paint listener from this object's listener
     * list.
     *
     * @param  l  listener to remove from view's list
     */
    public void removeViewPaintListener( ConceptViewPaintListener l );

    /**
     * Repaints this component.
     */
    public void repaint();

    public void repaint(int x,int y,int width,int height);

    public Matrix33 getViewTransform();

    public Matrix33 getMoveTransform();

    public double getWinWidth();

    public double getWinHeight();

    public Point getMidPoint();

    public void setScale(double newScale);

    public double getScale();

    public VisualEntity findVisualEntity(double x, double y);

    public void fireContentsResize();

//     public void layout(Layout layout,Graph graph);

    public DPoint activateEntity( VisualEntity activator);

    public Rectangle getBounds();

    abstract public double getHSpacing();
    abstract public double getVSpacing();
    abstract public VisualEntity getRoot();
    abstract public void setRoot(VisualEntity vent);

    public BufferedImage createPrintImage();

    public void drawSelectedNodes();

    public void setOffsets(double xoffset, double yoffset, boolean redraw);

    public void dataChanged(VertexList vl, RelationList rl);


}//ConceptView