/*********************************************************************
 *
 * DESCRIPTION:
 *      This files defines the interface for document classes in the
 *      program. Any class wishing to act as a document must
 *      implement this class.
 *
 ********************************************************************/

package jcreek.cke.document;

import jcreek.cke.command.CreekCommand;
import jcreek.cke.graph.CreekModel;
import jcreek.representation.Context;
import jcreek.cke.graph.Graph;
import jcreek.cke.conceptview.*;

import javax.swing.event.UndoableEditListener;


/**
 * This defines the interface for which all document classes will
 * implement. Classes that implement this interface can act as the
 * document objects within the Creek program, coordinating the
 * model, view, and disk files associated with a graph.
 *
 * @author
 * @version
 */

public interface CreekDocument extends Context
{
    /**
     * Registers an UndoableEditListener with this document.
     * As commands are performed on the document undoable
     * edit events will be sent out to all the listeners.
     *
     * @param  l  listern to add to this document
     * @see #removeUndoableEditListener
     */
    public void addUndoableEditListener( UndoableEditListener l );

    /**
     * The document is opened.
     */
    public void open();

    /**
     * This is provided as a convenience to the document so
     * it may free any resources its been using or remove
     * objects from change lists. It should not try to save
     * the document at this time.
     */
    public void close();


    /**
     * Returns the reference to the object acting as the model
     * for this document. The particular implementation of this
     * model depends on the implemention of CreekDocument.
     *
     * @return  reference to model object
     */
    public CreekModel getModel();

    /**
     * Returns the Title of the document is reading
     * and writing to.
     *
     * @return  name of the document
     */
    public String getTitle();


    /**
     * Returns the reference to the object acting as the view
     * for this document. The particular implementation of this
     * view depends on the implemention of CreekDocument.
     *
     * @return  reference to view object
     */
    public ConceptView getView();

    /**
     * Returns the current dirty status of this document. If the
     * document is dirty and needs to be saved, true will be
     * returned.
     *
     * @return  true if dirty, false if not dirty
     */
    public boolean isDirty();

    /**
     * Returns true if the document is currently named.
     *
     * @return  true if titled, false if unnamed
     */
    public boolean isTitled();

    /**
     * Perform the given command and fire an undoable edit event.
     *
     * @param  command  command to be performed
     * @return  true if command was performed successfully
     */
    public boolean performCommand( CreekCommand command );

    /**
     * Removes an UndoableEditListener from this document.
     *
     * @param  l  listern to remove from this document
     * @see #addUndoableEditListener
     */
    public void removeUndoableEditListener( UndoableEditListener l );

    /**
     * Reverts this document (the graph data) to the last-saved
     * version of the knowledgeModel.
     *
     * @return  true if successful, false if error
     */
    public boolean revert();

    public boolean revert(Graph g,String title);
    /**
     * Saves this document (the graph data) to the associated Context
     * in the KnowledgeModel.The format is determined by the CreekFormat
     * class used by the particular CreekDocument implementation.
     *
     * @return  true if successful, false if error
     */
//    public boolean save();

    /**
     * Saves this document (the graph data) to a new Document, and
     * hereinafter is associated with this new Document.
     *
     * @param  file  new file to save to
     * @return  true if successful, false if error
     */
//    public boolean saveAs( File file );


    /**
     * Tells the document whether it has a title or not.
     *
     * @param  titled  true if document has a titled
     */
    public void setTitled( boolean titled );

    /**
     * Sets the view this document is associated with. Documents
     * should not allow setting the view more than once.
     *
     * @param  view  view object
     */
    public void setView( ConceptView view );

    /**
     * Sets the file this document is associated with. Documents
     * should not allow setting the file more than once.
     *
     * @param  file  file object
     */
//    public void setFile( File file );

    /**
     * Returns the reference to the file object for this document.
     *
     * @return  reference to file object
     */
//    public File getFile();


    public void setTitle(String title);




}//CreekDocument