package jcreek.cke.document;


import jcreek.cke.conceptview.ConceptView;
import jcreek.cke.graph.CreekModel;
import jcreek.cke.graph.GraphModel;
import jcreek.cke.command.CreekCommand;
import jcreek.cke.CreekDocumentGroup;
import jcreek.representation.KnowledgeModel;
import jcreek.representation.*; import jcreek.representation.cbr.*;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import javax.swing.event.EventListenerList;
import javax.swing.event.UndoableEditListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.undo.UndoableEdit;
import java.util.Vector;



/**
 * Title:        JavaCreek
 * Description:  A Case Based Reasoning approach developed in Java
 * Copyright:    Copyright (c) 1999
 * Company:      Phil Systems
 * @author Victor Dias
 * @version
 */

public class KnowledgeModelController
{

    /**
     * The knowledge Model for this document
     */
     protected KnowledgeModel km;
    /**
     * Object supporting the Serialize format.
     */
    protected SerializeFormat format;
    /**
     * File to read and write to.
     */
    protected File file;
    /**
     * True if document is dirty and needs saving.
     */
    protected boolean dirty;
    /**
     * True if the document is titled and has a name.
     */
    protected boolean titled;
    /**
     * List of event listeners.
     */
    protected EventListenerList listeners;

    /**
     * Reference to the single instance of this class.
     */
    protected static KnowledgeModelController instance;


    /**
     * Creates a new SerializeGraphDocument object. It defaults to
     * a untitled document with no associated file or view.
     */
    protected KnowledgeModelController()
    {
        // create necessary objects
        listeners = new EventListenerList();
        format = new SerializeFormat();
    } // GMLAdjListDocument

    /**
     * Registers an UndoableEditListener with this document.
     * As commands are performed on the document undoable
     * edit events will be sent out to all the listeners.
     *
     * @param  l  listern to add to this document
     * @see #removeUndoableEditListener
     */
    public void addUndoableEditListener( UndoableEditListener l ) {
        listeners.add( UndoableEditListener.class, l );
    } // addUndoableEditListener

    /**
     * We're being shutdown, remove listeners from lists.
     */
    public void close() {
        format = null;
        file = null;
        listeners = null;
    } // close

    /**
     * Fires an undoable edit event to all the listeners
     * registered for such events.
     *
     * @param  ue  undoable edit
     */
    protected void fireEdit( UndoableEdit ue ) {
        UndoableEditEvent e = new UndoableEditEvent( this, ue );
        // Get the listener list as class/instance pairs.
        Object[] list = listeners.getListenerList();
        // Process the listeners last to first.
        // List is in pairs: class, instance
        for ( int i = list.length - 2; i >= 0; i -= 2 ) {
            if ( list[i] == UndoableEditListener.class ) {
                UndoableEditListener l = (UndoableEditListener)list[i+1];
                l.undoableEditHappened( e );
            }
        }
    } // fireEdit

    /**
     * Returns the reference to the file object for this document.
     *
     * @return  reference to file object
     */
    public File getFile() {
        return file;
    } // getFile

    /**
     * Returns the reference to the object acting as the model
     * for this document. The particular implementation of this
     * model depends on the implemention of CreekDocument.
     *
     * @return  reference to model object
     */
    public KnowledgeModel getModel() {
        return km;
    } // getModel

    /**
     * Returns the name of the file this document is reading
     * and writing to. Currently this does not include the
     * path to the file.
     *
     * @return  name of the document
     */
    public String getTitle() {
        return file.getName();
    } // getTitle


    /**
     * Returns the current dirty status of this document. If the
     * document is dirty and needs to be saved, true will be
     * returned.
     *
     * @return  true if dirty, false if not dirty
     */
    public boolean isDirty() {
        return dirty;
    } // isDirty

    /**
     * Returns true if the file is currently named.
     *
     * @return  true if titled, false if unnamed
     */
    public boolean isTitled() {
        return titled;
    } // isTitled

    /**
     * Perform the given command and fire an undoable edit event.
     *
     * @param  command  command to be performed
     * @return  true if command was performed successfully
     */
    public boolean performCommand( CreekCommand command ) {
        boolean success = command.perform();
        if ( success ) {
            fireEdit( command );
        }
        return success;
    } // performCommand


    /**
     * The state of the model changed, mark ourselves dirty. If
     * we're newly marked dirty, notify the document group that
     * we've become dirty.
     *
     * @param  e  change event
     */
    public void propertyChange( PropertyChangeEvent e ) {
    } // propertyChange

    /**
     * Removes an UndoableEditListener from this document.
     *
     * @param  l  listern to remove from this document
     * @see #addUndoableEditListener
     */
    public void removeUndoableEditListener( UndoableEditListener l ) {
        if ( listeners != null ) {
            listeners.remove( UndoableEditListener.class, l );
        } else {
            System.err.println( "ERROR: listener list is null" );
        }
    } // removeUndoableEditListener

    /**
     * Reverts this document (the graph data) to the last-saved
     * version of the disk file. If the document is untitled, then
     * the graph will simply be cleared and nothing will be read
     * from file.
     *
     * @return  true if successful, false if error
     */
    public boolean revert() {
        if ( titled ) {
            // read in the vertex and edge list from the file
            FileInputStream fis = null;
            try {
                fis = new FileInputStream( file );
            } catch ( FileNotFoundException fnfe ) {
                return false;
            }
/*            if ( !format.read(fis,km ) ) {
                return false;
            }*/
        }
/*        // replace the existing vertices and edges in the
        // graph with these vertices and edges
        model =          */
        dirty = false;
        return true;
    } // revert


    /**
     * Opens a test Model
     * @return  true if successful, false if error
     */
    public boolean openTestModel() {
      km = jcreek.examples.FruitModel.generateModel();
      return true;

    } // openTestAction

    /**
     * Saves this document (the graph data) to the associated disk
     * file. The format is determined by the CreekFormat class used
     * by the particular CreekDocument implementation.
     *
     * @return  true if successful, false if error
     */
    public boolean save() {
        return true;
    } // save

    /**
     * Saves this document (the graph data) to a new file, and
     * hereinafter is associated with this new file.
     *
     * @param  file  new file to save to
     * @return  true if successful, false if error
     */
    public boolean saveAs( File file ) {
        this.file = file;
        return save();
    } // saveAs

    /**
     * Sets the file this document is associated with. Documents
     * should not allow setting the file more than once.
     *
     * @param  file  file object
     */
    public void setFile( File file ) {
        if ( this.file == null ) {
            this.file = file;
        }
    } // setFile

    /**
     * Tells the document whether it has a title or not.
     *
     * @param  titled  true if document has a titled
     */
    public void setTitled( boolean titled ) {
        this.titled = titled;
    } // setTitled


    /**
     * Returns the reference to the single instance of this class.
     * If an instance does not exist it will be created.
     *
     * @return  the instance of this class
     */
    public static KnowledgeModelController instanceOf()
    {
        if ( instance == null )
        {
            instance = new KnowledgeModelController();
        }
        return instance;
    } // instanceOf

    public Entity[] getConceptList()
    {
        Entity[] conArray = new Entity[0];
        if(km !=null)
        {
          return km.getEntities();
        }
        else
         return conArray;
    }//getConceptList

}//knowledgeModelController
