/*********************************************************************
 *
 * DESCRIPTION:
 *      This class implements the CreekDocument interface using the
 *      Serialization file format and the Graph model.
 *
 ********************************************************************/
package jcreek.cke.document;

import jcreek.cke.graph.CreekModel;
import jcreek.cke.graph.GraphModel;
import jcreek.cke.command.CreekCommand;
import jcreek.cke.CreekDocumentGroup;
import jcreek.representation.Context;
import jcreek.representation.KnowledgeModel;
import jcreek.cke.graph.Graph;
import jcreek.representation.ModelChangeEvent;
import jcreek.representation.ModelChangeListener;

import jcreek.representation.RelationType;

import jcreek.representation.Entity;
import jcreek.representation.Relation;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.ClassNotFoundException;
import java.lang.Class;
import javax.swing.event.EventListenerList;
import javax.swing.event.UndoableEditListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.undo.UndoableEdit;

import jcreek.cke.conceptview.*;
import java.io.ObjectInputStream;


/**
 * Implements the CreekDocument interface for the serialize file format
 * with the graph model. This is instantiated when the
 * CreekDocumentGroup opens a file with the extension "km".
 *
 * @author
 * @version
 * @see
 * @see
 * @see
 */
public class SerializeGraphDocument implements CreekDocument, PropertyChangeListener, ModelChangeListener {
    /**
     * Our model.
     */
    protected GraphModel model;
    /**
     * Our view.
     */
    protected transient ConceptView view;
    /**
     * Object supporting the Serialize format.
     */
//     protected SerializeFormat format;

    /**
     * True if document is dirty and needs saving.
     */
    protected transient boolean dirty;
    /**
     * True if the document is titled and has a name.
     */
    protected boolean titled;

    protected String title;

    /**
     * File to read and write to.
     */
//     protected File file;
    /**
     * List of event listeners.
     */
    protected transient EventListenerList listeners;

    /**
     * Creates a new SerializeGraphDocument object. It defaults to
     * a untitled document with no associated file or view.
     */
    public SerializeGraphDocument() {
        // create necessary objects
//        format = new SerializeFormat();
        model = new GraphModel();
        listeners = new EventListenerList();
        // register a change listener with model
        model.addChangeListener( this );
    } // GMLAdjListDocument

    /**
     * Registers an UndoableEditListener with this document.
     * As commands are performed on the document undoable
     * edit events will be sent out to all the listeners.
     *
     * @param  l  listern to add to this document
     * @see #removeUndoableEditListener
     */
    public void addUndoableEditListener( UndoableEditListener l ) {
        listeners.add( UndoableEditListener.class, l );
    } // addUndoableEditListener


    public void open() {
        getKnowledgeModel().addModelChangeListener(this);
    }

    /**
     * We're being shutdown, remove listeners from lists.
     */
    public void close() {
        CreekDocumentGroup.instanceOf().getKM().removeModelChangeListener(this);
        if ( view != null ) {
            view.detach();
            view = null;
        }
/*
        if(model != null)
        {
            model.removeChangeListener( this );
            model = null;
        }

        format = null;
        listeners = null;

*/
    } // close

    /**
     * Fires an undoable edit event to all the listeners
     * registered for such events.
     *
     * @param  ue  undoable edit
     */
    protected void fireEdit( UndoableEdit ue ) {
        UndoableEditEvent e = new UndoableEditEvent( this, ue );
        // Get the listener list as class/instance pairs.
        Object[] list = listeners.getListenerList();
        // Process the listeners last to first.
        // List is in pairs: class, instance
        for ( int i = list.length - 2; i >= 0; i -= 2 ) {
            if ( list[i] == UndoableEditListener.class ) {
                UndoableEditListener l = (UndoableEditListener)list[i+1];
                l.undoableEditHappened( e );
            }
        }
    } // fireEdit


    /**
     * Returns the reference to the object acting as the model
     * for this document. The particular implementation of this
     * model depends on the implemention of CreekDocument.
     *
     * @return  reference to model object
     */
    public CreekModel getModel() {
        return model;
    } // getModel

    /**
     * Returns the name of the file this document is reading
     * and writing to.
     *
     * @return  name of the document
     */
    public String getTitle() {
          if(titled)
              return title;
          else
            return "No name";

//        return file.getName();
    } // getTitle


    public void setTitle(String title)
    {
        this.title =  title;
    }

    /**
     * Returns the reference to the object acting as the view
     * for this document. The particular implementation of this
     * view depends on the implemention of CreekDocument.
     *
     * @return  reference to view object
     */
    public ConceptView getView() {
        return view;
    } // getView

    /**
     * Returns the current dirty status of this document. If the
     * document is dirty and needs to be saved, true will be
     * returned.
     *
     * @return  true if dirty, false if not dirty
     */
    public boolean isDirty() {
        return dirty;
    } // isDirty

    /**
     * Returns true if the file is currently named.
     *
     * @return  true if titled, false if unnamed
     */
    public boolean isTitled() {
        return titled;
    } // isTitled

    /**
     * Perform the given command and fire an undoable edit event.
     *
     * @param  command  command to be performed
     * @return  true if command was performed successfully
     */
    public boolean performCommand( CreekCommand command ) {
        boolean success = command.perform();
        if ( success ) {
            fireEdit( command );
        }
        return success;
    } // performCommand


    /**
     * The state of the model changed, mark ourselves dirty. If
     * we're newly marked dirty, notify the document group that
     * we've become dirty.
     *
     * @param  e  change event
     */
    public void propertyChange( PropertyChangeEvent e ) {
        if ( e.getPropertyName().equals( "change" ) ) {
            // The model structure changed, make sure to mark
            // ourselves dirty.
            boolean oldState = dirty;
            dirty = true;
            if ( oldState == false ) {
                // notify document group we're dirty
                CreekDocumentGroup.instanceOf().markDocumentDirty( this );
            }
        if(view !=null)
          view.dataChanged( model.getEntities(),model.getRelations() );

        }
        if ( e.getPropertyName().equals( "selection" ) ) {
                // notify document group we're dirty
                CreekDocumentGroup.instanceOf().markSelection( this );
            }

    } // propertyChange

    /**
     * Removes an UndoableEditListener from this document.
     *
     * @param  l  listern to remove from this document
     * @see #addUndoableEditListener
     */
    public void removeUndoableEditListener( UndoableEditListener l ) {
        if ( listeners != null ) {
            listeners.remove( UndoableEditListener.class, l );
        } else {
            System.err.println( "ERROR: listener list is null" );
        }
    } // removeUndoableEditListener

    /**
     * Reverts this document (the graph data) to the last-saved
     * version of the disk file. If the document is untitled, then
     * the graph will simply be cleared and nothing will be read
     * from file.
     *
     * @return  true if successful, false if error
     */
    public boolean revert() {
/*
        Graph grp = new Graph();
        StringBuffer savedTitle = new StringBuffer();
        if ( titled ) {
            // read in the vertex and edge list from the file
            FileInputStream fis = null;
            try {
                fis = new FileInputStream( file );
            } catch ( FileNotFoundException fnfe ) {
                return false;
            }
            if ( !format.read( fis, grp ,savedTitle) ) {
                return false;
            }
        }
        // replace the existing nodes and relation in the
        // graph with these nodes and relations
        model.replace( grp );
        this.title = savedTitle.toString();
        dirty = false;
*/
        return true;
    } // revert

    public boolean revert(Graph g,String savedTitle) {
//        Graph grp = new Graph();
//        grp.copy(g);
        // replace the existing nodes and relation in the
        // graph with these nodes and relations
//        model.replace( grp );
        model.replace(g);
        this.title = savedTitle;
/*        if(getFile()!=null)
            save(); */
        dirty = false;
        return true;
    } // revert



    /**
     * Saves this document (the graph data) to the associated disk
     * file. The format is determined by the CreekFormat class used
     * by the particular CreekDocument implementation.
     *
     * @return  true if successful, false if error
     */
    /*
    public boolean save() {
        KnowledgeModel km = CreekDocumentGroup.instanceOf().getKM();
        if(!km.hasContext(this))
        {
            km.addContext(this);
        }
        else
        {
            km.removeContext(this);
            km.addContext(this);
        }
        return true;
        FileOutputStream fos = null;
        try {
            fos = new FileOutputStream( file );
        } catch ( IOException ioe ) {
            return false;
        }

        if ( !format.write( fos, model.getGraph(),getTitle()) ) {
            return false;
        }
        dirty = false;
        return true;
    } // save
    */

    /**
     * Saves this document (the graph data) to a new file, and
     * hereinafter is associated with this new file.
     *
     * @param  file  new file to save to
     * @return  true if successful, false if error
     */
/*    public boolean saveAs( String title ) {
//        this.title = title;
        return save();
    } // saveAs */


    /**
     * Saves this document (the graph data) to a new file, and
     * hereinafter is associated with this new file.
     *
     * @param  file  new file to save to
     * @return  true if successful, false if error
     */
/*    public boolean saveAs( File file ) {
        this.file = file;
        return save();
    } // saveAs*/

    /**
     * Tells the document whether it has a title or not.
     *
     * @param  titled  true if document has a titled
     */
    public void setTitled( boolean titled ) {
        this.titled = titled;
    } // setTitled

    /**
     * Sets the view this document is associated with. Documents
     * should not allow setting the view more than once.
     *
     * @param  view  view object
     */
    public void setView( ConceptView view ) {
        if ( this.view == null ) {
            this.view = view;
            view.attach(model);
        }
    } // setView

    public String toString()
    {
        return getTitle();
    }//toString


    public KnowledgeModel getKnowledgeModel()
    {
        return CreekDocumentGroup.instanceOf().getKM();
    }

    /**
     * Sets the file this document is associated with. Documents
     * should not allow setting the file more than once.
     *
     * @param  file  file object
     */
/*    public void setFile( File file ) {
        if ( this.file == null ) {
            this.file = file;
        }
    } // setFile  */

    /**
     * Returns the reference to the file object for this document.
     *
     * @return  reference to file object
     */
/*    public File getFile() {
        return file;
    } // getFile
*/

  public void modelChanged(ModelChangeEvent e)
  {
    if(e.getType() == ModelChangeEvent.ENTITY_REMOVED)
      updateModel((Entity) e.getSource());
    if(e.getType() == ModelChangeEvent.RELATION_REMOVED)
      updateModel((Relation) e.getSource());
  }


  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException
  {
    in.defaultReadObject();
    model.addChangeListener(this);
    listeners = new EventListenerList();
  }

  protected void updateModel(Entity c)
  {
    model.removeEntity((Entity)c);
  }

  protected void updateModel(Relation c)
  {
    model.removeRelation((Relation)c);
  }




} // SerializeGraphDocument





