package jcreek.cke.graph;

import java.io.Serializable;

import java.util.Vector;
import jcreek.cke.conceptview.VisualRelation;
import jcreek.cke.conceptview.VisualEntity;


public class RelationList implements Cloneable,Serializable {

    /**
     * List of vertices.
     */
    protected Vector relations;

    /**
     * No-arg constructor; creates the vector to hold the
     * list of vertices.
     */
    public RelationList() {
        relations = new Vector();
    } //RelationList

    /**
     * creates the vector to hold the
     * list of relations.
     */
    public RelationList(VisualRelation[] relList) {
        relations = new Vector();
        for(int i=0;i<relList.length;i++)
        {
              add(relList[i]);
        }
    } // VertexList

    /**
     * Add a new VisualRelation to this list. Won't add null references.
     *
     * @param  node new VisualRelation to add to list
     */
    public synchronized void add( VisualRelation edge ) {
        if ( edge != null ) {
            if(!relations.contains(edge))
                relations.addElement( (Object)edge );
        }
    } // add

    /**
     * Removes all of the relations from the list.
     */
    public synchronized void clear() {
        relations.clear();
    } // clear

    /**
     * Creates a shallow copy of this VisualRelation list, cloning the
     * list of the relations into the new object.
     *
     * @return  clone of this relation list
     */
    public Object clone() {
        RelationList vl = new RelationList();
        vl.relations = (Vector)relations.clone();
        return vl;
    } // clone

    /**
     * Makes a deep copy of this relation list, cloning each of
     * the vertices in the list.
     *
     * @return  copy of this relation list
     */
    public RelationList copy() {
        RelationList vl = new RelationList();
        int size = relations.size();
        VisualRelation v;
        try
        {
            for ( int i = 0; i < size; i++ ) {
                v = (VisualRelation)relations.elementAt( i );
                vl.add( (VisualRelation)v.clone() );
            }
        }
        catch(CloneNotSupportedException ex)
        {
            ex.printStackTrace();
        }
        return vl;
    } // copy

    /**
     * Get the VisualRelation stored at the specified index in the list.
     *
     * @param  index number of the nth VisualRelation to get (0-based)
     * @return the VisualRelation specified, or null if not in list
     */
    public VisualRelation get( int index ) {
        try {
            return (VisualRelation)relations.elementAt( index );
        } catch ( ArrayIndexOutOfBoundsException e ) {
            return null;
        }
    } // get

    /**
     * Find the index of the element in the list.
     *
     * @param  node VisualRelation to search for
     * @return position of element in list; -1 if not found
     */
    public int indexOf( VisualRelation edge ) {
        return relations.indexOf( (Object)edge );
    } // indexOf

    /**
     * Combines the elements from <code>list</code> into this
     * list of relations. Relations in <code>list</code> will only
     * be added if there is nothing similar in our list.
     *
     * @param  list  list of vertices to add
     */
    public void join( RelationList list ) {
        int nlSize = list.size();
        int size = relations.size();
        for ( int i = 0; i < nlSize; i++ ) {
            // Assume VisualRelation does not exist.
            boolean found = false;
            VisualRelation v1 = list.get( i );
            for ( int j = 0; j < size; j++ ) {
                // For each of their relations, loop through all
                // the relations in our list looking for a match.
                VisualRelation v2 = get( j );
                if ( ( v1 == v2 ) || v1.equals( v2 ) ) {
                    // We found a match!
                    found = true;
                    break;
                }
            }
            if ( !found ) {
                // VisualRelation is not in our list, add it.
                relations.addElement( (Object)v1 );
            }
        }
    } // join

    /**
     * Removes the specified element from the list given the VisualRelation itself.
     *
     * @param  edgeToRemove  VisualRelation to remove from list
     * @return  true if the VisualRelation was in the list, false otherwise
     */
    public synchronized boolean remove( VisualRelation edgeToRemove ) {
        return relations.removeElement( (Object)edgeToRemove );
    } // remove

    /**
     * Removes the specified element from the list given the index
     * of the VisualRelation within the list.
     *
     * @param  edgeToRemove  index of VisualRelation to remove from list
     * @return  true if the VisualRelation was in the list, false otherwise
     */
    public synchronized boolean remove( int edgeToRemove ) {
        try {
            relations.removeElementAt( edgeToRemove );
            return true;
        } catch ( ArrayIndexOutOfBoundsException e ) {
            return false;
        }
    } // remove

    /**
     * Returns all the edges that are connected to the given
     * vertex. Useful when removing a vertex from the graph.
     *
     * @param  v         vertex to find edges for
     * @return  list of edges connected to vertex; null if no match
     */
    public RelationList find( VisualEntity v ) {
        // Start by creating a new relation list.
        RelationList rl = new RelationList();
        int i = 0;
        VisualRelation re1 = get( i );
            // If undirected look for edges going to or leaving from v.
            // Run through our edge list looking for edges
            // that contain vertices matching the parameter
            // (either the "from" or the "to", we'll take both).
            while ( re1 != null ) {
                if ( re1.getFrom() == v ) {
                    // found an edge
                    rl.add( re1 );
                }
                if ( re1.getTo() == v ) {
                    // found an edge
                    rl.add( re1 );
                }
                // Go to the next edge in our list.
                i++;
                re1 = get( i );
            }
        // Check if we found any matches.
        if ( rl.size() == 0 ) {
            return null;
        } else {
            return rl;
        }
    }

    /**
     * Returns the size of the relation list.
     *
     * @return number of relations in list
     */
    public int size() {
        return relations.size();
    } // size

    /**
     * Error-checking method that verifies the data and structure
     * of the vertex list object.
     *
     * @return  true if okay, false if error
     */
    public boolean verify() {
	return true;
    } // verify



}//RelationList