package jcreek.cke.graph;

import java.util.Vector;
import java.io.*;

import jcreek.cke.conceptview.VisualEntity;


/**
 * A VertexList encapsulates a list of vertices. It can hold
 * as many vertices as needed.
 *
 * @author  Nathan Fiedler
 * @version 1.0  5/5/97
 * @see java.util.Vector
 */
public class VertexList implements Cloneable,Serializable {
    /**
     * List of vertices.
     */
    protected Vector vertices;

    /**
     * No-arg constructor; creates the vector to hold the
     * list of vertices.
     */
    public VertexList() {
        vertices = new Vector();
    } // VertexList

    /**
     * creates the vector to hold the
     * list of vertices.
     */
    public VertexList(VisualEntity[] veList) {
        vertices = new Vector();
        for(int i=0;i<veList.length;i++)
        {
              add(veList[i]);
        }
    } // VertexList
    /**
     * Add a new VisualEntity to this list. Won't add null references.
     *
     * @param  node new VisualEntity to add to list
     */
    public synchronized void add( VisualEntity node ) {
        if ( node != null ) {
            if(!vertices.contains(node))
              vertices.addElement( (Object)node );
        }
    } // add

    /**
     * Removes all of the vertices from the list.
     */
    public synchronized void clear() {
        vertices.clear();
    } // clear

    /**
     * Creates a shallow copy of this VisualEntity list, cloning the
     * list of the vertices into the new object.
     *
     * @return  clone of this vertex list
     */
    public Object clone() {
        VertexList vl = new VertexList();
        vl.vertices = (Vector)vertices.clone();
        return vl;
    } // clone

    /**
     * Makes a deep copy of this vertex list, cloning each of
     * the vertices in the list.
     *
     * @return  copy of this vertex list
     */
    public VertexList copy() {
        VertexList vl = new VertexList();
        int size = vertices.size();
        VisualEntity v;
        try
        {
            for ( int i = 0; i < size; i++ ) {
                v = (VisualEntity)vertices.elementAt( i );
                vl.add( (VisualEntity)v.clone() );
            }
        }
        catch(CloneNotSupportedException ex)
        {
            ex.printStackTrace();
        }
        return vl;
    } // copy

    /**
     * Get the VisualEntity stored at the specified index in the list.
     *
     * @param  index number of the nth VisualEntity to get (0-based)
     * @return the VisualEntity specified, or null if not in list
     */
    public VisualEntity get( int index ) {
        try {
            return (VisualEntity)vertices.elementAt( index );
        } catch ( ArrayIndexOutOfBoundsException e ) {
            return null;
        }
    } // get

    /**
     * Find the index of the element in the list.
     *
     * @param  node VisualEntity to search for
     * @return position of element in list; -1 if not found
     */
    public int indexOf( VisualEntity node ) {
        return vertices.indexOf( (Object)node );
    } // indexOf

    /**
     * Combines the elements from <code>list</code> into this
     * list of vertices. Vertices in <code>list</code> will only
     * be added if there is nothing similar in our list.
     *
     * @param  list  list of vertices to add
     */
    public void join( VertexList list ) {
        int nlSize = list.size();
        int size = vertices.size();
        for ( int i = 0; i < nlSize; i++ ) {
            // Assume VisualEntity does not exist.
            boolean found = false;
            VisualEntity v1 = list.get( i );
            for ( int j = 0; j < size; j++ ) {
                // For each of their vertices, loop through all
                // the vertices in our list looking for a match.
                VisualEntity v2 = get( j );
                if ( ( v1 == v2 ) || v1.equals( v2 ) ) {
                    // We found a match!
                    found = true;
                    break;
                }
            }
            if ( !found ) {
                // VisualEntity is not in our list, add it.
                vertices.addElement( (Object)v1 );
            }
        }
    } // join

    /**
     * Removes the specified element from the list given the VisualEntity itself.
     *
     * @param  nodeToRemove  VisualEntity to remove from list
     * @return  true if the VisualEntity was in the list, false otherwise
     */
    public synchronized boolean remove( VisualEntity nodeToRemove ) {
        return vertices.removeElement( (Object)nodeToRemove );
    } // remove

    /**
     * Removes the specified element from the list given the index
     * of the VisualEntity within the list.
     *
     * @param  nodeToRemove  index of VisualEntity to remove from list
     * @return  true if the VisualEntity was in the list, false otherwise
     */
    public synchronized boolean remove( int nodeToRemove ) {
        try {
            vertices.removeElementAt( nodeToRemove );
            return true;
        } catch ( ArrayIndexOutOfBoundsException e ) {
            return false;
        }
    } // remove

    /**
     * Returns the size of the vertex list.
     *
     * @return number of vertices in list
     */
    public int size() {
        return vertices.size();
    } // size

    /**
     * Error-checking method that verifies the data and structure
     * of the vertex list object.
     *
     * @return  true if okay, false if error
     */
    public boolean verify() {
	// look for duplicate VisualEntity entries
	int size = size();
	VisualEntity v1, v2;
	for ( int i = 0; i < size; i++ ) {
	    for ( int j = 0; j < size; j++ ) {
		// It's slow but it works. Simply look for entries
		// that are duplicates.
		if ( i != j ) {
		    v1 = get( i );
		    v2 = get( j );
		    if ( ( v1 == v2 ) || ( v1.equals( v2 ) ) ) {
			return false;
		    }
		}
	    }
	}
	// If we made it this far, list must be okay.
	return true;
    } // verify


} // VertexList




