package jcreek.cke.importexport;

import jcreek.representation.KnowledgeModel;
import jcreek.representation.inference.ActivationConstants;

/**
 * Some general methods used to ease the export process.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class KnowledgeModelExportUtilities {
	private KnowledgeModel km;

	/**
	 * The constructor that only sets the JavaCreek Knowledge Model
	 * to use in the export process.
	 *
	 * @param km the JavaCreek Knowledge Model that the data should
	 * be exported from
	 */
	public KnowledgeModelExportUtilities(KnowledgeModel km) {
		this.km = km;
	}

	/**
	 * Determines if a text is a text and not equal 'null' or
	 * of length 0. This is e.g. important for 'description'
	 * fields in the knowledge model.
	 *
	 * @param text the text to check
	 *
	 * @return true if the text is of length more than 0 and not 'null',
	 * otherwise returns false.
	 */
	public boolean hasLength(String text) {
		if ((text != null) && (text.trim().length() > 0))
			return true;
		else
			return false;
	} // hasLength

	/**
	 * Gets tyhe string representation of the strength value. If the value
	 * is out of range, then the default value will be returned.
	 *
	 * @param strength the value to convert
	 * @param includeDefault if set to false, then if the value is the default
	 * value or out of range then 'null' will be returned, otherwise the real value
	 * will be returned.
	 *
	 * @return a string representation of the value. Does return the default
	 * value if out of range.
	 */
	public String getStrength(double strength, boolean includeDefault) {
		if ((0 <= strength) && (strength <= 1)) {
			if ((includeDefault) && (strength == 0.5))
				return String.valueOf(strength);
			else
				return null;
		}
		else {
			if (includeDefault)
				return String.valueOf(0.5);
			else
				return null;
		}
	} // getStrength

	/**
	 * Gets the discrete variable of the double value as a string of. The value
	 * will be maped againts the predefained values (see ActivationConstants below).
	 * If the value is not valid according to these values the default value will
	 * be returned.
	 *
	 * @param importance the value to convert
	 * @param includeDefault if set to false, then if the value is the default
	 * value or out of range then 'null' will be returned, otherwise the real value
	 * will be returned.
	 *
	 * @return returns a string representing the value. If the value isn't
	 * valid, then the default value will be returned.
	 *
	 * @see jcreek.representation.inference.ActivationConstants
	 */
	public String getImportance(double importance, boolean includeDefault) {
		/**
		 * From ActivationConstants:
		  final static double NECCESSARY          = 0.95,
		                      CHARACTERISTIC      = 0.8,
		                      INFORMATIVE         = 0.5,
		                      DEFAULT_IMPORTANCE  = 0.5,
		                      IRRELEVANT          = 0.05;
		*/
		if (importance
			>= ((ActivationConstants.NECCESSARY + ActivationConstants.CHARACTERISTIC) / 2)) {
			return "neccessary";
		}
		else if (
			importance
				>= ((ActivationConstants.CHARACTERISTIC + ActivationConstants.INFORMATIVE) / 2)) {
			return "characteristic";
		}
		else if (
			importance
				>= ((ActivationConstants.INFORMATIVE + ActivationConstants.IRRELEVANT) / 2)) {
			if (includeDefault)
				return "informative";
			else
				return null;
		}
		else if (
			(importance >= 0)
				&& (importance
					< ((ActivationConstants.INFORMATIVE + ActivationConstants.IRRELEVANT) / 2))) {
			return "irrelevant";
		}
		else {
			if (includeDefault)
				return "default_importance";
			else
				return null;
		}
	} // getImportance

	/**
	 * Gets the discrete variable of the double value as a string of. The value
	 * will be maped againts the predefained values (see ActivationConstants below).
	 * If the value is not valid according to these values the default value will
	 * be returned.
	 *
	 * @param predictiveStrength the value to convert
	 * @param includeDefault if set to false, then if the value is the default
	 * value or out of range then 'null' will be returned, otherwise the real value
	 * will be returned.
	 *
	 * @return returns a string representing the value. If the value isn't
	 * valid, then the default value will be returned.
	 *
	 * @see jcreek.representation.inference.ActivationConstants
	 */
	public String getPredictiveStrength(
		double predictiveStrength,
		boolean includeDefault) {
		/**
		  final static int  SUFFICIENT                  = 1,
		                    STRONGLY_INDICATIVE         = 2,
		                    DEFAULT_PREDICTIVE_STRENGTH = 2,
		                    INDICATIVE                  = 3,
		                    SPURIOUS                    = 4;
		*/
		if (predictiveStrength
			>= ((ActivationConstants.SPURIOUS + ActivationConstants.INDICATIVE) / 2)) {
			return "spurious";
		}
		else if (
			predictiveStrength
				>= ((ActivationConstants.INDICATIVE + ActivationConstants.STRONGLY_INDICATIVE)
					/ 2)) {
			return "indicative";
		}
		else if (
			predictiveStrength
				>= ((ActivationConstants.STRONGLY_INDICATIVE + ActivationConstants.SUFFICIENT)
					/ 2)) {
			if (includeDefault)
				return "strongly_indicative";
			else
				return null;
		}
		else if (
			(predictiveStrength >= 0)
				&& (predictiveStrength
					< ((ActivationConstants.STRONGLY_INDICATIVE + ActivationConstants.SUFFICIENT)
						/ 2))) {
			return "sufficient";
		}
		else {
			if (includeDefault)
				return "default_predictive_strength";
			else
				return null;
		}
	} // getPredictiveStrength
}