package jcreek.cke.importexport;

import jcreek.representation.cbr.Case;


import jcreek.representation.Entity;

import jcreek.representation.KnowledgeModel;
import jcreek.representation.Relation;
import jcreek.representation.RelationType;
import jcreek.representation.inference.ActivationConstants;

/**
 * Some general methods used to ease the import process.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class KnowledgeModelImportUtilities {
        private KnowledgeModel km;

        /**
         * The constructor that only sets the JavaCreek Knowledge Model
         * to use in the import process.
         *
         * @param km the JavaCreek Knowledge Model that the data should
         * be imported into
         */
        public KnowledgeModelImportUtilities(KnowledgeModel km) {
                this.km = km;
        }

        /**
         * Creates a new relation type if not already exists.
         *
         * @param classDescriptor the class descriptor from which this new relation type
         * should inherit functionallity from
         * @param name the name of the new relation type
         * @param inverseName the inverse name of the new relation type
         * @param defaultStrengthStr the default explanation strength, a value between 0 and 1.
         * If this value is not in between 0 and 1 then a default value will be used which is 0.5
         * @param description the description of the new relation type
         *
         * @throws ClassDescriptorDoNotExist if the classDescriptor does not exists
         * @throws RelationTypeAlreadyExist in a relation type already exist
         * @throws EqualRelationTypeAlreadyExist if an equal relation type already exist
         * @throws RelationTypeNameAlreadyInUse if a realtion with an equal name exist
         * @throws CompilationFault if it wasn't possible to compile the new relation type,
         * which e.g. may happen in cases where the classpath doesn't contain tool.jar
         */
        public void addNewRelationType(
                String classDescriptor,
                String name,
                String inverseName,
                String defaultStrengthStr,
                String description)
                throws
                        ClassDescriptorDoNotExist,
                        RelationTypeAlreadyExist,
                        EqualRelationTypeAlreadyExist,
                        RelationTypeNameAlreadyInUse,
                        CompilationFault {

                boolean cdFound = false;
                boolean nameFound = false;
                boolean inverseFound = false;

                if (km.getEntity(classDescriptor) != null)
                        cdFound = true;
                if (km.getRelationType(name) != null)
                        nameFound = true;
                if (km.getRelationType(inverseName) != null)
                        inverseFound = true;

                // now check the result
                // if not found then end the hole process here
                if (!cdFound)
                        throw new ClassDescriptorDoNotExist(classDescriptor);

                // if both are found then need to do some more checking, it might be that
                // a relation use both the name and the inverse name, but still the
                // class descriptor might differ or that the names are used in two different
                // relations etc.
                if (nameFound && inverseFound) {
                        if (km
                                .getRelationType(name)
                                .getInverseName()
                                .equalsIgnoreCase(inverseName)) {
                                // ok this is the same relation names, but does it use the same class descriptor?


                                                // now we know that an equal relation exist and throws a warning
                                                throw new EqualRelationTypeAlreadyExist(name, inverseName);



                        } else {
                                // now we know that the names are used in two different relation types
                                throw new RelationTypeNameAlreadyInUse(name + "' and '" + inverseName);
                        }
                } else if (nameFound) {
                        throw new RelationTypeNameAlreadyInUse(name);
                } else if (inverseFound) {
                        throw new RelationTypeNameAlreadyInUse(inverseName);
                }

                // now check the defaultStrength
                double strength = getStrength(defaultStrengthStr);

                // now check the description
                if (description == null) {
                        description = "";
                }

                RelationType relationRCD =
                        km.getRelationType(classDescriptor);
                // PS! unable to catch the exception that might come from the compilation of this class
                try {
                        new RelationType(
                                km,
                                relationRCD,
                                name,
                                description,
                                inverseName,
                                description,
                                strength);
                }
                catch (Exception ex)
                {
                  throw new CompilationFault();
                }

        } // addNewRelationType

        /**
         * Adds a new entity if not already exist. If entity exist
         * then throws a warning exception that the entity exist.
         *
         * @param name the name of the new entity
         * @param description the description of the new entity
         *
         * @throws EntityAlreadyExist if an entity by the same name already exist
         */
        public void addEntity(String name, String description, String valueClass, String value)
                throws EntityAlreadyExist {
                if (km.getEntity(name) == null) {
                        // check the description
                        if (description == null) {
                                description = "";
                        }
                        try {
                                Entity ent = new Entity(km, name, description);
                                if(valueClass != null && !valueClass.equals(""))
                                {
                                  Class valClass = Class.forName(valueClass);
                                  java.lang.reflect.Constructor con = valClass.getConstructor(new Class[] {String.class});
                                  Object ob = con.newInstance(new Object[] {value});
                                  ent.setEntityObject((java.io.Serializable) ob);
                                }

                        } catch (jcreek.representation.NameAlreadyExistException ex) {
                                throw new EntityAlreadyExist(name);
                        } catch (Exception ex)
                        {
                          ex.printStackTrace();
                        }

                } else {
                        throw new EntityAlreadyExist(name);
                }
        } // addEntity

        /**
         * Adds a new relation between two entities. Both the entities and the
         * relation type must exist in able to add a new relation.
         *
         * @param fromEntity the name of the entity from where the new relation should come from
         * @param relationType the relation type to use
         * @param toEntity the name of the entity to where the the new relation should end
         * @param strengthStr the default strength, a value between 0 and 1.
         * If this value is not in between 0 and 1 then a default value will be used which is 0.5
         *
         * @throws EqualRelationAlreadyExist if an equal relation between these two entities
         * already exist
         * @throws EntityDoNotExist if one of the entities specified doesn't exist
         * @throws RelationTypeDoNotExist if the relation type specified doesn't exist
         */
        public void addRelation(
                String fromEntity,
                String relationType,
                String toEntity,
                String strengthStr)
                throws EqualRelationAlreadyExist, EntityDoNotExist, RelationTypeDoNotExist {

                // now check the defaultStrength
                double strength = getStrength(strengthStr);

                // test if have a relation allready, if that is the case, do not add another one
                // this method should perhaps be a bit better, in that it replaces a relation
                // that is considered of less importance

                // if an equal relation already exist then throw a warning exception
                if (equalRelation(fromEntity, relationType, toEntity)) {
                        throw new EqualRelationAlreadyExist(fromEntity, toEntity);
                }

                if (km.getEntity(fromEntity) == null) {
                        throw new EntityDoNotExist(fromEntity);
                } else if (km.getEntity(relationType) == null) {
                        throw new RelationTypeDoNotExist(relationType);
                } else if (km.getEntity(toEntity) == null) {
                        throw new EntityDoNotExist(toEntity);
                }

                try {
                        ((Entity) km.getEntity(fromEntity)).addRelation(
                                relationType,
                                (Entity) km.getEntity(toEntity),
                                strength);
                } catch (Exception ex) {
                }
        }

        /**
         * Adds a new case to the Knowledge Model of Creek. In able for this
         * process to be successful, the case must not exist.
         *
         * @param name the name of the new case to create
         * @param description the description of the new case
         *
         * @throws CaseAlreadyExist if a case with a equal name already exist in the
         * Knowledge Model
         */
        public void addCase(String name, String description)
                throws CaseAlreadyExist {
                if (km.getEntity(name) == null) {
                        // check the description
                        if (description == null) {
                                description = "";
                        }
                        try {
                                // also set the status to of the case to unsolved as default, this
                                // might be changed if the case got a solution, see also 'addHasSolution()'
                                 (new Case(km, name, description)).setStatus(Case.UNSOLVEDCASE);
                        } catch (Exception ex) {
                                throw new CaseAlreadyExist(name);
                        }
                } else {
                        throw new CaseAlreadyExist(name);
                }
        } // addCase

        /**
         * Adds a finding to an already existing case of the Knowledge Model of Creek.
         * In addition, the relation type and the entity specified must exist in order for
         * this process to be successful.
         *
         * @param caseName the name of the case to add a new finding
         * @param relationType the relation type for this relation. This relation type
         * should either be of HasFinding class or a subclass of this. There is not done
         * any checking to ensure that this is followed.
         * @param toEntity the name of the entity that is the finding of this case
         * @param importance the default importance of this finding, a value between 0 and 1.
         * If this value is not in between 0 and 1 then a default value will be used which is 0.5
         * @param predictiveStrength the default predictive strength of this finding, a value
         * between 0 and 1. If this value is not in between 0 and 1 then a default value will
         * be used which is 0.5
         *
         * @throws EqualRelationAlreadyExist if an equal relation already exist
         * @throws EntityDoNotExist if the entity of this finding does not exist
         * @throws RelationTypeDoNotExist if the relation type specified does not exist
         */
        public void addHasFinding(
                String caseName,
                String relationType,
                String toEntity,
                String importance,
                String predictiveStrength)
                throws EqualRelationAlreadyExist, EntityDoNotExist, RelationTypeDoNotExist {

                // now check the strengths
                double importanceVal = getImportance(importance);
                double predictiveStrengthVal = getPredictiveStrength(predictiveStrength);

                // test if have a relation already, if that is the case, do not add another one
                // this method should perhaps be a bit better, in that it replaces a relation
                // that is considered of less importance

                // if an equal relation already exist then throw a warning exception
                if (equalRelation(caseName, relationType, toEntity)) {
                        throw new EqualRelationAlreadyExist(caseName, toEntity);
                }

                if (km.getEntity(caseName) == null) {
                        throw new EntityDoNotExist(caseName);
                } else if (km.getEntity(relationType) == null) {
                        throw new RelationTypeDoNotExist(relationType);
                } else if (km.getEntity(toEntity) == null) {
                        throw new EntityDoNotExist(toEntity);
                }

                try {
                        new Case(km.getEntity(caseName)).addFinding(
                                relationType,
                                (Entity) km.getEntity(toEntity),
                                importanceVal,
                                predictiveStrengthVal);
                } catch (Exception ex) {
                }
        } // addHasFinding

        /**
         * Adds a solution to an already existing case of the Knowledge Model of Creek.
         * In addition, the entity specified must exist in order for this process to
         * be successful.
         *
         * @param caseName the name of the case to add a new finding
         * @param status the status of this solution, that is, SOLVED or UNSOLVED
         * @param toEntity the name of the entity that is the solution of this case
         * @param importance the default importance of this solution, a value between 0 and 1.
         * If this value is not in between 0 and 1 then a default value will be used which is 0.5
         * @param predictiveStrength the default predictive strength of this solution, a value
         * between 0 and 1. If this value is not in between 0 and 1 then a default value will
         * be used which is 0.5
         *
         * @throws EqualRelationAlreadyExist if an equal relation already exist
         * @throws EntityDoNotExist if the entity of this finding does not exist
         */
        public void addHasSolution(
                String caseName,
                String status,
                String toEntity,
                String importance,
                String predictiveStrength)
                throws EqualRelationAlreadyExist, EntityDoNotExist {

          try
          {

                // now check the strengths
                double importanceVal = getImportance(importance);
                double predictiveStrengthVal = getPredictiveStrength(predictiveStrength);

                // first do a general check independend of solved/unsolved status
                if (km.getEntity(caseName) == null) {
                        throw new EntityDoNotExist(caseName);
                }
                Case myCase = new Case(km.getEntity(caseName));


                // now check status value
                int statusVal;
                if (status.equalsIgnoreCase("solved")) {
                        statusVal = Case.SOLVEDCASE;
                } else {
                        statusVal = Case.UNSOLVEDCASE;
                }

                // now set the status
                myCase.setStatus(statusVal);

                if (toEntity == null) {
                        // only return since no solution to add, and this is valid
                        return;
                }
                // if an equal relation already exist then throw a warning exception
                if (equalRelation(caseName, "has solution", toEntity)) {
                        throw new EqualRelationAlreadyExist(caseName, toEntity);
                }
                if (km.getEntity(toEntity) == null) {
                        throw new EntityDoNotExist(toEntity);
                }

                // now add the solution
                try {
                        myCase.addFinding(
                                "has solution",
                                (Entity) km.getEntity(toEntity),
                                importanceVal,
                                predictiveStrengthVal);
                } catch (Exception ex) {
                        // do nothing since cheked for this already
                }
          }
          catch(jcreek.representation.EntityTypeNotMatchedException e)
          {
            e.printStackTrace();
          }
        } // addHasSolution

        /**
         * Performs an easy test to see if the two concepts already have a relation.
         *
         * @param from the name of the entity from where the relation come from
         * @param relationType the relation type these two entities could be using
         * @param to the name of the entity to where the the relation end
         *
         * @return if they do have an relation already it then returns true, otherwise false.
         */
        private boolean equalRelation(String from, String relationType, String to) {
                Relation[] relations = ((Entity) km.getEntity(from)).getRelations();
                for (int i = 0; i < relations.length; i++) {
                        Entity currEnt = (Entity) relations[i].getValue();
                        if ((currEnt.getName().equalsIgnoreCase(to))
                                && relations[i].getName().equalsIgnoreCase(relationType)) {
                                return true;
                        }
                }
                return false;
        } // haveRelation

        /**
         * Checks a possible strength value if it is a number in the range
         * between 0 and 1. If it is null or not a number then the default value
         * will be returned.
         *
         * @param strengthStr the strength value in textual form
         *
         * @return a value between 0 and 1 according to the conversion of the
         * strength string. If the strength string is not a number between 0 and
         * 1 it will return the default value which is 0.5.
         */
        private double getStrength(String strengthStr) {
                if (strengthStr != null) {
                        Double strengthDbl = new Double(strengthStr);
                        if (strengthDbl.isNaN()) {
                                return 0.5;
                        } else {
                                double strength = strengthDbl.doubleValue();
                                // now check the defaultStrength
                                if (strength < 0 || 1 < strength) {
                                        // then set to default
                                        return 0.5;
                                } else {
                                        return strength;
                                }
                        }
                } else {
                        return 0.5;
                }
        } // getStrength

        /**
         * Gets the number value of the importance string value. If the string value
         * is either not recognized or null, the default value will be returned.
         *
         * @param importance the string variant of the importance factor
         *
         * @return the double value of the string constant if found, otherwise will
         * return the default value.
         *
         * @see jcreek.representation.inference.ActivationConstants
         */
        private double getImportance(String importance) {
                if (importance != null) {
                        if (importance.equalsIgnoreCase("NECCESSARY")) {
                                return ActivationConstants.NECCESSARY;
                        } else if (importance.equalsIgnoreCase("CHARACTERISTIC")) {
                                return ActivationConstants.CHARACTERISTIC;
                        } else if (importance.equalsIgnoreCase("INFORMATIVE")) {
                                return ActivationConstants.INFORMATIVE;
                        } else if (importance.equalsIgnoreCase("IRRELEVANT")) {
                                return ActivationConstants.IRRELEVANT;
                        } else {
                                return ActivationConstants.DEFAULT_IMPORTANCE;
                        }
                } else {
                        return ActivationConstants.DEFAULT_IMPORTANCE;
                }
        } // getImportance

        /**
         * Gets the number value of the predictive strength string value. If the string value
         * is either not recognized or null, the default value will be returned.
         *
         * @param importance the string variant of the predictive strength factor
         *
         * @return the double value of the string constant if found, otherwise will
         * return the default value.
         *
         * @see jcreek.representation.inference.ActivationConstants
         */
        private double getPredictiveStrength(String predictiveStrength) {
                if (predictiveStrength != null) {
                        if (predictiveStrength.equalsIgnoreCase("SUFFICIENT")) {
                                return ActivationConstants.SUFFICIENT;
                        } else if (predictiveStrength.equalsIgnoreCase("STRONGLY_INDICATIVE")) {
                                return ActivationConstants.STRONGLY_INDICATIVE;
                        } else if (predictiveStrength.equalsIgnoreCase("INDICATIVE")) {
                                return ActivationConstants.INDICATIVE;
                        } else if (predictiveStrength.equalsIgnoreCase("SPURIOUS")) {
                                return ActivationConstants.SPURIOUS;
                        } else {
                                return ActivationConstants.DEFAULT_PREDICTIVE_STRENGTH;
                        }
                } else {
                        return ActivationConstants.DEFAULT_PREDICTIVE_STRENGTH;
                }
        } // getPredictiveStrength

        //******************************************
        // Exception Classes
        //******************************************

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to use a class descriptor that does not exist.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class ClassDescriptorDoNotExist extends Exception {
                public ClassDescriptorDoNotExist(String classDescriptor) {
                        super(
                                "[Fatal error] The class descriptor '"
                                        + classDescriptor
                                        + "' does not exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to create a new relation type with either a name
         * or inverse name that is already in use.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class RelationTypeNameAlreadyInUse extends Exception {
                public RelationTypeNameAlreadyInUse(String name) {
                        super(
                                "[Fatal error] The relation type name '"
                                        + name
                                        + "' is already in use.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to create a new relation type that already exist but
         * is not equal, that is, uses another class descriptor.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class RelationTypeAlreadyExist extends Exception {
                public RelationTypeAlreadyExist(String name, String inverse) {
                        super(
                                "[Fatal error] The relation type '"
                                        + name
                                        + "' with inverse '"
                                        + inverse
                                        + "' already exist but uses another class descriptor.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to use a relation type that does not exist.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class RelationTypeDoNotExist extends Exception {
                public RelationTypeDoNotExist(String relationType) {
                        super(
                                "[Fatal error] The relation type '"
                                        + relationType
                                        + "' does not exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to create a new relation type that already exist
         * and is equal, that is, uses the same class descriptor.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class EqualRelationTypeAlreadyExist extends Exception {
                public EqualRelationTypeAlreadyExist(String name, String inverse) {
                        super(
                                "[Warning] An equal relation type to '"
                                        + name
                                        + "' with inverse '"
                                        + inverse
                                        + "' already exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to add a relation between two entities and these
         * entities already have an equal relation between them.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class EqualRelationAlreadyExist extends Exception {
                public EqualRelationAlreadyExist(String from, String to) {
                        super(
                                "[Warning] An equal relation from entity '"
                                        + from
                                        + "' to  entity '"
                                        + to
                                        + "' already exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to create an entity that already exist.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class EntityAlreadyExist extends Exception {
                public EntityAlreadyExist(String name) {
                        super("[Warning] The entity '" + name + "' already exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to create a case that already exist.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class CaseAlreadyExist extends Exception {
                public CaseAlreadyExist(String name) {
                        super("[Warning] The case '" + name + "' already exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * an attempt to use an entity that does not exist.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class EntityDoNotExist extends Exception {
                public EntityDoNotExist(String name) {
                        super("[Fatal error] The entity '" + name + "' does not exist.");
                }
        }

        /**
         * An exception that is thrown in cases where there have been
         * a fault in the compilation of a new relation type. In most
         * cases this is due to that the tools.jar library file is not
         * specified in the classpath or a wrong version is in use.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        public class CompilationFault extends Exception {
                public CompilationFault() {
                        super(
                                "[Fatal error] A compilation error did occur. Please check"
                                        + " that 'tools.jar' is correctly set in the classpath.");
                }
        }
}
