package jcreek.cke.importexport.jcxml;

import java.awt.Frame;

import jcreek.cke.browser.SimpleFileFilter;
import jcreek.cke.importexport.ProcessControlInterface;
import jcreek.cke.importexport.ProcessStatusWindow;
import jcreek.representation.KnowledgeModel;
import jcreek.cke.visualizer.AppSettings;
import javax.swing.JFileChooser;

/**
 * This class is the one having the overall control of the export process of
 * the Knowledge Model of Creek to JCXML files.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class JCXMLexport extends Thread implements ProcessControlInterface {
	private ExportThread exportThread = null;
	private Frame parent = null;
	private KnowledgeModel km = null;
	private boolean exportDone = false;
	private boolean aborted = false;
	private String jcxmlFile = null;
	private ProcessStatusWindow statusWindow = null;
	private AppSettings props = null;

	/**
	 * Constructor for JCXMLexport, which does set some basic settings
	 * before the process of exporting the Knowledge Model may start.
	 *
	 * @param parent the parent frame window.
	 * @param km the JavaCreek knowledge model to export the model to.
	 */
	public JCXMLexport(Frame parent, KnowledgeModel km) {
		this.parent = parent;
		this.km = km;
		this.statusWindow =
			new ProcessStatusWindow(parent, "JCXML Export Status", this);
		props = AppSettings.instanceOf();
		props.load("Creek.ini");
	}

	/**
	 * Checks if the process was aborted;
	 *
	 * @return true if the porcess was aborted, otherwise false.
	 */
	public boolean isAborted() {
		return aborted;
	} // isAborted

	/**
	 * Starts the process by first showing the FileChooser dialog
	 * then starts the export process, by first starting the thread
	 * and gives feedback of the process to the JCXML status dialog.
	 */
	public void startExport() {
		JFileChooser fc =
			new JFileChooser(props.getString(JCXMLpreferences.LAST_DIRECTORY));
		fc.setMultiSelectionEnabled(false);
		fc.addChoosableFileFilter(
			new SimpleFileFilter(new String[] { "jcxml" }, "JCXML Files (*.jcxml)"));
		if (fc.showSaveDialog(parent) == JFileChooser.APPROVE_OPTION) {
			props.setString(
				JCXMLpreferences.LAST_DIRECTORY,
				fc.getSelectedFile().getPath());
			props.commit();
			jcxmlFile = fc.getSelectedFile().getAbsolutePath();
                        if(!jcxmlFile.toLowerCase().endsWith(".jcxml"))
                          jcxmlFile += ".jcxml";
			// now start the thread
			exportThread = new ExportThread();
			exportThread.start();
			// wait until done or aborted, update the status dialog while waiting
			// open the status dialog in the process
			statusWindow.show();
		}
	} // startExport

	/**
	 * Aborts the export process.
	 */
	public void abortProcess() {
		// check if not already been aborted
		if (exportThread != null) {
			// interrupt the thread
			Thread moribund = exportThread;
			exportThread = null;
			moribund.interrupt();
			aborted = true;
		}
	} // abortProcess

	//****************************************
	// Inner classes
	//****************************************
	/**
	 * This class contains the different stages of the export process. It has
	 * been put in a thread so that the user might have a more responsive GUI.
	 *
	 * @author Stein L. Tomassen, NTNU 2002
	 */
	protected class ExportThread extends Thread {
		/**
		 * Does a check if aborted before every new task
		 *
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			statusWindow.setCancelButton();
			if (jcxmlFile != null) {
				statusWindow.appendText(
					"Starts the creation of '" + jcxmlFile + "'...");
				statusWindow.appendText("Starts export of the knowledge model...");
				statusWindow.appendText(
					"-----------------------------------------------------");
				JCXMLexportParser creator =
					new JCXMLexportParser(
						jcxmlFile,
						km,
						props.getBoolean(JCXMLpreferences.ABORT_KEY),
						statusWindow);
				creator.startCreate();
				if (creator.isAbortOnFatalErrorsEnabled()
					&& creator.didFatalErrorOccur()) {
					aborted = true;
					statusWindow.appendText(
						"Export to JCXML was aborted because of a fatal error...");
				} else {
					creator.writeToDisk();
					statusWindow.appendText(
						"-----------------------------------------------------");
					statusWindow.appendText(
						"Export of the knowledge model to JCXML was successful...");
				}
			}
			// when finished
			statusWindow.setCloseButton();
			exportDone = true;
		} // run

		/**
		 * The default message given if the process was aborted.
		 */
		public void abortedMessage() {
			statusWindow.appendText("Export of JCXML was aborted...");
			statusWindow.setCloseButton();
		} // abortedMessage
	} // ExportThread
}