package jcreek.cke.importexport.jcxml;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Vector;
import jcreek.cke.importexport.xml.DOMdocumentSerializer;
import jcreek.cke.importexport.xml.DefaultGeneralParser;
import jcreek.cke.importexport.KnowledgeModelExportUtilities;
import jcreek.cke.importexport.ProcessStatusWindowInterface;
import jcreek.representation.cbr.Case;
import jcreek.representation.KnowledgeModel;
import jcreek.representation.Relation;
import jcreek.representation.Entity;
import jcreek.representation.RelationType;
import jcreek.representation.cbr.*;
import jcreek.representation.EntityTypeNotMatchedException;
import jcreek.cke.visualizer.AppSettings;

import org.apache.xerces.dom.TextImpl;
import org.apache.xerces.dom.DocumentImpl;
import org.apache.xerces.dom.DOMImplementationImpl;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

/**
 * Does the parsing of the Creek Knowledge Model and extracts
 * the complete model to JCXML.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class JCXMLexportParser extends DefaultGeneralParser {
	private boolean abortOnFatalErrors = false;
	private boolean fatalError = false;
	private String fileName;
	private Document document = null;
	private Node aboutRoot;
	private Node modelRoot;
	private KnowledgeModel km;
	private KnowledgeModelExportUtilities kmeu;
	private ProcessStatusWindowInterface statusWindow;
	private boolean includeDefault = false;

	/**
	 * Does set some basic settings before any parsing may begin.
	 *
	 * @param fileName the name and path of the output file
	 * @param km the Knowledge Model in Creek to export
	 * @param abortOnFatalErrors set to true if the export should be
	 * aborted in case of a fatal error, if not set to false
	 * @param statusWindow the window where the process status
	 * messages should be displayed
	 */
	public JCXMLexportParser(
		String fileName,
		KnowledgeModel km,
		boolean abortOnFatalErrors,
		ProcessStatusWindowInterface statusWindow) {
		this.fileName = fileName;
		this.statusWindow = statusWindow;
		this.km = km;
		this.abortOnFatalErrors = abortOnFatalErrors;
		kmeu = new KnowledgeModelExportUtilities(km);
		AppSettings props = AppSettings.instanceOf();
		props.load("Creek.ini");
		includeDefault = props.getBoolean(JCXMLpreferences.INCLUDE_DEFAULT);
	} // JCXMLexportParser

	/**
	 * Starts the creation of the JCXML document.
	 */
	public void startCreate() {
		try {
			createBasicDocument();
			handleAboutElement();
			handleNewRelationElement();
			handleEntityElement();
			handleRelationElement();
			handleCaseElement();
		} catch (Exception ex) {
			fatalError = true;
			statusWindow.appendText(ex.getMessage());
                        ex.printStackTrace();
		}
	} // startCreate

	/**
	 * Checks if abortOnFatalErrors is set or not.
	 *
	 * @return true if bortOnFatalErrors is enabled, otherwise false.
	 */
	public boolean isAbortOnFatalErrorsEnabled() {
		return abortOnFatalErrors;
	} // isAbortOnFatalErrorsEnabled

	/**
	 * Checks if an fatal error did occur.
	 *
	 * @return true if an fatal error did occur, otherwise false.
	 */
	public boolean didFatalErrorOccur() {
		return fatalError;
	} // didFatalErrorOccur

	/**
	 * Writes the DOM document to disk.
	 */
	public void writeToDisk() {
		DOMdocumentSerializer serializer = new DOMdocumentSerializer();
		serializer.serializeToFile(document, fileName);
	} // writeToDisk

	/**
	 * Creates a basic document. Must be run before trying to add any elements.
	 * Is run by the startCreate() method.
	 */
	protected void createBasicDocument() {
		// prepare the document
		document = new DocumentImpl();

		// create root element
		Element root = document.createElement("jcreekmodel");
		// set JCXML Schema default values
		root.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
		root.setAttribute("xsi:noNamespaceSchemaLocation", "jcreekmodel.xsd");
		// now add root to Document
		document.appendChild(root);

		// the about element
		Element about = document.createElement("about");
		// set aboutRoot for easier access
		aboutRoot = about;
		// now append the about element to the root
		root.appendChild(about);

		// the model element
		Element model = document.createElement("model");
		// set modelRoot for easier access
		modelRoot = model;
		// now append the model element to the root
		root.appendChild(model);
	} // createBasicDocument

	/**
	 * Does the handling of the general information about the knowledge model.
	 * Does only insert the data that is available, and that is basically the
	 * name of the model. Does also insert the date that this model was exported.
	 */
	protected void handleAboutElement() {
		/**************************************
		<about>
			<title>Car Example</title>
			<name>CarExample</name>
			<creator>Stein L. Tomassen</creator>
			<date>2002-05-23</date>
			<description>A car example knowledge model</description>
			<language>en</language>
		</about>
		***************************************/
		statusWindow.appendText("******************************************");

		// title element
		String title = km.getTitle();
		if (kmeu.hasLength(title)) {
			Element titleElem = document.createElement("title");
			titleElem.appendChild(document.createTextNode(title));
			aboutRoot.appendChild(titleElem);

			statusWindow.appendText("Title: " + title);
		} else {
			// could ask for this data...
		}

		// name element
		String name = km.getName();
		// should at least always be something... :-)
		Element nameElem = document.createElement("name");
		nameElem.appendChild(document.createTextNode(name));
		aboutRoot.appendChild(nameElem);
		statusWindow.appendText("Name: " + name);

		// creater element
		// could ask for this data...

		// date element
		String date = (new SimpleDateFormat("yyyy-MM-dd")).format(new Date());
		Element dateElem = document.createElement("date");
		dateElem.appendChild(document.createTextNode(date));
		aboutRoot.appendChild(dateElem);
		statusWindow.appendText("Date: " + date);

		// description element
		String description = km.getDescription();
		if (kmeu.hasLength(description)) {
			Element descriptionElem = document.createElement("description");
			descriptionElem.appendChild(document.createTextNode(description));
			aboutRoot.appendChild(descriptionElem);
			statusWindow.appendText("Description: " + description);
		} else {
			// could ask for this data...
		}

		// language element
		// could be asked for...

		statusWindow.appendText("******************************************");
	} // handleAboutElement

	/**
	 * Handels the relations that are an extension to the isopod model, that
	 * is new relations. Only the new relations are added to the JCXML. If
	 * a value collected is not valid, i.e. strength is out of range or e.g.
	 * no description, then these fields are not added to the JCXML output.
	 */
	protected void handleNewRelationElement() {
		/**************************************
		<newRelationType classDescriptor="has finding" name="has colour"
		 inverseName="colour of" defaultStrength="0.5"
		 description="describes the color of an entity"/>
		***************************************/
		// 'defaultStrength' and 'description' are not mandatory

		// container for relations already handeled
		Vector relationsVect = new Vector();

/*		Concept[] conceptList =
			km.getClassDescriptor(Entity.class).getInstances();

		for (int i = 0; i < conceptList.length; i++) {
			Relation[] relationList = ((Entity) conceptList[i]).getRelations();
			for (int ii = 0; ii < relationList.length; ii++) {
				if (relationList[ii].getClassDescriptor() != null) { */

                Entity[] rcds = km.getRelationType("Relation").getAllSubclasses();
                Entity[] descriptors = new Entity[rcds.length+1];
                descriptors[0] = km.getRelationType("Relation");
                System.arraycopy(rcds, 0, descriptors, 1, rcds.length);
                for(int i=0;i<descriptors.length;i++)
                {
                  try
                  {
                      RelationType current = new RelationType(descriptors[i]);
					// collect relation data
					String name = descriptors[i].getName();
					String classDescriptor;
                                        if(current.getImmediateSuperclasses().length == 0)
                                        {
                                          System.out.println("relationtype "+name+ " has no superclass");
                                          classDescriptor = null;
                                        }
                                        else
                                          classDescriptor = current.getImmediateSuperclasses()[0].getName();
					String inverseName =
						current.getInverseName();
					String defaultStrength = ""+current.getExplanationStrength();
//						kmeu.getStrength
//							km.getRelationType(name).getExplanationStrength(),
//							includeDefault);
					String description =
						descriptors[i].getDescription();

					// check if already been added or a standard relation
					// seems like all ISOPOD relations ends with 'Relation', so test for that string
					if (!(relationsVect.contains(name)
						|| relationsVect.contains(inverseName))
//						&& !classDescriptor.endsWith("Relation")
                                                ) {
						// create the element
						Element relationType = document.createElement("newRelationType");
						relationType.setAttribute("classDescriptor", classDescriptor);
						relationType.setAttribute("name", name);
						relationType.setAttribute("inverseName", inverseName);
						// check if include. If null then do not include
						if (defaultStrength != null)
							relationType.setAttribute("defaultStrength", defaultStrength);
						// check description
						if (kmeu.hasLength(description))
							relationType.setAttribute("description", description);
						// add the element to the model element
						modelRoot.appendChild(relationType);

						// report what has been done
						statusWindow.appendText(
							"Added a Relation Type => '"
								+ name
								+ "' <"
								+ classDescriptor
								+ "> '"
								+ inverseName
								+ "' : ["
								+ defaultStrength
								+ "] : {"
								+ description
								+ "}");

						// now add the names to the vector of hadeled relations
						relationsVect.add(name);
						relationsVect.add(inverseName);
					}
                  }
                  catch(EntityTypeNotMatchedException e)
                  {
                    e.printStackTrace();
                  }
                }


	} // handleNewRelationElement

	/**
	 * Does the handling of the entities (concepts) of the knowledge
	 * model. If the entity has a description then that is added as well.
	 */
	protected void handleEntityElement() {
		/**************************************
		<entity name="engine status" description="engine status"/>
		***************************************/
		// 'description' is not mandatory

		Entity[] conceptList =
			km.getEntities();

		for (int i = 0; i < conceptList.length; i++) {
			if (!Case.matches(conceptList[i])) {
				// collect entity data
                                Entity fromEntity = ((Entity) conceptList[i]);
				String name = conceptList[i].getName();
				String description = conceptList[i].getDescription();

				// create the element
				Element entity = document.createElement("entity");
				entity.setAttribute("name", name);

                                if(fromEntity.getEntityObject() != null)
                                {
                                  entity.setAttribute("valueClass", fromEntity.getEntityObject().getClass().getName());
                                  entity.setAttribute("value",fromEntity.getEntityObject().toString());
                                }


				// check description
				if (kmeu.hasLength(description)) {
					entity.setAttribute("description", description);
				}
				// add the element to the model element
				modelRoot.appendChild(entity);

				// report what has been done
				statusWindow.appendText(
					"Added a Entity => '" + name + "' : {" + description + "}");

/*
                                Relation[] relationList = fromEntity.getLocalRelations("Relation");
                                for (int ii = 0; ii < relationList.length; ii++) {
                                        // do not hadel cases here!!!
                                        if (!relationList[ii].getValue().getClass().equals(Case.class)) {
                                                // collect the values
                                                String relationType = relationList[ii].getName();
                                                String toEntity = ((Entity) relationList[ii].getValue()).getName();
                                                String strength =
                                                        kmeu.getStrength(
                                                                relationList[ii].getExplanationStrength(),
                                                                includeDefault);

                                                // check if this triple is not already added and also sees to it
                                                // that relations that where inherited is not added as well
                                                if (
//                                                    !relationsVect.contains(fromEntity + relationType + toEntity) &&
                                                        !relationList[ii].isInherited()) {
                                                        // create the element
                                                        Element relation = document.createElement("relation");
                                                        relation.setAttribute("fromEntity", fromEntity.getName());
                                                        relation.setAttribute("relationType", relationType);
                                                        relation.setAttribute("toEntity", toEntity);
                                                        // check if include. If null then do not include
                                                        if (strength != null)
                                                                relation.setAttribute("strength", strength);
                                                        // add the element to the model element
                                                        modelRoot.appendChild(relation);

                                                        // add the relation to the vector
                                                        // first the normal unique triple
//                                                        relationsVect.add(fromEntity + relationType + toEntity);
                                                        // then the inverse unique triple
//                                                        relationsVect.add(
//                                                                toEntity + relationList[ii].getInverseName() + fromEntity);

                                                        // report what has been done
                                                        statusWindow.appendText(
                                                                "Added a Relation => '"
                                                                        + fromEntity
                                                                        + "' <"
                                                                        + relationType
                                                                        + "> '"
                                                                        + toEntity
                                                                        + "' : ["
                                                                        + strength
                                                                        + "]");
                                                }
                                        }
				}
*/

			}
		}
	} // handleEntityElement

	/**
	 * Does the handling of the relations among the entities. Does either
	 * add the <i>normal-way</i> or the <i>inverse-way</i> of the relation,
	 * not both. If the relation has a valid strength, then that value is
	 * added as well.
	 * <p>
	 * Does not add relations that are not directly connected between two
	 * entities.
	 */
	protected void handleRelationElement() {
		/**************************************
		<relation fromEntity="battery status" relationType="has instance" toEntity="OK battery" strength="0.5"/>
		***************************************/
		// 'strength' is not mandatory

		Vector relationsVect = new Vector();

		Entity[] conceptList = km.getEntities();

		for (int i = 0; i < conceptList.length; i++) {
			if (!Case.matches(conceptList[i])) {
				String fromEntity = conceptList[i].getName();
				// change the belove to contain relations from KM
				Relation[] relationList =
					((Entity) conceptList[i]).getLocalRelations();
				for (int ii = 0; ii < relationList.length; ii++) {
					// do not hadel cases here!!!
					if (!Case.matches(relationList[ii].getValue())) {
						// collect the values
						String relationType = relationList[ii].getName();
						String toEntity = ((Entity) relationList[ii].getValue()).getName();
						String strength =
							kmeu.getStrength(
								relationList[ii].getExplanationStrength(),
								includeDefault);

						// check if this triple is not already added and also sees to it
						// that relations that where inherited is not added as well
						if (!relationsVect.contains(fromEntity + relationType + toEntity)
							&& !relationList[ii].isInherited()) {
							// create the element
							Element relation = document.createElement("relation");
							relation.setAttribute("fromEntity", fromEntity);
							relation.setAttribute("relationType", relationType);
							relation.setAttribute("toEntity", toEntity);
							// check if include. If null then do not include
							if (strength != null)
								relation.setAttribute("strength", strength);
							// add the element to the model element
							modelRoot.appendChild(relation);

							// add the relation to the vector
							// first the normal unique triple
							relationsVect.add(fromEntity + relationType + toEntity);
							// then the inverse unique triple
							relationsVect.add(
								toEntity + relationList[ii].getInverseName() + fromEntity);

							// report what has been done
							statusWindow.appendText(
								"Added a Relation => '"
									+ fromEntity
									+ "' <"
									+ relationType
									+ "> '"
									+ toEntity
									+ "' : ["
									+ strength
									+ "]");
						}
					}
				}
			}
		}
	} // handleNewRelationElement

	/**
	 * Does the handling of the cases.
	 * <p>
	 * Does in the same process take care of the findings and the possible solution
	 * of the case as well. If the case doesn't have a solution then no solution will
	 * be added. A solution might either have as status 'solved' or 'unsolved'. If
	 * the case got a description then that is added as well. If the 'importance' or
	 * the 'predictiveStrength' values of the findings or the solution then these values
	 * are added as well.
	 */
	protected void handleCaseElement() {
		/**************************************
		<case name="Car Case #1">
			<finding relationType="has colour" toEntity="red" importance="0.1" predictiveStrength="0.5"/>
			<finding relationType="has engine status" toEntity="engine not starting" importance="0.5"/>
			<finding relationType="has battery status" toEntity="empty battery" predictiveStrength="0.8"/>
			<solution status="solved" toEntity="good" predictiveStrength="0.8"/>
		</case>
		***************************************/
		// 'finding', 'importance', 'predictiveStrength', 'solution', 'solution:toEntity',
		// 'solution:predictiveStrength', and 'solution:importance' are not mandatory
		// 'status' might either be 'solved' or 'unsolved', if no solution element then the case
		// is in 'processing' state

		Entity[] conceptList = km.getEntities();

		for (int i = 0; i < conceptList.length; i++) {
			if (Case.matches(conceptList[i])) {
				// get the case data
                              try
                              {
                                Case currentCase = new Case(conceptList[i]);
				String name = currentCase.getName();
				String description = currentCase.getDescription();

				// create the case element
				Element caseElement = document.createElement("case");
				caseElement.setAttribute("name", name);

				// check description
				if (kmeu.hasLength(description)) {
					caseElement.setAttribute("description", description);
				}
				// add the element to the model element
				modelRoot.appendChild(caseElement);

				// report what has been done
				statusWindow.appendText(
					"Added a Case => '" + name + "' : {" + description + "}");

				// now get the findings
				Relation[] findingsList =
					currentCase.getRelations();
				// take also hold on the solution if found
				Relation solutionRelation = null;
				for (int ii = 0; ii < findingsList.length; ii++) {
					if (!findingsList[ii].getRelationType().equals(km.getRelationType(CBRModel.HAS_SOLUTION))) {
						// collect the finding data
						String relationType = findingsList[ii].getName();
						String toEntity = ((Entity) findingsList[ii].getValue()).getName();
						String importance =
							kmeu.getImportance(
								findingsList[ii].getExplanationStrength(),
								includeDefault);
						String predictiveStrength =
							kmeu.getPredictiveStrength(
								findingsList[ii].getInverse().getExplanationStrength(),
								includeDefault);

						// create the element
						Element finding = document.createElement("finding");
						finding.setAttribute("relationType", relationType);
						finding.setAttribute("toEntity", toEntity);
						// check if include. If null then do not include
						if (importance != null)
							finding.setAttribute("importance", importance);
						// check if include. If null then do not include
						if (predictiveStrength != null)
							finding.setAttribute("predictiveStrength", predictiveStrength);

						// add the element to the case element
						caseElement.appendChild(finding);

						// report what has been done
						statusWindow.appendText(
							"=> Added a Finding => '"
								+ name
								+ "' <"
								+ relationType
								+ "> '"
								+ toEntity
								+ "' : ["
								+ importance
								+ " | "
								+ predictiveStrength
								+ "]");
					} else {
						// take hold on the solution if found
						solutionRelation = findingsList[ii];
					}
				}

				// check if the case got a solution
				if (solutionRelation != null) {
					// collect the solution data
					String status;
					if (currentCase.getStatus() == Case.SOLVEDCASE) {
						status = "solved";
					} else {
						status = "unsolved";
					}
					String toEntity = ((Entity) solutionRelation.getValue()).getName();
					String importance =
						kmeu.getImportance(
							solutionRelation.getExplanationStrength(),
							includeDefault);
					String predictiveStrength =
						kmeu.getPredictiveStrength(
							solutionRelation.getInverse().getExplanationStrength(),
							includeDefault);

					// create the element
					Element solution = document.createElement("solution");
					solution.setAttribute("status", status);
					solution.setAttribute("toEntity", toEntity);
					// check if include. If null then do not include
					if (importance != null)
						solution.setAttribute("importance", importance);
					// check if include. If null then do not include
					if (predictiveStrength != null)
						solution.setAttribute("predictiveStrength", predictiveStrength);

					// add the element to the case element
					caseElement.appendChild(solution);

					// report what has been done
					statusWindow.appendText(
						"=> Added a Solution => '"
							+ name
							+ "' <-> '"
							+ toEntity
							+ "' : ["
							+ importance
							+ " | "
							+ predictiveStrength
							+ " | "
							+ status
							+ "]");
				}
                              }
                              catch(EntityTypeNotMatchedException e)
                              {
                                // Should never happen.
                                e.printStackTrace();
                              }

			}
		}
	} // handleCaseElement
} // JCXMLexportParser
