package jcreek.cke.importexport.jcxml;

import java.awt.Frame;

import jcreek.cke.browser.SimpleFileFilter;
import jcreek.cke.util.FileUtilities;
import jcreek.cke.importexport.xml.XMLvalidater;
import jcreek.cke.importexport.ProcessControlInterface;
import jcreek.cke.importexport.ProcessStatusWindow;
import jcreek.representation.KnowledgeModel;
import jcreek.cke.visualizer.AppSettings;
import javax.swing.JFileChooser;
import java.io.StringReader;
import java.io.InputStreamReader;
import java.io.Reader;

/**
 * This class is the one having the overall control of the import process of
 * JCXML files to the Knowledge Model of Creek.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class JCXMLimport extends Thread implements ProcessControlInterface {
        private ImportThread importThread = null;
        private Frame parent = null;
        private KnowledgeModel km = null;
        private boolean importDone = false;
        private boolean aborted = false;
        private String jcxmlFile = null;
        private ProcessStatusWindow statusWindow = null;
        private AppSettings props = null;

        /**
         * Constructor for JCXMLimport, which sets some basic settings.
         *
         * @param parent the parent frame window.
         * @param km the JavaCreek knowledge model to import the model to.
         */
        public JCXMLimport(Frame parent, KnowledgeModel km) {
                this.parent = parent;
                this.km = km;
                this.statusWindow =
                        new ProcessStatusWindow(parent, "JCXML Import Status", this);
                props = AppSettings.instanceOf();
                props.load("Creek.ini");
        } // JCXMLimport

        /**
         * Checks if the process was aborted;
         *
         * @return true if the porcess was aborted, otherwise false.
         */
        public boolean isAborted() {
                return aborted;
        }

        /**
         * Starts the process by first showing the FileChooser dialog
         * then starts the import process, by first starting the thread
         * and gives feedback of the process to the JCXML status dialog.
         */
        public void startImport() {
                JFileChooser fc =
                        new JFileChooser(props.getString(JCXMLpreferences.LAST_DIRECTORY));
                fc.setMultiSelectionEnabled(false);
                fc.addChoosableFileFilter(
                        new SimpleFileFilter(new String[] { "jcxml" }, "JCXML Files (*.jcxml)"));
                if (fc.showOpenDialog(parent) == JFileChooser.APPROVE_OPTION) {
                        props.setString(
                                JCXMLpreferences.LAST_DIRECTORY,
                                fc.getSelectedFile().getPath());
                        props.commit();
                        jcxmlFile = fc.getSelectedFile().getAbsolutePath();
                        // now start the thread
                        importThread = new ImportThread();
                        importThread.start();
                        // wait until done or aborted, update the status dialog while waiting
                        // open the status dialog in the process
                        statusWindow.show();
                }
        }

        /**
         * Aborts the import process. This is the implementation of the
         * method from 'ProcessControlInterface'. This method could be used
         * by other classes to abort a process.
         */
        public void abortProcess() {
          // check if not already been aborted
                if (importThread != null) {
                        // interrupt the thread
                        importThread = null;
                        aborted = true;
                }

        } // abortProcess

        //****************************************
        // Inner classes
        //****************************************
        /**
         * This class contains the different stages of the import process. It has
         * been put in a thread so that the user might have a more responsive GUI.
         *
         * @author Stein L. Tomassen, NTNU 2002
         */
        protected class ImportThread extends Thread {
                /**
                 * Does a check if aborted before every new task
                 *
                 * @see java.lang.Runnable#run()
                 */
                public void run() {
                        statusWindow.setCancelButton();
                        if (jcxmlFile != null) {
                                // satt slik forelpig
                                String schemaPath = props.getString(JCXMLpreferences.SCHEMA_KEY);
                                XMLvalidater validator;

                                if (aborted) {
                                        abortedMessage();
                                        return;
                                        }

                                Reader schemaSource;
                                if (!FileUtilities.exists(schemaPath))
                                {
                                  schemaPath = getClass().getClassLoader().getResource("jcreek/cke/resources/jcreekmodel.xsd").toExternalForm();
                                }

                                if(schemaPath != null && schemaPath.length() > 0)
                                {
                                        statusWindow.appendText(
                                                "Starts the validation of '"
                                                        + jcxmlFile
                                                        + "',\n"
                                                        + "this might take several minutes...");
                                        validator = new XMLvalidater(jcxmlFile, schemaPath);

                                        if (validator.isValid()) {
                                                statusWindow.appendText("Is valid...");
                                                if (aborted) {
                                                        abortedMessage();
                                                        return;
                                                }
                                                statusWindow.appendText("Starts import of the knowledge model...");
                                                statusWindow.appendText(
                                                        "-----------------------------------------------------");
                                                JCXMLimportParser trav =
                                                        new JCXMLimportParser(
                                                                validator.getDocument(),
                                                                km,
                                                                props.getBoolean(JCXMLpreferences.ABORT_KEY),
                                                                statusWindow);
                                                trav.startTraverse();
                                                if (trav.isAbortOnFatalErrorsEnabled()
                                                        && trav.didFatalErrorOccur()) {
                                                        aborted = true;
                                                        statusWindow.appendText(
                                                                "Import of JCXML was aborted because of a fatal error...");
                                                } else {
                                                        statusWindow.appendText(
                                                                "-----------------------------------------------------");
                                                        statusWindow.appendText("Import of JCXML successful...");
                                                }
                                        } else {
                                                statusWindow.appendText(validator.getAllFaultMsg().trim());
                                                statusWindow.appendText("Is not valid...");
                                        }
                                } else {
                                        aborted = true;
                                        statusWindow.appendText(
                                                "'jcreekmodel.xsd' was not found. Please set the correct\n"
                                                        + "settings in the Preferences of this JCXML Import Wizard");
                                }
                        }
                        // when finished
                        statusWindow.setCloseButton();
                        importDone = true;
                } // run

                /**
                 * The default message given if the process was aborted.
                 */
                public void abortedMessage() {
                        statusWindow.appendText("Import of JCXML was aborted...");
                        statusWindow.setCloseButton();
                } // abortedMessage
        } // ImportThread
}
