package jcreek.cke.importexport.jcxml;

import jcreek.cke.importexport.xml.DOMutilities;
import jcreek.cke.importexport.KnowledgeModelImportUtilities;
import jcreek.cke.importexport.ProcessStatusWindowInterface;
import jcreek.representation.KnowledgeModel;

import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Takes as input a DOM Document of a JCXML XML file that is assumed valid,
 * and traverses that document. Finds tags that are valid according to the
 * Schema of JCXML and constructs a Knowledge Model of the content of the
 * JCXML. if the abortOnFatalErrors is set to true the process will be
 * aborted if a fatal error should occur.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class JCXMLimportParser {
	private boolean abortOnFatalErrors = false;
	private boolean fatalError = false;
	private Document document = null;
	private KnowledgeModel km;
	private KnowledgeModelImportUtilities kmiu;
	private ProcessStatusWindowInterface statusWindow;

	/**
	 * Constructor that takes as input a DOM Document of a JCXML XML
	 * file that is assumed valid, and traverses that document. Finds tags
	 * that are valid according to the Schema of JCXML and constructs a
	 * Knowledge Model of the content of the JCXML. if the abortOnFatalErrors
	 * is set to true the process will be aborted if a fatal error should occur.
	 *
	 * @param document the DOM document containing the JavaCreek Knowledge Model.
	 * @param km the knowledge model to build
	 * @param abortOnFatalErrors abort if a fatal error occurs or just continue?
	 * @param statusWindow the window where the process information should be displayed
	 */
	public JCXMLimportParser(
		Document document,
		KnowledgeModel km,
		boolean abortOnFatalErrors,
		ProcessStatusWindowInterface statusWindow) {
		this.statusWindow = statusWindow;
		this.km = km;
		this.document = document;
		this.abortOnFatalErrors = abortOnFatalErrors;
		kmiu = new KnowledgeModelImportUtilities(km);
	} // JCXMLimportParser

	/**
	 * Starts the traversal of the DOM document.
	 */
	public void startTraverse() {
		try {
			traverse(document);
		} catch (Exception ex) {
			fatalError = true;
			statusWindow.appendText(ex.getMessage());
                        ex.printStackTrace();
		}
	} // startTraverse

	/**
	 * Checks if abortOnFatalErrors is enabled or not.
	 *
	 * @return true if bortOnFatalErrors is enabled, otherwise false.
	 */
	public boolean isAbortOnFatalErrorsEnabled() {
		return abortOnFatalErrors;
	} // isAbortOnFatalErrorsEnabled

	/**
	 * Checks if an fatal error did occur.
	 *
	 * @return true if an fatal error did occur, otherwise false.
	 */
	public boolean didFatalErrorOccur() {
		return fatalError;
	} // didFatalErrorOccur

	/**
	 * A recursive method that traverses the Document and locates the
	 * different Elements needed to collect the knowledge model
	 * contained in the XML file.
	 *
	 * @param node the node with childs to traverse. If the node is null
	 * no traversal will be done.
	 */
	private void traverse(Node node) throws Exception {
		// is there anything to do?
		if ((node == null) && (fatalError && abortOnFatalErrors)) {
			return;
		}

                Vector entityList = new Vector();
                Vector relationList = new Vector();
                Vector caseList = new Vector();

		int type = node.getNodeType();
		switch (type) {
			// handle Document node, that is only to traverse the childs...
			case Node.DOCUMENT_NODE :
				{
					traverse(((Document) node).getDocumentElement());
					break;
				} // case Node.DOCUMENT_NODE :
				// handle the Element node, this is where the main part is
			case Node.ELEMENT_NODE :
				{
					// handle the About element
					if (node.getNodeName().equalsIgnoreCase("about")) {
						handleAboutElement(node);
						// insert a return here so that the child elements of
						// the About element isn't handled more than once.
						return;
					} else if (node.getNodeName().equalsIgnoreCase("newRelationType")) {
						handleNewRelationElement(node);
						return;
					} else if (node.getNodeName().equalsIgnoreCase("entity")) {
						if (!fatalError) {
//                                                  entityList.add(node);
							handleEntityElement(node);
						}
						return;
					} else if (node.getNodeName().equalsIgnoreCase("relation")) {
						if (!fatalError) {
//                                                  relationList.add(node);
							handleRelationElement(node);
						}
						return;
					} else if (node.getNodeName().equalsIgnoreCase("case")) {
						if (!fatalError) {
//                                                  caseList.add(node);
							handleCaseElement(node);
						}
						// insert a return here so that the child elements of
						// the Case element isn't handled more than once.
						return;
					} else {
						// Elements here might eg. be 'jcreekmodel', 'model', ...
						Node child = node.getFirstChild();
						while ((child != null) && !(fatalError && abortOnFatalErrors)) {
							traverse(child);
							child = child.getNextSibling();
						}
					}
				} // case Node.ELEMENT_NODE
		} // switch(type)
/*
                for(int i=0;i<entityList.size();i++)
                  handleEntityElement((Node) entityList.elementAt(i));

                for(int i=0;i<caseList.size();i++)
                  handleCaseElement((Node) entityList.elementAt(i));

                for(int i=0;i<entityList.size();i++)
                  handleRelationElement((Node) entityList.elementAt(i));
*/
	} // traverse

	/**
	 * Does the handling of the general information about the knowledge model.
	 * <p>
	 * The About element contains all the basic information of the
	 * Knowledge Model in this JCXML file.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleAboutElement(Node node) {
		/**************************************
		<about>
			<title>Car Example</title>
			<name>CarExample</name>
			<creator>Stein L. Tomassen</creator>
			<date>2002-05-23</date>
			<description>A car example knowledge model</description>
			<language>en</language>
		</about>
		***************************************/
		statusWindow.appendText("******************************************");
		Node child = node.getFirstChild();
		while (child != null) {
			if (child.getNodeType() == Node.ELEMENT_NODE) {
				Node grandChild = child.getFirstChild();
				while (grandChild != null) {
					if (grandChild.getNodeType() == Node.TEXT_NODE) {
						if (child.getNodeName().equalsIgnoreCase("title")) {
                                                  System.out.println("km: "+km+ "gc: "+grandChild);
							km.setTitle(grandChild.getNodeValue());
							statusWindow.appendText("Title: " + grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("name")) {
							km.setName(grandChild.getNodeValue());
							statusWindow.appendText("Name: " + grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("creator")) {
							statusWindow.appendText("Creator: " + grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("date")) {
							statusWindow.appendText("Date: " + grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("description")) {
							km.setDescription(grandChild.getNodeValue());
							statusWindow.appendText(
								"Descrption: " + grandChild.getNodeValue());
						} else if (child.getNodeName().equalsIgnoreCase("language")) {
							statusWindow.appendText("Language: " + grandChild.getNodeValue());
						}
					}
					grandChild = grandChild.getNextSibling();
				}
			} // if (child.getNodeType()...
			child = child.getNextSibling();
		}
		statusWindow.appendText("******************************************");
	} // handleAboutElement

	/**
	 * Does the handling of new relations.
	 * <p>
	 * If a 'classDescriptor' does not exists, it then displays a fatal error message
	 * in the process output view. What it could have done is to put the 'new relation'
	 * aside for handling later on, but since all new relations are a subclass of the
	 * 'basic relations' and the new relations will not add any new functionality then
	 * I don't quite see the point. Instead there is a demand that all new relations
	 * should be a direct subclass of the 'basic relations.
	 * <p>
	 * If a relation already exists then the method will display a warning that the
	 * relation already exists.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleNewRelationElement(Node node) {
		/**************************************
		<newRelationType classDescriptor="has finding" name="has colour"
		 inverseName="colour of" defaultStrength="0.5"
		 description="describes the color of an entity"/>
		***************************************/
		// 'defaultStrength' and 'description' are not mandatory

		String classDescriptor =
			DOMutilities.getAttributeValueByName(
				"classDescriptor",
				node.getAttributes());
		String name =
			DOMutilities.getAttributeValueByName("name", node.getAttributes());
		String inverseName =
			DOMutilities.getAttributeValueByName("inverseName", node.getAttributes());
		String defaultStrength =
			DOMutilities.getAttributeValueByName(
				"defaultStrength",
				node.getAttributes());
		String description =
			DOMutilities.getAttributeValueByName("description", node.getAttributes());

		// now add the new relation
		try {
			kmiu.addNewRelationType(
				classDescriptor,
				name,
				inverseName,
				defaultStrength,
				description);
			statusWindow.appendText(
				"Added a Relation Type => '"
					+ name
					+ "' <"
					+ classDescriptor
					+ "> '"
					+ inverseName
					+ "' : ["
					+ defaultStrength
					+ "] : {"
					+ description
					+ "}");
		} catch (KnowledgeModelImportUtilities.CompilationFault ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
			statusWindow.appendText(ex.getMessage());
		} catch (KnowledgeModelImportUtilities.EqualRelationTypeAlreadyExist ex) {
			statusWindow.appendText(ex.getMessage());
		} catch (KnowledgeModelImportUtilities.ClassDescriptorDoNotExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
			statusWindow.appendText(ex.getMessage());
		} catch (KnowledgeModelImportUtilities.RelationTypeAlreadyExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
			statusWindow.appendText(ex.getMessage());
		} catch (KnowledgeModelImportUtilities.RelationTypeNameAlreadyInUse ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
			statusWindow.appendText(ex.getMessage());
		}
	} // handleNewRelationElement

	/**
	 * Does the handling of the entities.
	 * <p>
	 * If an entity already exists a warning message will be displayed.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleEntityElement(Node node) {
		/**************************************
		<entity name="engine status" description="engine status"/>
		***************************************/
		// 'description' is not mandatory
		String name =
			DOMutilities.getAttributeValueByName("name", node.getAttributes());
		String description =
			DOMutilities.getAttributeValueByName("description", node.getAttributes());
                String valueClass =
			DOMutilities.getAttributeValueByName("valueClass", node.getAttributes());
                String value =
			DOMutilities.getAttributeValueByName("value", node.getAttributes());

		// now add the new entity
		try {
			kmiu.addEntity(name, description, valueClass, value);
			statusWindow.appendText(
				"Added a Entity => '" + name + "' : {" + description + "}");
		} catch (KnowledgeModelImportUtilities.EntityAlreadyExist ex) {
			statusWindow.appendText(ex.getMessage());
		}
	} // handleEntityElement

	/**
	 * Does the handling of the relations among the entities.
	 * <p>
	 * If a relation already exist a warning message will be displayed. If
	 * the relation refers to an entity or a relation type that does not
	 * exist a fatal error message will be displayed.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleRelationElement(Node node) {
		/**************************************
		<relation fromEntity="battery status" relationType="has instance" toEntity="OK battery" strength="0.5"/>
		***************************************/
		// 'strength' is not mandatory
		String fromEntity =
			DOMutilities.getAttributeValueByName("fromEntity", node.getAttributes());
		String relationType =
			DOMutilities.getAttributeValueByName(
				"relationType",
				node.getAttributes());
		String toEntity =
			DOMutilities.getAttributeValueByName("toEntity", node.getAttributes());
		String strength =
			DOMutilities.getAttributeValueByName("strength", node.getAttributes());

		try {
			kmiu.addRelation(fromEntity, relationType, toEntity, strength);
			statusWindow.appendText(
				"Added a Relation => '"
					+ fromEntity
					+ "' <"
					+ relationType
					+ "> '"
					+ toEntity
					+ "' : ["
					+ strength
					+ "]");
		} catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
			statusWindow.appendText(ex.getMessage());
		} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
			statusWindow.appendText(ex.getMessage());
		} catch (KnowledgeModelImportUtilities.RelationTypeDoNotExist ex) {
			if (abortOnFatalErrors) {
				fatalError = true;
			}
			statusWindow.appendText(ex.getMessage());
		}

	} // handleNewRelationElement

	/**
	 * Does the handling of the cases.
	 * <p>
	 * Does in the same process take care of the findings and the possible solution
	 * of the case as well. If the case doesn't have a solution the status of the
	 * case will be set to 'proccessed'.
	 * <p>
	 * Displayes a warning message if the case already exist. Does also display a
	 * warning message if a finding or solution already exist. If the finding or
	 * solution refers to an entity or a relation type that does not exist a fatal
	 * error message will be displayed.
	 *
	 * @param node the node containing the element being analysed.
	 */
	private void handleCaseElement(Node node) {
		//		statusWindow.appendText("Handling Case element...");
		/**************************************
		<case name="Car Case #1">
			<finding relationType="has colour" toEntity="red" importance="0.1" predictiveStrength="0.5"/>
			<finding relationType="has engine status" toEntity="engine not starting" importance="0.5"/>
			<finding relationType="has battery status" toEntity="empty battery" predictiveStrength="0.8"/>
			<solution status="solved" toEntity="good" predictiveStrength="0.8"/>
		</case>
		***************************************/
		// 'finding', 'importance', 'predictiveStrength', 'solution', 'solution:toEntity',
		// 'solution:predictiveStrength', and 'solution:importance' are not mandatory
		// 'status' might either be 'solved' or 'unsolved', if no solution element then the case
		// is in 'processing' state

		// also handle the subelements
		// but first handle the attributes
		String name =
			DOMutilities.getAttributeValueByName("name", node.getAttributes());
		String description =
			DOMutilities.getAttributeValueByName("description", node.getAttributes());

		// now create the case
		try {
			kmiu.addCase(name, description);
			statusWindow.appendText(
				"Added a Case => '" + name + "' : {" + description + "}");
		} catch (KnowledgeModelImportUtilities.CaseAlreadyExist ex) {
			statusWindow.appendText(ex.getMessage());
		}

		// even if the case should already exists there is no problem to continue
		// it might be that new findings or solution could be added.
		// now continue with the findings and a possible solution
		Node child = node.getFirstChild();
		while (child != null && !(fatalError && abortOnFatalErrors)) {
			if (child.getNodeType() == Node.ELEMENT_NODE) {
				// the findings
				if (child.getNodeName().equalsIgnoreCase("finding")) {
					// first collect the attrbutes
					String relationType =
						DOMutilities.getAttributeValueByName(
							"relationType",
							child.getAttributes());
					String toEntity =
						DOMutilities.getAttributeValueByName(
							"toEntity",
							child.getAttributes());
					String importance =
						DOMutilities.getAttributeValueByName(
							"importance",
							child.getAttributes());
					String predictiveStrength =
						DOMutilities.getAttributeValueByName(
							"predictiveStrength",
							child.getAttributes());

					// then add the finding
					try {
						kmiu.addHasFinding(
							name,
							relationType,
							toEntity,
							importance,
							predictiveStrength);
						statusWindow.appendText(
							"=> Added a Finding => '"
								+ name
								+ "' <"
								+ relationType
								+ "> '"
								+ toEntity
								+ "' : ["
								+ importance
								+ " | "
								+ predictiveStrength
								+ "]");
					} catch (
						KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
						statusWindow.appendText(ex.getMessage());
					} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						statusWindow.appendText(ex.getMessage());
					} catch (KnowledgeModelImportUtilities.RelationTypeDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						statusWindow.appendText(ex.getMessage());
					}
				}
				// the possible solution
				else if (child.getNodeName().equalsIgnoreCase("solution")) {
					// first collect the attrbutes
					String status =
						DOMutilities.getAttributeValueByName(
							"status",
							child.getAttributes());
					String toEntity =
						DOMutilities.getAttributeValueByName(
							"toEntity",
							child.getAttributes());
					String importance =
						DOMutilities.getAttributeValueByName(
							"importance",
							child.getAttributes());
					String predictiveStrength =
						DOMutilities.getAttributeValueByName(
							"predictiveStrength",
							child.getAttributes());

					// then add the solution
					try {
						kmiu.addHasSolution(
							name,
							status,
							toEntity,
							importance,
							predictiveStrength);
						statusWindow.appendText(
							"=> Added a Solution => '"
								+ name
								+ "' <-> '"
								+ toEntity
								+ "' : ["
								+ importance
								+ " | "
								+ predictiveStrength
								+ " | "
								+ status
								+ "]");
					} catch (
						KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex) {
						statusWindow.appendText(ex.getMessage());
					} catch (KnowledgeModelImportUtilities.EntityDoNotExist ex) {
						if (abortOnFatalErrors) {
							fatalError = true;
						}
						statusWindow.appendText(ex.getMessage());
					}
				}
			} // if (child.getNodeType()...
			child = child.getNextSibling();
		} // while
	} // handleCaseElement
} // class JCXMLimportParser