package jcreek.cke.importexport.owl;

import java.awt.Frame;
import java.io.*;
import jcreek.cke.browser.SimpleFileFilter;
import jcreek.cke.importexport.ProcessControlInterface;
import jcreek.cke.importexport.ProcessStatusWindow;
import jcreek.representation.KnowledgeModel;
import jcreek.representation.sw.OWLexportParser;

import jcreek.cke.visualizer.AppSettings;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

/**
 * This class is the one having the overall control of the export process of
 * the Knowledge Model of Creek to OWL ontology representation
 *
 * Copied from JCXMLexport by Tomassen, 2002
 *
 * @author Mikael Kirkeby Fidjeland, 2006
 */
public class OWLexport extends Thread implements ProcessControlInterface {

    private static String DEFAULT_NS = "http://creek.idi.ntnu.no/owl/example/";

    private ExtThread exportThread = null;
    private Frame parent = null;
    private KnowledgeModel km = null;
    private boolean exportDone = false;
    private boolean aborted = false;
    private String owlFile = null;
    private String namespace = null;
    private ProcessStatusWindow statusWindow = null;
    private AppSettings props = null;

    /**
     * Constructor for OWLexport, which does set some basic settings
     * before the process of exporting the Knowledge Model may start.
     *
     * @param parent the parent frame window.
     * @param km the JavaCreek knowledge model to export the model to.
     */
    public OWLexport(Frame parent, KnowledgeModel km) {
        this.parent = parent;
        this.km = km;
        this.statusWindow =
            new ProcessStatusWindow(parent, "OWL Export Status", this);
        props = AppSettings.instanceOf();
        props.load("Creek.ini");
    }

    /**
     * Checks if the process was aborted;
     *
     * @return true if the porcess was aborted, otherwise false.
     */
    public boolean isAborted() {
        return aborted;
    } // isAborted

    /**
     * Starts the process by first showing the FileChooser dialog
     * then starts the export process, by first starting the thread
     * and gives feedback of the process to the OWL status dialog.
     */
    public void startExport() {
        JFileChooser fc =  new JFileChooser();
        fc.setMultiSelectionEnabled(false);
        fc.addChoosableFileFilter(
                                  new SimpleFileFilter(new String[] { "owl" }, "OWL Files (*.owl)"));
        if (fc.showSaveDialog(parent) == JFileChooser.APPROVE_OPTION) {
            props.setString("", fc.getSelectedFile().getPath());
            props.commit();
            owlFile = fc.getSelectedFile().getAbsolutePath();
            if(!owlFile.toLowerCase().endsWith(".owl"))
                owlFile += ".owl";
            // get the namespace
            JOptionPane inputBox = new JOptionPane();
            namespace = inputBox.showInputDialog(parent, "Namespace",
                                                 DEFAULT_NS +
                                                 fc.getSelectedFile().getName());
            if (namespace == null || namespace == "")
                namespace = DEFAULT_NS;

            // now start the thread
            exportThread = new ExtThread();
            exportThread.start();
            // wait until done or aborted, update the status dialog while waiting
            // open the status dialog in the process
            statusWindow.show();
        }
    } // startExport

    /**
     * Aborts the export process.
     */
    public void abortProcess() {
        // check if not already been aborted
        if (exportThread != null) {
            // interrupt the thread
            Thread moribund = exportThread;
            exportThread = null;
            moribund.interrupt();
            aborted = true;
        }
    } // abortProcess

    //****************************************
    // Inner classes
    //****************************************

    protected class ExtThread extends Thread {
        /**
         * Does a check if aborted before every new task
         *
         * @see java.lang.Runnable#run()
         */
        public void run() {
            statusWindow.setCancelButton();
            if (owlFile != null) {
                statusWindow.appendText("Starts the creation of '" + owlFile + "'...");
                OWLexportParser creator = new OWLexportParser(km, namespace);
                creator.createOntology();
                try {
                    statusWindow.appendText("writing...");
                    creator.write(owlFile);
                    statusWindow.appendText("done");
                }
                catch (FileNotFoundException ex) {
                    statusWindow.appendText(ex.getMessage());
                    ex.printStackTrace();
                    statusWindow.appendText("Error: File not found");
                }
                catch (IOException ex) {
                    statusWindow.appendText(ex.getMessage());
                    ex.printStackTrace();
                    statusWindow.appendText("Error: IO exception");
                }
            }
            // when finished
            statusWindow.setCloseButton();
            exportDone = true;
        } // run

        /**
         * The default message given if the process was aborted.
         */
        public void abortedMessage() {
            statusWindow.appendText("Export of OWL was aborted...");
            statusWindow.setCloseButton();
        } // abortedMessage
    } // ExtThread
}
