package jcreek.cke.importexport.xml;

import org.apache.xerces.dom.TextImpl;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.io.InputStream;
import java.io.Reader;

/**
 * The intention of this class is to have a class that may be used
 * in most situations where basic settings of a DOM parser are needed.
 * This parser may as well be more specialized by creating a class
 * extending this class.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class DefaultGeneralParser {
	//********************************************
	// Constants
	//********************************************
	/** Xerces validation by parser feature */
	public static final int VALIDATION = 0;
	/** Xerces namespaces consideration by parser feature */
	public static final int NAMESPACES = 1;
	/** Xerces schema support by parser feature */
	public static final int SCHEMA_SUPPORT = 2;
	/** Xerces full schema support by parser feature */
	public static final int SCHEMA_FULL_SUPPORT = 3;
	/** Xerces deffered dom support by parser feature */
	public static final int DEFFERED_DOM = 4;
	/** Xerces include ignorable whitespaces when parsing feature */
	public static final int INCLUDE_IGNORABLE_WHITESPACE = 5;

	//********************************************
	// Data
	//********************************************
	// some default settings for validator
	// set to default according to Apache XML project
	private static boolean _validation = false;
	private static boolean _nameSpaces = true;
	private static boolean _schemaSupport = true;
	private static boolean _schemaFullSupport = false;
	private static boolean _deferredDOM = true;
	private static boolean _includeIgnorableWhitespace = true;
	/** The parser wrapper */
	private DOMparserWrapper _parser;

	//********************************************
	// Public methods
	//********************************************

	/**
	 * Initializes the parser and sets the features of the parser
	 * to the default values if not set otherwise by the setFeature method.
	 *
	 * @throws Exception throws an exception if an fatal error occurs
   *
   * @see jcreek.cke.importexport.xml.DefaultGeneralParser#setFeature(int, boolean)
	 */
	public void initParser() throws Exception {
		_parser =
			(DOMparserWrapper) new jcreek.cke.importexport.xml.XercesParserWrapper();

		_parser.setFeature(
			"http://apache.org/xml/features/dom/defer-node-expansion",
			_deferredDOM);
		_parser.setFeature("http://xml.org/sax/features/validation", _validation);
		_parser.setFeature("http://xml.org/sax/features/namespaces", _nameSpaces);
		_parser.setFeature(
			"http://apache.org/xml/features/validation/schema",
			_schemaSupport);
		_parser.setFeature(
			"http://apache.org/xml/features/validation/schema-full-checking",
			_schemaFullSupport);
		_parser.setFeature(
			"http://apache.org/xml/features/dom/include-ignorable-whitespace",
			_includeIgnorableWhitespace);
	} // initParser

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#parse(String)
	 */
	public Document parse(String url) throws Exception {
		return _parser.parse(url);
	}

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#parse(InputStream)
	 */
	public Document parse(InputStream istream) throws Exception {
		return _parser.parse(istream);
	}

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#parse(Reader)
	 */
	public Document parse(Reader reader) throws Exception {
		return _parser.parse(reader);
	}

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#parse(Document)
	 */
	public Document parse(Document document) throws Exception {
		return _parser.parse(document);
	}

	/**
	 * Set features of the parser. Note that a feature will not be
   * put in action before the method initParser() is called.
	 *
	 * @param feature the feature to set, declared as static constants
	 * for this class
	 * @param state set to true to enabled the specified feature and false
	 * to disable the feature.
   *
   * @see jcreek.cke.importexport.xml.DefaultGeneralParser#initParser()
	 */
	public void setFeature(int feature, boolean state) {
		switch (feature) {
			case VALIDATION :
				{
					_validation = state;
					break;
				}
			case NAMESPACES :
				{
					_nameSpaces = state;
					break;
				}
			case SCHEMA_SUPPORT :
				{
					_schemaSupport = state;
					break;
				}
			case SCHEMA_FULL_SUPPORT :
				{
					_schemaFullSupport = state;
					break;
				}
			case DEFFERED_DOM :
				{
					_deferredDOM = state;
					break;
				}
			default :
				{
					return;
				}
		}
	} // setFeature

	/**
	 * Checks if there where any faults reported. If there where any faults, then
	 * that indicate that the document is not valid
	 *
	 * @return true if any faults where reported, else returning false
	 */
	public boolean anyFaults() {
		if ((getNumberOfErrors() + getNumberOfFatalErrors() + getNumberOfWarnings()
			> 0))
			return true;
		else
			return false;
	} // anyFaults

	/**
	 * A method that returns all the error messages that have occured during the parsing
	 * gathered in one string.
	 *
	 * @return the string containing all the error messages that have occured.
	 */
	public String getAllFaultMsg() {
		return getFatalErrorsMsg()
			+ '\n'
			+ getErrorsMsg()
			+ '\n'
			+ getWarningsMsg();
	}

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#getNumberOfErrors()
	 */
	public int getNumberOfErrors() {
		return _parser.getNumberOfErrors();
	} // getNbrOfErrors

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#getErrorsMsg()
	 */
	public String getErrorsMsg() {
		return _parser.getErrorsMsg();
	} // getErrorsMsg

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#getNumberOfWarnings()
	 */
	public int getNumberOfWarnings() {
		return _parser.getNumberOfWarnings();
	} // getNbrOfWarnings

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#getWarningsMsg()
	 */
	public String getWarningsMsg() {
		return _parser.getWarningsMsg();
	} // getWarningsMsg

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#getNumberOfFatalErrors()
	 */
	public int getNumberOfFatalErrors() {
		return _parser.getNumberOfFatalErrors();
	} // getNbrOfFatalErrors

	/**
	 * @see jcreek.cke.importexport.xml.DOMparserWrapper#getFatalErrorsMsg()
	 */
	public String getFatalErrorsMsg() {
		return _parser.getFatalErrorsMsg();
	} // getFatalErrorsMsg
} // class AbstractDefaultParser