package jcreek.cke.importexport.xml;

import java.io.Reader;
import java.io.StringReader;

import jcreek.cke.util.FileUtilities;
import jcreek.cke.importexport.ProcessStatusWindowInterface;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The intention of this class is to help validate a XML file by
 * specifying a Schema file. The XML file is first coarse tested
 * if it is a valid XML file, and if that the case then the correct
 * Schema path (the local Schema) is set. Then the validation against
 * the specified Schema is performed. In case of faults during the
 * validation, the error-messages are logged and may be retrieved.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class XMLvalidater extends DefaultGeneralParser {
	//********************************************
	// Data variables
	//********************************************
	/** is the XML file valid according to Schema? */
	private boolean _valid = false;
	/** indicating if there where any faults */
	private boolean _faults = false;
	private String _errorMsg = null;
	private boolean _done = false;
	private String _root = null;
	private String _schemaPath = null;
	private Document _document = null;

	//********************************************
	// Constructors
	//********************************************

  /**
   * Initializes the parser for validation by specifying a file path to the
   * DOM document and the schema path. If the validation was successful
   * may be checked with the method: isValid()
   *
   * @param filepath the file path to the DOM document to validate
   * @param schemaPath the path to the local schema to validate against
   *
   * @see jcreek.cke.helpers.xml.XMLvalidater#isValid()
   */
	public XMLvalidater(String filepath, String schemaPath) {
		_schemaPath = schemaPath;
		this.initParser(new StringReader(FileUtilities.readFromFile(filepath)));
	} // XMLvalidater

  /**
   * Initializes the parser for validation by specifying a java.io.Reader with the
   * DOM document and the schema path. If the validation was successful
   * may be checked with the method: isValid()
   *
   * @param source a Reader with the DOM document to validate
   * @param schemaPath the path to the local schema to validate against
   *
   * @see jcreek.cke.helpers.xml.XMLvalidater#isValid()
   */
	public XMLvalidater(Reader source, String schemaPath) {
		_schemaPath = schemaPath;
		this.initParser(source);
	} // XMLvalidater

	/**
	 * Initializes the parser for validation by specifying a
   * DOM document and the schema path. If the validatton was successful
   * may be checked with the method: isValid()
   *
	 * @param document the DOM document to validate
	 * @param schemaPath the path to the local schema to validate against
   *
   * @see jcreek.cke.helpers.xml.XMLvalidater#isValid()
	 */
	public XMLvalidater(Document document, String schemaPath) {
		_schemaPath = schemaPath;
		try {
			_root = getDocRootNodeName(document);
			correctSchemaPath(document);

			if (!_faults) {
				// then validate message
				super.setFeature(DefaultGeneralParser.VALIDATION, true);
				super.setFeature(DefaultGeneralParser.SCHEMA_FULL_SUPPORT, true);
				super.initParser();
				_document = super.parse(document);
			}
		} catch (Exception ex) {
			_faults = true; // indicating a fault
			_errorMsg = "Unable To Set Parser Features";
			_document = null;
		}
	} // XMLvalidater

	//********************************************
	// Public Methods
	//********************************************

	/**
	 * A method that returns the DOM document.
   *
	 * @return the DOM document specified in the constructor
	 */
	public Document getDocument() {
		return _document;
	}

	/**
	 * Checks to see if the message is a valid message according
   * to the specified schema in the constructor
	 *
	 * @return true if it is a valid message, otherwise false will be returned
	 */
	public boolean isValid() {
		return (!super.anyFaults() && !_faults);
	} // isValid

	/**
	 * Returns the schema file name and path of the message
	 *
	 * @return the schema file name
	 */
	public String getSchemaPath() {
		return _schemaPath;
	} // getSchemaPath

	/**
	 * A method that retrieves the node name of the root node of
   * the DOM document node.
	 *
	 * @param document the DOM document to handle
	 * @return the node name of the root node if found, does otherwise return null
	 */
	public String getDocRootNodeName(Document document) {
		// is there anything to do?
		if (document == null) {
			return null;
		}
		// handle the child nodes of the document
		NodeList children = document.getChildNodes();
		if (children != null) {
			int len = children.getLength();
			int i = 0;
			while (i < len
				&& (children.item(i).getNodeType() != Node.ELEMENT_NODE)) {
				i++;
			}
			if (i < len)
				return children.item(i).getNodeName();
			else
				return null;
		} else
			return null;
	} // getRoot

	/**
	 * Returns the Exception message
	 *
	 * @return a string with the Exception message
	 */
	public String getErrorMsg() {
		if (_errorMsg != null)
			return _errorMsg + '\n' + super.getAllFaultMsg();
		else
			return super.getAllFaultMsg();
	} // getExceptionMsg

	//********************************************
	// Protected methods
	//********************************************

	/**
	 * Initializes the parser. Used by the constructors.
	 *
	 * @param source the source of the document to parse
	 */
	protected void initParser(Reader source) {
		try {
			// first do element check to identify arkmin_message and type,
			// and set to the correct arkmin_schema according to version
			super.setFeature(DefaultGeneralParser.VALIDATION, false);
			super.setFeature(DefaultGeneralParser.SCHEMA_FULL_SUPPORT, true);
			super.initParser();
			Document document_1 = super.parse(source);
			_root = getDocRootNodeName(document_1);
			correctSchemaPath(document_1);

			if (!_faults) {
				// then validate message
				super.setFeature(DefaultGeneralParser.VALIDATION, true);
				super.setFeature(DefaultGeneralParser.SCHEMA_FULL_SUPPORT, true);
				super.initParser();
				_document = super.parse(document_1);
			}
		} catch (Exception ex) {
			_faults = true; // indicating a fault
			_errorMsg = "Unable To Set Parser Features";
			_document = null;
		}
	} // initParser

	/**
	 * A method that tries to locate where in the DOM document the Schema
   * is set. Nevertheless, the correct schema path is set to the local
   * schema. This procedure is important to ensure that the schema path
   * is correct in able to validate the document.
   * <p>
   * This method is recursive, but ends when the correct schema path is set.
   * The schema used is the one specified in the constructors, and this
   * method is also called by the constructors.
   *
	 * @param node the root node of the DOM document
	 */
	protected void correctSchemaPath(Node node) {
		// is there anything to do?
		if (node == null) {
			return;
		}
		int type = node.getNodeType();
		switch (type) {
			case Node.ELEMENT_NODE :
				{
					// first check if root node
					if (node.getNodeName().equalsIgnoreCase(_root)) {
						// then set to the correct local schema. Overwrites the attribute if already exists
						if (_schemaPath != null && _schemaPath.length() > 0) {
							// now check if file really exists
//							if (FileUtilities.exists(_schemaPath)) {
								// set the "new" attribute
								((Element) node).setAttribute(
									"xsi:noNamespaceSchemaLocation",
									_schemaPath);
/*							} else {
								_faults = true; // indicating a fault
								_errorMsg = "Schema Not Found";
							} */
						} else {
							_faults = true; // indicating a fault
							_errorMsg = "Schema Path Properties Not Found";
						}
						_done = true;
						return;
					} else {
						// handle the child nodes
						NodeList children = node.getChildNodes();
						if (children != null) {
							int len = children.getLength();
							int i = 0;
							while (i < len && (!_faults && !_done)) {
								correctSchemaPath(children.item(i));
								i++;
							}
						}
					}
					break;
				} // case Node.ELEMENT_NODE:
			case Node.DOCUMENT_NODE :
				{
					// handle the child nodes
					NodeList children = node.getChildNodes();
					if (children != null) {
						int len = children.getLength();
						int i = 0;
						while (i < len && (!_faults && !_done)) {
							correctSchemaPath(children.item(i));
							i++;
						}
					}
					break;
				} // case Node.DOCUMENT_NODE:
		}
	} // correctSchemaPath
}