/***********************************************************************
 *
 * DESCRIPTION:
 *      This file defines the class for the move tool. The move tool is
 *      used to move entities and relatios around a graph. This tool is
 *      not used for creating new entities or relations.
 *
 **********************************************************************/

package jcreek.cke.tool;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import javax.swing.Action;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenuItem;
import javax.swing.JComponent;
import javax.swing.JButton;
import java.awt.Component;
import java.awt.Cursor;
import javax.swing.JInternalFrame;
import javax.swing.SwingUtilities;
import javax.swing.JMenu;
import javax.swing.ImageIcon;
import java.net.URL;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.IOException;
import java.awt.Toolkit;
import java.io.Serializable;
import java.util.Vector;

import jcreek.cke.graph.CreekModel;
import jcreek.cke.conceptview.*;
import jcreek.cke.document.CreekDocument;
import jcreek.cke.CreekDocumentGroup;
import jcreek.cke.graph.VertexList;
import jcreek.cke.command.EntityMoveCommand;
import jcreek.cke.command.PasteCommand;
import jcreek.cke.util.DPoint;
import jcreek.cke.util.Matrix33;
import java.util.Iterator;

import jcreek.representation.Relation;
import jcreek.representation.Entity;

import jcreek.representation.RelationType;
import jcreek.cke.conceptview.DrawingSurface;
import jcreek.cke.CreekDocumentGroup;
import jcreek.cke.MainWindow;
import jcreek.cke.command.ActivateEntityCommand;
import jcreek.cke.command.ActivateRelationCommand;
import jcreek.cke.command.DeactivateRelationCommand;
import jcreek.cke.browser.Visualizer;
import jcreek.cke.util.SelectionSet;
import jcreek.cke.util.EntitySelection;
import jcreek.cke.graph.RelationList;
import jcreek.representation.KnowledgeModel;
import jcreek.cke.visualizer.AppSettings;

/**
 * Tool for moving entities and relations around a graph; acts as a
 * go-between for the user and the graph. This class takes user
 * input and coordinates the interaction by talking to the model
 * and view to provide appropriate feedback to the user. In this
 * way, the view and model appear to be operating as one entity.
 *      <p>
 * User can select a entity or relation by clicking on it. If user
 * clicks in empty space and drags the mouse they can select any
 * number of entities that fall within the bounding rectangle.
 * To select multiple objects, user must hold down Control key.
 *
 * @author
 * @version
 */

public class MoveTool implements CreekTool,
                       ConceptViewPaintListener,
                       PropertyChangeListener,
                       DropTargetListener,
                       DragGestureListener,
                       DragSourceListener
{
    /**
     * Model that we're associated with.
     */
    protected CreekModel model;
    /**
     * View that we're associated with.
     */
    protected ConceptView view;
    /**
     * Document this mediator is associated with.
     */
    protected CreekDocument document;
    /**
     * Vertex the user pressed on in mousePressed(). Could be
     * null if the user didn't press on a entity.
     */
    protected VisualEntity pressVertex;
    /**
     * Vertex the rightClicked. Could be
     * null if the user didn't rightClick on a entity.
     */
    protected VisualEntity propVertex;
    /**
     * List of the selected entities in the order they were
     * selected by the user.
     */
    protected VertexList selectedList;
    /**
     * True if mouse was pressed on the view before being released.
     * Helps avoid some mis-firings of the mouse buttons.
     */
    protected boolean wasPressed;
    /**
     * True if user is selecting entities by defining a region.
     * This is true when user pressed but not on a entity and
     * begins dragging the mouse.
     */
    protected boolean isSelectRegion;
    /**
     * True if user has pressed CTRL key.
     */
    protected boolean isCTRLDown;
    /**
     * The list of the entities being moved around. These
     * entities are a copy of the originals and can be modified.
     */
    protected VertexList movingVertices;
    /**
     * First mouse press position when dragging started.
     */
    protected Point firstMousePos;
    /**
     * Most recent mouse position while dragging.
     */
    protected Point lastMousePos;

    public PropertyWindow propWin;

    protected JPopupMenu relationList = new JPopupMenu("test");

    ItemListener  relationListHandler = new ItemListener() {
        public void itemStateChanged(ItemEvent e) {
            toggleRelation(propVertex, ((JCheckBoxMenuItem) e.getItem()).getText());
      }
    } ;

    protected DragSource      dragSource = null;
    private   ArrayList       dragGestureRecognizers=null;
    public static final DataFlavor dataFlavor =
          new DataFlavor(SelectionSet.class,"Array of VisualEntities");

    /**
     * origin must be made to go in the Transferable somehow.
     */
    private   Point           origin = null;

    /**
     * Creates a new MoveTool object.
     */
    public MoveTool() {
        selectedList = new VertexList();
        firstMousePos = new Point();
        lastMousePos = new Point();
        propWin = (PropertyWindow)new MoveProperty();
        dragSource = DragSource.getDefaultDragSource();
        dragGestureRecognizers = new ArrayList();
    } // MoveTool

    /**
     * Attach to the given model and view objects. This begins
     * user interaction and will imply that this tool is the
     * currently active tool. The tool should add itself as a
     * listener to the view object to begin interaction with
     * the user.
     *
     * @param  model  CreekModel to attach to
     * @param  view   ConceptView to attach to
     */
    public void attach( CreekModel model, ConceptView view,JInternalFrame win ) {
        if ( ( model == this.model ) && ( view == this.view ) ) {
            // Nothing has changed, nothing to do.
            return;
        }
        // Detach in case our manager forgot to.
        detach();

        // Now attach to new model and view.
        this.model = model;
        this.view = view;
        CreekDocumentGroup dg = CreekDocumentGroup.instanceOf();
        document = dg.getDocument( view );
        view.addViewListener( this );
        view.addViewPaintListener( this );
        model.addChangeListener( this );
        //recreate the selected list for tool
        VisualEntity[] ves = model.getSelectedEntities();
        for(int i=0;i<ves.length;i++)
        {
            selectedList.add(ves[i]);
        }
//        win.setContentPane(propWin.getComponent());
        propWin.attach(model,view,win);
	createDragGestureRecognizer(view, this);
    } // attach

    /**
     * Detach from the model and view and prepare to be
     * deactivated. This does not imply that this object is
     * about to be destroyed. The tool should remove itself
     * from the view listener list and cease interaction
     * with the user.
     * <p>
     * Note that this method may be called more than once
     * between calls to <code>attach()</code>
     */
    public void detach() {
        if ( view != null ) {
            destroyDragGestureRecognizer(view,this);
            view.removeViewPaintListener( this );
            view.removeViewListener( this );
            view = null;
        }
        if ( model != null ) {
            model.removeChangeListener( this );
            // forget the selected entities
//            model.unselectAllVertices(); comented to retain selection
            model = null;
        }
        selectedList.clear();
        document = null;
        propWin.detach();
    } // detach


    /**
     * Find an entity close to the given x,y position.
     *
     * @param  x  mouse x position
     * @param  y  mouse y position
     */
    protected VisualEntity findNode( int x, int y ) {
        return view.findVisualEntity(x,y);
   } // findEdge

    /**
     * If we were moving entities, clean up now.
     */
    protected void finishMoving() {
        if ( movingVertices != null ) {
            movingVertices.clear();
            movingVertices = null;
            //got to add a recal method
            view.fireContentsResize();
        }
        else
            view.repaint();

    } // finishMoving

    /**
     * Returns the list of selected entities, in the order they
     * were selected by the user. Often called by another class
     * that is going to perform some operation on the graph,
     * such as a search.
     *
     * @return  selected entity list
     */
    public VertexList getSelected() {
        return selectedList;
    } // getSelected

    /**
     * We ignore this event.
     *
     * @param  e  mouse event
     */
    public void mouseClicked( MouseEvent e ) {
//        ((DrawingSurface)view).center();
    } // mouseClicked

    /**
     * The user is dragging the mouse so lets see about moving
     * the selected entities.
     *
     * @param  me  mouse event
     */
    public void mouseDragged( MouseEvent me ) {

        if(me.isAltDown()&&me.isShiftDown())
            return;
        if(SwingUtilities.isLeftMouseButton(me))
        {
            if ( !wasPressed ) {
                return;
            }
            int deltaX = me.getX()-lastMousePos.x;
            int deltaY = me.getY()-lastMousePos.y ;
            lastMousePos.x =  me.getX();
            lastMousePos.y =  me.getY();

            if ( ( isSelectRegion == false ) &&
                 ( movingVertices == null ) &&
                 ( Math.abs( lastMousePos.x - firstMousePos.x ) < 3 ) &&
                 ( Math.abs( lastMousePos.y - firstMousePos.y ) < 3 ) ) {
                // We ignore tiny mouse movements until we being
                // doing something, such as moving vertices or
                // selecting a region.
                return;
            }

            // First make sure the vertex the user clicked on is
            // selected (putting it in the selected list). This
            // makes it possible to move several vertices at once.
            if ( ( pressVertex != null ) &&
                 ( pressVertex.isSelected() == false ) ) {
                model.setSelected( pressVertex, true );
                selectedList.add( pressVertex );
            }

            // If there are some vertices selected, deal with moving
            // them around the view.
            if ( ( pressVertex != null ) && ( selectedList.size() > 0 ) ) {
                if ( movingVertices == null ) {
                    // Make a copy of the selected vertices so we can
                    // modify their positions based off the originals.
                    // We'll use the copy in resetMovingVertices() to
                    // restore the original positions.
                    movingVertices = selectedList.copy();
                }
                VisualEntity v1;
                int size = movingVertices.size();
                for ( int i = 0; i < size; i++ ) {
                    // Move the original vertices based off the
                    // position of the copies.
                    v1 = selectedList.get( i );
                    v1.slide(deltaX,deltaY);
                }
//                view.drawSelectedNodes();
            } else if ( pressVertex == null ) {
                // User didn't click on a vertex and they are
                // defining a significant region.
                isSelectRegion = true;
            }
            view.repaint();
        }

    } // mouseDragged

    /**
     * We ignore this event.
     *
     * @param  e  mouse event
     */
    public void mouseEntered( MouseEvent e ) {
    } // mouseEntered

    /**
     * We ignore this event.
     *
     * @param  me  mouse event
     */
    public void mouseExited( MouseEvent me ) {
        ((JComponent)view).setCursor(Cursor.getDefaultCursor());
    } // mouseExited

    /**
     * We ignore this event.
     *
     * @param  e  mouse event
     */
    public void mouseMoved( MouseEvent me ) {

        if(findNode(me.getX(),me.getY())!=null)
          ((JComponent)view).setCursor(new Cursor( Cursor.MOVE_CURSOR ));
        else
          ((JComponent)view).setCursor(Cursor.getDefaultCursor());


    } // mouseMoved

    /**
     * User has pressed a mouse button on us. Interpret the click
     * as clicking a entity and save the entity reference. May be
     * null if the user didn't click on a entity.
     *
     * @param  me  mouse event for this canvas
     */
    public void mousePressed( MouseEvent me ) {
        if(me.isAltDown()&&me.isShiftDown())
            return;
        if(SwingUtilities.isLeftMouseButton(me))
        {
            // Remember the mouse position for when we have to
            // move entities around the view.
            firstMousePos.x = me.getX();
            firstMousePos.y = me.getY();
            lastMousePos.x = firstMousePos.x;
            lastMousePos.y = firstMousePos.y;
            pressVertex = findNode(firstMousePos.x, firstMousePos.y);
            wasPressed = true;
            isSelectRegion = false;
            if ( !me.isControlDown() ) {
              if((pressVertex == null)||(!pressVertex.isSelected()))
              {
                  // If user is not holding down the Ctrl key, deselect all.
                  if ( selectedList.size() > 0 ) {
                      selectedList.clear();
                  }
                  model.unselectAllVertices();
  //                model.unselectAllEdges();
              }//selection vertex check loop
            }// Control button loop
        }//left mouse button loop
        if (me.isPopupTrigger())
        {
            firstMousePos.x = me.getX();
            firstMousePos.y = me.getY();
            propVertex = findNode(firstMousePos.x, firstMousePos.y);
            if(propVertex!=null)
                if(!propVertex.isSelected())
                {
                      if ( selectedList.size() > 0 ) {
                          selectedList.clear();
                      }
                      model.unselectAllVertices();
                      model.setSelected( propVertex, true );
                      selectedList.add( propVertex );
                }
            formPopup(me.getX(),me.getY());
        }

    } // mousePressed

    /**
     * User has released the mouse button. We now check to see if
     * the user was moving entities, selecting a region, selecting
     * a entity, or selecting an relation.
     *
     * @param  me  mouse event for this canvas
     */
    public void mouseReleased( MouseEvent me ) {
        if(SwingUtilities.isLeftMouseButton(me))
        {
            if ( wasPressed == false ) {
                // This avoids the case where the user clicks outside
                // of the view and drags the mouse onto the view.
                // This takes care of menu selections or clicking
                // the "Ok" button of a dialog shown over the view.
                return;
            }
            wasPressed = false;
            int mX = me.getX();
            int mY = me.getY();
            if ( movingVertices != null ) {
                // We were moving vertices around the screen.
                resetMovingVertices();
                double deltaX = mX - firstMousePos.x;
                double deltaY = mY - firstMousePos.y;
                EntityMoveCommand action = new EntityMoveCommand
                    ( document, selectedList, ((int)deltaX),((int) deltaY));
                document.performCommand( action );
            } else if ( isSelectRegion ) {
                // User might have been selecting a region. Be sure
                // to invoke repaint, in the event the user didn't
                // capture any vertices in the region.
                selectRegion();
                isSelectRegion = false;
//                view.repaint(0,0,((JComponent)view).getWidth(),((JComponent)view).getHeight());
            } else if ( pressVertex != null ) {
                if ( pressVertex.isSelected() ) {
                    if(me.isControlDown())
                    {
                        model.setSelected( pressVertex, false );
                        selectedList.remove(pressVertex);
                    }
                    else
                    {
                          // If user is not holding down the Ctrl key, deselect all.
                          if ( selectedList.size() > 0 ) {
                              selectedList.clear();
                          }
                          model.unselectAllVertices();
                          model.setSelected( pressVertex, true );
                          selectedList.add( pressVertex );
                    }
                } else {
                    model.setSelected( pressVertex, true );
                    selectedList.add( pressVertex );
                }
            }
//            view.repaint();
            finishMoving();
            pressVertex = null;
        }
        if (me.isPopupTrigger())
        {
            firstMousePos.x = me.getX();
            firstMousePos.y = me.getY();
            propVertex = findNode(firstMousePos.x, firstMousePos.y);
            if(propVertex!=null)
               if(!propVertex.isSelected())
                {
                      if ( selectedList.size() > 0 ) {
                          selectedList.clear();
                      }
                      model.unselectAllVertices();
                      model.setSelected( propVertex, true );
                      selectedList.add( propVertex );
                }
            formPopup(me.getX(),me.getY());
        }
    } // mouseReleased

    /**
     * Called after the view draws the graph. Gives the
     * listener a chance to draw something to the view
     * after everything else has been drawn. The order in
     * which listeners are called is non-deterministic.
     *
     * @param  event  view paint event
     */
    public void paintAfter( ConceptViewPaintEvent event ) {
        if ( isSelectRegion ) {
            // Draw a bounding rectangle to show what will be
            // selected.
            Graphics2D g = event.getGraphics();
            g.setColor( Color.black );
            float dash[] = { 2.0f };
            BasicStroke stroke = new BasicStroke( 1.0f, BasicStroke.CAP_BUTT,
                                                  BasicStroke.JOIN_MITER,
                                                  10.0f, dash, 0.0f );
            g.setStroke( stroke );
            int x = (int)Math.min( firstMousePos.x, lastMousePos.x );
            int y = (int)Math.min( firstMousePos.y, lastMousePos.y );
            int w = (int)Math.max( firstMousePos.x, lastMousePos.x ) - x;
            int h = (int)Math.max( firstMousePos.y, lastMousePos.y ) - y;
            g.drawRect( x, y, w, h );
        }
    } // paintAfter

    /**
     * Called before the view draws the graph. Gives the
     * listener a chance to draw something to the view
     * before anything else appears. The order in which
     * listeners are called is non-deterministic.
     *
     * @param  event  view paint event
     */
    public void paintBefore( ConceptViewPaintEvent event ) {
        // We have nothing to do for now.
    } // paintBefore

    /**
     * The state of the object we're listening to has changed.
     * This means we should check if any of the selected entities
     * have been removed.
     *
     * @param  e  change event
     */
    public void propertyChange( PropertyChangeEvent e ) {
        if ( e.getPropertyName().equals( "selection" ) ) {
           return;
        }

    } // propertyChange

    /**
     * Moves all the moving entities back to their original
     * positions by using the original positions stored in
     * the copy of the entities.
     */
    protected void resetMovingVertices() {
        int size = selectedList.size();
        for ( int i = 0; i < size; i++ ) {
            VisualEntity selV = selectedList.get( i );
            VisualEntity origV = movingVertices.get( i );
            selV.setPosition( origV.getPosition().x, origV.getPosition().y );
        }
    } // resetMovingVertices

    /**
     * Called when user was defining a rectangular region with
     * which to select entities. Determines which entities are
     * within the rectangle and marks them selected.
     */
    protected void selectRegion() {
        // determine the bounding rectangle based on mouse positions
        int x1 = (int)Math.min( firstMousePos.x, lastMousePos.x );
        int y1 = (int)Math.min( firstMousePos.y, lastMousePos.y );
        int x2 = (int)Math.max( firstMousePos.x, lastMousePos.x );
        int y2 = (int)Math.max( firstMousePos.y, lastMousePos.y );
        // for each vertex in the model...
        VertexList vl = model.getEntities();
        int size = vl.size();
        Matrix33 viewTransform = view.getViewTransform();
        for ( int i = 0; i < size; i++ ) {
            VisualEntity v = vl.get( i );
               DPoint pos = v.getPosition();
               pos.transform(viewTransform);
            // ...if the vertex is within the region, and it is
            // not currently selected, mark it selected now
               if(pos.x >= x1 && pos.x <= x2 && pos.y >= y1 && pos.y <= y2)
               {
                  model.setSelected( v, true );
                  selectedList.add( v );
              }
        }
    } // selectRegion

    /**
     * Returns a string representation of this object.
     *
     * @return  string representing this object
     */
    public String toString() {
        return "MoveTool";
    } // toString


    /**
     * Returns the position of the mouse,this will return the
     * first Mouse Position or last Mouse Position depending
     * on the current Tool and the state of the tool
     *
     * @return Point position of the mouse
     */
     public Point getMousePosition()
     {
        return firstMousePos;
     }



  protected void formPopup(int x,int y)
  {

      relationList.removeAll();
      relationList.add(createMenuItem("paste"));
      if(propVertex != null)
      {
        relationList.add(createMenuItem("cut"));
        relationList.add(createMenuItem("copy"));
        relationList.addSeparator();
        relationList.add(createMenuItem("openEntity"));
        relationList.add(createRelationsMenu());
        relationList.add(createMenuItem("entityProperty"));
        relationList.add(createMenuItem("openCaseView"));
        relationList.add(createMenuItem("openCaseMatch"));
        relationList.addSeparator();
        relationList.add(createMenuItem("deactivate"));
        relationList.addSeparator();
        relationList.add(createMenuItem("delete"));
      }
      Visualizer viz = Visualizer.getFrame((Component)view);
      relationList.show(viz.getWin(document).getContentPane(), x, y);
  }

    protected JMenuItem createMenuItem( String cmd ) {
        MainWindow mw = MainWindow.instanceOf();
        // create menu item and set image and text
        JMenuItem mi = new JMenuItem
            ( mw.getResourceString( cmd + Visualizer.labelSuffix ) );
        URL url = mw.getResource( cmd + Visualizer.imageSuffix );
        if ( url != null ) {
            mi.setHorizontalTextPosition( JButton.RIGHT );
            mi.setIcon( new ImageIcon( url ) );
        }

        // set menu action command
        mi.setActionCommand( cmd );
        Action a = mw.getAction( cmd );
        if ( a != null ) {
            // set up the action to listen for events
            mi.addActionListener( a );
            mi.setEnabled( a.isEnabled() );
        } else {
            // no action! disable menu item
            mi.setEnabled( false );
        }
        return mi;
    } // createMenuItem

  protected JMenu createRelationsMenu()
  {
        JMenu rel = new JMenu(
          MainWindow.instanceOf().getResourceString("relations"+Visualizer.labelSuffix));
        URL url = MainWindow.instanceOf().getResource("relations"+ Visualizer.imageSuffix);
        if ( url != null ) {
            rel.setHorizontalTextPosition( JButton.RIGHT );
            rel.setIcon( new ImageIcon( url ) );
        }
        RelationType[] rels = propVertex.getEntity().getRelationTypes();

        for(int i = rels.length-1;i >= 0;i--)
        {
          Relation[] r = propVertex.getEntity().getLocalRelations(rels[i]);



//            actualRelations = (Entity.class.isAssignableFrom(r[j].getFrom().getClass()) && Entity.class.isAssignableFrom(r[j].getValue().getObject().getClass()));

          if(r.length > 0)
          {
            JCheckBoxMenuItem mi = new JCheckBoxMenuItem(rels[i].getName(), propVertex.relationIsActive(rels[i]));
            mi.addItemListener(relationListHandler);
            rel.add(mi);
          }
          else
          {
            JMenuItem mi = new JMenuItem(rels[i].getName());
            mi.setEnabled(false);
            rel.add(mi);
          }
        }
        return rel;
  }

  protected void toggleRelation(VisualEntity e, String relationName)
  {
      if(e.relationIsActive(relationName))
        deactivateRelation(relationName);
      else
        activateRelation(relationName);
  }

    public void deactivateRelation(String relationName)
    {
      Iterator iterator = propVertex.getEdgeIterator();
      Relation relation;
      VisualRelation visualRelation;
      while(iterator.hasNext())
      {
            visualRelation = (VisualRelation)iterator.next();
            relation = visualRelation.getRelation();
            if(!relation.isInherited() && (relation.getRelationType().getName().equals(relationName) && relation.getFrom() == propVertex.getEntity()) ||
              (relation.getRelationType().getInverseName().equals(relationName) && relation.getValue() == propVertex.getEntity()))
            {
                DeactivateRelationCommand delRelCmd = new DeactivateRelationCommand(document,visualRelation);
                if(document.performCommand(delRelCmd)==false)
                    JOptionPane.showMessageDialog
                        ( Visualizer.getFrame((Component)view),
                          MainWindow.instanceOf().getResourceString( "edgeCloseMsg" ),
                          MainWindow.instanceOf().getResourceString( "userError"+Visualizer.titleSuffix ),
                          JOptionPane.ERROR_MESSAGE );

            }
      }
    }

    protected void activateRelation(String relationType)
    {
      Relation[] rels = propVertex.getEntity().getRelations(relationType);
      for(int i=rels.length-1;i>=0;i--)
        if(!rels[i].isInherited())
          activateRelation(rels[i]);
    }

    protected void activateRelation(Relation r)
    {
      if(r.getFrom().equals(propVertex.getEntity()))
      {
          VisualEntity veto = activateEntity((Entity)r.getValue(),propVertex);
          ActivateRelationCommand arAction = new ActivateRelationCommand( document,new VisualRelation(propVertex, veto, r) );
          if ( document.performCommand( arAction ) == false )
          {
                JOptionPane.showMessageDialog
                    ( Visualizer.getFrame((Component)view),
                      MainWindow.instanceOf().getResourceString( "edgeAddMsg" ),
                      MainWindow.instanceOf().getResourceString("userError"+Visualizer.titleSuffix),
                      JOptionPane.ERROR_MESSAGE );
          }
      }
      else if(r.getValue().equals(propVertex.getEntity()))
      {
          VisualEntity vef = activateEntity((Entity)r.getFrom(),propVertex);
          ActivateRelationCommand arAction = new ActivateRelationCommand( document,new VisualRelation(vef, propVertex, r) );
          if ( document.performCommand( arAction ) == false )
          {
                JOptionPane.showMessageDialog
                    ( Visualizer.getFrame((Component)view),
                      MainWindow.instanceOf().getResourceString( "edgeAddMsg" ),
                      MainWindow.instanceOf().getResourceString("userError"+Visualizer.titleSuffix),
                      JOptionPane.ERROR_MESSAGE );
          }
      }
    }


    public VisualEntity activateEntity(Entity ed,VisualEntity activator)
    {
        VisualEntity ved = model.findVisualEntity(ed);
        if(ved == null)
        {
            DPoint tempPoint = view.activateEntity(activator);
            ved = new VisualEntity(ed,(int)tempPoint.x,(int)tempPoint.y);
            ActivateEntityCommand aed = new ActivateEntityCommand(document,ved);
            if(!document.performCommand(aed))
                return null;
        }
        return ved;
    }



/**
 * Methods For Drag and Drop
 */

    /**
     * DropTargetListener methods
     */
    public void drop(DropTargetDropEvent e)
    {
		Transferable trans;
		ConceptView cView;
                SelectionSet s=null;
		if( e.isDataFlavorSupported( dataFlavor ) == true ) {
			if( (e.getDropAction() & DnDConstants.ACTION_COPY_OR_MOVE) != 0 ) {
				if(e.isLocalTransfer() == false ) {
					System.err.println("Intra-JVM Transfers not implemented for figures yet.");
					return;
				}
				e.acceptDrop( e.getDropAction()  );
                                try
                                {
                                   s = (SelectionSet)(e.getTransferable().getTransferData(dataFlavor));
                                }
                                catch(UnsupportedFlavorException ex)
                                {
                                    Toolkit.getDefaultToolkit().beep();
                                    ex.printStackTrace();
                                }
                                catch(Exception exe)
                                {
                                    Toolkit.getDefaultToolkit().beep();
                                    exe.printStackTrace();
                                }
				if( s != null ) {
					cView = (ConceptView) e.getDropTargetContext().getComponent();
                                        Visualizer viz = Visualizer.getFrame((Component)cView);
                                        CreekDocument document = CreekDocumentGroup.instanceOf().getDocument(cView);
//                                        viz.getWin(document).setSelected(true);
                                        VertexList vl= s.getVertexList();
                                        RelationList rl = s.getRelationList();
                                        initialiseEntities(vl);
                                        initialiseRelations(rl);
                                        PasteCommand pc = new PasteCommand(document,vl,
                                                          rl,e.getLocation());
                                        try
                                        {
                                            if(!document.performCommand(pc))
                                                throw new Exception();
                                        }
                                        catch(Exception es)
                                        {
                                            es.printStackTrace();
                                        }

					e.getDropTargetContext().dropComplete(true);
				}
				else
					e.getDropTargetContext().dropComplete(false);

			}
			else
				e.rejectDrop();
            }
    }

    public void dragEnter(DropTargetDragEvent e)
    {
    }

    public void dragExit(DropTargetEvent e)
    {
    }
    public void dragOver(DropTargetDragEvent e)
    {
    }

    public void dropActionChanged(DropTargetDragEvent e)
    {
    }


    /**
     * DragGestureRecognized for DragGestureListener
     */
    public void dragGestureRecognized(DragGestureEvent e)
    {
                EntitySelection s=null;
		Component c = e.getComponent();
                if(selectedList.size()==0)
                    return;
		if(ConceptView.class.isInstance( c ) ) {
			boolean found = false;
			ConceptView cv = (ConceptView) c;
			Point p = e.getDragOrigin();
//			while ( itr.hasNext() ) {
//				if( ((Figure) itr.next()).containsPoint( p.x, p.y ) ) {
					found = true;
//					break;
//				}
//			}
                    if ( document != null )
                    {
                        VisualEntity ve[] = document.getModel().getSelectedEntities();
                        VertexList vl   = new  VertexList();
                        RelationList rl = new RelationList();
                        for(int k=0;k<ve.length;k++)
                        {
                            DPoint pt = ve[k].getPosition();
                            vl.add(new VisualEntity(ve[k].getEntity(),(int)pt.x,(int)pt.y));
                        }

                        for(int i =0;i<ve.length;i++)
                        {
                            Iterator iterator = ve[i].getEdgeIterator();
                            VisualRelation visualRelation ;
                            while(iterator.hasNext())
                            {
                                visualRelation = (VisualRelation)iterator.next();
                                if(isSelected(visualRelation,ve))
                                {
                                      rl.add(new VisualRelation(getVisualEntity(visualRelation.getFrom(),vl)
                                            ,getVisualEntity(visualRelation.getTo(),vl),
                                            visualRelation.getRelation()));
                                }

                            }
                        }
                        if(ve.length>0)
                        {
                              try
                              {
                                   s = new EntitySelection(new SelectionSet(vl,rl)
                                      ,dataFlavor);
                              }
                              catch(Exception ex)
                              {
                                  ex.printStackTrace();
                              }
                        }
                    }
		if( found == true ) {
				origin = p;

				e.getDragSource().startDrag(
								e,
								DragSource.DefaultCopyDrop,
								s,
								this);
			}
		}
    }


    /**
     * DragSourceListener Methods
     */
    public void dragDropEnd(DragSourceDropEvent e)
    {
    }

    public void dragEnter(DragSourceDragEvent e)
    {
    }

    public void dragExit(DragSourceEvent e)
    {
    }

    public void dragOver(DragSourceDragEvent e)
    {
    }

    public void dropActionChanged(DragSourceDragEvent e)
    {
    }


	/**
	 * Used to create the gesture recognizer which in effect turns on draggability.
	 */
	private void createDragGestureRecognizer(ConceptView dv, DragGestureListener dgl) {
		if(Component.class.isInstance( dv ) ) {
			Component c = (Component) dv;
			DragGestureRecognizer dgr =
						new MyDGR(
                                                        dragSource,
							c,
							DnDConstants.ACTION_COPY_OR_MOVE,
							this);
			dragGestureRecognizers.add( dgr );
		}
	}
	/**
	 * Used to destroy the gesture listener which ineffect turns off dragability.
	 */
	private void destroyDragGestureRecognizer(ConceptView dv, DragGestureListener dgl) {
		Iterator i = dragGestureRecognizers.iterator();
		while(i.hasNext()) {
			DragGestureRecognizer dgr = (DragGestureRecognizer)i.next();
			if ( dgr.getComponent() == dv ) {
				dgr.removeDragGestureListener( this );
				dgr.setComponent( null );
				i.remove();
				break;
			}
		}
	}


	public boolean canPerformAction(int action, Point location) {
              return true;
	}



  public boolean isSelected(VisualRelation vr,VisualEntity[] ves)
  {
      boolean from = false, to=false;
      for(int i=0;i<ves.length;i++)
      {
          if(ves[i]==vr.from)
              from=true;
          if(ves[i]==vr.to)
              to=true;
          if(from && to)
            return true;
      }

      return false;
  }

  public VisualEntity getVisualEntity(VisualEntity ve,VertexList vl)
  {
      for(int i=0;i<vl.size();i++)
      {
          if(ve.getEntity()==vl.get(i).getEntity())
              return vl.get(i);
      }

      return null;
  }

  protected void initialiseEntities(VertexList vl)
  {
      VisualEntity node;
      CreekDocumentGroup dg = CreekDocumentGroup.instanceOf();
      KnowledgeModel km = dg.getKM();
      for(int i=0;i<vl.size();i++)
      {
          node = vl.get(i);
//          node.getEntity() = (Entity)km.getConcept((node.getEntity().getName()));
      }
  }

  protected void initialiseRelations(RelationList rl)
  {
      VisualRelation edge;
      CreekDocumentGroup dg = CreekDocumentGroup.instanceOf();
      KnowledgeModel km = dg.getKM();
      AppSettings appSettings = AppSettings.instanceOf();
      for(int i=0;i<rl.size();i++)
      {
          edge = rl.get(i);
          Relation[] currentRelations =  edge.getFrom().getEntity().getRelations( );
          for(int k=0;k<currentRelations.length;k++)
              if((edge.getRelation().getName().equalsIgnoreCase(currentRelations[k].getName()))
              &&(edge.getTo().getEntity().getName().equalsIgnoreCase(((Entity)currentRelations[k].getValue()).getName()))
              &&(edge.getFrom().getEntity().getName().equalsIgnoreCase(((Entity)currentRelations[k].getFrom()).getName())))
              {
//                  edge.getRelation() = currentRelations[k];
//                  appSettings.addChangeListener(edge);
                  break;
              }
      }
  }

}//MoveTool








