package jcreek.cke.util;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * Contains some basic file utilities that could be very handy
 * in certain situations.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class FileUtilities {
	//
	// Date methods
	//

	/**
	 * A method that returns a date formatted as "yyyy-MM-dd HH:mm:ss:SSS".
   *
   * @return the formatted date as a string
	 */
	public static String getLogFormatedDate() {
		SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss:SSS");
		return formatter.format(new Date());
	}

	/**
	 * A method that returns a date formatted as "yyyyMMddHHmmssSSS".
   *
   * @return the formatted date as a string
	 */
	public static String getFileFormatedDate() {
		SimpleDateFormat formatter = new SimpleDateFormat("yyyyMMddHHmmssSSS");
		return formatter.format(new Date());
	}

	//
	// File methods
	//
	/**
	 * A method that guaranties that a directory path ends with a slash.
   *
	 * @param path the path to check and alter
	 * @return the path with ending slash
	 */
	public static String setPathSlash(String path) {
		if (path.endsWith("\\") || path.endsWith("/"))
			return path;
		else
			// since it does not mather to Java when doing filehandling if a path
			// has '\' or '/' as directory divider, then '/' will be used
			return path + "/";
	} // setPathSlash

	/**
	 * A method that returns a filename formatted as "[path]/yyyyMMddHHmmssSSS-[Name].xml".
   *
	 * @param path of the file
	 * @param name of the file
	 * @return the formatted file as a string
	 */
	public static String setXMLOutputFormat(String path, String name) {
		return setPathSlash(path) + getFileFormatedDate() + "-" + name + ".xml";
	} // setXMLOutputFormat

	/**
	 * Checks that the specified pathname is an existing file or directory.
   *
	 * @param pathname of the specified file or directory
	 * @return true if the specified file or directory exist, otherwise false.
	 */
	public static boolean exists(String pathname) {
		try {
			File file = new File(pathname);
			return file.exists();
		} catch (NullPointerException ex) {
			// no pathname give so it does not exists
			return false;
		}
	} // exists

	/**
	 * Checks if the specified pathname is a directory
   *
	 * @param pathname the pathname to check
	 * @return true if the pathname is a directory, otherwise false
	 */
	public static boolean isDirectory(String pathname) {
		try {
			File file = new File(pathname);
			return file.isDirectory();
		} catch (NullPointerException ex) {
			// no pathname give so it does not exists
			return false;
		}
	} // isDirectory

	/**
   * Checks if the specified filename is a file
   *
   * @param filename the filename to check
   * @return true if the filename is a file, otherwise false
	 */
	public static boolean isFile(String filename) {
		try {
			File file = new File(filename);
			return file.isFile();
		} catch (NullPointerException ex) {
			// no pathname give so it does not exists
			return false;
		}
	} // isFile

	/**
	 * Deletes the specified file or directory if exists. Note that the directory
   * must be empty to be removed.
   *
	 * @param pathname of the selected file or directory
	 * @return true if removal was successful, otherwise false
	 */
	public static boolean delFile(String filename) {
		try {
			File file = new File(filename);
			return file.delete();
		} catch (NullPointerException ex) {
			// no pathname give so it does not exists
			return false;
		}
	} // remove

	/**
	 * Gets the content of the specified file in string format
	 *
	 * @param filename name of the specified file
	 * @return the content of the specified file as string, does otherwise
   * return null in a case of an error
	 */
	public static String readFromFile(String filename) {
		try {
			BufferedReader in =
				new BufferedReader(new InputStreamReader(new FileInputStream(filename)));
			String retVal = "";
			String str = in.readLine();
			while (str != null) {
				retVal += str.trim() + '\n';
				str = in.readLine();
			}
			// close the stream
			in.close();
			return retVal;
		} catch (Exception ex) {
			System.err.println(ex.getMessage());
			return null;
		}
	} // readFromFile

	/**
	 * Gets the filenames and directories of the specified directory
   *
	 * @param pathname of the specified directory
	 * @return a string with the directory content delimited with a new line character, otherwise returns nothing if an error occurred
	 */
	public static String getDirContent(String pathname) {
		try {
			// create a file object
			File file = new File(pathname);
			// check if a directory
			String retVal = "";
			if (file.isDirectory()) {
				String files[] = file.list();
				for (int i = 0; i < files.length; i++)
					retVal += files[i] + '\n';
			}
			return retVal;
		} catch (Exception ex) {
			return "";
		}
	} // getDirContent

	/**
	 * Writes the content to the specified file.
   *
	 * @param filename name of the specified file
	 * @param content that should be written to the specified file
	 * @param append true if append to file if it exist or false if overwrite the file if it exist
	 * @return true if successful, otherwise false
	 */
	public static boolean writeToFile(
		String filename,
		String content,
		boolean append) {
		try {
			// create output stream
			DataOutputStream out =
				new DataOutputStream(
					new BufferedOutputStream(new FileOutputStream(filename, append)));
			// now write to file
			out.writeChars(content);
			// now flush the stream before close
			out.flush();
			// now close the stream
			out.close();
			// return ok
			return true;
		} catch (Exception ex) {
			return false;
		}
	} // writeToFile

	/**
	 * Writes the content to the specified file. Note that this method
   * overwrites the file it it exist.
   *
   * @param filename name of the specified file
   * @param content that should be written to the specified file
   * @return true if successful, otherwise false
	 */
	public static boolean writeToFile(String filename, String content) {
		return writeToFile(filename, content, false);
	} // writeToFile
}