package jcreek.cke.util;

import org.apache.regexp.RE;

/**
 * Contains some basic text utilities that could be very handy
 * in certain situations.
 *
 * @author Stein L. Tomassen, NTNU 2002
 */
public class TextUtilities {

	/**
	 * Replaces a matching regular expression pattern with a substitute text
	 *
	 * @param pattern the regular expression pattern to match
	 * @param txt the text that should be searched and replaced within
	 * @param substitute the text that should replace the matching text found
	 * @param caseDependent set to true if the matching against the specified pattern
   * should be case dependent, or set to false if it should be case independent
	 * @return the new text with the replacements where match where found
	 * @throws Exception if a pattern failure occurs
	 */
	public static String replaceMatch(
		String pattern,
		String txt,
		String substitute,
		boolean caseDependent)
		throws Exception {
		RE txtRE;
		if (caseDependent)
			txtRE = new RE(pattern);
		else
			txtRE = new RE(pattern, RE.MATCH_CASEINDEPENDENT);
		return txtRE.subst(txt, substitute);
	} // replaceMatch

	/**
	 * Strips a HTML file for it's HTML tags by using regular expressions. Does also convert
	 * some of the most common special characters into readable characters.
	 *
	 * @param html the HTML document to strip for HTML tags
	 * @return the document stripped for tags
	 * @throws Exception if an pattern failure occurs
	 */
	public static String stripHTML(String html) throws Exception {
		// patterns to substitute
		String commentPattern = "<!--(.|\\n)*?-->";
		String scriptPattern = "<(\\s)*?script(.|\n)*?/script(\\s)*?>";
		String htmlPattern = "<(.|\n)+?>";

		// remove comment tags before anye other html tags
		html = replaceMatch(commentPattern, html, "", false);
		// remove script tags before anye other html tags
		html = replaceMatch(scriptPattern, html, "", false);
		// remove ordinary html tags
		html = replaceMatch(htmlPattern, html, "", false);
		// replace some chars
		/**
		 * could be optimized by by doing a search for "&.{2,7};"
		 *  when found look up in a hash table with special chars
		 *  and replace that string found with the one found in the
		 *  hash table for that special char
		 *  Not a priority right now...
		 */
		html = replaceMatch("&aelig;", html, "", true);
		html = replaceMatch("&Aelig;", html, "", true);
		html = replaceMatch("&oslash;", html, "", true);
		html = replaceMatch("&Oslash;", html, "", true);
		html = replaceMatch("&aring;", html, "", true);
		html = replaceMatch("&Aring;", html, "", true);
		html = replaceMatch("&amp;", html, "&", true);
		html = replaceMatch("&eacute;", html, "", true);
		html = replaceMatch("&lt;", html, "<", true);
		html = replaceMatch("&gt;", html, ">", true);
		html = replaceMatch("&quot;", html, "\"", true);
		html = replaceMatch("&apos;", html, "'", true);
		html = replaceMatch("&laquo;", html, "", true);
		html = replaceMatch("&raquo;", html, "", true);
		html = replaceMatch("&deg;", html, "", true);
		html = replaceMatch("&nbsp;", html, " ", true);
		html = replaceMatch("&#124;", html, "|", true);
		html = replaceMatch("&copy;", html, "", true);
		html = replaceMatch("&#149;", html, "*", true);
		// remove those special chars not found, perhaps
		// not a very good idea, but...
		html = replaceMatch("&.{2,6};", html, "", true);
		// clean up
		// remove all double or more spaces
		html = replaceMatch("(\\s){3,}", html, "\n", true);
		return html;
	} // stripHTML

	/**
   * This method tries to find the start and end of a substring, by
   * specifying the content of the start and end of the substring. The
   * substring is specified by giving as parameters some of the text
   * at the start and the end of the substring. For example, lets say
   * you would search for the text between two HTML tags, e.g. '&lt;UL&gt;' and
   * '&lt;/UL&gt;', if the pattern where found the the starting and ending
   * position of the pattern are returned.<p>
   *
	 * Note, that this algorithm is a very simple one, and it should perhaps
   * be using regular expression for the string to match ot make it more robust.
	 *
	 * @param rawText the text to performe the serach within
	 * @param startPos the starting position for where in the rawText the
   * search should begin
	 * @param startPattern a string containing the text to match which should
   * be the start of the substring
	 * @param endPattern a string containing the text to match which should
   * be the end of the substring
	 * @param ignoreCase set to true if the matching against the specified pattern
   * should be case dependent, or set to false if it should be case independent
	 * @param includePatterns set to true if the patterns should be included
   * in the result.
	 * @return an array with two values, that is the starting and ending
   * positions respectively, if they where found. If no match where found,
   * null will be returned.
	 */
	public static int[] findSubStringFromToPos(
		StringBuffer rawText,
		int startPos,
		String startPattern,
		String endPattern,
		boolean ignoreCase,
		boolean includePatterns) {

		int start = findSubStringStartPos(rawText, startPos, startPattern, ignoreCase);
		if (start > 0) {
			int end = findSubStringStartPos(rawText, start, endPattern, ignoreCase);
			if (end > 0) {
				// set to last position
				end += endPattern.length();
			}
			// not include taggs?
			if (!includePatterns) {
				start += startPattern.length();
				end -= endPattern.length();
			}

			int retVal[] = { start, end };
			return retVal;
		}
		return null;
	} // findSubStringFromToPos

	/**
	 * Similar to findSubStringFromToPos, but returns only the ending position.
	 *
   * @param rawText the text to performe the serach within
   * @param startPos the starting position for where in the rawText the
   * search should begin
   * @param subString a string containing the text to match
   * @param ignoreCase set to true if the matching against the specified pattern
   * should be case dependent, or set to false if it should be case independent
   * @return the ending position of the substring found, does otherwise return -1
   * if no substring where found.
	 */
	public static int findSubStringEndPos(
		StringBuffer rawText,
		int startPos,
		String subString,
		boolean ignoreCase) {

		int endPos =
			findSubStringStartPos(rawText, startPos, subString, ignoreCase);

		if (endPos != -1) {
			return endPos + subString.length();
		} else {
			return endPos;
		}
	} // findSubStringEndPos

	/**
   * Similar to findSubStringFromToPos, but returns only the starting position.
	 *
   * @param rawText the text to performe the serach within
   * @param startPos the starting position for where in the rawText the
   * search should begin
   * @param subString a string containing the text to match
   * @param ignoreCase set to true if the matching against the specified pattern
   * should be case dependent, or set to false if it should be case independent
   * @return the starting position of the substring found, does otherwise return -1
   * if no substring where found.
	 */
	public static int findSubStringStartPos(
		StringBuffer rawText,
		int startPos,
		String subString,
		boolean ignoreCase) {

		// first do some basic checks
		if (rawText == null)
			return -1;
		else if (subString == null)
			return -1;

		// if this far, then continue...
		int length = rawText.length();
		StringBuffer sb;
		// if startPos is within the length of the text then proceed
		if (startPos < length) {
			try {
				// ignoreCase?
				if (ignoreCase) {
					sb = new StringBuffer(rawText.toString().toLowerCase());
					subString = subString.toLowerCase();
				} else {
					sb = rawText;
				}
				// set correct startPos
				if (startPos < 0)
					startPos = 0;
				char firstChar = subString.charAt(0);
				boolean found = false;
				int i = startPos;
				while ((!found) && (i < length)) {
					// first try to match the first char of the subString
					if (sb.charAt(i) == firstChar) {
						// then check the rest
						int ii = 1;
						boolean subFound = true;
						while ((subFound) && (ii < subString.length()) && (i + ii <= length)) {
							found = true;
							if (sb.charAt(i + ii) != subString.charAt(ii)) {
								// check if ignorable characters
								if (!(isIgnorableChar(sb.charAt(i + ii))
									|| isIgnorableChar(subString.charAt(ii)))) {
									// dismatch of characters, abort
									subFound = false;
								}
							}
							ii++;
						}
						if (subFound == found) {
							// match found
							return i;
						} else {
							// no match found, continue the search
							found = false;
							i += ii;
						}
					}
					i++;
				}
				return -1;
			} catch (Exception ex) {
				// if exception then out of range or something, therefore no match
				return -1;
			}
		} else {
			return -1;
		}
	} // findSubStringStartPos

	/**
	 * Checks if the specified character is an ignorable character. An ignorable
   * character is either a newLine, tab, form feed, return or a space.
   *
	 * @param chr the character to check
	 * @return true if the specified character is an ignorable
   * character, otherwise false is returned
	 */
	protected static boolean isIgnorableChar(char chr) {
		switch (chr) {
			case '\n' :
			case '\t' :
			case '\f' :
			case '\r' :
			case ' ' :
				return true;
			default :
				return false;
		}
	} // isIgnorableChar
}
