/*********************************************************************
 *
 * DESCRIPTION:
 *      This file defines the AppSettings class which holds a set of
 *      properties. Each property is stored with a key value that
 *      is used to access the property value. This class is also
 *      capable of saving the settings to a file and restoring from
 *      that file later.
 *
 *      We use the java.util.Properties class for our storage since
 *      it offers the load() and save() methods for easily storing
 *      the properties to a file.
 *
 ********************************************************************/
package jcreek.cke.visualizer;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.File;
import java.util.Properties;
import java.util.Vector;
import java.util.Enumeration;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;
import java.net.URL;
import java.util.Hashtable;
import java.util.StringTokenizer;

/**
 * Implements a properties storage class. This is used to store a variety
 * of properties, each associated with a key string. Each key string must
 * be unique or otherwise you will overwrite existing properties in the
 * table. The properties can be saved to a file and restored at a later
 * time, allowing persistent data storage.
 *
 * <p>
 * This class implements the Singleton design pattern to ensure that only
 * one instance of this class exists in the system. To get the single
 * instance you can call the instanceOf() method.
 * </p>
 *
 * @author
 * @version
 *
 */

public class AppSettings
{

    public static final String RELATION_NAME = "relationName";
    public static final String ISSUBCLASS = "SubclassOfRelation";
    public static final String ISINSTANCE = "InstanceOfRelation";
    public static final String HORIZONTAL_ALLIGNMENT = "horizontalAllign";
    public static final String VERTICAL_ALLIGNMENT = "verticalAllign";

    /**
     * Name of the property that determines which look & feel
     * is to be used in the user interface. Accepts string values.
     */
    public static final String LOOK_FEEL = "lookFeel";

    protected Properties table;

    protected EventListenerList changeListeners;

    protected static AppSettings instance;

    protected String iniFileName;

    protected AppSettings()
   {
	table = new Properties();
        changeListeners = new EventListenerList();
   }


    public void addChangeListener( ChangeListener listener )
   {
        changeListeners.add( ChangeListener.class, listener );
        listener.stateChanged( new ChangeEvent( this ) );
   }

    public void clear()
   {
	table.clear();
    }

    public void commit()
   {
	if ( iniFileName != null )
	{
	    save( iniFileName, "Option Settings" );
	}
	 else
	{
	    save( "Creek.ini", "Option settings" );
	}
        fireChange();
    }

    public boolean contains( String key )
   {
	return table.containsKey( key );
    }

    protected void fireChange()
   {
      if ( changeListeners == null )
          return;
      ChangeEvent c = new ChangeEvent( this );
      Object[] listeners = changeListeners.getListenerList();
      for ( int i = listeners.length - 2; i >= 0; i -= 2 )
      {
        if ( listeners[i] == ChangeListener.class )
        {
            ChangeListener cl = (ChangeListener)listeners[i+1];
            cl.stateChanged( c );
        }
      }
  }


    public boolean getBoolean( String key )
   {
	String sVal = table.getProperty( key );
	if ( sVal == null )
	{
	    return false;
	}
	Boolean bVal = Boolean.valueOf( sVal );
	return bVal.booleanValue();
    }

    public double getDouble( String key )
   {
	String sVal = table.getProperty( key );
	if ( sVal == null )
	{
	    return 0;
	}
	Double dVal = Double.valueOf( sVal );
	return dVal.doubleValue();
    }

    public String getString( String key )
   {
        String sVal = table.getProperty( key );
        if ( sVal == null )
       {
            return "";
        }
	return sVal;
    }

    public String getCurrentDirectory()
    {
      String dir = getString("currentDirectory");
      if(dir == null || dir.equals(""))
        dir = System.getProperty("user.dir");
      return dir;
    }

    public void setCurrentDirectory(String dir)
    {
      System.out.println(" dir = '"+dir+"' - '"+File.separator+"' - "+dir.lastIndexOf(File.separator));
      if(dir.lastIndexOf(File.separatorChar) > -1)
        dir = dir.substring(0, dir.lastIndexOf(File.separatorChar));
      System.out.println("setCurrentDirectory to '"+dir+"'");
      setString("currentDirectory", dir);
      commit();
    }

    public static AppSettings instanceOf()
   {
        if ( instance == null )
       {
            instance = new AppSettings();
       }
        return instance;
    }

    public boolean isEmpty()
    {
	return table.isEmpty();
    }

    public boolean load( String fileName )
   {
	try
        {
	    FileInputStream fis = new FileInputStream( fileName );
	    table.load( fis );
	}
	catch ( FileNotFoundException e )
	{
	    return false;
	}
	catch ( IOException e )
	{
	    return false;
	}
	iniFileName = fileName;
        fireChange();
	return true;
    }

    public void remove( String key )
   {
	table.remove( key );
    }

    public void removeChangeListener( ChangeListener listener )
   {
        changeListeners.remove( ChangeListener.class, listener );
    }

    public boolean save( String fileName, String comment )
   {
	try
	{
	    FileOutputStream fos = new FileOutputStream( fileName );
	    table.store( fos, comment );
	}
	catch ( IOException e )
	{
	    return false;
	}
        if ( iniFileName == null )
       {
            iniFileName = fileName;
        }
	return true;
    }

    public void setBoolean( String key, boolean value )
   {
	Boolean bVal = new Boolean( value );
	String sVal = bVal.toString();
	table.put( key, sVal );
    }

    public void setDouble( String key, double value )
   {
	Double dVal = new Double( value );
	String sVal = dVal.toString();
	table.put( key, sVal );
    }

    public void setString( String key, String value )
    {
	table.put( key, value );
    }

   public void setRelation( String key, int value )
   {
        table.put(key, Integer.toString(value));
    }

    public int getRelation(String key)
    {
        String sVal = table.getProperty(key );
        if ( sVal == null )
            return 0;
        return Integer.parseInt(sVal);
    }

    //rh doc:
    //this function is called in the visualizer createtaxy() when the overview buuton is selected

    public String[] getSelectedRelations()
    {
        Vector relat =  new Vector();
        String relationClassName,tmpRelationClassName;
        Enumeration relNames =  table.propertyNames();
        boolean bVal = false;
        while(relNames.hasMoreElements())
        {
            relationClassName = (String)relNames.nextElement();
            if(relationClassName.endsWith("Relation"))
            {
              bVal=getBoolean(relationClassName);
              tmpRelationClassName = (String) relationClassName.substring(0,relationClassName.indexOf("Relation"));
              if(bVal)
                relat.addElement(tmpRelationClassName);
            }
        }
        String[] relationClasses = new String[relat.size()];
        relat.copyInto(relationClasses);
        return relationClasses;
    }

    public int size()
    {
	return table.size();
    }


}


