package jcreek.cke.visualizer;

import jcreek.cke.util.*;
import jcreek.cke.graph.*;
import jcreek.cke.conceptview.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


public  class DragTool implements ConceptViewListener
 {
    /**
     * View that we're associated with.
     */
    protected ConceptView view;

    /**
     * Vertex the user pressed on in mousePressed(). Could be
     * null if the user didn't press on a entity.
     */
    protected VisualEntity pressVertex;
    /**
     * List of the selected entities in the order they were
     * selected by the user.
     */
    protected VertexList selectedList;
    /**
     * True if mouse was pressed on the view before being released.
     * Helps avoid some mis-firings of the mouse buttons.
     */
    protected boolean wasPressed;
    /**
     * The list of the entities being moved around. These
     * entities are a copy of the originals and can be modified.
     */
    protected VertexList movingVertices;
    /**
     * The list of the relations being moved around.
     */
    protected Point firstMousePos;
    /**
     * Most recent mouse position while dragging.
     */
    protected Point lastMousePos;

    /**
     * Creates a new DragTool object.
     */
    public DragTool() {
        selectedList = new VertexList();
        firstMousePos = new Point();
        lastMousePos = new Point();
    } // DragTool

    public void attach( ConceptView view) {
        this.view = view;
        view.addViewListener(this);
    }

    public void detach() {
        if ( view != null ) {
            view.removeViewListener( this );
            view = null;
        }
    }



    public void mouseClicked( MouseEvent e ) {
    } // mouseClicked

    public void mouseDragged( MouseEvent me ) {
        // Ignore drags without matching presses, and ignore drags
        // of insignificant movement (i.e. less than 3 pixels).

        DPoint p = new DPoint(me.getX(),me.getY());
        p.transform(view.getMoveTransform());

        if ( !wasPressed ) {
            return;
        }
        lastMousePos.x =  me.getX();
        lastMousePos.y =  me.getY();

        if ( ( movingVertices == null ) &&
             ( Math.abs( lastMousePos.x - firstMousePos.x ) < 3 ) &&
             ( Math.abs( lastMousePos.y - firstMousePos.y ) < 3 ) ) {
            // We ignore tiny mouse movements until we being
            // doing something, such as moving vertices or
            // selecting a region.
            return;
        }

        // First make sure the vertex the user clicked on is
        // selected (putting it in the selected list). This
        // makes it possible to move several vertices at once.
        if ( ( pressVertex != null ) &&
             ( pressVertex.isSelected() == false ) ) {
             pressVertex.setSelected(true);
            selectedList.add( pressVertex );
        }

        // If there are some vertices selected, deal with moving
        // them around the view.
        if ( ( pressVertex != null ) && ( selectedList.size() > 0 ) ) {
            if ( movingVertices == null ) {
                // Make a copy of the selected vertices so we can
                // modify their positions based off the originals.
                // We'll use the copy in resetMovingVertices() to
                // restore the original positions.
                movingVertices = selectedList.copy();
            }
                VisualEntity v1 = pressVertex;
                VisualEntity v2 = pressVertex;
                DPoint pos = v1.getPosition();
                pos.transform(view.getViewTransform());
                int deltaX = lastMousePos.x - (int)pos.x;
                int deltaY = lastMousePos.y - (int)pos.y;
            int size = movingVertices.size();
            for ( int i = 0; i < size; i++ ) {
                // Move the original vertices based off the
                // position of the copies.
                v1 = selectedList.get( i );
                v2 = movingVertices.get( i );
                v1.slide(deltaX,deltaY);
            }
        }
        view.repaint();

    } // mouseClicked

    public void mouseEntered( MouseEvent e ) {
    } // mouseClicked

    public void mouseExited( MouseEvent e ) {
    } // mouseClicked

    public void mouseMoved( MouseEvent e ) {
    } // mouseClicked

    public void mousePressed( MouseEvent me ) {
        if ( !me.isControlDown() ) {
            // If user is not holding down the Ctrl key, deselect all.
            for(int i=0;i<selectedList.size();i++)
                  ((VisualEntity)selectedList.get(i)).setSelected(false);
            if ( selectedList.size() > 0 ) {
                selectedList.clear();
            }

        }
        // Remember the mouse position for when we have to
        // move entities around the view.
        firstMousePos.x = me.getX();
        firstMousePos.y = me.getY();
        lastMousePos.x = firstMousePos.x;
        lastMousePos.y = firstMousePos.y;
        pressVertex = view.findVisualEntity((double)firstMousePos.x,(double)firstMousePos.y);
        wasPressed = true;
    } // mouseClicked

    public void mouseReleased( MouseEvent me ) {
        if ( wasPressed == false ) {
            // This avoids the case where the user clicks outside
            // of the view and drags the mouse onto the view.
            // This takes care of menu selections or clicking
            // the "Ok" button of a dialog shown over the view.
            return;
        }
        wasPressed = false;
        int mX = me.getX();
        int mY = me.getY();
        // See if user released on an edge.
        if ( movingVertices != null ) {
/*            // We were moving vertices around the screen.
            resetMovingVertices();
            double deltaX = mX - firstMousePos.x;
            double deltaY = mY - firstMousePos.y;
            EntityMoveCommand action = new EntityMoveCommand
                ( document, selectedList, ((int)deltaX),((int) deltaY), 0 );
            document.performCommand( action );
*/
        } else if ( pressVertex != null ) {
            // Pressed on a vertex, toggle the vertex selection
            // and remember the last vertices selected.
            if ( pressVertex.isSelected() ) {
                pressVertex.setSelected(false);
                selectedList.remove(pressVertex);
            } else {
                pressVertex.setSelected(true);
                selectedList.add( pressVertex );
            }
        }
        finishMoving();
        pressVertex = null;
    } // mouseClicked


    /**
     * If we were moving entities, clean up now.
     */
    protected void finishMoving() {
        if ( movingVertices != null ) {
            movingVertices.clear();
            movingVertices = null;
            //got to add a recal method
            view.fireContentsResize();
            view.repaint();
        }

    } // finishMoving

    /**
     * Returns the list of selected entities, in the order they
     * were selected by the user. Often called by another class
     * that is going to perform some operation on the graph,
     * such as a search.
     *
     * @return  selected entity list
     */
    public VertexList getSelected() {
        return selectedList;
    } // getSelected




 }//DragTool
