package jcreek.cli;

import org.jatha.*;
import org.jatha.compile.LispPrimitive;
import org.jatha.compile.*;
import org.jatha.dynatype.*;
import org.jatha.machine.*;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.representation.inference.*;
import java.util.*;
import java.io.IOException;

/**
 * Extends a Jatha LISP machine with primitives to access the jcreek.representation
 * package. This serves as a Command Line interface to the Creek representation.
 *
 * These primitives are added:
 * (creek-entity '<entityID/name>)  ==> (<name> <description>)
 * (creek-relation '<relationID>) ==> (<from> <type> <value> <strength> <isInherited>)
 * (creek-relations '<entityID/name>)  ==> (<from> <type> <value> <strength> <isInherited>)*
 * (creek-explain '<relationID>) ==> ((((<from> <type> <value> <strength> <isInherited>)*) <strength>)*)
 * (creek-entities) ==> (<entityName>*)
 * (creek-entity-object '<id/name>) ==> <entity object>
 *
 * (creek-add-entity '<name> '<description>) ==> T/nil
 * (creek-add-relation '<from> '<type> '<value>) ==> T/nil
 * (creek-set-relation-strength' <name> <strength>) ==> T/nil
 * (creek-set-entity-name '<entityID/name> <name) ==> T/nil
 * (creek-set-entity-description '<entityID/name> <description>) ==> T/nil
 * (creek-set-entity-object '<entityID/name> <entity object>) ==> T/nil
 * (creek-remove-entity '<entityID/name>) ==> T/nil
 * (creek-remove-relation '(<fromID> <typeID> <valueID> [strength] [isInherited])) ==> T/nil
 *
 * (creek-load '<.KM file>) ==> T/nil
 * (creek-save '<.KM file>) ==> T/nil
 *
 * NOT IMPLEMENTED YET: (creek-register-model-listener (<lambda-exp> (<changeID> <sourceID>)) ==> <listenerID>
 * NOT IMPLEMENTED YET: (creek-deregister-model-listener <listenerID>) ==> T
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class CreekJatha extends Jatha
{
  KnowledgeModel km;
  public CreekJatha(KnowledgeModel km)
  {
    super(true, true);
    this.km = km;
  }

  public void setKnowledgeModel(KnowledgeModel km)
  {
    this.km = km;
  }

  public KnowledgeModel getKnowledgeModel()
  {
    return km;
  }

  public void init()
  {
    super.init();
    COMPILER.Register(new CreekEntity(this));
    COMPILER.Register(new CreekRelations(this));
    COMPILER.Register(new CreekRelation(this));
    COMPILER.Register(new CreekEntities(this));
    COMPILER.Register(new CreekExplain(this));
    COMPILER.Register(new CreekEntityObject(this));
    COMPILER.Register(new CreekAddEntity(this));
    COMPILER.Register(new CreekAddRelation(this));
    COMPILER.Register(new CreekRemoveRelation(this));
    COMPILER.Register(new CreekRemoveEntity(this));
    COMPILER.Register(new CreekSetEntityName(this));
    COMPILER.Register(new CreekSetEntityDesc(this));
    COMPILER.Register(new CreekSetRelationStrength(this));
    COMPILER.Register(new CreekSetEntityObject(this));
    COMPILER.Register(new CreekSave(this));
    COMPILER.Register(new CreekLoad(this));
  }

  public static void main(String argv[])
  {

      LocalKnowledgeModel km = new LocalKnowledgeModel();

      CreekJatha cj = new CreekJatha(km);
      cj.init();
      cj.start();

  }

  public LispValue makeRelation(Relation rel)
  {
    ArrayList list = new ArrayList();
    list.add(toLisp(rel.getFrom().getName()));
    list.add(toLisp(rel.getRelationType().getName()));
    list.add(toLisp(rel.getValue().getName()));
    list.add(toLisp(new Double(rel.getExplanationStrength())));
    list.add(toLisp(new Boolean(rel.isInherited())));

    LispValue lispList = makeList(list);
    return lispList;
  }

  public Relation relationToJava(LispValue arg)
  {
    if(arg.basic_listp())
    {
      Collection relList = (Collection) arg.toCollection();
      if(relList.size() >= 3)
      {
        try
        {
          Iterator it = relList.iterator();
          String from = (String) ((LispValue) it.next()).toJava();
          String type = (String) ((LispValue) it.next()).toJava();
          String value = (String) ((LispValue) it.next()).toJava();

          Entity fromEnt = km.getEntity(from);
          Relation[] rels = fromEnt.getRelations(type);
          for(int i=0;i<rels.length;i++)
            if(rels[i].getRelationType().getName().equalsIgnoreCase(type) && rels[i].getValue().getName().equalsIgnoreCase(value))
              return rels[i];
        }
        catch(ClassCastException e) { }
        catch(NullPointerException e) { }
      }
    }
    return null;
  }
}

class CreekEntity extends LispPrimitive
{
//  CreekJatha lisp;
  public CreekEntity(CreekJatha lisp)
  {
    super(lisp, "CREEK-ENTITY", 1);   // 1 argument
//    this.lisp = lisp;
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue arg = machine.S.pop();

    System.out.println("Arg is "+arg.toString()+" as Java '"+arg.toJava().toString()+"' "+arg.toJava().getClass());
    Entity ent = km.getEntity(arg.toJava());
    if(ent != null)
    {
      System.out.println("Found "+ent);
      LispValue list = getLisp().makeList(getLisp().makeString(ent.getName()), getLisp().makeString(ent.getDescription()));
      System.out.println("=> "+list.toString());
      machine.S.push(list);
    }
    else
      machine.S.push(getLisp().NIL);

    machine.C.pop();
  }
}

class CreekRelations extends LispPrimitive
{
//  CreekJatha lisp;
  public CreekRelations(CreekJatha lisp)
  {
    super(lisp, "CREEK-RELATIONS", 1);   // 1 argument
//    this.lisp = lisp;
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue arg = machine.S.pop();

//    System.out.println("Arg is "+arg.toString()+" as Java '"+arg.toJava().toString()+"' "+arg.toJava().getClass());
    Entity ent = km.getEntity(arg.toJava());
    if(ent != null)
    {
      ArrayList list = new ArrayList();
      Relation[] rels = ent.getRelations();
      for(int i=0;i<rels.length;i++)
        list.add(((CreekJatha) getLisp()).makeRelation(rels[i]));

      LispValue lispList = getLisp().makeList(list);

      System.out.println("=> "+lispList.toString());
      machine.S.push(lispList);
    }
    else
      machine.S.push(getLisp().NIL);

    machine.C.pop();
  }
}

class CreekRelation extends LispPrimitive
{
  public CreekRelation(CreekJatha lisp)
  {
    super(lisp, "CREEK-RELATION", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue arg = machine.S.pop();

//    System.out.println("Java arg: "+arg.toJava().getClass()+" '"+arg.toJava().toString()+"'");
    Relation rel = km.getRelation(arg.toJava());
    if(rel != null)
    {
      machine.S.push(((CreekJatha) getLisp()).makeRelation(rel));
    }
    else
      machine.S.push(getLisp().NIL);

    machine.C.pop();
  }
}

class CreekEntities extends LispPrimitive
{
  public CreekEntities(CreekJatha lisp)
  {
    super(lisp, "CREEK-ENTITIES", 0);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();

    Iterator it = km.getEntityIterator();
    ArrayList list = new ArrayList();
    while(it.hasNext())
      list.add(getLisp().makeString(((Entity) it.next()).getName()));

    machine.S.push(getLisp().makeList(list));
    machine.C.pop();
  }
}

class CreekExplain extends LispPrimitive
{
  public CreekExplain(CreekJatha lisp)
  {
    super(lisp, "CREEK-EXPLAIN", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue arg = machine.S.pop();

    Relation myRel = ((CreekJatha) getLisp()).relationToJava(arg);

    if(myRel != null && myRel.isInherited())
    {
      try
      {
        Explanation[] exps = myRel.getExplanations();
        ArrayList explanationList = new ArrayList();
        for(int i=0;i<exps.length;i++)
        {
          ArrayList thisExp = new ArrayList();
          for(int j=0;j<exps[i].size();j++)
            thisExp.add(((CreekJatha) getLisp()).makeRelation(exps[i].relationAt(j)));
          ArrayList withExp = new ArrayList();
          withExp.add(((CreekJatha) getLisp()).makeList(thisExp));
          withExp.add(((CreekJatha) getLisp()).toLisp(new Double(exps[i].getStrength())));
          explanationList.add(((CreekJatha) getLisp()).makeList(withExp));
        }
        machine.S.push(getLisp().makeList(explanationList));
        machine.C.pop();
        return;
      }
      catch(TaskTooComplexException e)
          { }
    }

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekEntityObject extends LispPrimitive
{
  public CreekEntityObject(CreekJatha lisp)
  {
    super(lisp, "CREEK-ENTITY-OBJECT", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue arg = machine.S.pop();

    Entity ent = km.getEntity(arg.toJava());
    if(ent != null && ent.getEntityObject() != null)
      machine.S.push(getLisp().toLisp(ent.getEntityObject()));
        else
      machine.S.push(getLisp().NIL);

    machine.C.pop();
  }
}

class CreekAddEntity extends LispPrimitive
{
  public CreekAddEntity(CreekJatha lisp)
  {
    super(lisp, "CREEK-ADD-ENTITY", 2);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue desc = machine.S.pop();
    LispValue name = machine.S.pop();

    try
    {
      new Entity(km, (String) name.toJava(), (String) desc.toJava());
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NameAlreadyExistException e) { }
    catch(ClassCastException e) { }

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekAddRelation extends LispPrimitive
{
  public CreekAddRelation(CreekJatha lisp)
  {
    super(lisp, "CREEK-ADD-RELATION", 3);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue value = machine.S.pop();
    LispValue type = machine.S.pop();
    LispValue from = machine.S.pop();

    try
    {
      new Relation(km.getEntity(from.toJava()), km.getRelationType((String) type.toJava()), km.getEntity(value.toJava()));
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { }
    catch(ClassCastException e) { }

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekSetEntityName extends LispPrimitive
{
  public CreekSetEntityName(CreekJatha lisp)
  {
    super(lisp, "CREEK-SET-ENTITY-NAME", 2);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue newName = machine.S.pop();
    LispValue entityID = machine.S.pop();

    try
    {
      km.getEntity(entityID.toJava()).setName((String) newName.toJava());
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { e.printStackTrace(); }
    catch(ClassCastException e) { e.printStackTrace(); }
    catch(NameAlreadyExistException e) { e.printStackTrace();}

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekSetEntityDesc extends LispPrimitive
{
  public CreekSetEntityDesc(CreekJatha lisp)
  {
    super(lisp, "CREEK-SET-ENTITY-DESCRIPTION", 2);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue newName = machine.S.pop();
    LispValue entityID = machine.S.pop();
    try
    {
      km.getEntity(entityID.toJava()).setDescription((String) newName.toJava());
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { }
    catch(ClassCastException e) { }

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekSetEntityObject extends LispPrimitive
{
  public CreekSetEntityObject(CreekJatha lisp)
  {
    super(lisp, "CREEK-SET-ENTITY-OBJECT", 2);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue entityID = machine.S.pop();
    LispValue newName = machine.S.pop();

    try
    {
      Object ob = newName.toJava();
      km.getEntity(entityID.toJava()).setEntityObject((java.io.Serializable) ob);
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { }
    catch(ClassCastException e) { }

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekRemoveEntity extends LispPrimitive
{
  public CreekRemoveEntity(CreekJatha lisp)
  {
    super(lisp, "CREEK-REMOVE-ENTITY", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue entityID = machine.S.pop();

    try
    {
      km.getEntity(entityID.toJava()).remove();
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { e.printStackTrace();}
    catch(ClassCastException e) { e.printStackTrace();}

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekSetRelationStrength extends LispPrimitive
{
  public CreekSetRelationStrength(CreekJatha lisp)
  {
    super(lisp, "CREEK-SET-RELATION-STRENGTH", 2);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue newName = machine.S.pop();
    LispValue relationID = machine.S.pop();

    try
    {
      Relation rel = ((CreekJatha) getLisp()).relationToJava(relationID);
      rel.setExplanationStrength(((Double) newName.toJava("Double")).doubleValue());
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { e.printStackTrace();}
    catch(ClassCastException e) { e.printStackTrace();}

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekRemoveRelation extends LispPrimitive
{
  public CreekRemoveRelation(CreekJatha lisp)
  {
    super(lisp, "CREEK-REMOVE-RELATION", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    KnowledgeModel km = ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue relationID = machine.S.pop();

    try
    {
      Relation rel = ((CreekJatha) getLisp()).relationToJava(relationID);
      rel.remove();
      machine.S.push(getLisp().T);
      machine.C.pop();
      return;
    }
    catch(NullPointerException e) { }
    catch(ClassCastException e) { }

    machine.S.push(getLisp().NIL);
    machine.C.pop();
  }
}

class CreekLoad extends LispPrimitive
{
  public CreekLoad(CreekJatha lisp)
  {
    super(lisp, "CREEK-LOAD", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    LocalKnowledgeModel km = (LocalKnowledgeModel) ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue fname = machine.S.pop();
    try
    {
      String fn = (String) fname.toJava();
      km.load(fn);
    }
    catch(Exception e)
    {
      e.printStackTrace();
      machine.S.push(getLisp().NIL);
      machine.C.pop();
      return;
    }
    machine.S.push(getLisp().T);
    machine.C.pop();
  }
}

class CreekSave extends LispPrimitive
{
  public CreekSave(CreekJatha lisp)
  {
    super(lisp, "CREEK-SAVE", 1);   // 1 argument
  }

  public void Execute(SECDMachine machine)
  {
    LocalKnowledgeModel km = (LocalKnowledgeModel) ((CreekJatha) getLisp()).getKnowledgeModel();
    LispValue fname = machine.S.pop();
    try
    {
      String fn = (String) fname.toJava();
      km.saveAs(fn);
    }
    catch(IOException e)
    {
      e.printStackTrace();
      machine.S.push(getLisp().NIL);
      machine.C.pop();
      return;
    }
    machine.S.push(getLisp().T);
    machine.C.pop();
  }
}
