package jcreek.examples;

import javax.swing.JFrame;
import java.io.IOException;

import jcreek.representation.KnowledgeModel;
import jcreek.representation.LocalKnowledgeModel;
import jcreek.representation.cbr.CBRModel;

import jcreek.representation.Entity;
import jcreek.representation.EntityTypeNotMatchedException;
import jcreek.representation.Relation;
import jcreek.representation.RelationType;
import jcreek.representation.NoSuchRelationTypeException;
import jcreek.representation.cbr.Case;
import jcreek.representation.SubModel;
import jcreek.representation.NumberEntity;
import jcreek.representation.NameAlreadyExistException;

import jcreek.reasoning.RetrieveResult;
import jcreek.reasoning.CaseComparison;
import jcreek.reasoning.ReuseResult;

import jcreek.gui.representation.FramePane;
import jcreek.gui.representation.CasePane;
import jcreek.gui.reasoning.RetrieveResultPane;
import jcreek.gui.reasoning.CaseComparisonPane;
import jcreek.gui.SubModelView;

import jcreek.util.CreekException;

/**
 * This class in an example on how to use jCreek through the Java API. It first
 * demonstrates creating a new KnowledgeModel using the jcreek.representation
 * package. Then it demonstrates, on this model, using a the jcreek.reasoning
 * engine to use Case-Based Reasoning on a sample problem case. Last, it shows
 * off some standard GUI elements of the jcreek.gui packages.
 *
 * The implementation is very heavily commented and may be read as a tutorial
 * to jCreek.
 *
 * @author Frode Srmo, small changes by Stein Brandser
 * @version 2.0
 */

public class CreekExample
{
  private CreekExample()
  {
  }

  public static void main(String argv[])
  {
    // Make some options to turn off GUI and I/O elements of this demo.
    boolean showGUI = true;
    boolean disk = true;
    if((argv.length > 0 && argv[0].equals("-nogui")) ||
       (argv.length > 1 && argv[1].equals("-nogui")))
      showGUI = false;
    if((argv.length > 0 && argv[0].equals("-nodisk")) ||
       (argv.length > 1 && argv[1].equals("-nodisk")))
      disk = false;

    long startupTime = System.currentTimeMillis();

    // Creating a new knowledge model.
    // This does not create an empty model, but defines a bare-bones model as
    // defined by jcreek.representation.CBRModel.
    // As KnowledgeModel is now an interface, we need to use an implementation
    // of the interface when making a new model.

    System.out.println("Making a new KnowledgeModel");
    KnowledgeModel km = new LocalKnowledgeModel();

    // First, we create a small model of our chosen domain, where we want to
    // diagnose car starting problems.
    try {
      addAttributes(km);
      addCausalModel(km);
      addCases(km);
    }
    catch (NameAlreadyExistException e) {
      // As Entities in jCreek must have unique names, it may happen that we try
      // to make an entity with a name that already exist. If this happens, this
      // exception is thrown.
      e.printStackTrace();
    }
    catch (NoSuchRelationTypeException e) {
      // When we refer to a relation type by name (a string) instead of using
      // a RelationType object, it may be that we use a relation type that is
      // not defined, e.g. through a typo. If so, this exception is thrown.
      e.printStackTrace();
    }
    // This concludes the generation of the simple model.

    displayModel(km, showGUI);
    int geSize = km.getInferenceCache().size();
    if(disk)
    {
      saveModel(km, "creekExample.km");
      km = loadModel("creekExample.km");
    }
    solveCBR(km, showGUI);

    System.out.println("\n=============================================");
    System.out.println("CreekExample finisihed in " +
                       (System.currentTimeMillis()-startupTime)+" ms.");
  }

  /**
   * Improvement to km.getEntity (more meaningful exception than NullPointer).
   * @todo Redesign km.getEntity (and fix broken code) instead of wrapping it.
   * @param km KnowledgeModel
   * @param name String
   * @return Entity
   * @author Stein Brandser
   */
  public static Entity getEntity(KnowledgeModel km, String name) {
    Entity entity = km.getEntity(name);
    if (entity == null) {
      // If you get an exception thrown from the line below, fix the calling
      // code (as given by the next line in the call stack). Let is ask for an
      // entity that actually exists.
      throw new IllegalArgumentException(
       "The entity `"+name+"' was not found in the model `"+km.getName()+"'."
      );
    }
    return entity;
  }


  /**
   * Generates a new KnowledgeModel for a simple car-start problem domain,
   * part 1 of 3.
   */
  public static void addAttributes(KnowledgeModel km)
      throws NameAlreadyExistException, NoSuchRelationTypeException
  {
    // We define the attributes the car case usually have. These are colour,
    // engine status, battery status and age.

    // Defining colour.
    Entity colour = new Entity(km, "colour", "the color of something");

    // The colour attribute is of type symbol. This means that its values are
    // not numeric but nominal (e.g "red" not #FF00000 or 0.74).
    Entity symbol = km.getEntity(CBRModel.SYMBOL);
    colour.addRelation(CBRModel.SUBCLASS_OF, symbol);

    // Defines the attributes of colour - red and blue.
    Entity red = new Entity(km, "red", "the colour red");
    Entity blue = new Entity(km, "blue", "the colour blue");

    // The "has instance" relation is used to describe that an entity is value
    // of a given attribute, e.g. "red" beeing a value of "colour" in this case.
    colour.addRelation(CBRModel.HAS_INSTANCE, red);
    colour.addRelation(CBRModel.HAS_INSTANCE, blue);

    // Last, we create a special relation type (slot) that is used in the
    // description of a car case. This relation type is a subclass of the
    // CBRModel.FINDING_RELATION relation type, and that means that when
    // this relation is found in a case it is taken as one of the properties
    // of the case that should be used for case matching. Other relations, such
    // as "instance of" and "has solution" is not used in the case matching
    // by the jcreek.reasoning package.
    // When creating a new relation-type we also at the same time define an
    // inverse relation.
    // In a finding-type relation, the strength of the relation is equal to the
    // importance (for the relation from the case) and the predictive strength
    // (for the relation to the case).
    RelationType hasColor = new RelationType(
        km,
        // The superclass of this relation:
        km.getRelationType(CBRModel.FINDING_RELATION),
        // The name, description and default strength of the relation:
        "has colour",
        "the color of an object",
        0.1,
        // Ditto for the inverse:
        "colour of",
         "the origin is the color of the target",
         0.1
        );


    // Defines the "engine status" attribute. This is also a symbol type attribute.
    Entity engineStatus = new Entity(km, "engine status",
                                     "the engine status of a car");
    engineStatus.addRelation(CBRModel.SUBCLASS_OF, symbol);

    // Defines the possible values of engine status, "running" and "not starting".
    Entity engineRunning = new Entity(km, "running", "The engine is OK");
    engineStatus.addRelation(CBRModel.HAS_INSTANCE, engineRunning);
    Entity engineNotStarting = new Entity(km, "not starting",
                                          "The engine will not start");
    engineStatus.addRelation(CBRModel.HAS_INSTANCE, engineNotStarting);

    // Defines a "has engine status" relation to use in a car case.
    RelationType hasEngineStatus = new RelationType(km,
        km.getRelationType(CBRModel.FINDING_RELATION),
        "has engine status", "the engine status of something", 0.8,
        "engine status of", "the origin is the engine status of the target",
        0.5
        );


    // Defines the "battery status" attribute (another symbol type attribue).
    Entity batteryStatus = new Entity(km, "battery status",
                                      "the battery status of a car");
    batteryStatus.addRelation(CBRModel.SUBCLASS_OF, symbol);

    // Defines values of "battery status".
    Entity goodBT = new Entity(km, "OK", "The battery is OK");
    batteryStatus.addRelation(CBRModel.HAS_INSTANCE, goodBT);
    Entity emptyBT = new Entity(km, "empty", "The battery is empty");
    batteryStatus.addRelation(CBRModel.HAS_INSTANCE, emptyBT);
    Entity lowBT = new Entity(km, "low", "The battery is low");
    batteryStatus.addRelation(CBRModel.HAS_INSTANCE, lowBT);

    // Defines a "has battery status" relation to use in a car case.
    RelationType hasBatteryStatus = new RelationType(km,
        km.getRelationType(CBRModel.FINDING_RELATION),
        "has battery status", "the battery status of something", 0.8,
        "battery status of", "the origin is the battery status of the target",
        0.8
        );


    // Defines the attribute type "age". This is a numeric attribute. We do not
    // need to define possible values for numeric attributes as they are created
    // on demand (as we will see later).
    Entity age = new Entity(km, "age", "the age of the car");
    age.addRelation(CBRModel.SUBCLASS_OF, km.getEntity(CBRModel.NUMBER));

    // Defines a new relation-type to use with the age attribute.
    RelationType hasAge = new RelationType(
      km, km.getRelationType(CBRModel.FINDING_RELATION),
      "has age", "the age of something", 0.5,
      "age of", "the origin is the age of the target", 0.4
    );


    // Models the solutions of a case. The solutions are modeled a special case
    // of an attribute.
    Entity solution = new Entity(km, "solution", "The solution of a case");
    solution.addRelation(CBRModel.SUBCLASS_OF, symbol);

    // There are two possible solutions - recharge battery and "no problem" ;).
    Entity charge = new Entity(km, "recharge battery",
                               "You should recharge the battery.");
    charge.addRelation(CBRModel.HAS_INSTANCE, solution);
    Entity noProblem = new Entity(km, "ignore it",
                                  "There doesn't seem to be a problem.");
    noProblem.addRelation(CBRModel.HAS_INSTANCE, solution);

    // There is no need to make a special relation type for the solution. There
    // already exists a generic "has solution" relation type in the CBRModel,
    // and we will use this when defining the cases.
  }


  /**
   * Generates a new KnowledgeModel for a simple car-start problem domain,
   * part 2 of 3.
   */
  public static void addCases(KnowledgeModel km) throws
      NameAlreadyExistException, NoSuchRelationTypeException {

    // First, we create an entity to represent the type of case that this model
    // contains.
    Entity car = new Entity(km,
                            "car case", // The (unique) name of this entity
                            "a car problem" // The description (optional - it may be null)
                            );

    // Then we add a relation to imply that this is a type of case. "instance of"
    // relation and the "Case" entity is defined by the CBRModel class and
    // automatically added to all new LocalKnowledgeModels. We could refer to
    // the "Case" entity and "instance of" relation type by name, but by using the
    // CBRModel constants, the code are resistant to eventual name changes
    // in the future.
    car.addRelation(CBRModel.INSTANCE_OF, km.getEntity(CBRModel.CASE));


    // Now we create the solved cases in the casebase

    // Create a new case (case1)
    Case case1 = new Case(km,
      "car case 1",
      // The (unique) name of this case. Can be just "Case#192"
      "Low battery, not starting.",
      // The description of the case (optional - can be null)
      car
      // The type of case this is. There is a constructor omitting this,
      // in which case it will be made an instance of the generic "Case"
      // made by the CBRModel.
    );

    // It is a red car with low battery power and the engine won't start.
    case1.addRelation("has colour", getEntity(km, "red"));
    case1.addRelation("has engine status", getEntity(km, "not starting"));
    case1.addRelation("has battery status", getEntity(km, "low"));

    // The age is defined using NumberEntity. This convinience class creates a
    // new entity which encapsulating the number 14 (or the new Integer(14)
    // object). It also automatically creates a "value of" relation from this
    // entity to the age entity, signifying that this entity is a value of age.
    Entity age = getEntity(km, "age");
    case1.addRelation("has age", new NumberEntity(km, new Integer(3), age));
    case1.addRelation(CBRModel.HAS_SOLUTION, getEntity(km, "recharge battery"));

    // Sets the status of this case solved.
    case1.setStatus(Case.SOLVEDCASE);


    // Creates another solved case.
    Case case2 = new Case(km, "car case 2", "Good battery, running.", car);

    case2.addRelation("has colour", getEntity(km, "blue"));
    case2.addRelation("has engine status", getEntity(km, "running"));
    case2.addRelation("has battery status", getEntity(km, "OK"));
    case2.addRelation("has age", new NumberEntity(km, new Integer(15), age));
    case2.addRelation("has solution", getEntity(km, "ignore it"));
    case2.setStatus(Case.SOLVEDCASE);
    /*
    try {
      Relation statusRel = case2.getRelation("has case status");
      statusRel.setValue(km.getEntity("solved case"));
    } catch(jcreek.representation.NoSuchRelationException e) { // getRelation
      System.out.println("KUNNE IKKE SETTE CASE-STATUSEN!");
    }
    */

    // Creates an unsolved case. The procedure is the same, except there is no
    // "has solution" relation, and the status is UNSOLVEDCASE.
    Case case3 = new Case(km, "new case", "Empty battery, not starting", car);
    case3.addRelation("has colour", getEntity(km, "blue"));
    case3.addRelation("has engine status", getEntity(km, "not starting"));
    case3.addRelation("has battery status", getEntity(km, "empty"));
    case3.addRelation("has age", new NumberEntity(km, new Integer(16), age));
    case3.setStatus(Case.UNSOLVEDCASE);
  }


  /**
   * Generates a new KnowledgeModel for a simple car-start problem domain,
   * part 3 of 3.
   */
  public static void addCausalModel(KnowledgeModel km) throws
      NameAlreadyExistException, NoSuchRelationTypeException {
    // In order for the reasoning mechanism in jCreek to match symbolic values
    // that are not the same (e.g. matching "low battery power" and "empty
    // battery" partially), there have to be a model explaining how they are
    // similar. This is done through (possibly simplified) causal models. A
    // very, very simple such model now follow.

    Entity starterStop = new Entity(km, "starter engine won't turn",
                                    "starter engine won't turn");
    Entity starterSlow = new Entity(km, "starter engine turns slowly",
                                    "starter engine turns slowly");

    // Both a low battery power and empty battery casues the "engine stopped"
    // state. This allows Creek to match lowBT and emptyBT partially.
    Entity emptyBT = getEntity(km, "empty");
    Entity lowBT = getEntity(km, "low");
    emptyBT.addRelation(CBRModel.CAUSES, starterStop);
    lowBT.addRelation(CBRModel.CAUSES, starterSlow);
    starterStop.addRelation(CBRModel.CAUSES, getEntity(km, "not starting"));
    starterSlow.addRelation(CBRModel.CAUSES, getEntity(km, "not starting"));
  }





  /**
   * Takes an already created model and displays the cases and entities of the
   * model in different ways.
   */
  public static void displayModel(KnowledgeModel km, boolean showGUI)
  {
    // The next step is to access the model and displaying the contents of it.
    // The easiest way to access an entity is by name, and the easiest way
    // of displaying it is by using the getFrameString() method. This will
    // display the name, description and all relations from the entity.

    try
    {
      Entity colour = getEntity(km, "colour");
      System.out.println("Displaying the frame string of 'colour':");
      System.out.println(colour.getFrameString());

      // We can also do this manually by first printing the name and description:
      System.out.println();
      System.out.println("Showing the 'colour' entity:");
      System.out.println("  Name: "+colour.getName());  // The name of the entity
      System.out.println("  Description: "+colour.getDescription());  // The description
      if(colour.hasEntityObject())                      // Checks if the colour entity encapsulates a Java Object.
        System.out.println("  Encapsulated object: "+colour.getEntityObject().toString()); // Displays this object, if so.

      // Then we can elect to access certain relations selectively. The first
      // method returns the "subclass of" relation from this entity. If it has no
      // such relation, a NoSuchRelationException is thrown. If it has multiple such
      // relations, the first one is chosen.
      System.out.println("  Subclass Of: "+colour.getRelation("subclass of").getValue());

      // This method of accessing relations return all the relations of the given type.
      Relation[] instances = colour.getRelations("has instance");
      System.out.println("  Instances: ");
      for(int i=0;i<instances.length;i++)
        System.out.println("        "+instances[i].getValue());

      // ...or we can just go through all relations and display them.
      System.out.println("  All relations: ");
      Relation[] relations = colour.getRelations();
      for(int i=0;i<relations.length;i++)
        System.out.println("        "+relations[i].toString());

// ***

      // There are also some nifty GUI elements in the jcreek.gui.representation
      // package that will display entities and cases.

      if(showGUI)
      {

        // The first component is the FramePane, a component that will display an
        // entity and the relations from it. It also has a possibility for editing
        // them.
        FramePane colourPane = new FramePane(colour, true);

        // These are usually subclasses of JComponent, so we have to make a frame to show them in.
        JFrame colourFrame = new JFrame("The 'colour' entity");
        colourFrame.setContentPane(colourPane);
        colourFrame.setSize(colourFrame.getPreferredSize());
        colourFrame.show();
      }

        // A similar component exists for cases. Note that the FramePane can also
        // be used to display a Case, as the Case is a subclass of Entity, but
        // the CasePane will make a more specialized GUI element.

        // Before we do this, we use the matchEntityType() method in KnowledgeModel
        // to return all entities of a certain type. In this instance, we use it to
        // get an array of all the entities that are also cases.
        Case[] cases = (Case[]) km.matchEntityType(new Case[0]);

      if(showGUI)
      {
        // Then we iterate through this array and use the CasePane to display the
        // cases.
        for(int i=0;i<cases.length;i++)
        {
          CasePane casePane = new CasePane(cases[i], true);
          JFrame caseFrame = new JFrame("The case '"+cases[i].getName()+"'");
          caseFrame.setContentPane(casePane);
          caseFrame.setSize(casePane.getPreferredSize());
          caseFrame.show();
        }
      }

      // Last, we will display the a part of the model as a semantic net by the use of
      // the SubmodelView component. This class is a convinience class for
      // easy access to the jcreek.gui.concepmap classes that allows for alot
      // more configurability of this view. If interested, check the javadoc
      // pages for this.

      // First we define a submodel of the whole model
      SubModel subModel = new SubModel();

      // Let's add the "Symbol" Entity. This is a special CBRModel entity
      // that is the superclass of all the types we made to describe cases.
      Entity entity = getEntity(km, "Symbol");
      subModel.add(entity);

      // As we have made all our entities (indirectly) subclasses or instances of
      // Entity already (with the exception of relation types), we can easily access
      // the entire tree by the getAllSubclasses() method (this will also follow
      // "instance of" relations).
      Entity[] allSubclasses = entity.getAllSubclasses();

      // We add these to the submodel as well.
      subModel.add(allSubclasses);

      // Now, we want to see the relations between these as well, not only the
      // entities themselves. This, we have to do explicitly. We can iterate
      // through the entities in the model and get the relations and add them to
      // the model.
      Entity[] entitiesInSubModel = subModel.getEntities();
      for(int i=0;i<entitiesInSubModel.length;i++)
      {
        // The getLocalRelations() will only return relations that are not inherited.
        Relation[] relationsFromEntity = entitiesInSubModel[i].getLocalRelations();
        subModel.add(relationsFromEntity);
      }

      if(showGUI)
      {
        // Now we are ready to display this submodel. The SubModelView will automatically
        // lay out the model using the jcreek.util.graph.layout.SpringEmbedderLayout
        // algorithm, but it is possible to manually change and set the positions of the
        // nodes. See the jcreek.gui.conceptmap package.
        SubModelView subModelView = new SubModelView(km, subModel, true, true);

        // This, too, is a JComponent, so we need a new frame for it.
        JFrame subModelFrame = new JFrame("The Symbol hierarchy");
        subModelFrame.setContentPane(subModelView);
        subModelFrame.setSize(800, 800);
        subModelFrame.show();

        // This concludes the displaying part, although the solveCBR method displays
        // some of the GUI elements for the reasoning engine (from jcreek.gui.reasoning).
      }
    }
    catch(CreekException e)
    {
      // The jcreek.util class CreekException is the superclass for all jCreek
      // exceptions. This allows for the lazy of us to catch these exception with
      // a single catch() when appropriate.
      e.printStackTrace();
    }
  }

  /**
   * Exampel of saving the given model to disk.
   */
  public static void saveModel(KnowledgeModel model, String filename)
  {
    // Saving a KnowledgeModel to disk is very easy. There are now methods for
    // saving and loading in the KnowledgeModel interface because some implementations
    // of it may for instance use an SQL relational database for storage, or
    // a client-server model where the model is stored remotely. Currently,
    // the only implementation is the LocalKnowledgeModel, and this implementation
    // offers both save and load methods. We can easily test if the model is of
    // type LocalKnowledgeModel, and if it isn't we can make a new LocalKnowledgeModel
    // and transfers the contents of the KnowledgeModel to this new model.

    LocalKnowledgeModel localModel = null;
    if(model instanceof LocalKnowledgeModel)
    {
      // It is already a LocalKnowledgeModel, so we can just cast it.
      // Normally, this will be the case, and it certainly is in this class
      // as the KnowledgeModel we create in generateModel() is of this type.
      localModel = (LocalKnowledgeModel) model;
    }
    else
    {
      // It is not a LocalKnowledgeModel, but we can transfer the contents to one.
      localModel = new LocalKnowledgeModel(model);
    }

    // Then we just save it, using the saveAs method. This ties a file name with
    // the model so subsequent calls to save() will use the same name.
    // The method have to be enclosed in a try-catch statement to catch IOExceptions,
    // like all Java file handling.
    try
    {
      localModel.saveAs(filename);
    }
    catch(IOException e)
    {
      e.printStackTrace();
    }
  }

  /**
   * Example of loading a KnowledgeModel from disk.
   */
  public static KnowledgeModel loadModel(String filename)
  {
    // Now, this is real simple. Just make a new LocalKnowledgeModel
    // with the filename as the sole argument in the constructor. This has to
    // be enclosed in a ClassNotFoundException and a IOException since it uses
    // object serialization.
    try
    {
      return new LocalKnowledgeModel(filename);
    }
    catch(IOException e)
    {
      e.printStackTrace();
    }
    catch(ClassNotFoundException e)
    {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * Example of the jCreek CBR mechanism (jcreek.reasoning) and some GUI
   * elements displaying the process.
   */
  public static void solveCBR(KnowledgeModel km, boolean showGUI)
  {
    // We assume the model created in generateModel() in this method. First
    // we start by getting the unsolved case we made.
    Entity unsolvedCase = getEntity(km, "new case");

    // Note that when we use the getEntity() method in the KnowledgeModel we are
    // not guaranteed to get an object that can be cast to the specific subclass
    // of Entity that we created it as. Entity and its subclasses are non-persistant
    // access classes only. This has changed from the previous version (0.95) of
    // jCreek. The effect of this is that we can NOT cast unsovledCase to Case. In
    // other words
    //   Case myCase = (Case) unsolvedCase
    // is likely to throw a ClassCastException. To turn an entity object into an
    // object of another EntityType (e.g. a case), you can use a special constructor
    // made for this:
    try
    {
      Case unsolved = new Case(unsolvedCase);

      // This will NOT create a new case in the model, but it will make a new
      // Case instance that will allow easy access to "Case"-type features of
      // this entity. This constructor may throw a EntityTypeNotMatched exception
      // if the entity you try to encapsulate in the new Case instance is not of
      // the form expected of this type of entity (in this example, an entity
      // matching the Case template is expected to be a direct or indirect instance
      // of the CBRModel.CASE entity).
      //
      // If this was hard to follow, just try to remember that when you want to
      // cast an Entity myEntity to the Entity subclass MyEntitySubclass, use the
      // MyEntitySubclass(myEntity) constructor. Exampels of Entity subclasses
      // are RelationType, NumberEntity and, of course, Case.

      // Back to the CBR reasoning engine. The first step of the CBR cycle is to
      // retrieve similar cases. This can be done very easily by making a new
      // instance of the RetrieveResult class:

      RetrieveResult retrieve = new RetrieveResult(unsolved);

      // This will automatically complete the Retrieve step of finding the solved
      // cases of the same type as the input case and ranking them accoring to
      // similarity. You can control this process more if you like by instead
      // using the new RetrieveResult(unsolved, false) constructor. This will
      // not automatically run the retrieve process, but allow you to call the
      // activate(), explain() and focus() substeps with different parameters.
      //
      // The similarity measure is defined by the various findings'
      // Comparator classes. The subclasses of Symbol use the SymbolComparator
      // class, the subclasses of Number use the NumberComparator class and so on.
      // To implement your own similaity measure, have a look at the NumberEntity
      // and NumberComparator classes - it is fairly easy to do.

      // The simplest way of displaying a RetrieveResult is to call the toString()
      // method in the instance and print it.
      System.out.println("Displaying our case retrival:");
      System.out.println(retrieve.toString());

      if(showGUI)
      {
        // There is a GUI element that will show up to the ten highest cases
        // matches for a RetrieveResult object. This is possibly the easiest way
        // to graphically display all the information in the RetriveResult object.
        JFrame retrieveFrame = new JFrame("Retrieve for '"+unsolved.getName()+"'");
        RetrieveResultPane retrievePane = new RetrieveResultPane(retrieve);
        retrieveFrame.setContentPane(retrievePane);
        retrieveFrame.setSize(800,600);
        retrieveFrame.show();
      }

      // There is quite alot of detail available through the RetrieveResult object,
      // as apparent from the RetrieveResultPane GUI object. This can obviously
      // be accessed through code as well. The most obvious one is that each
      // pair of cases compared is represented by a CaseComparison object. We
      // can for instance get the CaseComparison that represents the comparison
      // of the input case with the best match case, and print that to standard
      // out:
      CaseComparison bestMatch = retrieve.getBestComparison();

      // You can access all of the CaseComparison object (one for each case the
      // input case is matched against) by using the getComparison(int) and
      // getComparisons() methods. These return CaseComparisons sorted so that
      // the best comparisons appear first.
      //
      // There are a ton of methods in a CaseComparison to access the findings
      // that are completely matched, partially matched and unmatched, to
      // explain each of these and so on. Refer to the CaseComparison javadoc
      // for more detail.
      System.out.println("The best case match is: ");
      System.out.println(bestMatch.toString());

      // You can also just go for the short story, although calling this an
      // explanation is a stretch:
      System.out.println();
      System.out.println(bestMatch.shortExplanation());
      // F. ex.: "The car problem 3 is 46% similar to car problem 1"

      if(showGUI)
      {
        // There is a GUI element for displaying a single CaseComparison as well:

        CaseComparisonPane ccPane = new CaseComparisonPane(bestMatch);
        JFrame ccFrame = new JFrame("Case Comparison for "+bestMatch.getOrigin().getName()+" and "+bestMatch.getTarget());
        ccFrame.setContentPane(ccPane); //setContentPane(ccPane.getLargePane());
        ccFrame.setSize(500, 400);
        ccFrame.show();
      }

      // As you can see, the CaseComparisonPane is (naturally) what is used internally
      // in the RetrieveResultPane.

      // The next step in the CBR process is Reuse, which will attempt to go
      // from the set of matched cases to a proposed solution. Currently, the
      // ReuseResult class implements a simple weighted vote of the k top cases
      // to do this (similar to kNN). If not specified, a default value is used.

      ReuseResult reuse = new ReuseResult(retrieve);

      // This can similarily be displayed to the user with the toString() method:
      System.out.println();
      System.out.println(reuse.toString());

      // Unfortunantly there are no GUI elements for ReuseResult yet, but the
      // getSolution() and getSolutionStrength() methods can be used to access
      // the contents.

      System.out.println("Proposed solution is: "+reuse.getSolution()+" with normalized weight: "+reuse.getSolutionStrength(reuse.getSolution()));

      // The last step is Retain, but at the moment this step is not implemented. However, it is easy to
      // add the solution from reuse to the unsolved case normally and the the status
      // to solved:
      reuse.getInputCase().addRelation(CBRModel.HAS_SOLUTION, reuse.getSolution());
      reuse.getInputCase().setStatus(Case.SOLVEDCASE);

      // By saving the model now, the case is effectively "retained" in a basic way, although no
      // updating of feature weights etc. have happened.
    }
    catch(EntityTypeNotMatchedException e)
    {
      // This is thrown when we try to encapsulate an entity into a Entity
      // subclass that it does not match the template of.
      e.printStackTrace();
    }
    catch(NoSuchRelationTypeException e)
    {
      e.printStackTrace();
    }
  }
}

