package jcreek.examples;

/**
 * A simple Knowledge Model Browser.
 * It basically is a JFrame showing FramePanes in JInternalFrames. Readers of
 * "Programmers Guide  modelling and matching" are (there) encouraged to use
 * it to visually inspect changes they have made to a model.
 *
 * Version 1.1: Extended the class to become a stand-alone application and
 * added an entity list on the left side (similar to TrollCreek Knowledge
 * editor) (Frode Srmo)
 *
 * @author Stein Brandser
 * @version 1.1
 */

import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.gui.representation.*;
import jcreek.gui.conceptmap.EntityList;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import java.io.*;

public class FrameBrowser extends JFrame {
  // Ignore this code if you goal is to understand the representation, and
  // not the GUI (using the Swing framework combined with custom elements).
  // If you want to make something similar a good starting point is http://
  // java.sun.com/docs/books/tutorial/uiswing/components/internalframe.html

  private final JDesktopPane layers = new JDesktopPane(); // holds internal frames.
  private KnowledgeModel km;

  private FrameBrowser() {} // remove default constructor.

  /**
   * Creates a new FrameBrowser for the KnowledgeModel supplied.
   *
   * @param km The Knowledge Model to open a frame browser for.
   * @param conceptName The name of the inital frame to show (optional).
   */
  public FrameBrowser(KnowledgeModel km, String conceptName) {
    if (km != null) {
      this.km = km;
    }
    else {
      throw new IllegalArgumentException("km required");
    }
    setTitle("FrameBrowser (viewing " + km.getName() + ")");

    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(layers, BorderLayout.CENTER);

    final EntityList entityList = new EntityList(km.getEntities());
    entityList.setPreferredSize(new Dimension(250, entityList.getMaximumSize().height));
    getContentPane().add(new JScrollPane(entityList), BorderLayout.EAST);
    entityList.addMouseListener(new MouseAdapter()
    {
      public void mouseClicked(MouseEvent e)
      {
        if(e.getClickCount() > 1)
        {
          Object ob = entityList.getSelectedValue();
          if(ob instanceof Entity)
          {
            open(((Entity) ob).getName());
          }
        }
      }
    });

    if (conceptName != null) open(conceptName);
  }

  /**
   * Adds another FramePane inside the Browser (container).
   * @param concept String - name of concept to open
   */
  public void open(String concept) {
    // Make a FramePane (which is able to make new ones in internalFrames):
    final FramePane pane = new FramePane(km.getEntity(concept));
    pane.setEditable(false);
    pane.removeLinkListener(pane.getOpenNewWindowLinkListener());
    pane.addLinkListener(new LinkListener() {
      public void linkClicked(LinkEvent e) {
        Entity ent = (Entity) e.getConcept();
        JInternalFrame frm = new JInternalFrame(ent.getName(), true, true, false, true);
        FramePane cp = new FramePane(ent);
        cp.setEditable(pane.isEditable());
        //cp.attach();
        cp.addLinkListener(this);
        layers.add(frm);
        frm.getContentPane().add(cp);
        frm.pack();
        frm.show();
      }

      public void linkAction(LinkEvent e) {}
    });

    // Add the pane to the FrameBrowser contained in a first JInternalFrame.
    JInternalFrame f = new JInternalFrame(concept, true, true, false, true);
    layers.add(f);
    f.getContentPane().add(pane);
    f.pack();
    f.show();
//    pack();
  }

  public static void main(String[] args) {
    File kmFile = null;
    if(args.length > 0)
    {
      kmFile = new File(args[0]);
    }
    if(kmFile == null || !kmFile.exists())
    {
      JFileChooser chooser = new JFileChooser();
      FileFilter filter = new FileFilter()
      {
        public String getDescription()
        {
          return "TrollCreek Knowledge Model";
        }

        public boolean accept(File f)
        {
          return f.getName().endsWith(".km");
        }
      };
      chooser.setFileFilter(filter);
      int returnVal = chooser.showOpenDialog(null);
      if(returnVal == JFileChooser.APPROVE_OPTION)
      {
        kmFile = chooser.getSelectedFile();
      }
    }


    try
    {
      LocalKnowledgeModel km = new LocalKnowledgeModel(kmFile);

      Entity initialFrame = null;
      if (args.length >= 2) {
        initialFrame = km.getEntity(args[1]);
      }
      if (initialFrame == null) {
        initialFrame = km.getEntities()[0];
      }
      FrameBrowser browser = new FrameBrowser(km, initialFrame.getName());
      browser.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      browser.setLocation(20, 20);
      browser.setSize(800, 600);
      browser.show();

    }
    catch(IOException e)
    {
      e.printStackTrace();
      System.exit(1);
    }
    catch(ClassNotFoundException e)
    {
      e.printStackTrace();
      System.exit(1);
    }

  }

}






