package jcreek.examples;

import jcreek.representation.KnowledgeModel;
import jcreek.representation.LocalKnowledgeModel;
import jcreek.representation.Entity;
import jcreek.representation.cbr.Case;
import jcreek.representation.RelationType;
import jcreek.representation.NoSuchRelationTypeException;
import jcreek.representation.NameAlreadyExistException;
import jcreek.representation.EntityTypeNotMatchedException;
import jcreek.gui.representation.FramePane; // to have a look on the model
import jcreek.reasoning.ReuseResult;
import jcreek.reasoning.RetrieveResult;
import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import jcreek.examples.FrameBrowser;

/**
 * Source code for "Programmers Guide  modelling and matching".
 * <p>Title: ModelligTutoral</p>
 * <p>Description: Contains the code described in the written tutorial.
 * This file is unnaturaly fragmented by coding standards, but this breaks
 * the process down into stages which can be taken stepwise (as documented
 * in the tutorial), and therby experimented with. </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: NTNU</p>
 * @author Stein Brandser
 * @version 1.0
 */
public class ModellingTutorial {

  // The commands that the main method will except:
  private static final String newKM = "new";
  private static final String addDM = "car";
  private static final String match = "match";
  private static final String doAll = "experiment";
  private static final String edit  = "browse";

  // Default name (fallback) to use if the user don't name a .km-file.
  private static final String defaultKM = "tutorial.km";


  /**
   * Parse the command line for arguments, run the corresponding method.
   * @param args String[] command line
   */
  public static void main(String[] args) {
    if (args.length == 0) {
      System.out.println("Missing argument, see the tutorial for a summary.");
    } else if
        (args[0].equals(newKM)) newKM(args); else if
        (args[0].equals(addDM)) addAttributesAndCases(args); else if
        (args[0].equals(match)) doMatch(args); else if
        (args[0].equals(doAll)) doAll(args); else if
        (args[0].equals(edit)) launchBrowser(args); else {
          System.out.println("Try a argument in {"
                             + edit +  ", "
                             + newKM + ", "
                             + addDM + ", "
                             + match + ", "
                             + doAll + "}.");
    }

  }


  public static void launchBrowser(String[] args) {
    if (args.length > 3) { //                 args        0   1   2
      System.out.println("Usage: ModellingTutorial "+edit+" [km [entity]]");
      System.exit(1);
    } else {
      KnowledgeModel km = loadKmFromFile(getFileName(args), edit); // args[1]
      String conceptName = (args.length > 2) ? args[2] : "entity";
      FrameBrowser b = new FrameBrowser(km, conceptName);
      b.show(); // show it.
    }
  }


  /**
   * Extract and return first argument, or default, from a String array.
   * @param argv String[] to extract first argument from.
   * @return String
   */
  protected static String getFileName(String[] args) {
    String filename = (args.length > 1) ? args[1] : defaultKM;
    if (!filename.endsWith(".km")) filename = filename + ".km";
    return filename;
  }

  /**
   * Give recovery feedback to user, call it from caught FileNotFoundExceptions.
   *
   * @param e Exception
   * @param cmd String
   */
  protected static void showMissingFileHint(Exception e, String cmd) {
    System.out.print("Missing " + e.getMessage() +". ");
    System.out.print("Specify another file (or use `"+newKM+"' instead ");
    System.out.println("of `"+cmd+"' to autogenerate a fresh file).");
  }


  /**
   * Load knowledge model from file successfully or exit program.
   * @param filename String
   * @return KnowledgeModel
   */
  protected static KnowledgeModel loadKmFromFile(String filename, String cmd) {
    KnowledgeModel km = null;
    try {
      km = new LocalKnowledgeModel(filename); // load model
    }
    catch (java.io.FileNotFoundException e) {
      showMissingFileHint(e, cmd);
      System.exit(2);
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
    catch (ClassNotFoundException e) {
      e.printStackTrace();
    }
    return km;
  }

  protected static boolean mayOverwrite(String filename) throws IOException {
    BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
    System.out.print("The file " + filename + " already exists, overwrite? ");
    String s = in.readLine();
    if (!s.toLowerCase().startsWith("y")) {
      return false;
    }
    return true;
  }


  public static void newKM(String[] args) {
    if (args.length > 2) {
      System.out.println("Usage: ModellingTutorial " + newKM + " [km]");
      System.exit(1);
    } else {
      String filename = getFileName(args);
      KnowledgeModel km = new LocalKnowledgeModel();
      try {
        if (!(new File(filename).exists())) {
          System.out.println("Making a new knowledge model in file: "+filename);
        } else { // file already exists
          if (mayOverwrite(filename)) {
            System.out.println("Overwriting "+filename+" with a new KM.");
          } else {
            System.out.println(filename+" left untouched.");
            System.exit(3);
          }
        }
        km.setName("model in "+filename);
        ((LocalKnowledgeModel)km).saveAs(filename);
      } catch (IOException e) {
        e.printStackTrace();
      }
    }
  }




  /**
   * Simply a wrapper to CreekExample.addAttributes(km) and addCases(km).
   * @param args String[], args[1] contains the filename of the km to load.
   * @see CreekExample.addCarDomainToModel
   */
  public static void addAttributesAndCases(String[] args) {
    if (args.length > 2) {
      System.out.println("Usage: ModellingTutorial "+addDM+" [km]");
      System.exit(1);
    } else {
      String filename = getFileName(args); // extract args[1]
      KnowledgeModel km = loadKmFromFile(filename, addDM);
      try {
        jcreek.examples.CreekExample.addAttributes(km);
        jcreek.examples.CreekExample.addCases(km);
        ( (LocalKnowledgeModel) km).saveAs(filename); // save the changes!
        System.out.println
          ("Added car domain to the model in `" + filename +"'.");
      }
      catch (NameAlreadyExistException e) { // from addAttributes
        System.out.println
          ("The model in `"+filename+"' already contains `"+e.getName()+ "'.");
        // See the comments about this exception in examples.CreekExample!
      }
      catch (NoSuchRelationTypeException e) { // from addAttributes and Cases
        e.printStackTrace();
        // See the comments about this exception in examples.CreekExample!
      }
      catch (IOException e) { // from saveAs
        e.printStackTrace();
      }
    }
  }





  public static void doMatch(String[] args) {
    if (args.length > 2) { //                 args      0     1
      System.out.println("Usage: ModellingTutorial "+match+" [km]");
      System.exit(1);
    } else {
      KnowledgeModel km = loadKmFromFile(getFileName(args), match); // args[1]
      theActualMatching(km);
    }
  }

  protected static void theActualMatching(KnowledgeModel km) {
    try {
      Entity entity = km.getEntity("new case");
      if (entity != null) {
        Case case3 = new Case(entity);
        // CBR: retrieve and reuse
        ReuseResult reuse = new ReuseResult(new RetrieveResult(case3));
        System.out.println("\n" + reuse.toString()); // print out solution.
        // (det skrives ut litt sty fra EntityComparison.active ogs).
      } else {
        System.out.println("Could not find the entity `new case' in the km, no match done");
        System.out.println("Change `"+match+"' into `"+addDM+"' to add it.");
      }
    } catch(EntityTypeNotMatchedException e) {
      e.printStackTrace();
    }
  }


  /**
   * Run an experiment with relation strength and/or causal model.
   * The color attribute is made very important with -s.
   * A causal model is added to the knowledge base with -c.
   * @param args String[]
   */
  public static void doAll(String[] args) {
    boolean importantColours = false, causalModel = false, v = false;
    for (int i = 1; i < args.length; i++) {
      if (args[i].equals("-s")) {
        importantColours = true;
      } else if (args[i].equals("-c")) {
        causalModel = true;
      } else if (args[i].equals("-v")) {
        v = true;
      } else {
        System.out.println("Usage: ModellingTutorial "+doAll+" [-v] [-s] [-c]");
        System.out.println("v: verbose, s: strength mod, c: causal mod");
        System.exit(1);
      }
    }
    try
    {
      if (v) System.out.println("* -v tells you what is going on.");
      if (v) System.out.println("* starting with a fresh model.");
      KnowledgeModel km = new LocalKnowledgeModel();
      if (v) System.out.println("* adding attributes");
      jcreek.examples.CreekExample.addAttributes(km);

      if (importantColours) {
        try {
          if (v) System.out.print("* -s redefining color as a strong trait");
          RelationType hasColour = new RelationType(km.getEntity("has colour"));
          hasColour.setExplanationStrength(0.36); // pvirker nye instanser
          //RelationType colourOf = new RelationType(km.getEntity("colour of"));
          //colourOf.setExplanationStrength(0.1); // pvirker nye instanser
        }
        catch (EntityTypeNotMatchedException e) {
          System.out.print(" [FAILED]");
        }
        finally {
          System.out.println();
        }
      }


      if (causalModel) {
        if (v) System.out.println("* -c adding causal model");
        jcreek.examples.CreekExample.addCausalModel(km);
        // try {((LocalKnowledgeModel)km).saveAs("CM");} catch(IOException e){};
      } else {
        if (v) System.out.println("* NOT adding causal model");
      }
      if (v) System.out.println("* adding cases");
      jcreek.examples.CreekExample.addCases(km);
      if (v) System.out.println("* performing case match");
      theActualMatching(km);
      if (v) System.out.println("* the end");
    }
    // from addAttributes og addCases:
    catch (NameAlreadyExistException e) { e.printStackTrace(); }
    catch (NoSuchRelationTypeException e) { e.printStackTrace(); }
  }


}
