package jcreek.gui;

/**
 * Generic graphical user interface element that will create a visual
 * representation of a part of a KnowledgeModel (a SubModel) as a directed
 * graph of nodes and edges.
 *
 * This class is really a specialized use of the jcreek.gui.conceptmap package,
 * but is easier to use for those who do not need all the customization options
 * of that package.
 *
 * The graph is layed out using the spring-embedder algorithm implemented in
 * jcreek.util.graph.layout.SpringEmbedderLayout.
 *
 * @author Frode Srmo
 * @version 1.0
 */

import jcreek.gui.conceptmap.*;
import jcreek.gui.conceptmap.tool.DragTool;
import jcreek.util.graph.layout.SpringEmbedderLayout;
import jcreek.util.graph.Node;
import java.util.Iterator;
import java.util.Vector;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import java.awt.*;
import java.awt.event.MouseEvent;

public class SubModelView extends ConceptMapView
{
  protected SubModel subModel;
  private boolean layoutOnResize = false;
  private int lastWidth = -1;
  private int lastHeight = -1;
  private SpringEmbedderLayout layout;
  private Vector nodes;

  /**
   * Creates a new SubModelView from the supplied KnowledgeModel and SubModel
   * objects.
   *
   * @param km The KnowledgeModel the SubModel is a part of.
   * @param model The SubModel to display.
   * @param manipulatable If this is true, the use may move the entities around
   * after the components is layed out and displayed.
   * @param layoutOnResize If this is true, the layout is redone after a resize of the component
   * to fill the available space.
   */

  public SubModelView(KnowledgeModel km,
                      SubModel model, boolean manipulatable, boolean layoutOnResize)
  {

    super(new ConceptMapModel(km));
    this.subModel = model;
    this.layoutOnResize = layoutOnResize;

    Entity[] ents = model.getEntities();
    for(int i=0;i<ents.length;i++)
      getModel().addEntity(ents[i], new Point(0,0));

    Relation[] rels = model.getRelations();
    for(int i=0;i<rels.length;i++)
      getModel().addRelation(rels[i]);

    this.nodes = new Vector(java.util.Arrays.asList(getModel().getNodes()));

    if(manipulatable)
      new DragTool(this);
  }

  public SubModel getSubModel()
  {
    return subModel;
  }
/*
  public Dimension getMaximumSize()
  {
    if(layoutOnResize)
      return new Dimension(10000, 10000);
    return super.getMaximumSize();
  }

  public Dimension getPreferredSize()
  {
    if(layoutOnResize)
      return new Dimension(10000, 10000);
    return super.getPreferredSize();
  }
*/
  public static void main(String argv[])
  {
    LocalKnowledgeModel km = new LocalKnowledgeModel();
    SubModel test = new SubModel();
    test.add(km.getEntities());
    test.add(km.getRelations());
    javax.swing.JFrame frame = new javax.swing.JFrame("Test");
    frame.getContentPane().setLayout(new BorderLayout());
    frame.getContentPane().add(new SubModelView(km, test, true, true), BorderLayout.CENTER);
    frame.setSize(800, 800);
    frame.show();
  }

  public void setBounds(int x, int y, int a, int b)
  {
    super.setBounds(x, y, a, b);

      if (layout == null && getModel() != null &&
          getSize().width > getController().getViewInsets().width * 2 &&
          this.getSize().height > getController().getViewInsets().height * 2)
      {
        firstResize();
      }
      else if (layout != null &&
               ( (lastWidth != getSize().width) ||
                (lastHeight != getSize().height)))
      {
//      firstResize();
        subsequentResize();
      }
  }

  protected void firstResize()
  {
    Rectangle bounds = new Rectangle(0,0, (int) Math.round(getSize().width/getScale()-getController().getViewInsets().width*2), (int) Math.round(getSize().height/getScale()-getController().getViewInsets().height*2));
    layout = new SpringEmbedderLayout();
    layout.randomizePositions((Node[]) nodes.toArray(new Node[nodes.size()]), bounds);
    layout.layout(getModel(),(Node[]) nodes.toArray(new Node[nodes.size()]), bounds);
    lastWidth = getSize().width;
    lastHeight = getSize().height;
  }

  protected void subsequentResize()
  {
    Rectangle bounds = new Rectangle(0,0, (int) Math.round(getSize().width/getScale()-getController().getViewInsets().width*2), (int) Math.round(getSize().height/getScale()-getController().getViewInsets().height*2));
    int dx = Math.abs(lastWidth-getSize().width);
    int dy = Math.abs(lastHeight-getSize().height);
//      System.out.println("dx: "+dx+ " dy: "+dy);

    double xFact = ((double) getSize().width)/lastWidth;
//      System.out.println("xFact: "+xFact+" as width: "+getSize().width+" and lastWidth "+lastWidth);
    double yFact = ((double) getSize().height)/lastHeight;
//      System.out.println("yFact: "+yFact);

    for(Iterator it = getModel().getNodeIterator();it.hasNext();)
    {
      Node node = (Node) it.next();
      node.setPosition((int) Math.round(node.getPosition().x*xFact), (int) Math.round(node.getPosition().y*yFact));
    }

    double temp = (dx+dy)/100;
//      if(temp > 5.0)
      temp = 3.5;
    layout.setInitialTemperature(temp);
    layout.layout(getModel(), (Node[]) nodes.toArray(new Node[nodes.size()]), bounds, 50);
    lastWidth = getSize().width;
    lastHeight = getSize().height;
  }
  public void fixEntity(Entity ent, Point pos)
  {
    VisualEntity remove = null;
    for(Iterator it=nodes.iterator();it.hasNext() && remove == null;)
    {
      VisualEntity ve = (VisualEntity) it.next();
//      System.out.println("Testing "+ve+" against "+ent);
      if(ve.getEntity().equals(ent))
        remove = ve;
    }
    if(remove != null)
    {
      nodes.remove(remove);
      remove.setPosition(pos);
    }
//    else
//      System.err.println("Failed to fix "+ent);
  }
}
