package jcreek.gui.chart;

import javax.swing.JPanel;
import java.awt.*;
import java.awt.image.*;

/**
 * Title:        JavaCreek
 * Description:  Creates a chart (2D bar graph) of any data described by the
 *               Chartable interface.
 * Copyright:    Copyright (c) 2002
 * Company:      NTNU
 * @author       Frode Srmo
 * @version 1.0
 */

public class JChart extends JPanel
{
  ChartModel model;
  BufferedImage drawBuffer;
  Color labelColor = Color.black;
  static Font labelFont = new Font("Arial", Font.BOLD, 14);
  static Font valueFont = new Font("Arial", Font.PLAIN, 10);

  /**
   *  Creates a new chart from the data provided by the Chartable.
   */
  public JChart(ChartModel model)
  {
    this.model = model;
  }

  /**
   * Sets the model used for describing the data in this chart.
   */
  public void setDataModel(ChartModel model)
  {
    this.model = model;
  }

  /**
   * Sets the color of the labels.
   */
  public Color getLabelColor()
  {
    return labelColor;
  }

  public void setLabelColor(Color c)
  {
    labelColor = c;
  }

  /**
   * Returns the data model for this chart.
   */
  public ChartModel getDataModel()
  {
    return model;
  }

  /**
   * Paints the chart.
   */
  public void paint(Graphics graphic)
  {
    Graphics2D g;

    if(drawBuffer == null || drawBuffer.getWidth() < this.getWidth() || drawBuffer.getHeight() < this.getHeight())
      drawBuffer = (BufferedImage) createImage(this.getWidth(), this.getHeight());

    g = drawBuffer.createGraphics();
//    g.setColor(view.getLineColor());
    g.setColor(getBackground());
    g.fillRect(0,0,getWidth(),getHeight());
    g.drawLine(2,0,20,this.getHeight()-20);
    g.drawLine(2,this.getHeight()-20,this.getWidth(),this.getHeight()-20);

    double maxValue = model.getMaxValue();
    int maxHeight = this.getHeight()-21;
    double heightStep = maxHeight/maxValue;
    double widthStep = (this.getWidth()-2)/model.getRowSize();
    for(int i=model.getRowSize()-1;i>=0;i--)
    {
      int bottomY=this.getHeight()-21;
      int fromX = (int) Math.round(widthStep*i);
      int width = (int) Math.round(widthStep-2.0);
      int colSize = model.getColSize(i);

      g.setColor(model.getBackgroundColor(i));
      g.fillRect(2+fromX, 0, width+2, this.getHeight());

      for(int j=0;j<colSize;j++)
      {
        int topY = bottomY-(int) Math.round(heightStep*model.getValue(j, i));
        if(topY < 0)
          topY = 0;
        int boxHeight = bottomY-topY-1;
        g.setColor(model.getValueColor(j, i));
//        System.out.println("Drawing col "+j+" row "+i+" at "+fromX+","+topY+" size "+(int) Math.round(widthStep-2.0)+","+(bottomY-topY)+" value is "+model.getValue(j, i));
        g.fill3DRect(3+fromX, topY, width, boxHeight,true);
        if(boxHeight >= g.getFontMetrics(valueFont).getHeight()+2)
        {
          g.setFont(valueFont);
          g.setColor(Color.white);
          g.drawString(model.getValueLabel(j, i),
                       3+fromX+(width-g.getFontMetrics().stringWidth(model.getValueLabel(j,i)))/2,
                       topY+(boxHeight-g.getFontMetrics().getHeight())/2+g.getFontMetrics().getAscent());
        }
        bottomY=topY;
      }
      g.setColor(labelColor);
      String label = model.getRowLabel(i);
      g.setFont(labelFont);
      g.drawString(model.getRowLabel(i),3+fromX+(width-g.getFontMetrics(labelFont).stringWidth(label))/2, this.getHeight()-8);
    }

    graphic.drawImage(drawBuffer, 0, 0, this);
    g.dispose();
  }

  /**
   * Returns the the value (row in the data model) that correspons to the graphical
   * position given by x and y (e.g. as given by the mouse position).
   */
  public int getValueForPos(int x, int y)
  {
    double maxValue = model.getMaxValue();
//    int colSize = model.getColSize();
    int maxHeight = this.getHeight()-21;
    int maxWidth = this.getWidth();
    double heightStep = maxHeight/maxValue;
    double widthStep = (maxWidth-2)/model.getRowSize();

    if(x < 22 || x > maxWidth || y > maxHeight || y < 1)
    {
      return -1;
    }

    int value = (int) Math.floor((x-22)/widthStep);
    int mod = (int) Math.round((x-22)%widthStep);

    if(mod < Math.round(widthStep-2.0))
    {
      return value;
    }
    else
      return -1;
  }
}
