package jcreek.gui.conceptmap;

import java.util.Iterator;
import java.util.Hashtable;
import java.util.ArrayList;
import java.io.*;
import jcreek.util.graph.*;

import jcreek.representation.KnowledgeModel;
import jcreek.representation.Entity;
import jcreek.representation.ModelChangeListener;
import jcreek.representation.ModelChangeEvent;
import jcreek.representation.Relation;
import jcreek.representation.SubModel;
import jcreek.representation.Context;
import jcreek.util.selection.*;
import java.awt.Point;
import java.awt.Rectangle;

/**
 * This is and extention of the Graph class that serves as a model (in
 * the model-view-controller paradigm) for the ConceptMapView. This model
 * is made with the assumption that the nodes and edges are VisualEntities and
 * VisualRelations.
 *
 * In addition to this, the model has a selction set that keeps the currently
 * selected entities at all times. This selection set is not saved when the model
 * is serialized.
 */

public class ConceptMapModel extends Graph implements ModelChangeListener, Serializable, Context
{
  private transient SelectionSet selectionSet = null;

  private Hashtable entityToVisual = new Hashtable();
  private Hashtable relationToVisual = new Hashtable();
  private KnowledgeModel km = null;

  // The boundary box of the model, i.e. the minimal rectangle containing all the edges.
  private Rectangle bounds = new Rectangle(0,0,0,0);
  private Rectangle innerBounds = new Rectangle(0,0,0,0);

  /**
   * Creates a new, empty ConceptMapModel.
   */
  public ConceptMapModel(KnowledgeModel km)
  {
    this.km = km;
    km.addModelChangeListener(this);
  }

  /**
   * Removes the ConceptMapModel and unregisteres it with the KnowledgeModel.
   */
  public void remove()
  {
    km.removeModelChangeListener(this);
  }

  /**
   * Adds an Entity to the ConceptMapModel by creating a VisualEntity that
   * is a wrapper for the Entity, containg the position of it within the model.
   *
   * @param entity The jcreek.representation Entity added to this ConceptMapModel
   * @param pos The initial position of the new VisualEntity.
   * @return The VisualEntity representing this entity.
   */
  public VisualEntity addEntity(Entity entity, Point pos)
  {
    if(entity == null)
      return null;

    if(contains(entity))
      return getVisualEntity(entity);

    VisualEntity ve = new VisualEntity(entity, pos);
    addNode(ve);
    return ve;
  }

  /**
   * Adds a Relation to the ConceptMap by creating a VisualRealtion that is a
   * wrapper for the Relation. If 'from' and 'value' entities of the relation is
   * not already added to the ConceptMapModel, this method will fail and return
   * null.
   *
   * @param relation The jcreek.representation Relation to represent.
   * @return The relation
   */
  public VisualRelation addRelation(Relation rel)
  {
    if(rel == null)
      return null;

    if(contains(rel) || contains(rel.getInverse()))
      return getVisualRelation(rel);

    VisualEntity from = getVisualEntity(rel.getFrom());
    VisualEntity value = getVisualEntity(rel.getValue());

    if(from == null || value == null)
      return null;

    VisualRelation vr = new VisualRelation(from, value, rel);
    addEdge(vr);
    return vr;
  }

  /**
   * Overrides the addNode() method in Graph to ensure this is a VisualEntity
   * and that all VisualEntities in the graph encapsulates entities in the same
   * KnowledgeModel.
   */
  public void addNode(Node node)
  {
    if(node instanceof VisualEntity)
    {
      if(km == null)
        km = ((VisualEntity) node).getEntity().getKnowledgeModel();
      else if(km.equals(((VisualEntity) node).getEntity().getKnowledgeModel()))
      {
        if(!contains(node))
          entityToVisual.put(((VisualEntity) node).getEntity().getID(), node);
        super.addNode(node);
      }
    }
    else
      throw new java.lang.ClassCastException("A "+node.getClass()+" supplied - only VisualEntity supported.");
  }

  public void removeNode(Node node)
  {

    if(node != null && contains(node))
    {
      entityToVisual.remove(((VisualEntity) node).getEntity().getID());
      getSelectionSet().remove(node);
    }
    super.removeNode(node);


  }

  /**
   * Returns the VisualEntity that represents the given entity in this model.
   */
  public VisualEntity getVisualEntity(Entity entity)
  {
    return (VisualEntity) entityToVisual.get(entity.getID());
  }

  /**
   * Returns true if this model has a node representing the given entity.
   */
  public boolean contains(Entity entity)
  {
    return entityToVisual.containsKey(entity.getID());
  }

  /**
   * Overrides the addEdge() method in Graph to ensure this is a VisualRelation.
   */
  public void addEdge(Edge edge)
  {
    if(edge instanceof VisualRelation)
    {
      if(!contains(edge))
        relationToVisual.put(((VisualRelation) edge).getRelation().getID(), edge);
      super.addEdge(edge);
    }
    else
      throw new java.lang.ClassCastException("A "+edge.getClass()+" supplied - only VisualRelation supported.");
  }

  public void removeEdge(Edge edge)
  {
    if(edge != null && contains(edge))
    {
      relationToVisual.remove(((VisualRelation) edge).getRelation().getID());
      getSelectionSet().remove(edge);
    }
    super.removeEdge(edge);
  }

  /**
   * Returns the VisualRelation that represents the given relation in this model.
   */
  public VisualRelation getVisualRelation(Relation rel)
  {
    return (VisualRelation) relationToVisual.get(rel.getID());
  }

  /**
   * Returns true if this model has an edge representing the given relation.
   */
  public boolean contains(Relation rel)
  {
    if(rel == null)
      return false;
    return relationToVisual.containsKey(rel.getID());
  }

  /**
   * Returns the SelectionSet of the model - i.e. the VisualEntities that are
   * currently marked as being selected.
   */
  public SelectionSet getSelectionSet()
  {
    if(selectionSet == null)
      selectionSet = new SelectionSet();
    return selectionSet;
  }

  /**
   * Returns the KnowledgeModel of this model.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    return km;
  }

  public void modelChanged(ModelChangeEvent e)
  {
    if((e.getType() == e.ENTITY_NAME_CHANGED && contains((Entity) e.getSource())) ||
       (e.getType() == e.ENTITY_DESC_CHANGED && contains((Entity) e.getSource())) ||
       (e.getType() == e.ENTITY_DATA_OBJECT_CHANGED && contains((Entity) e.getSource()))  )
      this.triggerNodeChangeEvent(new GraphChangeEvent(getVisualEntity((Entity) e.getSource()), GraphChangeEvent.NODE_CHANGED));

    if(e.getType() == e.ENTITY_REMOVED && contains((Entity) e.getSource()))
    {
      removeNode(getVisualEntity((Entity) e.getSource()));
    }

    if((e.getType() == e.RELATION_DESC_CHANGED && contains((Relation) e.getSource())) ||
       (e.getType() == e.RELATION_EXPLANTION_STRENGTH_CHANGED && contains((Relation) e.getSource())) ||
       (e.getType() == e.RELATION_NAME_CHANGED && contains((Relation) e.getSource())) ||
       (e.getType() == e.RELATION_TYPE_CHANGED && contains((Relation) e.getSource())))
      this.triggerNodeChangeEvent(new GraphChangeEvent(getVisualRelation((Relation) e.getSource()), GraphChangeEvent.EDGE_CHANGED));

    if((e.getType() == e.RELATION_FROM_CHANGED && contains((Relation) e.getSource())) ||
       (e.getType() == e.RELATION_FROM_CHANGED && contains((Relation) e.getSource())))
    {
      Relation rel = (Relation) e.getSource();
      removeEdge(getVisualRelation(rel));
      addRelation(rel);
    }

    if(e.getType() == e.RELATION_REMOVED && contains((Relation) e.getSource()))
    {
      Relation rel = (Relation) e.getSource();
      removeEdge(getVisualRelation(rel));
    }
  }

  /**
   * Returns a Submodel that contains all the entities and relations currently
   * activated in this Concept Map.
   */
  public SubModel toSubModel()
  {
    SubModel model = new SubModel();
    for(Iterator nodes = getNodeIterator(); nodes.hasNext();)
    {
      VisualEntity ve = (VisualEntity) nodes.next();
      model.add(ve.getEntity());
    }

    for(Iterator edges = getEdgeIterator(); edges.hasNext();)
    {
      VisualRelation ve = (VisualRelation) edges.next();
      model.add(ve.getRelation());
    }


    return model;
  }

  private void writeObject(java.io.ObjectOutputStream out) throws IOException
  {
    out.writeObject(nodeList);
    out.writeObject(entityToVisual);
    out.writeObject(edgeList);
    out.writeObject(relationToVisual);
    out.writeObject(bounds);
    out.writeObject(innerBounds);
    out.writeObject(km);
  }

  private void readObject(ObjectInputStream in)
      throws IOException, ClassNotFoundException
  {
    nodeList = (ArrayList) in.readObject();
    entityToVisual = (Hashtable) in.readObject();
    edgeList = (ArrayList) in.readObject();
    relationToVisual = (Hashtable) in.readObject();
    bounds = (Rectangle) in.readObject();
    innerBounds = (Rectangle) in.readObject();
    km = (KnowledgeModel) in.readObject();

    for(Iterator nodes = getNodeIterator(); nodes.hasNext();)
    {
      VisualEntity ve = (VisualEntity) nodes.next();
      ve.connectAfterLoad(getKnowledgeModel());
    }

    for(Iterator edges = getEdgeIterator(); edges.hasNext();)
    {
      VisualRelation ve = (VisualRelation) edges.next();
      ve.connectAfterLoad(getKnowledgeModel());
    }
  }
}
