package jcreek.gui.conceptmap;

/**
 * <p>Title: ConceptMap</p>
 * <p>Description: Creates a "drawing area" for concept map like representations.
 *
 * This class serves as a view in the traditional model-view-controller design,
 * where the ConceptMapModel is the model and the ConceptMapController is the
 * controller.
 *
 * @author Victor Dias, Frode Srmo
 * @version 1.0
 */

import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.*;
import java.awt.geom.*;
import java.awt.datatransfer.*;
import java.awt.dnd.DropTarget;
import java.awt.dnd.*;
import java.awt.event.*;
import jcreek.util.graph.*;
import jcreek.util.*;
import jcreek.gui.conceptmap.tool.ConceptMapTool;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.util.selection.*;

public class ConceptMapView extends JComponent implements GraphChangeListener, SelectionChangeListener
{
  protected ConceptMapModel concepMapModel;
  protected ConceptMapController conceptMapController;

  protected double scale = 1.0;
  protected Rectangle bounds = new Rectangle(0,0,300,300);
  protected Point offset = new Point(0,0);
  protected BufferedImage backImage = null;
  private boolean remakeBackimage = true;
  private boolean suspendBoundBoxUpdate = false;

  private transient ConceptMapTool currentTool;

  /**
   * Creates a new ConceptMap using the supplied ConceptMapModel as the model of this
   * view.
   *
   * @param graph The ConceptMapModel to be displayed in this ConceptMap.
   * @param configuration The ConceptMapController as the configuration of the view.
   */
  public ConceptMapView(ConceptMapModel model, ConceptMapController configuration)
  {
    super();
    this.concepMapModel = model;
    this.conceptMapController = configuration;
    model.addGraphChangeListener(this);
    model.getSelectionSet().addSelectionChangeListener(this);
    calculateBounds();
  }

  /**
   * Creates a new ConceptMap using the supplied ConceptMapmodel as the model and
   * the default ConceptMapController as the configuration.
   *
   * @param graph The ConceptMapModel to be displayed in this ConceptMap.
   */
  public ConceptMapView(ConceptMapModel model)
  {
    this(model, new DefaultConceptMapController());
  }

  /**
   * Returns the model (ConceptMapModel object) this ConceptMap is a display of.
   *
   * @return concepMapModel The ConceptMapModel object.
   */
  public ConceptMapModel getModel()
  {
    return concepMapModel;
  }

  /**
   * Returns the controller (ConceptMapController object) for this ConceptMap.
   */
  public ConceptMapController getController()
  {
    return conceptMapController;
  }

  /**
   * Sets the controlle (ConceptMapController object) for this ConceptMap. This
   * object serves as a configuration-object for the view.
   *
   * @param controller The ConceptMapController object.
   */
  public void setController(ConceptMapController controller)
  {
    this.conceptMapController = controller;
  }

  /**
   * (Re)calculates the outer boundaries of the ConceptMap. This is done based
   * on the highest and lowest x and y positions on visible nodes in the graph.
   * In addition a constand is added in each direction. This method should
   * not normally be required to called from the outside of this class.
   */
  public void calculateBounds()
  {
    bounds = new Rectangle(0,0, 100, 100);
    for(Iterator it = concepMapModel.getNodeIterator();it.hasNext();)
    {
      Node node = (Node) it.next();
      checkBoundsOnNode(node);
    }
    bounds.width -= bounds.x;
    bounds.height -= bounds.y;
    setPreferredSize(new Dimension((int) Math.round((bounds.width+conceptMapController.getViewInsets().width*2)*scale),
                                   (int) Math.round((bounds.height+conceptMapController.getViewInsets().height*2)*scale)));
    setMaximumSize(getPreferredSize());
//    if(getParent() != null)
//      setBounds(getBounds().x, getBounds().y, getMaximumSize().width, getMaximumSize().height);
    synchronized(getClass())
    {
      remakeBackimage = true;
    }
    repaint();
  }

  /**
   * Suspends or restarts the recalculation of the bounding box when a node is added or
   * removed from the model.
   */
  public void suspendBoundingBoxUpdate(boolean susp)
  {
    if(!susp && suspendBoundBoxUpdate)
      calculateBounds();
    suspendBoundBoxUpdate = susp;
  }

  private void checkBoundsOnNode(Node node)
  {
    if(node.getPosition().x < bounds.x)
      bounds.x = (int) Math.floor(node.getPosition().x);
    if(node.getPosition().y < bounds.y)
      bounds.y = (int) Math.floor(node.getPosition().y);
    if(node.getPosition().x > bounds.width)
      bounds.width = (int) Math.floor(node.getPosition().x);
    if(node.getPosition().y > bounds.height)
      bounds.height = (int) Math.floor(node.getPosition().y);
  }

  /**
   * The bounds of the rectangle (using model coordinates) displayed in this
   * component.
   */
  private Rectangle getBoundRectangle()
  {
    return bounds;
  }

  private Point getOffset()
  {
    return new Point(getVisibleRect().x, getVisibleRect().y);
  }

  /**
   * Returns the rectangle in model space coordinates that represents what
   * this view displays of the model.
   */
  public Rectangle getModelVisibleRect()
  {
    return new Rectangle(viewToModelTransform(new Point(0,0)), new Dimension((int) Math.round(getVisibleRect().width*scale),(int) Math.round(getVisibleRect().height*scale)));
  }

  /**
   * Returns the bounding box, in model space coordinates of all the nodes.
   */
  public Rectangle getBoundingBox()
  {
    return bounds;
  }

  /**
   * Change the center position of the ConceptMap view to match this coordinate
   * in model space.
   *
   * @param p A point in the model space.
   */
  public void centerOn(Point2D p)
  {
    int centerX = (int) Math.round(p.getX()*scale);
    int centerY = (int) Math.round(p.getY()*scale);

    Rectangle newViewport = new Rectangle(-getVisibleRect().width/2, centerY-getVisibleRect().height/2,
        centerX+getVisibleRect().width/2, centerY+getVisibleRect().height/2);
    scrollRectToVisible(newViewport);
    repaint();
  }


  /**
   * Centers the view on a node in the graph.
   */
  public void centerOn(Node node)
  {
    centerOn(new Point2D.Double(node.getPosition().x, node.getPosition().y));
  }

  /**
   * Returns the point located closest to the center of this view, in model
   * space coordinates.
   */

  public Point getCenterPoint()
  {
    return viewToModelTransform(new Point(getVisibleRect().width/2+getVisibleRect().x, getVisibleRect().height/2+getVisibleRect().y));
  }

  /**
   * Transforms a point in model space to a point in view space.
   *
   * @param modelSpacePoint A point in the model space.
   * @return The corresponding point in the view space.
   */
  public Point modelToViewTransform(Point modelSpacePoint)
  {
    return new Point((int) Math.round((modelSpacePoint.x-bounds.x+conceptMapController.getViewInsets().width)*scale),
    (int) Math.round((modelSpacePoint.y-bounds.y+conceptMapController.getViewInsets().height)*scale));
//    return new Point((int) Math.round((modelSpacePoint.x-getOffset().x-bounds.x)*scale),
//    (int) Math.round((modelSpacePoint.y-getOffset().y-bounds.y)*scale));
  }

/*  public Point modelToViewTransform(Point p)
  {
    return modelToViewTransform(new Point((int) Math.round(p.x), (int) Math.round(p.y)));
  } */

  /**
   * Transforms a point in view space to a point in model space.
   *
   * @param viewSpacePoint A point in the view space.
   * @return The corresponding point in the model space.
   */
  public Point viewToModelTransform(Point viewSpacePoint)
  {
    return new Point((int) Math.round(viewSpacePoint.x/scale+bounds.x-conceptMapController.getViewInsets().width),
    (int) Math.round(viewSpacePoint.y/scale+bounds.y-conceptMapController.getViewInsets().height));
//    return new Point((int) Math.round((viewSpacePoint.x+getOffset().x+bounds.y)/scale),
//    (int) Math.round((viewSpacePoint.y+getOffset().y+bounds.y)/scale));
  }

  /**
   * Changes the scale of the view. The scale is given as a double larger
   * than 0.0.
   *
   * @param newScale The new scale to set this view to.
   */
  public void setScale(double newScale)
  {
    Point center = getCenterPoint();

    setPreferredSize(new Dimension((int) Math.round((bounds.width+conceptMapController.getViewInsets().width*2)*newScale),
                                   (int) Math.round((bounds.height+conceptMapController.getViewInsets().height*2)*newScale)));
    setMaximumSize(getPreferredSize());
    this.scale = newScale;

//    entityFont = new Font("Arial",Font.PLAIN,(int) Math.round(14*scale));
//    relationFont = new Font("Arial",Font.PLAIN,(int) Math.round(12*scale));

    centerOn(center);
    backImage = null;
    revalidate();
  }

  /**
   * Returns the current scale as a double larger than 0.0.
   *
   * @return The current scale.
   */
  public double getScale()
  {
    return scale;
  }

  /**
   * Overrides the standard setSize() to maintain the center point when the size
   * is changed.
   */
  public void setBounds(int x, int y, int a, int b)
  {
    Point p = getCenterPoint();
    super.setBounds(x,y,a,b);
//    if(p != null)
//      centerOn(p);
    backImage = null;
    revalidate();
  }

  public VisualEntity findVisualEntity(Point viewPoint)
  {
     VisualEntity node = null;

     Point modelPoint = viewToModelTransform(viewPoint);
//     System.out.println("After transform: "+modelPoint.x+" "+modelPoint.y);

     VisualEntity tmpnode;
     Point pos;
     Dimension bbox;


     double closest = (conceptMapController.getEntitySize().width + conceptMapController.getEntitySize().height);
     if(concepMapModel != null)
     {
       for(Iterator it = concepMapModel.getNodeIterator(); it.hasNext();)
       {
         tmpnode = (VisualEntity) it.next();
         if(tmpnode != null)
         {
           pos = tmpnode.getPosition();
//           System.out.print("  -> "+pos.x+" "+pos.y+"   ("+tmpnode.getEntity()+")");

//           bbox = tmpnode.getBoundingBox();

           pos.x = Math.abs(pos.x - modelPoint.x);
           pos.y = Math.abs(pos.y - modelPoint.y);

//           System.out.println("  Distance: - "+pos.x+" "+pos.y);

//           if(pos.x < bbox.width / 2.0 + 1 && pos.y < bbox.height / 2.0 + 1)
//           {
             if(pos.x + pos.y < closest)  // rough estimate of closeness
             {
               closest = pos.x + pos.y;
               node = tmpnode;
             }
//           }
         }
       }
     }
     return node;
  }//findVisualEntity

  public VisualRelation findVisualRelation(Point p)
  {
    double closest = -1;
    VisualRelation current = null;
    Point pos = viewToModelTransform(p);

    int extra = 20;

    for(Iterator it = concepMapModel.getEdgeIterator(); it.hasNext();)
    {
      VisualRelation ve = (VisualRelation) it.next();

//      System.out.println("Testing for "+ve+"   - from: "+ve.getFrom().getPosition()+"  to: "+ve.getTo().getPosition());
      if(((pos.x < ve.getFrom().getPosition().x+extra && pos.x > ve.getTo().getPosition().x-extra) ||
          (pos.x > ve.getFrom().getPosition().x+extra && pos.x < ve.getTo().getPosition().x-extra)) &&
         ((pos.y < ve.getFrom().getPosition().y+extra && pos.y > ve.getTo().getPosition().y-extra) ||
          (pos.y > ve.getFrom().getPosition().y+extra && pos.y < ve.getTo().getPosition().y-extra)))
      {
        double distance = relationDistance(ve, pos);
//        System.out.println("Distance to "+ve+" is "+distance);
        if(closest == -1 || distance < closest)
        {
          current = ve;
          closest = distance;
        }
      }
    }
    if(closest < 20)
      return current;
    return null;
  }

  private double relationDistance(VisualRelation rel, Point pos)
  {
    double d = (1.0d*(rel.getTo().getPosition().y-rel.getFrom().getPosition().y))/(rel.getTo().getPosition().x-rel.getFrom().getPosition().x);
    return Math.abs(d*pos.x-pos.y-(d*rel.getFrom().getPosition().x-rel.getFrom().getPosition().y));
  }

  /**
   * Sets the current tool of this view. This is usually done automatically
   * by the tool's constructor.
   *
   * @param tool The new current tool
   */
  public void setTool(ConceptMapTool tool)
  {
    this.currentTool = tool;
  }

  /**
   * Returns the current tools used on this view.
   */
  public ConceptMapTool getCurrentTool()
  {
    return currentTool;
  }

  /**
   * (Re)paints this component.
   */
  public void paintComponent(Graphics g)
  {
//    System.out.println("Bounding box: "+getBounds());
//    System.out.println("Visible model space: "+getViewport());

    super.paintComponent(g);
    synchronized(getClass())
    {
      if(backImage == null || remakeBackimage)
      {
        int width = getPreferredSize().width;
        if(width < getSize().width)
          width = getSize().width;
        int height = getPreferredSize().height;
        if(height < getSize().height)
          height = getSize().height;
        backImage = (BufferedImage) createImage(width, height);
        remakeBackimage = false;
      }
    }

    Rectangle clipArea = getVisibleRect();

    Graphics2D graphics = (Graphics2D) backImage.createGraphics();
    graphics.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                               RenderingHints.VALUE_ANTIALIAS_ON );
    graphics.setColor(Color.white);
    graphics.setPaintMode();
    graphics.fillRect(clipArea.x, clipArea.y, clipArea.width, clipArea.height);


    graphics.setColor(Color.black);

    Rectangle virtualClipArea = new Rectangle((int) Math.round(clipArea.x*scale), (int) Math.round(clipArea.y*scale),
        (int) Math.round(clipArea.width*scale), (int) Math.round(clipArea.height*scale));
    drawObjects(graphics, clipArea);

    if(currentTool != null)
        currentTool.paintTool(graphics);

//    g.setPaintMode();
    g.drawImage(backImage, 0, 0, this);


//    firePaintAfter( (Graphics2D)g );
    graphics.dispose();
  }

  public void drawObjects(Graphics2D graphics, Rectangle clipArea)
  {
      VisualEntity v;
      VisualRelation vr;

      // Then draw all of the vertices.
      if ( getModel() != null )
      {

        for(Iterator it = getModel().getNodeIterator(); it.hasNext(); )
        {
          Object node = it.next();
          if(node instanceof VisualEntity)
          {
            v = (VisualEntity) node;
//            if(v.getDrawBounds(graphics).intersects(clipArea)) ;
              drawNode(graphics, v);
          }
        }

/*        if(temporaryEntities != null)
        {
          for(Iterator it=temporaryEntities.iterator();it.hasNext();)
          {
            drawNode(graphics, (VisualEntity) it.next());
          }
        } */

        // Then draw all of the relations.
        for(Iterator it = getModel().getEdgeIterator(); it.hasNext(); )
        {
          Object rel = it.next();
          if(rel instanceof VisualRelation)
          {
            vr = (VisualRelation) rel;
            if((vr.getFrom().getPosition().x <= clipArea.x+clipArea.width &&
               vr.getFrom().getPosition().x >= clipArea.x &&
               vr.getFrom().getPosition().y <= clipArea.y+clipArea.height &&
               vr.getFrom().getPosition().y >= clipArea.y) ||
               (vr.getTo().getPosition().x <= clipArea.x+clipArea.width &&
               vr.getTo().getPosition().x >= clipArea.x &&
               vr.getTo().getPosition().y <= clipArea.y+clipArea.height &&
               vr.getTo().getPosition().y >= clipArea.y)) ;
              drawEdge(graphics, vr, 0);
          }
        }
    }
  }

  public void drawNode(Graphics2D graphics, VisualEntity entity)
  {
     Color defColor = graphics.getColor();

     double w = conceptMapController.getEntitySize().width* scale;
     double h = conceptMapController.getEntitySize().height * scale;

     if(w < 1)
        w = 1;
     if(h < 1)
        h = 1;

     Point modelPoint = new Point(entity.getPosition().x, entity.getPosition().y);

     Point viewPoint = modelToViewTransform(new Point((int) entity.getPosition().x, (int) entity.getPosition().y));
//     System.out.println("Painting "+entity.getEntity()+" at "+viewPoint.x+" "+viewPoint.y);

     int x = viewPoint.x;
     int y = viewPoint.y;

     if(entity.isSelected())
       graphics.setColor(conceptMapController.getEntitySelectedColor(entity.getEntity()));
     else
       graphics.setColor(conceptMapController.getEntityColor(entity.getEntity()));

     graphics.fillOval((int)(x - w / 2), (int)(y - h / 2),
                             (int)w, (int)h);

     //draw label
     if(conceptMapController.showEntityNames(scale))
     {
       graphics.setFont(conceptMapController.getEntityFont(entity.getEntity(), scale));
       graphics.setColor(conceptMapController.getEntityFontColor(entity.getEntity()));
       double widest = 0.0;
       double lx, ly;
       FontMetrics fm = graphics.getFontMetrics();
       widest = fm.stringWidth(entity.getEntity().getName());
       lx = x - widest / 2.0;
       ly = y + h / 2.0 + fm.getAscent() + 1.0;
       graphics.drawString(entity.getEntity().getName(), (int)lx, (int)ly);
     }

     //draw selection
     if(!entity.isSelected())
        return;

     graphics.setColor(conceptMapController.getEntitySelectedColor(entity.getEntity()));

    // Draw selection components.
     /*
     Point tr = new Point((int) Math.round(x + w / 2 + 1),
                            (int) Math.round(y - h / 2 - 1));
     Point bl = new Point((int) Math.round(x - w / 2.0 - 1),
                            (int) Math.round(y + h / 2.0 + 1));

     graphics.drawRect(x - 4, y - 1, 8, 2);
     graphics.drawRect(x - 1, y - 4, 2, 8);
     graphics.setColor(Color.white);
     graphics.drawLine(x - 3, y, x + 3, y);
     graphics.drawLine(x, y - 3, x, y + 3);

    graphics.setColor(defColor);
     */
  }

  public void drawEdge(Graphics2D graphics, VisualRelation relation, int cnt)
  {
    Color defColor = graphics.getColor();



//    int cnt = 1; //from.getRelationCount(to,this);
//    QuadCurve2D.Double curve=new QuadCurve2D.Double();

    Point p1to, p2to;
    p1to = new Point(relation.getFrom().getPosition().x, relation.getFrom().getPosition().y);
    p2to = new Point(relation.getTo().getPosition().x, relation.getTo().getPosition().y);

    Point p1 = intersectWithLineTo(p1to, p2to);

    Point p2 = intersectWithLineTo(p2to, p1to);


    double dX = p1.x - p2.x;
    double dY = p1.y - p2.y;

    if(dX==0 && dY==0)
        return;//when the points are one and the same

    Point ctrl = modelToViewTransform(getCtrlPoint(p2,p1,1));

    Point viewPoint1 = modelToViewTransform(p1);
    Point viewPoint2 = modelToViewTransform(p2);

    if(getModel().getSelectionSet().contains(relation))
      graphics.setColor(conceptMapController.getRelationSelectedColor(relation.getRelation()));
    else
      graphics.setColor(conceptMapController.getRelationColor(relation.getRelation()));
    if(cnt <= 1)
    {
      graphics.drawLine(viewPoint1.x, viewPoint1.y, viewPoint2.x, viewPoint2.y);

      /*
      Point pfrom = new Point(p1to);
      if(transform != null)
        pfrom.transform(transform);
      if((int)pfrom.x == (int)p1.x && (int)pfrom.y == (int)p1.y)
         pfrom.x -= 10.0;
      */

      drawArrow(graphics, viewPoint2, viewPoint1);
    }
/*    else
    {
      curve.setCurve((int) Math.round(p2.x),(int) Math.round(p2.y),(int) Math.round(ctrl.x),(int) Math.round(ctrl.y),(int) Math.round(p1.x),(int) Math.round(p1.y));
      graphics.draw(curve);
    } */
// Draw Label
    if(conceptMapController.showRelationNames(scale))
    {
      double degree;
        if(dX==0 && dY>0)
        degree= Math.PI/2;
      else if(dX==0 && dY<0)
        degree=Math.PI*3/2;
      else
        degree=Math.atan(dY/dX);

      String text = relation.getRelation().getName();
      AffineTransform rat  =  AffineTransform.getRotateInstance(
          degree);
      Font iFont = graphics.getFont();
      FontMetrics fm = graphics.getFontMetrics();
      int ascent = fm.getAscent();

      Point textPoint = new Point();

      double length = Math.sqrt(dX*dX+dY*dY);
      double center = (length-fm.stringWidth(text))/(2*length);

      if(p2.x < p1.x)
      {
        textPoint.x = (int) Math.round(((double) (p1.x-p2.x))*center+p2.x);
        textPoint.y = (int) Math.round(((double) (p1.y-p2.y))*center+p2.y);
      }
      else
      {
        textPoint.x = (int) Math.round(((double) (p2.x-p1.x))*center+p1.x);
        textPoint.y = (int) Math.round(((double) (p2.y-p1.y))*center+p1.y);
      }

      textPoint = modelToViewTransform(textPoint);

      graphics.setColor(conceptMapController.getRelationFontColor(relation.getRelation()));
      graphics.setFont(conceptMapController.getRelationFont(relation.getRelation(), scale).deriveFont(rat));
      graphics.drawString(text, (int)textPoint.x,(int) textPoint.y);  //draw text
      graphics.setFont(iFont);
      graphics.setColor(defColor);
    }


  }//draw



  private Point getCtrlPoint(Point fromPos,Point toPos,int cnt)
  {
    Point ctrl=new Point();//control point to draw the curve

    ctrl.translate((int) Math.round(150*Math.cos(Math.toRadians(45+2*cnt))+fromPos.x),
                     (int) Math.round(150*Math.sin(Math.toDegrees(45+2*cnt ))+fromPos.y));

    if(fromPos.x > toPos.x)
    {
          ctrl.translate((int) Math.round(150*Math.cos(Math.toRadians(45+2*cnt))+toPos.x),
                         (int) Math.round(150*Math.sin(Math.toDegrees(45+2*cnt ))+toPos.y));//when rotated it must becomes exactly how it was on the opp quadrant

    }

    return ctrl;
  }

  public Point intersectWithLineTo(Point to, Point from)
  {
    int y_ = from.y;
    int x_ = from.x;

    int dy = to.y - y_;
    int dx = to.x - x_;

    // If point is inside the node, return the centerpoint.
    if(dx * dx / (conceptMapController.getEntitySize().width * conceptMapController.getEntitySize().width) + dy * dy / (conceptMapController.getEntitySize().height * conceptMapController.getEntitySize().height)
       <= .25)
      return new Point(x_, y_);

    if(dx == 0)
    {
      if(dy > 0)
        return new Point(x_, y_ + conceptMapController.getEntitySize().height / 2);
      else
        return new Point(x_, y_ - conceptMapController.getEntitySize().height / 2);
    }
    double slope = (1.0+dy) / dx;
    double x, y;

//          double tempx = (width/2)*Math.cos(Math.atan(slope));

    x = 1.0 * conceptMapController.getEntitySize().width * conceptMapController.getEntitySize().height /
        (2.0 * Math.sqrt(conceptMapController.getEntitySize().height * conceptMapController.getEntitySize().height
                                        + slope * slope *
                                        conceptMapController.getEntitySize().width * conceptMapController.getEntitySize().width));
    if(to.x < x_)
      x = -x;
    y = slope * x;

    return new Point((int) Math.round(x_ + x), (int) Math.round(y_ + y));
  }


  private void drawArrow(Graphics2D graphics, Point p1, Point p2)
  {
    double dx = p1.x - p2.x;
    double dy = p1.y - p2.y;

    double length = Math.sqrt(dx * dx + dy * dy);

    Point p3 = new Point(p2.x, p2.y);

    double arrow_size = 6;

    // Move 5 pixels back.
    p3.x += (arrow_size*scale) / length * dx;
    p3.y += (arrow_size*scale) / length * dy;

    Point p4 = new Point(p3.x, p3.y);

    // Out 4 pixels.
    p4.x += (arrow_size*scale) * .7 / length * dy;
    p4.y -= (arrow_size*scale) * .7 / length * dx;

    p3.x -= (arrow_size*scale) * .7 / length * dy;
    p3.y += (arrow_size*scale) * .7 / length * dx;

    graphics.drawLine((int)p2.x, (int)p2.y, (int)p4.x, (int)p4.y);
    graphics.drawLine((int)p2.x, (int)p2.y, (int)p3.x, (int)p3.y);
  }//drawArrow

/*  public void repaintMe()
  {
    Component c = this;
//    while(c != null)
//    {
//      System.out.println(c);
//      c = c.getParent();
//    }

/*    if(getParent() != null && getParent().getParent() != null && getParent().getParent() instanceof JScrollPane)
    {
      System.out.println("Repaint scrollpane");
      getParent().getParent().validate();
      getParent().getParent().repaint();
    }
    else
      repaint();
  }
*/
  public void nodeMoved(GraphChangeEvent e)
  {
    if(!suspendBoundBoxUpdate)
      calculateBounds();
    repaint();
  }
  public void nodeChanged(GraphChangeEvent e)
  {
    repaint();
  }
  public void nodeAdded(GraphChangeEvent e)
  {
    if(!suspendBoundBoxUpdate)
      calculateBounds();
    repaint();
  }
  public void nodeRemoved(GraphChangeEvent e)
  {
    if(!suspendBoundBoxUpdate)
      calculateBounds();
    repaint();
  }
  public void edgeAdded(GraphChangeEvent e)
  {
    repaint();
  }
  public void edgeRemoved(GraphChangeEvent e)
  {
    repaint();
  }
  public void edgeChanged(GraphChangeEvent e)
  {
    repaint();
  }
  public void selectionClear(SelectionChangeEvent e)
  {
    repaint();
  }
  public void selectionRemove(SelectionChangeEvent e)
  {
    repaint();
  }
  public void selectionAdd(SelectionChangeEvent e)
  {
    repaint();
  }

  public void setDropEnable(boolean b)
  {
    try
    {
      setDropTarget(new DropTarget());
      this.getDropTarget().addDropTargetListener(new ConceptMapViewDropHandler());
    }
    catch(TooManyListenersException e)
    {
      e.printStackTrace();
    }
  }

  class ConceptMapViewDropHandler implements DropTargetListener
  {
    VisualEntity ent = null;
    private boolean canImport = false;

    ConceptMapViewDropHandler()
    {
    }

    private boolean actionSupported(int action)
    {
      return true;
    }

    public void dragEnter(DropTargetDragEvent e)
    {
      DataFlavor[] flavors = e.getCurrentDataFlavors();

      if(Arrays.asList(flavors).contains(EntityTransferable.getEntityDataFlavor()))
        canImport = true;

      int dropAction = e.getDropAction();

      if (canImport && actionSupported(dropAction))
      {
        e.acceptDrag(dropAction);
      }
      else
      {
        e.rejectDrag();
      }
    }

    public void dragOver(DropTargetDragEvent e)
    {
      int dropAction = e.getDropAction();

      if (canImport && actionSupported(dropAction)) {
        e.acceptDrag(dropAction);
      } else {
        e.rejectDrag();
      }
    }

    public void dragExit(DropTargetEvent e)
    {
    }

    public void drop(DropTargetDropEvent e)
    {
      int dropAction = e.getDropAction();

      if(canImport && actionSupported(dropAction))
      {
        Transferable transfer = e.getTransferable();
        try
        {
          Object entityID = (Object) transfer.getTransferData(EntityTransferable.getEntityDataFlavor());
          Entity toAdd = getModel().getKnowledgeModel().getEntity(entityID);
          Point pos = viewToModelTransform(e.getLocation());
          if(getModel().contains(toAdd))
          {
            getModel().getVisualEntity(toAdd).setPosition(pos);
          }
          else
            getModel().addEntity(toAdd, pos);
          e.dropComplete(true);
          repaint();
        }
        catch(Exception ex)
        {
          e.dropComplete(false);
        }


        JComponent c = (JComponent)e.getDropTargetContext().getComponent();
        TransferHandler importer = c.getTransferHandler();
      }
      else
        e.rejectDrop();
    }

    public void dropActionChanged(DropTargetDragEvent e)
    {
      int dropAction = e.getDropAction();

      if (canImport && actionSupported(dropAction))
      {
        e.acceptDrag(dropAction);
      }
      else
      {
        e.rejectDrag();
      }
    }
  }
}
