package jcreek.gui.conceptmap;

/**
 * A specialized JList subclass made for displaying entities. It contains a
 * special ListCellRenderer that will use information from a ConceptMapController
 * to display the entities using similar graphics as a ConceptMapView. It is also
 * able to use clipboard (Java datatransfer) and Drag-and-Drop from ConceptMapViews
 * and other EntityLists
 *
 * @author Frode Srmo
 * @version 1.0
 */

import javax.swing.*;
import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.util.graph.*;
import java.util.*;

public class EntityList extends JList implements GraphChangeListener
{
  private ConceptMapModel model = null;
  private ConceptMapController controller;
  private Entity[] entities;

  /**
   * Creates a new EntityList of the supplied list of entities using the drawing
   * configurations of the supplied ConceptMapController.
   *
   * @param controller The ConceptMapController to use to draw the entities.
   * @param entities The entities to display.
   */
  public EntityList(ConceptMapController controller, Entity[] entities)
  {
    super(entities);
    this.controller = controller;
    this.entities = entities;
    init();
  }

  /**
   * Creates a new EntityList of the supplied entities using the default drawing
   * configuration from the DefaultConceptMapController.
   *
   * @param entities The entities to display.
   */
  public EntityList(Entity[] entities)
  {
    this(new DefaultConceptMapController(), entities);
  }

  /**
   * Creates a new EntityList using from the supplied ListModel using the
   * drawing configurations of the supplied ConceptMapController.
   *
   * @param controller The ConceptMapController to use to draw the entities.
   * @param model The ListModel to use.
   */
  public EntityList(ConceptMapController controller, ListModel model)
  {
    super(model);
    this.controller = controller;
    init();
  }

  /**
   * Creates a new EntityList using from the supplied ListModel using the default drawing
   * configuration from the DefaultConceptMapController.
   *
   * @param controller The ConceptMapController to use to draw the entities.
   */
  public EntityList(ListModel model)
  {
    this(new DefaultConceptMapController(), model);
  }

  private void init()
  {
    this.setCellRenderer(new ConceptMapEntityListRenderer());
    this.setTransferHandler(new EntityListTransferHandler());
//    this.setBackground(Color.darkGray);
  }

  public void setConceptMapModel(ConceptMapModel model)
  {
    if(model != null)
      model.addGraphChangeListener(this);
    else if(this.model != null)
      this.model.removeGraphChangeListener(this);
    this.model = model;
  }

  public void paintComponent(Graphics g)
  {
    Graphics2D g2 = (Graphics2D) g;
    g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                               RenderingHints.VALUE_ANTIALIAS_ON );
    super.paintComponent(g2);
  }

  public void nodeMoved(GraphChangeEvent e)
  {
  }
  public void nodeChanged(GraphChangeEvent e)
  {
    repaint();
  }
  public void nodeAdded(GraphChangeEvent e)
  {
    repaint();
  }
  public void nodeRemoved(GraphChangeEvent e)
  {
    repaint();
  }
  public void edgeAdded(GraphChangeEvent e)
  {
  }
  public void edgeRemoved(GraphChangeEvent e)
  {
  }
  public void edgeChanged(GraphChangeEvent e)
  {
  }


  class ConceptMapEntityListRenderer extends JComponent implements ListCellRenderer
  {
    private Entity myEntity;
    ConceptMapEntityListRenderer()
    {
      super();
      int height = controller.getEntitySize().height+10;
      int fontHeight = this.getFontMetrics(controller.getEntityFont(null, 1.0)).getHeight()+10;

      if(height < fontHeight)
        height = fontHeight;

      setPreferredSize(new Dimension(getPreferredSize().width, height));
      setMinimumSize(getPreferredSize());
//      this.setBorder(BorderFactory.createEtchedBorder());
    }

    public void setEntity(Entity myEntity)
    {
      this.myEntity = myEntity;
    }

    public void paintComponent(Graphics g)
    {
//      int height = controller.getEntitySize().height+10;
      int fontHeight = this.getFontMetrics(controller.getEntityFont(myEntity, 1.0)).getHeight()+10;

//      if(height < fontHeight)
//        height = fontHeight;

      if(model != null && model.contains(myEntity))
      {
        g.setColor(getBackground().darker());
        g.fillRect(0,0,getSize().width, getSize().height);
        g.setColor(controller.getEntityColor(myEntity).brighter());
        g.fillOval(10,(getSize().height-controller.getEntitySize().height)/2, controller.getEntitySize().width, controller.getEntitySize().height);
        g.setColor(controller.getEntityFontColor(myEntity).brighter());
      }
      else
      {
        g.setColor(getBackground());

        g.fillRect(0,0,getSize().width, getSize().height);
        g.setColor(controller.getEntityColor(myEntity));
        g.fillOval(10,(getSize().height-controller.getEntitySize().height)/2, controller.getEntitySize().width, controller.getEntitySize().height);
        g.setColor(controller.getEntityFontColor(myEntity));
      }

      g.setFont(controller.getEntityFont(myEntity, 1.0));
      int count = 0;
      while(count < 6 && this.getFontMetrics(getFont()).stringWidth(myEntity.getName())+controller.getEntitySize().width+40 > getSize().width)
      {
        g.setFont(g.getFont().deriveFont(getFont().getSize2D()-1.0f));
        count++;
      }
      g.drawString(myEntity.getName(), 20+controller.getEntitySize().width, getSize().height/2-this.getFontMetrics(g.getFont()).getHeight()/2+this.getFontMetrics(g.getFont()).getMaxAscent());
      g.setColor(Color.gray);
      g.drawLine(0,getSize().height-1,getWidth(),getSize().height-1);
//      paintBorder(g);
    }

    public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus)
    {
      if(!(value instanceof Entity))
        return new JLabel(value.toString());
      setEntity((Entity) value);
      setComponentOrientation(list.getComponentOrientation());
      if (isSelected) {
        setBackground(list.getSelectionBackground());
        setForeground(list.getSelectionForeground());
      }
      else
      {
        setBackground(list.getBackground());
        setForeground(list.getForeground());
      }
      return this;
    }
  }

  class EntityListTransferHandler extends TransferHandler
  {
    Icon dataIcon;
    public EntityListTransferHandler()
    {
      super();
      setDropTarget(null);
    }

    public boolean importData(JComponent c, Transferable t)
    {
      System.out.println("importData");
      try
      {
        if(t.isDataFlavorSupported(EntityTransferable.getEntityDataFlavor()))
        {
          Object entityIDs = (Object) t.getTransferData(EntityTransferable.getEntityDataFlavor());
        }
        return true;
      }
      catch(Exception e)
      {
      }
      return false;
    }

    protected void exportDone(JComponent c, Transferable data, int action)
    {
      System.out.println("exportDone");
      repaint();
    }

    public boolean canImport(JComponent c, DataFlavor[] flavors)
    {
      return false;
//      System.out.println("canImport");
//      return !c.equals(this) && !Arrays.asList(flavors).contains(EntityTransferable.getEntityDataFlavor());
    }

    protected Transferable createTransferable(JComponent c)
    {
      System.out.println("createTransferable");
//      renderer = new ConceptMapEntityListRenderer();
//      renderer.setEntity((Entity) ((EntityList) c).getSelectedValue());

      dataIcon = new Icon() {
        public int getIconWidth()
        {
          return 20;
        }
        public int getIconHeight()
        {
          return 20;
        }
        public void paintIcon(Component c, Graphics g, int x, int y)
        {
          g.setColor(Color.BLUE);
          g.fillOval(x,y, 20, 20);
        }

      };

      return new EntityTransferable((Entity) ((EntityList) c).getSelectedValue());
    }

    public int getSourceActions(JComponent c)
    {
        return COPY_OR_MOVE;
    }

    public Icon getVisualRepresentation(Transferable data)
    {
      System.out.println("!");
      return dataIcon;
    }
  }
}
