package jcreek.gui.conceptmap;

/**
 * Simple implementation of a graph node that in addition stores a reference
 * to a Creek Entity.
 *
 * @author Frode Srmo, Victor Dias
 * @version 1.1
 * @see jcreek.graph.Node
 * @see jcreek.representation.Entity
 */

import java.awt.Point;

import jcreek.representation.Entity;
import jcreek.representation.KnowledgeModel;

import jcreek.util.graph.*;

public class VisualEntity extends Node
{
  private transient Entity creekLEntity;
  private Object entityID;

//    protected Color DEFAULT_COLOR = new Color(134,134,174);
//    protected Color DEFAULT_COLOR = new Color(149,149,255);
//    protected Color FONT_COLOR = new Color(0,0,0);

  private static final long serialVersionUID=6685398730103073216L;


  /**
   * Creates a new VisualEntity.
   *
   * @param entity The entity in a KnowledgeModel represented by this VisualEntity.
   * @param startX The x coordinate in a 2-d space representing the initial position of the VisualEntity.
   * @param startY The y coordinate in a 2-d space representing the initial position of the VisualEntity.
   */
  public VisualEntity(Entity entity, int startX, int startY)
  {
    super(startX, startY);
    creekLEntity = entity;
    entityID = entity.getID();
  }

  /**
   * Creates a new VisualEntity.
   *
   * @param entity The entity in a KnowledgeModel represented by this VisualEntity.
   * @param pos The point in a 2-d space representing the initial position of the VisualEntity.
   */
  public VisualEntity(Entity entity, Point pos)
  {
    this(entity, pos.x, pos.y);
  }

  /**
   * Returns true if this VisualEntity is a representation of the given entity,
   * i.e. if getEntity().equals(e).
   */
  public boolean represents(Entity e)
  {
    return creekLEntity.equals(e);
  }

  public Object clone()
  {
    VisualEntity copy = new VisualEntity(getEntity(), getPosition().x, this.getPosition().y);
    copy.graph = getGraph();
    return copy;
  }

  /**
   * Add this entity to the selected set of entities in the current graph.
   */
  public void setSelected(boolean selected)
  {
    if(getGraph() instanceof ConceptMapModel)
    {
      ((ConceptMapModel) getGraph()).getSelectionSet().add(this);
    }
  }

  /**
   * Returns true if this VisualEntity is part of the selection set of the current
   * graph.
   */
  public boolean isSelected()
  {
    return ((ConceptMapModel) getGraph()).getSelectionSet().contains(this);
  }

  /**
   * Returns the Creek Entity this VisualEntity encapsulates.
   */
  public Entity getEntity()
  {
    return creekLEntity;
  }

  /**
   * Overridden to fulfill hasCode contract as equals() is overridden.
   */
  public int hashCode()
  {
    return 11+creekLEntity.hashCode();
  }

  /**
   * Two VisualEntities are considered equal if they encapsulate the same Creek
   * Entity.
   */
  public boolean equals(Object ob)
  {
    if(ob instanceof VisualEntity)
      return ((VisualEntity) ob).creekLEntity.equals(creekLEntity);
    else
      return false;
  }

  /**
   * Returns a string representation of this VisualEntity, including the name
   * of the underlaying entity and the position of the node.
   */
  public String toString()
  {
    return "VisualEntity for "+getEntity().getName()+" at "+getPosition().x+", "+getPosition().y;
  }

  /**
   * After the VisualEntity is loaded, this method must be used to connect it
   * to a KnowledgeModel.
   */
  void connectAfterLoad(KnowledgeModel model)
  {
    creekLEntity = model.getEntity(entityID);
  }

}
