package jcreek.gui.conceptmap.tool;

/**
 * Tool for using on a ConceptMapView that will allow the user to edit the model
 * by adding relations and entities.
 *
 * When pressing down the mouse button on a VisualEntity in the view, the default
 * action of this tool is to start drawing a relation. The target of this relation
 * is taken to be at the point where the mouse button is released.
 *
 * If no current relation target is set (using the setCurrentRelationType() )
 * method, or the CTRL-button is held while pressing the mouse button, this
 * tool will behave as a DragTool - i.e. start moving the entity around.
 *
 * When the user clicks anywhere that is not occupied by a VisualEntity, a new
 * VisualEntity will be created (if and only if setAllowNewEntities() is set to
 * true, which is default).
 *
 * @author Frode Srmo
 * @version 1.0
 */

import java.awt.event.*;
import java.awt.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import java.awt.dnd.DragSourceContext;
import javax.swing.*;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.gui.conceptmap.*;
import java.util.*;

public class BuildTool extends DragTool // implements KeyListener
{
  private RelationType currentRelationType = null;
  private boolean allowNewEntities = true;
  private boolean makeNewRelation = false;

  private static Stroke myStroke = new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND, 0.0f, new float[] { 3.0f, 3.0f }, 0.0f);
  private TransferHandler transferHandler;
  private TransferHandler origHandler;

  private Point initialPoint;
  VisualRelation pressRelation;

  private Action removeAction = new AbstractAction("Delete")
  {
      public void actionPerformed(ActionEvent e)
      {
        removeSelection();
      }
  };


  /**
   * Creates a new BuildTool.
   */
  public BuildTool()
  {
    super();
  }

  /**
   * Creates a new BuildTool and registers it with this ConceptMapView.
   *
   * @param conceptMapView The ConceptMapView to add this tool to.
   */
  public BuildTool(ConceptMapView conceptMapView)
  {
    super(conceptMapView);
  }

  public void attach(ConceptMapView view)
  {
    super.attach(view);
    origHandler = view.getTransferHandler();
    view.setTransferHandler(new ConceptMapViewTransferHandler(view, this));

    Container cont = view.getParent();
    view.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.
          getKeyStroke(KeyEvent.VK_DELETE, 0), "Delete");
    view.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.
          getKeyStroke(KeyEvent.VK_BACK_SPACE, 0), "Delete");
    view.getActionMap().put("Delete", removeAction);
  }

  public Action getRemoveAction()
  {
    return removeAction;
  }

  public void detach()
  {
    view.setTransferHandler(origHandler);
    Container cont = view.getParent();

    view.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).
          remove(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
    view.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).
          remove(KeyStroke.getKeyStroke(KeyEvent.VK_BACK_SPACE, 0));
    view.getActionMap().remove("Delete");

    super.detatch();
  }

  /**
   * Sets the current relation type to be drawn when creating new relations.
   *
   * @param drawType The RelationType to make when drawing new relations.
   */
  public void setCurrentRelationType(RelationType drawType)
  {
    this.currentRelationType = drawType;
  }

  /**
   * Returns the current relation type used when making new relations.
   */
  public RelationType getCurrentRelationType()
  {
    return currentRelationType;
  }

  /**
   * Sets if this tools is allowed to create new entities or not. Default is true.
   *
   * @param allowEntities If true, new entities are allowed to be made, if false, not.
   */
  public void setAllowNewEntities(boolean allowEntities)
  {
    this.allowNewEntities = allowEntities;
  }

  /**
   * Returns true if this tools is allowed to make new entities.
   */
  public boolean getAllowNewEntities()
  {
    return allowNewEntities;
  }

  public void mouseClicked( MouseEvent e )
  {
    view.getModel().getSelectionSet().clear();

    VisualEntity ve = view.findVisualEntity(e.getPoint());
    if(ve != null)
    {
      view.getModel().getSelectionSet().add(ve);
    }
    else
    {
      VisualRelation vr = view.findVisualRelation(e.getPoint());

      if(vr != null)
      {
        view.getModel().getSelectionSet().add(vr);
//        view.repaint();
      }
      else if(allowNewEntities)
        createVisualEntity(e.getX(), e.getY());
    }
//    view.repaint();
  } // mouseClicked

  public void mouseDragged( MouseEvent me )
  {

    if(makeNewRelation)
      continueMakeRelation(me.getX(), me.getY());
    else if(pressRelation != null)
    {
      if(me.getPoint().distance(initialPoint) > 5)
      {
        view.getModel().getSelectionSet().clear();
        view.getModel().getSelectionSet().add(pressRelation);
        view.getTransferHandler().exportAsDrag(view, me, TransferHandler.MOVE);
      }
    }
    else
      continueDrag(me.getX(), me.getY());
  }

  public void mouseEntered( MouseEvent e )
  {
  } // mouseClicked

  public void mouseExited( MouseEvent e )
  {
    if(pressVertex != null && view != null && !makeNewRelation)
    {
      view.getTransferHandler().exportAsDrag(view, e, TransferHandler.MOVE);
      stopDrag();
    }
  } // mouseClicked

  public void mouseMoved( MouseEvent e )
  {
  } // mouseClicked

  public void mousePressed( MouseEvent me )
  {
    view.getModel().getSelectionSet().clear();
    VisualEntity ve = view.findVisualEntity(me.getPoint());
    if(ve == null)
    {
      VisualRelation vr = view.findVisualRelation(me.getPoint());
      if(vr != null)
      {
        pressRelation = vr;
        initialPoint = me.getPoint();
      }
    }
    else if(!me.isControlDown() && me.getButton() == MouseEvent.BUTTON1)
    {
      startDrag(me.getX(), me.getY());
    }
    else if(ve != null)
    {
      startMakeRelation(me.getX(), me.getY());
    }
  } // mouseClicked

  public void mouseReleased( MouseEvent me )
  {
    if(makeNewRelation)
      stopMakeRelation();
    else
      stopDrag();
    if(pressRelation != null)
      pressRelation = null;
  }
/*
  public void keyReleased(KeyEvent e)
  {
  }

  public void keyPressed(KeyEvent e)
  {
  }

  public void keyTyped(KeyEvent e)
  {
    switch(e.getKeyCode())
    {
      case KeyEvent.VK_DELETE:
        removeSelection(); break;
      case KeyEvent.VK_BACK_SPACE:
        removeSelection(); break;
    }
  }
*/
  protected synchronized void removeSelection()
  {
    System.out.println("removeSelection()");
    for(Iterator it =((Set) view.getModel().getSelectionSet().clone()).iterator();it.hasNext();)
    {
      Object next = it.next();
      if(next instanceof VisualRelation)
        view.getModel().removeEdge((VisualRelation) next);
      if(next instanceof VisualEntity)
        view.getModel().removeNode((VisualEntity) next);
    }
  }

  protected void startMakeRelation(int x, int y)
  {
    makeNewRelation = true;
    startDrag(x, y);
  }

  protected void continueMakeRelation(int x, int y)
  {
    if (!wasPressed || pressVertex == null)
    {
      return;
    }
    lastMousePos.x =  x;
    lastMousePos.y =  y;

    if ( ( currentRelationType != null ) &&
         ( Math.abs( lastMousePos.x - firstMousePos.x ) < 3 ) &&
         ( Math.abs( lastMousePos.y - firstMousePos.y ) < 3 ) )
    {
      // We ignore tiny mouse movements until we being
      // doing something, such as moving vertices or
      // selecting a region.
      return;
    }

    // First make sure the vertex the user clicked on is
    // selected (putting it in the selected list). This
    // makes it possible to move several vertices at once.
    pressVertex.setSelected(true);
    view.repaint();

  }

  protected void stopMakeRelation()
  {
    makeNewRelation = false;
    VisualEntity target = view.findVisualEntity(lastMousePos);
    if(currentRelationType != null && target != null)
    {
      Relation newRel = pressVertex.getEntity().addRelation(currentRelationType, target.getEntity());
      VisualRelation vr = new VisualRelation(pressVertex, target, newRel);
      view.getModel().addEdge(vr);
    }
    else
      view.repaint();
  }

  protected void createVisualEntity(int x, int y)
  {
    String name = "New Entity";
    int i=1;
    Entity newEnt = null;
    while(newEnt == null)
    {
      try
      {
        newEnt = new Entity(view.getModel().getKnowledgeModel(), name, "");
      }
      catch(NameAlreadyExistException e)
      {
        newEnt = null;
        i++;
        name = "New Entity ("+i+")";
      }
    }

    view.getModel().addEntity(newEnt, view.viewToModelTransform(new Point(x, y)));
  }

  public void paintTool(Graphics g)
  {
    if(makeNewRelation)
    {
//      System.out.println("paintTool");
      Graphics2D g2 = (Graphics2D) g;
      g2.setColor(Color.BLUE);
      Stroke cs = g2.getStroke();
      g2.setStroke(myStroke);
      Point viewP = view.modelToViewTransform(pressVertex.getPosition());
      g2.drawLine(viewP.x, viewP.y, lastMousePos.x, lastMousePos.y);
      g2.setStroke(cs);
    }
  }
}

class ConceptMapViewTransferHandler extends TransferHandler
{
//  Icon dataIcon;
  ConceptMapView view;
  BuildTool tool;
//  Image image;
//  private static MySwingDragGestureRecognizer recognizer = null;
  public ConceptMapViewTransferHandler(ConceptMapView view, BuildTool tool)
  {
    super();
    this.view = view;
    this.tool = tool;
/*    image = view.getToolkit().getImage(
            getClass().getResource("/jcreek/its/trashcan-small.jpg"));
    MediaTracker tracker = new MediaTracker(view);
    tracker.addImage(image,0);

    try
    {
      tracker.waitForAll();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    } */
  }
/*
  public void exportAsDrag(JComponent comp, InputEvent e, int action)
  {
    int srcActions = getSourceActions(comp);
    int dragAction = srcActions & action;
    if (! (e instanceof MouseEvent)) {
      // only mouse events supported for drag operations
      dragAction = NONE;
    }
    if (dragAction != NONE && !GraphicsEnvironment.isHeadless()) {
      if (recognizer == null) {
        recognizer = new SwingDragGestureRecognizer(new MyDragHandler());
      }
      recognizer.gestured(comp, (MouseEvent)e, srcActions, dragAction);
    } else {
      exportDone(comp, null, NONE);
    }
  }*/

  public boolean importData(JComponent c, Transferable t)
  {
//    System.out.println("ConceptMapView.importData");
    // Handled by the ConceptMapViewDropHandler.
    return false;
  }

  protected void exportDone(JComponent source, Transferable data, int action)
  {
//    System.out.println("ConceptMapView.exportDone "+source);
//    if(action == TransferHandler.MOVE)
//    {

//    }

//    view.repaint();
  }

  public boolean canImport(JComponent c, DataFlavor[] flavors)
  {
//    System.out.println("ConceptMapView.canImport");
    return !c.equals(this) && !Arrays.asList(flavors).contains(EntityTransferable.getEntityDataFlavor());
  }

  protected Transferable createTransferable(JComponent c)
  {
    if(!c.equals(view))
      return null;
//    System.out.println("ConceptMapView.createTransferable");
//      renderer = new ConceptMapEntityListRenderer();
//      renderer.setEntity((Entity) ((EntityList) c).getSelectedValue());

/*    dataIcon = new Icon() {
      public int getIconWidth()
      {
        return 20;
      }
      public int getIconHeight()
      {
        return 20;
      }
      public void paintIcon(Component c, Graphics g, int x, int y)
      {
        g.setColor(Color.BLUE);
        g.fillOval(x,y, 20, 20);
      }

    }; */

    if(tool.pressRelation != null)
    {
      Transferable t = new RelationTransferable(tool.pressRelation.getRelation());
      tool.pressRelation = null;
      return t;
    }
    else
      return new EntityTransferable(tool.pressVertex.getEntity());
  }

  public int getSourceActions(JComponent c)
  {
      return COPY_OR_MOVE;
  }
/*
  public Icon getVisualRepresentation(Transferable data)
  {
    System.out.println("!");
    return dataIcon;
  }

  private class MyDragHandler implements DragGestureListener, DragSourceListener
  {

    private boolean scrolls;

       // --- DragGestureListener methods -----------------------------------

    public void dragGestureRecognized(DragGestureEvent dge)
    {
      JComponent c = (JComponent) dge.getComponent();
      ConceptMapViewTransferHandler th = (ConceptMapViewTransferHandler) c.getTransferHandler();
      Transferable t = th.createTransferable(c);
      if (t != null) {
        scrolls = c.getAutoscrolls();
        c.setAutoscrolls(false);
        try {
          dge.startDrag(view.getToolkit().getDefaultToolkit().createCustomCursor(image, new Point(10, 10), "test")
                        , image, new Point(10, 10), t, this);
//          dge.startDrag(null, t, this);
          return;
        } catch (RuntimeException re) {
          c.setAutoscrolls(scrolls);
        }
      }

      th.exportDone(c, null, NONE);
    }

    // --- DragSourceListener methods -----------------------------------

    public void dragEnter(DragSourceDragEvent dsde)
    {

    }

    public void dragOver(DragSourceDragEvent dsde) {
    }

    public void dragExit(DragSourceEvent dsde) {
    }

    public void dragDropEnd(DragSourceDropEvent dsde)
    {
      DragSourceContext dsc = dsde.getDragSourceContext();
      JComponent c = (JComponent)dsc.getComponent();
      if (dsde.getDropSuccess()) {
        ((ConceptMapViewTransferHandler) c.getTransferHandler()).exportDone(c, dsc.getTransferable(), dsde.getDropAction());
      } else {
        ((ConceptMapViewTransferHandler) c.getTransferHandler()).exportDone(c, null, NONE);
      }
      c.setAutoscrolls(scrolls);
    }

    public void dropActionChanged(DragSourceDragEvent dsde) {
    }
  }
  private static class MySwingDragGestureRecognizer extends DragGestureRecognizer {

      MySwingDragGestureRecognizer(DragGestureListener dgl) {
          super(DragSource.getDefaultDragSource(), null, NONE, dgl);
      }

      void gestured(JComponent c, MouseEvent e, int srcActions, int action) {
          setComponent(c);
          setSourceActions(srcActions);
          appendEvent(e);
          fireDragGestureRecognized(action, e.getPoint());
      }

      protected void registerListeners() {
      }

      protected void unregisterListeners() {
      }

  }*/

}
