package jcreek.gui.conceptmap.tool;

/**
 * This class implements a ConceptMapTool that when attached to a ConceptMapView
 * will allow the user to:
 *   * Drag and drop individual entities around the ConceptMap.
 *   * Manipulate the selection set by:
 *      - Setting the selection set to an entity when left-clicked on.
 *      - Adding/removing an entity to the selection set when CTRL-left-cliced on.
 *   * Moving the selection set around when one of the entities in the set is
 *     moved.
 */

//import jcreek.util.*;
import jcreek.util.graph.*;
import jcreek.gui.conceptmap.*;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


public class DragTool implements ConceptMapTool, MouseListener, MouseMotionListener
 {
    /**
     * View that we're associated with.
     */
    protected ConceptMapView view;

    /**
     * Vertex the user pressed on in mousePressed(). Could be
     * null if the user didn't press on a entity.
     */
    protected VisualEntity pressVertex;

    /**
     * True if mouse was pressed on the view before being released.
     * Helps avoid some mis-firings of the mouse buttons.
     */
    protected boolean wasPressed;

    /**
     * The set of moving entites.
     */
    protected HashSet movingEntities;

    /**
     * The list of the relations being moved around.
     */
    protected Point firstMousePos;
    /**
     * Most recent mouse position while dragging.
     */
    protected Point lastMousePos;

    /**
     * Creates a new DragTool object.
     */
    public DragTool()
    {
        firstMousePos = new Point();
        lastMousePos = new Point();
    } // DragTool

    public void paintTool(Graphics g)
    {
    }

    /**
     * Creates a new DragTool and attaches it to the supplied ConceptMapView.
     *
     * @param view The ConceptMapView to attach this tool to.
     */
    public DragTool(ConceptMapView view)
    {
      this();
      attach(view);
    }

    public void attach(ConceptMapView view)
    {
        this.view = view;
        view.setTool(this);
        view.addMouseListener(this);
        view.addMouseMotionListener(this);
    }

    public void detatch()
    {
        if ( view != null ) {
            view.removeMouseListener( this );
            view = null;
        }
    }

    public void mouseClicked( MouseEvent e )
    {
    } // mouseClicked

    public void mouseDragged( MouseEvent me )
    {
      continueDrag(me.getX(), me.getY());
    }

    public void mouseEntered( MouseEvent e )
    {
    } // mouseClicked

    public void mouseExited( MouseEvent e )
    {
    } // mouseClicked

    public void mouseMoved( MouseEvent e )
    {
    } // mouseClicked

    public void mousePressed( MouseEvent me )
    {
        if ( !me.isControlDown() ) {
            // If user is not holding down the Ctrl key, deselect all.
            view.getModel().getSelectionSet().clear();
        }

        startDrag(me.getX(), me.getY());
        // Remember the mouse position for when we have to
        // move entities around the view.
    } // mouseClicked

    public void mouseReleased( MouseEvent me )
    {
      stopDrag();
    }

    protected void startDrag(int x, int y)
    {
      firstMousePos.x = x;
      firstMousePos.y = y;

      lastMousePos.x = firstMousePos.x;
      lastMousePos.y = firstMousePos.y;

      pressVertex = view.findVisualEntity(firstMousePos);
      view.getModel().getSelectionSet().add(pressVertex);

//        System.out.println("PressVertex: "+pressVertex);
      wasPressed = true;
    }

    protected void continueDrag(int x, int y)
    {
      // Ignore drags without matching presses, and ignore drags
      // of insignificant movement (i.e. less than 3 pixels).

      if (!wasPressed || pressVertex == null)
      {
          return;
      }
      lastMousePos.x =  x;
      lastMousePos.y =  y;

      if ( ( movingEntities == null ) &&
           ( Math.abs( lastMousePos.x - firstMousePos.x ) < 3 ) &&
           ( Math.abs( lastMousePos.y - firstMousePos.y ) < 3 ) ) {
          // We ignore tiny mouse movements until we being
          // doing something, such as moving vertices or
          // selecting a region.
          return;
        }
      // First make sure the vertex the user clicked on is
      // selected (putting it in the selected list). This
      // makes it possible to move several vertices at once.
      pressVertex.setSelected(true);

      // If there are some vertices selected, deal with moving
      // them around the view.
      if (!view.getModel().getSelectionSet().isEmpty())
      {
          if (movingEntities == null )
          {
              // Make a copy of the selected vertices so we can
              // modify their positions based off the originals.
              // We'll use the copy in resetMovingVertices() to
              // restore the original positions.
              movingEntities = new HashSet(view.getModel().getSelectionSet());
              for(Iterator it=movingEntities.iterator();it.hasNext();)
              {
                ((VisualEntity) it.next()).suspendEventReport(true);
              }
          }
          VisualEntity v1 = pressVertex;
          Point pos = view.modelToViewTransform(v1.getPosition());

          int deltaX = lastMousePos.x - pos.x;
          int deltaY = lastMousePos.y - pos.y;
          for (Iterator it=movingEntities.iterator();it.hasNext();)
          {
            // Move the original vertices based off the
            // position of the copies.
            ((VisualEntity) it.next()).translate(deltaX,deltaY);
          }
          view.repaint();
        }
    }

    protected void stopDrag()
    {
      if ( wasPressed == false )
      {
        // This avoids the case where the user clicks outside
        // of the view and drags the mouse onto the view.
        // This takes care of menu selections or clicking
        // the "Ok" button of a dialog shown over the view.
        return;
      }
      wasPressed = false;

      if (movingEntities != null && pressVertex != null )
      {
        // Pressed on a vertex, toggle the vertex selection
        // and remember the last vertices selected.
        if ( pressVertex.isSelected() ) {
          pressVertex.setSelected(false);
        } else {
          pressVertex.setSelected(true);
        }
      }
      if(movingEntities != null)
      {
        for(Iterator it=movingEntities.iterator();it.hasNext();)
        {
          ((VisualEntity) it.next()).suspendEventReport(false);
        }
      }
      movingEntities = null;
      pressVertex = null;
    }

 }//DragTool
