package jcreek.gui.cycle;

/**
 * Creek Cycle is an example simple CBR application that uses the existing GUI
 * components of Creek to perform the full Case-Based Reasoninc Cycle.
 *
 * @author Frode Srmo
 * @version 1.0
 */

import java.awt.*;
import javax.swing.*;
import java.awt.event.ActionEvent;
import jcreek.representation.*;
import jcreek.representation.cbr.*;
import jcreek.reasoning.*;
import jcreek.gui.reasoning.*;
import jcreek.gui.representation.*;

public class Cycle extends JFrame
{
  private KnowledgeModel km;
  private int cycleStatus = -1;
  private Case currentCase;

  private CasePane inputPane = null;
  private RetrieveResultPane retrievePane = null;
  private ReuseResultPane reusePane = null;
  private CasePane retainPane = null;

  public static final int CYCLE_NEW_CASE = 0;
  public static final int CYCLE_RETRIEVE = 1;
  public static final int CYCLE_REUSE = 2;
  public static final int CYCLE_REVISE_AND_RETAIN = 3;

  public Action nextStep = new NextStepAction("Next Step >>");
  public Action previousStep = new PreviousStepAction("<< Previous Step");
  public Action openModel = new OpenModelAction("Open Creek Model");

  public Cycle()
  {
    super("Creek Cycle");
    setJMenuBar(buildMenu());
  }

  private JMenuBar buildMenu()
  {
    JMenuBar menu = new JMenuBar();
    JMenu fileMenu = new JMenu("File");
    fileMenu.add(openModel);
    fileMenu.add("Save");
    fileMenu.add("Save as...");
    fileMenu.addSeparator();
    fileMenu.add("Exit");

    JMenu actionMenu = new JMenu("Actions");
    actionMenu.add("New Case");
    actionMenu.add(nextStep);
    actionMenu.add(previousStep);

    JMenu helpMenu = new JMenu("Help");
    helpMenu.add("About Cycle");

    menu.add(fileMenu);
    menu.add(actionMenu);
    menu.add(helpMenu);
    return menu;
  }

  public void openKnowledgeModel(KnowledgeModel km)
  {
    this.km = km;
    openNewCase();
    setTitle(km.getName()+" - Creek Cycle");
  }

  public void setCycleStatus(int status)
  {
    cycleStatus = status;

    if(cycleStatus == CYCLE_NEW_CASE)
    {
      nextStep.setEnabled(true);
      previousStep.setEnabled(false);

      if (inputPane == null)
      {
        inputPane = new CasePane(currentCase);
        inputPane.setShowCaseStatus(false);
        inputPane.setEditable(true);
      }

      currentCase.setStatus(Case.UNSOLVEDCASE);
      setContentPane(inputPane);
      validate();
      repaint();
    }
    else if(cycleStatus == CYCLE_RETRIEVE)
    {
      nextStep.setEnabled(true);
      previousStep.setEnabled(true);

      if(retrievePane == null)
      {
        retrievePane = new RetrieveResultPane(new RetrieveResult(currentCase));
      }
      else
        retrievePane.updateContents();

      setContentPane(retrievePane);
    }
    else if(cycleStatus == CYCLE_REUSE)
    {
      nextStep.setEnabled(true);
      previousStep.setEnabled(true);

      if(reusePane == null)
      {
        if (retrievePane == null)
          retrievePane = new RetrieveResultPane(new RetrieveResult(currentCase));

        RetrieveResult retrieve = retrievePane.getRetrieveResult();
        reusePane = new ReuseResultPane(new ReuseResult(retrieve));
        reusePane.attach();
      }
      else
      {
        ReuseResult reuse = reusePane.getReuseResult();
        reuse.reset();
        reuse.activate();
        reuse.explain();
        reuse.focus();
      }
      setContentPane(reusePane);
      validate();
      repaint();
    }
    else if(cycleStatus == CYCLE_REVISE_AND_RETAIN)
    {
      nextStep.setEnabled(true);
      previousStep.setEnabled(true);

      if(retainPane == null)
      {
        retainPane = new CasePane(currentCase);
        retainPane.setCaseStatusDisplay(CasePane.SOLVED_STATUS_EDITABLE);

        currentCase.setStatus(Case.PROCESSCASE);

      }
      if(reusePane != null)
      {
        ReuseResult reuse = reusePane.getReuseResult();
        currentCase.setSolution(reuse.getSolution());
      }

      setContentPane(reusePane);
      validate();
      repaint();
    }
  }

  public int getCycleStatus()
  {
    return cycleStatus;
  }

  public KnowledgeModel getKnowledgeModel()
  {
    return km;
  }

  public void openNewCase()
  {
    currentCase = null;
    int cases = km.matchEntityType(new Case[0]).length;
    while(currentCase == null)
    {
      try
      {
        currentCase = new Case(km,"Case "+cases, "Case "+cases);
      }
      catch(NameAlreadyExistException e)
      {
        cases++;
      }
    }

    inputPane = null;
    retrievePane = null;
    reusePane = null;
    retainPane = null;

    setCycleStatus(CYCLE_NEW_CASE);
  }

  public static void main(String argv[])
  {
    Cycle cycle = new Cycle();
    cycle.setSize(800,600);
    cycle.show();
  }

  class NextStepAction extends AbstractAction
  {
    public NextStepAction(String name)
    {
      super(name);
    }

    public void actionPerformed(ActionEvent e)
    {
      if(getCycleStatus() >= CYCLE_NEW_CASE)
      {
        if(getCycleStatus() == CYCLE_REVISE_AND_RETAIN)
          openNewCase();
        else
        {
          setCycleStatus(getCycleStatus() + 1);
        }
      }
    }
  }

  class PreviousStepAction extends AbstractAction
  {
    public PreviousStepAction(String name)
    {
      super(name);
    }

    public void actionPerformed(ActionEvent e)
    {
      if(getCycleStatus() > CYCLE_NEW_CASE)
      {
        setCycleStatus(getCycleStatus()-1);
      }
    }
  }

  class OpenModelAction extends AbstractAction
  {
    public OpenModelAction(String name)
    {
      super(name);
    }

    public void actionPerformed(ActionEvent e)
    {
      try
      {
        openKnowledgeModel(new LocalKnowledgeModel("creekExample.km"));
      }
      catch(Exception ex)
      {
        ex.printStackTrace();
      }
    }
  }
}
