package jcreek.gui.reasoning;

/**
 * This Swing GUI element provides a framework for creating custom GUI elements
 * for different similarity measures as implemented by subclasses of the
 * jcreek.reasoning.EntityComparison class. Subclasses of this class do not
 * have to implement a subclass of this class - this class will be used as a
 * default for displaying the results. EntityComparisonPane uses the toSubModel()
 * and shortExplanation() methods in order to do so.
 *
 * While custom EntityComparison classes are found by lookup in the knowledge
 * model, the GUI elements are not referenced in this way in order to avoid
 * dependencies between model and the GUI packages. In order to associate a new
 * similarity measure (EntityComparison subclass) with its corresponding
 * GUI element, please edit the jcreek/reasoning/comparisonpanes.properties file.
 * Adding the line jcreek.reasoning.MyComparison = jcreek.gui.reasoning.MyComparisonPane
 * will signify that the MyComparisonPane class will be used to display the
 * MyComparison similarity measure. This resource file is used by the static
 * method createComparisonPane(EntityComparison) method of this class. It will
 * generate the most specific GUI elment possible to display the EntiyComparison
 * given as an argument. This method is used by RetrieveResultPane, CaseComparisonPane
 * and FrameMatchPane to create some of its elements. This means that to have GUI
 * elements for new similarity measures form an integral part of the existing GUI
 * applications, all a developer have to do is to create a subclass of this class
 * and add it to the comparisonpanes.properties file.
 *
 * There are two types of EntityComparisonPanes - large and small. Small panes
 * are usually displayed as a part of a larger GUI pane, while the large panes
 * are displayed by filling the whole window. The latter case is particularly
 * useful for recursive cases, e.g. with CaseComparisonPane, where a case can contain
 * a subcase that will also use a CaseComparisonPane and so on. By setting
 * this pane to be of the large type, we can ensure that the nested CaseComparisonPanes
 * will not be limited to an ever decreasing part of the screen realestate.
 *
 * Note that as EntityComparisons are immutable, these elements displaying them
 * are also immutable and will not update if the underlaying model is changed.
 *
 * @author Frode Srmo
 * @version 2.0
 */

import java.awt.GridLayout;
import javax.swing.*;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.util.*;
import java.io.InputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.reasoning.*;

public class EntityComparisonPane extends JPanel
{
  private EntityComparison comparison;
  protected int type;
  protected static Properties comparisonPanes;

  public static final int PANE_TYPE_LARGE = 0;
  public static final int PANE_TYPE_SMALL = 1;

  /**
   * Creates a new EntityComparisonPane.
   *
   * Note that a subclass of EntityComparisonPane must implement a constructor
   * with this signature as it is assumed to exist by the createComparisonPane
   * method.
   *
   * @param comp The EntityComparison object to display.
   */
  public EntityComparisonPane(EntityComparison comp)
  {
    super();
    this.comparison = comp;
    this.type = PANE_TYPE_SMALL;
    buildPane();
  }

  /**
   * Returns the type of component this is. A large component means that it should
   * fill the screen realestate as far as possible, while a small pane means that
   * it may be nested.
   *
   * @return int corresponding to one of the constants PANE_TYPE_SMALL or PANE_TYPE_LARGE
   * as defined by this class.
   */
  public int getPaneType()
  {
    return type;
  }

  /**
   * Builds this Swing component.
   */
  protected void buildPane()
  {
    setLayout(new BorderLayout());
    add(buildSubModelView(), BorderLayout.CENTER);
    add(buildDescriptionPane(), BorderLayout.EAST);
  }

  protected JComponent buildDescriptionPane()
  {
    JPanel info = new JPanel();
    info.setLayout(new GridLayout(6,2));
    info.add(new JLabel("Origin"));
    info.add(new JLabel(comparison.getOrigin().getValueString()));
    info.add(new JLabel("Target"));
    info.add(new JLabel(comparison.getTarget().getValueString()));
    info.add(new JLabel("Status"));
    info.add(new JLabel(comparison.getStatusText()));
    info.add(new JLabel("Activation strength"));
    info.add(new JLabel(""+(int) Math.round(comparison.getActivationStrength()*100)+"%"));
    info.add(new JLabel("Explanation strength"));
    info.add(new JLabel(""+(int) Math.round(comparison.getExplanationStrength()*100)+"%"));
    info.add(new JLabel("Total strength"));
    info.add(new JLabel(""+(int) Math.round(comparison.getStrength()*100)+"%"));
    JPanel explanationPanel = new JPanel(new BorderLayout());
    explanationPanel.add(info, BorderLayout.NORTH);
    JTextArea explanation = new JTextArea(10, 30);
    explanation.setEditable(false);
    explanation.setWrapStyleWord(true);
    explanation.setLineWrap(true);
    explanation.setText(comparison.shortExplanation());
    explanationPanel.add(new JScrollPane(explanation), BorderLayout.CENTER);
    return explanationPanel;
  }

  protected JComponent buildSubModelView()
  {
    return new ComparisonSubModelView(comparison.toSubModel(), comparison.getOrigin(), comparison.getTarget());
  }


  /**
   * Returns the EntityComparison this pane displays.
   *
   * @return EntityComparison
   */
  public EntityComparison getEntityComparison()
  {
    return comparison;
  }

  /**
   * Adds a "Back" button to the bottom of the pane and adds this pane to the
   * given container. When pressed, the Back button will change the contents
   * of the container to the previous pane.
   * @param containerPane The container component to place this component in.
   * @param preivousPane The "previous" component in the stack - this will replace this
   * pane as the component viewd in the containerPane when the Back-button is pressed.
   */
  public void addBackButton(JComponent containerPane, JComponent previousPane)
  {
    JPanel buttonPane = new JPanel();
    buttonPane.setBorder(BorderFactory.createEmptyBorder(2,2,2,2));
    buttonPane.add(new JButton(new BackAction(containerPane, previousPane)));
    JPanel container = new JPanel(new BorderLayout());
    container.add(this, BorderLayout.CENTER);
    container.add(buttonPane, BorderLayout.SOUTH);
    containerPane.removeAll();
    containerPane.add(container);
  }

  /**
   * Builds a EntityComparisonPane for the given EntityComparison. In doing this,
   * the file jcreek/reasoning/comparisonpanes.properties is used to find the most
   * specific pane possible. By this means that the class of the comparison given
   * will be looked up in the comparisonpanes.properties file. If it exists there,
   * this pane will be used. If not, the same process is repeated for the superclass
   * of the comparison and so on, until a suitable pane is found. If no pane is
   * found, an EntityComparisonPane is used by default.
   *
   * @param comp The comparison object to generate a Swing GUI element to display.
   * @return The Swing GUI element that displays the given comparison.
   */
  public static EntityComparisonPane createComparisonPane(EntityComparison comp)
  {
    KnowledgeModel km = comp.getKnowledgeModel();

    if(comparisonPanes == null)
    {
      try
      {
        comparisonPanes = new Properties();
        InputStream in = comp.getClass().getResourceAsStream("/jcreek/gui/reasoning/comparisonpanes.properties");
        if(in != null)
          comparisonPanes.load(in);
        else
          System.err.println( "EntityComparisonPane error: jcreek.gui.reasoning.comparisonpanes.properties failed to load!" );
      }
      catch ( java.io.IOException ex )
      {
          System.err.println( "EntityComparisonPane error: jcreek.gui.reasoning.comparisonpanes.properties failed to load!" );
          ex.printStackTrace();
      }

      if(comparisonPanes == null)
      {
        return new EntityComparisonPane(comp);
      }
    }

    Class compClass = comp.getClass();
    while(compClass != EntityComparison.class && compClass != Object.class)
    {
      if(comparisonPanes.containsKey(compClass.getName()))
      {
        String paneClassName = comparisonPanes.getProperty(compClass.getName());
        try
        {
          Class paneClass = Class.forName(paneClassName);
          Constructor construct = paneClass.getConstructor(new Class[]
            {compClass});
          return (EntityComparisonPane) construct.newInstance(new Object[]
                                {comp});
        }
        catch(Exception e)
        {
          System.err.println("EntityComparisonPane.createComparisonPane() : Failed to create "+paneClassName+" for "+comp.getClass()+".");
          e.printStackTrace();
        }
      }
      compClass = compClass.getSuperclass();
    }

    // Default
    return new EntityComparisonPane(comp);
  }

  class BackAction extends AbstractAction
  {
    JComponent container;
    JComponent previous;
    BackAction(JComponent container, JComponent previous)
    {
      super("Back");
      this.container = container;
      this.previous = previous;
    }
    public void actionPerformed(ActionEvent e)
    {
      container.removeAll();
      container.add(previous);
      container.validate();
      container.repaint();
    }
  }
}
