package jcreek.gui.reasoning;

import jcreek.gui.chart.*;
import jcreek.gui.conceptmap.DefaultConceptMapController;
import jcreek.gui.*;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.reasoning.*;

import javax.swing.*;
import java.awt.Color;
import java.awt.Point;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.*;

/**
 * This Swing component displays the contents of a ReuseResult. It contains a
 * bar graph showing the relative size of the different solutions suggested
 * by the top k cases (as given by the ReuseResult object) and displays the
 * textual description of the selected solution as well as the submodel for
 * the causal chains from the input case findings to the solution.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class ReuseResultPane extends JPanel implements ChartModel, ReasoningUpdateListener
{
  private ReuseResult reuse;
  private JPanel modelPanel = new JPanel(new BorderLayout());
  private boolean active = false;
  private JTextArea solutionText;
  private JLabel solutionLabel;
  private JChart chart;
  private int selected = 0;

  public ReuseResultPane(ReuseResult reuse)
  {
    super(new BorderLayout());
    this.reuse = reuse;
    if(reuse.isFocused())
      buildPane();
    else
      buildWaitPane();
  }

  void buildPane()
  {
    this.removeAll();
    JPanel bottomPanel = new JPanel(new BorderLayout());
    JPanel textPanel = new JPanel(new BorderLayout());
    textPanel.setMinimumSize(new Dimension(200, 200));
    textPanel.setPreferredSize(new Dimension(200, 300));
    solutionLabel = new JLabel("Solution description");
    solutionText = new JTextArea(10, 30);
//    textExplanationArea.setRows(10);
    solutionText.setEditable(false);
    solutionText.setWrapStyleWord(true);
    solutionText.setLineWrap(true);
    textPanel.add(solutionLabel,BorderLayout.NORTH);
    textPanel.add(solutionText,BorderLayout.CENTER);
    bottomPanel.add(new JScrollPane(textPanel), BorderLayout.EAST);

    JPanel chartPanel = new JPanel(new BorderLayout());
    chartPanel.add(new JLabel("Possible solutions"), BorderLayout.NORTH);
    chart = new JChart(this);
    chart.addMouseListener(new MouseAdapter() {
       public void mouseClicked(MouseEvent e)
       {
         setSolutionSelected(chart.getValueForPos(e.getX(), e.getY()));
       }
    });
    chart.setMinimumSize(new Dimension(200, 200));
    chart.setPreferredSize(new Dimension(300, 300));
    chartPanel.add(chart, BorderLayout.CENTER);
    chartPanel.setBorder(BorderFactory.createEtchedBorder());
    bottomPanel.add(chartPanel);
    add(bottomPanel, BorderLayout.SOUTH);
    modelPanel.setBorder(BorderFactory.createEtchedBorder());
    add(modelPanel, BorderLayout.CENTER);
    if(reuse.getSolution() != null)
      setSolutionSelected(0);
  }

  synchronized void buildWaitPaneIS()
  {
    SwingUtilities.invokeLater(new Runnable()
    {
        public void run()
        {
          buildWaitPane();
        }
    });
  }

  void buildWaitPane()
  {
    removeAll();
    setLayout(new BorderLayout());
    JProgressBar bar = new JProgressBar();
    bar.setIndeterminate(true);
    bar.setString("Waiting for Reuse to finish");
    bar.setStringPainted(true);
    add(bar, BorderLayout.SOUTH);
    validate();
    repaint();
  }

  public ReuseResult getReuseResult()
  {
    return reuse;
  }


  void setSolutionSelected(int selected)
  {
    this.selected = selected;
    Entity solution = reuse.getSolution(selected);
    modelPanel.removeAll();
    modelPanel.add(new JLabel("Causal model for possible solution '"+solution.getName()+"'"), BorderLayout.NORTH);
    modelPanel.add(new ReuseModelView(), BorderLayout.CENTER);
    solutionText.setText(solution.getDescription());
    validate();
    repaint();
  }

  // Implements methods for the ChartModel

  public double getValue(int col, int row)
  {
    Case c = reuse.getCasesForSolution(row)[col];
    for(int i=0;i<reuse.getK();i++)
    {
      if(c.equals(reuse.getCase(i)))
        return reuse.getCaseStrength(i);
    }
    return 0.0;
  }
  public String getRowLabel(int i)
  {
    return reuse.getSolution(i).getName();
  }
  public int getRowSize()
  {
    return reuse.size();
  }
  public int getColSize(int row)
  {
    return reuse.getCasesForSolution(row).length;
  }

  public double getMaxValue()
  {
    return 1.0;
  }

  public Color getValueColor(int col, int row)
  {
    return Color.blue;
  }

  public Color getBackgroundColor(int row)
  {
    if(row == selected)
     return new Color(150,150,240);
    else
      return Color.white;
  }

  public void attach()
  {
    if(!active)
    {
      active = true;
      reuse.getRetrieveResult().addReasoningUpdateListener(this);
      reuse.addReasoningUpdateListener(this);
    }
  }

  public void detach()
  {
    if(active)
    {
      active = false;
      reuse.getRetrieveResult().removeReasoningUpdateListener(this);
      reuse.removeReasoningUpdateListener(this);
    }
  }

  /**
   * Called whenever the update() method is called.
   */
  public void reasoningUpdateStarted(ReasoningUpdateEvent e)
  {
    if(e.getSource().equals(reuse))
    {

    }
    else
    {
      reuse.reset();
    }
  }

  /**
   * Called whenever a reasoning state is changed.
   * @param e ReasoningUpdateEvent
   */
  public void reasoningStateChanged(ReasoningUpdateEvent e)
  {
//    System.out.println("reasoningStateChanged in "+e.getSource()+".");
    if(e.getReasoningStep().isFocused())
    {
      if(e.getSource().equals(reuse))
      {
//        System.out.println("Building proper pane.");
        SwingUtilities.invokeLater(new Runnable()
        {
          public void run()
          {
            buildPane();
            validate();
            repaint();
          }
        });
      }
      else
      {
        // Executing this step may take a while, so it we operate under the assumption
        // that the executing thread is not dependendant on low latency. This should be
        // OK, since the thread have alredy executed at least one of the reasoning steps
        // in RetrieveResult().
        reuse.activate();
        reuse.explain();
        reuse.focus();
      }
    }
    if(e.getReasoningStep().getState() == ReuseResult.REUSE_STEP+ReuseResult.INITIAL_STATE && e.getSource().equals(reuse))
    {
      buildWaitPaneIS();
    }
  }

  /**
   * Called whenever the reasoning is aborted for some reason.
   * @param e ReasoningUpdateEvent
   */
  public void reasoningAborted(ReasoningUpdateEvent e)
  {
    if(e.getSource().equals(reuse))
    {

    }
    else
    {
      reuse.reset();
      buildWaitPaneIS();
    }
  }


  public String getValueLabel(int col, int row)
  {
    return reuse.getCasesForSolution(row)[col].getName();
  }

  class ReuseModelView extends SubModelView
  {
    ReuseModelView()
    {
      super(reuse.getKnowledgeModel(), reuse.getCausalModel(selected), true, true);
      this.setScale(0.8);
      this.setController(new DefaultConceptMapController()
      {
        public Dimension getViewInsets() {
          return new Dimension(60, 60);
        }
      });
    }

    public void firstResize()
    {
      fixEntity(reuse.getSolution(selected), new Point(getWidth(), getHeight()/2));
      fixEntity(reuse.getInputCase(), new Point(0, getHeight()/2));
      Relation[] findings = reuse.getInputCase().getRelations(reuse.getKnowledgeModel().getRelationType(CBRModel.FINDING_RELATION));
      for(int i=0;i<findings.length;i++)
        fixEntity(findings[i].getValue(), new Point(120, getHeight()*(i+1)/(findings.length+1)));
      super.firstResize();
    }
  }
}
