package jcreek.gui.representation;

import jcreek.representation.*;
import jcreek.representation.cbr.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.util.Hashtable;
import java.util.Arrays;
import jcreek.util.EntityNameComparator;

/**
 * GUI element for creating or manipulating a Creek case. An
 * instance of this class encapsulates a Case (as defined in jcreek.representation)
 * and allows the manipulation of it.</p>
 * @author Frode Srmo
 * @see jcreek.representation.Case
 * @version 1.0
 */

public class CasePane extends FramePane
{
  private int solvedStatus;
  private boolean showCaseType = true;

  /** The solved status of the case is not shown in the GUI element. */
  public static final int SOLVED_STATUS_HIDDEN = 1;

  /** The solved status of the case is visible but can't be changed. */
  public static final int SOLVED_STATUS_VISIBLE = 2;

  /** The solved status of the case is visible and can be changed. */
  public static final int SOLVED_STATUS_EDITABLE = 3;

  /** The strings displayed for the different case statuses. */
  public static final String[] CASE_STATUS_NAMES = new String[] {"Solved", "Processed", "Unsolved"}  ;

  private Hashtable caseElements = new Hashtable();

  protected JPanel elementPanel;
  protected JComponent caseStatus;

  protected CaseTableModel caseTableModel;
  protected JComponent addValue;
  protected JTable caseTable;
  protected JButton addSolutionButton;
  protected JComboBox addSolution;


  /**
   * Creates a new CasePane encapuslating an existing case. Changes done to this
   * case in the GUI will be applied to this case.
   *
   * @param existingCase The case displayed by this CasePane
   * @param linker If true, adds the default link listener that will open a new
   *               FramePane in a new JFrame when clicked.
   */
  public CasePane(Case existingCase, boolean linker)
  {
    super(existingCase, linker);
  }

  /**
   * Creates a new CasePane encapuslating an existing case. Changes done to this
   * case in the GUI will be applied to this case.
   *
   * @param existingCase The case displayed by this CasePane
   */
  public CasePane(Case existingCase)
  {
    this(existingCase, false);
  }

  /**
   * Creates a new CasePane for a new, empty case (instance of jcreek.representation.Case).
   * This case can be accessed through the getCase() method.
   *
   * @param km The KnowledgeModel the new case should be created in.
   */
  public CasePane(KnowledgeModel km)
  {
    super(km);
    int cases = km.matchEntityType(new Case[0]).length;
    Case newCase = null;
    while(newCase == null)
    {
      try
      {
        newCase = new Case(km,"Case "+cases, "Case "+cases);
      }
      catch(NameAlreadyExistException e)
      {
        cases++;
      }
    }
    setCase(newCase);
  }

  protected void initTable()
  {
    caseTable = new JTable();
    if(caseTableModel == null)
      if(getCase() != null)
        caseTableModel = new CaseTableModel(getCase().getStatus() != Case.UNSOLVEDCASE, new Relation[0]);
      else
        caseTableModel = new CaseTableModel(false, new Relation[0]);
    else
      caseTableModel.setShowStrengths(getCase().getStatus() != Case.UNSOLVEDCASE);
    caseTable.setModel(caseTableModel);
    caseTableModel.setTable(caseTable);
    caseTable.getTableHeader().setReorderingAllowed(false);

    if(caseTableModel != null)
      caseTableModel.setEditable(true);
    frameTableModel = caseTableModel;
  }


  public void setShowCaseStatus(boolean status)
  {
    showCaseType = status;
  }

  public boolean getShowCaseStatus()
  {
    return showCaseType;
  }

  /**
   * Sets the case to be displayed and modified by this CasePane.
   *
   * @param existingCase The case object.
   */
  public void setCase(Case existingCase)
  {
    setEntity(existingCase);
  }

  public void setEntity(Entity entity)
  {
    boolean a = isAttached();
    if(a)
      detach();
    myEntity = entity;
    buildFramePane();
    if(a)
      attach();
  }

  /**
   * Returns the case displayed and modified by this CasePane.
   *
   * @return The case object.
   */
  public Case getCase()
  {
    return (Case) getEntity();
  }


  /**
   * Sets if the status of the case should be hidden, visible or editable. If
   * it is hidden it will not be shown in the GUI element at all. If it is
   * visible, it will be shown but impossible to change. If it is editable, it
   * is visible and possible to change through the GUI. This value is separate
   * from the setEditable() value.
   *
   * @param display One of CasePane.SOLVED_STATUS_EDITABLE, CasePane.SOLVED_STATUS_HIDDEN,
   * or CasePane.SOLVED_STATUS_VISIBLE.
   */
  public void setCaseStatusDisplay(int display)
  {
    solvedStatus = display;
  }


  public int getCaseStatusDisplay()
  {
    return solvedStatus;
  }

  protected void addElementFor(Relation r, CasePaneElement c)
  {
    caseElements.put(r, c);
  }

  protected CasePaneElement getElementFor(Relation r)
  {
    return (CasePaneElement) caseElements.get(r);
  }

  protected void removeElementFor(Relation r)
  {
    if(caseElements.containsKey(r))
    {
//      ((CasePaneElement) caseElements.get(r)).setActive(false);
      caseElements.remove(r);
    }
  }

  protected synchronized void buildFramePane()
  {
//    System.out.println("buildFramePane (case)");
    removeAll();


    if(getCase() == null)
    {
      buildEmptyPane();
      return;
    }

    caseTableModel.formatTable(caseTable, getCase().getKnowledgeModel());


    JPanel mainPanel = buildHeaderPane();

    if(getCaseStatusDisplay() != SOLVED_STATUS_HIDDEN)
    {
      JPanel statusPanel = new JPanel();

      statusPanel.setLayout(new BorderLayout());
      statusPanel.add(new JLabel("Case status:"), BorderLayout.NORTH);
      statusPanel.setBorder(BorderFactory.createEmptyBorder(0,0,10,0));
      if(getCaseStatusDisplay() == SOLVED_STATUS_VISIBLE || !isEditable())
        caseStatus = new JLabel(getCase().getStatusString());
      else
      {
        caseStatus = new JComboBox(new String[] {getCase().getKnowledgeModel().getEntity(CBRModel.UNSOLVED_CASE).getName(),
                                   getCase().getKnowledgeModel().getEntity(CBRModel.PROCESSED_CASE).getName(),
                                   getCase().getKnowledgeModel().getEntity(CBRModel.SOLVED_CASE).getName()
             });

        ((JComboBox) caseStatus).setSelectedItem(getCase().getStatusString());
        ((JComboBox) caseStatus).addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e)
          {
            if(!((JComboBox) caseStatus).getSelectedItem().equals(getCase().getStatusString()))
            {
              getCase().setStatus(((JComboBox) caseStatus).getSelectedIndex());
              initTable();
              updatePane();
            }
          }
        });
      }
      statusPanel.add(caseStatus, BorderLayout.SOUTH);
      mainPanel.add(statusPanel);
    }

    if(showCaseType)
    {
      try
      {
        Relation instanceOf = getCase().getRelation("instance of");
        RelationType[] rcds = new RelationType[] { getCase().getKnowledgeModel().getRelationType("has subclass") };
        JPanel solutionPanel = new JPanel();
        solutionPanel.setLayout(new BorderLayout());
        solutionPanel.add(new JLabel("Case Type:"));
        if(isEditable())
          solutionPanel.add(EntityComponentFactory.getSymbolComboBox(instanceOf, EntityComponentFactory.getSubEntities(getCase().getKnowledgeModel().getEntity("Case"), rcds, true)), BorderLayout.SOUTH);
        else
          solutionPanel.add(new JLabel(instanceOf.getValue().toString()), BorderLayout.SOUTH);
        solutionPanel.setBorder(BorderFactory.createEmptyBorder(0,0,10,0));
        mainPanel.add(solutionPanel);
      }
      catch(NoSuchRelationException e) { }
      catch(NoSuchRelationTypeException e) { }
    }

    if(getCase().getStatus() != Case.UNSOLVEDCASE)
    {
      Relation[] solutions = getCase().getRelations("has solution");

      if(solutions.length > 0)
      {
        JPanel solutionPanel = new JPanel();
        solutionPanel.setLayout(new BorderLayout());
        solutionPanel.add(new JLabel("Solution:"));
        solutionPanel.add(EntityComponentFactory.getEntityComponent(solutions[0], isEditable()), BorderLayout.SOUTH);
        solutionPanel.setBorder(BorderFactory.createEmptyBorder(0,0,10,0));
        mainPanel.add(solutionPanel);
      }
      else
      {
        if(isEditable())
        {
          JPanel solutionPanel = new JPanel();
          solutionPanel.setLayout(new BorderLayout());

          Entity valueType = null;
          try
          {
             valueType = myEntity.getKnowledgeModel().getEntity("has solution").getRelation(new String[] {"has value", "has subclass", "has instance"}).getValue();
          }
          catch(NoSuchRelationException e)
          {
          }
          catch(NoSuchRelationTypeException e)
          {
          }

          if(valueType == null)
            valueType = myEntity.getKnowledgeModel().getEntity("Entity") ;

          Entity ents[] = EntityComponentFactory.getSubEntities(valueType, new RelationType[] {
              myEntity.getKnowledgeModel().getRelationType("has instance"),
              myEntity.getKnowledgeModel().getRelationType("has subclass"),
              myEntity.getKnowledgeModel().getRelationType("has value") }, false);

          Arrays.sort(ents, new EntityNameComparator());

          Object[] choices = new Object[ents.length+1];
          choices[0] = "<Choose value>";
          System.arraycopy(ents,0,choices,1,ents.length);

          addSolution = new JComboBox(choices);
          addSolution.setSelectedIndex(0);
          addSolution.setPreferredSize(new Dimension(120, getPreferredSize().height));
          addSolution.setMinimumSize(new Dimension(120, getPreferredSize().height));
//      addEntity.setFont(entityName.getFont());
          ActionListener solutionActiveListener = new ActionListener()
          {
            public void actionPerformed(ActionEvent e)
            {
              try
              {
                if(addSolution.getSelectedIndex() != 0)
                  addSolutionButton.setEnabled(true);
                else
                  addSolutionButton.setEnabled(false);
              }
              catch(NumberFormatException ex)
              {
                addSolutionButton.setEnabled(false);
              }
            }
          };

          addSolution.addActionListener(solutionActiveListener);


          addSolutionButton = new JButton("Set Solution");
          addSolutionButton.setEnabled(false);
          addSolutionButton.addActionListener(new ActionListener()
          {
            public void actionPerformed(ActionEvent e)
            {
              try
              {
                myEntity.addRelation("has solution", (Entity) addSolution.getSelectedItem());
              }
              catch(ClassCastException ex)
              {
                // This should not happen - the button is only active when the relation may be added,
                // but just to be sure.
                ex.printStackTrace();
              }
              catch(NoSuchRelationTypeException ex)
              {
                ex.printStackTrace();
              }
            }
          });

          solutionPanel.add(addSolutionButton, BorderLayout.EAST);
          solutionPanel.add(addSolution, BorderLayout.CENTER);
          solutionPanel.setBorder(BorderFactory.createEmptyBorder(0,0,10,0));
          mainPanel.add(solutionPanel);
        }
      }
    }

    Relation[] findings = getCase().getRelations("has finding");

    caseTableModel.setRelations(findings);
    mainPanel.add(caseTable.getTableHeader());
    mainPanel.add(caseTable);

    if(isEditable())
    {
      addPanel = new JPanel();
      addPanel.setLayout(new BoxLayout(addPanel, BoxLayout.X_AXIS));

      Entity[] rcds = myEntity.getKnowledgeModel().getRelationType("has finding").getAllSubclasses();
      Arrays.sort(rcds, new EntityNameComparator());
      Object[] choices = new Object[rcds.length+1];
      choices[0] = "<Choose relation>";
      System.arraycopy(rcds,0,choices,1,rcds.length);

      int width = caseTable.getPreferredSize().width;

      addRelationType = new JComboBox(choices);
      addRelationType.setSelectedIndex(0);
      addRelationType.setPreferredSize(new Dimension(140, getPreferredSize().height));
      addRelationType.setMinimumSize(new Dimension(140, getPreferredSize().height));
      addRelationType.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          if(addRelationType.getSelectedIndex() != 0)
          {
            try
            {
              RelationType rcd = (RelationType)new RelationType( (Entity)
                  addRelationType.getSelectedItem());

              if (addEntity.getSelectedIndex() != 0)
                addButton.setEnabled(true);
              else
                addButton.setEnabled(false);

              Entity valueEntity = null;
              try {
                valueEntity = rcd.getRelation(new String[] {"has value",
                                              "has instance", "has subclass"}).
                    getValue();
              }
              catch (NoSuchRelationException ex) {}
              catch (NoSuchRelationTypeException ex) {}
              if (valueEntity == null)
                valueEntity = rcd.getKnowledgeModel().getEntity("Entity");

              Entity[] possibleValues = EntityComponentFactory.getSubEntities(
                  valueEntity,
                  new RelationType[] {valueEntity.getKnowledgeModel().
                  getRelationType("has value"),
                  valueEntity.getKnowledgeModel().getRelationType(
                  "has instance"),
                  valueEntity.getKnowledgeModel().getRelationType(
                  "has subclass")}
                  , false);

              Arrays.sort(possibleValues, new EntityNameComparator());
              Object[] choices = new Object[possibleValues.length + 1];
              choices[0] = "<Choose value>";
              System.arraycopy(possibleValues, 0, choices, 1,
                               possibleValues.length);
              addEntity.setModel(new DefaultComboBoxModel(choices));
            }
            catch(EntityTypeNotMatchedException ex){ }
          }
          else
            addButton.setEnabled(false);
        }
      });

      Entity[] ents = myEntity.getKnowledgeModel().getEntities();
      Arrays.sort(ents, new EntityNameComparator());

      choices = new Object[ents.length+1];
      choices[0] = "<Choose value>";
      System.arraycopy(ents,0,choices,1,ents.length);

      addEntity = new JComboBox(choices);
      addEntity.setSelectedIndex(0);
      addEntity.setPreferredSize(new Dimension(120, getPreferredSize().height));
      addEntity.setMinimumSize(new Dimension(120, getPreferredSize().height));
//      addEntity.setFont(entityName.getFont());
      ActionListener buttonActiveListener = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          try
          {
            if(addEntity.getSelectedIndex() != 0 && addRelationType.getSelectedIndex() != 0)
              addButton.setEnabled(true);
            else
              addButton.setEnabled(false);
          }
          catch(NumberFormatException ex)
          {
            addButton.setEnabled(false);
          }
        }
      };

      addEntity.addActionListener(buttonActiveListener);


      addButton = new JButton("Add");
      addButton.setEnabled(false);
      addButton.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          try
          {
            RelationType rcd = (RelationType) new RelationType((Entity) addRelationType.getSelectedItem());
            Relation rel = rcd.newInstance(myEntity, (Entity) addEntity.getSelectedItem());
          }
          catch(EntityTypeNotMatchedException ex)
          {
            ex.printStackTrace();
          }
          catch(ClassCastException ex)
          {
            // This should not happen - the button is only active when the relation may be added,
            // but just to be sure.
            ex.printStackTrace();
          }
          catch(NumberFormatException ex)
          {
            // This should not happen - the button is only active when the relation may be added,
            // but just to be sure.
            ex.printStackTrace();
          }

        }
      });

      addPanel.add(addRelationType);
      addPanel.add(addEntity);
      addPanel.add(addButton);
      addPanel.setBorder(BorderFactory.createEmptyBorder(10,0,0,0));
      mainPanel.add(addPanel);
    }



/*    caseElements = new Hashtable();

    JPanel solutionPanel = new JPanel(new BorderLayout());
    Relation[] solutions = getCase().getRelations("has solution");

    JPanel secondaryLabelPanel = new JPanel();
    secondaryLabelPanel.setLayout(new GridLayout(0,2));
    secondaryLabelPanel.setBackground(getBackground());
    secondaryLabelPanel.add(new JLabel("Importance"));
    secondaryLabelPanel.add(new JLabel("Predictive Strength"));


    if(solutions.length > 0)
    {
      JLabel solutionLabel;
      if(solutions.length == 1)
        solutionLabel = new JLabel("Solution");
      else
        solutionLabel = new JLabel("Solutions");

      mainPanel.add(solutionLabel);

      solutionPanel.setLayout(new GridLayout(0, 3));
/*    solutionPanel.add(new JLabel("Relation"));
      solutionPanel.add(new JLabel("Value"));

      solutionPanel.add(secondaryLabelPanel);
      solutionPanel.setBackground(Color.white);

      for(int i=0;i<solutions.length;i++)
      {
        CasePaneElement cpe = new CasePaneElement(solutions[i]);
//      cpe.setActive(isActive());
        cpe.setEditable(isEditable());
//        cpe.setBorder(BorderFactory.createEmptyBorder(5,2,5,2));
        addElementFor(solutions[i], cpe);
        cpe.buildCasePaneElement(solutionPanel);
      }
      solutionPanel.setBorder(BorderFactory.createLineBorder(Color.black, 1));

      solutionPanel.setPreferredSize(new Dimension(350, solutionPanel.getPreferredSize().height));

      mainPanel.add(solutionPanel);
    }

    JPanel findingsPanel = new JPanel(new BorderLayout());
    Relation[] findings = getCase().getRelations("has finding");

/*    secondaryLabelPanel = new JPanel();
    secondaryLabelPanel.setBackground(getBackground());
    secondaryLabelPanel.setLayout(new GridLayout(0,2));
    secondaryLabelPanel.add(new JLabel("Importance"));
    secondaryLabelPanel.add(new JLabel("Predictive Strength"));

    if(findings.length > 0)
    {
      JLabel findingsLabel;
      if(findings.length == 1)
        findingsLabel = new JLabel("Finding");
      else
        findingsLabel = new JLabel("Findings");

      mainPanel.add(findingsLabel);

      findingsPanel.setLayout(new GridLayout(0, 3));
/*      findingsPanel.add(new JLabel("Relation"));
      findingsPanel.add(new JLabel("Value"));
      findingsPanel.add(secondaryLabelPanel);
      findingsPanel.setBackground(Color.white);

      for(int i=0;i<findings.length;i++)
      {
        CasePaneElement cpe = new CasePaneElement(findings[i]);
//      cpe.setActive(isActive());
        cpe.setEditable(isEditable());
//        cpe.setBorder(BorderFactory.createEmptyBorder(5,2,5,2));
        addElementFor(findings[i], cpe);
        cpe.buildCasePaneElement(findingsPanel);
//        findingsPanel.add(cpe);
      }
      findingsPanel.setBorder(BorderFactory.createLineBorder(Color.black, 1));

      findingsPanel.setPreferredSize(new Dimension(350, findingsPanel.getPreferredSize().height));

      mainPanel.add(findingsPanel);
    }
*/
    mainPanel.setBorder(BorderFactory.createEmptyBorder());
    setLayout(new FlowLayout(FlowLayout.LEFT));
    add(mainPanel);
    isUpdatingGUI = false;
  }
/*
  public static void main(String argv[])
  {
    KnowledgeModel km = examples.FruitModel.generateModel();
    try
    {
      Case case12 = new Case(km.getEntity("Case#12"));
      JFrame frame = new JFrame("New Case");
      CasePane cp = new CasePane(case12);
      cp.setEditable(true);
      cp.attach();
      cp.addLinkListener(openNewWindow);
      frame.setContentPane(cp);
      frame.setSize(cp.getPreferredSize().width+12, cp.getPreferredSize().height+50);
//    Case case12 = cp.getCase();


      JFrame frame2 = new JFrame(case12.getName());
      FramePane fp = new FramePane(case12);
      fp.setEditable(true);
      fp.attach();
      frame2.setContentPane(new JScrollPane(fp));
      frame2.setSize(fp.getPreferredSize().width+12, cp.getPreferredSize().height+50);

      fp.addLinkListener(openNewWindow);

    CaseComparison comp = new CaseComparison(km.getEntity("Case#6"), km.getEntity("Case#9"));
    comp.activate();
    comp.explain();

    System.out.println(comp);

      JFrame frame3 = new JFrame("Matching case#12");
      RetrieveResultPane rr = new RetrieveResultPane(case12);
      rr.attach();
      frame3.setContentPane(rr);
      frame3.setSize(600, 600);

      frame2.show();
      frame.show();
      frame3.show();
    }
    catch(jcreek.util.CreekException e)
    {
      e.printStackTrace();
    }
  } */

/*  class FindingValueComboBoxModel extends DefaultComboBoxModel
  {
    FindingValueComboBoxModel()
    {
      super();
    }


  } */
}



