package jcreek.gui.representation;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.reasoning.NumberComparison;
import jcreek.reasoning.SymbolComparison;
import jcreek.representation.NumberEntity;
import jcreek.util.EntityNameComparator;
import javax.swing.*;
import java.awt.event.*;
import java.awt.Color;
import java.util.Vector;
import java.util.Arrays;

/**
 * Class that produces Swing component to edit different types
 * of Entities in relations. </p>
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class EntityComponentFactory
{
  /**
   * This is a factory class that will just produce objects - it should not
   * be instanciated itself.
   */
  private EntityComponentFactory()
  {
  }

  /**
   * Returns a JComponent that is made to display or edit the value component of the
   * supplied relation.
   */
  public static JComponent getEntityComponent(Relation rel, boolean editable)
  {
    Entity valueEntity = rel.getValue();
    if(valueEntity == null)
      return new JLabel("<No Value>");

    Entity valueType = null;
    try
    {
      // Try to find a has-value relation from the relation type.
      try
      {
        Relation valueTypeRelation = rel.getRelationType().getRelation(new String[] {"has value", "has instance", "has subclass"});
        valueType = valueTypeRelation.getValue();
      }
      catch(NoSuchRelationException e)
      {
        valueType = rel.getKnowledgeModel().getEntity("Entity");
      }

      Entity comparator = valueType.getRelation("has comparator").getValue();

      if(comparator.equals(CBRModel.NUMBER_COMPARISON))
      {
        Object valueObject = valueEntity.getEntityObject();
        if(valueObject == null)
        {
          valueEntity = new DoubleEntity(rel.getKnowledgeModel(), 0.0, valueType);
          valueObject = valueEntity.getEntityObject();
        }
        JTextField doubleField = new JTextField(valueObject.toString());
        if(editable)
        {
          doubleField.addFocusListener(new DoubleFocusListener(rel, valueEntity, doubleField));
        }
        doubleField.setEditable(editable);
        return doubleField;
      }
    }
    catch(NoSuchRelationException e)
    {
    }
    catch(NoSuchRelationTypeException e)
    {
    }

//      if(comparator.equals(valueEntity.getKnowledgeModel().getClassDescriptor(SymbolComparison.class)))
    if(editable)
    {
      if(valueType == null)
        valueType = rel.getKnowledgeModel().getEntity("Entity");
      Entity[] ents = getSubEntities(valueType, new RelationType[] { valueType.getKnowledgeModel().getRelationType("has value"),
                                     valueType.getKnowledgeModel().getRelationType("has instance"),
                                     valueType.getKnowledgeModel().getRelationType("has subclass") }, false );

      return getSymbolComboBox(rel, ents);
    }


    return new JLabel(valueEntity.toString());
  }

  public static JComboBox getSymbolComboBox(Relation rel, Entity[] values)
  {
    Arrays.sort(values, new EntityNameComparator());
    JComboBox box = new JComboBox(new EntityBoxModel(rel, values));
    box.setSelectedItem(rel.getValue());
//    System.out.println("Setting selected to "+rel.getValue()+" for "+rel);
    box.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {
        JComboBox myBox = ((JComboBox) e.getSource());
        ((EntityBoxModel) myBox.getModel()).rel.setValue((Entity) myBox.getSelectedItem());
      }
    });
    return box;
  }

  public static Entity[] getSubEntities(Entity ent, RelationType[] relationClassDescriptors, boolean includeRelationValue)
  {
//    System.out.println("Finding for "+ent);
    Vector result = new Vector();
    Vector queue = new Vector();


    queue.add(ent);
    if(includeRelationValue)
      result.add(ent);
    while(!queue.isEmpty())
    {
      Entity next = (Entity) queue.get(0);
      queue.remove(0);
      if(next != null)
      {
//      System.out.println("Next is "+next);
        for (int i = 0; i < relationClassDescriptors.length; i++)
        {
          Relation[] rels = next.getLocalRelations(relationClassDescriptors[i]);
          for (int j = 0; j < rels.length; j++)
          {
            if (!result.contains(rels[j].getValue()))
            {
              result.add(rels[j].getValue());
              queue.add(rels[j].getValue());
            }
          }
        }
      }
    }
    return (Entity[]) result.toArray(new Entity[0]);
  }

  static class EntityBoxModel extends DefaultComboBoxModel
  {
    Relation rel;
    Entity[] values;
    EntityBoxModel(Relation rel, Entity[] values)
    {
      super(values);
      this.rel = rel;
      this.values = values;
    }


  }

  static class DoubleFocusListener implements FocusListener
  {
    Entity ent;
    JTextField field;
    Relation rel;
    DoubleFocusListener(Relation rel, Entity ent, JTextField field)
    {
      this.ent = ent;
      this.field = field;
      this.rel = rel;
    }

    public void focusGained(FocusEvent e)
    {
    }

    public void focusLost(FocusEvent e)
    {
      try
      {
        Double val = new Double(field.getText());
        ent.setEntityObject(val);
        if(rel.getValue() != ent)
          rel.setValue(ent);
      }
      catch(NumberFormatException ex)
      {
        field.setBorder(BorderFactory.createLineBorder(Color.red,1));
      }
    }
  }
}
