package jcreek.gui.representation;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.util.EntityNameComparator;
import javax.swing.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.*;
import java.io.Serializable;
import java.util.Arrays;

/**
 * GUI element for creating or manipulating a Creek entity. An
 * instance of this class encapsulates an Entiy (as defined in jcreek.representation)
 * and allows the manipulation of it.
 *
 * @author Frode Srmo
 * @see jcreek.representation.Entity
 * @version 1.0
 */

public class FramePane extends JComponent implements ModelChangeListener
{
  protected Entity myEntity;
  private boolean editable = true;
  private int solvedStatus;
  private boolean active = false;
  protected boolean isUpdatingGUI = false;

  public JTextField entityName;
  protected JTextArea entityDescription;
  protected JPanel elementPanel;
  protected JPanel addPanel;
  protected JButton addButton;
  protected JComboBox addRelationType;
  protected JComboBox addEntity;
  protected JComboBox entityObjectClassSelector;
  protected JTextField addStrength;

  protected FrameTableModel frameTableModel;
  protected JTable frameTable;

  protected boolean displayInheritedToggle = true;
  protected boolean displayInherited = true;

  /**
   * Creates and empty FramePane.
   */
  public FramePane()
  {
    super();
    initTable();
  }

  /**
   * Creates a new FramePane encapsulating a new entity.
   *
   * @param km The KnowledgeModel to create the new entity in.
   */
  public FramePane(KnowledgeModel km)
  {
    super();
    int news = 0;
    String name = "New Entity";
    while (km.hasEntity(name) && km.hasEntity(name + " (" + news + ")"))
      news++;

    initTable();
    try {
      Entity newEntity = new Entity(km, name, name);
      setEntity(newEntity);
    }
    catch (NameAlreadyExistException e) {
    }

  }

  /**
   * Creates a new FramePane encapsulating the given entity.
   *
   * @param frame The entity to show.
   * @param linker If true, adds the default LinkListener that will make and
   *   display a new FramePane in a new JFrame if a link is clicked.
   */
  public FramePane(Entity frame, boolean linker)
  {
    this(frame, linker, false, true);
  }

  /**
   * Creates a new FramePane encapsulating the given entity.
   *
   * @param frame The entity to show.
   */
  public FramePane(Entity frame)
  {
    this(frame, false, false, true);
  }

  /**
   * Creates a new FramePane enacapsulating the given entity
   *
   * @param frame The entity to display.
   * @param linker
   * @param displayInheritedToggle If true, displays a toggle that will allow
   *        the user to change if inherited relations are displayed or not.
   * @param displayInherited If true, displays the inherited relations.
   */
  public FramePane(Entity frame, boolean linker, boolean displayInheritedToggle,
                   boolean displayInherited)
  {
    this.displayInheritedToggle = displayInheritedToggle;
    this.displayInherited = displayInherited;
    initTable();
    setEntity(frame);
    if (linker)
      addLinkListener(getOpenNewWindowLinkListener());
  }

  /**
   * Returns true if this gui element displays the toggle to change between
   * displaying inherited relations or only local relations.
   */
  public boolean showInheritedToggle()
  {
    return displayInheritedToggle;
  }

  /**
   * Returns true if this gui element shows inherited relations in addition to
   * the local ones.
   */
  public boolean showInheritedRelations()
  {
    return displayInherited;
  }

  /**
   * Sets if this gui element is supposed to display only local relations or
   * all relations.
   *
   * @param showInheritedRelations If true, all relations (including inherited
   *        relations) are displayed. If not true, only local relations are shown.
   */
  public void setShowInheritedRelations(boolean showInheritedRelations)
  {
    if (displayInherited != showInheritedRelations) {
      displayInherited = showInheritedRelations;
      updatePane();
    }
  }

  protected void initTable()
  {
    frameTable = new JTable();
    frameTableModel = new FrameTableModel(new Relation[0]);
    frameTable.setModel(frameTableModel);
    frameTableModel.setTable(frameTable);
    frameTable.getTableHeader().setReorderingAllowed(false);
    frameTableModel.setEditable(true);
  }

  /**
   * When called, this method will register ModelChangeListeners with all the
   * relations in the table to update it automatically if a change is registered
   * in any of them.
   *
   * Before the table is disposed, the detach() method should be called to
   * make sure the listeners are removed.
   */
  public synchronized void attach()
  {
    if (!active) {
      active = true;
      if (myEntity != null) {
//        myEntity.addModelChangeListener(this);
        myEntity.getKnowledgeModel().addModelChangeListener(this);

        /*        if(frameTableModel != null)
                {
                  Relation[] rels = frameTableModel.getRelations();
                  for(int i=0;i<rels.length;i++)
                    rels[i].removeModelChangeListener(this);
                } */
      }
    }
  }

  /**
   * Removes the change-listeners.
   */
  public synchronized void detach()
  {
    if (active) {
      active = false;
      if (myEntity != null) {
        myEntity.getKnowledgeModel().removeModelChangeListener(this);

        /*         if(frameTableModel != null)
                 {
                   Relation[] rels = frameTableModel.getRelations();
                   for(int i=0;i<rels.length;i++)
                     rels[i].removeModelChangeListener(this);
                 }*/
      }
    }
  }

  /**
   * Returns true if the FramePanel updates when other elements changes
   * the case or other relevant parts of the model, false if not.
   */
  public boolean isAttached()
  {
    return active;
  }

  public void modelChanged(ModelChangeEvent e)
  {
    entityObjectClassSelector = null;
    updatePane();
  }

  /**
   * Registers a request to update its contents. It will return at once and not
   * wait for this to happen. This is usually in response to a change in the
   * data model. Although this method affects the Swing component, it can be
   * called outside the Swing event thread.
   *
   * Repeated calls to this method before the pane is actually updated will not
   * result in the pane redrawing multiple times - it functions much as the
   * repaint() method.
   */
  public synchronized void updatePane()
  {
//    System.out.println("UpdatePane in "+getClass());
    if (!isUpdatingGUI) {
      SwingUtilities.invokeLater(new Runnable()
      {
        public void run()
        {
          buildFramePane();
          validate();
          repaint();
        }
      });
      isUpdatingGUI = true;
    }
  }

  /**
   * Sets if the case displayed by this FramePane should be editable or through
   * the GUI or not.
   *
   * @param editable True if the case should be editable, false if not.
   */
  public void setEditable(boolean editable)
  {
    this.editable = editable;

    if (frameTableModel != null)
      frameTableModel.setEditable(editable);

    updatePane();
  }

  /**
   * Returns true if the case encapsulated by this FramePane can be modified
   * through the GUI, or fasle if it is merely displaed.
   */
  public boolean isEditable()
  {
    return editable;
  }

  /**
   * Sets the entity to be displayed and modified by this FramePane.
   *
   * @param entity The case object.
   */
  public void setEntity(Entity entity)
  {
    boolean a = isAttached();
    if (a)
      detach();
    myEntity = entity;
    entityObjectClassSelector = null;
    buildFramePane();
    if (a)
      attach();
  }

  /**
   * Returns the entity this FramePane displays.
   */
  public Entity getEntity()
  {
    return myEntity;
  }

  private void setTableRelations(Relation[] rels)
  {
    if (isAttached()) {
      Relation[] old = frameTableModel.getRelations();
    }
    frameTableModel.setRelations(rels);
  }

  /**
   * Builds a pane that is displayed when no entity is selected.
   */
  protected void buildEmptyPane()
  {
    setLayout(new BorderLayout());
    add(new JLabel("<No Entity Selected>", JLabel.CENTER), BorderLayout.CENTER);
  }

  /**
   * Build the entity object part of the frame pane, containing information
   * about what object is stored within the frame.
   *
   * @return JPanel
   */
  protected JPanel buildEntityObjectPane()
  {
//    System.out.println("Rebuilding for " + getEntity() + "object is: " +
//                       getEntity().getEntityObject());

    if (!isEditable() && getEntity().getEntityObject() == null)
      return null;

    JPanel entityObjectPanel = new JPanel(new BorderLayout());
//    entityObjectPanel.setBorder(BorderFactory.createTitledBorder("Entity Object"));

    JPanel classListPanel = new JPanel(new BorderLayout());
    classListPanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));
    classListPanel.add(new JLabel("Entity Object Class"), BorderLayout.NORTH);

    JPanel valuePanel = new JPanel(new BorderLayout());
    valuePanel.add(new JLabel("Entity Object Value"), BorderLayout.NORTH);
    valuePanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));

    if (isEditable())
    {
      if (entityObjectClassSelector == null)
      {
        Class[] supported = EntityTypeFieldFactory.getClassesSupported();

        Object[] list;
        Object noEntityObject = "<None>";
        Object selected;
        if (getEntity().getEntityObject() == null) {
          selected = noEntityObject;
          list = new Object[supported.length + 1];
          list[0] = noEntityObject;
        }
        else
        {
          Class entityObjectClass = getEntity().getEntityObject().getClass();
          if (java.util.Arrays.asList(supported).contains(entityObjectClass)) {
            list = new Object[supported.length + 1];
            list[0] = noEntityObject;
            selected = entityObjectClass;
          }
          else {
            list = new Object[supported.length + 2];
            list[0] = noEntityObject;
            list[1] = getEntity().getEntityObject().getClass().getName() +
                " (no edit method)";
            selected = list[1];
          }
        }
        System.arraycopy(supported, 0, list, list.length - supported.length,
                         supported.length);

        entityObjectClassSelector = new JComboBox(list);

        entityObjectClassSelector.setSelectedItem(selected);
        entityObjectClassSelector.addActionListener(new ActionListener()
        {
          public void actionPerformed(ActionEvent e)
          {
            updatePane();
          }
        });
      }
      classListPanel.add(entityObjectClassSelector, BorderLayout.CENTER);
    }
    else
      classListPanel.add(new JLabel(getEntity().getEntityObject().getClass().
                                    getName()), BorderLayout.CENTER);

    entityObjectPanel.add(classListPanel, BorderLayout.NORTH);



    if (isEditable())
    {
      if(
        entityObjectClassSelector != null &&
        entityObjectClassSelector.getSelectedItem() instanceof Class)
      {
        Class type = (Class) entityObjectClassSelector.getSelectedItem();
        final EntityTypeField comp = EntityTypeFieldFactory.getEntityTypeField(
            type);

        comp.addCellEditorListener(new CellEditorListener()
        {
          public void editingStopped(ChangeEvent e)
          {
            Object value = comp.getCellEditorValue();
            if (value != null && value instanceof Serializable)
              getEntity().setEntityObject( (Serializable) value);
          }

          public void editingCanceled(ChangeEvent e)
          {
            // Don't update the contents.
          }
        });
        valuePanel.add(comp.getEditorComponent(getEntity().getEntityObject()),
                       BorderLayout.CENTER);
        entityObjectPanel.add(valuePanel, BorderLayout.CENTER);
      }
    }
    else
    {
      EntityTypeField comp = EntityTypeFieldFactory.getEntityTypeField(getEntity().getEntityObject().getClass());
      if(comp != null)
      {
        valuePanel.add(comp.getRendererComponent(getEntity().getEntityObject()),
                       BorderLayout.CENTER);
      }
      else
        valuePanel.add(new JLabel(getEntity().getEntityObject().toString()), BorderLayout.CENTER);

      entityObjectPanel.add(valuePanel, BorderLayout.CENTER);
    }

    return entityObjectPanel;
  }

  /**
   * Build the header pane with the name and description of the frame.
   */
  protected JPanel buildHeaderPane()
  {
    JPanel mainPanel = new JPanel();

    mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));

    JPanel namePanel = new JPanel();
    namePanel.setLayout(new BorderLayout());
    namePanel.add(new JLabel("Name:"), BorderLayout.NORTH);

    entityName = new JTextField(myEntity.getName(), 35);

    if (isEditable()) {
      entityName.addFocusListener(new FocusListener()
      {
        public void focusLost(FocusEvent e)
        {
          if (myEntity != null && !myEntity.getName().equals(entityName.getText())) {
            try {
              myEntity.setName(entityName.getText());
            }
            catch (NameAlreadyExistException ex) {
              JOptionPane.showMessageDialog(entityName,
                                            "A concept with the name '" +
                                            entityName.getText() +
                  "' already exists in this knowledge model.",
                                            "Name change error",
                                            JOptionPane.ERROR_MESSAGE);
              entityName.setText(myEntity.getName());
            }
          }
        }

        public void focusGained(FocusEvent e)
        {}
      });
      entityName.setEditable(true);
    }
    else
      entityName.setEditable(false);

    namePanel.add(entityName, BorderLayout.CENTER);

    mainPanel.add(namePanel);

    /*    if(myEntity != null && myEntity.getEntityObject() != null)
        {

          JPanel valuePanel = new JPanel();
          valuePanel.setLayout(new BorderLayout());
          valuePanel.add(new JLabel("Value:"), BorderLayout.NORTH);

          JComponent entityValue = ValueComponentFactory.getValueComponent(myEntity, isEditable());

          valuePanel.add(entityValue, BorderLayout.CENTER);
          valuePanel.setBorder(BorderFactory.createEmptyBorder(10,0,0,0));
          mainPanel.add(valuePanel);
     }*/

    JPanel objectPanel = buildEntityObjectPane();
    if (objectPanel != null)
      mainPanel.add(objectPanel);
      //   }

    JPanel descPanel = new JPanel();
    descPanel.setLayout(new BorderLayout());
    descPanel.add(new JLabel("Description:"), BorderLayout.NORTH);
    descPanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 10, 0));
    entityDescription = new JTextArea(myEntity.getDescription(), 6, 35);
    entityDescription.setLineWrap(true);
    entityDescription.setWrapStyleWord(true);
    entityDescription.setBorder(entityName.getBorder());
    entityDescription.setBackground(entityName.getBackground());
    entityDescription.setEditable(isEditable());
    if (isEditable()) {
      entityDescription.addFocusListener(new FocusListener()
      {
        public void focusLost(FocusEvent e)
        {
          if (myEntity != null &&
              !myEntity.getDescription().equals(entityDescription.getText())) {
            myEntity.setDescription(entityDescription.getText());
          }
        }

        public void focusGained(FocusEvent e)
        {}
      });
    }
    descPanel.add(entityDescription, BorderLayout.CENTER);

    if (showInheritedToggle()) {
      JPanel togglePanel = new JPanel();
      JRadioButton showInherited = new JRadioButton("All Relations",
          showInheritedRelations());
      JRadioButton showLocal = new JRadioButton("Local Only",
                                                !showInheritedRelations());
      ButtonGroup group = new ButtonGroup();
      group.add(showInherited);
      group.add(showLocal);
      ActionListener listen = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          if (e.getActionCommand().equals("All Relations"))
            setShowInheritedRelations(true);
          else
            setShowInheritedRelations(false);
        }
      };

      showInherited.addActionListener(listen);
      showLocal.addActionListener(listen);
      togglePanel.add(showInherited);
      togglePanel.add(showLocal);
      descPanel.add(togglePanel, BorderLayout.SOUTH);
    }
    mainPanel.add(descPanel);

    return mainPanel;
  }

  /**
   * (Re)builds the pane.
   */
  protected synchronized void buildFramePane()
  {
    removeAll();

    if (myEntity == null) {
      buildEmptyPane();
      return;
    }

    frameTableModel.formatTable(frameTable, myEntity.getKnowledgeModel());

    JPanel mainPanel = buildHeaderPane();

    if (displayInherited)
      frameTableModel.setRelations(myEntity.getRelations());
    else
      frameTableModel.setRelations(myEntity.getLocalRelations());

    mainPanel.add(frameTable.getTableHeader());
    mainPanel.add(frameTable);

    if (isEditable()) {
      addPanel = new JPanel();
      addPanel.setLayout(new BoxLayout(addPanel, BoxLayout.X_AXIS));

      Entity[] rcds = myEntity.getKnowledgeModel().getRelationType("Relation").
          getAllSubclasses();
      Arrays.sort(rcds, new EntityNameComparator());

      Object[] choices = new Object[rcds.length + 1];
      choices[0] = "<Choose relation>";
      System.arraycopy(rcds, 0, choices, 1, rcds.length);

      int width = frameTable.getPreferredSize().width;

      addRelationType = new JComboBox(choices);
      addRelationType.setSelectedIndex(0);
      addRelationType.setPreferredSize(new Dimension(140,
          getPreferredSize().height));
      addRelationType.setMinimumSize(new Dimension(140,
          getPreferredSize().height));
      addRelationType.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          if (addRelationType.getSelectedIndex() != 0) {
            try {
              RelationType rcd = new RelationType( (Entity) addRelationType.
                                                  getSelectedItem());
              addStrength.setText(Double.toString(rcd.getExplanationStrength()));

              if (addEntity.getSelectedIndex() != 0)
                addButton.setEnabled(true);
              else
                addButton.setEnabled(false);
            }
            catch (EntityTypeNotMatchedException ex) {
              addButton.setEnabled(false);
            }
          }
          else
            addButton.setEnabled(false);
        }
      });

      Entity[] ents = myEntity.getKnowledgeModel().getEntities();
      Arrays.sort(ents, new EntityNameComparator());

      choices = new Object[ents.length + 1];
      choices[0] = "<Choose value>";
      System.arraycopy(ents, 0, choices, 1, ents.length);

      addEntity = new JComboBox(choices);
      addEntity.setSelectedIndex(0);
      addEntity.setPreferredSize(new Dimension(120, getPreferredSize().height));
      addEntity.setMinimumSize(new Dimension(120, getPreferredSize().height));
//      addEntity.setFont(entityName.getFont());
      ActionListener buttonActiveListener = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          try {
            Double strength = new Double(addStrength.getText());
            if (addEntity.getSelectedIndex() != 0 &&
                addRelationType.getSelectedIndex() != 0)
              addButton.setEnabled(true);
            else
              addButton.setEnabled(false);
          }
          catch (NumberFormatException ex) {
            addButton.setEnabled(false);
          }
        }
      };

      addEntity.addActionListener(buttonActiveListener);

      addStrength = new JTextField("<Strength>", 6);
      addStrength.setPreferredSize(new Dimension(60, getPreferredSize().height));
      addStrength.getDocument().addDocumentListener(new DocumentListener()
      {
        public void changedUpdate(DocumentEvent e)
        {
          check();
        }

        public void insertUpdate(DocumentEvent e)
        {
          check();
        }

        public void removeUpdate(DocumentEvent e)
        {
          check();
        }

        private void check()
        {
          try {
            Double strength = new Double(addStrength.getText());
            if (addEntity.getSelectedIndex() != 0 &&
                addRelationType.getSelectedIndex() != 0)
              addButton.setEnabled(true);
            else
              addButton.setEnabled(false);
          }
          catch (NumberFormatException ex) {
            addButton.setEnabled(false);
          }
        }
      });
      addStrength.addActionListener(buttonActiveListener);

      addButton = new JButton("Add");
      addButton.setEnabled(false);
      addButton.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          try {
            RelationType rcd = (RelationType)new RelationType( (Entity)
                addRelationType.getSelectedItem());
            Relation rel = rcd.newInstance(myEntity,
                                           (Entity) addEntity.getSelectedItem());
            double strength = new Double(addStrength.getText()).doubleValue();
            if (strength != rcd.getExplanationStrength())
              rel.setExplanationStrength(strength);
          }
          catch (EntityTypeNotMatchedException ex) {
            // Should not happen since only relation types are listed.
            // ex.printStackTrace();
          }
          catch (ClassCastException ex) {
            // This should not happen - the button is only active when the relation may be added,
            // but just to be sure.
            ex.printStackTrace();
          }
          catch (NumberFormatException ex) {
            // This should not happen - the button is only active when the relation may be added,
            // but just to be sure.
            ex.printStackTrace();
          }

        }
      });

      addPanel.add(addRelationType);
      addPanel.add(addEntity);
      addPanel.add(addStrength);
      addPanel.add(addButton);
      addPanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));
      mainPanel.add(addPanel);
    }

    mainPanel.setBorder(BorderFactory.createEmptyBorder());
    setLayout(new FlowLayout(FlowLayout.LEFT));
    add(mainPanel);

    isUpdatingGUI = false;
  }

  public void addLinkListener(LinkListener listener)
  {
    frameTableModel.addLinkListener(listener);
  }

  public void removeLinkListener(LinkListener listener)
  {
    frameTableModel.removeLinkListener(listener);
  }

  protected LinkListener openNewWindow = new LinkListener()
  {
    public void linkClicked(LinkEvent e)
    {
      Entity ent = (Entity) e.getConcept();
      JFrame frame = new JFrame(ent.getName());
      FramePane cp = new FramePane(ent);
      cp.setEditable(isEditable());
      cp.attach();
      cp.addLinkListener(openNewWindow);
      frame.setContentPane(new JScrollPane(cp));
      frame.setSize(cp.getPreferredSize().width + 12,
                    cp.getPreferredSize().height + 50);
      frame.show();
    }

    public void linkAction(LinkEvent e)
    {}
  };

  public LinkListener getOpenNewWindowLinkListener()
  {
    return openNewWindow;
  }

  public JTable getFrameTable()
  {
    return frameTable;
  }
/*
  public static void main(String argv[])
  {
    KnowledgeModel km = jcreek.examples.FruitModel.generateModel();
    try {
      Case case12 = (Case)new Case(km.getEntity("Case#12"));
      JFrame frame = new JFrame(case12.getName());
      FramePane cp = new FramePane(case12);
      cp.setEditable(true);
      cp.attach();
      frame.setContentPane(new JScrollPane(cp));
      frame.setSize(cp.getPreferredSize().width + 12,
                    cp.getPreferredSize().height + 50);

      cp.addLinkListener(cp.openNewWindow);

      frame.show();
    }
    catch (jcreek.util.CreekException e) {e.printStackTrace();
    }
  }
      }
     */ }



