package jcreek.reasoning;

import java.util.Vector;

import jcreek.representation.*; import jcreek.representation.cbr.*;
/**
 * <p>Title: CBRReasoningStep</p>
 * <p>Description: Abstract class used to implement a step in the Case-Based Reasoning
 * process. This class and the classes it implemetns are not thread safe.</p>
 * @author Frode Srmo
 * @version 1.0
 */

public abstract class CBRReasoningStep
{
  protected KnowledgeModel km;
  protected volatile int state;
  protected Vector updateListeners = new Vector();
  private volatile boolean aborted = false;

  protected volatile int counter;
  protected volatile int counterMax;

  public static final int INITIAL_STATE = 0;
  public static final int ACTIVATED_STATE = 1;
  public static final int EXPLAINED_STATE = 2;
  public static final int FOCUSED_STATE = 3;

  /**
   * Creates a new CBRReasoningStep using the model specified.
   */
  public CBRReasoningStep(KnowledgeModel model)
  {
    this.km = model;
  }

  /**
   * Resets the reasoning to the initial step.
   */
  public abstract void reset();

  /**
   * Updates the reasoning, e.g. when there has been a change in the case-base.
   */
  public void update()
  {
    int oldState = getState();
    fireReasoningUpdateStarted();
    reset();
    if(oldState%10 >= FOCUSED_STATE)
    {
      activate();
      explain();
      focus();
    }
    else if(oldState%10 >= EXPLAINED_STATE)
    {
      activate();
      explain();
    }
    else if(oldState%10 >= ACTIVATED_STATE)
    {
      activate();
    }
  }

  /**
   * Runs the "activate" step in this reasoning step. In general, this step use
   * an heuristic that is computationally inexpensive to provide a rough sort of
   * the elements of this reasoning step.
   *
   * The subclass implementing this subclass should have the contents synchronized
   * over the 'this' object.
   */
  public abstract void activate();

  /**
   * Runs the "explain" step in this reasoning step. Use a more computationally
   * expensive method to further compare the most interesting elements from the
   * activate() step.
   *
   * The subclass implementing this subclass should have the contents synchronized
   * over the 'this' object.
   */
  public abstract void explain();

  /**
   * Trims down the number of elements in the reasoning step, using information from
   * the activate() and explani() methods.
  *
  * The subclass implementing this subclass should have the contents synchronized
  * over the 'this' object.
   */
  public abstract void focus();

  protected void setState(int state)
  {
    if(this.state != state)
    {
      this.state = state;
      fireReasoningStateChanged();
    }
  }

  /**
   * Returns the state of the reasoning step. T
   */
  public int getState()
  {
    return state;
  }

  /**
   * Returns true if this reasoning step has been activated.
   */
  public boolean isActivated()
  {
    return state%10 >= ACTIVATED_STATE;
  }

  /**
   * Returns true if this reasoning step has been explained.
   */
  public boolean isExplained()
  {
    return state%10 >= EXPLAINED_STATE;
  }

  /**
   * Returns true if this reasoning step has been focused.
   */
  public boolean isFocused()
  {
    return state%10 >= FOCUSED_STATE;
  }

  /**
   * Returns the KnowledgeModel this reasoning is executed on.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    return km;
  }

  /**
   * The activate/explain/focis step is usually excecuted on N cases, and this
   * method returns how many iterations the current task is performing. Useful
   * for making progressbars etc.
   *
   * The value from this method can be used as the value for setMaximum() in
   * JProgressBar for example. Minimum is always 0.
   */
  public int getIterationMax()
  {
    return counterMax;
  }

  /**
   * The activate/explain/focis step is usually excecuted on N cases, and this
   * method returns how far the current task is progressed. Useful
   * for making progressbars etc.
   *
   * The value from this method can be used to set the value of setValue() of
   * JProgressBar for example.
   */
  public int getIteration()
  {
    return counter;
  }

  /**
   * If called, it aborts the current reasoning process and resets the status to
   * INITIAL_STATE. The call does not return until the reasoning method currently
   * active (if any) exists.
   */
  public void abort()
  {
    // This will signal the reasoning processes to abort.
    aborted = true;
    synchronized(this)
    {
      // When this semaphore is held, we know that the reasoning process has
      // aborted as they have to hold it while doing reasoning.
      reset();
      aborted = false;
    }

    // This is done outside the synchronized statement to avoid deadlocks in
    // executing unknown event calls.
    fireReasoningAborted();

  }

  /**
   * This method may be used to check at intervals during the reasoning process if
   * another thread has requested that the reasoning is aborted.
   */
  public boolean isAborted()
  {
    if(aborted)
      System.err.println("Aborted");
    return aborted;
  }

  /**
   * Adds a listener that will be notified when the reasoning status enters a new
   * status or when it is restarted because of an update to the underlaying model.
   *
   * @param listener The listener to add.
   */
  public void addReasoningUpdateListener(ReasoningUpdateListener listener)
  {
    updateListeners.add(listener);
  }

  /**
   * Removes a listener from the list.
   * @param listener The listener to remove.
   */
  public void removeReasoningUpdateListener(ReasoningUpdateListener listener)
  {
    updateListeners.remove(listener);
  }

  void fireReasoningAborted()
  {
    ReasoningUpdateEvent e = new ReasoningUpdateEvent(this, ReasoningUpdateEvent.REASONING_ABORTED);
    for(int i=0;i<updateListeners.size();i++)
      ((ReasoningUpdateListener) updateListeners.get(i)).reasoningAborted(e);
  }

  void fireReasoningUpdateStarted()
  {
    ReasoningUpdateEvent e = new ReasoningUpdateEvent(this, ReasoningUpdateEvent.REASONING_UPDATESTARTED);
    for(int i=0;i<updateListeners.size();i++)
      ((ReasoningUpdateListener) updateListeners.get(i)).reasoningUpdateStarted(e);
  }

  void fireReasoningStateChanged()
  {
    ReasoningUpdateEvent e = new ReasoningUpdateEvent(this, ReasoningUpdateEvent.REASONING_STATECHANGED);
    for(int i=0;i<updateListeners.size();i++)
      ((ReasoningUpdateListener) updateListeners.get(i)).reasoningStateChanged(e);
  }

}
