package jcreek.reasoning;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.representation.inference.*;
import jcreek.representation.cbr.*;
import jcreek.util.LRUCache;
import java.util.*;

/**
 * <p>A comparison between two cases.
 * <p>The comparison is comparable if the entities have one parent in common.
 * Two entities are compared by comparing the relations of the entities,
 * i.e. by creating relation comparisons of every possible combination,
 * (followed by activation and explanation) and computing the strength.
 *
 * @author Ellen Lippe, modified by Frode Srmo of NTNU 2001-2002
 * @version 1.1
*/
public class CaseComparison extends EntityComparison
{
   private static final int DEBUG_LEVEL = 0;

  /**
   * The list of all relation comparisons.
   * The list consist of matched and explained relation comparisons.
   */

   /** The list of all finding comparisons */
   protected Vector comparisons;

   /** Hashtable of finding-relation => comparisons involving that finding-relation. */
   protected Hashtable findingToComparisons;

   /** List of unmatched target relations. */
   protected Vector unmatchedTarget;

   /** List of unmatched target relations. */
   protected Vector unmatchedOrigin;

   /** The sum of all the relvance strengths of the target cases findings. */
   protected double relevanceStrengthSum;

//   private double relevanceStrengthSumPenalty = 0.0;

   /**
    * Constant that gives the maximum penalty for unmatched origin findings.
    */
   public static final double UNMATCHED_ORIGIN_PENALTY = 0.01;
//------------------------------------------------------------------------------

  /**
   * Creates a new comparison between the origin case and the target case
   * with default depth depth.
   * This consists of initializing the activation comparison list and the
   * explanation comparison list with the origin and target relations,
   * ready to be activated or explained.
   *
   * @param origin the case to compare from
   * @param target the case to compare to
   * @param findingList list of findings that should be compared.
   *
   * @see DEFAULT_DEPTH
   */
  public CaseComparison(Entity origin, Entity target)
  {
    super(origin, target);
//    this.origin = origin;
//    this.target = target;

  }//CaseComparison

//------------------------------------------------------------------------------

  public static double getRelevanceFactor(Relation finding)
  {
    return (finding.getExplanationStrength()+finding.getInverse().getExplanationStrength())/2;
  }

  /**
   * Resets the comparison.
   */
  public void reset()
  {
    if(DEBUG_LEVEL > 0)
      System.out.println("Reset in casecomparison between "+getOrigin()+" and "+getTarget());
    super.reset();
    comparisons = new Vector();
    findingToComparisons = new Hashtable();

//    long time = System.currentTimeMillis();
    unmatchedOrigin = new Vector(java.util.Arrays.asList(getOriginRelations()));
    unmatchedTarget = new Vector(java.util.Arrays.asList(getTargetRelations()));
    relevanceStrengthSum = 0.0;

    for(int j=0;j<unmatchedTarget.size();j++)
    {
      Relation targetRel = (Relation) unmatchedTarget.elementAt(j);
      relevanceStrengthSum += getRelevanceFactor(targetRel);

      for(int i=0;i<unmatchedOrigin.size();i++)
      {
        Relation originRel = (Relation) unmatchedOrigin.elementAt(i);
        if(DEBUG_LEVEL > 1)
          System.out.println("--> "+originRel+" == "+targetRel+"  ?");
        if(findingsComparable(originRel, targetRel))
        {
          if(DEBUG_LEVEL > 1)
            System.out.println("       comparable");
          CaseFindingComparison cfc = new CaseFindingComparison(originRel, targetRel);

          Vector v = null;
          if(findingToComparisons.containsKey(targetRel))
            v = (Vector) findingToComparisons.get(targetRel);
          else
            v = new Vector();

          v.addElement(cfc);
          findingToComparisons.put(targetRel, v);

          comparisons.add(cfc);
        }

      }
    }
//    System.out.println("reset done: RelevanceStrengthSum is "+relevanceStrengthSum);
/*
//    long time2 = System.currentTimeMillis();
    int originCount=0;
    int targetCount=0;
    while(originCount < unmatchedOrigin.size() && targetCount < unmatchedTarget.size())
    {
      int originAdv = originCount;
      boolean match = false;
      while(originAdv < unmatchedOrigin.size())
      {
        if(findingsComparable(((Relation) unmatchedOrigin.elementAt(originAdv)), ((Relation) unmatchedTarget.elementAt(targetCount))))
        {
          match = true;
          break;
        }
        else
          originAdv++;
      }

      if(match)
      {
        originCount = originAdv;
        Relation originRel = (Relation) unmatchedOrigin.elementAt(originCount);
        Relation targetRel = (Relation) unmatchedTarget.elementAt(targetCount);

        CaseFindingComparison cfc = new CaseFindingComparison(originRel, targetRel);

        Vector v = null;
        if(findingToComparisons.containsKey(targetRel))
          v = (Vector) findingToComparisons.get(targetRel);
        else
          v = new Vector();

        v.addElement(cfc);
        findingToComparisons.put(targetRel, v);

        comparisons.add(cfc);
      }
      originCount++;
      targetCount++;
    }
//    time2 = time2-time;
//      time3 = time3-time;
//    time = System.currentTimeMillis()-time;
//    System.out.println("CaseComparison between "+getOrigin().getName()+" and "+getTarget().getName()+" took "+time+" ms "+"("+time2+")");
    */
  }//reset

//------------------------------------------------------------------------------

  public boolean findingsComparable(Relation origin, Relation target)
  {

    if(target == null || origin == null)
    {
      if(DEBUG_LEVEL > 2)
        System.out.println("      NOT COMPARABLE - origin or target is null.");
      return false;
    }

    try
    {
      Relation originValOf = origin.getValue().getLocalRelation(new String[] {"value of", CBRModel.INSTANCE_OF, CBRModel.SUBCLASS_OF});

      Relation targetValOf = target.getValue().getLocalRelation(new String[] {"value of", CBRModel.INSTANCE_OF, CBRModel.SUBCLASS_OF});

      if((originValOf == null || targetValOf == null) && !originValOf.getValue().equals(targetValOf.getValue()))
      {
        if(DEBUG_LEVEL > 2)
          System.out.println("      NOT COMPARABLE - values of target or origin is null or are not equal.");
        return false;
      }
    }
    catch(NoSuchRelationException e)
    {
    }
    catch(NoSuchRelationTypeException e)
    {
    }


    if(!origin.getRelationType().isAssignableFrom(target.getRelationType()) &&
       !target.getRelationType().isAssignableFrom(origin.getRelationType()))
    {
      if(DEBUG_LEVEL > 2)
        System.out.println("      NOT COMPARABLE - relation-types do not match.");
      return false;
    }

    return true;
  }

  /**
   * Activates the comparison and returns the activation strenght.
   * i.e. matches the activation comparison list to put all relations in
   * correct lists (matched or unmatched) and to obtain the relation comparisons,
   * and then compute the strength.
   *
   * This method should not be called to often.
   * Instead, call 'activate' once
   * and then 'getStrength' each time you need the activation strength.
   *
   * @return the activation strength of the comparison
   */
  protected double updateActivationStrength()
  {
    if(DEBUG_LEVEL > 0)
      System.out.println("updateActivationStrength in casecomparison between "+getOrigin()+" and "+getTarget());

    //return 0 if there are no origin and/or target
    if (getOrigin() == null || getTarget() == null)
    {
      return 0;
    }//if null

    //return 1 if the target equals the origin
    else if (getOrigin().equals(getTarget()))
    {
      return 1;
    }//if equal

    //compute the strength if the origin and target differ
    else
    {
      //compute the strength
      double strength = 0.0;

      for(Enumeration enum=findingToComparisons.keys();enum.hasMoreElements();)
      {
        Relation finding = (Relation) enum.nextElement();
//        System.out.println("Case "+getOrigin()+" to "+getTarget()+" Activation for "+finding);
        Vector comps = (Vector) findingToComparisons.get(finding);
        double highestCompStrength = 0.0;
        CaseFindingComparison highestComp = null;
        for(int i=0;i<comps.size();i++)
        {
          CaseFindingComparison comp = (CaseFindingComparison) comps.elementAt(i);
          comp.setRelevance(0.0);
          comp.activate();
          if(DEBUG_LEVEL > 1)
            System.out.println("   -> "+comp.getOriginRelation()+" strength "+comp.getActivationStrength());
          if(!comp.isUnmatched())
          {
            unmatchedOrigin.remove(comp.getOriginRelation());
            unmatchedTarget.remove(comp.getTargetRelation());
            if(comp.getStrength() > highestCompStrength)
            {
              highestComp = comp;
              highestCompStrength = comp.getStrength();
            }
          }

          if(DEBUG_LEVEL > 1)
          System.out.println("Comp ("+comp.hashCode()+") "+comp.getOriginRelation()+" compared to "+comp.getTargetRelation()+" has strength "+comp.getStrength());
        }

        double relevanceStrength = getRelevanceFactor(finding);
        if(highestComp != null)
        {
          highestComp.setRelevance(relevanceStrength);
          strength += highestCompStrength * relevanceStrength;
        }
      }

//      System.out.println("RelevanceStrength in activate is "+relevanceStrengthSum+" strength is "+strength);
      if (relevanceStrengthSum > 0)
      {
        strength = strength/relevanceStrengthSum;
      }//if
      else
      {
        return 0;
      }

      //return the strength
      return strength;
    }//else

  }//getActivationStrength

//------------------------------------------------------------------------------

  /**
   * Explains the comparison and returns the explanation strenght.
   * i.e. matches the explanation comparison list to put all relations in
   * correct lists (matched or unmatched) and to obtain the relation comparisons,
   * and then compute the strength.
   * If the comparison is not activated already, this will be done first.
   *
   * This method should not be called to often.
   * Instead, call 'explain' once
   * and then 'getStrength' each time you need the explanation strength.
   *
   * @return the explanation strength of the comparison
   */
   protected double updateExplanationStrength()
   {
     if(DEBUG_LEVEL > 0)
       System.out.println("updateExplanationStrength in casecomparison between "+getOrigin()+" and "+getTarget());

      //return 0 if there are no origin and/or target
      if (getOrigin() == null || getTarget() == null)
      {
        return 0;
      }//if null

      //return 1 if the target equals the origin
      else if (getOrigin().equals(getTarget()))
      {
        return 1;
      }//if equal

      //compute the strength if the origin and target differ
      else
      {

        //compute the strength
        double strength = 0.0;

        for(Enumeration enum=findingToComparisons.keys();enum.hasMoreElements();)
        {
          Relation finding = (Relation) enum.nextElement();
          if(unmatchedTarget.contains(finding))
          {
            Vector comps = (Vector) findingToComparisons.get(finding);
            double highestCompStrength = 0.0;
            CaseFindingComparison highestComp = null;
            for(int i=0;i<comps.size();i++)
            {
              CaseFindingComparison comp = (CaseFindingComparison) comps.elementAt(i);
              comp.explain();
              comp.setRelevance(0.0);
              if(!comp.isUnmatched())
              {
                unmatchedOrigin.remove(comp.getOriginRelation());
                unmatchedTarget.remove(comp.getTargetRelation());
                if(comp.getStrength() > highestCompStrength)
                {
                  highestComp = comp;
                  highestCompStrength = comp.getStrength();
                }
              }

              if(DEBUG_LEVEL > 1)
                System.out.println("Comp ("+comp.hashCode()+") "+comp.getOriginRelation()+" compared to "+comp.getTargetRelation()+" has strength "+comp.getStrength());
            }
            double relevanceStrength = getRelevanceFactor(finding);
            if(highestComp != null)
            {
              highestComp.setRelevance(relevanceStrength);
              strength += highestCompStrength * relevanceStrength;
            }
          }
        }

        if (relevanceStrengthSum > 0)
        {
          strength = strength/relevanceStrengthSum;
        }//if
        else
        {
          return 0;
        }

        //return the strength
        return strength;
      }//else

  }//getExplanationStrength


//------------------------------------------------------------------------------

  /**
   * Returns all the relations of origin entity of the comparison.
   *
   * @return the relations of the origin entity of the comparison.
   */
  public Relation[] getOriginRelations()
  {
    return getOriginRelations(getKnowledgeModel().getRelationType(CBRModel.FINDING_RELATION));
  }//getOriginRelations

//------------------------------------------------------------------------------

  /**
   * Returns all the relations of the given type of origin entity of the comparison.
   *
   * @param relationClass the class of relations to return for the origin entity.
   * @return the relations of the given type origin entity of the comparison.
   */
  public Relation[] getOriginRelations(RelationType relationClass)
  {
    if (getOrigin() != null && relationClass != null)
      return removeNegated(getOrigin().getRelations(relationClass, new PlausibleInheritanceMethod(getKnowledgeModel())));
    else
     return new Relation[0];
  }//getOriginRelations()

//------------------------------------------------------------------------------

  /**
   * Returns all the relations of the target entity of the comparison.
   *
   * @return the relations of the target entity of the comparison.
   */
  public Relation[] getTargetRelations()
  {
    return getTargetRelations(getKnowledgeModel().getRelationType(CBRModel.FINDING_RELATION));
  }//getTargetRelations

//------------------------------------------------------------------------------

  /**
   * Returns all the relations of the given type of target entity of the comparison.
   *
   * @param relationClass the class of relations to return for the target entity.
   * @return the relations of the given type target entity of the comparison.
   */
  public Relation[] getTargetRelations(RelationType relationClass)
  {
    if (getTarget() != null && relationClass != null)
    {
      Relation rels[] = getTarget().getRelations(relationClass,
                                      new
                                      SortedPlausibleInheritanceMethod(getKnowledgeModel()));
      return removeNegated(rels);
    }
    else
     return new Relation[0];
  }//getTargetRelations()


  /**
   * Removes negated relations (relations with strength 0.0) from list.
   *
   * @param List of relations, possibly containing negated relations
   * @return List of relations with negated relations removed.
   */
  private Relation[] removeNegated(Relation[] relationList)
  {
    ArrayList relList = new ArrayList();
    for(int i=0;i<relationList.length;i++)
      if(!relationList[i].isNegated())
        relList.add(relationList[i]);
//      else
//        System.out.println("Negated and removed: "+relationList[i]);
    return (Relation[]) relList.toArray(new Relation[relList.size()]);
  }

//------------------------------------------------------------------------------

  /**
   * Returns the list of all finding comparisons.
   *
   * @return list of all finding comparisons
   */
  public CaseFindingComparison[] getFindingComparisons()
  {
    return (CaseFindingComparison[]) comparisons.toArray(new CaseFindingComparison[0]);
  }//getFindingComparisons

//------------------------------------------------------------------------------

  /**
   * Returns the list of completely matched finding comparisons.
   *
   * @return list of completely matched finding comparisons
   */
  public CaseFindingComparison[] getCompletelyMatchedComparisons()
  {
    //get all finding comparisons
    CaseFindingComparison[] all = getFindingComparisons();

    //extract correct comparisons
    Vector matchVector = new Vector();
    for (int i=0; i<all.length; i++)
    {
      if (all[i].isCompletelyMatched())
      {
        matchVector.add(all[i]);
      }//if
    }//for all

    //put the correct comparisons in an array
    CaseFindingComparison[] comps = new CaseFindingComparison[matchVector.size()];
    matchVector.copyInto(comps);
    return comps;
  }//getCompletelyMatchedComparisons

//------------------------------------------------------------------------------

  /**
   * Returns the list of partly matched finding comparisons.
   *
   * @return list of partly matched finding comparisons
   */
  public CaseFindingComparison[] getPartlyMatchedComparisons()
  {
    //get all finding comparisons
    CaseFindingComparison[] all = getFindingComparisons();

    //extract correct comparisons
    Vector matchVector = new Vector();
    for (int i=0; i<all.length; i++)
    {
      if (all[i].isPartlyMatched())
      {
        matchVector.add(all[i]);
      }//if
    }//for all

    //put the correct comparisons in an array
    CaseFindingComparison[] comps = new CaseFindingComparison[matchVector.size()];
    matchVector.copyInto(comps);
    Arrays.sort(comps, new Comparator()
    {
      public int compare(Object ob1, Object ob2)
      {
        CaseFindingComparison f1 = (CaseFindingComparison) ob1;
        Entity from1 = f1.fromTarget.getValue();
        String name1 = from1.getName();
        CaseFindingComparison f2 = (CaseFindingComparison) ob2;
        Entity from2 = f2.fromTarget.getValue();
        String name2 = from2.getName();

        int ret = name1.compareTo(name2);
        if(ret == 0)
        {
          return new Double(f2.getRelevance()).compareTo(new Double(f1.getRelevance()));
        }
        else
          return ret;
      }
    });

    return comps;
  }//getPartlyMatchedComparisons

//------------------------------------------------------------------------------

  /**
   * Returns the list of unmatched finding comparisons.
   *
   * @return list of unmatched finding comparisons
   */
  public CaseFindingComparison[] getUnmatchedComparisons()
  {
    //get all finding comparisons
    CaseFindingComparison[] all = getFindingComparisons();

    //extract correct comparisons
    Vector matchVector = new Vector();
    for (int i=0; i<all.length; i++)
    {
      if (all[i].isUnmatched())
      {
        matchVector.add(all[i]);
      }//if
    }//for all

    //put the correct comparisons in an array
    CaseFindingComparison[] comps = new CaseFindingComparison[matchVector.size()];
    matchVector.copyInto(comps);
    return comps;
  }//getUnmatchedComparisons

//------------------------------------------------------------------------------

  /**
   * Return the finding-relations in the origin case that are currently
   * not matched with any relations in the target case.
   *
   */
  public Relation[] getUnmatchedOriginRelations()
  {
    return (Relation[]) unmatchedOrigin.toArray(new Relation[0]);
  }

  /**
   * Return the finding-relations in the target case that are currently
   * not matched with any relations in the origin case.
   *
   */
  public Relation[] getUnmatchedTargetRelations()
  {
    return (Relation[]) unmatchedTarget.toArray(new Relation[0]);
  }

  /**
   * Returns the sum of the relevance factors of all findings in the target
   * case. This can be used to normalize the relative relevance of the findings.
   * @return double
   */
  public double getRelevanceFactorSum()
  {
    return relevanceStrengthSum;
  }

  /**
   * Returns the total strength of the case comparison, defined as:
   *   activation strength+explanation strength-penalty for unmatched origin findings.
   * @return double
   */
  public double getStrength()
  {
    double strength = super.getStrength()-getUnmatchedOriginRelations().length*getUnmatchedPenalty();
    if(strength >= 0.0)
      return strength;
    else
      return 0.0;
  }

  /**
   * Returns the penalty applied to the score for a case for each unmatched origin case finding.
   * @return double
   */
  public double getUnmatchedPenalty()
  {
    return UNMATCHED_ORIGIN_PENALTY/getOriginRelations().length;
  }

  /**
   * Returns the list of explained matched finding comparisons.
   *
   * @return list of explained matched finding comparisons
   */
  public CaseFindingComparison[] getExplainedMatchedComparisons()
  {
    //get all finding comparisons
    CaseFindingComparison[] all = getFindingComparisons();

    //extract correct comparisons
    Vector matchVector = new Vector();
    for (int i=0; i<all.length; i++)
    {
      if (all[i].isExplainMatched())
      {
        matchVector.add(all[i]);
      }//if
    }//for all

    //put the correct comparisons in an array
    CaseFindingComparison[] comps = new CaseFindingComparison[matchVector.size()];
    matchVector.copyInto(comps);
    return comps;
  }//getExplainedMatchedComparisons

//------------------------------------------------------------------------------

  public boolean hasFindings()
  {
    return (getFindingComparisons().length > 0);
  }//hasFindings

//------------------------------------------------------------------------------

  /**
   * Returns the string representation of the entity comparison
   *
   * @return the string representation of the entity comparison
   */
  public String toString()
  {
    //initialize the string representation
    String str = new String();
    //add the properties of the super class
    str = str + super.toString();
    str = str + "\n";

    //add facts about relation comparisons
      str = str + "Unmatched relation comparisons:          ";
      str = str + getUnmatchedComparisons().length + "\n";
      str = str + "Partially matched relation comparisons:  ";
      str = str + getPartlyMatchedComparisons().length + "\n";
      str = str + "Completely matched relation comparisons: ";
      str = str + getCompletelyMatchedComparisons().length + "\n";
      str = str + "Explained relation comparisons:          ";
      str = str + getExplainedMatchedComparisons().length + "\n";

    //print the relation comparisons
    str = str + "\n" + "The finding comparisons:";
    CaseFindingComparison[] comps = getFindingComparisons();
    for (int i=0; i<comps.length; i++)
    {
      str = str + "\n\t" + comps[i].getOrigin()+" and "+comps[i].getTarget()+" (relevance "+comps[i].getRelevance()+") has strength "+comps[i].getStrength()+" - explainstr: "+comps[i].getExplanationStrength();
    }//for comparisons


    //return the string
    return str;
  }//toString


  /**
   * Returns a submodel of the entities compared.
   */
  public SubModel toSubModel()
  {
    SubModel model = super.toSubModel();

    CaseFindingComparison[] comps = getCompletelyMatchedComparisons();
    for(int i=0;i<comps.length;i++)
    {
      model.add(comps[i].getOriginRelation());
      model.add(comps[i].getTargetRelation());
      model = model.union(comps[i].toSubModel());
    }

    comps = getPartlyMatchedComparisons();
    for(int i=0;i<comps.length;i++)
    {
      model.add(comps[i].getOriginRelation());
      model.add(comps[i].getTargetRelation());
      model.union(comps[i].toSubModel());
    }

    Relation[] rels = getUnmatchedOriginRelations();
    for(int i=0;i<rels.length;i++)
    {
      model.add(rels[i].getValue());
      model.add(rels[i]);
    }

    rels = getUnmatchedTargetRelations();
    for(int i=0;i<rels.length;i++)
    {
      model.add(rels[i].getValue());
      model.add(rels[i]);
    }


    model.add(getTarget());
    return model;
  }

//------------------------------------------------------------------------------
  public static Entity[] getCompleteFindingList(Case cases[])
  {
    if(cases.length < 1)
      return new Entity[0];

    LRUCache ic = cases[0].getKnowledgeModel().getInferenceCache();
    Object id = new Object[] {"getCompleteFindingList"};
    if(ic.isCached(id))
      return (Entity[]) ic.get(id);

    Vector findings = new Vector();
    for(int i=cases.length-1;i>=0;i--)
    {
      Relation[] finding = cases[i].getRelations(CBRModel.FINDING_RELATION);
      for(int j=finding.length-1;j>=0;j--)
      {
        try
        {
          Entity value = finding[j].getValue();
          if(value != null)
          {
            Relation valueOf = value.getRelation(new String[] {"value of", CBRModel.INSTANCE_OF, CBRModel.SUBCLASS_OF});
            if(!findings.contains(valueOf.getValue()))
            {
              findings.addElement(valueOf.getValue());
            }
          }
        }
        catch(NoSuchRelationException e)
        {
          e.printStackTrace();
        }
        catch(NoSuchRelationTypeException e)
        {
          e.printStackTrace();
        }
      }
    }
    Entity[] ents = (Entity[]) findings.toArray(new Entity[0]);
    ic.put(id, ents);
    return ents;
  }
/*
  public static Relation[] getCaseFindingList(Entity myCase, Entity[] findingList)
  {
    // THIS IS SLOW - 100 ms +
    InferenceCache ic = myCase.getKnowledgeModel().getInferenceCache();
    Object[] id = new Object[] {"getCaseFindingList", myCase};
    if(ic.isCached(id))
      return (Relation[]) ic.get(id);

//    System.out.print("!");

    Relation ret[] = new Relation[findingList.length];
    for(int i=findingList.length-1;i>=0;i--)
      ret[i] = getFinding(myCase, findingList[i]);
    ic.put(id, ret);
    return ret;
  }

  public static Relation getFinding(Entity myCase, Entity findEnt)
  {
//    InferenceCache ic = myCase.getKnowledgeModel().getInferenceCache();
//    Object[] id = new Object[] {"getFinding", myCase, findEnt};
/*
    if(ic.isCached(id))
    {
      Object ob = ic.get(id);
      if(ob.equals(new Integer(0)))
        return null;
      else
        return (Relation) ob;
    }

    System.out.print(".");*/
/*    Relation[] finding = myCase.getRelations("has finding");
    for(int j=finding.length-1;j>=0;j--)
    {
      try
      {
        Entity value = finding[j].getValue();
        if(value != null)
        {
          Relation valueOf = value.getRelation(new String[] {"value of", "instance of", "subclass of"});
          if(valueOf.getValue().equals(findEnt))
          {
//            ic.put(id, finding[j]);
            return finding[j];
          }
        }
      }
      catch(NoSuchRelationException e)
      {}
    }
//    ic.put(id, new Integer(0));
    return null;
  }
*/
}//Class EntityComparison
