package jcreek.reasoning;

import jcreek.representation.*; import jcreek.representation.cbr.*;

/**
 * <p>Title: CaseFindingComparison </p>
 * <p>Description: Structure used when matching two cases. This structure contains a matching of
 * two findings of the cases. This is not really a </p>
 * @author Frode Srmo
 * @version 1.0
 */

public class CaseFindingComparison extends EntityComparison
{
  protected Relation fromOrigin;
  protected Relation fromTarget;
  protected EntityComparison comparison;
  protected double relevance;

  /**
   * Creates a new CaseFindingComparison comparing the two relations given.
   * The EntityComparison.createComparison() method is used to create the actual
   * comparator - this class just wraps around that and also contains the relations
   * and relevance strength.
   *
   * @param fromOrigin The relation from the origin (new) case.
   * @param fromTarget The relation from the target (retrieved) case.
   */
  public CaseFindingComparison(Relation fromOrigin, Relation fromTarget)
  {
    super(fromOrigin.getValue(), fromTarget.getValue());
    this.fromOrigin = fromOrigin;
    this.fromTarget = fromTarget;
    this.comparison = EntityComparison.createComparison(fromOrigin.getValue(), fromTarget.getValue());
    reset();
  }

  public void reset()
  {
    if(comparison != null)
      comparison.reset();
  }

  public void activate()
  {
    comparison.activate();
  }

  public void explain()
  {
    comparison.explain();
  }

  public int getStatus()
  {
    return comparison.getStatus();
  }

  public SubModel toSubModel()
  {
    SubModel model = comparison.toSubModel();
    model.add(getTargetRelation());
    model.add(getOriginRelation());
    return model;
  }

  public String shortExplanation()
  {
    return comparison.shortExplanation();
  }
  /**
   * Updates the activation strength.
   */
  public double updateActivationStrength()
  {
    return comparison.updateActivationStrength();
  }

  /**
   * Updates the explanation strength.
   */
  public double updateExplanationStrength()
  {
    return comparison.updateExplanationStrength();
  }

  /**
   * Gives the encapsulated comparison.
   */
  public EntityComparison getComparison()
  {
    return comparison;
  }

  /**
   * Sets the relevance weight of this CaseFindingComparison. This may vary
   * from getTargetRelation().getRelevance() since that relation may be matched
   * to more than one relation in the origin (new) case. If this is the case,
   * the relevance is divided by the number of such matches.
   */
  public void setRelevance(double d)
  {
    relevance = d;
  }

  /**
   * Returns the relevance weight of this comparison.This may vary
   * from getTargetRelation().getRelevance() since that relation may be matched
   * to more than one relation in the origin (new) case. If this is the case,
   * the relevance is divided by the number of such matches.
   */
  public double getRelevance()
  {
    return relevance;
  }

  /**
   * Returns the finding-relation from the origin (new) case that is the basis
   * of this comparison.
   */
  public Relation getOriginRelation()
  {
    return fromOrigin;
  }

  /**
   * Return the finding-relation from the target (retrieved) case that is the basis
   * of this comparison.
   */
  public Relation getTargetRelation()
  {
    return fromTarget;
  }

  /**
   * Returns the comparison comparing the entities that are the values of the
   * finding-relations of the origin (new) and target (retrieved) case.
   */
  public EntityComparison getEntityComparison()
  {
    return comparison;
  }

  /**
   * Returns the activation strength of the comparisons.
   */

  public double getActivationStrength()
  {
    return comparison.getActivationStrength();
  }

  /**
   * Returns the explanation strength of the comparisons.
   */

  public double getExplanationStrength()
  {
    return comparison.getExplanationStrength();
  }

}
