package jcreek.reasoning;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.util.LRUCache;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;

/**
 * <p>
 * An abstract class for comparison of two entities.
 * <p>
 * The similarity of two entities is measured by a strength.
 * The strength is only computed if the comparison is comparable and the depth
 * is above the threshold depth for comparison.
 * The definition of comparable depends on whether the comparison is between
 * entities or relations.
 * Implementations of this class must therefore include a method for determining
 * whether the comparison is comparable or not.
 * They must include algorithms on how to compute the comparison strength.
 *
 * @author Ellen Lippe, modified by Frode Srmo of NTNU 2001-2002
 * @version 1.1
 */

public abstract class EntityComparison implements Comparable
{

  /** The status of the comparison before activation */
  public static final int INITIAL_STATUS = 0;
  /** The status of the comparison after activation */
  public static final int ACTIVATED_STATUS = 1;
  /** The status of the comparison after explanation */
  public static final int EXPLAINED_STATUS = 2;

  /** The initial strength of the comparison */
  protected static final double INITIAL_STRENGTH = 0.0;

  /** The origin concept */
  protected Entity origin;
  /** The target concept */
  protected Entity target;

  /** The activation strength */
  protected double activationStrength = INITIAL_STRENGTH;
  /** The explanation strength */
  protected double explanationStrength = INITIAL_STRENGTH;


  /** The status */
  private int status = INITIAL_STATUS;


//------------------------------------------------------------------------------

  /**
   * Constructs a comparison between the origin concept and the target concept
   * with given depth.
   *
   * @param origin the concept to compare from
   * @param target the concept to compare to
   * @param depth the depth of the comparison
   */
  public EntityComparison(Entity origin, Entity target)
  {
    //set the origin and target concepts and the depth
    this.origin = origin;
    this.target = target;
    reset();
  }//Class EntityComparison


  protected EntityComparison()
  {

  }
//------------------------------------------------------------------------------

  /**
   * Computes the activation strenght of the comparison.
   * This method should not be called to often.
   * Instead, call 'activate' and then 'getActivationStrength'
   * When implemented by subclasses, this method should consist of the
   * activation between the comparing concepts
   *
   * @return the activation strength of the comparison
   */
  protected abstract double updateActivationStrength();

  /**
   * Updates the explanation strenght of the comparison.
   * This method should not be called to often.
   * Instead, call 'explain' and then 'getStrength'
   * When implemented by subclasses, this method should consist of the
   * explanation between the comparing concepts
   *
   * @return the explanation strength of the comparison
   */
  protected abstract double updateExplanationStrength();

  /**
   * Tests if the entities in the comparison is comparable.
   * @return true if the comparison is comparable.
   */
  public boolean isComparable()
  {
    return checkIsComparable(origin, target);
  }

  /**
   * Tests if the two given entities are comparable.
   *
   * @return true if these entities are possible to compare.
   */

  public static boolean checkIsComparable(Entity e1, Entity e2)
  {
    if(e1.equals(e1.getKnowledgeModel().getEntity(CBRModel.NO_VALUE)) ||
       e2.equals(e2.getKnowledgeModel().getEntity(CBRModel.NO_VALUE)))
      return false;

    KnowledgeModel km = e1.getKnowledgeModel();
    LRUCache ic = km.getInferenceCache();
    Object id[] = new Object[] {"checkIsComparable", e1, e2};
//    if(ic.isCached(id))
//    {
//      System.out.print(",");
//      return ((Boolean) ic.get(id)).booleanValue();
//    }
//    System.out.print(".");


    try
    {
      Entity comp = e1.getRelation(CBRModel.HAS_COMPARATOR).getValue();
//      System.out.println("=> Comparator of "+e1+" is "+comp);
      if(e1.getRelation(CBRModel.HAS_COMPARATOR).getValue().equals(e2.getRelation(CBRModel.HAS_COMPARATOR).getValue()))
      {
//        Entity comp = e1.getRelation(CBRModel.HAS_COMPARATOR).getValue();
//        System.out.println("Comp is "+comp);
        Class compClass = (Class) comp.getEntityObject();
        Method compMethod = compClass.getMethod("isComparable", new Class[] {Entity.class, Entity.class });
        Object ob = compMethod.invoke(null, new Object[] {e1, e2 });
//        ic.put(id, ob);
//        System.out.println("isComparable "+e1.getName()+" and "+e2.getName()+" "+ob);
        return ((Boolean) ob).booleanValue();
      }
    }
    catch(NoSuchRelationException e)
    {
//      e.printStackTrace();
    }
    catch(NoSuchRelationTypeException e)
    {
//      e.printStackTrace();
    }
    catch(NoSuchMethodException e)
    {
//      e.printStackTrace();
    }
    catch(ClassCastException e)
    {
//      e.printStackTrace();
    }
    catch(java.lang.reflect.InvocationTargetException e)
    {
      e.printStackTrace();
    }
    catch(IllegalAccessException e)
    {
      e.printStackTrace();
    }
    catch(NullPointerException e)
    {
//      e.printStackTrace();
    }

//    ic.put("checkIsComparable", id, new Boolean(false));
    return false;
/*
    KnowledgeModel km = e1.getKnowledgeModel();
    InferenceCache ic = km.getInferenceCache();
    Long id = new Long(e1.hashCode()+e2.hashCode());
    if(ic.isCached("isComparable", id))
      return ic.get("isComparable", id).equals(new Integer(1));

    try
    {
      if(e1.getRelation("has comparator").getValue().equals(e2.getRelation("has comparator").getValue()))
      {
        Entity valueType;
//        System.out.println("Attempting to find (symbol) valuetype of "+e1+" and "+e2);
        Relation[] originValueOf = e1.getRelations("value of");
        Relation[] targetValueOf = e2.getRelations("value of");

        for(int i=originValueOf.length-1;i>=0;i--)
        {
          for(int j=targetValueOf.length-1;j>=0;j--)
          {
//            System.out.println("Attempted match: "+originValueOf[i]+" and "+targetValueOf[j]);
            if(originValueOf[i].getValue().equals(targetValueOf[j].getValue()))
            {
              valueType = originValueOf[i].getValue();
//              System.out.println("Found value type - it is "+valueType);
              ic.put("isComparator", id, new Integer(1));
              return true;
            }
          }
        }
      }
    }
    catch(NoSuchRelationException e)
    {
    }
    ic.put("isComparator", id, new Integer(0));
    return false;

    */
  }

  public static boolean isComparable(Entity e1, Entity e2)
  {
    return e1 != e1.getKnowledgeModel().getEntity(CBRModel.NO_VALUE) &&
           e2 != e2.getKnowledgeModel().getEntity(CBRModel.NO_VALUE);

  }

  /**
   * Resets the activation/explanation.
   * i.e sets the status to INITIAL_STATUS and strength to INITIAL_STRENGTH.
   */
  public void reset()
  {
    status = INITIAL_STATUS;
    activationStrength = INITIAL_STRENGTH;
    explanationStrength = INITIAL_STRENGTH;
  }//reset

  /**
   *
   */
  public void update()
  {
    int oldStatus = status;
    reset();
    if(oldStatus == ACTIVATED_STATUS)
    {
      activate();
    }
    else if(oldStatus == EXPLAINED_STATUS)
    {
      activate();
      explain();
    }
  }
//------------------------------------------------------------------------------

  /**
   * Activates the comparison between the origin and target concepts.
   * i.e sets the status to ACTIVATED_STATUS
   * and computes the activation strength if the concepts are comparable
   * and the depth is above threshold.
   *
   * @see isComparable()
   */
  public void activate()
  {
    if (status != INITIAL_STATUS)
    {
      reset();
    }//if not initial

    status = ACTIVATED_STATUS;

    activationStrength = updateActivationStrength();
  }//activate

//------------------------------------------------------------------------------

  /**
   * Tests whether the comparison is activated.
   * The comparison is activated if the status is ACTIVATED_STATUS
   *
   * @return true if the comparison is activated
   */
  public boolean isActivated()
  {
    return (getStatus() == ACTIVATED_STATUS);
  }//isActivated

//------------------------------------------------------------------------------

  /**
   * Explains the comparison between the origin and target concepts.
   * i.e sets the status to EXPLAINED_STATUS
   * and computes the explanation strength if the concepts are comparable
   * and the depth is above threshold.
   *
   * @see isComparable()
   */
  public void explain()
  {
    if (!isActivated())
    {
      activate();
    }//if not activated

    explanationStrength = updateExplanationStrength();
    status = EXPLAINED_STATUS;

  }//explain

//------------------------------------------------------------------------------

  /**
   * Tests whether the comparison is explained.
   * The comparison is activated if the status is EXPLAINED_STATUS
   *
   * @return true if the comparison is explained
   */
  public boolean isExplained()
  {
    return (getStatus() == EXPLAINED_STATUS);
  }//isExplained

//------------------------------------------------------------------------------

  /**
   * Returns the knowledge model of the origin concept.
   *
   * @return the knowledge model of the origin concept.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    if (getOrigin() != null)
      return getOrigin().getKnowledgeModel();
    else
      return null;
  }//getKnowledgeModel

//------------------------------------------------------------------------------

  /**
   * Returns the origin concept of the comparison.
   *
   * @return the origin concept of the comparison.
   */
  public Entity getOrigin()
  {
    return origin;
  }//getOrigin

//------------------------------------------------------------------------------

  /**
   * Returns the target concept of the comparison.
   *
   * @return the target concept of the comparison.
   */
  public Entity getTarget()
  {
    return target;
  }//getTarget

//------------------------------------------------------------------------------

  /**
   * Returns the status of the comparison.
   *
   * @return the status of the comparison
   *
   * @see INITIAL_STATUS
   * @see ACTIVATED_STATUS
   * @see  EXPLAINED_STATUS
   */
  public int getStatus()
  {
    return status;
  }//getStatus

//------------------------------------------------------------------------------

  /**
   * Returns the status of the comparison as a text string.
   *
   * @return the status of the comparison as a text string
   *
   * @see INITIAL_STATUS
   * @see ACTIVATED_STATUS
   * @see  EXPLAINED_STATUS
   */
  public String getStatusText()
  {
    switch (getStatus())
    {
      case INITIAL_STATUS : return "Initial status";
      case ACTIVATED_STATUS : return "Activated status";
      case EXPLAINED_STATUS: return "Explained status";
      default: return "No status";  //should never occur!!
    }//switch
  }//getStatus

//------------------------------------------------------------------------------

  /**
   * Returns the activation strength of the comparison.
   */
  public double getActivationStrength()
  {
    return activationStrength;
  }//getActivationStrength()

//------------------------------------------------------------------------------

  /**
   * Returns the explanation strength of the comparison.
   */
  public double getExplanationStrength()
  {
    return explanationStrength;
  }//getExplanationStrength


  /**
   * Returns a submodel of the entities compared.
   */
  public SubModel toSubModel()
  {
    SubModel model = new SubModel();
    model.add(getOrigin());
    model.add(getTarget());
    return model;
  }
//------------------------------------------------------------------------------

  /**
   * Returns the strength of the comparison
   *
   * @return the strenght of the comparison
   */
  public double getStrength()
  {
    if (isExplained())
      return (getActivationStrength() + getExplanationStrength());
    else if (isActivated())
      return getActivationStrength();
    else return 0.0;
  }//getStrength

  /**
   * Returns the strength of the comparison as a percentage int.
   */

  public int getPercentageStrength()
  {
    return (int) Math.round(getStrength()*100.0);
  }
//------------------------------------------------------------------------------

  /**
   * Formula used for calculate the strength of two explanations when they are
   * parallell, i.e. either can be used.
   * Assumptions:   0 <= s1 and s2 <= 1.0
   *                addParallell(s1, s2) == addParallell(s2, s1)
   */

  public static double addParallell(double s1, double s2)
  {
    if(s1 > s2)
      return s1;
    else
      return s2;

//    return s1 + (1.0 - s1) * s2;
  }

//------------------------------------------------------------------------------

  /**
   * Formula used for calculate the strength of two explanations when they are
   * serial, i.e. both must apply.
   * Assumptions:   0 <= s1 and s2 <= 1.0
   *                addParallell(s1, s2) == addParallell(s2, s1)
   */

  public static double addSerial(double s1, double s2)
  {
    return s1 * s2;
  }//addSerial

//-----------------------------------------------------------------------------

  /**
   * Tests whether the comparison is completely matched.
   * The comparison is completely matched if the activation strength is 1 or more.
   *
   * @return true if the comparison is completely matched.
   */
  public boolean isCompletelyMatched()
  {
    return (getActivationStrength() >= 1);
  }//isCompletelyMatched

//-----------------------------------------------------------------------------

  /**
   * Tests whether the comparison is partly matched.
   * The comparison is partly matched if the activation strength is
   * between 0 and 1.
   *
   * @return true if the comparison is partly matched.
   */
  public boolean isPartlyMatched()
  {
    return (getStrength() > 0 && getStrength() < 1);
  }//isPartlyMatched

//-----------------------------------------------------------------------------

  /**
   * Tests whether the comparison is unmatched.
   * The comparison is unmatched if the activation strength is 0 or less.
   *
   * @return true if the comparison is unmatched.
   */
  public boolean isUnmatched()
  {
    return (getStrength() <= 0);
  }//isUnmatched

//-----------------------------------------------------------------------------

  /**
   * Tests whether the comparison is explained matched.
   * The comparison is explained matched if the explanation strength is above 0.
   *
   * @return true if the comparison is explained matched.
   */
  public boolean isExplainMatched()
  {
    return (getExplanationStrength() > 0);
  }//isExplainedMatched

//------------------------------------------------------------------------------

  /**
   * Gives a short textual representation (a sentence or so) describing the
   * comparison.
   */

  public String shortExplanation()
  {
    return "The "+getOrigin().getValueString()+" is "+(int) Math.round(getStrength()*100)+"% similar "+
        "to "+getTarget().getValueString();
  }

  /**
   * Returns the string representation of the comparison
   *
   * @return the string representation of the comparison
   */
  public String toString()
  {
    //initialize the string representation
    String str = new String();
    //add the properties of the comparison
    str = str + "Origin:   " + getOrigin().toString() + "\n";
    str = str + "Target:   " + getTarget().toString() + "\n";
    str = str + "Status:   " + getStatusText() + "\n";
    str = str + "Strength: " + new Double(getStrength()).toString() + "\n";
 //   str = str + "A Strength: " + new Double(getActivationStrength()).toString() + "\n";
 //   str = str + "E Strength: " + new Double(getExplanationStrength()).toString() + "\n";
    return str;
  }//toString

  public static EntityComparison createComparison(Entity origin, Entity target)
  {
    LRUCache ic = origin.getKnowledgeModel().getInferenceCache();
    Object id[] = new Object[] {"createComparison", origin, target};

    if(ic.isCached(id))
    {
      Object o = ic.get(id);
//      System.out.println("comparison between "+origin+" and "+target+" is cached.");
      if(o.equals(new Integer(0)))
        return new NoMatchComparison(origin, target);

      EntityComparison comp = (EntityComparison) o;
      if(comp.getOrigin().equals(origin) && comp.getTarget().equals(target))
        return (EntityComparison) o;
    }

//    System.out.println("createComparison hard way");

    if(checkIsComparable(origin, target))
    {
      try
      {
        ClassEntity comparatorECD = new ClassEntity(origin.getRelation(CBRModel.HAS_COMPARATOR).getValue());
//        System.out.println("Has comparator "+comparatorECD);
        Class comparatorClass = comparatorECD.asClass();
        Class entityClass = origin.getClass();
        Constructor comparatorConst = null;
        while(comparatorConst == null && !entityClass.equals(Object.class))
        {
          try
          {
            comparatorConst = comparatorClass.getConstructor(new Class[] {entityClass, entityClass});
          }
          catch(java.lang.NoSuchMethodException e)
          {
            entityClass = entityClass.getSuperclass();
          }
        }

        if(comparatorConst != null)
        {
          try
          {
//            System.out.println(" ---> instanciating "+entityClass);
            EntityComparison comp = (EntityComparison) comparatorConst.newInstance(new Object[] {origin, target});
            ic.put(id, comp);
            return comp;
          }
          catch(java.lang.reflect.InvocationTargetException e)
          {
            e.printStackTrace();
          }
          catch(IllegalAccessException e)
          {
            e.printStackTrace();
          }
          catch(InstantiationException e)
          {
            e.printStackTrace();
          }
        }
        else
        {
            System.out.println("No common superclass comparison");
        }
      }
      catch(NoSuchRelationException e)
      {
            e.printStackTrace();
      }
      catch(NoSuchRelationTypeException e)
      {
            e.printStackTrace();
      }
      catch(ClassCastException e)
      {
            e.printStackTrace();
      }
      catch(EntityTypeNotMatchedException e)
      {
        e.printStackTrace();
      }
    }
//    else
//      System.out.println("Not comparable");

    ic.put(id, new Integer(0));
//    System.out.println("Failed to create comparison for "+origin.getName()+" and "+target.getName());
    return new NoMatchComparison(origin, target);

  }

  public int compareTo(Object a)
  {

    return new Double(((EntityComparison) a).getStrength()).compareTo(new Double(getStrength()));
  }
//------------------------------------------------------------------------------

  static class NoMatchComparison extends EntityComparison
  {
    NoMatchComparison(Entity e1, Entity e2)
    {
      super(e1, e2);
    }

    public double updateExplanationStrength() { return 0.0; }
    public double updateActivationStrength() { return 0.0; }
  }


}//Class EntityComparison


