package jcreek.reasoning;

/**
 * NumberComparison compares two entities containing numbers - i.e. where
 * the entities has a data object (as returned by getEntityObject()) which are both
 * subclasses of the java.lang.Number class. This can be floats, doubles, ints
 * and longs.
 * <p>
 * In activate, the comparison follows the "value of" relation from the entities to find a
 * common superclass for the two entities. If no such superclass is found, the
 * numbers are deemed incomparable. If found, it retrieves all values of this
 * superclass-entity by following all "has instance" relations in order to
 * normalize the range. The similarity of the numbers are given by:<br>
 * <ul>1-abs(e1 - e2) <br>
 * where e1 is the normalized value of the number contained in entity 1, and e2
 * is the normalized value of the number contained in entity 2.
 * </ul>
 * <p>
 * The explain step contain no further matching.
 * </p>
 * @author Frode Srmo
 * @version 1.0
 */

import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.util.LRUCache;
import java.util.ArrayList;

public class NumberComparison extends EntityComparison
{
  protected Entity valueType = null;
  protected ArrayList relList = null;
  protected boolean notComparable = false;

  /**
   * Creates a new NumberComparison, comparing the origin and target entities
   * given as parameters.
   */

  public NumberComparison(Entity origin, Entity target)
  {
    super(origin, target);
//    System.out.println("Creating new NumberComparison between "+origin+" and "+target);
  }

  /**
   * Nothing is done in the explanation step.
   *
   * @return Always 0.0 .
   */
  public double updateExplanationStrength()
  {
    return 0.0;
  }

  /**
   * The activation step attempts to find an entity having a "has instance" relation
   * to both the origin and target entities. All values from this super-entity
   * is then used to normalize the values of the two entities.
   */
  public double updateActivationStrength()
  {
    if(isComparable() &&
       getOrigin().getEntityObject() != null && Number.class.isAssignableFrom(getOrigin().getEntityObject().getClass()) &&
       getTarget().getEntityObject() != null && Number.class.isAssignableFrom(getTarget().getEntityObject().getClass()))
    {
      Number originValue = (Number) getOrigin().getEntityObject();
      Number targetValue = (Number) getTarget().getEntityObject();

      if(getNumberNormalizer().hasRange())
        return 1.0 - Math.abs(normalize(originValue)-normalize(targetValue));
      else
        return 0.0;
    }
    return 0.0;
  }

  /**
   * Checks if the entities are comparable. In order to be so, they must both
   * have the same comparator and also have a "value of" relation to a single
   * common entity.
   */
  public boolean isComparable()
  {
    if(notComparable)
      return false;
    if(valueType != null)
      return true;

    if(getOrigin().equals(getOrigin().getKnowledgeModel().getEntity(CBRModel.NO_VALUE)) ||
       getTarget().equals(getTarget().getKnowledgeModel().getEntity(CBRModel.NO_VALUE)))
      return false;

    relList = new ArrayList();
//    System.out.println("Attempting to find valuetype of "+getOrigin()+" and "+getTarget());
    Relation[] originValueOf = getOrigin().getLocalRelations(new String[] {"value of", "instance of", "subclass of"});
    Relation[] targetValueOf = getTarget().getLocalRelations(new String[] {"value of", "instance of", "subclass of"});

    for(int i=originValueOf.length-1;i>=0;i--)
    {
      for(int j=targetValueOf.length-1;j>=0;j--)
      {
//        System.out.println("Attempted match: "+originValueOf[i]+" and "+targetValueOf[j]);
        if(originValueOf[i].getValue().equals(targetValueOf[j].getValue()))
        {
          relList.add(originValueOf[i]);
          relList.add(targetValueOf[j]);
          valueType = originValueOf[i].getValue();
//          System.out.println("Found value type - it is "+valueType);
          return true;
        }
      }
    }
    notComparable = true;
    return false;
  }

  protected NumberNormalizer getNumberNormalizer()
  {
    if(isComparable())
    {
      LRUCache ic = getOrigin().getKnowledgeModel().getInferenceCache();
      Object id[] = new Object[] {"NumberNormalizer", valueType} ;
      if(ic.isCached(id))
        return (NumberNormalizer) ic.get(id);

      NumberNormalizer normalizer = new NumberNormalizer(valueType);
      ic.put(id, normalizer);

      return normalizer;
    }
    else
      return null;
  }

  public double normalize(Number number)
  {
    return getNumberNormalizer().normalize(number);
  }

  public String shortExplanation()
  {
    if(!isComparable() || valueType != null)
    {
      if(isCompletelyMatched())
        return "The values of "+valueType.getValueString()+" are the same.";
      else if(isPartlyMatched())
        return "The numbers "+getOrigin().getValueString()+" and "+getTarget().getValueString()+" are "+
        (int) Math.round(getStrength()*100)+"% similar, as the values of "+valueType.getValueString()+
        " varies from "+getNumberNormalizer().min+" to "+getNumberNormalizer().max;
      return "The values do not match.";
    }
    return "The values are not comparable.";
  }

  public SubModel toSubModel()
  {
    SubModel subModel = super.toSubModel();
    if(valueType != null)
    {
      subModel.add(valueType);
      subModel.add((Relation[]) relList.toArray(new Relation[relList.size()]));
    }
    return subModel;
  }
}




/**
 * Class used to normalize the value of a parameter.
 */
class NumberNormalizer
{
  protected Entity valueType;
  protected double min;
  protected double max;

  NumberNormalizer(Entity valueType)
  {
    this.valueType = valueType;
    updateMinMax();
  }

  protected void updateMinMax()
  {
    Relation[] valueRels = valueType.getLocalRelations(new String[] {"has value", "has instance", "has subclass" });

    for(int i=valueRels.length-1;i>=0;i--)
    {
      Entity valueEntity = valueRels[i].getValue();
      if(valueEntity.getEntityObject() != null && Number.class.isAssignableFrom(valueEntity.getEntityObject().getClass()))
      {
        double value = ((Number) valueEntity.getEntityObject()).doubleValue();
        if(i==valueRels.length-1 || min > value)
          min = value;
        if(i==valueRels.length-1 || max < value)
          max = value;
      }
    }
  }

  public boolean hasRange()
  {
    return max != min;
  }

  protected double normalize(Number value)
  {
    double val = value.doubleValue();
    return (val-min)/(max-min);
  }
}
