package jcreek.reasoning;

/**
 * Compares two entities representing partitions. The entities should contain
 * an object that is a valid ID to a partition registered with the knowledge
 * model.
 * <p>
 * The comparison measures the number of entities and relations that are shared
 * between the two partitions and normalizes the result from 0 to 1 (where 1
 * means that the origin partition is a subset of the target partition.
 *
 * @author Frode Srmo
 * @version 1.0
 */

import jcreek.representation.*; import jcreek.representation.cbr.*;

public class PartitionComparison extends EntityComparison
{
  boolean notComparable = false;

  /**
   * Creates a new NumberComparison, comparing the origin and target entities
   * given as parameters.
   */

  public PartitionComparison(Entity origin, Entity target)
  {
    super(origin, target);
  }

  /**
   * Nothing is done in the explanation step.
   *
   * @return Always 0.0 .
   */
  public double updateExplanationStrength()
  {
    return 0.0;
  }

  /**
   * The activation step attempts to find the partitions and measure the similarity
   * between them by comparing the number og relations and entities they have in
   * common to the total number of relations and entities in the target entity's
   * partition.
   */
  public double updateActivationStrength()
  {
//    System.out.println("updateActivateStrength for "+getOrigin()+" and "+getTarget());
    SubModel commonModel = getOverlapModel();
//    System.out.println("commonModel "+commonModel);
    SubModel unionModel = getUnionModel();
//    System.out.println("unionModel "+unionModel);
    if(commonModel != null && unionModel != null)
    {
      if(unionModel.size() != 0.0)
        return ((double) commonModel.size()) / ((double) unionModel.size());
    }

    return 0.0;
  }

  /**
   * Returns a submodel containing the overlap of the partitions compared by
   * this object.
   */
  public SubModel getUnionModel()
  {
    Object originPartitionID = getOrigin().getEntityObject();
    Object targetPartitionID = getTarget().getEntityObject();
    if(originPartitionID != null && targetPartitionID != null)
    {
      Partition originPartition = getOrigin().getKnowledgeModel().getPartition(
          originPartitionID);
      Partition targetPartition = getTarget().getKnowledgeModel().getPartition(
          targetPartitionID);
      if (originPartition != null && targetPartition != null)
      {
        SubModel unionModel = new SubModel();
        unionModel.add(originPartition);
        unionModel.add(targetPartition);

        return unionModel;
      }
    }
    return null;
  }


  /**
   * Returns a submodel containing the overlap of the partitions compared by
   * this object.
   */
  public SubModel getOverlapModel()
  {
    Object originPartitionID = getOrigin().getEntityObject();
    Object targetPartitionID = getTarget().getEntityObject();
    if(originPartitionID != null && targetPartitionID != null)
    {
      Partition originPartition = getOrigin().getKnowledgeModel().getPartition(
          originPartitionID);
//      System.out.println("  -> origin partition: "+originPartition);
      Partition targetPartition = getTarget().getKnowledgeModel().getPartition(
          targetPartitionID);
//      System.out.println("  -> target partition: "+targetPartition);
      if (originPartition != null && targetPartition != null)
      {
        SubModel difModel = new SubModel();
        difModel.add(originPartition);
        difModel.remove(targetPartition);
        SubModel commonModel = new SubModel();
        commonModel.add(originPartition);
        commonModel.remove(difModel);
        return commonModel;
      }
    }
    return null;
  }


  /**
   * Checks if the entities are comparable. In order to be so, they must both
   * have the same comparator and also have a "value of" relation to a single
   * common entity.
   */
  public boolean isComparable()
  {
//    System.out.println("Checking if "+getOrigin()+" and "+getTarget()+" is comparable");

    if(notComparable)
      return false;


    Object originPartitionID = getOrigin().getEntityObject();
    Object targetPartitionID = getTarget().getEntityObject();
    if(originPartitionID != null && targetPartitionID != null)
    {
//      System.out.println("  -> origin: "+originPartitionID+" target: "+targetPartitionID);
      Partition originPartition = getOrigin().getKnowledgeModel().getPartition(
          originPartitionID);
//      System.out.println("  -> origin partition: "+originPartition);
      Partition targetPartition = getTarget().getKnowledgeModel().getPartition(
          targetPartitionID);
//      System.out.println("  -> target partition: "+targetPartition);
      if (originPartition != null && targetPartition != null) {
        return true;
      }
    }
    notComparable = true;
    return false;
  }

  public String shortExplanation()
  {
    if(isComparable())
    {
      if(isCompletelyMatched())
        return "The partitions overlap completely";

      if(isPartlyMatched())
        return "The partitions overlap on "+(int) Math.round(this.getActivationStrength()*100)+"%.";

      return "The values do not match.";
    }
    return "The values are not comparable.";
  }

  public String toString()
  {
    String ret = shortExplanation();
    if(isComparable())
    {
      ret += "\nOverlap model is:\n" + getOverlapModel().toString();
    }
    return ret;
  }

  public SubModel toSubModel()
  {
    SubModel subModel = super.toSubModel();
    subModel.add(getOrigin());
    subModel.add(getTarget());
    return subModel;
  }

}
