package jcreek.reasoning;

import jcreek.representation.*;
import jcreek.representation.cbr.*;
import java.util.Vector;
import java.util.Arrays;

/**
 * <p>Title: RetriveResult</p>
 * <p>Description: Runs the first step (activate) of the CBR Retrive process. When a
 * new instance of this class is created, it creates an instance of CaseComparison
 * between the input case (problem case) and the cases we want to compare this case
 * to.</p>
 * @author Frode Srmo 2002, based on version by Martha Drum Jhre 2000.
 * @version 1.0
 */

public class RetrieveResult extends CBRReasoningStep
{
  private final int DEBUG_LEVEL = 0;

  /**
   * The input case (the problem case description).
   */
  protected Case inputCase;

  /**
   * Vector with a list of all the CaseComparisons.
   */
  protected Vector caseComparisons = new Vector();
  /**
   * The explain() and focus() steps operate on a subset of the caseComparisons
   * vector. This variable keeps track of the percieved size for these.
   */
  protected int caseComparisonSize = 0;

  protected int lastExplainMax = 0;
  protected double lastExplainThreshold = 0;
  protected int lastFocusMax = 0;
  protected double lastFocusThreshold = 0;
  protected Entity[] findingList;

  /**
   * The ID for the Retrieve step.
   */
  public static final int RETRIEVE_STEP = 10;

  // Defaults

  /**
   * DEFAULT_EXPLAIN_MAX_CASES - the default value of how many cases are explained.
   * This value is used by the explain() method to call the explain(int maxExplain, int explainThreshold)
   * method as the value of maxExplain. This corresponds to the maximum number of
   * case-comparisons the explain() step is executed on.
   */
  public static int DEFAULT_EXPLAIN_MAX_CASES = 500;

  /**
   * DEFAULT_EXPLAIN_THRESHOLD - the default value of the activation strength
   * threshold used by the explain() method to call the explain(int maxExplain, int explainThreshold)
   * method, using this value as the value of explainThreshold. This corresponds to the minimum
   * strength a case-comparison must have from the activate() step for the explain()
   * step to be executed.
   */
  public static double DEFAULT_EXPLAIN_THRESHOLD = 0.00;

  /**
   * DEFAULT_FOCUS_MAX_CASES - the default value for how many case-comparisons are kept after
   * the focus() method is run.
   */
  public static int DEFAULT_FOCUS_MAX_CASES = 3;

  /**
   * DEFAULT_FOCUS_THRESHOLD - the default minimum strength a case-comparison must have
   * in order to be kept when the focus() method is run.
   */
  public static double DEFAULT_FOCUS_THRESHOLD = 0.00;


  /**
   * Creates a new RetriveResult, comparing the input case to the set of cases
   * given in the compares set.
   *
   * @param input The input case (the new problem case description).
   * @param compares The set of cases to compare the input case to.
   * @param auto If true, the activate(), explain() and focus() methods are automatically run with default values.
   */
  public RetrieveResult(Case input, Case[] compares, boolean auto)
  {
    super(input.getKnowledgeModel());
    inputCase = input;


    // Initializing the vector of cases.
    counterMax = compares.length;
    counter = 0;
    for(int i=compares.length-1;i>=0;i--)
    {
      counter++;
      if(DEBUG_LEVEL > 1)
        System.out.println("Checking case "+compares[i]+" has status "+compares[i].getStatus());
      if(!compares[i].getID().equals(input.getID()) && compares[i].getStatus() == Case.SOLVEDCASE)
      {
        EntityComparison comp = new CaseComparison(input, compares[i]);
        if(DEBUG_LEVEL > 1)
          System.out.println(" "+i+" -> "+comp);
        caseComparisons.addElement(comp);
      }
    }

    reset();

    if(auto)
    {
      activate();
      explain();
      focus();
    }
  }

  /**
   * Creates a RetriveResult comparing the input case to all the solved cases in the
   * casebase.
   *
   * @param input The input case (the new problem case description).
   * @param auto If true, the activate(), explain() and focus() methods are automatically run with default values.   */
  public RetrieveResult(Case input, boolean auto)
  {
    this(input, (Case[]) input.getKnowledgeModel().matchEntityType(new Case[0]), auto);
  }

  /**
   * Creates a RetrieveResult comparing the input case to all the solved cases in the
   * casebase, and automatically runs activate(), explain() and focus().
   */
  public RetrieveResult(Case input)
  {
    this(input, true);
  }

  /**
   * Resets the state of this RetrieveResult to the initial state.
   */
  public void reset()
  {
    setState(RETRIEVE_STEP+INITIAL_STATE);
    counter = 0;
    counterMax = 0;
  }


  /**
   * Updates the CaseComparisons if any changes are done in the knowledgemodel.
   */
  public synchronized void update()
  {
    int oldState = getState();
    fireReasoningUpdateStarted();
    reset();
    if(isFocused())
    {
      activate();
      explain(lastExplainMax, lastExplainThreshold);
      focus(lastFocusMax, lastFocusThreshold);
    }
    else if(isExplained())
    {
      activate();
      explain(lastExplainMax, lastExplainThreshold);
    }
    else if(isActivated())
    {
      activate();
    }
  }

  /**
   * Sorts the caseComparisons vector.
   */
  protected void sortComparisons()
  {
    CaseComparison[] obs = (CaseComparison[]) caseComparisons.toArray(new CaseComparison[0]);
    Arrays.sort(obs);
    caseComparisons = new Vector(Arrays.asList(obs));
  }

  /**
   * Runs the activate step of the Retrieve. This runs a computationally inexpensive
   * method to get a rough idea of what cases are more similar to the input case.
   */
  public void activate()
  {
    synchronized(this)
    {
      counterMax = caseComparisons.size();
      counter = 0;
      for (int i = caseComparisons.size() - 1; i >= 0; i--)
      {
        if(isAborted())
          return;

        counter++;
        CaseComparison comp = (CaseComparison) caseComparisons.elementAt(i);
        if (comp != null) {
          comp.reset();
          comp.activate();
        }
        else
          caseComparisons.removeElementAt(i);
      }
      sortComparisons();
      setState(RETRIEVE_STEP + ACTIVATED_STATE);
      counterMax = 0;
      counter = 0;
    }
  }

  /**
   * Runs the activate step of the Retrieve. This runs a computationally inexpensive
   * method to get a rough idea of what cases are more similar to the input case.
   *
   * @param maxExplain The maximum number of case-comparisons to run the explain step on.
   * The case-comparisons with the highest strengths from the activate step is preferred.
   */
  public void explain(int maxExplain, double explainThreshold)
  {
    synchronized(this)
    {
      int i;
      counterMax = Math.min(size(), maxExplain);
      counter = 0;
      for (i = 0; i < caseComparisons.size(); i++)
      {
        if(isAborted())
          return;

        CaseComparison comp = (CaseComparison) caseComparisons.elementAt(i);
        if (comp.getStrength() > explainThreshold && i < maxExplain)
        {
          comp.explain();
          counter++;
        }
        else
        {
          break;
        }
      }
      caseComparisonSize = i;
      sortComparisons();
      setState(RETRIEVE_STEP + EXPLAINED_STATE);
      lastExplainMax = maxExplain;
      lastExplainThreshold = explainThreshold;
      counterMax = 0;
      counter = 0;
    }
  }

  /**
   * Runs the explain step using default values for max cases and threshold.
   */
  public void explain()
  {
    explain(DEFAULT_EXPLAIN_MAX_CASES, DEFAULT_EXPLAIN_THRESHOLD);
  }

  /**
   * Shrinks the number of case-comparisons.
   *
   * @param noCases Max number of case-comparisons to keep.
   * @param threshold All kept case-comparisons must have a strength of at least this value.
   */
  public synchronized void focus(int noCases, double threshold)
  {
    synchronized(this)
    {
      int i;
      counterMax = size();
      counter = 0;
      for (i = 0; i < size(); i++) {
        if (isAborted())
          return;

        counter++;
        CaseComparison comp = (CaseComparison) caseComparisons.elementAt(i);
        if (comp.getStrength() < threshold || i > noCases) {
//        System.out.println("i: "+i+" noCases: "+noCases);
          break;
        }
      }
      caseComparisonSize = i - 1;
      setState(RETRIEVE_STEP + FOCUSED_STATE);
      lastFocusMax = noCases;
      lastFocusThreshold = threshold;
      counterMax = 0;
      counter = 0;
    }
  }

  /**
   * Shrinks the number of case-comparisons, using default values for the maximum
   * number of cases to keep and for case-comparison threshold strength.
   */
  public void focus()
  {
    focus(DEFAULT_FOCUS_MAX_CASES, DEFAULT_FOCUS_THRESHOLD);
  }

  /**
   * Gives the set of CaseComparisons for the current reasoning step. After the
   * explain and focus steps, this set may be smaller than the initial set.
   */
  public CaseComparison[] getComparisons()
  {
    return (CaseComparison[]) caseComparisons.subList(0, size()-1).toArray(new CaseComparison[0]);
  }

  /**
   * Gives all case-comparisons, including those comparisons which has not been
   * through the explain and focus steps.
   */
  public CaseComparison[] getAllComparisons()
  {
    return (CaseComparison[]) caseComparisons.toArray(new CaseComparison[0]);
  }

  /**
   * Gives the CaseComparison object that represents the best match with a
   * solved case.
   */
  public CaseComparison getBestComparison()
  {
    if(size() > 0)
      return (CaseComparison) caseComparisons.elementAt(0);
    else
      return null;
  }

  /**
   * Gives the CaseComparison at the place specified.
   */
  public CaseComparison getComparison(int i)
  {
    if(i < caseComparisons.size())
      return (CaseComparison) caseComparisons.elementAt(i);
    else
      return null;
  }

  /**
   * Returns the input case (problem case).
   */
  public Case getInputCase()
  {
    return inputCase;
  }

  /**
   * Returns the case that is the best match for the input case.
   */
  public Case getBestCase()
  {
    return (Case) getBestComparison().getTarget();
  }

  /**
   * Returns the number of case-comparisons at this step in the Retrieve process.
   */
  public int size()
  {
    if(caseComparisonSize != 0)
      return caseComparisonSize;
    else
      return caseComparisons.size();
  }

  /**
   * Returns the number of case-comparisons created originally - i.e. before
   * the numbers are reduced by explain() and focus().
   */
  public int realsize()
  {
    return caseComparisons.size();
  }

  /**
   * Displays the RetrieveResult as a string.
   */
  public String toString()
  {
    String ret = "RetrieveResult for "+getInputCase()+".\n";
    if(isFocused())
      ret += "Status is focused.\n";
    else if(isExplained())
          ret += "Status is explained.\n";
    else if(isActivated())
      ret += "Status is activated.\n";
    ret+="Matched to: \n";
    for(int i=0;i<size();i++)
      ret+="\t"+getComparison(i).getTarget()+" strength "+getComparison(i).getStrength()+"\n";
    return ret;
  }
}
