package jcreek.reasoning;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Comparator;import java.util.Iterator;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import jcreek.representation.cbr.*;

/**
 * <p>Title: ReuseResult</p>
 * <p>Description: Uses the cases retrieved by a RetrieveResult object to select a solultion.
 * This class uses the kNN algorithm, using the distance as votes for k nearest neighbours.</p>
 * @author Frode Srmo
 * @version 1.0
 */

public class ReuseResult extends CBRReasoningStep
{
  protected RetrieveResult retrieve;
  protected int k = -1;
  protected boolean defaultK = false;

  protected Hashtable solutionStrength = new Hashtable();
  protected Entity[] solutions;

  /**
   * The ID for the Reuse step.
   */
  public static final int REUSE_STEP = 20;


  /**
   * Creates a new ReuseResult object, using k first cases of the RetrieveResult
   * to vote on a solution.
   *
   * @param retrieve The retrieved set of cases to reuse.
   * @param k How many of the best retrieved cases to use when selecting a solution.
   * @param automatic If true, activate, explain and focus is automatically run when constructed.
   */
  public ReuseResult(RetrieveResult retrieve, int k, boolean automatic)
  {
    super(retrieve.getKnowledgeModel());
    this.retrieve = retrieve;
    this.k = k;
    reset();
    if(automatic)
    {
      activate();
      explain();
      focus();

    }
  }


  /**
   * Creates a new ReuseResult object, using k first cases of the RetrieveResult
   * to vote on a solution.
   *
   * This constructor automatically calls activate, explain and focus.
   *
   * @param retrieve The retrieved set of cases to reuse.
   * @param k How many of the best retrieved cases to use when selecting a solution.
   */

  public ReuseResult(RetrieveResult retrieve, int k)
  {
    this(retrieve, k, true);
  }


  /**
   * Creates a new ReuseResult object, using the the focused cases of the RetrieveResult
   * to vote on a solution.
   *
   * This constructor automatically calls activate, explain and focus.
   */
  public ReuseResult(RetrieveResult retrieve)
  {
    this(retrieve, retrieve.size());
    defaultK = true;
  }

  /**
   * Returns the RetrieveResult object used as a prerequisite for this ReuseResult
   * object.
   */
  public RetrieveResult getRetrieveResult()
  {
    return retrieve;
  }

  /**
   * Returns how many of the cases are used to vote on the solution.
   */
  public int getK()
  {
    if(defaultK)
      return retrieve.realsize();
    else
      return Math.min(k, retrieve.realsize());
  }

  /**
   * Returns the input case (problem description).
   */
  public Case getInputCase()
  {
    return retrieve.getInputCase();
  }

  /**
   * Returns the matched cases used to vote on the solution.
   */
  public Case getCase(int i)
  {
    return (Case) retrieve.getComparison(i).getTarget();
  }

  public double getCaseStrength(int i)
  {
    return retrieve.getComparison(i).getStrength();
  }

  /**
   * Resets to the initial state.
   */
  public void reset()
  {
    solutions = new Entity[0];
    solutionStrength = new Hashtable();
  }

  /**
   * Activate: Retrieves the solution of the k best cases and vote
   */
  public void activate()
  {
    synchronized(this)
    {
//      System.out.println("Retrieve size: "+retrieve.realsize()+" k: "+k+" getK "+getK());
      Vector sols = new Vector();
      double totalVote = 0.0;
      for (int i = 0; i < getK(); i++)
      {
        if(isAborted())
          return;

        Case thisCase = getCase(i);
//        System.out.println("Case "+i+" is "+thisCase.toString());
        try {
          Entity solution = thisCase.getRelation("has solution").getValue();
//        System.out.println(" ---> solution: "+solution);
          if (solution != null)
          {
            double vote;
            if (!solutionStrength.containsKey(solution))
            {
              vote = 0.0;
              sols.addElement(solution);
            }
            else
            {
              vote = ( (Double) solutionStrength.get(solution)).doubleValue();
            }
            vote += getCaseStrength(i);
            totalVote += getCaseStrength(i);
            solutionStrength.put(solution, new Double(vote));
          }
        }
        catch (NoSuchRelationException e){
        }
        catch (NoSuchRelationTypeException e)
        {
        }
      }

      // Normalizes the solution strength so the sum of all strengths are 1.
      for (Iterator it = solutionStrength.keySet().iterator(); it.hasNext(); )
      {
        Object key = it.next();
        solutionStrength.put(key, new Double(
            ( (Double) solutionStrength.get(key)).doubleValue() / totalVote
            ));
      }

      solutions = (Entity[]) sols.toArray(new Entity[0]);
      Arrays.sort(solutions, new Comparator() {
        public int compare(Object a, Object b) {
          Double d = new Double(getSolutionStrength( (Entity) b));
          return d.compareTo(new Double(getSolutionStrength( (Entity) a)));
        }
      });

      setState(REUSE_STEP + ACTIVATED_STATE);
    }
  }

  /**
   * Explain step: Currently does nothing.
   */
  public void explain()
  {
    setState(REUSE_STEP+EXPLAINED_STATE);
  }

  /**
   * Focus step: Currently does nothing.
   */
  public void focus()
  {
    setState(REUSE_STEP+FOCUSED_STATE);
  }

  /**
   * Updates the solutions (e.g. if any changes are done in the knowledgemodel).
   */
  public synchronized void update()
  {
    retrieve.update();
    super.update();
  }

  /**
   * Returns the cases supporting the given solution.
   *
   * @param solutionNumber The solution number
   * @return Case[]
   */
  public Case[] getCasesForSolution(int solutionNumber)
  {
    Entity solution = getSolution(solutionNumber);
    ArrayList cases = new ArrayList(getK());
    for(int i=0;i<getK();i++)
    {
      Case c = getCase(i);
      if(solution.equals(c.getSolution()))
      {
        cases.add(c);
      }
    }
    return (Case[]) cases.toArray(new Case[cases.size()]);
  }

  /**
   * Returns the solution selected.
   *
   * @return An entity representing the solution.
   */
  public Entity getSolution()
  {
    return getSolution(0);
  }

  /**
   * Returns the i best solution found.
   *
   * @param i The number solution (i.e. no 1 is the best, 2 is second best etc.)
   * @return The entity representing the solution.
   */
  public Entity getSolution(int i)
  {
    if(solutions != null && solutions.length > i)
      return solutions[i];
    else
      return null;
  }

  /**
   * Returns the (non-normalized) strength of the votes solution no i have recieved.
   *
   * @param i The ith solution (from the best to the worst).
   * @return The non-normalized votes recieved by this solution.

   */
  public double getSolutionStrength(int i)
  {
    Entity ent = getSolution(i);
    return getSolutionStrength(ent);
  }

  /**
   * Returns the (non-normalized) strength of the votes solution ent have recieved.
   *
   * @param ent The Entity representing the solution.
   * @return The non-normalized votes recieved by this solution.
   */
  public double getSolutionStrength(Entity ent)
  {
    if(ent != null)
      return ((Double) solutionStrength.get(ent)).doubleValue();
    else
      return 0.0;
  }

  public int size()
  {
    if(solutions != null)
      return solutions.length;
    else
      return 0;
  }

  /**
   * Returns a SubModel that contains all findings of the input case and all
   * relations that is included in an explanation path for any inherited causal
   * relation to the solution, as well as the solution itself.
   *
   * @param solutionNumber The solution number to provide a causal model for
   * @return SubModel
   */
  public SubModel getCausalModel(int solutionNumber)
  {
    Entity solution = getSolution(solutionNumber);
    SubModel causalModel = new SubModel();

    causalModel.add(solution);
    Case input = getInputCase();
    causalModel.add(input);

    Relation[] findings = input.getRelations( km.getRelationType(CBRModel.FINDING_RELATION) );
    for(int j=0;j<findings.length;j++)
    {
      causalModel.add(findings[j].getValue());
      causalModel.add(findings[j]);
      Relation[] causalRelations = findings[j].getValue().getRelations(new RelationType[] {km.
          getRelationType(CBRModel.CAUSAL_RELATION),
          km.getRelationType(CBRModel.INVERSE_CAUSAL_RELATION)}
          );
      for (int i = 0; i < causalRelations.length; i++)
      {
        if (causalRelations[i].getValue().equals(solution))
        {
          causalModel.addExpanded(causalRelations[i]);
        }
      }
    }
    return causalModel;
  }

  /**
   * Returns a string saying what solution is chosen.
   */
  public String toString()
  {
    String ret;
    if(size() > 0)
     ret = "Solution for "+getInputCase().toString()+" is "+getSolution().toString()+".";
    else
     ret = "No solution found for "+getInputCase().toString()+".";

    for(int i=0;i<size();i++)
      ret += "\n\t"+getSolution(i).toString()+" has strength "+getSolutionStrength(i)+".";
    return ret;
  }
}
