package jcreek.reasoning;

/**
* StringComparison compares two entities containing strings - i.e. where
* the entities has a data object (as returned by getEntityObject()) which are both
* subclasses of the java.lang.String class.
* <p>
* The explain step contain no further matching.
* </p>
* @author Elise Bakke
* @version 1.0
*/

import jcreek.representation.*; import jcreek.representation.cbr.*;
import java.util.ArrayList;
import java.util.*;

public class StringComparison extends EntityComparison
{

  protected Entity valueType = null;
  protected ArrayList relList = null;
  protected boolean notComparable = false;

  private double similarwords = -1.0;
  private double numberofwords = -1.0;

  /**
   * Creates a new StringComparison, comparing the origin and target entities
   * given as parameters.
   */

  public StringComparison(Entity origin, Entity target)
  {
    super(origin, target);
//    System.out.println("Creating new StringComparison between "+origin+" and "+target);
  }

  /**
   * Nothing is done in the explanation step.
   *
   * @return Always 0.0 .
   */
  public double updateExplanationStrength()
  {
    return 0.0;
  }

  public double updateActivationStrength()
  {
    if(isComparable() &&
       getOrigin().getEntityObject() != null &&
       String.class.isAssignableFrom(getOrigin().getEntityObject().getClass())&&
       getTarget().getEntityObject() != null &&
       String.class.isAssignableFrom(getTarget().getEntityObject().getClass()))
    {
      String originString = (String) getOrigin().getEntityObject();
      String targetString = (String) getTarget().getEntityObject();

      StringTokenizer st1 = new StringTokenizer(originString);
      StringTokenizer st2 = new StringTokenizer(targetString);

      similarwords=0;
      numberofwords = st2.countTokens();
      ArrayList a1 = new ArrayList();

      while (st1.hasMoreTokens()){
        a1.add(st1.nextToken());
      }
//compares string2 to string1 (in arraylist a1).
      while (st2.hasMoreTokens()){
        String current = st2.nextToken();
        if (a1.contains(current)){
          similarwords++;
        }
      }
//similarity metric for string2 compared to string1.
//based on the number of similar words
// in ratio to the total number of words in st2

      double metric = (similarwords/numberofwords);

      return metric;

    }
    return 0.0;
  }

  /**
   * Checks if the entities are comparable. In order to be so, they must both
   * have the same comparator and also have a "value of" relation to a single
   * common entity.
   */
  public boolean isComparable()
  {
    if(notComparable)
      return false;
    if(valueType != null)
      return true;

    if(getOrigin().equals(getOrigin().getKnowledgeModel().getEntity(CBRModel.NO_VALUE)) ||
       getTarget().equals(getTarget().getKnowledgeModel().getEntity(CBRModel.NO_VALUE)))
      return false;

    relList = new ArrayList();
    Relation[] originValueOf = getOrigin().getLocalRelations(new String[] {"value of", "instance of", "subclass of"});
    Relation[] targetValueOf = getTarget().getLocalRelations(new String[] {"value of", "instance of", "subclass of"});

    for(int i=originValueOf.length-1;i>=0;i--)
    {
      for(int j=targetValueOf.length-1;j>=0;j--)
      {
        if(originValueOf[i].getValue().equals(targetValueOf[j].getValue()))
        {
          relList.add(originValueOf[i]);
          relList.add(targetValueOf[j]);
          valueType = originValueOf[i].getValue();
//          System.out.println("Found value type - it is "+valueType);
          return true;
        }
      }
    }
    notComparable = true;
    return false;
  }

  public String shortExplanation()
  {
    if(!isComparable() || valueType != null)
    {
      if(isCompletelyMatched())
        return "The values of "+valueType.getValueString()+" are the same.";
      else if(isPartlyMatched())
        return "The numbers "+getOrigin().getValueString()+" and "+getTarget().getValueString()+" are "+
        (int) Math.round(getStrength()*100)+"% similar";
      return "The values do not match.";
    }
    return "The values are not comparable.";
  }

  public SubModel toSubModel()
  {
    SubModel subModel = super.toSubModel();
    if(valueType != null)
    {
      subModel.add(valueType);
      subModel.add((Relation[]) relList.toArray(new Relation[relList.size()]));
    }
    return subModel;
  }
}
