package jcreek.reasoning;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

/**
 * <P>Defines a comparison between two symbols. This is the default comparator
 * of entities in a domain.
 *
 * @author Frode Srmo of NTNU 2002
 * @version 1.0
 */

public class SymbolComparison extends EntityComparison
{

  protected Vector symbolConvergenceList = new Vector();

  protected Entity valueType;
  protected boolean notComparable = false;


  /**
   * Creates a new comparison between the origin and target entities.
   *
   * @param origin the entity to compare from
   * @param target the entity to compare to
   */
  public SymbolComparison(Entity origin, Entity target)
  {
    super(origin, target);
  }

  public double updateActivationStrength()
  {
//    System.out.print("   SC: "+getTarget()+" and "+getOrigin()+" equals? ");
/*    if(getOrigin() != null && getTarget() != null && getOrigin().equals(getTarget()))
      System.out.println("yes");
    else
      System.out.println("no"); */
    if(getOrigin() != null && getTarget() != null && getOrigin().equals(getTarget()))
      return 1.0;
    else
      return 0.0;
  }

  /**
   * Returns the explanation strength of leaf findings.
   */
  protected double updateExplanationStrength()
  {
//    if(isComparable())
//    {
//      System.out.println("Testing explanation symbol comparison between "+getOrigin()+" and "+getTarget()+":");

      double strength = 0.0;

      RelationType causalRCD =
       getOrigin().getKnowledgeModel().getRelationType(CBRModel.CAUSAL_RELATION);

      //get the values of the findings
      Entity origin = (Entity)getOrigin();
      Entity target = (Entity)getTarget();

      // Temp list of possible convergence points, indexed by the convergence
      // entities themselves.
      Hashtable findConvIndex;

      symbolConvergenceList = new Vector();

      findConvIndex =  findConvergenceWithRelations( new RelationType[] { causalRCD } );
      findConvIndex.putAll(findConvergenceWithRelations( new RelationType[] { causalRCD.getInverseType() } ));


      // Going through the index to evaluate all the potential FindingConvergence
      // instances found so far. This is neccessary as there may be some
      // instances that have origin relations but no target relations.
      for(Enumeration pointEnum = findConvIndex.keys();pointEnum.hasMoreElements();)
      {
        Entity point = (Entity) pointEnum.nextElement();
        SymbolConvergence fc = (SymbolConvergence) findConvIndex.get(point);
        if(fc.hasConvergence())
        {
//          System.out.println("\tConvergence in point "+fc.getConvergencePoint()+" with strength "+fc.getStrength()+"\n"+fc.toString("\t\t"));
          symbolConvergenceList.addElement(fc);
        }//if convergence
//        else
//        {
//          System.out.println("\tNot convergence in point " + fc.getConvergencePoint());
//        }
      }//for convergence index

      if(symbolConvergenceList.size() > 0)
      {
        strength = 0.0;
        // Going through the FindingConvergene instances for this pair of findings
        // and adding their
        for(int i=symbolConvergenceList.size()-1;i>=0;i--)
        {
          SymbolConvergence fc = (SymbolConvergence) symbolConvergenceList.elementAt(i);
          strength = addParallell(strength, fc.getStrength());
        }//for finding convergences
//        System.out.println("\tTotal strength is : "+strength);
//        strength = strength / symbolConvergenceList.size();
      }//if finding convergences

    return strength;
//    }
//    else
//      return 0.0;
  }//getLeafExplanationStrength

  private Hashtable findConvergenceWithRelations(RelationType[] relTypes)
  {
    Entity origin = getOrigin();
    Entity target = getTarget();

    Hashtable findConvIndex = new Hashtable();

    // Adding the origin entity itself as a possible point.
    findConvIndex.put(origin, new SymbolConvergence(origin, target, origin));

    // Finding the relations from the origin finding.
    Relation[] originRels = origin.getRelations( relTypes );

//    System.out.println("\tOrigin rels: ");
    // Going through the relations from the origin finding.
    for(int i=originRels.length-1;i>=0;i--)
    {

      if(originRels[i].getExplanationStrength() > 0.0)
      {
//      System.out.println("\t\t"+originRels[i]);
        Entity point = (Entity) originRels[i].getValue();
        SymbolConvergence fc;

        // Checks if there already exists a FindingConvergence for the point.
        if (findConvIndex.containsKey(point))
          fc = (SymbolConvergence) findConvIndex.get(point);
        else {
          // The point does not already exist we must create it.
          fc = new SymbolConvergence(getOrigin(), getTarget(), point);
          findConvIndex.put(point, fc);
        }

        // Adding this particular relations to the list of relations going to
        // that point.
        fc.addOriginRelation(originRels[i]);
      }
    }

    // Fidning the relations from the target finding.
    Relation[] targetRels = target.getRelations( relTypes);

//    System.out.println("\tTarget rels:");

    // Going through the relations from the target finding.
    for(int i=targetRels.length-1;i>=0;i--)
    {
      if(targetRels[i].getExplanationStrength() > 0.0)
      {
//      System.out.println("\t\t"+targetRels[i]);
        Entity point = (Entity) targetRels[i].getValue();
        SymbolConvergence fc;

        // Checks if there already exists a FindingConvergence for the point.
        // If not there is no need to create it as that means there is no
        // relation between the origin entity to that point.
        if (findConvIndex.containsKey(point)) {
          fc = (SymbolConvergence) findConvIndex.get(point);
          // Adding this particular relations to the list of relations going to
          // that point.
          fc.addTargetRelation(targetRels[i]);
        } //if
      }
    }//for target relations

    return findConvIndex;
  }

  /**
   * Returning the list of FindingConvergence instances found for this FindingComparison.
   */

  public SymbolConvergence[] getSymbolConvergences()
  {
    return (SymbolConvergence[]) symbolConvergenceList.toArray(new SymbolConvergence[0]);
  }

//-----------------------------------------------------------------------------

  private String capitalize(String str)
  {
    if(str.length() > 1)
      return str.substring(0, 1).toUpperCase()+str.substring(1);
    return str;
  }

  public String shortExplanation()
  {
//    return shortExplanation("");

    if(isCompletelyMatched())
      return getTarget().getName()+" and "+getOrigin().getName()+" are the same.";
    else if(isPartlyMatched())
    {
      SymbolConvergence strongestConvergence = null;
      for (int i=symbolConvergenceList.size()-1;i>=0;i--)
      {
        SymbolConvergence currentConvergence = (SymbolConvergence) symbolConvergenceList.elementAt(i);
        if(strongestConvergence == null || currentConvergence.getStrength() > strongestConvergence.getStrength())
        {
          strongestConvergence = currentConvergence;
        }
      }
      if(strongestConvergence != null)
      {
        return capitalize(getTarget().getName()+" and "+getOrigin().getName()+" match partially because "+strongestConvergence.strongestExplanation());
      }

      return "No idea";
      /*
      String ret = getTarget().getName()+" and "+getOrigin().getName()+" match partially because both implicate ";
      for(int i=symbolConvergenceList.size()-1;i>=0;i--)
      {
        ret += ((SymbolConvergence) symbolConvergenceList.elementAt(i)).getConvergencePoint().getValueString();
        if(i > 1)
          ret += ", ";
        else if(i == 1)
          ret += " and ";
        else
          ret +=".";
      }
      return ret; */
    }
    else
      return getTarget().getName()+" and "+getOrigin().getName()+" do not match.";

  }

  /**
   * Returns a short explanation.
   */
  public String shortExplanation(String buf)
  {
    String str = "";
    if(isCompletelyMatched())
    {
      str += "\n"+buf+getTarget().getName()+" and "+
        getOrigin().getName()+" are the same.";
    }
    else if(isPartlyMatched())
    {
/*
      SymbolConvergence strongest;

      */


      str += "\n"+buf+getTarget().getName()+" and "+
        getOrigin().getName()+" match partially because: ";
      str += "(str: "+getStrength()+" act: "+getActivationStrength()+" exp: "+getExplanationStrength()+")";
      for(int i=symbolConvergenceList.size()-1;i>=0;i--)
      {
        str += "\n"+buf+"\t"+((SymbolConvergence) symbolConvergenceList.elementAt(i)).shortExplanation(buf+"\t");
      }

    }
    else
    {
      str = "\n"+buf+getTarget().getName()+" and "+
        getOrigin().getName()+" do not match.";
    }
    return str;
  }//shortExplanation

//------------------------------------------------------------------------------


  /**
   * Checks if the entities are comparable. In order to be so, they must both
   * have the same comparator and also have a "value of" relation to a single
   * common entity.
   */
  public boolean isComparable()
  {
    if(notComparable)
      return false;


    notComparable = !isComparable(getOrigin(), getTarget());
    return !notComparable;
  }

  public static boolean isComparable(Entity e1, Entity e2)
  {
    Entity valueType;
//    return true;
//    System.out.println("Attempting to find (symbol) valuetype of "+e1+" and "+e2);

    if(e1.equals(e1.getKnowledgeModel().getEntity(CBRModel.NO_VALUE)) ||
       e2.equals(e1.getKnowledgeModel().getEntity(CBRModel.NO_VALUE)))
      return false;

    if(e1.equals(e2))
      return true;

/*
    Relation[] originValueOf = e1.getLocalRelations(new String[] {"value of", "instance of", "subclass of"});
    Relation[] targetValueOf = e2.getLocalRelations(new String[] {"value of", "instance of", "subclass of"});

    for(int i=originValueOf.length-1;i>=0;i--)
    {
      for(int j=targetValueOf.length-1;j>=0;j--)
      {
//        System.out.println("Attempted match: "+originValueOf[i]+" and "+targetValueOf[j]);
        if(originValueOf[i].getValue().equals(targetValueOf[j].getValue()))
        {
          valueType = originValueOf[i].getValue();
//          System.out.println("Found value type - it is "+valueType);
          return true;
        }
      }
    }
  */
    return true;
  }


  /**
   * Returns the string representation of the finding comparison
   *
   * @return the string representation of the finding comparison
   */
  public String toString()
  {
    String ret = "";
    for(int i=0;i<symbolConvergenceList.size();i++)
    {
      ret += ((SymbolConvergence) symbolConvergenceList.get(i)).toString()+"\n\n";
    }
    return ret;
  }//toString

  public SubModel toSubModel()
  {
    SubModel model = super.toSubModel();
    for(int i=0;i<symbolConvergenceList.size();i++)
    {
      SymbolConvergence conv = ((SymbolConvergence) symbolConvergenceList.elementAt(i));
      model.union(conv.toCompactSubModel());
    }
    return model;
  }

  public SubModel toCompleteSubModel()
  {
    SubModel model = super.toSubModel();
    for(int i=0;i<symbolConvergenceList.size();i++)
    {
      SymbolConvergence conv = ((SymbolConvergence) symbolConvergenceList.elementAt(i));
      model.union(conv.toSubModel());
    }
    return model;
  }

//-----------------------------------------------------------------------------

}
