package jcreek.reasoning;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import java.util.Vector;
import java.io.Serializable;

/**
 * Describes a common point in the causal or statistical model implicated by
 * findings of two or more cases. This point (actually an entity) may be implicated
 * through one or more relations from each finding to the convergence-entity.
 * This serves as an explanation of why the  findings may play the same role
 * in some way, and thus the FindingConvergence have a composite strength for
 * this explanation, as well as methods for getting natural language text
 * for this explanation.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class SymbolConvergence implements Serializable
{
  private Entity convergencePoint;

  private Entity origin;
  private Entity target;

  private double strength = -1.0;

  private Vector originRelations = new Vector();
  private Vector targetRelations = new Vector();

  /**
   * Constructor for creating a new FindingConvergence. Should normally only be
   * done by the FindingComparison class.
   * @param origin The has-finding relation between the origin case and the finding entity.
   * @param target The has-finding relation between the target case and the finding entity.
   * @param convergencePoint The entity implicated by both findings.
   */
  SymbolConvergence(Entity origin, Entity target, Entity convergencePoint)
  {
    this.convergencePoint = convergencePoint;
    this.origin = origin;
    this.target = target;

  }

  /**
   * Constructor for creating a FindingConvergence from an array of others.
   * This does not guarantee that the resulting FindingConvergence is singular.
   * All FindingConvergence instances used in this method are assumed to have a
   * common convergence point and target finding. If this is not the case, a
   * NotCompatibleException is thrown.
   */

  SymbolConvergence(SymbolConvergence[] fcs) throws NotCompatibleException
  {
    if(fcs.length > 0)
    {
      this.convergencePoint = fcs[0].convergencePoint;
      this.origin = fcs[0].origin;
      this.target = fcs[0].target;
      this.targetRelations = fcs[0].targetRelations;
      this.originRelations = fcs[0].originRelations;
      this.strength = -1;

      for(int i=fcs.length;i>=1;i--)
      {
        this.union(fcs[i]);
      }
    }
  }

  /**
   * Merges the given SymbolConvergence with this SymbolConvergence. They
   * are assumed to share a convergencePoint and target finding - if not,
   * a NotCompatibleException is thrown.
   *
   * @param fc A FindingConvergence instance merged with this instance. The
   * fc instance is not changed.
   * @return Reference to this FindingConvergence.
   */

  public SymbolConvergence union(SymbolConvergence fc) throws NotCompatibleException
  {
    if(getConvergencePoint() != fc.getConvergencePoint() ||
      getTarget() != fc.getTarget())
    {
      throw new NotCompatibleException(this, fc);
    }
    else
    {
      originRelations.addAll(fc.originRelations);
      targetRelations.addAll(fc.targetRelations);
      if(getOrigin() != fc.getOrigin())
        origin = null;
      strength = -1;
    }
    return this;
  }

  /**
   * Add a relation implicating the convergencePoint entity from the origin
   * case finding. Normally only done by FindingComparison.
   */
  void addOriginRelation(Relation r)
  {
    if(!originRelations.contains(r) && r.getExplanationStrength() > 0.0)
    {
      originRelations.addElement(r);
      strength = -1.0;
    }
  }

  /**
   * Add a relation implicating the convergencePoint entity from the target
   * case finding. Normally only done by FindingComparison.
   */
  void addTargetRelation(Relation r)
  {
    if(!targetRelations.contains(r) && r.getExplanationStrength() > 0.0)
    {
      targetRelations.addElement(r);
      strength = -1.0;
    }
  }

  /**
   * Returns true if this FeindingConvergence is singular, i.e. if it explains
   * from a single finding to a single finding.
   */

  public boolean isSingular()
  {
    return target != null && origin != null;
  }

  /**
   * True if the FindingConvergence has proofs that the convergencePoint is
   * implicated by both findings.
   *
   * This only makes sense if there are single target and origins, i.e. if
   * isSingular() is true.
   */
  public boolean hasConvergence()
  {
    return target != null && origin != null &&
            (!targetRelations.isEmpty() || target.equals(convergencePoint)) &&
            (!originRelations.isEmpty() || origin.equals(convergencePoint));
  }

  /**
   * Gives a list of all relations (composite or direct) from the origin
   * case finding to the convergence entity.
   */
  public Relation[] getOriginRelations()
  {
    return (Relation[]) originRelations.toArray(new Relation[0]);
  }

  /**
   * Gives a list of all relations (composite or direct) from the target
   * case finding to the convergence entity.
   */
  public Relation[] getTargetRelations()
  {
    return (Relation[]) targetRelations.toArray(new Relation[0]);
  }

  /**
   * Returns the entity that is the convergence point for this FindingConvergence.
   */
  public Entity getConvergencePoint()
  {
    return convergencePoint;
  }

  /**
   * Returns the has-finding relation from the origin case to the origin finding.
   */
  public Entity getOrigin()
  {
    return origin;
  }

  /**
   * Returns the has-finding relation from the target case to the target finding.
   */
  public Entity getTarget()
  {
    return target;
  }

  /**
   * Gives the strength this FindingConvergence offers to support that the origin
   * and target findings play equal roles.
   */
  public double getStrength()
  {
    if(strength == -1.0)
      calcStrength();
    return strength;
  }

  private void calcStrength()
  {
    double originStrength = 0.0;
    if(getOrigin().equals(getConvergencePoint()))
    {
      originStrength = 1.0;
    }
    else
    {
      Relation[] originRels = getOriginRelations();
      for(int i=originRels.length-1;i>=0;i--)
/*        if(originRels[i] == null)
          originStrength = 1.0;
      else */
        originStrength = EntityComparison.addParallell(originStrength, originRels[i].getExplanationStrength());
    }

    double targetStrength = 0.0;

    if(getTarget().equals(getConvergencePoint()))
    {
      targetStrength = 1.0;
    }
    else
    {
      Relation[] targetRels = getTargetRelations();
      for(int i=targetRels.length-1;i>=0;i--)
/*        if(targetRels[i] == null)
          targetStrength = 1.0;
      else */
         targetStrength = EntityComparison.addParallell(targetStrength, targetRels[i].getExplanationStrength());
    }

    strength = EntityComparison.addSerial(originStrength, targetStrength);
  }

  /**
   * Gives a short, natural language representation of this structure, of the
   * form "<origin finding name> implicates <convergence point name>, which is also
   * implicated by <target finding name>".
   */
  public String shortExplanation(String buf)
  {
    String res = "";
    Vector unique = new Vector();

    for(int i=originRelations.size()-1;i>=0;i--)
    {
      String fromName = ((Relation) originRelations.elementAt(i)).getFrom().getName();
      if(!unique.contains(fromName))
        unique.addElement(fromName);
    }
    for(int i=unique.size()-1;i>=0;i--)
      if(i > 1)
        res += (String) unique.elementAt(i)+", ";
      else if(i == 1)
        res += (String) unique.elementAt(i)+" and ";
      else
        res += (String) unique.elementAt(i);

    res += " implicates "+getConvergencePoint().getName();

    if(getTarget() != getConvergencePoint())
      res += " which is also implicated by "+getTarget().getName();

    return res;
  }

  /**
   * Gives the strongest explanation in support of this point.
   */
  public String strongestExplanation()
  {
    Relation originStrongest = null;
    for(int i=originRelations.size()-1;i>=0;i--)
    {
      Relation currentOrigin = (Relation) originRelations.get(i);
      if(originStrongest == null || currentOrigin.getExplanationStrength() > originStrongest.getExplanationStrength())
        originStrongest = currentOrigin;
    }
    Relation targetStrongest = null;
    for(int i=targetRelations.size()-1;i>=0;i--)
    {
      Relation currentTarget = (Relation) targetRelations.get(i);
      if(targetStrongest == null || currentTarget.getExplanationStrength() > targetStrongest.getExplanationStrength())
        targetStrongest = currentTarget;
    }

//    System.out.println("targetStrongest: "+targetStrongest+" originStrongest: "+originStrongest);


    if(getConvergencePoint().equals(getOrigin()) && targetStrongest != null)
      if(targetStrongest.isInherited())
        return targetStrongest.getExplanationModel().getStrongestExplanation(targetStrongest.getRelationType(), targetStrongest.getValue()).text()+
        ".  This means that "+getOrigin().getName()+" possibly "+targetStrongest.getNaturalName()+" "+getTarget()+".";
      else
        return getTarget()+" "+targetStrongest.getNaturalName()+" "+getOrigin().getName()+".";
    if(getConvergencePoint().equals(getTarget()) && originStrongest != null)
      if(originStrongest.isInherited())
        return originStrongest.getExplanationModel().getStrongestExplanation(originStrongest.getRelationType(), originStrongest.getValue()).text()+
        ".  This means that "+getTarget().getName()+" possibly "+originStrongest.getNaturalName()+" "+getOrigin()+".";
      else
        return getOrigin()+" "+originStrongest.getNaturalName()+" "+getTarget().getName();

    if(targetStrongest != null && originStrongest != null)
    {
      String ret = "";
      try
      {
        if (targetStrongest.isInherited())
          ret = targetStrongest.getExplanationModel()
              .getStrongestExplanation(
              targetStrongest.getRelationType(),
              targetStrongest.getValue())
              .text() + ", which means that ";
      }
      catch(NullPointerException e)
      {
        System.err.println("targetStrongest: "+targetStrongest);
        System.err.println("targetStrongest.getExplanationModel(): "+targetStrongest.getExplanationModel());
        System.err.println("targetStrongest.getRelationType(): "+targetStrongest.getRelationType());
        System.err.println("targetStrongest.getValue(): "+targetStrongest.getValue());
        System.err.println("strongestExplanation: " + targetStrongest.getExplanationModel()
              .getStrongestExplanation(
              targetStrongest.getRelationType(),
              targetStrongest.getValue()));
        System.err.println("text: "+targetStrongest.getExplanationModel()
              .getStrongestExplanation(
              targetStrongest.getRelationType(),
              targetStrongest.getValue())
              .text());

      }
      ret += getTarget()+(targetStrongest.isInherited()?" possibly":"")+" "+targetStrongest.getNaturalName()+" "+getConvergencePoint().getName()+".\n\n";
      ret = ret;

      String add = "";
      if(originStrongest.isInherited())
        add += originStrongest.getExplanationModel().
            getStrongestExplanation(originStrongest.getRelationType(), originStrongest.getValue())
              .text()+", which means that ";
      ret += capitalize(add+getOrigin().getName())+(originStrongest.isInherited()?" possibly":"")+" "+originStrongest.getNaturalName()+" "+getConvergencePoint().getName()+". ";
      return ret;
    }
    return "";
  }

  private String capitalize(String str)
  {
    if(str.length() > 1)
      return str.substring(0, 1).toUpperCase()+str.substring(1);
    return str;
  }

  public String toString()
  {
    return toString("");
  }

  public String toString(String pre)
  {
    String res = pre+"The "+getTarget().getName()+" and "+getOrigin().getName()+" has the convergence point "+getConvergencePoint().getName()+"\n";
    res+= pre+"Origin relations implicating: \n";
    for(int i=originRelations.size()-1;i>=0;i--)
      res += pre+"   "+originRelations.elementAt(i).toString()+"\n";

    res+= pre+"Origin target relations implicating: \n";
    for(int i=targetRelations.size()-1;i>=0;i--)
      res += "   "+targetRelations.elementAt(i).toString()+"\n";
    res += pre+"Strength of convergence point: "+getStrength();
    return res;
  }

  /**
   * Transforms this FindingComparison to a submodel.
   */
  public SubModel toSubModel()
  {
    SubModel sm = new SubModel();
    sm.add(getConvergencePoint());
    sm.add(getTarget());
    if(getOrigin() != null)
      sm.add(getOrigin());

    sm.addExpanded(getTargetRelations());
    sm.addExpanded(getOriginRelations());
    return sm;
  }

  /**
   * Transforms this FindingComparison to a compact submodel.
   */
  public SubModel toCompactSubModel()
  {
    SubModel sm = new SubModel();
    sm.add(getConvergencePoint());
    sm.add(getTarget());
    if(getOrigin() != null)
      sm.add(getOrigin());

    sm.add(getTargetRelations());
    sm.add(getOriginRelations());
    return sm;
  }
}
