package jcreek.reasoning;

/**
* URLComparison compares two entities containing urls - i.e. where
* the entities has a data object (as returned by getEntityObject()) which are both
* subclasses of the java.net.URL class.
*
* The explain step contain no further matching.
* </p>
* @author Elise Bakke
* @version 1.0
*/

import jcreek.representation.*; import jcreek.representation.cbr.*;
//import jcreek.util.LRUCache;
import java.util.ArrayList;
import java.util.*;
import java.net.URL;

public class URLComparison extends EntityComparison
{

  protected Entity valueType = null;
  protected ArrayList relList = null;
  protected boolean notComparable = false;

  private double score = -1.0;
  //1 point for similar protocoll, 3 for reference equivalent hosts, 3 for same port number on
  //host, 3 for same fragment of file
  private double fullscore= 10.0;

  /**
   * Creates a new URLComparison, comparing the origin and target entities
   * given as parameters.
   */

  public URLComparison (Entity origin, Entity target)
  {
    super(origin, target);
//    System.out.println("Creating new URLComparison between "+origin+" and "+target);
  }

  /**
   * Nothing is done in the explanation step.
   *
   * @return Always 0.0 .
   */
  public double updateExplanationStrength()
  {
    return 0.0;
  }

  public double updateActivationStrength()
  {
    if(isComparable() &&
       getOrigin().getEntityObject() != null &&
       URL.class.isAssignableFrom(getOrigin().getEntityObject().getClass())&&
       getTarget().getEntityObject() != null &&
       URL.class.isAssignableFrom(getTarget().getEntityObject().getClass()))
    {
      URL originURL = (URL) getOrigin().getEntityObject();
      URL targetURL = (URL) getTarget().getEntityObject();

      if (targetURL.equals(originURL)){
        return 1.0;
      }
else{
  score=0;
  //compare to see if the protocols are similar
  String originProtocol = originURL.getProtocol();
  String targetProtocol = targetURL.getProtocol();
  if(targetProtocol.equals(originProtocol)){
    score++;
  }
  //test to see if they have reference equivalent hosts
  //compares strings that contain IPv6 adresses
  String originHost = originURL.getHost();
  String targetHost = targetURL.getHost();
  if(targetHost.equals(originHost)){
    score=score+3;
    //test to see it the URLs use the same port number on host
    //no need to do that unless the hosts are the same
    //if they use the same port on different hosts it would not matter
    int originPort = originURL.getPort();
    int targetPort = targetURL.getPort();
    if(originPort!=-1 && originPort==targetPort){
      score = score +3;
    }
  }
  //test to see if the filename is the same for both URLs
  //the getFile() is the equvalent of the concatenation of getPath() and getQuery()
  String originFilename = originURL.getFile();
  String targetFilename = targetURL.getFile();
  if(targetFilename.equals(originFilename)){
    score = score + 3;
  }
//similarity metric for targetURL compared with originURL
  double metric = (score/fullscore);
  return metric;
}
}
    return 0.0;
  }

  /**
   * Checks if the entities are comparable. In order to be so, they must both
   * have the same comparator and also have a "value of" relation to a single
   * common entity.
   */
  public boolean isComparable()
  {
    if(notComparable)
      return false;
    if(valueType != null)
      return true;

    if(getOrigin().equals(getOrigin().getKnowledgeModel().getEntity(CBRModel.NO_VALUE)) ||
       getTarget().equals(getTarget().getKnowledgeModel().getEntity(CBRModel.NO_VALUE)))
      return false;

    relList = new ArrayList();
  //    System.out.println("Attempting to find valuetype of "+getOrigin()+" and "+getTarget());
  Relation[] originValueOf = getOrigin().getLocalRelations(new String[] {"value of", "instance of", "subclass of"});
  Relation[] targetValueOf = getTarget().getLocalRelations(new String[] {"value of", "instance of", "subclass of"});

  for(int i=originValueOf.length-1;i>=0;i--)
    {
      for(int j=targetValueOf.length-1;j>=0;j--)
      {
      //        System.out.println("Attempted match: "+originValueOf[i]+" and "+targetValueOf[j]);
      if(originValueOf[i].getValue().equals(targetValueOf[j].getValue()))
      {
          relList.add(originValueOf[i]);
          relList.add(targetValueOf[j]);
          valueType = originValueOf[i].getValue();
//          System.out.println("Found value type - it is "+valueType);
          return true;
        }
      }
    }
    notComparable = true;
    return false;
  }

public String shortExplanation()
{
  if(!isComparable() || valueType != null)
  {
    if(isCompletelyMatched())
      return "The urls "+valueType.getValueString()+" are the same.";
    else if(isPartlyMatched())
      return "The urls "+getOrigin().getValueString()+" and "+getTarget().getValueString()+" are "+
          (int) Math.round(getStrength()*100)+"% similar";
    return "The urls do not match.";
  }
  return "The urls are not comparable.";
}

  public SubModel toSubModel()
  {
    SubModel subModel = super.toSubModel();
    if(valueType != null)
    {
      subModel.add(valueType);
      subModel.add((Relation[]) relList.toArray(new Relation[relList.size()]));
    }
    return subModel;
  }
}

