package jcreek.representation;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Hashtable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import jcreek.util.LRUCache;

/**
 * An abstract implementation of the methods of the KnowledgeModel interface
 * that is not dependent on the mode of storage.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public abstract class AbstractKnowledgeModel implements KnowledgeModel
{
  protected transient ArrayList modelChangeListeners;

  /**
   * Creates a new KnowledgeModel by importing the contents of the supplied model.
   *
   * @param model The KnowledgeModel to import.
   */
  public AbstractKnowledgeModel(KnowledgeModel model)
  {
    importModel(model);
  }

  protected AbstractKnowledgeModel()
  {
    super();
  }

  /**
   * Removes any existing content from the KnowledgeModel and replaces it with
   * the contents of the current model. This is useful for transferring models
   * across different implemetnations of this interface, e.g. from a KnowledgeModel
   * storing the data locally and a KnowledgeModel storing the information in a
   * database.
   *
   * It will not transfer non-permanent data like change listeners.
   *
   * @param model The KnowledgeModel that will have its contents copied into this model.
   */
  public void importModel(KnowledgeModel model)
  {
    // Removes all the relations in the current model.
    Iterator relationIterator = getRelationIterator();
    while (relationIterator.hasNext())
      ( (Relation) relationIterator.next()).remove();

    // Removes all the entities in the current model.
    Iterator entityIterator = getEntityIterator();
    while (entityIterator.hasNext())
      ( (Entity) entityIterator.next()).remove();

    mergeModel(model);
  }

  /**
   * This method merges the contents of a given model into the current model.
   * Non-permanent data is not merged, i.e. listeners and such. The model that
   * is merged into this model is not changed.
   *
   * The merge will only contain one copy of each entity that has the same name
   * in both models. If there exists a relation between two entities in the
   * current model, another relation of the same relation type between the same
   * entities will not be imported even though they may contain the same entities.
   *
   * Partitions are imported if they have different names than the already existing
   * partitions. Already existing partitions will not be changed.
   *
   * @param model The knowledgemodel to merge into this model.
   */
  public void mergeModel(KnowledgeModel model)
  {
    // Imports all the entities in the given model.
    try
    {
      Iterator entityIterator = model.getEntityIterator();
      Hashtable entityOldToNewID = new Hashtable();

//      System.out.println("Importing entities.");
      while(entityIterator.hasNext())
      {
        Entity foreignEntity = (Entity) entityIterator.next();
        System.out.print("--> "+foreignEntity.getName());
        if(!hasEntity(foreignEntity.getName()))
        {
          Entity newEntity = new Entity(this, foreignEntity.getName(), foreignEntity.getDescription());
          EntityData newEntityData = newEntity.getEntityData();
          newEntityData.setCreationTime(foreignEntity.getCreationTime());
          newEntity.setEntityObject(foreignEntity.getEntityObject());
          entityOldToNewID.put(foreignEntity.getID(), newEntityData.getID());
//          System.out.println(" ... was not found, created");
        }
        else
        {
          entityOldToNewID.put(foreignEntity.getID(),
                               getEntity(foreignEntity.getName()).getID());
//          System.out.println(" ... was matched to "+getEntity(foreignEntity.getName()).getID());
        }
      }

//      System.out.println("\nImporting relations.");
      Iterator relationIterator = model.getRelationIterator();
      Hashtable relationOldToNewID = new Hashtable();
      while(relationIterator.hasNext())
      {
        Relation foreignRelation = (Relation) relationIterator.next();
//        System.out.print("\t--> "+foreignRelation.toString());
        try
        {
          // Trying to find an existing relation of the same type between the same
          // entities in the current model.
          Relation existingRelations[] = getEntity(entityOldToNewID.get(
              foreignRelation.getFrom().getID())).
              getLocalRelations(new RelationType(getEntity(entityOldToNewID.get(
              foreignRelation.getRelationType().getID()))));
          boolean exists = false;
          for (int i = 0; i < existingRelations.length && !exists; i++)
            if (existingRelations[i].getValue().getID().equals(entityOldToNewID.
                get(foreignRelation.getValue().getID())))
            {
              // A match to an existing relation was found. Using this instead of
              // creating a new one.
              exists = true;
              relationOldToNewID.put(foreignRelation.getID(), existingRelations[i].getID());
//              System.out.println(" ... was matched to existing relation "+existingRelations[i]);
            }
          if (!exists)
          {
            // No relation that match this merged relation was found in the current
            // model, so we add a new relation.
            Relation newRelation = new Relation(
                getEntity(entityOldToNewID.get(foreignRelation.getFrom().getID())),
                new RelationType(getEntity(entityOldToNewID.get(foreignRelation.getRelationType().getID()))),
                getEntity(entityOldToNewID.get(foreignRelation.getValue().getID())));
            newRelation.setExplanationStrength(foreignRelation.
                getExplanationStrength());

            newRelation.getRelationData().setCreationTime(foreignRelation.getCreationTime());
            relationOldToNewID.put(foreignRelation.getID(), newRelation.getID());
//            System.out.println(" ... was not found, created");
          }
        }
        catch(EntityTypeNotMatchedException e)
        {
          // This should not be possible...
          e.printStackTrace();
        }
      }

//      System.out.println("\nAdding overrides");
      // Another iteration through the relations to add the overrides.
      // This can not be done in the first iteration as this may refer relations
      // in the foreign model that is not yet created in the current model.
      relationIterator = model.getRelationIterator();
      while(relationIterator.hasNext())
      {
        Relation foreignRelation = (Relation) relationIterator.next();
        System.out.println("\t--> "+foreignRelation);
        Relation[] overrides = foreignRelation.getOverridingRelations();
        for (int i=0; i < overrides.length; i++)
          getRelation(relationOldToNewID.get(foreignRelation.getID())).
              addOverriding(getRelation(relationOldToNewID.get(overrides[i].getID())));
      }

      // Adding the partitions.
      System.out.println("\nAdding Partitions");
      Iterator partitionIterator = model.getPartitionIterator();
      while(partitionIterator.hasNext())
      {
        Partition foreignPartition = (Partition) partitionIterator.next();
        System.out.println("\t --> "+foreignPartition);
        if(!hasPartition(foreignPartition.getName()))
        {
          Partition newPartition = new Partition(this, foreignPartition.getName());
          newPartition.getPartitionData().setCreationTime(foreignPartition.getCreationTime());

          Entity foreignPartitionEntities[] = foreignPartition.getEntities();
          for (int i = 0; i < foreignPartitionEntities.length; i++)
            newPartition.add(
              getEntity(entityOldToNewID.get(foreignPartitionEntities[i].getID())));

          Relation foreignPartitionRelations[] = foreignPartition.getRelations();
          for (int i = 0; i < foreignPartitionRelations.length; i++)
            newPartition.add(
              getRelation(relationOldToNewID.get(foreignPartitionRelations[i].getID())));
        }
      }

      setName(model.getName());
      setDescription(model.getDescription());
      setAuthor(model.getAuthor());
      setTitle(model.getTitle());
//      System.out.println("Done.");
    }
    catch(NameAlreadyExistException e)
    {
      // This should never happen. If it does, we abort the import.
      e.printStackTrace(System.err);
    }
  }

  /**
   * Returns an array of all entities in the model that match a given EntityType.
   *
   * @param arr An array of the type of EntityType to find.
   */
  public EntityType[] matchEntityType(EntityType[] arr)
  {
//    System.out.println("matchEntity for : "+arr.getClass().getComponentType()+" with hash "+arr.getClass().getComponentType().hashCode());
    Object key[] = new Object[] {"_knowledgemodel_match_entity_type", arr.getClass().getComponentType() };
    if(getInferenceCache().isCached(key))
      return (EntityType[]) ((ArrayList) getInferenceCache().get(key)).toArray(arr);

    try
    {
      Class entityTypeClass = arr.getClass().getComponentType();
      Method matchMethod = entityTypeClass.getMethod("matches", new Class[] { Entity.class });
      Constructor entityTypeConstructor = entityTypeClass.getConstructor(new Class[] { Entity.class });
      ArrayList result = new ArrayList();

      Iterator entityIterator = getEntityIterator();

      while(entityIterator.hasNext())
      {
        Entity next = (Entity) entityIterator.next();
        try
        {
          if(((Boolean) matchMethod.invoke(null, new Object[] { next })).booleanValue())
          {
            if(next.getClass().equals(entityTypeClass))
              result.add(next);
            result.add(entityTypeConstructor.newInstance(new Object[] { next }));
          }
        }
        catch(IllegalAccessException e)
        {
          e.printStackTrace(System.err);
        }
        catch(IllegalArgumentException e)
        {
          e.printStackTrace(System.err);
        }
        catch(InvocationTargetException e)
        {
          e.printStackTrace(System. err);
        }
      }
      getInferenceCache().put(key, result);

      return (EntityType[]) result.toArray(arr);
    }
    catch(NoSuchMethodException e)
    {
      e.printStackTrace(System.err);
    }
    catch(InstantiationException e)
    {
      e.printStackTrace(System.err);
    }
    return new EntityType[0];
  }

  /**
   * Find a relation class in the model of the given name. If no relation type
   * is indexed under that name, null is returned.
   *
   * @param name The name of the relation class to be found
   * @return The RelationType object for this type
   */
  public RelationType getRelationType(String name)
  {
    Entity type = getEntity(name);
    if(type instanceof RelationType)
      return (RelationType) type;

    synchronized(this)
    {
      if(type != null)
      {
        try
        {
          RelationType typeRT = new RelationType(type);
          return typeRT;
//            entityCache.remove(type);
        }
        catch(EntityTypeNotMatchedException e)
        {
          // This should never happen.
          e.printStackTrace(System.err);
        }
      }
    }
    return null;
  }

  /**
   * Returns an array of all relations in the model. This may be prohibitive for
   * large models since it forces all relations to be stored in memory at once.
   */
  public Relation[] getRelations()
  {
    String key = "_abstract_knowledgemodel_get_relations";
    if(getInferenceCache().isCached(key))
      return (Relation[]) getInferenceCache().get(key);

    ArrayList list = new ArrayList();
    for(Iterator it = getRelationIterator();it.hasNext();)
    {
      list.add(it.next());
    }

    Relation[] ret = (Relation[]) list.toArray(new Relation[list.size()]);
    getInferenceCache().put(key, ret);
    return ret;
  }

  /**
   * Returns an array of all entities in the model. This may be prohibitive for
   * large models since it forces all entities to be stored in memory at once.
   */
  public Entity[] getEntities()
  {
    String key = "_abstract_knowledgemodel_get_entities";
    if(getInferenceCache().isCached(key))
      return (Entity[]) getInferenceCache().get(key);

    ArrayList list = new ArrayList();
    for(Iterator it = getEntityIterator();it.hasNext();)
    {
      list.add(it.next());
    }

    Entity[] ret = (Entity[]) list.toArray(new Entity[list.size()]);
    getInferenceCache().put(key, ret);
    return ret;
  }

  /**
   * Returns an array of all partitions in this KnowledgeModel.
   */
  public Partition[] getPartitions()
  {
    String key = "_abstract_knowledgemodel_get_partition";
    if(getInferenceCache().isCached(key))
      return (Partition[]) getInferenceCache().get(key);

    ArrayList list = new ArrayList();
    for(Iterator it = getPartitionIterator();it.hasNext();)
    {
      list.add(it.next());
    }

    Partition[] ret = (Partition[]) list.toArray(new Partition[list.size()]);
    getInferenceCache().put(key, ret);
    return ret;
  }

  /**
   * Registrers a listener that will get notified if there are any changes to
   * the model.
   *
   * @param listen The ModelChangeListener to add.
   */
  public void addModelChangeListener(ModelChangeListener listen)
  {
    synchronized(this)
    {
      if(modelChangeListeners == null)
        modelChangeListeners = new ArrayList();
      modelChangeListeners.add(listen);
    }
  }
  /**
   * Removes a ModelChangeListener listening to changes to this model.
   *
   * @param listen The ModelChangeListener to remove.
   */
  public void removeModelChangeListener(ModelChangeListener listen)
  {
    synchronized(this)
    {
      if(modelChangeListeners != null)
        modelChangeListeners.remove(listen);
    }
  }
}
