package jcreek.representation;

import jcreek.util.CreekException;

/**
 * The basic epistemological model is a barebones top-level model required
 * by the representation and inference structures. The assumptions ma
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class BasicModel implements EpistemologicalModel
{
  // BASIC STRUCTURE

  /** Top level concept. Everything is a subclass/instance of this. */
  public static final String THING = "_thing";

  /** Top level entity. Every concept that is not a relation type is a subclass/instanec of this. */
  public static final String ENTITY = "_entity";

  /** Top level relation type. All relation types are subclasses of this relation type. */
  public static final String RELATION = "_relation";

  /** The inverse relation type of RELATION. All the inverse relation types are structured under this. */
  public static final String INVERSE_RELATION ="_inverse relation";

  // REQUIRED ENTITY TYPES

  /** Entity type for signaling that a relation has a missing value, e.g. "car has colour no value". */
  public static final String NO_VALUE = "_no value";

  /** Entity type for entities representing numbers. */
  public static final String NUMBER = "_number";

  /** Entity type for entities representing symbols. */
  public static final String SYMBOL = "_symbol";

  // RELATION TYPES REQUIRED BY jcreek.representation.inference.SubclassInheritanceMethod

  /** Inverse of "has subclass". Organizes the taxonomy of the model. Core relation type for the jcreek.inference.SubclassInheritanceMethod inference. */
  public static final String SUBCLASS_OF = "_subclass of";

  /** Organizes the taxonomy of the model. Core relation type for the jcreek.representation.inference.SubclassInheritanceMethod inference. */
  public static final String HAS_SUBCLASS = "_has subclass";

  /** Relation between a data type and an instance (value) of that data type. */
  public static final String HAS_INSTANCE = "_has instance";

  /** Inverse of INSTANCE_OF. */
  public static final String INSTANCE_OF = "_instance of";

  // RELATION TYPES REQUIRED BY jcreek.representation.inference.PlausibleInheritanceMethod

  /** Used to describe plausible inheritance rules for use with jcreek.representation.inference.PlausibleInheritanceMethod. May only be used between relation types. */
  public static final String TRANSFERS = "_transfers";

  /** Inverse of the TRANSFERS relation type. */
  public static final String INHERITS_OVER = "_inherits over";

  // RELATION TYPES REQUIRED BY jcreek.representation.RelationType

  /** This relation represents the link between a relation type and its inverse type. This type is its own inverse. */
  public static final String HAS_INVERSE = "_has inverse";

  /** The default explanation strength for a relation type, i.e. the explanation strength used when creating a new relationship of this type. */
  public static final String HAS_DEFAULT_EXPLANATION_STRENGTH = "_default explanation strength";

  /** Inverse of HAS_DEFAULT_EXPLANATION_STRENGTH */
  public static final String DEFAULT_EXPLANATION_STRENGTH_OF ="_default explanation strength of";

  /** The value type of a relation type. See setValueType(). */
  public static final String HAS_VALUE_TYPE = "_has_value type";

  /** Inverse of HAS_VALUE_TYPE */
  public static final String VALUE_TYPE_OF = "_value type of";

  public void generateEpistemologicalModel(KnowledgeModel model)
  {
    generateBasicModel(model);
  }

  public static void generateBasicModel(KnowledgeModel model)
  {
    try
    {
      model.getInferenceCache().suspendFlush(true);

      // Starts by creating the very basic hierarchy.
      Entity thing = new Entity(model, "Thing", "");
      model.registerEntity(THING, thing);
      Entity entity = new Entity(model, "Entity", "");
      model.registerEntity(ENTITY, entity);


      // The "Relation", "Relation (inverse)", "has inverse" and "subclass of" relations are created
      // without using the RelationType normal constructor since they are used within the RelationType
      // class.
      RelationType relation = new RelationType(new Entity(model, "Relation", ""));
      model.registerEntity(RELATION, relation);
      RelationType inverseRelation = new RelationType(new Entity(model,
        "Relation (inverse)", ""));
      model.registerEntity(INVERSE_RELATION, inverseRelation);

      RelationType hasInverse = new RelationType(new Entity(model,
        "has inverse", ""));
      model.registerEntity(HAS_INVERSE, hasInverse);

      RelationType subclassOf = new RelationType(new Entity(model,
        "subclass of", ""));
      model.registerEntity(SUBCLASS_OF, subclassOf);
      RelationType hasSubclass = new RelationType(new Entity(model,
        "has subclass", ""));
      model.registerEntity(HAS_SUBCLASS, hasSubclass);

      // We are now at the point where it is safe to add relations that do not have an inverse relation.
      // The "has inverse" relation itself is the only example of this so far, but is also all we need
      // to establish inverse relation types for the other relations.
      relation.addRelation(HAS_INVERSE, inverseRelation, 1.0);
      subclassOf.addRelation(HAS_INVERSE, hasSubclass, 1.0);

      // At this point, we can use "subclass of".
      thing.addRelation(HAS_SUBCLASS, relation, 1.0);
      thing.addRelation(HAS_SUBCLASS, entity, 1.0);
      relation.addRelation(HAS_SUBCLASS, inverseRelation, 1.0);
      relation.addRelation(HAS_SUBCLASS, hasSubclass, 1.0);
      inverseRelation.addRelation(HAS_SUBCLASS, subclassOf, 1.0);
      relation.addRelation(HAS_SUBCLASS, hasInverse);
      inverseRelation.addRelation(HAS_SUBCLASS, hasInverse);


      // Adds the rest of the relation types used by the normal RelationType constructor.
      RelationType hasDefaultExplanationStrength = new RelationType(new Entity(
        model, "has default explanation strength", ""));
      RelationType defaultExplanationStrengthOf = new RelationType(new Entity(
        model, "default explanation strength of", ""));

      hasDefaultExplanationStrength.addRelation(HAS_INVERSE,
                                                defaultExplanationStrengthOf,
                                                1.0);

      model.registerEntity(HAS_DEFAULT_EXPLANATION_STRENGTH,
                           hasDefaultExplanationStrength);
      model.registerEntity(DEFAULT_EXPLANATION_STRENGTH_OF,
                           defaultExplanationStrengthOf);


      inverseRelation.addRelation(HAS_DEFAULT_EXPLANATION_STRENGTH,
                                  new DoubleEntity(model, 1.0, ""), 1.0);
      subclassOf.addRelation(HAS_DEFAULT_EXPLANATION_STRENGTH,
                             new DoubleEntity(model, 0.9, ""), 1.0);
      hasSubclass.addRelation(HAS_DEFAULT_EXPLANATION_STRENGTH,
                              new DoubleEntity(model, 0.9, ""), 1.0);
      hasDefaultExplanationStrength.addRelation(
        HAS_DEFAULT_EXPLANATION_STRENGTH, new DoubleEntity(model, 1.0, ""), 1.0);
      defaultExplanationStrengthOf.addRelation(HAS_DEFAULT_EXPLANATION_STRENGTH,
                                               new DoubleEntity(model, 1.0, ""),
                                               1.0);

      hasDefaultExplanationStrength.addRelation(SUBCLASS_OF, relation);
      defaultExplanationStrengthOf.addRelation(SUBCLASS_OF, inverseRelation);

      // At this point we have established enough of the model to use the RelationType constructor and
      // the addRelation() method in Entity normally.

      // The rest of the relation hierarchy

      RelationType hasInstance = new RelationType(model, relation,
        "has instance",
        "The target is an instance of the origin",
        "instance of",
        "The target is the class of the origin.", 0.9);
      RelationType instanceOf = hasInstance.getInverseType();
      model.registerEntity(HAS_INSTANCE, hasInstance);
      model.registerEntity(INSTANCE_OF, instanceOf);

      RelationType transfers = new RelationType(model, relation,
        "transfers",
        "The target relation-type is transferred over the origin relation-type.",
        "inherits over",
        "The target relation-type is inherited over the origin relation-type",
                                                1.0);
      model.registerEntity(TRANSFERS, transfers);
      model.registerEntity(INHERITS_OVER, transfers.getInverseType());

      RelationType hasValueType = new RelationType(model, relation,
        "has value type",
        "The origin relation-type can only have instances of the target entity as its value type, i.e all relations of this type should have an instance of this entity as its value.",
        "value type of ",
        "The target is a relation-type this entity is the value type for",
        1.0);
      model.registerEntity(HAS_VALUE_TYPE, hasValueType);
      model.registerEntity(VALUE_TYPE_OF, hasValueType.getInverseType());


      // Creating the entity hierarchy.

      Entity number = new Entity(model, "Number", "");
      number.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(NUMBER, number);

      Entity symbol = new Entity(model, "Symbol", "");
      symbol.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(SYMBOL, symbol);

      Entity noValue = new Entity(model, "No Value", "");
      noValue.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(NO_VALUE, noValue);

      // Plausible Inheritance Rules

      // Sublass of and instance of transfers all other relation types...

      Relation transferBySubclass = relation.addRelation(INHERITS_OVER,
        subclassOf);

      Relation transferByInstanceOf = relation.addRelation(INHERITS_OVER,
        instanceOf);

      // ...except themselves and their inverses.

      Relation r;

      r = subclassOf.addRelation(INHERITS_OVER, subclassOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferBySubclass);

      r = hasSubclass.addRelation(INHERITS_OVER, subclassOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferBySubclass);

      r = subclassOf.addRelation(INHERITS_OVER, instanceOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferByInstanceOf);

      r = hasSubclass.addRelation(INHERITS_OVER, instanceOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferByInstanceOf);

      r = instanceOf.addRelation(INHERITS_OVER, subclassOf);
      r.negate();
      r.getInverse().negate();

      r = hasInstance.addRelation(INHERITS_OVER, subclassOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferBySubclass);

      r = instanceOf.addRelation(INHERITS_OVER, instanceOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferByInstanceOf);

      r = hasInstance.addRelation(INHERITS_OVER, instanceOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferByInstanceOf);

      // ...also the "has inverse" and "inverse of" relations should not be inherited.

      r = hasInverse.addRelation(INHERITS_OVER, subclassOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferBySubclass);

      r = hasInverse.addRelation(INHERITS_OVER, instanceOf);
      r.negate();
      r.getInverse().negate();
      r.addOverriding(transferByInstanceOf);

    }
    catch (CreekException e)
    {
      e.printStackTrace();
    }
    model.getInferenceCache().suspendFlush(false);

  }
}
