package jcreek.representation;

import java.io.Serializable;
import java.util.*;

import jcreek.representation.inference.InheritanceMethod;
import jcreek.representation.inference.PlausibleInheritanceMethod;

import jcreek.representation.inference.*;

/**
 * The Entity Class encapsulates a node in the semantic network, and has references
 * to all relations from and to the entity.
 *
 * An Entity instance actually encapsulates the node, and there may be several
 * instances of Entity representing each node, like a pointer. This means that
 * an Entity can't be serialized directly (but a KnowledgeModel object may be),
 * and a clone created of an Entity will not make another node in the semantic
 * network but another pointer to the same node.
 *
 * @author Frode Srmo
 * @version 2.0
 */

public class Entity implements Comparable, Cloneable
{

  /**
   * Reference to the KnowledgeModel.
   */
  private transient KnowledgeModel knowledgeModel;

  /**
   * Reference to the EntityData that contains the data for the entity.
   */
  private transient EntityData entityData;

  /**
   * The current partition used as context in this entity.
   */
  private transient Partition currentPartition;

  /**
   * Creates a new Entity object that encapsulates the given EntityData object.
   *
   * This method of creating an entity should only be used by a KnowledgeModel
   * object that need to encapsulate an existing EntityData object.
   *
   * @param km The KnowledgeModel this Entity is a part of.
   * @param entityData The EntityData object to encapsulate.
   */
  Entity(KnowledgeModel km, EntityData entityData)
  {
    this.knowledgeModel = km;
    this.entityData = entityData;
  }

  /**
   * Create a new entity with the given name, description and creator object.
   *
   * @param knowledgeModel The KnowledgeModel this concept is a part of.
   * @param name The name of the concept. This name must be unique in the KnowledgeModel unless the
   * model is set to accept duplicates. If it is not, a ConceptAlreadyExistException is thrown.
   * @param description An (optional) description of the entity.
   */
  public Entity(KnowledgeModel knowledgeModel, String name, String description) throws NameAlreadyExistException
  {
    this.knowledgeModel = knowledgeModel;
    try
    {
      this.entityData = knowledgeModel.makeEntity();
      knowledgeModel.registerEntity(name, this);
      entityData.setName(name);
      entityData.setCreationTime(System.currentTimeMillis());
      setDescription(description);
      knowledgeModel.fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.ENTITY_ADDED));
    }
    catch(NameAlreadyExistException e)
    {
      // Removes this entity from the KnowledgeModel if the name already exists.
      entityData.remove();

      // Passes the exception.
      throw e;
    }
  }

  /**
   * Added for compatibility reason. The User object is removed from the represetnation package
   * since it was not used, so this constructor should be replaced by the Entity(KnowledgeModel, String, String)
   * constructor.
   *
   * @deprecated.
   * @param knowledgeModel The KnowledgeModel this concept is a part of.
   * @param name The name of the concept. This name must be unique in the KnowledgeModel unless the
   * model is set to accept duplicates. If it is not, a ConceptAlreadyExistException is thrown.
   * @param description An (optional) description of the entity.
   * @param user The user creating this object, now ignored.
   */
  public Entity(KnowledgeModel knowledgeModel, String name, String description, Object user) throws NameAlreadyExistException
  {
    this(knowledgeModel, name, description);
  }

  /**
   * Removes the entity. This will also remove all the realtions from and to the entity.
   * If this entity represents a relation type, all relations of this type will also
   * be removed.
   */
  public void remove()
  {
    // Removes my instances
    Relation instances[] = entityData.getRelationInstances();
    for(int i=0;i<instances.length;i++)
    {
      instances[i].remove();
    }

    // Removes my relations.
    Relation relations[] = getLocalRelations();
    for(int i=0;i<relations.length;i++)
      relations[i].remove();

    entityData.remove();
    knowledgeModel.fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.ENTITY_REMOVED));
  }

  /**
   * Returns the EntityData object in case a subclass needs to manipulate it directly.
   */
  protected EntityData getEntityData()
  {
    return entityData;
  }

  /**
   * Returns the unique ID of this relation. The kind of object this method returns
   * depends on the implementation of the EntityData class.
   */
  public Object getID()
  {
    return entityData.getID();
  }

  /**
   * Adds a relation.
   *
   * @param relationType The RelationType of the new relation to create.
   * @param entity The Entity that is the value of this relation.
   */
  public Relation addRelation(RelationType relationType, Entity value)
  {
    return relationType.newInstance(this, value);
  }

  /**
   * Adds a relation with the given strength.
   *
   * @param relationType The name of the relation type of the new relation to create.
   * @param entity The Entity that is the value of this relation.
   * @param strength A double value between 0 and 1 representing the relation's strength.
   */
  public Relation addRelation(RelationType relationType, Entity value, double strength)
  {
    return relationType.newInstance(this, value, strength);
  }

  /**
   * Adds a relation.
   *
   * @param String Name of relationtype.
   * @param Entity Reference to entity to create relation to.
   */
  public Relation addRelation(String relationType, Entity value) throws NoSuchRelationTypeException
  {
    RelationType rcd = getKnowledgeModel().getRelationType(relationType);
    if(rcd == null)
      throw new NoSuchRelationTypeException(relationType);
    else
      return rcd.newInstance(this, value);
  }

  /**
   * Adds a relation with the default strength.
   *
   * @param relationType The name of the relation typeype of the new relation to create.
   * @param entity The name of the entity that is the value of this relation.
   * @param strength A double value between 0 and 1 representing the relation's strength.
   */
  public Relation addRelation(String relationType, String value) throws NoSuchRelationTypeException, NoSuchEntityException
  {
    RelationType rcd = getKnowledgeModel().getRelationType(relationType);
    Entity valueEntity = getKnowledgeModel().getEntity(value);
    if(rcd == null)
      throw new NoSuchRelationTypeException(relationType);
    else if(valueEntity == null)
      throw new NoSuchEntityException(value);
    else
      return rcd.newInstance(this, valueEntity);
  }

  /**
   * Adds a relation with a given strength.
   *
   * @param relationType the name of relationtype.
   * @param value the reference to entity to create relation to.
   * @param explanationStrength the explanation strength for this relation
   */
  public Relation addRelation(String relationType, Entity value, double explanationStrength) throws NoSuchRelationTypeException
  {
    RelationType rcd = getKnowledgeModel().getRelationType(relationType);
    if(rcd == null)
      throw new NoSuchRelationTypeException(relationType);
    else
    {
    	Relation relation = rcd.newInstance(this, value, explanationStrength);
    	return relation;
    }
  }

  /**
   * Adds a relation with a given strength.
   *
   * @param relationType the name of relationtype.
   * @param value the reference to entity to create relation to.
   * @param explanationStrength the explanation strength for this relation
   */
  public Relation addRelation(String relationType, String value, double explanationStrength) throws NoSuchRelationTypeException, NoSuchEntityException
  {
    RelationType rcd = getKnowledgeModel().getRelationType(relationType);
    Entity valueEntity = getKnowledgeModel().getEntity(value);
    if(rcd == null)
      throw new NoSuchRelationTypeException(relationType);
    else if(valueEntity == null)
      throw new NoSuchEntityException(value);
    {
      Relation relation = rcd.newInstance(this, valueEntity, explanationStrength);
      return relation;
    }
  }

  /**
   * Return all the names of the relation types from this element. This function
   * is different from getRelations in that it also seperates between inverse names and
   * names of relations.
   *
   * Should maybe be replaced by getToRelations() and getFromRelations().
   */

  public String[] getRelationNames()
  {
    ArrayList names = new ArrayList();
    Relation[] rels = getRelations(knowledgeModel.getRelationType(BasicModel.RELATION));

    for(int i=rels.length-1;i>=0;i--)
    {
        String name = rels[i].getName();
      if(!names.contains(name))
        names.add(name);
    }

    return (String[]) names.toArray(new String[names.size()]);
  }

  /**
   * Returns all the different relation types from or to the entity.
   */
  public RelationType[] getRelationTypes()
  {
    ArrayList descs = new ArrayList();
    Relation[] rels = getRelations(knowledgeModel.getRelationType(BasicModel.RELATION));

    for(int i=rels.length-1;i>=0;i--)
    {
      RelationType rcd = rels[i].getRelationType();
      if(!descs.contains(rcd))
        descs.add(rcd);
    }

    return (RelationType[]) descs.toArray(new RelationType[descs.size()]);
  }

  /**
   * Returns all relations of the type relationClass that lead from or to this
   * entity.
   **/
  public Relation[] getLocalRelations(RelationType relationClass)
  {
    return getLocalRelations(new RelationType[] {relationClass});
  }

  /**
   * Returns all relations from or to this entity with the specified name.
   *
   * @param relationName   The name of the relation type to return.
   */

  public Relation[] getLocalRelations(String relationName)
  {
    RelationType rcd = knowledgeModel.getRelationType(relationName);
    if(rcd == null)
    {
      return new Relation[0];
    }

    return getLocalRelations(rcd);
  }

  /**
   * Returns all relations from or to this entity with any of the specified names.
   *
   * @param relationName   The name of the relation type to return.
   */
  public Relation[] getLocalRelations(String[] relationName)
  {
    RelationType[] relationClassDescriptors = new RelationType[relationName.length];
    for(int i=0;i<relationName.length;i++)
    {
      relationClassDescriptors[i] = knowledgeModel.getRelationType(relationName[i]);
    }
    return getLocalRelations(relationClassDescriptors);
  }

  /**
   * Returns all relations - local and inherited - for this entity.
   * This method use plausible inheritance.
   */
  public Relation[] getRelations()
  {
    return getRelations(knowledgeModel.getRelationType(BasicModel.RELATION), new PlausibleInheritanceMethod(knowledgeModel));
  }

  /**
   * Returns all relations from or to this entity with the specified name.
   *
   * @param relationName   The name of the relation type to return.
   */
  public Relation[] getRelations(String relationName)
  {
    RelationType rcd = knowledgeModel.getRelationType(relationName);
    if(rcd == null)
    {
      return new Relation[0];
    }

    return getRelations(rcd);
  }

  /**
   * Returns all relations from or to this entity with the specified names.
   *
   * @param relationName   The list of names of the relation types to return.
  */
  public Relation[] getRelations(String[] relationName)
  {
    RelationType[] relationClassDescriptors = new RelationType[relationName.length];
    for(int i=0;i<relationName.length;i++)
    {
      relationClassDescriptors[i] = knowledgeModel.getRelationType(relationName[i]);
    }

    return getLocalRelations(relationClassDescriptors);
  }

  /**
   * Returns all relations - local and inherited - for this entity.
   * Only relations that are of the class specified, or direct or indirect
   * subclasses of it will be returned. This method use plausible inheritance.
   *
   * @param   relationClass   The RelationType for the type of
   *                          relations returned.
   */
  public Relation[] getRelations(RelationType relationClass)
  {
    return getRelations(new RelationType[] {relationClass});
  }

  /**
   * Returns all relations - local and inherited - for this entity.
   * Only relations that are of the classes specified, or direct or indirect
   * subclasses of it will be returned. This method use plausible inheritance.
   *
   * @param   relationClass   The set of RelationTypes for the type of
   *                          relations returned.
   */
  public Relation[] getRelations(RelationType[] relationClass)
  {
    return getRelations(relationClass, new PlausibleInheritanceMethod(knowledgeModel));
//    return getRelations(relationClass, new SubclassInheritanceMethod());
  }

  /**
   * Returns all relations - local and inherited - for this entity using the
   * supplied inheritance method. Only relations that are of
   * the class specified, or direct or indirect subclasses of it will be
   * returned.
   *
   * @param   relationClass   The RelationType for the type of
   *                          relations returned.
   *          method          The inheritance method used.
   */
  public Relation[] getRelations(RelationType relationClass, InheritanceMethod method)
  {
    return getRelations(new RelationType[] { relationClass }, method);
  }

  /**
   * The same as the getRelations(RelationType, InheritanceMethod) method,
   * but it accepts a set of RelationTypes and will return relations
   * with relation-types that match or are subclasses of any of the supplied
   * RelationTypes.
   *
   * @param   relationClass   The set of RelationType for the type of
   *                          relations returned.
   *          method          The inheritance method used.
   */
  public Relation[] getRelations(RelationType relationClass[], InheritanceMethod method)
  {
    ArrayList rels = new ArrayList();
    Object key[] = new Object[3+relationClass.length];
    key[0] = method.getClass();
    key[1] = "Entity.getRelations";
    key[2] = entityData;
    for(int i=0;i<relationClass.length;i++)
      key[i+3] = relationClass[i];
    if(getKnowledgeModel().getInferenceCache().isCached(key))
    {
//      System.out.print(".");
      rels = (ArrayList) getKnowledgeModel().getInferenceCache().get(key);
    }
    else
    {
//      System.out.print(",");
      Relation[] inherited = method.inherit(this);
      for(int i=inherited.length-1;i>=0;i--)
      {
        boolean added = false;
        for(int j=relationClass.length-1;j>=0 && !added;j--)
          if(relationClass[j].isAssignableFrom(inherited[i].getRelationType()))
          {
            rels.add(inherited[i]);
            added = true;
          }
      }
      getKnowledgeModel().getInferenceCache().put(key ,rels);
    }
    return (Relation[]) rels.toArray(new Relation[rels.size()]);
  }

  /**
   * Returns the first relation of a given type. If no relations exists of this
   * type, a NoSuchRelationException is thrown.
   *
   * @param rcd The RelationType of the relation-type to find.
   */
  public Relation getRelation(RelationType rcd) throws NoSuchRelationException
  {
    Relation[] rels = getRelations(rcd);
    if(rels.length == 0)
      throw new NoSuchRelationException(rcd, this);
    else
      return rels[0];
  }

  /**
   * Returns the first relation in a set of given types. If no relations exists of any
   * of these types, a NoSuchRelationException is thrown.
   *
   * @param rcd The set of RelationType of the relation-type to find.
   */
  public Relation getRelation(RelationType[] rcd) throws NoSuchRelationException
  {
    Relation[] rels = getRelations(rcd);
    if(rels.length == 0)
      throw new NoSuchRelationException(rcd, this);
    else
      return rels[0];
  }

  /**
   * Returns the first relation of a given type (local relations only, no inherited relations).
   * If no relations exists of this type, a NoSuchRelationException is thrown.
   *
   * @param rcd The RelationType of the relation-type to find.
   */
  public Relation getLocalRelation(RelationType rcd) throws NoSuchRelationException
  {
    Relation[] rels = getLocalRelations(rcd);
    if(rels.length == 0)
      throw new NoSuchRelationException(rcd, this);
    else
      return rels[0];
  }

  /**
   * Returns the first relation of a given set of types (local relations only, no inherited relations).
   * If no relations exists of this type, a NoSuchRelationException is thrown.
   *
   * @param rcd The RelationType of the relation-type to find.
   */
  public Relation getLocalRelation(RelationType[] rcd) throws NoSuchRelationException
  {
    Relation[] rels = getLocalRelations(rcd);
    if(rels.length == 0)
      throw new NoSuchRelationException(rcd, this);
    else
      return rels[0];
  }

  /**
   * Returns the first relation of a given type. If no relations exists of this
   * type, a NoSuchRelationException is thrown.
   *
   * @param relationName The name of the relation-type to find.
   */
  public Relation getRelation(String relationName) throws NoSuchRelationException, NoSuchRelationTypeException
  {
    RelationType rcd = knowledgeModel.getRelationType(relationName);
    if(rcd == null)
      throw new NoSuchRelationTypeException(relationName);
    return getRelation(rcd);
  }

  /**
   * Given an array of strings, this method retrieves the set of relations
   * where the name of the relation-type (the RelationType) match
   * any of the given relation name, and returns the first from that set.
   * If no such relations exisst, a NoSuchRelationException is thrown.
   */
  public Relation getRelation(String relationName[]) throws NoSuchRelationException, NoSuchRelationTypeException
  {
    RelationType[] relationClassDescriptors = new RelationType[relationName.length];
    for(int i=0;i<relationName.length;i++)
    {
      relationClassDescriptors[i] = knowledgeModel.getRelationType(relationName[i]);
      if(relationClassDescriptors[i] == null)
        throw new NoSuchRelationTypeException(relationName[i]);
    }
    return getRelation(relationClassDescriptors);
  }

  /**
   * Returns the first relation of a given type (only local relations, no inherited relations).
   * If no relations exists of this type, a NoSuchRelationException is thrown.
   *
   * @param relationName The name of the relation-type to find.
   */
  public Relation getLocalRelation(String relationName) throws NoSuchRelationException, NoSuchRelationTypeException
  {
    RelationType rcd = knowledgeModel.getRelationType(relationName);
    if(rcd == null)
      throw new NoSuchRelationTypeException(relationName);
    return getLocalRelation(rcd);
  }


  /**
   * Given an array of strings, this method retrieves the set of local relations
   * where the name of the relation-type (the RelationType) match
   * any of the given relation name, and returns the first from that set.
   * If no such relations exisst, a NoSuchRelationException is thrown.
   */
  public Relation getLocalRelation(String relationName[]) throws NoSuchRelationException, NoSuchRelationTypeException
  {
    RelationType[] relationClassDescriptors = new RelationType[relationName.length];
    for(int i=0;i<relationName.length;i++)
    {
      relationClassDescriptors[i] = knowledgeModel.getRelationType(relationName[i]);
      if(relationClassDescriptors[i] == null)
        throw new NoSuchRelationTypeException(relationName[i]);
    }
    return getLocalRelation(relationClassDescriptors);
  }

  /**
   * Returns the first relation of the given type to the given value.
   *
   * @param relationType The relation type of the relation to return
   * @param value The value of the relation
   * @return The first relation matching the above criteria
   */
  public Relation getLocalRelation(RelationType relationType, Entity value) throws NoSuchRelationException
  {
    Relation[] rels = getLocalRelations(relationType);

    for(int i=0;i<rels.length;i++)
    {
      if(rels[i].getValue().equals(value))
        return rels[i];
    }

    throw new NoSuchRelationException(relationType, this);
  }

  /**
   * Returns the first relation of the given type to the given value and explanation strength.
   *
   * @param relationType The relation type of the relation to return
   * @param value The value of the relation
   * @param strength The explanation strength of the relation
   * @return The first relation matching the above criteria
   */
  public Relation getLocalRelation(RelationType relationType, Entity value, double strength) throws NoSuchRelationException
  {
    Relation[] rels = getLocalRelations(relationType);

//    System.out.println(this.getFrameString());
    for(int i=0;i<rels.length;i++)
    {
      if(rels[i].getValue().equals(value) && rels[i].getExplanationStrength() == strength)
        return rels[i];
//      else
//        System.err.println("Failed: "+rels[i].getValue()+" with "+value+" - "+rels[i].getExplanationStrength()+" with "+strength);
    }


    throw new NoSuchRelationException(relationType, this);
  }


  /**
   * Returns all local relations from or to this entity with any of the specified
   * relation types.
   *
   * @param types   An array of the types of relations to return.
   */
  public Relation[] getLocalRelations(RelationType[] types)
  {
    int typeHash = 0;
    for(int i=0;i<types.length;i++)
    {
      if(types[i] != null)
        typeHash += types[i].hashCode();
    }
    Object key[] = new Object[] { "_entity_get_local_relations", getID(), new Integer(typeHash)};
    if(getKnowledgeModel().getInferenceCache().isCached(key))
    {
      return (Relation[]) getKnowledgeModel().getInferenceCache().get(key);
    }
    ArrayList typeSet = new ArrayList();
    for(int i=0;i<types.length;i++)
    {
      if(types[i] != null)
      {
        typeSet.add(types[i]);
        typeSet.addAll(types[i].getAllSubclassesCollection());
      }
    }

    Relation[] res = entityData.getLocalRelations((Entity[]) typeSet.toArray(new Entity[typeSet.size()]));
    getKnowledgeModel().getInferenceCache().put(key, res);
    return res;
  }

  /**
   * Returns all local relations from this entity.
   */
  public Relation[] getLocalRelations()
  {
    return entityData.getLocalRelations();
  }

  /**
   * Attempts to set the name of this entity.
   * If another concept of this name already exists, a ConceptAlreadyExistException will be thrown.
   *
   * Names starting with "_" is reserverd for system use.
   *
   * @param name The (new) name to assign to this entity.
   */
  public void setName(String name) throws NameAlreadyExistException
  {
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.ENTITY_NAME_CHANGED, getName(), name);
    knowledgeModel.registerEntity(name, this);
    knowledgeModel.unregisterEntity(getName());
    entityData.setName(name);
    knowledgeModel.fireModelChangeEvent(e);
  }

  /**
   * Returns the name of this entity.
   */
  public String getName()
  {
    return entityData.getName();
  }

  /**
   * Sets the current partition context of this entity.
   *
   * @param partition The partition to set as current.
   */
  void setCurrentPartition(Partition partition)
  {
    currentPartition = partition;
  }

  /**
   * Returns the current partition, if set.
   */
  public Partition getCurrentPartition()
  {
    return currentPartition;
  }

  /**
   * Alternative toString method that should construct a brief string of the entity.
   * @return String representation of the value
   */
  public String getValueString()
  {
    if(getEntityObject() != null)
      return getEntityObject().toString()+ " ("+getName()+")";
    else
      return getName();
  }

  /**
   * Constructs the string representation of the Object conatined in the value. The default implementation should call the toString  method of the contained object.
   * @return String Representation of the object.
   */
  public String toString()
  {
//    return "Entity '"+getValueString()+"'";
    return getValueString();
  }

  /**
   * Returns a string representation of this entity, including all relations from it.
   */
  public String getFrameString()
  {
    String ret = toString()+" :\n";
    Relation[] rels=getRelations((RelationType) getKnowledgeModel().getRelationType(BasicModel.RELATION));
//    Relation[] rels=getLocalRelations((RelationType) getKnowledgeModel().getRelationType(RELATION));
    for(int i=0;i<rels.length;i++)
    {
      ret += "\t"+rels[i].getName()+"\t"+rels[i].getValue().getName()+" ("+rels[i].getExplanationStrength()+")"+(rels[i].isInherited()?" inherited. Originals are "+java.util.Arrays.asList(rels[i].getOriginal()):"")+(rels[i].isOverridden()?("\n\t\t Overridden by "+rels[i].getOverridenBy()[0]):"")+"\n";
      if(rels[i].isInherited())
      {
        try
        {
          Explanation[] exps = rels[i].getExplanations();
          ret+="\t\t\t"+exps.length+" explanations:\n";
          for(int j=exps.length-1;j>=0;j--)
            ret+="\t\t\t"+exps[j].toString()+"\n";
        }
        catch(TaskTooComplexException e)
        {
          ret += "<Too Complex>\n";
        }
      }
    }
    return ret;
  }

  /**
   * Sets an object as beeing encapsulated by this entity.
   */
  public void setEntityObject(Serializable data)
  {
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.ENTITY_DATA_OBJECT_CHANGED, getEntityObject(), data);
    entityData.setEntityObject(data);
    knowledgeModel.fireModelChangeEvent(e);

  }

  /**
   * Removes the EntityData-object encapsulated by this entity. The object itself
   * is not removed from memory, but removed from beeing encapsulated by the
   * entity.
   */
  public void removeEntityData()
  {
    entityData.setEntityObject(null);
  }

  /**
   * Returns the object encapsulated by this entity.
   */
  public Serializable getEntityObject()
  {
    return entityData.getEntityObject();
  }

  /**
   * Returns true if this Entity encapsulates a Java object.
   */
  public boolean hasEntityObject()
  {
    return entityData.getEntityObject() != null;
  }

  /**
   * Registers a relation as beeing an instance of this entity (which then
   * represents a relation type).
   *
   * @param relationData The new relation instance.
   */
  void addRelationInstance(Relation relation)
  {
    entityData.addRelationInstance(relation);
  }

  /**
   * Removes a relation as beeing an instance of this entity (which then
   * represents a relation type).
   *
   * @param relation The relation to remove.
   */
  void removeRelationInstance(Relation relation)
  {
    entityData.removeRelationInstance(relation);
  }

  /**
   * Adds a relation to the list of relations from this entity.
   *
   * @param relation The relation to add.
   */
  void addRelationReference(Relation relation)
  {
    entityData.addRelation(relation);
    knowledgeModel.fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.ENTITY_RELATION_ADDED, null, relation));
  }

  /**
   * Removes a relation to the list of relation from this entity.
   *
   * @param relation The relation to remove.
   */
  void removeRelation(Relation relation)
  {
    entityData.removeRelation(relation);
    knowledgeModel.fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.ENTITY_RELATION_REMOVED, relation, null));
  }

  /**
   * Sorting function, sorts the entity according to name.
   */
  public int compareTo(Object o)
  {
    return getName().compareTo(((Entity) o).getName());
  }

  /**
   * Two entities are considered equal if they represent the same EntityData
   * object.
   */
  public boolean equals(Object ob)
  {
    if(ob instanceof Entity)
      return entityData.equals(((Entity) ob).entityData);
    return false;
  }

  /**
   * Overrides the hashCode() method to match the class' new equals() method.
   */
  public int hashCode()
  {
    return 3+7*entityData.hashCode();
  }

  /**
   * Returns a description of this entity.
   */
  public String getDescription()
  {
    return entityData.getDescription();
  }

  /**
   * Sets the description of this entity.
   */
  public void setDescription(String description)
  {
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.ENTITY_DESC_CHANGED, getDescription(), description);
    entityData.setDescription(description);
    knowledgeModel.fireModelChangeEvent(e);
  }

  /**
   * Returns the knowledgemodel this concept belongs to.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    return knowledgeModel;
  }

  /**
   * Creates a clone of this Entity. This will not actually make a new node in
   * the semantic network, but another pointer to the same node.
   */

  public Object clone()
  {
    Entity newEnt = new Entity(knowledgeModel, entityData);
    newEnt.currentPartition = currentPartition;
    return newEnt;
  }

  /**
   * Returns the time of creation of this concept, in the the format as
   * System.currentTimeMillis();
   */
  public long getCreationTime()
  {
    return entityData.getCreationTime();
  }

  /**
   * Returns the direct superclass(es) of this entity.
   */
  public Entity[] getImmediateSuperclasses()
  {
    Collection result = getImmediateSuperclassesCollection();
    return (Entity[]) result.toArray(new Entity[result.size()]);
  }

  public Collection getImmediateSuperclassesCollection()
  {
    if(getKnowledgeModel().getEntity(BasicModel.RELATION).getID().equals(getID()))
       return new HashSet();

    Relation[] rts = getEntityData().getLocalRelations();
    HashSet result = new HashSet();
    for(int i=0;i < rts.length; i++)
      if(rts[i].getRelationType().equals(getKnowledgeModel().getRelationType(BasicModel.SUBCLASS_OF)) ||
         rts[i].getRelationType().equals(getKnowledgeModel().getRelationType(BasicModel.INSTANCE_OF)))
         result.add(rts[i].getValue());
    return result;
  }

  /**
   * Returns all the direct subclasses of this entity.
   */
  public Entity[] getImmediateSubclasses()
  {
    Collection result = getImmediateSubclassesCollection();
    return (Entity[]) result.toArray(new Entity[result.size()]);
  }

  public Collection getImmediateSubclassesCollection()
  {
    Relation[] rts = getEntityData().getLocalRelations();
    HashSet result = new HashSet();
    for(int i=0;i < rts.length; i++)
      if(rts[i].getRelationType().equals(getKnowledgeModel().getRelationType(BasicModel.HAS_SUBCLASS)) ||
         rts[i].getRelationType().equals(getKnowledgeModel().getRelationType(BasicModel.HAS_INSTANCE)))
        result.add(rts[i].getValue());
    return result;
  }

  /**
   * Returns all direct and indirect subclasses of this entity. An indirect
   * subclass is a entity that can be linked to this entity through
   * several "has subclass" relations.
   */
  public Entity[] getAllSubclasses()
  {
    Collection done = getAllSubclassesCollection();
    return (Entity[]) done.toArray(new Entity[done.size()]);
  }

  public Collection getAllSubclassesCollection()
  {
    Object key[] = new Object[] { "_entity_get_all_subclasses_collection", getID()};
    if(getKnowledgeModel().getInferenceCache().isCached(key))
      return (Collection) getKnowledgeModel().getInferenceCache().get(key);

    ArrayList queue = new ArrayList();
    queue.add(this);
    for(int i=0;i<queue.size();i++)
    {
      Entity next = (Entity) queue.get(i);
      Collection adds = next.getImmediateSubclassesCollection();
      for(Iterator it=adds.iterator();it.hasNext();)
      {
        Object nextOb = it.next();
        if(!queue.contains(nextOb))
          queue.add(nextOb);
      }
    }

    getKnowledgeModel().getInferenceCache().put(key, queue);
    return queue;
  }

  /**
   * Returns all direct and indirect subclasses of this entity.
   */
  public Entity[] getAllSuperclasses()
  {
    Collection done = getAllSuperclassesCollection();
    return (Entity[]) done.toArray(new Entity[done.size()]);
  }

  public Collection getAllSuperclassesCollection()
  {
    Object key[] = new Object[] { "_entity_get_all_superclasses_collection", getID()};
    if(getKnowledgeModel().getInferenceCache().isCached(key))
      return (Collection) getKnowledgeModel().getInferenceCache().get(key);
    ArrayList queue = new ArrayList();
    queue.add(this);
    for(int i=0;i<queue.size();i++)
    {
      Entity next = (Entity) queue.get(i);
      Collection adds = next.getImmediateSuperclassesCollection();
      for(Iterator it=adds.iterator();it.hasNext();)
      {
        Object nextOb = it.next();
        if(!queue.contains(nextOb))
          queue.add(nextOb);
      }
    }
    getKnowledgeModel().getInferenceCache().put(key, queue);
    return queue;
  }

  /**
   * Returns true if this entity is a direct or indirect superclass of the
   * given entity.
   *
   * @param subclassType A relation type to test.
   */
  public boolean isAssignableFrom(Entity subclassType)
  {
    return subclassType.equals(this) || subclassType.getAllSuperclassesCollection().contains(this);
  }

}



