package jcreek.representation;

import java.io.Serializable;

/**
 * This interface is used to access the data encapsulated by an Entity object.
 * Typically, a KnowledgeModel-implementation will also implement this interface.
 *
 * The contract of this interface specifices that:
 *  - Each EntityData object can be encapsulated by multiple Entity objects. This
 * may for instance happen when a relation type is used both in an Entity and a
 * RelationType object.
 *  - The contract does NOT state that the data is stored locally in the EntityData
 * object, and this can not be assumed. This means that the EntityData object can be
 * an interface to a standard SQL database or a concurrently accessed knowledge model
 * server.
 *  - The data represented by an EntityData object must be possible to access concurrently.
 * If an implementation of EntityData allows multiple instances of EntityData to
 * exist at once, the access must allow concurrentl, thread-safe even access to the data
 * across the EntityData instances.
 *  - All manipulations of this object is done either from a KnowledgeModel or Entity
 * object (or a subclass of these). EntityData classes should be viewed as internal
 * classes to the Entity, Relation and KnowledgeModel classes.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public interface EntityData
{
  /**
   * Returns a unique ID for the entity this EntityData represents.
   */
  public Object getID();

  /**
   * Sets the name of this entity.
   *
   * @param name The new name of the entity.
   */
  public void setName(String name);

  /**
   * Gives the name of this entity.
   */
  public String getName();

  /**
   * Sets the description of this entity.
   *
   * @param description The description String.
   */
  public void setDescription(String description);

  /**
   * Returns the description of this entity.
   *
   * @param description The description String.
   */
  public String getDescription();


  /**
   * Registers a relation as beeing an instance of this entity (which then
   * represents a relation type).
   *
   * @param relationData The new relation instance.
   */
  public void addRelationInstance(Relation relation);

  /**
   * Removes a relation as beeing an instance of this entity (which then
   * represents a relation type).
   *
   * @param relation The relation to remove.
   */
  public void removeRelationInstance(Relation relation);

  /**
   * Adds a relation to the list of relations from this entity.
   *
   * @param relation The relation to add.
   */
  public void addRelation(Relation relation);

  /**
   * Removes a relation to the list of relation from this entity.
   *
   * @param relation The relation to remove.
   */
  public void removeRelation(Relation relation);

  /**
   * Returns the Relation objects that are of
   * the relation type represented by this entity. Only entities that represent
   * relation types are expected to have such instances.
   */
  public Relation[] getRelationInstances();

  /**
   * Returns the time of creation for this entity in the same format as System.getCurrentMillis();
   */
  public long getCreationTime();

  /**
   * Sets the creation time. This is normally not necessary as the creation time
   * is automatically set when the entity is created, but in some odd cases
   * (like the importModel method in KnowledgeModel), it can be required.
   *
   * @param time The time in the format of System.getCurrentMillis() this entity was created.
   */
  public void setCreationTime(long time);

  /**
   * Returns the local relations from the entity represented by this object, where
   * the type of each relation must be in the array of relation types supplied to
   * this method.
   *
   * @param relationTypes An array of the relation types of relations to return.
   */
  public Relation[] getLocalRelations(Entity[] relationTypes);

  /**
   * Returns all the relations from this entity.
   */
  public Relation[] getLocalRelations();

  /**
   * Returns the current (if any) data object encapsulated by this entity.
   */
  public Serializable getEntityObject();

  /**
   * Sets data object encapsulated by this entity.
   *
   * @param object Any Java object that implements the Serializable interface.
   */
  public void setEntityObject(Serializable object);

  /**
   * Permanently deletes the entity.
   */
  public void remove();
}