package jcreek.representation;

/**
 * <p>Title: EntityType</p>
 * <p>Description: Represents an encapsulated entity that is a certain type of
 * frame.
 *
 * This class does not really add anything to the representation, but provide
 * an easy way to make Java classes that match different patterns in the
 * representation and provide methods to support them.
 *
 * To create a new EntityType, create a subclass of this class. You can then use
 * the matchEntityType(EntityType[] entityTypeArray) to return an array of all
 * entities in the model that match the pattern represented by this type.
 *
 * Although the supplied constructor will create an EntityType instance from an
 * existing entity, it is encouraged to create new constructors that actually
 * creates new entities of the given type in the model.
 *
 * For exampels of this, see the Case and NumberEntity classes
 *
 * @author Frode Srmo
 * @version 1.0
 */

public abstract class EntityType extends Entity
{
  /**
   * Creates a new EntityType encapsulating an existing entity. Using this
   * constructor implies that the given entity should have a structure that matches
   * the type of frame reprsented by the subclass of this EntityType.
   *
   * @param km The KnowledgeModel this EntityType is a part of.
   * @param entity The entity to match to attempt to match this EntityType.
   */
  public EntityType(Entity entity) throws EntityTypeNotMatchedException
  {
    super(entity.getKnowledgeModel(), entity.getEntityData());
    if(!matches(entity))
      throw new EntityTypeNotMatchedException(entity, this.getClass());
  }

  /**
   * Creates a new instance by creating a new node in the semantic network of
   * this particular type.
   */
  public EntityType(KnowledgeModel model, String name, String description) throws NameAlreadyExistException
  {
    super(model, name, description);
  }

  protected EntityType(Entity entity, boolean ignore)
  {
    super(entity.getKnowledgeModel(), entity.getEntityData());
  }

  /**
   * Should return true if the given entity matches the pattern represented by
   * this class of EntityType.
   */
  public static boolean matches(Entity entity)
  {
    return true;
  }
}