package jcreek.representation;

import java.util.Iterator;
import jcreek.util.LRUCache;

/**
 * The KnowledgeModel is the main interface to the model of entities and
 * relations.
 *
 * As the entities can be compared to a node and a relation to an edge, the
 * KnowledgeModel can be thought of as the graph containing these nodes and edges.
 *
 * The KnowledgeModel interface provides a uniform way of accessing the data
 * that is independent on how this data is stored. Any code that access and
 * manipulate model contents should refer to this interface instead of any
 * specific implementations of the interface.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public interface KnowledgeModel
{
  /**
   * Removes any existing content from the KnowledgeModel and replaces it with
   * the contents of the current model. This is useful for transferring models
   * across different implemetnations of this interface, e.g. from a KnowledgeModel
   * storing the data locally and a KnowledgeModel storing the information in a
   * database.
   *
   * @param model The KnowledgeModel that will have its contents copied into this model.
   */
  public void importModel(KnowledgeModel model);

  /**
   * This method merges the contents of a given model into the current model.
   * Non-permanent data is not merged, i.e. listeners and such. The model that
   * is merged into this model is not changed.
   *
   * The merge will only contain one copy of each entity that has the same name
   * in both models. If there exists a relation between two entities in the
   * current model, another relation of the same relation type between the same
   * entities will not be imported even though they may contain the same entities.
   *
   * Partitions are imported if they have different names than the already existing
   * partitions. Already existing partitions will not be changed.
   *
   * @param model The knowledgemodel to merge into this model.
   */
  public void mergeModel(KnowledgeModel model);


  /**
   * Sets name of model.
   *
   * @param name The name of the model.
   */
  public void setName(String name);

  /**
   * Returns the name of the model.
   */
  public String getName();

  /**
   * Sets description of model.
   *
   * @param description The description of the model.
   */
  public void setDescription(String description);

  /**
   * Returns the description of the model.
   */
  public String getDescription();

  /**
   * Sets title of model.
   *
   * @param title The title of the model.
   */
  public void setTitle(String title);

  /**
   * Returns the title of the model. If none is set and there is a file
   * associated with the model, the filename is returned.
   */
  public String getTitle();

  /**
   * Returns the author of the model.
   */
  public String getAuthor();

  /**
   * Sets the author of the model.
   *
   * @param authorName The name of the author.
   */
  public void setAuthor(String authorName);

  /**
   * Find an entity in the model with the given name or ID.
   *
   * @param id The name or ID of the entity to be found.
   * @return The entity-object, or null if no entity with that name or ID is found.
   */
  public Entity getEntity(Object id);

  /**
   * Find a relation in the model with the given ID.
   *
   * @param id The unique identifier for this relation.
   */
  public Relation getRelation(Object id);

  /**
   * Indexes a given entity with the given name. This is automatically done with the name when
   * generating an entity, but additional indexes may be made. An name may only
   * be used once, or a NameAleadyExistException will be returned.
   *
   * @param name The name to assoicate with this entity.
   * @param ent The Entity object.
   */
  public void registerEntity(String name, Entity ent) throws NameAlreadyExistException;

  /**
   * Removes the index with the given name from the entity index.
   *
   * @param name The name (key) of the index entry to remove.
   */
  public void unregisterEntity(String name);

  /**
   * Returns true if an entity with the given name exists in the KnowledgeModel or
   * an entity is registered with this key.
   *
   * @param id The index key of the entity.
   */
  public boolean hasEntity(Object id);

  /**
   * Returns the number of entities in the model.
   */
  public int entitySize();

  /**
   * Returns the number of relations in the model.
   */
  public int relationSize();

  /**
   * Returns an iterator of all the entities in the model.
   */
  public Iterator getEntityIterator();

  /**
   * Returns an iterator of all the realtions in the model.
   */
  public Iterator getRelationIterator();

  /**
   * Returns an array of all entities in the model. This may be prohibitive for
   * large models since it forces all entities to be stored in memory at once.
   */
  public Entity[] getEntities();

  /**
   * Returns an array of all relations in the model. This may be prohibitive for
   * large models since it forces all relations to be stored in memory at once.
   */
  public Relation[] getRelations();

  /**
   * Returns this models inference cache. This cache is flushed each time
   * the model is changed, but can be used to cache inferences done with the
   * basis of the current model.
   */
  public LRUCache getInferenceCache();

  /**
   * Returns an array of all entities in the model that match a given EntityType.
   *
   * @param arr An array of the type of EntityType to find.
   */
  public EntityType[] matchEntityType(EntityType[] arr);

  /**
   * Find a relation class in the model of the given name.
   *
   * @param name The name of the relation class to be found
   * @return The relation-class.
   */
  public RelationType getRelationType(String name);

  /**
   * Registrers a listener that will get notified if there are any changes to
   * the model.
   *
   * @param listen The ModelChangeListener to add.
   */
  public void addModelChangeListener(ModelChangeListener listen);

  /**
   * Removes a ModelChangeListener listening to changes to this model.
   *
   * @param listen The ModelChangeListener to remove.
   */
  public void removeModelChangeListener(ModelChangeListener listen);

  /**
   * Creates a new data object for storing data for an Entity. This is a part
   * of the creation of a new Entity, and should only be used from an Entity
   * constructor.
   *
   * The new entity will not be index under a name.
   */
  public EntityData makeEntity();

  /**
   * Creates a new data object for storing data for a Relation. This is a part
   * of the creation of a new Relation, and should only be used from a Relation
   * constructor.
   *
   * The Relation is required to have from, value and type entities.
   *
   * @param from The Entity this relation is from (the origin).
   * @param value The Entity this relation is to (the target).
   * @param type The Entity that represents the type of the relation.
   */
  public RelationData makeRelation(Entity from, Entity type, Entity value);

  /**
   * Creates a new data object for storing data for a Partition. This is part
   * of the creation of a new Partition and should only be used from the Partition
   * constructor.
   *
   * The Partition is required to have a unique name or a NameAlreadyExistException
   * is thrown.
   */
  public PartitionData makePartition() throws NameAlreadyExistException;

  /**
   * Returns an array of all partitions in this KnowledgeModel.
   */
  public Partition[] getPartitions();

  /**
   * Returns an iterator for all the partitions in this KnowledgeModel.
   */
  public Iterator getPartitionIterator();

  /**
   * Returns the partition with the given name or ID. If no parition with this name
   * is found, null is returned.
   *
   * @param id The name or ID of the partition to return.
   */
  public Partition getPartition(Object id);

  /**
   * Registers a partition under the given name in this models partition index.
   * The partition can then be accessed through the getPartition(Object id) method
   * using this string. If there is already a partition registered with this name,
   * a NameAlreadyExistException will be thrown.
   *
   * @param name The name (key) to register the partition under.
   * @param partition The Partition to register under this name.
   */
  public void registerPartition(String name, Partition partition) throws NameAlreadyExistException;

  /**
   * Unregisters an index entry from the parition index.
   *
   * @param name The name of the index entry to remove.
   */
  public void unregisterPartition(String name);

  /**
   * Returns true if the KnowledgeModel contains a Partition with the given
   * ID or name.
   *
   * @param id The name (string) or id of a partition to test for.
   */
  public boolean hasPartition(Object id);

  /**
   * Notifies the listeners that a change has occured in the model.
   *
   * @param event The event object reporting this change.
   */
  public void fireModelChangeEvent(ModelChangeEvent event);



  //
  // Compatibility with CKE - should eventually be removed.
  //

  /**
   * Registers a context with the KnowledgeModel.
   *
   * Only here  for compatibility reasons. Should be removed.
   */
  public void addContext(Context c);

  /**
   * Returns an array of all registered contexts.
   *
   * Only here  for compatibility reasons. Should be removed.
   */
  public Context[] getContexts();

  /**
   * Returns all context object registrered with the KnowledgeModel that is
   * of the Class given as a parameter to the method. Objects that are subclasses
   * of the given class is also returned.
   *
   */
  public Context[] getContexts(Class c);

  /**
   * Removes the given context from this KnowledgeModel.
   *
   * Only here  for compatibility reasons. Should be removed.
   */
  public void removeContext(Context c);

  /**
   * Returns true if the give context is registred with the KnowledgeModel.
   *
   * Only here  for compatibility reasons. Should be removed.
   */
  public boolean hasContext(Context c);
}
