package jcreek.representation;

import java.io.Serializable;
import java.util.*;

/**
 * Implementation of EntityData that will store data locally in memory or by
 * serialization.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class LocalEntityData implements EntityData, Serializable
{
  private String name;
  private Integer id;
  private String description;
  private long creationTime;
  private Serializable entityObject;
  private ArrayList relations;
  private ArrayList instances;
  transient LocalKnowledgeModel knowledgeModel;
  private transient HashMap relationIndex;

  LocalEntityData(LocalKnowledgeModel km, Integer id)
  {
    this.knowledgeModel = km;
    this.id = id;
    relations = new ArrayList();
    instances = new ArrayList();
  }

  /**
   * Returns a unique ID for the entity this EntityData represents.
   */
  public Object getID()
  {
    return id;
  }

  /**
   * Sets the name of this entity.
   *
   * @param name The new name of the entity.
   */
  public void setName(String name)
  {
    synchronized(knowledgeModel)
    {
      this.name = name;
    }
  }

  /**
   * Gives the name of this entity.
   */
  public String getName()
  {
    return name;
  }

  /**
   * Sets the description of this entity.
   *
   * @param description The description String.
   */
  public void setDescription(String description)
  {
    synchronized(knowledgeModel)
    {
      this.description = description;
    }
  }

  /**
   * Returns the description of this entity.
   *
   * @param description The description String.
   */
  public String getDescription()
  {
    return description;
  }

  /**
   * Returns the Relation objects that are of
   * the relation type represented by this entity. Only entities that represent
   * relation types are expected to have such instances.
   */
  public Relation[] getRelationInstances()
  {
    synchronized(knowledgeModel)
    {
      Relation[] rels = new Relation[instances.size()];
      int i = 0;
      for(Iterator it=instances.iterator();it.hasNext();)
      {
        rels[i] = knowledgeModel.getRelation(it.next());
        i++;
      }
      return rels;
    }
  }

  /**
   * Registers a relation as beeing an instance of this entity (which then
   * represents a relation type).
   *
   * @param relationData The new relation instance.
   */
  public void addRelationInstance(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      instances.add(relation.getID());
    }
  }

  /**
   * Removes a relation as beeing an instance of this entity (which then
   * represents a relation type).
   *
   * @param relation The relation to remove.
   */
  public void removeRelationInstance(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      instances.remove(relation.getID());
    }
  }

  /**
   * Adds a relation to the list of relations from this entity.
   *
   * @param relation The relation to add.
   */
  public void addRelation(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      relations.add(relation.getID());
      relationIndex = null;
//      indexRelation(relation);
    }
  }

  /**
   * Removes a relation to the list of relation from this entity.
   *
   * @param relation The relation to remove.
   */
  public void removeRelation(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      relations.remove(relation.getID());
      relationIndex = null;
    }
  }

  /**
   * Returns the time of creation for this entity in the same format as System.getCurrentMillis();
   */
  public long getCreationTime()
  {
    return creationTime;
  }

  /**
   * Sets the creation time.
   *
   * @param time The time in the format of System.getCurrentMillis() this entity was created.
   */
  public void setCreationTime(long time)
  {
    synchronized(knowledgeModel)
    {
      creationTime = time;
    }
  }

  /**
   * Returns the local relations from the entity represented by this object, where
   * the type of each relation must be in the array of relation types supplied to
   * this method.
   *
   * @param relationTypes An array of the relation types of relations to return.
   */
  public Relation[] getLocalRelations(Entity[] relationTypes)
  {
    synchronized(knowledgeModel)
    {
      HashSet typeSet = new HashSet();
      for(int i=0;i<relationTypes.length;i++)
        typeSet.add(relationTypes[i].getID());

      if(relationIndex == null)
        relationIndex = new HashMap();

      if(relationIndex.containsKey(typeSet))
        return (Relation[]) relationIndex.get(typeSet);

      ArrayList result = new ArrayList();
      for(Iterator it=relations.iterator();it.hasNext();)
      {
        Relation next = (Relation) knowledgeModel.getRelation(it.next());
        if(typeSet.contains(next.getRelationType().getID()))
        {
          result.add(next);
        }
      }

      Relation res[] = (Relation[]) result.toArray(new Relation[result.size()]);
      relationIndex.put(typeSet, res);
      return res;

      /*
      if(relationIndex == null)
        buildRelationIndex();

      HashSet relationSet = new HashSet();
      for(int i=0;i<relationTypes.length;i++)
      {
        if(relationIndex.containsKey(relationTypes[i].getID()))
          relationSet.addAll((ArrayList) relationIndex.get(relationTypes[i].getID()));
      }
      return (Relation[]) relationSet.toArray(new Relation[relationSet.size()]);
      */
    }
  }

  /**
   * Returns all the relations from this entity.
   */
  public Relation[] getLocalRelations()
  {
    ArrayList ret = new ArrayList();
    for(Iterator it=relations.iterator();it.hasNext();)
      ret.add(knowledgeModel.getRelation(it.next()));
    return (Relation[]) ret.toArray(new Relation[ret.size()]);
  }


  private void buildRelationIndex()
  {
    relationIndex = new HashMap();
    ArrayList relVector = new ArrayList();
    for(Iterator it=relations.iterator();it.hasNext();)
    {
      indexRelation(knowledgeModel.getRelation(it.next()));
    }
  }

  private void indexRelation(Relation r)
  {
    if(relationIndex == null)
      relationIndex = new HashMap();

    RelationType type = r.getRelationType();

    ArrayList v;
    if(relationIndex.containsKey(type.getID()))
      v = (ArrayList) relationIndex.get(type.getID());
    else
      v = new ArrayList();
    v.add(r);
    relationIndex.put(type.getID(), v);
    /*
    Collection types = type.getAllSuperclassesCollection();
    types.add(type);

    for(Iterator it=types.iterator();it.hasNext();)
    {
      RelationType currentType = (RelationType) it.next();
      System.out.println("   -> "+currentType);
      ArrayList v;
      if(relationIndex.containsKey(currentType.getID()))
        v = (ArrayList) relationIndex.get(currentType.getID());
      else
        v = new ArrayList();
      v.add(r);
      relationIndex.put(currentType.getID(), v);
    } */
  }

  /**
   * Returns the current (if any) data object encapsulated by this entity.
   */
  public Serializable getEntityObject()
  {
    return entityObject;
  }

  /**
   * Sets data object encapsulated by this entity.
   *
   * @param object Any Java object that implements the Serializable interface.
   */
  public void setEntityObject(Serializable object)
  {
    synchronized(knowledgeModel)
    {
      entityObject = object;
    }
  }

  /**
   * Permanently deletes the entity.
   */
  public void remove()
  {
    synchronized(knowledgeModel)
    {
      knowledgeModel.removeEntityData(this);
    }
  }

  private void readObject(java.io.ObjectInputStream stream)
       throws java.io.IOException, ClassNotFoundException
  {
    stream.defaultReadObject();
  }
   private void writeObject(java.io.ObjectOutputStream stream)
       throws java.io.IOException
   {
     stream.defaultWriteObject();
   }
}
