package jcreek.representation;

import java.util.HashMap;
import java.util.Vector;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.Iterator;
import java.io.*;
import java.util.zip.*;

import jcreek.representation.cbr.CBRModel;
import jcreek.util.LRUCache;

/**
 * The KnowledgeModel is the main interface to the model of entities and
 * relations.
 *
 * As the entities can be compared to a node and a relation to an edge, the
 * KnowledgeModel can be thought of as the graph containing these nodes and edges.
 *
 * The LocalKnowledgeModel class implements a KnowledgeModel where all the data
 * is read from a binary or XML file and stored permanently in memory until
 * explicitly saved by using the various save() commands. It can be accessed
 * concurrently by multiple threads (it is thread safe), but this is done by
 * only allowing one thread to do any access to the model at a time.
 *
 * @author Frode Srmo
 * @version 1.0
 */
public class LocalKnowledgeModel extends AbstractKnowledgeModel implements Serializable
{
  //
  // FIELDS
  //
  protected transient LRUCache inferenceCache;

  /**
   * The name of the Knowledge Model. Used for filenames and such.
   */
  protected String name = "NoName";

  /**
   * A description of this current Knowledge Model.
   */
  protected String description = "";

  /**
   * The title of the knowledgemodel. More full-length than name.
   */
  protected String title = "";

  /**
   * File to read and write to.
   */
  protected transient File file;

  /**
   * Marks if this model has been changed since last save.
   */
  protected transient boolean dirty = false;

  /**
   * The author of the KnowledgeModel
   */
  protected String author ="";

  /**
   * Index of all registered entities. This is no longer guaranteed to be a
   * one-to-one index for all indexes.
   */
//  private transient HashMap entityIndex = new HashMap();

  /**
   * Caches the entities for the getEntity() and related methods.
   *
   * This should probably be made to an LRUCache?
   */
  private transient HashMap entityCache = new HashMap(200);

  /**
   * Caches the relations for the getRelation() and related methods.
   */
   private transient HashMap relationCache = new HashMap(200);

   /**
    * Caches the relations for the getPartition() and related methods.
    */
   private transient HashMap partitionCache = new HashMap();

  /**
   * Persistent version of this index.
   */
  private HashMap entityIDIndex = new HashMap();


  /**
   * Index of all registered partitions.
   */
//  private transient HashMap partitionIndex = new HashMap();

  private HashMap partitionIDIndex = new HashMap();

  /**
   * Tracks the ID numbers assigned to EntityData and RelationData objects. This
   * object contains the reference to the last number assigned.
   */
  private int lastIdentifier = 0;

  /**
   * The mapping of EntityData objects from EntityData identifiers. This is
   * done using a LinkedHashMap so that traversal of the list is also reasonably
   * quick.
   */
  private LinkedHashMap entityDataMap = new LinkedHashMap();

  /**
   * The mapping of RelationData objects from RelationData identifiers. This is
   * done using a LinkedHashMap so that traversal of the list is also reasonably
   * quick.
   */
  private LinkedHashMap relationDataMap = new LinkedHashMap();

  /**
   * The mapping of the PartitionData object from PartitionData identifiers.
   */
  private LinkedHashMap partitionDataMap = new LinkedHashMap();


  private Vector contexts = new Vector();

  private static final long serialVersionUID=-4904879461577094564L;

  //
  // CONSTRUCTORS
  //

  /**
   * Creates a new knowledge model based on the epistemological model supplied.
   */
  public LocalKnowledgeModel(EpistemologicalModel epistemologicalModel)
  {
    epistemologicalModel.generateEpistemologicalModel(this);
  }

  /**
   * Create a new, empty knowledge model.
   *
   * This constructor uses the jcreek.representation.cbr.CBRModel as its
   * epistemological model.
   */
  public LocalKnowledgeModel()
  {
    CBRModel.generateCBRModel(this);
  }

  /**
   * Creates a new KnowledgeModel instance by loading it from the filename supplied.
   */
  public LocalKnowledgeModel(String filename) throws IOException, ClassNotFoundException
  {
    this.load(filename);
  }

  /**
   * Creates a new KnowledgeModel instance by loading it from the file supplied.
   */
  public LocalKnowledgeModel(File file) throws IOException, ClassNotFoundException
  {
    this.load(file);
  }

  /**
   * Creates a new KnowledgeModel instance by loading it from the supplied stream.
   */
  public LocalKnowledgeModel(InputStream input) throws IOException, ClassNotFoundException
  {
    this.load(input);
  }


  /**
   * Creates a new KnowledgeModel by importing the contents of the supplied model.
   *
   * @param model The KnowledgeModel to import.
   */
  public LocalKnowledgeModel(KnowledgeModel model)
  {
    super(model);
  }


  //
  // ACCESS/CHANGE METHODS
  //

  /**
   * Sets name of model.
   *
   * @param name The name of the model.
   */
  public void setName(String name)
  {
    String oldValue;
    synchronized(this)
    {
      oldValue = this.name;
      this.name = name;
    }
    fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.MODEL_NAME_CHANGED, oldValue, name));
  }

  /**
   * Returns the name of the model.
   */
  public String getName()
  {
      return name;
  }

  /**
   * Sets description of model.
   *
   * @param description The description of the model.
   */
  public void setDescription(String description)
  {
    String oldValue;
    synchronized(this)
    {
      oldValue = this.description;
      this.description = description;
    }
    fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.MODEL_DESC_CHANGED, oldValue, description));
  }

  /**
   * Returns the description of the model.
   */
  public String getDescription()
  {
    return description;
  }

  /**
   * Sets title of model.
   *
   * @param title The title of the model.
   */
  public void setTitle(String title)
  {
    String oldValue;
    synchronized(this)
    {
      oldValue = this.title;
      this.title = title;
    }
    fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.MODEL_TITLE_CHANGED, oldValue, title));
  }

  public static void main(String argv[])
  {
    try
    {
      LocalKnowledgeModel km = new LocalKnowledgeModel();
      Entity test = new Entity(km, "Test", "test");
      System.out.println("getEntity: " + km.getEntity("Test") + "  has: " +
                         km.hasEntity("Test"));
      test.remove();
      System.out.println("getEntity: " + km.getEntity("Test") + "  has: " +
                         km.hasEntity("Test"));
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }

  }

  /**
   * Returns the title of the model. If none is set and there is a file
   * associated with the model, the filename is returned.
   */
  public String getTitle()
  {
    synchronized(this)
    {
      if(title != null)
        return title;
      else
        if(file != null)
          return file.getName();
      return null;
    }
  }

  /**
   * Returns the author of the model.
   */
  public String getAuthor()
  {
    return author;
  }

  /**
   * Sets the author of the model.
   *
   * @param authorName The name of the author.
   */
  public void setAuthor(String authorName)
  {
    String oldValue;
    synchronized(this)
    {
      oldValue = this.author;
      this.author = authorName;
    }
    fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.MODEL_AUTHOR_CHANGED, oldValue, author));
  }

  /**
   * Sets the file this KnowledgeModel is associated with.
   * @param  file  file object
   */
  public void setFile( File file )
  {
    this.file = file;
  } // setFile


  /**
   * Returns the file this KnowledgeModel is associated with (if any).
   */
  public File getFile()
  {
      return file;
  }

  /**
   * Returns true if this KnowledgeModel has a file associated with it.
   */
  public boolean hasFile()
  {
    return file != null;
  }

  /**
   * Find an entity in the model with the given name or ID.
   *
   * @param id The name (String) or identifier (Object) for an entity
   * @return The entity-object, or null if no entity with that name or ID is found.
   */
  public Entity getEntity(Object id)
  {
    synchronized(this)
    {
      if(id instanceof String)
        id = ((String) id).toLowerCase();

      if(entityCache.containsKey(id))
        return (Entity) entityCache.get(id);


      EntityData data = (EntityData) entityDataMap.get(id);
      Entity result = null;
      if(data != null)
      {
        result = new Entity(this, data);
      }
      if(id instanceof String)
      {
        result = getEntity(entityIDIndex.get(id));
      }
      if(result != null)
        entityCache.put(id, result);

      return result;
    }
  }

  /**
   * Find a relation in the model with the given ID.
   *
   * @param id The unique identifier for this relation.
   */
  public Relation getRelation(Object id)
  {
    if(relationCache.containsKey(id))
    {
      return (Relation) relationCache.get(id);
    }
    synchronized(this)
    {
      RelationData data = (RelationData) relationDataMap.get(id);
      if(data != null)
      {
        Relation res = new Relation(this, data);
        relationCache.put(id, res);
        return res;
      }
      return null;
    }
  }

  /**
   * Find a relation class in the model of the given name. If no relation type
   * is indexed under that name, null is returned.
   *
   * @param name The name of the relation class to be found
   * @return The RelationType object for this type
   */
  public RelationType getRelationType(String name)
  {
    Entity type = getEntity(name);
    if(type instanceof RelationType)
    {
      return (RelationType) type;
    }

    synchronized(this)
    {
      if(type != null)
      {
        try
        {
          RelationType typeRT = new RelationType(type);
          entityCache.remove(name.toLowerCase());
          entityCache.put(name.toLowerCase(), typeRT);
          return typeRT;
        }
        catch(EntityTypeNotMatchedException e)
        {
          // This should never happen.
          e.printStackTrace(System.err);
        }
      }
    }
    return null;
  }


  /**
   * Indexes a given entity with the given id. This is automatically done when
   * generating an entity.
   *
   * @param id The name of the
   * @param ent The Concept object.
   */
  public void registerEntity(String id, Entity ent) throws NameAlreadyExistException
  {
    synchronized(this)
    {
      if(hasEntity(id))
        throw new NameAlreadyExistException(id);
//      entityIndex.put(id.toLowerCase(), ent);
      entityCache.put(id, ent);
      entityIDIndex.put(id.toLowerCase(), ent.getID());
    }
  }

  /**
   * Removes the index with the given name from the entity index.
   *
   * @param name The name (key) of the index entry to remove.
   */
  public void unregisterEntity(String name)
  {
    synchronized(this)
    {
      entityCache.remove(name);
      entityIDIndex.remove(name.toLowerCase());
    }
  }

  /**
   * Returns true if an entity with the given name or id exists in the KnowledgeModel.
   *
   * @param id The key of the concept.
   */
  public boolean hasEntity(Object id)
  {
    synchronized(this)
    {
      if(id instanceof String)
      {
        String idStr = (String) id;
        if(entityIDIndex.containsKey(idStr.toLowerCase()))
          return true;
      }
      return entityDataMap.containsKey(id);
    }
  }

  /**
   * Returns true if this model contains the supplied entity.
   *
   * @param entity The entity to check if exists in the model.
   */
  public boolean hasEntity(Entity entity)
  {
    synchronized(this)
    {
      return entityDataMap.containsKey(entity.getID());
    }
  }

  /**
   * Returns true if this model contains the supplied relation.
   *
   * @param relation The relation to check if exists in the model.
   */
  public boolean hasRelation(Relation relation)
  {
    synchronized(this)
    {
      return relationDataMap.containsKey(relation.getID());
    }
  }

  /**
   * Returns the number of entities in the model.
   */
  public int entitySize()
  {
    synchronized(this)
    {
      return entityDataMap.size();
    }
  }

  /**
   * Returns the number of relations in the model.
   */
  public int relationSize()
  {
    synchronized(this)
    {
      return relationDataMap.size();
    }
  }

  /**
   * Returns an iterator of all the entities in the model.
   *
   * The snapshot of the entities in the model is created from
   */
  public Iterator getEntityIterator()
  {
    synchronized(this)
    {
      return new EntityIterator(this, entityDataMap.values().iterator());
    }
  }

  /**
   * Returns an iterator of all the realtions in the model.
   */
  public Iterator getRelationIterator()
  {
    synchronized(this)
    {
      return new RelationIterator(this, relationDataMap.values().iterator());
    }
  }

  /**
   * Creates a new data object for storing data for an Entity. This is a part
   * of the creation of a new Entity, and should only be used from an Entity
   * constructor.
   *
   */
  public EntityData makeEntity()
  {
    synchronized(this)
    {
      lastIdentifier++;
      Integer id = new Integer(lastIdentifier);
      EntityData newData = new LocalEntityData(this, id);
      entityDataMap.put(id, newData);
      return newData;
    }
  }

  /**
   * Creates a new data object for storing data for a Relation. This is a part
   * of the creation of a new Relation, and should only be used from a Relation
   * constructor.
   *
   * The Relation is required to have from, value and type entities.
   *
   * @param from The Entity this relation is from (the origin).
   * @param value The Entity this relation is to (the target).
   * @param type The Entity that represents the type of the relation.
   */
  public RelationData makeRelation(Entity from, Entity type, Entity value)
  {
    synchronized(this)
    {
      lastIdentifier++;
      Integer id = new Integer(lastIdentifier);
      RelationData newData = new LocalRelationData(this, id, from, value, type);
      relationDataMap.put(id, newData);

      return newData;
    }
  }

  /**
   * Removes an EntityData object from the index. Should only be called from the
   * EntityData object itself.
   */
  void removeEntityData(LocalEntityData entityData)
  {
    synchronized(this)
    {
      entityDataMap.remove(entityData.getID());
      ArrayList removeList = new ArrayList();

      // Removing from the entityIDIndex
      for(Iterator it=entityIDIndex.keySet().iterator();it.hasNext();)
      {
        String key = (String) it.next();
        if(((Object) entityIDIndex.get(key)).equals(entityData.getID()))
        {
          removeList.add(key);
        }
      }
      for(Iterator it=removeList.iterator();it.hasNext();)
      {
        String id = (String) it.next();
        unregisterEntity(id);
      }
    }
  }

  /**
   * Removes an EntityData object from the index. Should only be called from the
   * EntityData object itself.
   */
  void removeRelationData(LocalRelationData relationData)
  {
    synchronized(this)
    {
      relationDataMap.remove(relationData.getID());

      ArrayList removeList = new ArrayList();
      for(Iterator it=relationCache.keySet().iterator();it.hasNext();)
      {
        Object key = it.next();
        if(((Object) relationCache.get(key)).equals(relationData.getID()))
          removeList.add(key);
      }
      for(Iterator it=removeList.iterator();it.hasNext();)
      {
        Object id = it.next();
        relationCache.remove(it.next());
      }
    }
  }

  /**
   * Removes a PartitionData object from the index. Should only be called from the
   * PartitionData object itself.
   */
  void removePartitionData(LocalPartitionData partitionData)
  {
    synchronized(this)
    {
      partitionDataMap.remove(partitionData.getID());
      ArrayList removeList = new ArrayList();
      for(Iterator it=partitionIDIndex.keySet().iterator();it.hasNext();)
      {
        String key = (String) it.next();
        if(((Object) partitionIDIndex.get(key.toLowerCase())).equals(partitionData.getID()))
          removeList.add(key);
      }
      for(Iterator it=removeList.iterator();it.hasNext();)
      {
        String id = (String) it.next();
        unregisterPartition(id);
      }
    }
  }

  /**
   * Creates a new data object for storing data for a Partition. This is part
   * of the creation of a new Partition and should only be used from the Partition
   * constructor.
   *
   */
  public PartitionData makePartition() throws NameAlreadyExistException
  {
    synchronized(this)
    {
      lastIdentifier++;
      Integer id = new Integer(lastIdentifier);
      PartitionData newData = new LocalPartitionData(this, id);
      partitionDataMap.put(id, newData);
      return newData;
    }
  }

  /**
   * Registers a partition under the given name in this models partition index.
   * The partition can then be accessed through the getPartition(Object id) method
   * using this string. If there is already a partition registered with this name,
   * a NameAlreadyExistException will be thrown.
   *
   * @param name The name (key) to register the partition under.
   * @param partition The Partition to register under this name.
   */
  public void registerPartition(String name, Partition partition) throws NameAlreadyExistException
  {
    synchronized(this)
    {
      if(hasPartition(name))
        throw new NameAlreadyExistException(name);
      partitionCache.put(name.toLowerCase(), partition);
      partitionIDIndex.put(name.toLowerCase(), partition.getID());
    }
  }

  /**
   * Unregisters an index entry from the parition index.
   *
   * @param name The name of the index entry to remove.
   */
  public void unregisterPartition(String name)
  {
    synchronized(this)
    {
      partitionCache.remove(name);
      partitionIDIndex.remove(name.toLowerCase());
    }
  }

  /**
   * Returns true if the KnowledgeModel contains a Partition with the given
   * ID or name.
   *
   * @param id The name (string) or id of a partition to test for.
   */
  public boolean hasPartition(Object id)
  {
    synchronized(this)
    {
      if(id instanceof String)
      {
        String idStr = (String) id;
        if(partitionIDIndex.containsKey(idStr.toLowerCase()))
          return true;
      }
      return partitionDataMap.containsKey(id);
    }
  }

  /**
   * Returns an iterator for all the partitions in this KnowledgeModel.
   */
  public Iterator getPartitionIterator()
  {
    synchronized(this)
    {
      return new PartitionIterator(this, partitionDataMap.values().iterator());
    }
  }

  /**
   * Returns the partition with the given name or ID. If no parition with this name
   * is found, null is returned.
   *
   * @param id The name or ID of the partition to return.
   */
  public Partition getPartition(Object id)
  {
    synchronized(this)
    {
      if(id instanceof String)
        id = ((String) id).toLowerCase();
      if(partitionCache.containsKey(id))
        return (Partition) partitionCache.get(id);


      PartitionData data = (PartitionData) partitionDataMap.get(id);
      Partition result = null;
      if(data != null)
      {
        result = new Partition(this, data);
      }
      if(id instanceof String)
      {
        result = getPartition(partitionIDIndex.get(id));
      }
      if(result != null)
        partitionCache.put(id, result);

      return result;
    }
  }

  //
  // SAVE/LOAD METHODS
  //

  /**
   * Saves this document (the data) to the associated disk
   * file.
   */
  public synchronized void save() throws IOException
  {
    synchronized(this)
    {
      FileOutputStream fileStream = new FileOutputStream(file);
      ObjectOutputStream out = new ObjectOutputStream(new DeflaterOutputStream(fileStream));
      writeObject(out);

      out.flush();
      out.close();
      dirty = false;
    }
  } // save

  /**
   * Saves this knowledgemodel (the data) to a new file, and
   * hereinafter is associated with this new file.
   *
   * @param  file  new file to save to
   */
  public void saveAs(File file) throws IOException
  {
    synchronized(this)
    {
      this.file = file;
      save();
    }
  } // saveAs

  /**
   * Saves the model to disk.
   *
   * @param name The filename of the model.
   */
  public void saveAs(String name) throws IOException
  {
    synchronized(this)
    {
      if(!name.endsWith(".km"))
        name = name+".km";
      saveAs(new File(name));
    }
  }

  /**
   * Load a model from disk.
   * @param name The filename of the model (without extension)
   */
  public void load(String name) throws ClassNotFoundException, IOException
  {
    load(new File(name));
  }

  /**
   * Loads a model from disk. This methods sets the file associated with this
   * model as in setFile() and calls load().
   *
   * @param file The file representing the model.
   */
  public void load(File file) throws ClassNotFoundException, IOException
  {
    synchronized(this)
    {
      this.file = file;
      load();
    }
  }

  /**
   * Loads a model from a stream.
   *
   * @param in Loads from the stream in.
   */
  public void load(InputStream in) throws ClassNotFoundException, IOException
  {
    synchronized(this)
    {
      ObjectInputStream input = new ObjectInputStream(new InflaterInputStream(in));
      readObject(input);

      if(inferenceCache != null)
        getInferenceCache().flush();

    }
  }

  /**
   * Recreates the relation type index. This is necessary because of a bug
   * in 0.65 branch.
   */
  public void reindex()
  {
    Relation[] rels = this.getRelations();
    for(int i=0;i<rels.length;i++)
    {
      Relation rel = rels[i];
      if(rel.getFrom() == null || rel.getValue() == null || rel.getRelationType() == null || rel.getInverse() == null)
      {
        System.out.println("Relation ["+rel.getFrom()+ " - " +rel.getRelationType()+ " - "+rel.getValue()+" ( "+rel.getInverse()+" ) is buggy and removed.");
        rel.remove();
      }
      else
      {
        if(!java.util.Arrays.asList(rel.getRelationType().getInstances()).contains(rel))
        {
//          System.out.println("Fixed reindex bug in [" + rel + "]");
          rel.getRelationType().addRelationInstance(rel);
        }
      }
    }
  }


  /**
   * Loads the model from the file associated with the model.
   */
  public void load() throws ClassNotFoundException, IOException
  {
    synchronized(this)
    {
      FileInputStream fileStream = new FileInputStream(file);
      ObjectInputStream input = new ObjectInputStream(new InflaterInputStream(fileStream));
      readObject(input);

      if(inferenceCache != null)
        getInferenceCache().flush();
    }
  }

  private void readObject(ObjectInputStream input) throws ClassNotFoundException, IOException
  {
    this.name = (String) input.readObject();
    this.description = (String) input.readObject();
    this.author = (String) input.readObject();
    this.title = (String) input.readObject();
    this.entityDataMap = (LinkedHashMap) input.readObject();
    this.relationDataMap = (LinkedHashMap) input.readObject();
    this.relationCache = new HashMap(relationDataMap.size());
    this.entityIDIndex = (HashMap) input.readObject();
    this.entityCache = new HashMap(entityIDIndex.size());
    this.partitionDataMap = (LinkedHashMap) input.readObject();
    this.partitionIDIndex = (HashMap) input.readObject();
    this.partitionCache = new HashMap(200);
    this.lastIdentifier = input.readInt();
    this.contexts = (Vector) input.readObject();

    for(Iterator it=entityDataMap.values().iterator();it.hasNext();)
      ((LocalEntityData) it.next()).knowledgeModel = this;
    for(Iterator it=relationDataMap.values().iterator();it.hasNext();)
      ((LocalRelationData) it.next()).knowledgeModel = this;
    for(Iterator it=partitionDataMap.values().iterator();it.hasNext();)
      ((LocalPartitionData) it.next()).knowledgeModel = this;

    // Force a reindex of relation types due to bug in 0.65 branch.
    reindex();
  }

  private void writeObject(ObjectOutputStream output) throws IOException
  {
    output.writeObject(name);
    output.writeObject(description);
    output.writeObject(author);
    output.writeObject(title);
    output.writeObject(entityDataMap);
    output.writeObject(relationDataMap);
    output.writeObject(entityIDIndex);
    output.writeObject(partitionDataMap);
    output.writeObject(partitionIDIndex);
    output.writeInt(lastIdentifier);
    output.writeObject(contexts);
  }

  //
  // Compatibility with CKE - should eventually be removed.
  //

  /**
   * Registers a context with the KnowledgeModel.
   */
  public void addContext(Context c)
  {
    dirty = true;
    if(!contexts.contains(c))
      contexts.add(c);
  }

  /**
   * Returns an array of all registered contexts.
   */
  public Context[] getContexts()
  {
    return (Context[]) contexts.toArray(new Context[0]);
  }

  /**
   * Returns all context object registrered with the KnowledgeModel that is
   * of the Class given as a parameter to the method. Objects that are subclasses
   * of the given class is also returned.
   */
  public Context[] getContexts(Class c)
  {
    Vector v = new Vector();
    for(int i=contexts.size()-1;i>=0;i--)
    {
      Context context = (Context) contexts.elementAt(i);
      if(context.getClass().isAssignableFrom(c))
        v.add(context);
    }
    return (Context[]) contexts.toArray(new Context[0]);
  }

  /**
   * Removes the given context from this KnowledgeModel.
   */
  public void removeContext(Context c)
  {
    dirty = true;
    contexts.removeElement(c);
  }

  /**
   * Returns true if the give context is registred with the KnowledgeModel.
   */
  public boolean hasContext(Context c)
  {
    return contexts.contains(c);
  }


  /**
   * Returns true if this knowledgemodel has been changed since last save.
   */
  public boolean isDirty()
  {
    return dirty;
  }

  /**
   * Returns this models inference cache. This cache is flushed each time
   * the model is changed, but can be used to cache inferences done with the
   * basis of the current model.
   */
  public synchronized LRUCache getInferenceCache()
  {
    if(inferenceCache == null)
    {
      inferenceCache = new LRUCache(50000);
    }
    return inferenceCache;
  }

  /**
   * Notifies the listeners that a change has occured in the model.
   *
   * @param event The event object reporting this change.
   */
  public void fireModelChangeEvent(ModelChangeEvent event)
  {
    ArrayList listenerClone;

    getInferenceCache().flush();

    if(modelChangeListeners != null)
    {
      // Creates a copy of the arraylist to avoid staying in a synchronized block
      // while executing events in external objects to avoid deadlock traps.
      synchronized(this)
      {
        listenerClone = (ArrayList) modelChangeListeners.clone();
      }

      if(modelChangeListeners != null)
      {
        for(int i=modelChangeListeners.size()-1;i>=0;i--)
          ((ModelChangeListener) modelChangeListeners.get(i)).modelChanged(event);
      }
    }
  }


  private static class EntityIterator implements Iterator
  {
    Iterator superIterator;
    KnowledgeModel km;

    EntityIterator(KnowledgeModel km, Iterator it)
    {
      superIterator = it;
      this.km = km;
    }

    public Object next()
    {
      return new Entity(km, (EntityData) superIterator.next());
    }

    public boolean hasNext()
    {
      return superIterator.hasNext();
    }

    public void remove()
    {
      superIterator.remove();
    }
  }

  private static class RelationIterator implements Iterator
  {
    Iterator superIterator;
    KnowledgeModel km;

    RelationIterator(KnowledgeModel km, Iterator it)
    {
      superIterator = it;
      this.km = km;
    }

    public Object next()
    {
      return new Relation(km, (RelationData) superIterator.next());
    }

    public boolean hasNext()
    {
      return superIterator.hasNext();
    }

    public void remove()
    {
      superIterator.remove();
    }
  }

  private static class PartitionIterator implements Iterator
  {
    Iterator superIterator;
    KnowledgeModel km;

    PartitionIterator(KnowledgeModel km, Iterator it)
    {
      superIterator = it;
      this.km = km;
    }

    public Object next()
    {
      return new Partition(km, (PartitionData) superIterator.next());
    }

    public boolean hasNext()
    {
      return superIterator.hasNext();
    }

    public void remove()
    {
      superIterator.remove();
    }
  }
}
