package jcreek.representation;

import java.util.*;

/**
 * Implements the PartitionData interface by storing the partition information
 * locally, in memory. Companion class to LocalKnowledgeModel.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class LocalPartitionData implements PartitionData, java.io.Serializable
{
  transient LocalKnowledgeModel knowledgeModel;
  private Integer id;
  private String name;
  private long creationTime;
  private HashSet entities = new HashSet();
  private HashSet relations = new HashSet();

  /**
   * Creates a new LocalPartitionData object. Should only be done from the
   * makePartition() method in the LocalKnowledgeModel
   */
  LocalPartitionData(LocalKnowledgeModel knowledgeModel, Integer id)
  {
    this.knowledgeModel = knowledgeModel;
    this.id = id;
  }

  /**
   * Sets the name of this partition.
   *
   * @param name The new name
   */
  public void setName(String name)
  {
    synchronized(knowledgeModel)
    {
      this.name = name;
    }
  }

  /**
   * Returns the name of this partition.
   */
  public String getName()
  {
    return name;
  }

  /**
   * Returns a unique ID for the entity this EntityData represents.
   */
  public Object getID()
  {
    return id;
  }

  /**
   * Returns the time of creation for this partition in the same format as System.getCurrentTimeMillis();
   */
  public long getCreationTime()
  {
    return creationTime;
  }

  /**
   * Sets the creation time. This is normally not necessary as the creation time
   * is automatically set when the entity is created, but in some odd cases
   * (like the importModel method in KnowledgeModel), it can be required.
   *
   * @param time The time in the format of System.getCurrentTimeMillis() this parition was created.
   */
  public void setCreationTime(long time)
  {
    synchronized(knowledgeModel)
    {
      creationTime = time;
    }
  }

  /**
   * Adds a relation to this partition.
   *
   * @param relation The relation to add.
   */
  public void addRelation(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      relations.add(relation.getID());
    }
  }

  /**
   * Removes a relation to the list of relation from this entity. Returns false
   * if the relation do not exist in the partition.
   *
   * @param relation The relation to remove.
   */
  public boolean removeRelation(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      return relations.remove(relation.getID());
    }
  }

  /**
   * Returns all the relations in this partition.
   */
  public Relation[] getRelations()
  {
    synchronized(knowledgeModel)
    {
      ArrayList list = new ArrayList();
      for(Iterator it = getRelationIterator();it.hasNext();)
      {
        list.add(it.next());
      }

      Relation[] ret = (Relation[]) list.toArray(new Relation[list.size()]);
      return ret;
    }
  }

  /**
   * Returns an iterator for all the relations in this partition.
   */
  public Iterator getRelationIterator()
  {
    synchronized(knowledgeModel)
    {
      return new RelationIterator(knowledgeModel, id, ((HashSet) relations.clone()).iterator());
    }
  }

  /**
   * Returns true if the given relation is a part of this partition.
   *
   * @param relation
   */
  public boolean hasRelation(Relation relation)
  {
    synchronized(knowledgeModel)
    {
      return relations.contains(relation.getID());
    }
  }

  /**
   * Adds an entity to this partition.
   *
   * @param entity The entity to add.
   */
  public void addEntity(Entity entity)
  {
    synchronized(knowledgeModel)
    {
      entities.add(entity.getID());
    }
  }

  /**
   * Removes an entity to the list of entity from this partition. Returns false
   * if the entity do not exist in the partition.
   *
   * @param entity The entity to remove.
   */
  public boolean removeEntity(Entity entity)
  {
    synchronized(knowledgeModel)
    {
      return relations.remove(entity.getID());
    }
  }

  /**
   * Returns all the entitys in this partition.
   */
  public Entity[] getEntities()
  {
    synchronized(knowledgeModel)
    {
      ArrayList list = new ArrayList();
      for(Iterator it = getEntityIterator();it.hasNext();)
      {
        list.add(it.next());
      }

      Entity[] ret = (Entity[]) list.toArray(new Entity[list.size()]);
      return ret;
    }
  }

  /**
   * Returns an iterator for all the entities in this partition.
   */
  public Iterator getEntityIterator()
  {
    synchronized(knowledgeModel)
    {
      return new EntityIterator(knowledgeModel, id, ((HashSet) entities.clone()).iterator());
    }
  }

  /**
   * Returns true if the given entity is a part of this partition.
   *
   * @param entity
   */
  public boolean hasEntity(Entity entity)
  {
    synchronized(knowledgeModel)
    {
      return entities.contains(entity.getID());
    }
  }

  /**
   * Permanently deletes the parition.
   */
  public void remove()
  {
    knowledgeModel.removePartitionData(this);
  }

  private static class EntityIterator implements Iterator
  {
    Iterator superIterator;
    KnowledgeModel km;
    Integer id;

    EntityIterator(KnowledgeModel km, Integer id, Iterator it)
    {
      superIterator = it;
      this.km = km;
      this.id = id;
    }

    public Object next()
    {
      Entity rel = (Entity) km.getEntity(superIterator.next()).clone();
      rel.setCurrentPartition(km.getPartition(id));
      return rel;
    }

    public boolean hasNext()
    {
      return superIterator.hasNext();
    }

    public void remove()
    {
      superIterator.remove();
    }
  }

  private static class RelationIterator implements Iterator
  {
    Iterator superIterator;
    KnowledgeModel km;
    Integer id;

    RelationIterator(KnowledgeModel km, Integer id, Iterator it)
    {
      superIterator = it;
      this.km = km;
      this.id = id;
    }

    public Object next()
    {
      Relation rel = (Relation) km.getRelation(superIterator.next()).clone();
      rel.setCurrentPartition(km.getPartition(id));
      return rel;
    }

    public boolean hasNext()
    {
      return superIterator.hasNext();
    }

    public void remove()
    {
      superIterator.remove();
    }
  }
}