package jcreek.representation;

import java.util.*;

/**
 * Implementation of RelationData that stores the relation data in memory or
 * by serialization.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class LocalRelationData implements RelationData, java.io.Serializable
{
  transient LocalKnowledgeModel knowledgeModel;
  private Integer id;
  private Integer fromID;
  private transient Entity from;
  private Integer valueID;
  private transient Entity value;
  private Integer typeID;
  private transient Entity type;
  private long creationTime;
  private ArrayList overrides;
  private ArrayList overriddenBy;
  private Integer inverseID;
  private transient Relation inverse;
  private String description;
  private double explanationStrength;

  LocalRelationData(LocalKnowledgeModel km, Integer id, Entity from, Entity value, Entity type)
  {
    knowledgeModel = km;
    this.id = id;
    this.from = from;
    this.fromID = (Integer) from.getID();
    this.value = value;
    this.valueID = (Integer) value.getID();
    this.type = type;
    this.typeID = (Integer) type.getID();
    this.explanationStrength = -1;
    overrides = new ArrayList();
    overriddenBy = new ArrayList();
  }

  /**
   * Returns a unique ID for the entity this elationData represents.
   */
  public Object getID()
  {
    return id;
  }

  /**
   * Sets the from-entity (the origin) of the relation.
   *
   * @param from The new from-entity of the relation.
   */
  public void setFrom(Entity from)
  {
    synchronized(knowledgeModel)
    {
      this.from = from;
      fromID = (Integer) from.getID();
    }
  }

  /**
   * Returns the entity from where the relation originates.
   */
  public Entity getFrom()
  {
    synchronized(knowledgeModel)
    {
      if(from == null)
        from = knowledgeModel.getEntity(fromID);
      return from;
    }
  }

  /**
   * Sets the value (the target) of the relation.
   *
   * @param value The new value of the relation.
   */
  public void setValue(Entity value)
  {
    synchronized(knowledgeModel)
    {
      this.value = value;
      valueID = (Integer) value.getID();
    }
  }

  /**
   * Returns the entity this relation is pointing at.
   */
  public Entity getValue()
  {
    synchronized(knowledgeModel)
    {
      if(value == null)
        value = knowledgeModel.getEntity(valueID);
      return value;
    }
  }

  /**
   * Returns the instance of Entity describing this relation's type.
   */
  public Entity getRelationType()
  {
    synchronized(knowledgeModel)
    {
      if(type == null)
        type = knowledgeModel.getEntity(typeID);
      return type;
    }
  }

  /**
   * Sets this relation's type.
   *
   * @param relationType The Entity describing this relation's type.
   */
  public void setRelationType(Entity relationType)
  {
    synchronized(knowledgeModel)
    {
      this.type = relationType;
      typeID = (Integer) type.getID();
    }
  }

  /**
   * Sets the desciption of this relation.
   *
   * @param description The String containing the description.
   */
  public void setDescription(String description)
  {
    synchronized(knowledgeModel)
    {
      this.description = description;
    }
  }

  /**
   * Returns the description of this entity.
   */
  public String getDescription()
  {
    return description;
  }

  /**
   * Returns the time of creation for this relation in the same format as System.getCurrentMillis();
   */
  public long getCreationTime()
  {
    return creationTime;
  }

  /**
   * Sets the creation time. T
   *
   * @param time The time in the format of System.getCurrentMillis() this relation was created.
   */
  public void setCreationTime(long time)
  {
    synchronized(knowledgeModel)
    {
      this.creationTime = time;
    }
  }

  /**
   * Sets the strength variable contains the explanation strength for this relation.
   *
   * @param value The explanation strength.
   */
  public void setExplanationStrength(double value)
  {
    synchronized(knowledgeModel)
    {
      this.explanationStrength = value;
    }
  }

  /**
   * Returns the explanation strength for this relation.
   */
  public double getExplanationStrength()
  {
    return explanationStrength;
  }

  /**
   * This relation is market as overriding another relation.
   *
   * @param override The relation to override
   */
  public void addOverriding(Relation override)
  {
    synchronized(knowledgeModel)
    {
      overrides.add(override.getID());
    }
  }

  /**
   * Removes a relation from beeing overriden by this relation.
   *
   * @param override The relation to remove from the overide list.
   */
  public void removeOverriding(Relation override)
  {
    synchronized(knowledgeModel)
    {
      overrides.remove(override.getID());
    }
  }

  /**
   * Returns the relation (if any) this relation overrides. See setOverriding().
   */
  public Relation[] getOverrides()
  {
    synchronized(knowledgeModel)
    {
      Relation rels[] = new Relation[overrides.size()];
      for(int i=0;i<rels.length;i++)
      {
        rels[i] = knowledgeModel.getRelation(overrides.get(i));
      }
      return rels;
    }
  }

  /**
   * Returns true if this relation overrides another relation. See setOverriding().
   */
  public boolean isOverriding()
  {
    synchronized(knowledgeModel)
    {
      return !overrides.isEmpty();
    }
  }

  /**
   * Returns true if there is at least one relation overriding this relation.
   * See setOverriding().
   */
  public boolean isOverridden()
  {
    synchronized(knowledgeModel)
    {
      return !overriddenBy.isEmpty();
    }
  }

  /**
   * This relation is market as beeing overriden by another relation.
   *
   * @param override The relation to override
   */
  public void addOverridenBy(Relation override)
  {
    synchronized(knowledgeModel)
    {
      overriddenBy.add(override.getID());
    }
  }

  /**
   * Removes a relation from beeing overriden by this relation.
   *
   * @param override The relation to remove from the overide list.
   */
  public void removeOverriddenBy(Relation override)
  {
    synchronized(knowledgeModel)
    {
      overriddenBy.remove(override.getID());
    }
  }

  /**
   * Returns which relations this relation is overriden by. See setOverriding().
   */
  public Relation[] getOverridenBy()
  {
    synchronized(knowledgeModel)
    {
      Relation rels[] = new Relation[overriddenBy.size()];
      for(int i=0;i<rels.length;i++)
      {
        rels[i] = knowledgeModel.getRelation(overriddenBy.get(i));
      }
      return rels;
    }
  }

  /**
   * Permanently removes the relation represented by this RelationData object.
   */
  public void remove()
  {
    synchronized(knowledgeModel)
    {
      knowledgeModel.removeRelationData(this);
    }
  }

  /**
   * Sets the inverse of this relation to be the given relation.
   *
   * @param relation The inverse of this relation.
   */
  public void setInverse(Relation inverse)
  {
    synchronized(knowledgeModel)
    {
      this.inverse = inverse;
      inverseID = (Integer) inverse.getID();
    }
  }

  /**
   * Returns the inverse of this relation. If it has none, null is returned.
   */
  public Relation getInverse()
  {
    if(inverse == null)
        inverse = knowledgeModel.getRelation(inverseID);
    return inverse;
  }
}