package jcreek.representation;

import jcreek.util.LRUCache;

/**
 * A Partition is a submodel of a KnowledgeModel. Unlike the SubModel class, the
 * Partition is a persistent part of the model, and can be looked up, retrieved
 * and accessed through a KnowledgeModel instance.
 *
 * In addition, it to limit the scope of model inference and browsing to a
 * partition so that e.g. an Entity object will only return relations from this
 * entity that are in the "current partition", and inheritance will only work
 * on relations and entities present in this partition. To do this, see the setCurrentPartition()
 * method in the Entity and Relation classes.
 *
 * It is also possible to register ModelChangeListener instances to a
 * partition. This will allow the listener to only get events when a change
 * occurs in the partition, i.e. an entity or relation is the partition is
 * changed or a change is done to the partition itself.
 *
 * Each Partition also have an inference cache associated with it. When possible,
 * it is wise to use this cache instead of the one from the KnowledgeModel class
 * as a Partition's inference cache is only flushed when a change occurs in the
 * Partition while the KnowledgeModel inference cache is flushed whenever any
 * change occurs.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class Partition extends SubModel
{
  private transient PartitionData partitionData;
  private transient KnowledgeModel knowledgeModel;

  /**
   * Creates a new Partition with the given name in the given KnowledgeModel.
   * If there already exists a partition in the model with this, a
   * NameAlreadyExistException is thrown.
   *
   * @param model The KnowledgeModel to make a new Partition in.
   * @param name The name of the new Parition.
   */
  public Partition(KnowledgeModel model, String name) throws NameAlreadyExistException
  {
    knowledgeModel = model;
    try
    {
      partitionData = model.makePartition();
      model.registerPartition(name, this);
      partitionData.setName(name);
      knowledgeModel.fireModelChangeEvent(new ModelChangeEvent(model, ModelChangeEvent.PARTITION_ADDED, null, this));
    }
    catch(NameAlreadyExistException e)
    {
      // Clean up if the exception is thrown.
      partitionData.remove();

      // Throw it onwards.
      throw e;
    }
  }

  /**
   * Creates a new Partition object from an existing PartitionData object. This
   * will not creat a new partition, but only a new wrapper clase for an existing
   * one. This should only be used by the KnowledgeModel and PartitionData classes.
   */
  Partition(KnowledgeModel model, PartitionData data)
  {
    knowledgeModel = model;
    partitionData = data;
  }

  /**
   * Sets the name of the Partition. If there is already a Partition with the
   * new name already exists in the KnowledgeModel, a NameAlreadyExistException
   * is thrown.
   *
   * @param name The new name of the Partition.
   */
  public void setName(String name) throws NameAlreadyExistException
  {
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.PARTITION_NAME_CHANGE, getName(), name);
    knowledgeModel.registerPartition(name, this);
    knowledgeModel.unregisterPartition(name);
    partitionData.setName(name);
    knowledgeModel.fireModelChangeEvent(e);
  }

  /**
   * Returns the name of this Partition.
   */
  public String getName()
  {
    return partitionData.getName();
  }

  /**
   * Adds an entity to the partition.
   *
   * @param entity The entity to add.
   */
  public void add(Entity entity)
  {
    Entity copy = (Entity) entity.clone();
    copy.setCurrentPartition(this);
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.PARTITION_ENTITY_ADDED, null, copy);
    partitionData.addEntity(copy);
    knowledgeModel.fireModelChangeEvent(e);
  }

  /**
   * Adds a set of entitites to the partition.
   *
   * @param ents The arrays of entities to add.
   */
  public void add(Entity[] ents)
  {
    for(int i=0;i<ents.length;i++)
      add(ents[i]);
  }

  /**
   * Adds a relation to the partition.
   *
   * @param r The relation to add.
   */
  public void add(Relation r)
  {
    Relation copy = (Relation) r.clone();
    copy.setCurrentPartition(this);
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.PARTITION_RELATION_ADDED, null, copy);
    partitionData.addRelation(copy);
    knowledgeModel.fireModelChangeEvent(e);
  }

  /**
   * Adds an array of relations to the partition.
   *
   * @param rels The array of relations to add.
   */
  public void add(Relation[] rels)
  {
    for(int i=0;i<rels.length;i++)
      add(rels[i]);
  }

  /**
   * Removes a relation from the partition. Returns false if the relation is not
   * present in the partition.
   */
  public boolean remove(Relation r)
  {
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.PARTITION_RELATION_REMOVED, null, r);
    boolean ret = partitionData.removeRelation(r);
    if(ret)
      knowledgeModel.fireModelChangeEvent(e);
    return ret;
  }

  /**
   * Removes an entity from the submodel. Return false if the relation
   * is not present in the partition.
   */
  public boolean remove(Entity ent)
  {
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.PARTITION_ENTITY_REMOVED, null, ent);
    boolean ret = partitionData.removeEntity(ent);
    if(ret)
      knowledgeModel.fireModelChangeEvent(e);
    return ret;
  }

  /**
   * Returns true if the submodel contains the entity.
   */
  public boolean contains(Entity ent)
  {
    return partitionData.hasEntity(ent);
  }

  /**
   * Returns true if the submodel contains the relation.
   */
  public boolean contains(Relation rel)
  {
    return partitionData.hasRelation(rel);
  }

  /**
   * Returns an array of all the entities in the submodel.
   */
  public Entity[] getEntities()
  {
    return partitionData.getEntities();
  }

  /**
   * Returns an array of all the relations in the submodel.
   */
  public Relation[] getRelations()
  {
    return partitionData.getRelations();
  }

  /**
   * Removes the partition from the KnowledgeModel.
   */
  public void remove()
  {
    partitionData.remove();
    knowledgeModel.fireModelChangeEvent(new ModelChangeEvent(knowledgeModel, ModelChangeEvent.PARTITION_REMOVED, this, null));
  }

  /**
   * Returns the PartitionData object in case a subclass needs to manipulate it directly.
   */
  protected PartitionData getPartitionData()
  {
    return partitionData;
  }

  /**
   * Returns the unique ID of this partition. The kind of object this method returns
   * depends on the implementation of the PartitionData class.
   */
  public Object getID()
  {
    return partitionData.getID();
  }

  /**
   * Returns the knowledgemodel this partition belongs to.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    return knowledgeModel;
  }

  /**
   * Returns the time of creation of this partition, in the the format as
   * System.currentTimeMillis();
   */
  public long getCreationTime()
  {
    return partitionData.getCreationTime();
  }

  /**
   * Two entities are considered equal if they represent the same PartitionData
   * object.
   */
  public boolean equals(Object ob)
  {
    if(ob instanceof Partition)
      return getPartitionData().equals(((Partition) ob).getPartitionData());
    return false;
  }

  public String toString()
  {
    return "Partition "+getName()+"\n"+super.toString();
  }

  /**
   * Overrides the hashCode() method to match the class' new equals() method.
   */
  public int hashCode()
  {
    return 3+7*getPartitionData().hashCode();
  }
}
