package jcreek.representation;

import java.util.Iterator;

/**
 * The PartitionData interface is an interface for storing data about partitions. An
 * implementation of the KnowledgeModel interface will also typically have a custom
 * implementation of this interface.
 *
 * The contract of this interface specifices that:
 *  - Each ParitionData object can be encapsulated by multiple Partition objects.
 *  - The contract does NOT state that the data is stored locally in the PartitionData
 * object, and this can not be assumed. This means that the PartitionData object can be
 * an interface to a standard SQL database or a concurrently accessed knowledge model
 * server.
 *  - The data represented by an PartitionData object must be possible to access concurrently.
 * If an implementation of PartitionData allows multiple instances of PartitionData to
 * exist at once, the access must allow concurrent, thread-safe even access to the data
 * across the PartitionData instances.
 *  - All manipulations of this object is done either from a KnowledgeModel or Partition
 * object (or a subclass of these). PartitionData classes should be viewed as internal
 * classes to the Partition, Entity, Relation and KnowledgeModel classes.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public interface PartitionData
{
  /**
   * Sets the name of this partition.
   *
   * @param name The new name
   */
  public void setName(String name);

  /**
   * Returns the name of this partition.
   */
  public String getName();

  /**
   * Returns a unique ID for the entity this EntityData represents.
   */
  public Object getID();

  /**
   * Returns the time of creation for this partition in the same format as System.getCurrentTimeMillis();
   */
  public long getCreationTime();

  /**
   * Sets the creation time. This is normally not necessary as the creation time
   * is automatically set when the entity is created, but in some odd cases
   * (like the importModel method in KnowledgeModel), it can be required.
   *
   * @param time The time in the format of System.getCurrentTimeMillis() this parition was created.
   */
  public void setCreationTime(long time);

  /**
   * Adds a relation to this partition.
   *
   * @param relation The relation to add.
   */
  public void addRelation(Relation relation);

  /**
   * Removes a relation to the list of relation from this entity. Returns false
   * if the relation do not exist in the partition.
   *
   * @param relation The relation to remove.
   */
  public boolean removeRelation(Relation relation);

  /**
   * Returns all the relations in this partition.
   */
  public Relation[] getRelations();

  /**
   * Returns an iterator for all the relations in this partition.
   */
  public Iterator getRelationIterator();

  /**
   * Returns true if the given relation is a part of this partition.
   *
   * @param relation
   */
  public boolean hasRelation(Relation relation);

  /**
   * Adds an entity to this partition.
   *
   * @param entity The entity to add.
   */
  public void addEntity(Entity entity);

  /**
   * Removes an entity to the list of entity from this partition. Returns false
   * if the entity do not exist in the partition.
   *
   * @param entity The entity to remove.
   */
  public boolean removeEntity(Entity entity);

  /**
   * Returns all the entitys in this partition.
   */
  public Entity[] getEntities();

  /**
   * Returns an iterator for all the entities in this partition.
   */
  public Iterator getEntityIterator();

  /**
   * Returns true if the given entity is a part of this partition.
   *
   * @param entity
   */
  public boolean hasEntity(Entity entity);

  /**
   * Permanently deletes the parition.
   */
  public void remove();
}