package jcreek.representation;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.*;

import jcreek.representation.inference.ExplanationModel;
import jcreek.representation.inference.Explanation;
import jcreek.representation.inference.TaskTooComplexException;
import jcreek.representation.inference.ActivationConstants;

/**
 * The Relation class represents a relation (like an edge in a graph) between
 * two entities. Associated with each relation is a RelationType object that
 * siginifies its type. If the RelationType has an inverse type, an inverse relation
 * is also created (in the other direction) between the two entities.
 *
 * @author Frode Srmo
 * @version 1.1
 */
public class Relation implements Cloneable
{
  //******************
  // OBJECT VARIABLES
  //******************

  private transient RelationData relationData;
  private transient KnowledgeModel knowledgeModel;
  private transient Vector originals;
  private transient ExplanationModel explanationModel;
  private transient Partition currentPartition;

  //*********
  // CONSTRUCTOR
  //*********


  /**
   * Creates a new relation without supplying a RelationType object (the RelationType
   * class is really a convinience class of a representation level higher than Relation
   * since the type of a relation in the formalism is represented by an Entity). This should
   * only be used to define the very core of the model i.e. the relation used to
   * define a relation type.
   */
  protected Relation(Entity from, Entity type, Entity value, double strength, boolean ignore)
  {
    knowledgeModel = type.getKnowledgeModel();
    relationData = knowledgeModel.makeRelation(from, type, value);
    relationData.setExplanationStrength(strength);
  }

  /**
   * Create a new relation of a given type from an origin entity to a target
   * entity. This automatically also creates a relation in the opposite direction
   * (from the target entity to the origin entity) of the opposite type.
   *
   * @param type The type of the relation to create.
   * @param from The origin entity - where the relation is from.
   * @param value The target entity - where the relation is pointing to.
   */
  public Relation(Entity from, RelationType type, Entity value)
  {
    this(from, type, value, type.getExplanationStrength(), false);
    relationData.setInverse(new Relation(value, type.getInverseType(), from, type.getInverseType().getExplanationStrength(), true));
    relationData.getInverse().getRelationData().setInverse(this);
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_ADDED));
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(getInverse(), ModelChangeEvent.RELATION_ADDED));
    from.addRelationReference(this);
    value.addRelationReference(getInverse());
    type.addRelationInstance(this);
    type.getInverseType().addRelationInstance(getInverse());
  }

  /**
   * Create a new relation of a given type from an origin entity to a target
   * entity. This automatically also creates a relation in the opposite direction
   * (from the target entity to the origin entity) of the opposite type.
   *
   * @param type The type of the relation to create.
   * @param from The origin entity - where the relation is from.
   * @param value The target entity - where the relation is pointing to.
   * @param strength The explanation strength of the relation.
   */

  public Relation(Entity from, RelationType type, Entity value, double strength)
  {
    this(from, type, value, strength, false);
    relationData.setInverse(new Relation(value, type.getInverseType(), from, strength, true));
    relationData.getInverse().getRelationData().setInverse(this);
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_ADDED));
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(getInverse(), ModelChangeEvent.RELATION_ADDED));
    from.addRelationReference(this);
    value.addRelationReference(getInverse());
    type.addRelationInstance(this);
    type.getInverseType().addRelationInstance(getInverse());
  }

  /**
   * Create a new relation of a given type from an origin entity to a target
   * entity. This automatically also creates a relation in the opposite direction
   * (from the target entity to the origin entity) of the opposite type.
   *
   * In addition, this relation is market as overriding another relation. This means
   * that if both relations are found to be (either directly or through inheritance)
   * present in an entity, this relation will supress the other relation so it will
   * not be returned.
   *
   * @param type The type of the relation to create.
   * @param from The origin entity - where the relation is from.
   * @param value The target entity - where the relation is pointing to.
   * @param override The relation that this relation overrides.
   */

  public Relation(Entity from, RelationType type, Entity value, Relation override)
  {
    this(from, type, value);
    addOverriding(override);
    getInverse().addOverriding(override.getInverse());
  }

  /**
   * Creates a new Relation from an existing RelationData object. This should
   * only be done by a KnowledgeModel implementation
   */
  Relation(KnowledgeModel model, RelationData data)
  {
    this.knowledgeModel = model;
    this.relationData = data;
  }

  //*********
  // METHODS
  //*********

  /**
   * Removes this relation. This will also automatically remove the inverse
   * of this relation.
   */
  public void remove()
  {
    try
    {
      removeMe();
      if(getInverse() != null)
        getInverse().removeMe();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
    catch(Error e)
    {
      e.printStackTrace();
    }

  }

  private void removeMe()
  {
    Relation overrides[] = relationData.getOverrides();
    for(int i=0;i<overrides.length;i++)
      if(overrides[i] != null)
        this.removeOverriding(overrides[i]);

    overrides = relationData.getOverridenBy();
    for(int i=0;i<overrides.length;i++)
      if(overrides[i] != null)
        overrides[i].removeOverriding(this);

    if(getFrom() != null)
      getFrom().removeRelation(this);
    if(getRelationType() != null)
      getRelationType().removeRelationInstance(this);

//    System.out.println("RemoveMe for "+toString()+ " - already removed from "+getFrom()+"s list.");
    getRelationData().remove();

    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_REMOVED));
  }


  /**
   * Sets the strength variable contains the explanation strength for this relation.
   * This value must be [0.0..1.0] - any value below this range is set to 0.0, any
   * value above this range is set to 1.0.
   *
   * If this realtion is temporary (e.g. inherited), this will confirm the relation
   * and set the local explanation strength - it will not change the explanation
   * strength of the original relation this relation is derrived from (e.g. inherited
   * from). See the confirm() method.
   *
   * @param value The explanation strength.
   */
  public void setExplanationStrength(double value)
  {
    confirm();
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.RELATION_REMOVED, new Double(getExplanationStrength()), new Double(value));
    relationData.setExplanationStrength(value);
    getKnowledgeModel().fireModelChangeEvent(e);
  }

  /**
   * Returns the strength variable contains the explanation strength for this relation.
   * If not set, the default for the class is used. If the task is too complex to
   * calculate, 0.0 will be returned.
   */
  public double getExplanationStrength()
  {
    if(isInherited())
    {
      try
      {
/*        Object[] key = { "relation.getExplanationStrength", getRelationType().getID(), getFrom().getID(), getValue().getID() };
        if(getKnowledgeModel().getInferenceCache().isCached(key))
        {
          System.out.print(";");
          return ((Double) getKnowledgeModel().getInferenceCache().get(key)).doubleValue();
        }
        System.out.print(":");*/
        double strength = getExplanationModel().getExplanationStrength(getRelationType(), getValue());
//        getKnowledgeModel().getInferenceCache().put(key, new Double(strength));
        return strength;
      }
      catch(TaskTooComplexException e)
      {
        return 0.0;
      }
    }
    return getLocalExplanationStrength();
  }

  double getLocalExplanationStrength()
  {
    return relationData.getExplanationStrength();
  }

  /**
   * Sets the explanation strength to 0 - in
   * effect representing explicitly that there is NOT a relationship of this kind between
   * the two entities. Useful for exceptions to general rules (i.e. all birds flies, except
   * penguins etc.)
   *
   * If this realtion is temporary (e.g. inherited), this will confirm the relation
   * and set the local explanation strength - it will not change the explanation
   * strength of the original relation this relation is derrived from (e.g. inherited
   * from). See the confirm() method.
   */

  public void negate()
  {
    setExplanationStrength(0.0);
  }

  /**
   * Returns true if this relation is negated, i.e. represents an explicit representation
   * of a lack of relation between two entities.
   */
  public boolean isNegated()
  {
    return getExplanationStrength() == 0.0;
  }

  /**
   * Returns the entity this relation is pointing at.
   */
  public Entity getValue()
  {
    return relationData.getValue();
  }

  /**
   * Returns the entity from where the relation originates.
   */
  public Entity getFrom()
  {
    return relationData.getFrom();
  }
  /**
   * getInverseName returns the name of the inverse relation.
   */
  public String getInverseName()
  {
    return getInverse().getName();
  }

  /**
   * Returns the name of this relation.
   */
  public String getName()
  {
    return getRelationType().getName();
  }

  /**
   * Returns true if this relation has an inverse relation. Normally all relations
   * have that.
   */
  public boolean hasInverse()
  {
    return getInverse() != null;
  }

  /**
   * Returns the natural inverse name of the relation. By "natural", is ment a
   * form that can be used to generate natural language sentences. For a relation,
   * this is usually a verb phrase.
   */
  public String getNaturalInverseName()
  {
    return getInverse().getNaturalName();
  }

  /**
   * Returns the natural name of the relation. By "natural", is ment a
   * form that can be used to generate natural language sentences. For a relation,
   * this is usually a verb phrase.
   */
  public String getNaturalName()
  {
    return getRelationType().getNaturalName();
  }

  /**
   * Returns the reverse of this relation, i.e. the relation going from the
   * target of this relation to the origin of this relation with a relation type
   * that is the inverse of this relation's type.
   */
  public Relation getInverse()
  {
    return relationData.getInverse();
  }

  /**
   * Returns an instance of RelationType describing this relations type.
   */
  public RelationType getRelationType()
  {
    // This should be changed somehow so it don't make new RelationType instances
    // all the time.

    Entity ent = relationData.getRelationType();
    if(ent instanceof RelationType)
      return (RelationType) ent;
    else
      try
      {
        if(ent == null)
          System.err.println("WARNING! relation type is null with relation from "+getFrom()+" to "+getValue());
        return new RelationType(ent);
      }
      catch(EntityTypeNotMatchedException e) { }
    return null;
  }

  /**
   * This relation is market as overriding another relation. This means
   * that if both relations are found to be (either directly or through inheritance)
   * present in an entity, this relation will supress the other relation so it will
   * not be returned.
   */
  public void addOverriding(Relation override)
  {
    relationData.addOverriding(override);
    override.addOverriddenBy(this);
    if(!isInherited())
      getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_OVERRIDE_ADDED, null, override));
  }

  /**
   * Removes a relation from the list of overridden relations.
   */
  public void removeOverriding(Relation override)
  {
    relationData.removeOverriding(override);
    override.removeOverriddenBy(this);
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_OVERRIDE_REMOVED, override, null));
  }

  /**
   * Returns the relations (if any) this relation overrides. See addOverriding().
   */
  public Relation[] getOverridingRelations()
  {
    return relationData.getOverrides();
  }

  /**
   * Returns true if this relation overrides another relation.See setOverriding().
   */
  public boolean isOverriding()
  {
    return relationData.isOverriding();
  }

  /**
   * Returns true if there is at least one relation overriding this relation.
   * See setOverriding().
   */
  public boolean isOverridden()
  {
    return relationData.isOverridden();
  }

  /**
   * Returns which relations this relation is overriden by. See setOverriding().
   */
  public Relation[] getOverridenBy()
  {
    return relationData.getOverridenBy();
  }

  protected void removeOverriddenBy(Relation rel)
  {
    relationData.removeOverriddenBy(rel);
  }

  protected void addOverriddenBy(Relation rel)
  {
    relationData.addOverridenBy(rel);
  }

  /**
  * Sets the value (the target) of the relation.
  *
  * This also updates the inverse relation.
  *
  * @param value The new value of the relation.
  */
  public void setValue(Entity value)
  {
    confirm();
    Entity oldValue = getValue();
    relationData.setValue(value);
    getInverse().getRelationData().setFrom(value);
    oldValue.removeRelation(getInverse());
    value.addRelationReference(getInverse());
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_VALUE_CHANGED, oldValue, value));
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(getInverse(), ModelChangeEvent.RELATION_FROM_CHANGED, oldValue, value));
  }//setValue


  /**
  * Sets the from-entity (the origin) of the relation.
  *
  * This also updates the inverse relation.
  *
  * @param from The new from-entity of the relation.
  */
  public void setFrom(Entity from)
  {
    confirm();
    Entity oldValue = getFrom();
    relationData.setFrom(from);
    getInverse().getRelationData().setValue(from);
    oldValue.removeRelation(this);
    from.addRelationReference(this);
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_FROM_CHANGED, oldValue, from));
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(getInverse(), ModelChangeEvent.RELATION_VALUE_CHANGED, oldValue, from));
  }//setValue

  /**
   * Sets the relation-type of this relation. This also changes the relation-type
   * of the inverse relation.
   *
   * @param relationType The new relation-type.
   */
  public void setRelationType(RelationType relationType)
  {
    confirm();
    RelationType oldValue = getRelationType();
    relationData.setRelationType(relationType);
    Entity oldInverse = getInverse().getRelationType();
    getInverse().getRelationData().setRelationType(relationType.getInverseType());
    oldValue.removeRelationInstance(this);
    relationType.addRelationInstance(this);
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(this, ModelChangeEvent.RELATION_TYPE_CHANGED, oldValue, relationType));
    getKnowledgeModel().fireModelChangeEvent(new ModelChangeEvent(getInverse(), ModelChangeEvent.RELATION_VALUE_CHANGED, oldInverse, relationType.getInverseType()));
  }

  /**
   * Two entities are considered equal if they represent the same RelationData
   * object.
   */
  public boolean equals(Object ob)
  {
    if(ob instanceof Relation)
      return getRelationData().equals(((Relation) ob).getRelationData());
    return false;
  }

  /**
   * Overrides the hashCode() method to match the class' new equals() method.
   */
  public int hashCode()
  {
    return 5+7*getRelationData().hashCode();
  }

  /**
   * Returns a description of this relation.
   */
  public String getDescription()
  {
    return relationData.getDescription();
  }

  /**
   * Sets the description of this relation.
   */
  public void setDescription(String description)
  {
    confirm();
    ModelChangeEvent e = new ModelChangeEvent(this, ModelChangeEvent.RELATION_TYPE_CHANGED, getDescription(), description);
    relationData.setDescription(description);
    getKnowledgeModel().fireModelChangeEvent(e);
  }

  /**
   * Returns the knowledgemodel this relation belongs to.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    return knowledgeModel;
  }

  /**
   * Returns the time of creation of this relation, in the the format as
   * System.currentTimeMillis();
   */
  public long getCreationTime()
  {
    return relationData.getCreationTime();
  }

  /**
   * When inheriting a relation, a copy of the relation is made and returned.
   *
   * This means that the original origin entity is changed to the entity this
   * relation is inherited to.
   *
   * @param inheritedTo The entity this relation shall be inherited to.
   * @param model The ExplanationModel justifying this inheritance.
   */
  public Relation inherit(Entity inheritedTo, ExplanationModel model)
  {
//    System.err.println("Relation.inherit "+this+" to "+inheritedTo);
    Relation ret=null;
    if(isInherited())
    {
      Relation rel = ((Relation) originals.get(0)).inherit(inheritedTo, model);
      for(int i=1;i<originals.size();i++)
        rel.addOriginal((Relation) originals.get(i));
      return rel;
    }
    else
    {
      ret = new Relation(getKnowledgeModel(), new TemporaryRelationData(inheritedTo, getValue(), getRelationType()));
      ret.getRelationData().setDescription(getDescription());

      Relation overrides[] = getOverridingRelations();
      for(int i=0;i<overrides.length;i++)
        ret.addOverriding(overrides[i]);

      overrides = getOverridenBy();
      for(int i=0;i<overrides.length;i++)
        ret.addOverriddenBy(overrides[i]);

      ret.originals = new Vector();
      ret.originals.add(this);
      ret.explanationModel = model;
      ret.addOriginal(this);

      if(hasInverse())
      {
        Relation inverse = new Relation(getKnowledgeModel(), new TemporaryRelationData(getValue(), inheritedTo, this.getRelationType().getInverseType()));
        inverse.getRelationData().setDescription(getInverse().getDescription());

        overrides = getInverse().getOverridingRelations();
        for(int i=0;i<overrides.length;i++)
          inverse.addOverriding(overrides[i]);

        overrides = getInverse().getOverridenBy();
        for(int i=0;i<overrides.length;i++)
          inverse.addOverriddenBy(overrides[i]);

        inverse.originals = new Vector();

        inverse.explanationModel = model;
        inverse.addOriginal(inverse);
        ret.getRelationData().setInverse(inverse);
        inverse.getRelationData().setInverse(ret);
      }
    }
//    System.err.println("Done inherit: "+ret+"   inverse: "+ret.getInverse());
    return ret;
  }

  /**
   * If this relation is inherited, this method returns the model that justifies
   * the inheritance. If it is not inherited, null is returned.
   */
  public ExplanationModel getExplanationModel()
  {

    if(!isInherited())
    {
      return null;
    }
    else
    {
      return explanationModel;
    }
  }

  /**
   * If this relation is inherited, this method returns a submodel of the
   * knowledgemodel which contains all relations and all entities
   * contained in all explanation-paths from the origin to the target entity.
   */
  public SubModel getPathModel()
  {
    if(!isInherited())
    {
      return null;
    }
    else
    {
      return explanationModel.getActivatedModel(getRelationType(), (Entity) getValue());
    }
  }

  /**
   * Returns an array of explanations for this relation. This is only valid if this
   * relation is composite, i.e. not a single relation. This task is NP-hard and as such
   * it may throw a TaskTooComplexException if the task is above a treshold which currently
   * is a constant.
   */
  public Explanation[] getExplanations() throws TaskTooComplexException
  {
    if(!isInherited())
      return null;
    else
      return explanationModel.getExplanations(getRelationType(), (Entity) getValue(), 10000);
  }

  /**
   * Returns true if this relation is inherited.
   */
  public boolean isInherited()
  {
    return !isPersistent();
  }

  /**
   * Returns true if this relation is composite, i.e. composed of a chain of relations.
   */
  public boolean isPersistent()
  {
    return !(relationData instanceof TemporaryRelationData);
  }

  /**
   * When this relation is an inherited relation, it returnes the original
   * relation it was inherited from.
   */
  public Relation[] getOriginal()
  {
    if(isInherited())
    {
      return (Relation[]) originals.toArray(new Relation[0]);
    }
    return new Relation[0];
  }

  /**
   * Returns true if the supplied relations is among the originals of this
   * relation. This is only relevant for inherited relations, and as such
   * this method will always return false for non-inherited relations.
   *
   * @param rel The relation to test.
   */
  public boolean hasOriginal(Relation rel)
  {
    return isInherited() && originals.contains(rel);
  }

  /**
   * If this relation is inherited, this method adds a relation that is counted as
   * beeing among the originals of this relations. This method is usually only
   * used by an InheritanceMethod.
   *
   * @param rel The (additional) original relation.
   */
  public void addOriginal(Relation rel)
  {
    if(isInherited())
      originals.add(rel);
  }

  /*
   * If this is an inherited relation, this method will make it permanent and add it to the local
   * lists of relations of the origin and target entities.
   */
  private void makePermanent()
  {
    if(isInherited())
    {
      System.err.println("Relation.makePermanent for "+this+" - inverse: "+getInverse());
      RelationData newRelationData = getKnowledgeModel().makeRelation(getFrom(), getRelationType(), getValue());
      newRelationData.setDescription(getDescription());
      newRelationData.setExplanationStrength(getExplanationStrength());

      Relation overrides[] = getOverridingRelations();
      Relation inverse = getInverse();
      relationData = newRelationData;

      for(int i=0;i<originals.size();i++)
      {
        Relation orig = (Relation) originals.get(i);
        newRelationData.addOverriding(orig);
        orig.addOverriddenBy(this);
      }

      for(int i=0;i<overrides.length;i++)
      {
        newRelationData.addOverriding(overrides[i]);
        overrides[i].addOverriddenBy(this);
      }



      if(inverse != null && inverse != this)
      {

        RelationData newInverseData = getKnowledgeModel().makeRelation(inverse.getFrom(), inverse.getRelationType(), inverse.getValue());
        newInverseData.setDescription(inverse.getDescription());
        newInverseData.setExplanationStrength(inverse.getExplanationStrength());

        overrides = inverse.getOverridingRelations();
        inverse.relationData = newInverseData;

        for(int i=0;i<inverse.originals.size();i++)
        {
          Relation orig = (Relation) inverse.originals.get(i);
          newInverseData.addOverriding(orig);
          orig.addOverriddenBy(inverse);
        }
        for(int i=0;i<overrides.length;i++)
        {
          newInverseData.addOverriding(overrides[i]);
          overrides[i].addOverriddenBy(inverse);
        }
        inverse.originals = null;
        inverse.explanationModel = null;

        getRelationData().setInverse(inverse);
        inverse.getRelationData().setInverse(this);
      }

      originals = null;
      explanationModel = null;

      getFrom().addRelationReference(this);
      getValue().addRelationReference(getInverse());
      getRelationType().addRelationInstance(this);
      getRelationType().getInverseType().addRelationInstance(getInverse());
    }
  }

  /**
   * If this relation is temporary (e.g. inherited), this method will make it
   * permanent and decouple it from its original source.
   */
  public void confirm()
  {
    if(isInherited())
      makePermanent();
  }

  /**
   * Returns this relation's RelationData object.
   */
  protected RelationData getRelationData()
  {
    return relationData;
  }

  /**
   * Sets the current partition context of this relation.
   *
   * @param partition The partition to set as current.
   */
  void setCurrentPartition(Partition partition)
  {
    currentPartition = partition;
  }

  /**
   * Returns the current partition, if set.
   */
  public Partition getCurrentPartition()
  {
    return currentPartition;
  }

  /**
   * Returns the unique ID of this relation. The kind of object this method returns
   * depends on the implementation of the RelationData method.
   */
  public Object getID()
  {
    return relationData.getID();
  }

  /**
   * Returns a clone of this relation.
   */
  public Object clone()
  {
    Relation rel = new Relation(knowledgeModel, relationData);
    if(originals != null)
      rel.originals = (Vector) originals.clone();
    rel.explanationModel = explanationModel;
    return rel;
  }

  public String toString()
  {
    return getFrom().getName()+" "+getName()+" "+getValue().getName()+(isNegated()?" (negated)":"")+" (strength "+getExplanationStrength()+")";
  }

  public static void main(String argv[])
  {
    try
    {
      LocalKnowledgeModel km = new LocalKnowledgeModel();
      Entity dings = new Entity(km, "Dings", "Dings");
      Entity dill = new Entity(km, "Dill", "Dill");

      Entity ting = new Entity(km, "Ting", "Ting");

      ting.addRelation(BasicModel.SUBCLASS_OF, dings);
      Relation origCauses = dings.addRelation("causes", dill);

      System.out.println(ting.getFrameString());


      Relation causes = ting.getRelation("causes");

      System.out.println(causes+" overridden by: "+java.util.Arrays.asList(causes.getOverridenBy())+" overrides "+java.util.Arrays.asList(causes.getOverridingRelations()));
      System.out.println(origCauses+" overridden by: "+java.util.Arrays.asList(origCauses.getOverridenBy())+" overrides "+java.util.Arrays.asList(origCauses.getOverridingRelations()));

      causes.confirm();
      causes.setExplanationStrength(1.0);
      System.out.println("Persistent? "+causes.isPersistent());
      km.getInferenceCache().flush();

      System.out.println();

      System.out.println(ting.getFrameString());
      Relation[] locals = ting.getLocalRelations(km.getRelationType("Relation"));
      for(int i=0;i<locals.length;i++)
        System.out.println("Local --> "+locals[i]);
      System.out.println();
      locals = ting.getLocalRelations();
      for(int i=0;i<locals.length;i++)
        System.out.println("Local --> "+locals[i]);

      System.out.println();
      System.out.println(causes+" overridden by: "+java.util.Arrays.asList(causes.getOverridenBy())+" overrides "+java.util.Arrays.asList(causes.getOverridingRelations()));
      System.out.println(origCauses+" overridden by: "+java.util.Arrays.asList(origCauses.getOverridenBy())+" overrides "+java.util.Arrays.asList(origCauses.getOverridingRelations()));

      causes.remove();

    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
  }
}



