package jcreek.representation;

import java.io.Serializable;

/**
 * This interface is used to access the data encapsulated by an Relation object.
 * Typically, a KnowledgeModel-implementation will also implement this interface.
 *
 * The contract of this interface specifices that:
 *  - Each RelationData object can be encapsulated by multiple Relation objects.
 *  - The contract does NOT state that the data is stored locally in the RelationData
 * object, and this can not be assumed. This means that the RelationData object can be
 * an interface to a standard SQL database or a concurrently accessed knowledge model
 * server.
 *  - The data represented by an RelationData object must be possible to access concurrently.
 * If an implementation of RelationData allows multiple instances of RelationData to
 * exist at once, the access must allow concurrentl, thread-safe even access to the data
 * across the RelationData instances.
 *  - All manipulations of this object is done either from a KnowledgeModel or Relation
 * object (or a subclass of these). RelationData classes should be viewed as internal
 * classes to the Relation and KnowledgeModel classes.
 *
 * @author Frode Srmo
 * @version 1.0
 */

public interface RelationData
{
  /**
   * Returns a unique ID for the entity this elationData represents.
   */
  public Object getID();

  /**
   * Sets the from-entity (the origin) of the relation.
   *
   * @param from The new from-entity of the relation.
   */
  public void setFrom(Entity from);

  /**
   * Returns the entity from where the relation originates.
   */
  public Entity getFrom();

  /**
   * Sets the value (the target) of the relation.
   *
   * @param value The new value of the relation.
   */
  public void setValue(Entity value);

  /**
   * Returns the entity this relation is pointing at.
   */
  public Entity getValue();

  /**
   * Returns the instance of Entity describing this relation's type.
   */
  public Entity getRelationType();

  /**
   * Sets this relation's type.
   *
   * @param relationType The Entity describing this relation's type.
   */
  public void setRelationType(Entity relationType);

  /**
   * Sets the desciption of this relation.
   *
   * @param description The String containing the description.
   */
  public void setDescription(String description);

  /**
   * Returns the description of this entity.
   */
  public String getDescription();

  /**
   * Returns the time of creation for this relation in the same format as System.getCurrentMillis();
   */
  public long getCreationTime();

  /**
   * Sets the creation time. This is normally not necessary as the creation time
   * is automatically set when the relation is created, but in some odd cases
   * (like the importModel method in KnowledgeModel), it can be required.
   *
   * @param time The time in the format of System.getCurrentMillis() this relation was created.
   */
  public void setCreationTime(long time);

  /**
   * Sets the strength variable contains the explanation strength for this relation.
   * This value must be [0.0..1.0] - any value below this range is set to 0.0, any
   * value above this range is set to 1.0.
   *
   * @param value The explanation strength.
   */
  public void setExplanationStrength(double value);

  /**
   * Returns the explanation strength for this relation.
   */
  public double getExplanationStrength();

  /**
   * This relation is market as overriding another relation.
   *
   * @param override The relation to override
   */
  public void addOverriding(Relation override);

  /**
   * Removes a relation from beeing overriden by this relation.
   *
   * @param override The relation to remove from the overide list.
   */
  public void removeOverriding(Relation override);

  /**
   * Returns the relation (if any) this relation overrides. See setOverriding().
   */
  public Relation[] getOverrides();

  /**
   * Returns true if this relation overrides another relation. See setOverriding().
   */
  public boolean isOverriding();

  /**
   * Returns true if there is at least one relation overriding this relation.
   * See setOverriding().
   */
  public boolean isOverridden();

  /**
   * This relation is market as beeing overriden by another relation.
   *
   * @param override The relation to override
   */
  public void addOverridenBy(Relation override);

  /**
   * Removes a relation from beeing overriden by this relation.
   *
   * @param override The relation to remove from the overide list.
   */
  public void removeOverriddenBy(Relation override);

  /**
   * Returns which relations this relation is overriden by. See setOverriding().
   */
  public Relation[] getOverridenBy();

  /**
   * Permanently removes the relation represented by this RelationData object.
   */
  public void remove();

  /**
   * Sets the inverse of this relation to be the given relation.
   *
   * @param relation The inverse of this relation.
   */
  public void setInverse(Relation inverse);

  /**
   * Returns the inverse of this relation. If it has none, null is returned.
   */
  public Relation getInverse();

  }
