package jcreek.representation;

import java.util.ArrayList;
import java.util.Arrays;
import jcreek.util.CreekException;

/**
 * <p>Title: RelationType</p>
 * <p>Description: Represents a type of relation.
 * @author Frode Srmo
 * @version 1.0
 */

public class RelationType extends EntityType
{
  /**
   * Creates a new relation type which is its own inverse type (i.e. it is
   * bidirectional).
   *
   * @param km The KnowledgeModel this RelationType is a part of.
   * @param superclass The superclass of the new RelationType.
   * @param name The name of the new RelationType.
   * @param description The description of the new RelationType.
   * @param defaultExplanationStrength The explanation strength used for new relations of this type.
   */
  public RelationType(KnowledgeModel km,
               RelationType superclass,
               String name,
               String description,
               double defaultExplanationStrength) throws NameAlreadyExistException
  {
    super(km, name, description);
    try
    {
      addRelation(BasicModel.SUBCLASS_OF, superclass);
      setExplanationStrength(defaultExplanationStrength);
    }
    catch(NoSuchRelationTypeException e)
    {
      // This implies someone did something stupid with the EpistemologicalModel.
      e.printStackTrace(System.err);
    }
  }

  /**
   * Creates a new relation type with an inverse type.
   *
   * @param km The KnowledgeModel this RelationType is a part of.
   * @param superclass The superclass of the new RelationType.
   * @param name The name of the new RelationType.
   * @param description The description of the new RelationType.
   * @param defaultExplanationStrength The explanation strength used when creating new relations of this type.
   * @param inverseName The name of the inverse of this RelationType.
   * @param inverseDesc The description of the inverse of this RelationType.
   * @param inverseDefaultExplanationStrength The explanation strength used for new relations of the inverse type.
   * @param creator The user creating this RelationType.
   */
  public RelationType(KnowledgeModel km,
                      RelationType superclass,
                      String relationName,
                      String description,
                      double defaultExplanationStrength,
                      String inverseName,
                      String inverseDesc,
                      double inverseDefaultExplanationStrength) throws NameAlreadyExistException
  {
    this(km, superclass, relationName, description, defaultExplanationStrength);
    try
    {
      RelationType inverse = new RelationType(km, superclass.getInverseType(), inverseName, inverseDesc, inverseDefaultExplanationStrength);
      addRelation(BasicModel.HAS_INVERSE, inverse);
    }
    catch(NoSuchRelationTypeException e)
    {
      e.printStackTrace(System.err);
    }
  }

  /**
   * Creates a new relation type with an inverse type, where the default explanation strength
   * of both are the same.
   *
   * @param km The KnowledgeModel this RelationType is a part of.
   * @param superclass The superclass of the new RelationType.
   * @param name The name of the new RelationType.
   * @param description The description of the new RelationType.
   * @param inverseName The name of the inverse of this RelationType.
   * @param inverseDesc The description of the inverse of this RelationType.
   * @param defaultExplanationStrength The explanation strength used for new relations of this type and the inverse type.
   * @param creator The user creating this RelationType.
   */
  public RelationType(KnowledgeModel km,
                      RelationType superclass,
                      String relationName,
                      String description,
                      String inverseName,
                      String inverseDesc,
                      double defaultExplanationStrength) throws NameAlreadyExistException
  {
    this(km, superclass, relationName, description, defaultExplanationStrength, inverseName, inverseDesc, defaultExplanationStrength);
  }
  /**
   * This constructor does not create a new relation type in the model, but merely
   * creates a new RelationType object to encapsulate an entity that represents
   * a RelationType already.
   */
  public RelationType(Entity relationTypeEntity) throws EntityTypeNotMatchedException
  {
    super(relationTypeEntity);
  }

  /**
   * Returns the RelationType object for the inverse of this type.
   *
   * This is found by following the "has inverse" relation. If there is no such
   * relation, it is assumed that this type has no inverse (is bidirectional) and
   * thus this RelationType is returned.
   */
  public RelationType getInverseType()
  {
    try
    {
      Relation rt[] = getEntityData().getLocalRelations(new Entity[] { getKnowledgeModel().getEntity(BasicModel.HAS_INVERSE) } );
      if(rt != null && rt.length > 0)
      {
        return new RelationType(rt[0].getValue());
      }
    }
    catch(EntityTypeNotMatchedException e)
     { }
/*    catch(NoSuchRelationException e)
     { }
    catch(NoSuchRelationTypeException e)
    {
      e.printStackTrace(System.err);
    } */
    return this;
  }

  /**
   * Returns true if this relation type has an inverse.
   */
  public boolean hasInverseType()
  {
    return getInverseType() != this;
  }

  /**
   * Returns a complete verb phrase useable for natural language generatation. If
   * a natural name is not defined, the name of the entity is returned.
   */
  public String getNaturalName()
  {
    if(getEntityObject() instanceof String)
    {
      return (String) getEntityObject();
    }
    return getName();
  }

  /**
   * Returns the name of the inverse relation-type.
   */
  public String getInverseName()
  {
    return getInverseType().getName();
  }

  /**
   * Returns the natural name (a verb phrase) of the inverse relation-type.
   */
  public String getNaturalInverseName()
  {
    if(getInverseType().getEntityObject() instanceof String)
    {
      return (String) getInverseType().getEntityObject();
    }
    return getInverseType().getName();
  }


  /**
   * Sets the natural name of this RelationType. This name should be a complete
   * verb phrase. So for instance while "color of" is an acceptable relation type
   * name, the natural language name should also include the auxilry verb and
   * determiner required in this case, i.e. "is the color of". To test, place a noun
   * before and after the phrase and see if the sentence fragment sounds natural.
   *
   * @param name The natural name.
   */
  public void setNaturalName(String name)
  {
    setEntityObject(name);
  }

  /**
   * Sets the default explanation strength of new relations created of this relation type.
   * This strength must be between 1.0 and 0.0.
   * If it is not, it is rounded to the closest value in the range.
   *
   * @param strength A double in the range [0.0, 1.0] that is the new default explanation strength.
   */
  public void setExplanationStrength(double strength)
  {
    if(strength > 1.0)
      strength = 1.0;
    if(strength < 0.0)
      strength = 0.0;

    try
    {
      Relation defExp = getLocalRelation(BasicModel.HAS_DEFAULT_EXPLANATION_STRENGTH);
      try
      {
        DoubleEntity valueEntity = new DoubleEntity(defExp.getValue());
        valueEntity.setValue(new Double(strength));
      }
      catch(EntityTypeNotMatchedException e)
      {
        defExp.setValue(new DoubleEntity(getKnowledgeModel(), strength, ""));
      }
    }
    catch(CreekException e)
    {
      try
      {
        addRelation(BasicModel.HAS_DEFAULT_EXPLANATION_STRENGTH, new DoubleEntity(getKnowledgeModel(), strength, ""));
      }
      catch(NoSuchRelationTypeException ex)
      {
        // This should never happen.
        ex.printStackTrace();
      }
    }
  }

  /**
   * Returns the default explanation strength of new relations created of this type.
   */
  public double getExplanationStrength()
  {
    try
    {
      Relation rel = getLocalRelation(BasicModel.HAS_DEFAULT_EXPLANATION_STRENGTH);
      return new DoubleEntity(rel.getValue()).asDouble().doubleValue();
    }
    catch(CreekException e)
    {
      return 0.01;
    }
  }

  /**
   * Returns the value type of this relation type. The meaning of value type
   * is that relations of this relation type is supposed to have a value that
   * is an instance or subclass of the entity representing the value.
   *
   * @return The value type entity.
   */

  public Entity getValueType()
  {
    try
    {
      Relation rel = getLocalRelation(BasicModel.HAS_VALUE_TYPE);
      return rel.getValue();
    }
    catch(CreekException e)
    {
      return getKnowledgeModel().getRelationType(BasicModel.THING);
    }
  }

  /**
   * Sets the value type of this relation type. The meaning of value type
   * is that relations of this relation type is supposed to have a value that
   * is an instance or subclass of the entity representing the value.
   *
   * @aram The value type entity.
   */

  public void setValueType(Entity type)
  {
    try
    {
      Relation rel = getLocalRelation(BasicModel.HAS_VALUE_TYPE);
      rel.setValue(type);
    }
    catch (CreekException e)
    {
      try
      {
        addRelation(BasicModel.HAS_VALUE_TYPE, type);
      }
      catch (CreekException ex)
      {
        // Should never happen
        ex.printStackTrace();
      }
    }
  }


  /**
   * Removes this relation type. This will also remove the inverse relation and all
   * instances of this relation and the inverse relation.
   */
  public void remove()
  {
    if(hasInverseType())
     getInverseType().removeSingle();
    super.remove();
  }

  private void removeSingle()
  {
    super.remove();
  }


  /**
   * Returns all instances of this relation type and all of its subclasses.
   */
  public Relation[] getInstances()
  {
    ArrayList instances = new ArrayList();
    Entity[] subclasses = getAllSubclasses();
    for(int i=0;i<subclasses.length;i++)
    {
      instances.addAll(Arrays.asList(subclasses[i].getEntityData().getRelationInstances()));
    }
    return (Relation[]) instances.toArray(new Relation[instances.size()]);
  }

  /**
   * Creates a new relation of the type represented by this RelationType
   * between the two supplied relations.
   *
   * @param from The origin entity of the new relation.
   * @param to The value entity of the new relation.
   */
  public Relation newInstance(Entity from, Entity to)
  {
    return new Relation(from, this, to);
  }

  /**
   * Creates a new relation of the type represented by this RelationType
   * between the two supplied relations.
   *
   * @param from The origin entity of the new relation.
   * @param to The value entity of the new relation.
   * @param explanationStrength The strength of the new relation.
   */

  public Relation newInstance(Entity from, Entity to, double explanationStrength)
  {
    return new Relation(from, this, to, explanationStrength);
  }

  /**
   * An Entity is a RelationType if and only if and only if it is an (indirect) subclass of the Relation
   * object.
   */
  public static boolean matches(Entity ent)
  {
//    System.err.print(".");
//    Object key = new Object[] { "_relation_type_matches", ent.getID()};
//    if(ent.getKnowledgeModel().getInferenceCache().contains(key))
//      return ((Boolean) ent.getKnowledgeModel().getInferenceCache().get(key)).booleanValue();

    boolean result = ent != null && ent.getKnowledgeModel().getEntity(BasicModel.RELATION).equals(ent) ||
        ent.getKnowledgeModel().getEntity(BasicModel.SUBCLASS_OF).equals(ent) ||
        ent.getKnowledgeModel().getEntity(BasicModel.HAS_SUBCLASS).equals(ent) ||
        (ent.getKnowledgeModel().getEntity(BasicModel.INSTANCE_OF) != null &&
        ent.getKnowledgeModel().getEntity(BasicModel.INSTANCE_OF).equals(ent)) ||
        (ent.getKnowledgeModel().getEntity(BasicModel.HAS_INSTANCE) != null &&
        ent.getKnowledgeModel().getEntity(BasicModel.HAS_INSTANCE).equals(ent)) ||
        ent.getKnowledgeModel().getEntity(BasicModel.HAS_INVERSE).equals(ent) ||
        ent.getKnowledgeModel().getEntity(BasicModel.HAS_DEFAULT_EXPLANATION_STRENGTH).equals(ent) ||
        ent.getKnowledgeModel().getEntity(BasicModel.DEFAULT_EXPLANATION_STRENGTH_OF).equals(ent) ||
        ent.getKnowledgeModel().getRelationType(BasicModel.RELATION).getAllSubclassesCollection().contains(ent) ||
        ent.getEntityData().getRelationInstances().length > 0;

//    ent.getKnowledgeModel().getInferenceCache().put(key, new Boolean(result));
    return result;
  }
}
