package jcreek.representation;

import java.io.Serializable;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class represents a set of RelationTypes.
 */

public class RelationTypeSet implements Serializable, Cloneable
{
  Vector rules = new Vector(0);

  /**
   * Create an empty RelationTypeSet.
   */
  public RelationTypeSet()
  {
  }

  /**
   * Creates a RelationTypeSet from the supplied RelationType. All
   * direct and indirect subclasses of the relation type represented by the
   * RelationType is included in the set.
   */

  public RelationTypeSet(RelationType rcd)
  {
//    System.out.println("Making RelationTypeSet for "+rcd.toString());
    Object id[] = new Object[] {"RelationTypeSet", this.getClass(), rcd };
    if(rcd.getKnowledgeModel().getInferenceCache().isCached(id))
      rules = (Vector) rcd.getKnowledgeModel().getInferenceCache().get(id);
    else
    {
      rules = new Vector(rcd.getAllSubclassesCollection());
      rules.add(rcd);
      rcd.getKnowledgeModel().getInferenceCache().put(id, rules);
    }
//    System.out.println("New RelationTypeSet: "+toString());
  }



  /**
   * Merges this RelationTypeSet with another RelationTypeSet in such a manner
   * that it will only contain the RelationTypes contained in both
   * sets.
   *
   * Only the RelationTypeSet in which the method is called is modified. The
   * returned RelationTypeSet is actually a reference to the set in which the
   * method is invoked.
   */

  public RelationTypeSet intersect(RelationTypeSet newRule)
  {
    for(int i=rules.size()-1;i>=0;i--)
    {
      if(!newRule.rules.contains(rules.elementAt(i)))
        rules.removeElementAt(i);
    }
    return this;
  }

  /**
   * Merges this RelationTypeSet with another RelationTypeSet in such a manner
   * that it will contain all the RelationTypes contained in either
   * set.
   *
   * Only the RelationTypeSet in which the method is called is modified. The
   * returned RelationTypeSet is actually a reference to the set in which the
   * method is invoked.
   */

  public RelationTypeSet union(RelationTypeSet newRule)
  {
    for(int i=newRule.rules.size()-1;i>=0;i--)
    {
      if(!rules.contains(newRule.rules.elementAt(i)))
        rules.addElement(newRule.rules.elementAt(i));
    }
    return this;
  }

  /**
   * Removes all RelationTypes found in set passed as a parameter
   * to the method. The returned set is guaranteed to have no
   * RelationTypes in common with the set passed to the method as
   * a parameter.
   *
   * Only the RelationTypeSet in which the method is called is modified. The
   * returned RelationTypeSet is actually a reference to the set in which the
   * method is invoked.
   */

  public RelationTypeSet subtract(RelationTypeSet newRule)
  {
    for(int i=rules.size()-1;i>=0;i--)
    {
      if(newRule.rules.contains(rules.elementAt(i)))
        rules.removeElementAt(i);
    }
    return this;
  }

  /**
   * Returnes true if this set is a subset of the supplied set.
   */

  public boolean isSubsetOf(RelationTypeSet rule)
  {
//    System.out.println("Comparing "+rule.toString());
//    System.out.println("   to "+this.toString());

    for(int i=rules.size()-1;i>=0;i--)
    {
      if(!rule.rules.contains(rules.elementAt(i)))
        return false;
    }
    return true;
  }

  /**
   * Adds the supplied RelationType to the set, but does not add
   * subclasses of the RelationType.
   */

  public void addElement(RelationType rcd)
  {
    rules.addElement(rcd);
  }

  /**
   * Removes the supplied RelationType to the set, but does not add
   * subclasses of the RelationType.
   */

  public void removeElement(RelationType rcd)
  {
    rules.removeElement(rcd);
  }

  /**
   * Returnes true if the set contains the supplied RelationType.
   */

  public boolean contains(RelationType rcd)
  {
    return rules.contains(rcd);
  }

  /**
   * Returnes true if the set is empty.
   */

  public boolean empty()
  {
    if(rules.size() == 0)
      return true;
    else
      return false;
  }

  protected Object clone()
  {
      RelationTypeSet tr = new RelationTypeSet();
      tr.rules = (Vector) rules.clone();
      return tr;
  }

  public RelationTypeSet copy()
  {
      return (RelationTypeSet) clone();
  }

  public String toString()
  {
    String ret = "";
    for(int i=rules.size()-1;i>=0;i--)
      ret += " "+((RelationType) rules.elementAt(i));
//      +"(hc "+((RelationType) rules.elementAt(i)).hashCode()+")";
    return "Transfer Rule ["+ret+"]";

  }

  public boolean equals(Object ob)
  {
    return ob.getClass().equals(getClass()) && ((RelationTypeSet) ob).isSubsetOf(this) && isSubsetOf((RelationTypeSet) ob);
  }

  public int hashCode()
  {
    int c = 19;
    for(int i=0;i<rules.size();i++)
      c += rules.get(i).hashCode();
    return c;
  }
}

