package jcreek.representation;

import java.util.HashSet;
import java.util.Set;
import java.util.Arrays;
import java.util.Iterator;

/**
 * Represents a submodel of a knowledgemodel. This means a set of entities and
 * relations. Each relation of entity may only exists once in the set.
 *
 * This class is not thread-safe.
 *
 * @author Frode Srmo
 * @version 2.0
 */

public class SubModel
{
  private HashSet entities;
  private HashSet relations;

  /**
   * Creates a new SubModel from an array of entities and relations to be
   * included in the submodel.
   */
  public SubModel(Entity[] ents, Relation[] rels)
  {
    entities = new HashSet(ents.length);
    add(ents);

    relations = new HashSet(rels.length);
    add(rels);
  }

  /**
   * Creates an empty submodel.
   */
  public SubModel()
  {
    entities = new HashSet();
    relations = new HashSet();
  }

  /**
   * Adds an entity to the submodel.
   *
   * @param e The entity to add.
   */
  public void add(Entity e)
  {
    if(e != null)
      entities.add(e);
  }

  /**
   * Adds a set of entitites to the submodel.
   *
   * @param ents The arrays of entities to add.
   */
  public void add(Entity[] ents)
  {
    if (ents != null)
      entities.addAll(Arrays.asList(ents));
  }

  /**
   * Adds a relation to the submodel.
   *
   * @param r The relation to add.
   */
  public void add(Relation r)
  {
    if(r != null)
      relations.add(r);
  }

  /**
   * Adds an array of relations to the submodel.
   *
   * @param rels The array of relations to add.
   */
  public void add(Relation[] rels)
  {
    if (rels != null)
    {
      relations.addAll(Arrays.asList(rels));
    }
  }

  /**
   * Removes a relation from the submodel. Return false if the relation
   * is not present in the submodel.
   */
  public boolean remove(Relation r)
  {
    return relations.remove(r);
  }

  /**
   * Adds a relation, but if the relation is composite, the submodel of containing
   * all relations of the composite structure is added instead.
   *
   * @param r The relation to (expand and) add.
   */
  public void addExpanded(Relation r)
  {
    if (r != null)
    {
      if (r.isInherited())
        union(r.getPathModel());
      else
        add(r);
    }
  }

  /**
   * Adds an array of relations to the submodel, but if the relation is composite,
   * the submodel of containing all relations of the composite structure is added
   * instead.
   */
  public void addExpanded(Relation[] rels)
  {
    if (rels != null)
    {
      for (int i = rels.length - 1; i >= 0; i--)
        addExpanded(rels[i]);
    }
  }

  /**
   * Removes an entity from the submodel. Return false if the relation
   * is not present in the submodel.
   */
  public boolean remove(Entity r)
  {
    return entities.remove(r);
  }

  /**
   * Returns true if the submodel contains the entity.
   */
  public boolean contains(Entity ent)
  {
    return entities.contains(ent);
  }

  /**
   * Returns true if the submodel contains the relation.
   */
  public boolean contains(Relation rel)
  {
    return relations.contains(rel);
  }

  /**
   * Returns an array of all the entities in the submodel.
   */
  public Entity[] getEntities()
  {
    return (Entity[]) entities.toArray(new Entity[0]);
  }

  /**
   * Returns an array of all the relations in the submodel.
   */
  public Relation[] getRelations()
  {
    return (Relation[]) relations.toArray(new Relation[relations.size()]);
  }

  /**
   * Return the number of entities and relations in the submodel.
   */
  public int size()
  {
    return relations.size();
  }

  /**
   * Returns true if model specified is submodel of this submodel,
   * i.e. if all the entities and all of the relations of the specified
   * model is also contained in this model.
   */
  public boolean contains(SubModel model)
  {
    Entity[] ents = model.getEntities();
    for (int i = ents.length - 1; i >= 0; i--)
      if (!contains(ents[i]))
        return false;

    Relation[] rels = model.getRelations();
    for (int i = rels.length - 1; i >= 0; i--)
      if (!contains(rels[i]))
        return false;

    return true;
  }

  /**
   * Changes this submodel to be the union of the specified model
   * and this submodel. The specified model is not changed.
   * I.e. this model will only contain the relations and entities contained
   * in both models. A reference to this model will be returned.
   */
  public SubModel union(SubModel other)
  {
    add(other);
    return this;
  }

  /**
   * Adds all relations and entities in the supplied submodel to this submodel.
   * The supplied submodel is not changed.
   *
   * @param other The submodel to add to this model.
   */
  public void add(SubModel other)
  {
    Entity[] ents = other.getEntities();
    Relation[] rels = other.getRelations();

    for (int i = ents.length - 1; i >= 0; i--)
      add(ents[i]);
    for (int i = rels.length - 1; i >= 0; i--)
      add(rels[i]);
  }

  /**
   * Intersects the specified model with this model. This model will be changed to
   * become the new submodel. The specified model will not be changed.
   * I.e. this model will now contain all the relations and entities of both the
   * specified model and this model. A reference to this model is returned.
   *
   * This method is deprecated because the name is really not descriptive.
   *  Use union(SubModel model) instead.
   *
   * @deprecated
   */
  public SubModel joinWith(SubModel other)
  {
    add(other);
    return this;
  }

  /**
   * Returns a Set that contains all the relations and entities in this model.
   * @return Set
   */
  public Set toSet()
  {
    HashSet ret = new HashSet();
    ret.addAll(entities);
    ret.addAll(relations);
    return ret;
  }
  /**
   * Removes all the relations and entities in the supplied model from this model.
   * If some relations and entities did not exist in this model, nothing will be done.
   * This model is effectively changed to be the intersection between the supplied
   * model and this model before the remove() operation. The supplied model is not
   * changed.
   *
   * @param other The model to intersect with this model.
   */
  public void remove(SubModel other)
  {
    HashSet newEnts = new HashSet();
    HashSet newRels = new HashSet();
    Entity nextEntity = null;
    for(Iterator it=entities.iterator();it.hasNext();nextEntity = (Entity) it.next())
      if(nextEntity != null && other.contains(nextEntity))
        newEnts.add(nextEntity);
    Relation nextRel = null;
    for(Iterator it=relations.iterator();it.hasNext();nextRel = (Relation) it.next())
      if(nextRel != null && other.contains(nextRel))
        newRels.add(nextRel);

    entities = newEnts;
    relations = newRels;
  }

  /**
   * Removes all contents (relations and entities) from the submodel.
   */
  public void clear()
  {
    Relation rels[] = getRelations();
    for(int i=0;i<rels.length;i++)
    {
      remove(rels[i]);
    }
    Entity ents[] = getEntities();
    for(int i=0;i<ents.length;i++)
    {
      remove(ents[i]);
    }
  }


  public String toString()
  {
    String res = "Entities:\n";
    Entity[] ents = getEntities();
    for(int i=ents.length-1;i>=0;i--)
      res += "   "+ents[i].toString()+"\n";
    res +="Relations:\n";
    Relation[] rels = getRelations();
    for(int i=rels.length-1;i>=0;i--)
      res += "   "+rels[i].toString()+"\n";
    return res;
  }
}



