package jcreek.representation;

/**
 * Represents a the data in a temporary relation, like in an inherited relation.
 *
 *
 * @author Frode Srmo
 * @version 1.0
 */


import java.util.*;
import java.io.Serializable;


public class TemporaryRelationData implements RelationData, Serializable
{
  private String id;
  private Entity from;
  private Entity value;
  private Entity type;
  private long creationTime;
  private ArrayList overrides;
  private ArrayList overriddenBy;
  private Relation inverse;
  private String description;
  private double explanationStrength;

  private static int temporaryCount = 0;

  TemporaryRelationData(Entity from, Entity value, Entity type)
  {
    this.from = from;
    this.value = value;
    this.type = type;
    this.explanationStrength = -1;
    overrides = new ArrayList();
    overriddenBy = new ArrayList();
    synchronized(this.getClass())
    {
      temporaryCount++;
      id = "_temp#"+temporaryCount;
    }
  }

  /**
   * Returns a unique ID for the entity this relationData represents.
   */
  public Object getID()
  {
    return id;
  }

  /**
   * Sets the from-entity (the origin) of the relation.
   *
   * @param from The new from-entity of the relation.
   */
  public void setFrom(Entity from)
  {
    this.from = from;
  }

  /**
   * Returns the entity from where the relation originates.
   */
  public Entity getFrom()
  {
    return from;
  }

  /**
   * Sets the value (the target) of the relation.
   *
   * @param value The new value of the relation.
   */
  public void setValue(Entity value)
  {
    this.value = value;
  }

  /**
   * Returns the entity this relation is pointing at.
   */
  public Entity getValue()
  {
      return value;
  }

  /**
   * Returns the instance of Entity describing this relation's type.
   */
  public Entity getRelationType()
  {
      return type;
  }

  /**
   * Sets this relation's type.
   *
   * @param relationType The Entity describing this relation's type.
   */
  public void setRelationType(Entity relationType)
  {
      this.type = relationType;
  }

  /**
   * Sets the desciption of this relation.
   *
   * @param description The String containing the description.
   */
  public void setDescription(String description)
  {
    this.description = description;
  }

  /**
   * Returns the description of this entity.
   */
  public String getDescription()
  {
    return description;
  }

  /**
   * Returns the time of creation for this relation in the same format as System.getCurrentMillis();
   */
  public long getCreationTime()
  {
    return creationTime;
  }

  /**
   * Sets the creation time. T
   *
   * @param time The time in the format of System.getCurrentMillis() this relation was created.
   */
  public void setCreationTime(long time)
  {
    this.creationTime = time;
  }

  /**
   * Sets the strength variable contains the explanation strength for this relation.
   *
   * @param value The explanation strength.
   */
  public void setExplanationStrength(double value)
  {
    this.explanationStrength = value;
  }

  /**
   * Returns the explanation strength for this relation.
   */
  public double getExplanationStrength()
  {
    return explanationStrength;
  }

  /**
   * This relation is market as overriding another relation.
   *
   * @param override The relation to override
   */
  public synchronized void addOverriding(Relation override)
  {
    overrides.add(override);
  }

  /**
   * Removes a relation from beeing overriden by this relation.
   *
   * @param override The relation to remove from the overide list.
   */
  public synchronized void removeOverriding(Relation override)
  {
    overrides.remove(override);
  }

  /**
   * Returns the relation (if any) this relation overrides. See setOverriding().
   */
  public synchronized Relation[] getOverrides()
  {
    return (Relation[]) overrides.toArray(new Relation[overrides.size()]);
  }

  /**
   * Returns true if this relation overrides another relation. See setOverriding().
   */
  public synchronized boolean isOverriding()
  {
    return !overrides.isEmpty();
  }

  /**
   * Returns true if there is at least one relation overriding this relation.
   * See setOverriding().
   */
  public synchronized boolean isOverridden()
  {
    return !overriddenBy.isEmpty();
  }

  /**
   * This relation is market as beeing overriden by another relation.
   *
   * @param override The relation to override
   */
  public synchronized void addOverridenBy(Relation override)
  {
    overriddenBy.add(override);
  }

  /**
   * Removes a relation from beeing overriden by this relation.
   *
   * @param override The relation to remove from the overide list.
   */
  public synchronized void removeOverriddenBy(Relation override)
  {
    overriddenBy.remove(override);
  }

  /**
   * Returns which relations this relation is overriden by. See setOverriding().
   */
  public synchronized Relation[] getOverridenBy()
  {
    return (Relation[]) overriddenBy.toArray(new Relation[overriddenBy.size()]);
  }

  /**
   * Does nothing in a temporary relation data object since it is not registered.
   */
  public void remove()
  {
  }

  /**
   * Sets the inverse of this relation to be the given relation.
   *
   * @param relation The inverse of this relation.
   */
  public void setInverse(Relation inverse)
  {
    this.inverse = inverse;
  }

  /**
   * Returns the inverse of this relation. If it has none, null is returned.
   */
  public Relation getInverse()
  {
    return inverse;
  }
}
