package jcreek.representation.cbr;

/**
 * A minimal epistemological model for using the case-based reasoning (CBR)
 * packages of jcreek (jcreek.reasoning) .
 *
 * @author Frode Srmo
 * @version 1.0
 */

import jcreek.representation.*;
import jcreek.util.CreekException;

public class CBRModel extends BasicModel
{
  public static final String HAS_INVERSE = "_has inverse";
  public static final String HAS_COMPARATOR = "_has comparator";
  public static final String COMPARATOR_OF ="_comparator of";
  public static final String HAS_CASE_STATUS = "_has case status";
  public static final String CASE_STATUS_OF = "_case status of";
  public static final String HAS_SOLUTION = "_has solution";
  public static final String SOLUTION_OF = "_solution of";

  public static final String FINDING_RELATION = "_finding relation";
  public static final String INVERSE_FINDING_RELATION = "_inverse finding relation";

  public static final String CAUSAL_RELATION = "_causal relation";
  public static final String INVERSE_CAUSAL_RELATION = "_inverse causal relation";
  public static final String CAUSES = "_causes";
  public static final String CAUSED_BY = "_caused by";
  public static final String IMPLIES = "_implies";
  public static final String IMPLIED_BY = "_implied by";

  public static final String CASE = "_case";
  public static final String JAVA_CLASS = "_java class";

  public static final String CASE_STATUS = "_case_status";
  public static final String PROCESSED_CASE = "_processed case";
  public static final String SOLVED_CASE = "_solved case";
  public static final String UNSOLVED_CASE = "_unsolved case";

  public static final String NUMBER_COMPARISON = "_number comparison";
  public static final String CASE_COMPARISON = "_case comparison";
  public static final String SYMBOL_COMPARIOSON = "_symbol comparison";

  public static final String STRING = "_string";
  public static final String URL = "_url";


  public void generateEpistemologicalModel(KnowledgeModel model)
  {
    generateCBRModel(model);
  }

  public static void generateCBRModel(KnowledgeModel model)
  {
    generateBasicModel(model);
    try
    {
      RelationType relation = model.getRelationType(RELATION);
      RelationType hasComparator = new RelationType(model, relation, "has comparator", "The target entity encapsulates a java class that implements a comparator for the origin class.",
        "comparator of", "The target class is a type of entity compared by the encapsulated class of the origin entity.", 1.0);
      model.registerEntity(HAS_COMPARATOR, hasComparator);
      model.registerEntity(COMPARATOR_OF, hasComparator.getInverseType());

      RelationType hasCaseStatus = new RelationType(model, relation, "has case status", "The target entity represents the status of the case (the origin).",
        "case status of", "The target is a case the origin is the status for.", 1.0);
      model.registerEntity(HAS_CASE_STATUS, hasCaseStatus);
      model.registerEntity(CASE_STATUS_OF, hasCaseStatus.getInverseType());

      RelationType hasSolution = new RelationType(model, relation, "has solution", "The target is the solution of the origin case.",
        "solution of", "The target is a case the origin is the solution for.", 1.0);
      model.registerEntity(HAS_SOLUTION, hasSolution);
      model.registerEntity(SOLUTION_OF, hasSolution.getInverseType());

      RelationType findingRelation = new RelationType(model, relation, "has finding", "Relation from a case to a finding.",
          "finding of", "Relation from a finding to a case", 0.5);
      model.registerEntity(FINDING_RELATION, findingRelation);
      model.registerEntity(INVERSE_FINDING_RELATION, findingRelation.getInverseType());

      RelationType causalRelation = new RelationType(model, relation, "Causal Relation", "A superclass for causal relations.",
          "Causal Relation (inverse)", "Superclass for invese causal relations.", 0.5);
      model.registerEntity(CAUSAL_RELATION, causalRelation);
      model.registerEntity(INVERSE_CAUSAL_RELATION, causalRelation.getInverseType());

      RelationType causes = new RelationType(model, causalRelation, "causes", "The target is caused by the origin",
          "caused by", "The target causes the origin", 0.7);
      model.registerEntity(CAUSES, causes);
      model.registerEntity(CAUSED_BY, causes.getInverseType());

      RelationType implies = new RelationType(model, causalRelation, "implies", "The target is implied (statistically) by the origin",
          "implied by", "The target implies (statistically) the origin", 0.4);
      model.registerEntity(IMPLIES, implies);
      model.registerEntity(IMPLIED_BY, implies.getInverseType());

      // Creating the entity hierarchy.
      Entity entity = model.getEntity(ENTITY);
      Entity symbol = model.getEntity(SYMBOL);
      Entity number = model.getEntity(NUMBER);

      Entity creekCase = new Entity(model, "Case", "");
      creekCase.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(CASE, creekCase);

      Entity string = new Entity(model, "String", "");
      string.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(STRING, string);

      Entity url = new Entity(model, "URL", "");
      url.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(URL, url);

      Entity javaClass = new Entity(model, "Java Class", "");
      javaClass.addRelation(SUBCLASS_OF, entity);
      model.registerEntity(JAVA_CLASS, javaClass);

      Entity caseStatus = new Entity(model, "Case Status", "");
      caseStatus.addRelation(SUBCLASS_OF, symbol);
      model.registerEntity(CASE_STATUS, caseStatus);

      Entity processedCase = new Entity(model, "Processed Case", "");
      processedCase.addRelation(INSTANCE_OF, caseStatus);
      model.registerEntity(PROCESSED_CASE, processedCase);

      Entity solvedCase = new Entity(model, "Solved Case", "");
      solvedCase.addRelation(INSTANCE_OF, caseStatus);
      model.registerEntity(SOLVED_CASE, solvedCase);

      Entity unsolvedCase = new Entity(model, "Unsolved Case", "");
      unsolvedCase.addRelation(INSTANCE_OF, caseStatus);
      model.registerEntity(UNSOLVED_CASE, unsolvedCase);

      Entity comparator = new Entity(model, "Comparator", "");
      comparator.addRelation(SUBCLASS_OF, javaClass);

      Entity caseComparison = new ClassEntity(model, jcreek.reasoning.CaseComparison.class, "");
      caseComparison.addRelation(INSTANCE_OF, comparator);
      caseComparison.addRelation(COMPARATOR_OF, creekCase);

      Entity numberComparison = new ClassEntity(model, jcreek.reasoning.NumberComparison.class, "");
      numberComparison.addRelation(INSTANCE_OF, comparator);
      numberComparison.addRelation(COMPARATOR_OF, number);

      Entity stringComparison = new ClassEntity(model, jcreek.reasoning.StringComparison.class, "");
      stringComparison.addRelation(INSTANCE_OF, comparator);
      stringComparison.addRelation(COMPARATOR_OF, string);

      Entity urlComparison = new ClassEntity(model, jcreek.reasoning.URLComparison.class, "");
      urlComparison.addRelation(INSTANCE_OF, comparator);
      urlComparison.addRelation(COMPARATOR_OF, url);

      Entity symbolComparison = new ClassEntity(model, jcreek.reasoning.SymbolComparison.class, "");
      symbolComparison.addRelation(INSTANCE_OF, comparator);
      symbolComparison.addRelation(COMPARATOR_OF, symbol);

    }
    catch(CreekException e)
    {
      e.printStackTrace();
    }
  }
}
