package jcreek.representation.cbr;

import java.util.Collection;
import jcreek.representation.*;
import jcreek.util.CreekException;

/**
 * A case is a particular type of entity that represents a case (in the Case-
 * Based Reasoning sense).
 *
 * @author Frode Srmo
 * @version 1.0
 */

public class Case extends EntityType
{
  public final static int SOLVEDCASE = 2;
  public final static int PROCESSCASE = 1;
  public final static int UNSOLVEDCASE = 0;

  /**
   * Creates a new Case.
   *
   * @param model The KnowledgeModel to create the new entity in.
   * @param name The name of the case to create.
   * @param description The description of the entity.
   */
  public Case(KnowledgeModel model, String name, String description) throws NameAlreadyExistException
  {
    this(model, name, description, model.getEntity(CBRModel.CASE));
    setStatus(UNSOLVEDCASE);
  }

  /**
   * Creates a new Case.
   *
   * @param model The KnowledgeModel to create the new entity in.
   * @param name The name of the case to create.
   * @param description The description of the entity.
   * @param caseType The entity representing the type of case this is.
   */
  public Case(KnowledgeModel model, String name, String description, Entity caseType) throws NameAlreadyExistException
  {
    super(model, name, description);
    setStatus(UNSOLVEDCASE);
    try
    {
      addRelation(CBRModel.INSTANCE_OF, caseType);
    }
    catch(NoSuchRelationTypeException e)
    {
      // Should never happen.
      e.printStackTrace();
    }
  }

  /**
   * Creates a Case from an existing Entity.
   */
  public Case(Entity ent) throws EntityTypeNotMatchedException
  {
    super(ent);
  }

  /**
   * Returns the solved status of this Case
   */
  public int getStatus()
  {
    try
    {
      Relation status = getRelation(CBRModel.HAS_CASE_STATUS);
      if(status.getValue().equals(getKnowledgeModel().getEntity(CBRModel.PROCESSED_CASE)))
        return PROCESSCASE;
      if(status.getValue().equals(getKnowledgeModel().getEntity(CBRModel.SOLVED_CASE)))
        return SOLVEDCASE;
    }
    catch(CreekException e)
    { }
    return UNSOLVEDCASE;
  }


  /**
   * Returns the solved status of this Case as an english string
   */
  public String getStatusString()
  {
    try
    {
      Relation status = getRelation(CBRModel.HAS_CASE_STATUS);
      if(status.getValue().equals(getKnowledgeModel().getEntity(CBRModel.PROCESSED_CASE)))
        return getKnowledgeModel().getEntity(CBRModel.PROCESSED_CASE).getName();
      if(status.getValue().equals(getKnowledgeModel().getEntity(CBRModel.SOLVED_CASE)))
        return getKnowledgeModel().getEntity(CBRModel.SOLVED_CASE).getName();
    }
    catch(CreekException e)
    { }
    return getKnowledgeModel().getEntity(CBRModel.UNSOLVED_CASE).getName();
  }

  /**
   * Returns the solution of this Case. If the case is unsolved or no solution
   * is registered, null is returned.
   *
   * @return Entity
   */
  public Entity getSolution()
  {
    if(getStatus() == UNSOLVEDCASE)
      return null;
    try
    {
      Relation hasSolution = getRelation(CBRModel.HAS_SOLUTION);
      return hasSolution.getValue();
    }
    catch(CreekException e)
    {
      return null;
    }
  }

  /**
   * Sets the solution of the case. A case may only have one solution using
   * this method.
   *
   * @param solution The entity representing the solution of the case.
   */
  public void setSolution(Entity solution)
  {
    Relation[] solutions = getRelations(CBRModel.HAS_SOLUTION);
    for(int i=solutions.length-1;i>=0;i--)
      solutions[i].remove();

    addRelation(getKnowledgeModel().getRelationType(CBRModel.HAS_SOLUTION), solution);
  }

  public void setStatus(int status)
  {
    Entity statusEntity = getKnowledgeModel().getEntity(CBRModel.UNSOLVED_CASE);
    try
    {

      if(status == SOLVEDCASE)
        statusEntity = getKnowledgeModel().getEntity(CBRModel.SOLVED_CASE);
      else if(status == PROCESSCASE)
        statusEntity = getKnowledgeModel().getEntity(CBRModel.PROCESSED_CASE);

      Relation statusRel = getRelation(CBRModel.HAS_CASE_STATUS);
      statusRel.setValue(statusEntity);
    }
    catch(CreekException e)
    {
      try
      {
        addRelation(CBRModel.HAS_CASE_STATUS, statusEntity);
      }
      catch(CreekException ex)
      {
        // This should never happen.
        ex.printStackTrace();
      }
    }
  }

  /**
   * Adds a Relation for a Case.
   * <p>
   * Added in the Case class for the ability to set Explanation and Predictive
   * Strength in an easy manner.
   * <p>
   * Stein L. Tomassen (30-05-2002)
   *
   * @param relationType the name of relationtype.
   * @param value the reference to the entity to create relation to.
   * @param importance the importance strength for this relation
   * @param predictiveStrength the predictive strength for this relation
   */
  public Relation addFinding(String relationType, Entity value, double importance, double predictiveStrength) throws NoSuchRelationTypeException
  {
    Relation rel = this.addRelation(relationType, value, importance);
    rel.getInverse().setExplanationStrength(predictiveStrength);
    return rel;
  } // addCaseRelation

  /**
   * An Entity is a RelationType if and only if and only if it is an (indirect) subclass of the Relation
   * object.
   */
  public static boolean matches(Entity ent)
  {
    Entity caseEnt = ent.getKnowledgeModel().getEntity(CBRModel.CASE);
    if(caseEnt != null)
    {
      Collection subclasses = caseEnt.getAllSubclassesCollection();
      if (subclasses.contains(ent))
        return true;
    }
    return false;
  }
}
