package jcreek.representation.inference;

import java.util.Vector;
import jcreek.representation.Entity;
import jcreek.representation.Relation;
import jcreek.representation.SubModel;

/**
 * Represents an explanation between two entities.
 * This type of explanation is a set of linked relations.
 */

public class Explanation
{
  Relation[] relations = new Relation[0];

  /**
   * Creates a new explanation from a set of relations.
   * The relations are assumed to be linked.
   */
  Explanation(Relation[] relations)
  {
    this.relations = relations;
  }

  /**
   * Returns the strength of this particular explanation.
   */
  public double getStrength()
  {
    double strength = 1.0;
    for(int i=relations.length-1;i>=0;i--)
      strength = strength * relations[i].getExplanationStrength();

    return strength;
  }

  public String toString()
  {
    if(relations.length < 1)
    {
      return "[Empty explanation]";
    }


    return text()+" (Strength is "+getStrength()+".)";
  }

  /**
   * Returns a submodel containing all relations and entities in this explanation.
   */
  public SubModel toSubModel()
  {
    SubModel me = new SubModel();

    for(int i=relations.length-1;i>=0;i--)
    {
      me.add(relations[i].getFrom());
      me.add(relations[i].getValue());
      me.add(relations[i]);
    }

    return me;
  }

  public boolean contains(Relation rel)
  {
    return toSubModel().contains(rel);
  }

  public boolean contains(Entity ent)
  {
    return toSubModel().contains(ent);
  }

  public static SubModel toSubModel(Explanation[] explanations)
  {
    SubModel activated = new SubModel();
    for(int i=0;i<explanations.length;i++)
    {
      for(int j=0;j<explanations[i].size();j++)
      {
        activated.add(explanations[i].relationAt(j).getFrom());
        activated.add(explanations[i].relationAt(j).getValue());
        activated.add(explanations[i].relationAt(j));

      }
    }
    return activated;
  }
  /**
   * Adds a relation to the end of the exlanation chain.
   */
  public void add(Relation r)
  {
    Relation[] newRels = new Relation[relations.length+1];
    newRels[relations.length] = r;
    for(int i=relations.length-1;i>=0;i--)
    {
      newRels[i] = relations[i];
    }
    relations = newRels;
  }

  /**
   * Returns a natural language version of this explanation.
   */
  public String text()
  {
    if(relations.length < 1)
      return "empty explanation!";

    String ret = "the "+relations[0].getFrom().getName();
    Entity last = relations[0].getFrom();
    for(int i=0;i<relations.length && last != null;i++)
    {
      if(i>0)
        ret = ret + " that";

        ret = ret + " "+relations[i].getNaturalName()+" "+relations[i].getValue().getValueString();
        if(Entity.class.isAssignableFrom(relations[i].getValue().getClass()))
          last = (Entity) relations[i].getValue();
        else
          last = null;

    }
    return ret;
  }

  public int size()
  {
    return relations.length;
  }

  public Relation relationAt(int i)
  {
    return relations[i];
  }
}



