package jcreek.representation.inference;

import java.util.*;
import jcreek.representation.*;

import jcreek.representation.inference.SpreadingActivationRule;

/**
 * This class represents a subset of the knowledge model resulting from a
 * spreading activation operation from the "origin" entity.
 *
 * This class is created by a SpreadingActivationRule and is not to be created directly.
 */

public class ExplanationModel extends SubModel
{
  Entity          origin;
  Hashtable       explanationStrengths = null;
  SpreadingActivationRule rule;
  KnowledgeModel  knowledgeModel;

  public double MINIMUM_STRENGTH_TRESHOLD = 0.001;

  /**
   * Creates an empty model, with only the origin entity.
   */
  ExplanationModel(Entity origin, SpreadingActivationRule rule)
  {
    super();
    this.origin = origin;
    this.rule = rule;
    knowledgeModel = origin.getKnowledgeModel();
    add(origin);
  }

  /**
   * Returns the explanation representing the strongest path from the origin entity
   * to the target entity.
   *
   * This implementation is not optimal - it can be done using a variation of Dijkstra's
   * shortest path algorithm to avoid the complexity of getPaths();
   */
  public Explanation getStrongestExplanation(Entity target)
  {
    Explanation[] paths = new Explanation[0];
    try
    {
      paths = getExplanations(target, 10000000);
    }
    catch(TaskTooComplexException e)
    {
       System.err.println("Too complex to find explanation for "+origin.getName()+" "+target.getName());
    }

    Explanation res = null;
    double strength = -0.5;

    for(int i=paths.length-1;i>=0;i--)
      if(paths[i].getStrength() > strength)
      {
        strength = paths[i].getStrength();
        res = paths[i];
      }

    return res;
  }

  /**
   * Returns the explanation representing the strongest path from the origin
   * entity to the target relation. In cases where the relation is inherited,
   * this path is the single strongest.
   *
   * This implementation is not optimal - it can be done using a variation of Dijkstra's
   * shortest path algorithm to avoid the complexity of getPaths();
   */

  public Explanation getStrongestExplanation(RelationType rcd, Entity target)
  {
    Explanation[] paths = new Explanation[0];
    try
    {
      paths = getExplanations(rcd, target, 1000000000);
    }
    catch(TaskTooComplexException e)
    {
      System.err.println("Too complex!!!");
    }

//    System.err.println("Size of explanations is "+paths.length);

    Explanation res = null;
    double strength = -0.5;

    for(int i=paths.length-1;i>=0;i--)
      if(paths[i].getStrength() > strength)
      {
        strength = paths[i].getStrength();
        res = paths[i];
      }

    return res;
  }
  /**
   * Exhaustive, depth-first search for paths within the submodel. As this is very computationally
   * though, it demands an integer of how high complexity is accetable and may throw
   * a TaskTooComplexException if the task exceeds the accepted complexity.
   *
   * @param target  The entity to look for paths to. (All the paths originate at the origin entity of
   *                the SpreadingActivationModel.)
   */

  public Explanation[] getExplanations(Entity target, int acceptedComplexity) throws TaskTooComplexException
  {
    int complexity = assessGetPathsComplexity();
//    if(complexity > acceptedComplexity)
//      throw new TaskTooComplexException("Explanation generation is too complex. Assessed complexity is "+complexity+".", complexity);

//    Vector explanations = new Vector();
    Vector path = new Vector();
    Vector entityPath = new Vector();
    entityPath.addElement(origin);
//    System.out.println("-- NEW:\n");

    Vector explanations = strongestPriFindPath(origin, target, new RelationType[] { target.getKnowledgeModel().getRelationType(BasicModel.RELATION) });
    Vector approvedExplanations = new Vector();

  /*  System.out.println("Found "+explanations.size()+ " explanation paths from "+origin.toString()+ " to "+target.toString());
    for(int i=explanations.size()-1;i>=0;i--)
    {
      Explanation current = (Explanation) explanations.elementAt(i);
      System.out.println("Explanation no: "+i+ ": "+current);

//      for(int j=0;j<current.size();j++)
//        System.out.println("  -> "+current.relationAt(j).toString());

      if(rule.accept(current))
      {
        approvedExplanations.addElement(current);
        System.out.println("\tAPPROVED");
      }
    }

    explanations = new Vector();
    approvedExplanations = new Vector();

    System.out.println("-- OLD:\n");

    recFindPath(origin, target, path, entityPath, explanations, "", 1.0);

//    System.out.println("NUMBER of explantions found: "+explanations.size());



    System.out.println("Found "+explanations.size()+ " explanation paths from "+origin.toString()+ " to "+target.toString());
    for(int i=explanations.size()-1;i>=0;i--)
    {
      Explanation current = (Explanation) explanations.elementAt(i);
      System.out.println("Explanation no: "+i+ ": "+current);

//      for(int j=0;j<current.size();j++)
//        System.out.println("  -> "+current.relationAt(j).toString());

      if(rule.accept(current))
      {
        approvedExplanations.addElement(current);
        System.out.println("\tAPPROVED");
      }
    }
    explanations = approvedExplanations;
*/
    Vector lastRels = new Vector();
    for(int i=explanations.size()-1;i>=0;i--)
    {
      Explanation current = (Explanation) explanations.get(i);
      if(current.size() > 0)
      {
        Relation rel = current.relationAt(current.size()-1);
        lastRels.add(rel);
      }
    }

//    System.out.println("Explanation model paths checking");
    approvedExplanations = (Vector) explanations.clone();

    for(int i=explanations.size()-1;i>=0;i--)
    {
      Explanation current = (Explanation) explanations.get(i);
      if(current.size() > 0)
      {
        Relation rel = current.relationAt(current.size()-1);

//        System.out.println("  -> "+rel.toString());
        if(rel.isOverridden())
        {
          boolean remove = false;
          Relation[] overriders = rel.getOverridenBy();
          for(int j=overriders.length-1;j>=0 && !remove;j--)
          {
            if(lastRels.contains(overriders[j]))
              remove = true;
          }
          if(remove)
          {
//            System.out.println("  ->   Explanation Relation "+rel+" is overridden. Removing.");
            approvedExplanations.removeElement(current);
          }
        }
      }
    }

    Explanation[] expArray = new Explanation[approvedExplanations.size()];
    approvedExplanations.copyInto(expArray);
    return expArray;
  }

  /**
   * Exhaustive, depth-first search for paths within the submodel. As this is potentiallt computationally
   * though, it demands an integer of how high complexity is accetable and may throw
   * a TaskTooComplexException if the task exceeds the accepted complexity.
   *
   * @param rcd     RelationType of the relation type we are interested in. Only explanations
   *                in which this relation-type is the last relation, are returned.
   * @param target  The relation to look for paths to. (All the paths originate at the origin entity of
   *                the Explanation.)
   */

  public Explanation[] getExplanations(RelationType rcd, Entity target, int acceptedComplexity) throws TaskTooComplexException
  {
    Object key[] = new Object[] { "explanation_model_get_explanations", this, rcd.getID(), target.getID() };
    if(target.getKnowledgeModel().getInferenceCache().isCached(key))
      return (Explanation[]) target.getKnowledgeModel().getInferenceCache().get(key);
    Explanation[] paths = getExplanations(target, acceptedComplexity);
    Vector approvedExplanations = new Vector();
//    System.out.println("getExplanations(rcd, target) - checking for approved relations over "+rcd.toString()+" to "+target.toString() );
    for(int i=paths.length-1;i>=0;i--)
    {
//      System.out.println(" approve? size is "+paths[i].size());
//      if(paths[i].size() > 0)
//       System.out.println("  rcd is "+paths[i].relationAt(paths[i].size()-1).getRelationType().toString());
      if(paths[i].size() > 0 && paths[i].relationAt(paths[i].size()-1).getRelationType().equals(rcd))
        approvedExplanations.addElement(paths[i]);
    }

    Explanation[] expArray = new Explanation[approvedExplanations.size()];
    approvedExplanations.copyInto(expArray);
    target.getKnowledgeModel().getInferenceCache().put(key, expArray);
    return expArray;
  }


  /**
   * Returns an integer assessing the maximal complexity of finding all paths
   * in a submodel. This is equal to no. of relations squared.
   */
  public int assessGetPathsComplexity()
  {
    return getRelations().length * getRelations().length;
  }

  private void recFindPath(Entity from, Entity to, Vector path, Vector entityPath, Vector explanations, String pre, double strength)
  {
    if(from.equals(to))
    {
//      System.out.println("Found explanation-path: ");
//      for(int i=0;i<path.size()-1;i++)
//        System.out.println("    --> "+((Relation) path.elementAt(i)).toString());
      Relation[] relArray = new Relation[path.size()];
      path.copyInto(relArray);
      explanations.addElement(new Explanation(relArray));

/*
      Relation[] rels = from.getLocalRelations(knowledgeModel.getRelationType(RELATION));
      for(int i=rels.length-1;i>=0;i--)
      {
        if(rels[i].getValue().equals(to))
        {
          path.addElement(rels[i]);
          relArray = new Relation[path.size()];
          path.copyInto(relArray);
          explanations.addElement(new Explanation(relArray));
          path.removeElement(rels[i]);
        }
       } */
    }
    else
    {
      Relation[] rels = from.getLocalRelations(knowledgeModel.getRelationType(BasicModel.RELATION));
      for(int i=rels.length-1;i>=0;i--)
      {
        if(!path.contains(rels[i]) &&
          !entityPath.contains(rels[i].getValue()) &&
          this.contains(rels[i]) &&
          Entity.class.isAssignableFrom(rels[i].getValue().getClass()))
        {
//          System.out.println(pre+"> "+from.toString()+" rel ["+rels[i].toString()+"]");
          double newStrength = strength * rels[i].getExplanationStrength();
          if(newStrength >= MINIMUM_STRENGTH_TRESHOLD)
          {
            path.addElement(rels[i]);
            entityPath.addElement(rels[i].getValue());
            recFindPath( (Entity) rels[i].getValue(), to, path, entityPath,
                        explanations, pre + "--", newStrength);
            entityPath.removeElement(rels[i].getValue());
            path.removeElement(rels[i]);
          }
        }
      }
    }
  }

  private Vector strongestPriFindPath(Entity from, Entity to, RelationType[] rcds)
  {
//    System.out.println("strongestPriFindPath "+from+" to "+to);
    ArrayList queue = new ArrayList(500);
    Vector explanations = new Vector(10);
    insertSorted(from.getLocalRelations(), new ExpPath(), queue);
    int iteration = 0;
    while(!queue.isEmpty() && ((iteration < 200 && !explanations.isEmpty()) || (iteration < 500)))
    {
      iteration++;
      ExpPath next = (ExpPath) queue.get(0);
      queue.remove(0);
///      System.out.println("\tAt "+next.getExplanation().toString());
      if(next.getLastEntity().equals(to))
      {
        Explanation nextExp = next.getExplanation();
        if(rule.accept(nextExp))
          explanations.add(nextExp);

        Relation[] rels = to.getLocalRelations(rcds);
        for(int i=rels.length-1;i>=0;i--)
        {
          if(rels[i].getValue().equals(to))
          {
            Explanation exp = new ExpPath(next, rels[i]).getExplanation();
            if(rule.accept(exp))
            {
              explanations.add( exp );
//              System.out.println("\t\tAccepted explanation "+exp);
            }
          }
         }


/*        Relation rels[] = to.getLocalRelations(rcds);
        for(int i=0;i<rels.length;i++)
        {
          if(contains(rels[i]))
          {
            Explanation exp = new ExpPath(next, rels[i]).getExplanation();
            if(rule.accept(exp))
            {
              explanations.add( exp );
//              System.out.println("\t\tAccepted explanation "+exp);
            }
          }
        } */
      }
      else
      {
        insertSorted(next.getLastEntity().getLocalRelations(), next, queue);
      }
    }
//    System.out.println("it: "+iteration+" for "+explanations.size());
    return explanations;
  }

  private void insertSorted(Relation[] rels, ExpPath path, ArrayList queue)
  {
    for(int i=rels.length-1;i>=0;i--)
    {
//      System.out.print("\t\tEvaluating adding..... "+rels[i]);
      if(this.contains(rels[i]) &&
        !path.contains(rels[i]) &&
        !path.contains(rels[i].getValue())
        )
//        Entity.class.isAssignableFrom(rels[i].getValue().getClass()))
      {
        ExpPath newExp = new ExpPath(path, rels[i]);
        int j=0;
        while(j<queue.size() && ((ExpPath) queue.get(j)).getStrength() > newExp.getStrength())
          j++;
        queue.add(j, newExp);
//        System.out.println("added at "+j);
      }
//      else
//        System.out.println("Not added.");
    }
  }

  class ExpPath
  {
    Relation[] explanation;
    HashMap containsTester = new HashMap();
    double strength = 1.0;
    public ExpPath()
    {
      explanation = new Relation[0];
    }

    public ExpPath(ExpPath existing, Relation r)
    {
      explanation = new Relation[existing.explanation.length+1];
      System.arraycopy(existing.explanation, 0, explanation, 0, existing.explanation.length);
      explanation[explanation.length-1] = r;
      strength = existing.getStrength()*r.getExplanationStrength();
    }

    public double getStrength()
    {
      return strength;
    }

    public boolean contains(Relation rel)
    {
      for(int i=0;i<explanation.length;i++)
        if(explanation[i].equals(rel))
          return true;
      return false;
    }

    public boolean contains(Entity ent)
    {
      for(int i=0;i<explanation.length;i++)
        if(explanation[i].getValue().equals(ent))
          return true;
      return false;
    }


    public Entity getLastEntity()
    {
      return explanation[explanation.length-1].getValue();
    }

    public Explanation getExplanation()
    {
      return new Explanation(explanation);
    }
  }

  /**
   * Returns a submodel containing all enties and relations of all paths from
   * the origin to the designated entity where the last step of all paths is
   * of the RelationType designated.
   */
  public SubModel getActivatedModel(RelationType rcd, Entity target)
  {
    Explanation[] explanations = new Explanation[0];
    try
    {
      explanations = getExplanations(rcd, target, 10000000);
    }
    catch(TaskTooComplexException e) { System.err.println("Too complex to find explanation for "+origin.getName()+" "+rcd.getName()+" "+target.getName()); }

    return Explanation.toSubModel(explanations);
  }
  /**
   * Returns the activation strength of the target in regards to the origin. This
   * calculation is done using only the entities and relations in the set.
   */
  public double getExplanationStrength(RelationType rcd, Entity target) throws TaskTooComplexException
  {
    Object key[] = new Object[] { "explanation_model_get_explanation_strength", this, rcd.getID(), target.getID() };
    if(target.getKnowledgeModel().getInferenceCache().isCached(key))
      return ((Double) target.getKnowledgeModel().getInferenceCache().get(key)).doubleValue();

    Explanation[] paths = new Explanation[0];

    paths = getExplanations(rcd, target, 10000000);
//    System.out.println(" ==== "+rcd.toString()+"   target: "+target+" getExplanationStrength(rcd,target): no paths: "+paths.length);

    double d = getExplanationStrength(paths, target);
    target.getKnowledgeModel().getInferenceCache().put(key, new Double(d));
    return d;
  }

  public double getExplanationStrength(Entity target) throws TaskTooComplexException
  {
    Explanation[] paths = new Explanation[0];

    paths = getExplanations(target, 10000000);

    return getExplanationStrength(paths, target);
  }

  // This method plain and simply do not work. As a temporary measure, we only
  // return the strongest explanation.
  public static double getExplanationStrength(Explanation[] paths, Entity target)
  {
    Explanation highExp = null;
    for(int i=0;i<paths.length;i++)
      if(highExp == null || highExp.getStrength() < paths[i].getStrength())
      {
        highExp = paths[i];
      }

    if(highExp != null)
      return highExp.getStrength();
    else
      return 0.0;
  }

    /*
    Hashtable strengths = new Hashtable();

    // Initializes the target strength to 0.0 .

    strengths.put("target", new Double(0.0));

    // Initializes all origins to 1.0 .

    for(int i=paths.length-1;i>=0;i--)
      strengths.put(paths[i].relationAt(0).getFrom(), new Double(1.0));

//  System.out.println(" In getExplanationStrength() for relation size of paths is "+paths.length+" target is "+target);
    for(int i=paths.length-1;i>=0;i--)
    {
      Explanation currentPath = paths[i];
//    System.out.println("Explanation "+currentPath.text()+" has size "+currentPath.size());
      for(int j=0;j<currentPath.size();j++)
      {
        Relation currentRelation = currentPath.relationAt(j);
        Object currentEntity;
        if(j+1 == currentPath.size())
          currentEntity = "target";
        else
          currentEntity = currentRelation.getValue();


//      System.out.println("Calculating for "+currentEntity.toString());

        double currentStrength;
        double currentK;
        if(!strengths.containsKey(currentEntity))
        {
//        System.out.println("Strength has does not contain entity.");
          currentStrength = 0.0;
          currentK = 1.0;
        }
        else
        {
          currentStrength = ((Double) strengths.get(currentEntity)).doubleValue();
          currentK = 1.0;
        }

        // Formula from Agnar's thesis, based on Mycin's formula fro propagating certainty factors along
        // a chain of rules.

        double lastStrength = ((Double) strengths.get(currentRelation.getFrom())).doubleValue();

//        System.out.println("Old value: "+currentStrength);
        currentStrength += currentK*lastStrength*currentRelation.getExplanationStrength()*(1.0-currentStrength);
//        System.out.println("New value: "+currentStrength+ " (lastStrength: "+lastStrength+")");

        strengths.put(currentEntity, new Double(currentStrength));
      }
    }
    return ((Double) strengths.get("target")).doubleValue();
  } */
}

