package jcreek.representation.inference;

import jcreek.representation.*;
import java.util.Hashtable;
import java.util.Vector;

public class PlausibleInheritanceMethod extends InheritanceMethod
{
  Hashtable importRule = null;
  Hashtable exportRule = null;
  Hashtable relationTransfers = new Hashtable();
  Vector needReenter = null;
  Vector inheritedRelations = null;

  RelationTypeSet  interestingRelations;
  /**
   * Creates a new PlausibleInheritanceMethod.
   * Only relations of the type specified by the parameter will be spread through
   * and accepted as inheritable.
   */
  public PlausibleInheritanceMethod(RelationType rcd)
  {
    interestingRelations = new RelationTypeSet(rcd);
    DEBUG = false;
  }

  /**
   * Creates a new PlausibleInheritanceMethod
   */
  public PlausibleInheritanceMethod(KnowledgeModel knowledgeModel)
  {
    DEBUG = false;
    interestingRelations = new RelationTypeSet(knowledgeModel.getRelationType(BasicModel.RELATION));
  }


  public boolean accept(Relation rel)
  {
    if(DEBUG)
      System.out.println("2: Checking relation '"+rel.toString()+"'");
    Entity from = rel.getFrom();
    Entity target = (Entity) rel.getValue();
    RelationTypeSet fromImportRule;
    if(importRule.containsKey(from))
    {
      fromImportRule = (RelationTypeSet) importRule.get(from);

      if(DEBUG)
        System.out.println("from-rule: "+fromImportRule.toString());

      // Finding the set of relationclasses this particular relationclass
      // transfers.
      RelationTypeSet transfers = findRelationTransfers(rel.getRelationType());
      if(DEBUG)
        System.out.println("transfers-rule: "+transfers.toString());

      RelationTypeSet targetImportRule = fromImportRule.copy();
      targetImportRule.intersect(transfers);
      if(DEBUG)
        System.out.println("intersected-rule: "+targetImportRule.toString());

      if(!targetImportRule.empty())
      {
        RelationTypeSet targetOriginalRule;
        if(importRule.containsKey(target))
        {
          // The target entity is already activated, and has a set of
          // relationclasses that can be inherited through it.
          targetOriginalRule = (RelationTypeSet) importRule.get(target);
          if(DEBUG)
            System.out.println("original-rule: "+targetOriginalRule.toString());

          // Checking if the current set of relationclasses is a subset
          // of the relationclasses the entity already transfers. If it isn't,
          // we need to spread through it again.
          if(!targetImportRule.isSubsetOf(targetOriginalRule))
          {
            if(DEBUG)
              System.out.println("Need to reenter "+target+"?");

            if(!needReenter.contains(target))
            {
              needReenter.addElement(target);
              if(DEBUG)
                System.out.println("Yep");
            }
            importRule.put(target, targetOriginalRule.union(targetImportRule));
          }
        }
        else
        {
          // It does not have a rule already, so we use the one
          importRule.put(target, targetImportRule);
        }
        if(fromImportRule.contains(rel.getRelationType()))
        {
          inheritedRelations.addElement(rel);
          if(DEBUG)
            System.out.println("    accepted (as transfer and inherited)");
        }
        else
        {
          if(DEBUG)
            System.out.println("    accepted (as transfer only)");
        }
        return true;
      }
      // -- //

      if(fromImportRule.contains(rel.getRelationType()))
      {
        if(DEBUG)
          System.out.println("   accepted (in own right only).");
        inheritedRelations.addElement(rel);
        return true;
      }
    }

    return false;
  }

  public boolean accept(Relation rel, Entity target)
  {
    if(DEBUG)
      System.out.println("Checking entity '"+target.toString()+"'");
    Entity from = rel.getFrom();

    // Finding the set of relationclasses that can be inherited to the 'from'
    // entity.
    RelationTypeSet fromImportRule;
    if(importRule.containsKey(from))
      return true;
    else
      if(inheritedRelations.contains(from))
        return true;
      else
        return false;
  }

  public boolean accept(Explanation path)
  {
    if(path.size() < 1)
      return false;

    RelationType lastRCD = path.relationAt(path.size()-1).getRelationType();
    for(int i=path.size()-2;i>=0;i--)
    {
      Relation rel = path.relationAt(i);
      RelationTypeSet transfers = findRelationTransfers(rel.getRelationType());
      if(!transfers.contains(lastRCD))
        return false;
    }
    return true;
  }

  public boolean acceptInheritance(Relation target)
  {
    if(interestingRelations.contains(target.getRelationType()) && inheritedRelations.contains(target))
      return true;
    else
      return false;
  }

  public boolean reenter(Entity target)
  {
    if(DEBUG)
      System.out.println("Checking if "+target+" should be re-entered");
    if(needReenter.contains(target))
    {
      needReenter.remove(target);
      return true;
    }
    return false;
  }

  public ExplanationModel execute(Entity origin)
  {
    if(DEBUG)
      System.out.println("Excute on "+origin);
    importRule = new Hashtable();
    exportRule = new Hashtable();
    needReenter = new Vector();
    inheritedRelations = new Vector();

    importRule.put(origin, interestingRelations);
    exportRule.put(origin, interestingRelations);

    return super.execute(origin);
  }

  public RelationTypeSet findRelationTransfers(RelationType rcd)
  {
    if(relationTransfers.containsKey(rcd))
     return (RelationTypeSet) relationTransfers.get(rcd);
    RelationType transferRelationDescriptor = rcd.getKnowledgeModel().getRelationType(BasicModel.TRANSFERS);
    Relation[] transferRelations = rcd.getRelations(transferRelationDescriptor, new SubclassInheritanceMethod());

    RelationTypeSet result = new RelationTypeSet();

    for(int i=transferRelations.length-1;i>=0;i--)
      if(!transferRelations[i].isNegated())
        try
        {
          result.union(new RelationTypeSet((RelationType) new RelationType(transferRelations[i].getValue())));
        }
        catch(EntityTypeNotMatchedException e) { }
    for(int i=transferRelations.length-1;i>=0;i--)
      if(transferRelations[i].isNegated())
        try
        {
          result.subtract(new RelationTypeSet((RelationType) new RelationType(transferRelations[i].getValue())));
        }
        catch(EntityTypeNotMatchedException e) { }

//    System.out.println("Relation transfer set for '"+rcd+"' is "+result);

    relationTransfers.put(rcd, result);
    return result;
  }

  public boolean equals(Object ob)
  {
    return ob == this || (ob.getClass().equals(getClass()) && ((PlausibleInheritanceMethod) ob).interestingRelations.equals(interestingRelations));
  }

  public int hashCode()
  {
    return 7+interestingRelations.hashCode();
  }


}


