package jcreek.representation.inference;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import java.util.Vector;
import java.util.Comparator;

/**
 * This class is an abstract class defining a rule for spreading activation through
 * a semantic network. For each spreading activation, one instance of this class is
 * created.
 */

public abstract class SpreadingActivationRule
{
  /**
   * Creates a new SpreadingActivationRule. This rule is limitied to spreading
   * to find relations of the type contained in the RelationTypeSet supplied.
   *
   * @param km    KnowledgeModel the rule will operate on. Included for future
   *              caching of inference by the KM.
   */
  public SpreadingActivationRule()
  {
  }

  /**
   * Returns true if we can accept the Entity target from the Relation from. Acceptance
   * means the entity will be added to the set of entites spread to.
   */
  public abstract boolean accept(Relation from, Entity target);

  /**
   * Returns true if we can accept the Relation target. Acceptance
   * means the entity will be added to the set of entites spread to.
   */
  public abstract boolean accept(Relation target);

  /**
   * Returns true if the rule can accept the Explanation path.
   */
  public abstract boolean accept(Explanation path);

  /**
   * This method is called on an entity that has already been spread through, to see if
   * we should spread through it again. It has the same function as enter(), but while
   * enter() is only called the FIRST time an entity is encountered during the spreading
   * activation, reenter() is called the following times.
   * WARNING: This method practically defines the complexity of the spreading operation.
   * If this method is always false, the complexity is linear with the number of concepts in
   * the model. If there is no cap on how many times an entity can be reconsidered, the
   * worst case complexity is infinite!
   * Complexity formula: O(c * n) where
   *  c = number of concepts in the model
   *  n = number of times a single entity can be reconsidered.
   */

  public boolean reenter(Entity target)
  {
    return false;
  }

  /**
   * Method for generating an instance of the model used in the execution of
   * the rule. This is a seperate method to allow subclasses to use subclasses
   * of ExplanationModel as their data object.
   */

  protected ExplanationModel generateModel(Entity origin)
  {
    return new ExplanationModel(origin, this);
  }


  /**
   * The basic spreading activation algorithm. This method uses accept(), and
   * renter(), and should not normally be rewritten.
   */

   public ExplanationModel execute(Entity origin)
   {
    KnowledgeModel knowledgeModel = origin.getKnowledgeModel();

    Vector queue = new Vector(0);
    ExplanationModel result = new ExplanationModel(origin, this);

    queue.addElement(origin);

    // Getting classdescriptor for the 'Relation' entity.
    RelationType relationRelationType = knowledgeModel.getRelationType(BasicModel.RELATION);

    int innerLoopCount = 0;

    while(!queue.isEmpty()) // && innerLoopCount < 120)
    {
      innerLoopCount++;
      // Popping the first entity of the queue, and ignoring it if it is also stored later in the queue.
      Entity currentEntity = null;
//      System.out.println("Queue size: "+queue.size());
      do
      {
        if(currentEntity != null)
        {

        }
        currentEntity = (Entity) queue.firstElement();
        queue.removeElementAt(0);
      }
      while(queue.contains(currentEntity));
//      System.out.println("Spreading to "+currentEntity);

      // Getting all local relations for the current entity.
      Relation[] localRelations = currentEntity.getLocalRelations(relationRelationType);

      // Circle through all relations in this entity.
      for(int i = localRelations.length-1;i>=0;i--)
      {
        if(accept(localRelations[i]))
        {
          result.add(localRelations[i]);

          // Finding target entity for the relation.
          Entity target;

          target = (Entity) localRelations[i].getValue();

          if(result.contains(target))
          {
            if(reenter(target))
            {
              queue.addElement(target);
            }
          }
          else
          {
            if(accept(localRelations[i], target))
            {
              result.add(target);
              queue.addElement(target);
            }
          }
        }
      }
    }
    return result;
   }
}



