package jcreek.representation.sw;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.io.*;
import java.net.URI;
import java.net.URISyntaxException;

// creek imports
import jcreek.representation.KnowledgeModel;
import jcreek.representation.BasicModel;
import jcreek.representation.cbr.Case;
import jcreek.representation.Relation;
import jcreek.representation.Entity;
import jcreek.representation.RelationType;
import jcreek.representation.EntityTypeNotMatchedException;
import jcreek.cke.importexport.KnowledgeModelExportUtilities;
import jcreek.representation.NoSuchRelationException;
import jcreek.representation.NumberEntity;
import jcreek.representation.DoubleEntity;

// jena imports
import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.rdf.model.*;

// predefined vocabularies
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.DC;
import com.hp.hpl.jena.vocabulary.RDFS;
import jcreek.representation.sw.CREEK;


/**
 * Does the parsing of the Creek Knowledge Model and extracts
 * the complete model to OWL.
 *
 * @author: Mikael Kirkeby Fidjeland, NTNU 2006
 *
 */
public class OWLexportParser {
    private KnowledgeModel km;
    private OntModel model;
    private String KM_NS;
    private String BASEURI;
    private KnowledgeModelExportUtilities kmeu;

    /**
     * Does set some basic settings before any parsing may begin.
     *
     * @param km the Knowledge Model in Creek to export
     * @param namespace the namespace for the ontology
     */
    public OWLexportParser(KnowledgeModel km, String namespace) {
        this.km = km;
        this.model = ModelFactory.createOntologyModel( OntModelSpec.OWL_MEM );
        this.BASEURI = namespace;
        this.KM_NS = this.BASEURI + "#";
        kmeu = new KnowledgeModelExportUtilities(km);
    } // OWLexportParser

    /**
     * Starts the creation of the OWL document.
     */
    public void createOntology() {
        model.removeNsPrefix("daml");
        model.setNsPrefix( "creek", CREEK.NS );
        model.setNsPrefix( "km", KM_NS );

        // ONTOLOGY META ELEMENT
        Ontology ont = model.createOntology(BASEURI );
        // title
        String title = km.getTitle();
        if (kmeu.hasLength(title))
            ont.addProperty( DC.title, title );
        // name
        String name = km.getName();
        ont.addProperty( RDFS.label, name );
        // creator
        String creator = km.getAuthor();
        if (kmeu.hasLength(creator))
            ont.addProperty( DC.creator, creator );
        // description element
        String description = km.getDescription();
        if (kmeu.hasLength(description))
            ont.addProperty(RDFS.comment, description);
        // import ISOPOD model element
        ont.addImport(CREEK.NAMESPACE);


        // CONCEPTS and RELATIONS
        Entity[] conceptList =  km.getEntities();
        for (int i = 0; i < conceptList.length; i++) {
            Entity fromEntity = ((Entity) conceptList[i]);
            if (fromEntity.getAllSuperclassesCollection().
                contains(km.getEntity(BasicModel.RELATION)))
                addProperty(fromEntity);
            else
                // drop the "NumberEntity*"'ies
                if (!fromEntity.getName().regionMatches(0, "NumberEntity", 0, 12))
                    addConcept(fromEntity);
        }
    }


    /**
     * Adds a Creek relation as an ObjectProperty to OntModel if
     * it is not added already <p>
     * Adds owl:inverseOf, rdfs:subPropertyOf, creek:hasDefaultExplanationStrength,
     * rdfs:label and rdfs:comment properties of the relation <p>
     * Recursive call to parent property and inverse property <p>
     *
     * @param kmRelation Creek Relation
     * @return The OntModel Property that were added, or the
     * corresponding property if it is allready added.
     */
    public Property addProperty(Entity kmRelation)
    {
        // make URI of relation and value names
        String label = kmRelation.getName();
        String propertyURI = nameToURI(label);
        Property swProperty = model.getObjectProperty(propertyURI);
        Property  swCreek = CREEK.m_model.getOntProperty(nameToURI(label, CREEK.NS));
        if (swCreek != null)
            // if it is a property defined in CREEK ISOPOD model
            return swCreek;
        else if (swProperty != null)
            // if it is defined already
            return swProperty;
        else
            // add it
            {
                String comment = kmRelation.getDescription();
                swProperty = model.createObjectProperty(propertyURI);
                // add label
                swProperty.addProperty( RDFS.label, label );
                // add comment
                if (kmeu.hasLength(comment))
                    swProperty.addProperty(RDFS.comment, comment);
                try
                    {
                        // add superclass
                        swProperty.addProperty( RDFS.subPropertyOf,
                                                addProperty(kmRelation.
                                                            getLocalRelation(BasicModel.SUBCLASS_OF).
                                                            getValue()));
                        // add inverse
                        swProperty.addProperty( OWL.inverseOf,
                                                addProperty(kmRelation.
                                                            getLocalRelation(BasicModel.HAS_INVERSE).
                                                            getValue()));

                        // add defaultexplanationstrength
                        swProperty.addProperty( CREEK.hasDefaultExplanationStrength,
                                                kmRelation.getLocalRelation(BasicModel.
                                                                            HAS_DEFAULT_EXPLANATION_STRENGTH).
                                                getValue().getEntityObject());
                        //  can check if the property transfers itself,
                        // then defined it as transitive in OWL
                        // ...
                    }
                catch (NoSuchRelationException e) {} // do nothing, e.g when no
                                                     // hasDefaultExplanationStrength exist
                catch (Exception e) { e.printStackTrace(); }
                return swProperty;
            }
    }


    /**
     * Adds a Creek concept as a Class to OntModel <p>
     * Adds all relations from this concept <p>
     * Recursive call to all relations from and concepts related to this concept <p>
     * @param kmConcept Creek Entity
     * @return The OntClass that were added to the OntModel
     */
    public OntResource addConcept(Entity kmConcept)
    {
        // collect entity data
        String label = kmConcept.getName();
        String conceptURI  = nameToURI(label);
        OntResource swCreek = CREEK.m_model.getOntResource(nameToURI(label, CREEK.NS));
        OntResource swConcept = model.getOntResource(conceptURI);
        if (swCreek != null)
            // if it is concept defined in CREEK ISOPOD model
            return swCreek;
        else if (swConcept != null)
            // if it is defined already
            return swConcept;
        else
            // add it
            {
                String comment = kmConcept.getDescription();
                swConcept =(OntResource) model.createClass(conceptURI);

                // add label
                swConcept.addProperty( RDFS.label, label );
                // add comment
                if (kmeu.hasLength(comment))
                    swConcept.addProperty(RDFS.comment, comment);

                // if Case, add casestatus
                if (Case.matches(kmConcept))
                    try {
                        switch (new Case(kmConcept).getStatus())
                            {
                            case Case.SOLVEDCASE:
                                swConcept.addProperty(CREEK.hasCaseStatus,
                                                      CREEK.SolvedCase);
                                break;
                            case Case.PROCESSCASE:
                                swConcept.addProperty(CREEK.hasCaseStatus,
                                                      CREEK.ProcessedCase);
                                break;
                            default:
                                swConcept.addProperty(CREEK.hasCaseStatus,
                                                      CREEK.UnsolvedCase);
                                break;
                            }
                    }
                    catch(Exception e)
                        {
                            e.printStackTrace();
                        }


                // add all other relations from this entity
                Relation[] relationList = kmConcept.getLocalRelations();
                for (int j = 0; j < relationList.length; j++)
                    {
                        Relation kmRelation = ((Relation) relationList[j]);
                        Entity kmObject = kmRelation.getValue();
                        if (!kmObject.getAllSuperclassesCollection().
                            contains(km.getEntity(BasicModel.RELATION)))
                            // only add if object of relation is not a relation itself
                            {
                                Property swProperty =
                                    addProperty((Entity) kmRelation.getRelationType());
                                // use RDF type for instantiation,
                                // and remove as instance of owl:Class
                                if (swProperty.equals(CREEK.instanceOf))
                                    {
                                        swConcept.addProperty(RDF.type,
                                                              addConcept(kmObject));
                                        swConcept.removeProperty(RDF.type, OWL.Class);
                                    }
                                // use RDFS subclassOf
                                else if (swProperty.equals(CREEK.subclassOf))
                                    swConcept.addProperty(RDFS.subClassOf,
                                                          addConcept(kmObject));
                                // drop hasInstance and hasSubclass
                                else if (!swProperty.equals(CREEK.hasSubclass) &&
                                         !swProperty.equals(CREEK.hasInstance))
                                    {
                                        String inverseSubjectURI =
                                            nameToURI(kmRelation.getValue().getName());
                                        String inversePropertyURI =
                                            nameToURI(kmRelation.getInverseName());
                                        OntResource swInverseSubject =
                                            model.getOntResource(inverseSubjectURI);
                                        Property swInverseProperty =
                                            model.getProperty(inversePropertyURI);
                                        // drop if inverse is already added
                                        if (swInverseSubject != null &&
                                            !swConcept.equals(swInverseSubject.
                                                              getPropertyValue(swInverseProperty)))
                                            swConcept.addProperty(swProperty,
                                                                  addConcept(kmObject));
                                        //handle [Double|Number]Entity
                                        else if (kmObject.getEntityObject() instanceof Double ||
                                                 kmObject.getEntityObject() instanceof Number ||
                                                 kmObject.getEntityObject() instanceof String)
                                            swConcept.addProperty(swProperty,
                                                                  kmObject.getEntityObject());
                                    }
                            }
                    }
                return swConcept;
            }
    }


    /**
     * Translates a name from Creek form to OWL/Java etc. form,
     * e.g. 'has finding' -> 'hasFinding'
     *
     * @param name Name in Creek form
     * @return name in OWL/Java etc form
     */
    public String changeName(String name)
    {
        char[] nameArray = name.toCharArray();

        if (nameArray.length == 0)
            return "";
        else
            {
                // substitute any ' ' or '(' and the following char
                // to only uppercase of the char
                int j = 0;
                int i;
                for (i = 0; i < nameArray.length - 1; i++)
                    if ((nameArray[i] == ' ' || nameArray[i] == '(' ))
                        {
                            name = name.substring(0,i-j) +
                                Character.toUpperCase(nameArray[i+1]) +
                                name.substring(i-j+2);
                            j++;
                        }
                // remove trailing ')'
                if (nameArray[i++] == ')')
                    name = name.substring(0, i-j-1) + name.substring(i-j);
            }
        // cannot have '#' in URIrefs
        return name.replace( '#', '_' );
    }

    /**
     * Translate Creek Entity name to URI. <p>
     * Removes spaces and capitalize character after space in names<p>
     * Adds namespace to name<p>
     * @param name Entity name in Creek knowledgemodel
     * @param namespace Namespace URI
     * @return URIref for Entity
     */
    public String nameToURI(String name, String namespace)
    {
        return namespace + changeName(name);
    }

    /**
     * Translate Creek Entity name to URI.<p>
     * Removes spaces and capitalize character after space in names<p>
     * Adds namespace to name, using default namespace<p>
     *
     * If name is already a valid URIref, just return it
     *
     * @param name Entity name in Creek knowledgemodel
     * @return URIref for Entity
     */
    public String nameToURI(String name)
    {
        try
            {
                // check if name is already a URIref
                URI uri = new URI(name);
                if (uri.getScheme() == null || uri.getScheme() == "")
                    return nameToURI(name, KM_NS);
                else
                    return name;
            }
        catch (URISyntaxException e)
            {
                return nameToURI(name, KM_NS);
            }
    }

    /**
     * Writes the Jena OntModel to disk
     * @param fileName Path to file
     */
    public void write(String fileName) throws FileNotFoundException, IOException
    {
        FileOutputStream out = new FileOutputStream(fileName);
        model.write(out, "RDF/XML-ABBREV", KM_NS.substring( 0,KM_NS.length() - 1 ));
        out.close();
    }

    /**
     * Writes the Jena OntModel to stdout
     */
    public void write()
    {
        model.write(System.out, "RDF/XML-ABBREV", KM_NS.substring( 0,KM_NS.length() - 1 ));
    }
}
