package jcreek.representation.sw;

import java.util.Iterator;
import java.lang.Double;
import java.io.*;
import java.net.UnknownHostException;

// creek imports
import jcreek.representation.KnowledgeModel;
import jcreek.representation.BasicModel;
import jcreek.representation.Relation;
import jcreek.representation.Entity;
import jcreek.representation.RelationType;
import jcreek.representation.EntityTypeNotMatchedException;
import jcreek.cke.importexport.KnowledgeModelImportUtilities;
import jcreek.representation.NoSuchRelationException;
import jcreek.representation.NumberEntity;
import jcreek.representation.DoubleEntity;
import jcreek.representation.LocalKnowledgeModel;
import jcreek.representation.cbr.CBRModel;

// jena imports
import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.rdf.model.*;

// predefined vocabularies
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.DC;
import com.hp.hpl.jena.vocabulary.RDFS;
import jcreek.representation.sw.CREEK;


/**
 * Does the parsing of an OWL ontology and imports it
 * into a Creek km.
 *
 * Creates a new empty ISOPOD model before importing
 *
 * @author: Mikael Kirkeby Fidjeland, NTNU 2006
 *
 */
public class OWLimportParser
{
    private static String DEFAULT_EXPL_STRENGTH = "0.5";
    private static String INVERSE_SUFFIX = "Inverse";

    private KnowledgeModel km;
    private OntModel model;
    private String importURI;
    private KnowledgeModelImportUtilities kmiu;
    private String NS;
    private String BASEURI;

    /**
     * Does the parsing of an ontology and imports it into
     * a Creek knowledge model
     *
     * @param importURI the URI of the ontology
     * @param km the knowledge model to import into
     */
    public OWLimportParser(String importURI, KnowledgeModel km)
    {
        this.importURI = importURI;
        this.km = km;

        OntDocumentManager mgr = new OntDocumentManager();
        // use OWL model, no reasoner
        OntModelSpec s = new OntModelSpec(OntModelSpec.OWL_MEM);
        mgr.addIgnoreImport(CREEK.getURI());
        s.setDocumentManager(mgr);
        this.model = ModelFactory.createOntologyModel(s);
        this.model.addSubModel(CREEK.m_model);
        kmiu = new KnowledgeModelImportUtilities(km);
    }


    /**
     * Starts the importing. Goes through all ObjectProperty's, DatatypeProperty's,
     * Class'es and Individual's in the model and adds them to the
     * knowledge model.
     */
    public void importOntology() throws IOException, ClassNotFoundException
    {
        BASEURI = importURI;
        NS = BASEURI + "#";

        // read input ontology
        model.read(importURI);

        // metadata
        Ontology ont = model.getOntology(BASEURI);
        if (ont != null)
            {
                km.setName(ont.getLabel(null));
                Statement title = ont.getProperty(DC.title);
                if (title != null)
                    km.setTitle(title.getObject().toString());
                else
                    km.setTitle(BASEURI);
                Statement author = ont.getProperty(DC.creator);
                if (author != null)
                    km.setAuthor(author.getObject().toString());
                km.setDescription(ont.getComment(null));
            }

        // add concepts
        for (Iterator i = model.listClasses(); i.hasNext();)
            addConcept((OntClass) i.next());
        for (Iterator i = model.listIndividuals(); i.hasNext();)
            addConcept((Individual) i.next());

    }

    /**
     * Adds the properties of the ontology as new relation types in Creek
     * Adds inverse, super property, default explanationstrength, transfers
     * and inherits over  relations
     *
     * @param prop the Ontology property to add
     * @return the name of the property added
     */
    public String addRelationType(OntProperty prop)
    {
        if (prop == null)
            // set default type to be HAS_FINDING
            return URItoName(CREEK.hasFinding.getURI());

        // name of relation
        String name = prop.getURI();
        if (name == null || name == "")
            // set default type to be HAS_FINDING
            return URItoName(CREEK.hasFinding.getURI());

        else if (CREEK.m_model.getOntProperty(name) != null)
            // resource is part of ISOPOD model
            return URItoName(name);

        else if (km.getRelationType(name) ==  null &&
                 model.getOntResource(prop) != null)
            //  it is not already added and is part of importing ontology
            {
                // strength of relation
                String strength;
                Statement tmpStmnt = prop.getProperty(CREEK.hasDefaultExplanationStrength);
                strength = tmpStmnt != null ? new Double(tmpStmnt.getDouble()).toString() :
                    DEFAULT_EXPL_STRENGTH;

                // comment
                String comment = prop.getComment(null);
                if (comment == null || comment == "")
                    comment = prop.getLabel(null);

                // inverse
                OntProperty invProp = getInverse(name, prop);
                String inv = invProp.getURI();

                try
                    {
                        // recursive call to add parent property
                        kmiu.addNewRelationType(addRelationType(prop.getSuperProperty()),
                                                name, inv, strength, comment);
                        addRelationProperties(name, prop, inv, invProp);
                    }
                catch(Exception e)
                    {
                        e.printStackTrace();
                    }
            }
        return name;
    }

    /**
     * Finds the inverse of a OntProperty, or generates
     * one if it does not exist
     *
     * @param name the name (or URIref) of the property
     * @param prop the property
     * @return the inverse of prop
     */
    private OntProperty getInverse(String name, OntProperty prop)
    {
        // inverse relation
        OntProperty invProp = prop.getInverse();
        if (invProp == null)
            {
                invProp = prop.getInverse();
                if (invProp == null)
                    {
                        invProp = prop.getInverseOf();
                        if (invProp == null)
                            {
                                invProp = model.createOntProperty(name + INVERSE_SUFFIX);
                            }
                    }
            }
        return invProp;
    }

    /**
     * Adds transfer and inheritsOver relations for a property
     * if they exist
     *
     * @param name Name (or URIref) for the property
     * @param prop Property
     * @param name Name (or URIref) for the inverse property
     * @param invProp The inverse of property
     */
    private void addRelationProperties(String name, OntProperty prop,
                                       String inv, OntProperty invProp) throws Exception
    {
        // transfers property
        Statement tmpRes = prop.getProperty(CREEK.transfers);
        Statement tmpRes2 = invProp.getProperty(CREEK.transfers);
        if (tmpRes != null)
            {
                String tmpURI = tmpRes.getResource().getURI();
                if (tmpURI != null)
                    {
                        OntProperty tmpProp = model.getOntProperty(tmpURI);
                        if (tmpProp == null)
                            tmpProp = model.createOntProperty(tmpURI);
                        kmiu.addRelation(name,
                                         addRelationType(CREEK.transfers),
                                         addRelationType(tmpProp),
                                         null);
                    }
            }
        else if (tmpRes2 != null)
            {
                String tmpURI = tmpRes2.getResource().getURI();
                if (tmpURI != null)
                    {
                        OntProperty tmpProp = model.getOntProperty(tmpURI);
                        if (tmpProp == null)
                            tmpProp = model.createOntProperty(tmpURI);
                        kmiu.addRelation(inv,
                                         addRelationType(CREEK.inheritsOver),
                                         addRelationType(tmpProp),
                                         null);
                    }
            }

        // inheritsover property
        tmpRes = prop.getProperty(CREEK.inheritsOver);
        tmpRes2 = invProp.getProperty(CREEK.inheritsOver);
        if (tmpRes != null)
            {
                String tmpURI  = tmpRes.getResource().getURI();
                if (tmpURI != null)
                    {
                        OntProperty tmpProp = model.getOntProperty(tmpURI);
                        if (tmpProp == null)
                            tmpProp = model.createOntProperty(tmpURI);
                        kmiu.addRelation(name,
                                         addRelationType(CREEK.inheritsOver),
                                         addRelationType(tmpProp),
                                         null);
                    }
            }
        else if (tmpRes2 != null)
            {
                String tmpURI  = tmpRes2.getResource().getURI();
                if (tmpURI != null)
                    {
                        OntProperty tmpProp = model.getOntProperty(tmpURI);
                        if (tmpProp == null)
                            tmpProp = model.createOntProperty(tmpURI);
                        kmiu.addRelation(inv,
                                         addRelationType(CREEK.inheritsOver),
                                         addRelationType(tmpProp),
                                         null);
                    }
            }
        // may need to loop through all properties for this
        // relation
    }





    /**
     * Adds a OntResource (Class or Individual) from the ontology as
     * an Enitiy in the knowledge model
     *
     *
     * @param concept the Class or Individual from the Ontology to add
     * @return the name of the entity added
     */
    public String addConcept(OntResource concept)
    {
        if (concept == null || concept.equals(OWL.Class))
            // set default class to Symbol
            return URItoName(CREEK.Symbol.getURI());

        // name of relation
        String name = concept.getURI();

        // anonymous node
        if (name == null || name == "")
            {
                name = concept.getId().toString();
                // need to add code here if we want to handle
                // anoymous nodes (owl:Restrictions, creek:Relation etc)
            }

        // if owl:ObjectPropety or owl:DatatypeProperty
        if (concept.isProperty())
            return URItoName(CREEK.Relation.getURI());

        if (CREEK.m_model.getOntResource(name) != null)
            // resource is  part of ISOPOD model
            return URItoName(name);
        else if (km.getEntity(name) == null &&
                 model.getOntResource(concept) != null)
            // it is not already added and is part of importing ontology
            {
                // description
                String description = "";
                String comment = concept.getComment(null);
                if (comment != null && comment != "")
                    description = comment;
                String label = concept.getLabel(null);
                if (label != null && label != "")
                    description = description + label;

                try
                    {
                        // if case
                        if (concept.hasRDFType(CREEK.Case))
                            kmiu.addCase(name, description);
                        else
                            kmiu.addEntity(name, comment, null, null);

                        //add all properties
                        for (StmtIterator i = concept.listProperties(); i.hasNext();)
                            {
                                Statement stmt = i.nextStatement();

                                String propURI = stmt.getPredicate().getURI();
                                OntProperty prop = model.getOntProperty(propURI);
                                // if property is not part of the importing ontology
                                if (prop == null)
                                    prop = model.createOntProperty(propURI);
                                // if comment or label,
                                if (prop.equals(RDFS.label) || prop.equals(RDFS.comment))
                                    continue;
                                // if rdfs:subClassOf change to creek:subClassOf
                                else if (prop.equals(RDFS.subClassOf))
                                    prop = CREEK.subclassOf;
                                // if rdf:type change to creek:instanceOf
                                else if (prop.equals(RDF.type))
                                    prop = CREEK.instanceOf;

                                String objURI;
                                try
                                    {
                                        objURI = stmt.getResource().getURI();
                                    }
                                catch (ResourceRequiredException ex)
                                    // handle Number Entities
                                    {
                                        handleLiteral(name, prop, stmt);
                                        continue;
                                    }
                                OntResource obj = model.getOntResource(objURI);
                                // if object is not part of the importing ontology
                                if (obj == null)
                                    obj = model.createOntResource(objURI);
                                else if (prop.equals(CREEK.hasCaseStatus) &&
                                         obj.equals(CREEK.SolvedCase))
                                    {
                                        kmiu.addHasSolution(name, "Solved", null, null, null);
                                        continue;
                                    }
                                try
                                    {
                                        kmiu.addRelation(name,
                                                         addRelationType(prop),
                                                         addConcept(obj),
                                                         null);
                                    }
                                catch (KnowledgeModelImportUtilities.EqualRelationAlreadyExist ex)
                                    {} // ignore this and continue loop
                            }
                    }
                catch(Exception e)
                    {
                        e.printStackTrace();
                    }
            }
        return name;
    }


    /**
     *  Adds a Relation where the object of a Statement is a literal (string, number..)
     *
     * @param name name (or URIref) of the property
     * @param ont the property
     * @param stmt the full Statement
     */
    private void handleLiteral (String name, OntProperty prop, Statement stmt) throws Exception
    {
        try
            {
                String value = Integer.toString(stmt.getLiteral().getInt());
                String entityName = "NumberEntity" + value;
                if (km.getEntity(entityName) == null)
                    kmiu.addEntity(entityName, "", "java.lang.Integer",
                                   value);
                kmiu.addRelation(name, addRelationType(prop), entityName, null);
            }
        catch (NumberFormatException notNumber)
            {} // ignore this
        // add code here to handel String or Doubles etc...
    }




    /** Translates a URI to the format used in jCreek
     *
     * @param URI Name of resource
     * @return Name in the form used internally in jCreek
     */
    public String URItoName(String URI)
    {
        char[] nameArray = URI.split("#")[1].toCharArray();
        String name = "";
        if (nameArray.length > 0)
            {
                // keep first character of string as is
                name = String.valueOf(nameArray[0]);
                for (int i = 1; i < nameArray.length; i++)
                    if (Character.isUpperCase(nameArray[i]))
                        name = name +  " " + Character.toLowerCase(nameArray[i]);
                    else
                        name = name + String.valueOf(nameArray[i]);
            }
        return name;
    }
}

