package jcreek.util;

/**
 * A standard, thread-safe Least Recent Used (LRU) cache that will store references
 * to object based on a (possibly array) key. The maximum size of the cache is
 * defined upon initialziation, and when the maximum number of elements have been
 * used, the least recently accessed item will be purged from the cache when a new
 * item is stored.
 *
 * @author       Frode Srmo
 * @version 1.2
 */

import java.util.HashMap;
import java.util.ArrayList;

public class LRUCache
{
  protected volatile HashMap index;
  protected volatile ArrayList queue;
  protected boolean suspendFlush = false;
  protected boolean flushedWhileSuspended = false;
  protected int maxSize;

  /**
   * Creates a new cache.
   *
   * @param maxSize The maximum number of elements in the cache.
   */
  public LRUCache(int maxSize)
  {
    this.maxSize = maxSize;
    index = new HashMap(maxSize);
    queue = new ArrayList(maxSize);
  }

  public int size()
  {
    return queue.size();
  }

  public synchronized void put(Object key, Object result)
  {
    if(queue.contains(key))
      queue.remove(key);
    else
      index.put(key, result);

    while(queue.size() >= maxSize)
    {
//      System.err.println("Queue size is "+queue.size()+" max is "+maxSize);
      index.remove(queue.get(queue.size()-1));
      queue.remove(queue.size()-1);
    }
    queue.add(0, key);
  }

  public synchronized void put(Object[] key, Object result)
  {
    put(new CacheKey(key), result);
  }

  public synchronized boolean isCached(Object key)
  {
    return index.containsKey(key);
  }

  public synchronized boolean isCached(Object[] key)
  {
    return isCached(new CacheKey(key));
  }

  public boolean contains(Object key)
  {
    return isCached(key);
  }

  public boolean contains(Object[] key)
  {
    return isCached(key);
  }

  public void remove(Object key)
  {
    if(queue.contains(key))
    {
      index.remove(key);
      queue.remove(key);
    }
  }

  public void remove(Object[] key)
  {
    remove(new CacheKey(key));
  }

  public synchronized Object get(Object key)
  {
    if(index.containsKey(key))
    {
      queue.remove(key);
      queue.add(0, key);
      return index.get(key);
    }
    return null;
  }

  public synchronized Object get(Object[] key)
  {
    return get(new CacheKey(key));
  }

  public synchronized void flush()
  {
//    if(queue.isEmpty())
//      return;
    if(!suspendFlush)
    {
      if(!queue.isEmpty())
      {
        int currentSize = queue.size();
        index = new HashMap(currentSize);
        queue = new ArrayList(currentSize);
      }
    }
    else
      flushedWhileSuspended = true;
  }

  public synchronized void suspendFlush(boolean b)
  {
    if(!b && flushedWhileSuspended)
    {
      suspendFlush = false;
      flush();
    }
    else
    {
      suspendFlush = b;
    }
    flushedWhileSuspended = false;
  }

  public String toString()
  {
    return "LRUCache with "+index.size();
  }

  class CacheKey
  {
    protected Object[] key;
    private int hashCode = -1;

    public CacheKey(Object[] key)
    {
      this.key = key;
    }

    public int hashCode()
    {
      if(hashCode == -1)
      {
        hashCode = 0;
        for(int i=key.length-1;i>=0;i--)
        {
          if(key[i] != null)
            hashCode += hashFunc(i)*key[i].hashCode();
        }
      }
      return hashCode;
    }

    int primes[] = new int[] {1, 3, 5, 7, 11, 13, 17, 19};
    private int hashFunc(int i)
    {
      return primes[i%primes.length];
    }

    public boolean equals(Object ob)
    {
      if(hashCode() == ob.hashCode() && ob.getClass() == getClass())
      {
        CacheKey ce = (CacheKey) ob;
        if(ce.key.length != key.length)
          return false;
        for(int i=key.length-1;i>=0;i--)
          if(!ce.key[i].equals(key[i]))
            return false;
        return true;
      }
      return false;
    }
  }
}


