package jcreek.util.graph;

/**
 * The default implementation of the Graph interface.
 *
 * In this implementation, it is illegal for the graph to contain edges from and
 * to nodes that do not exist in the graph.
 *
 * This implementation is thread-safe.
 *
 * @author Frode Sro
 * @version 1.0
 */

import java.util.*;

public class Graph implements java.io.Serializable
{
  protected ArrayList nodeList;
  protected ArrayList edgeList;
  protected transient ArrayList changeListeners;

  /**
   * Creates a new, empty graph.
   */
  public Graph()
  {
      nodeList  = new ArrayList();
      edgeList = new ArrayList();
  }

  /**
   * Adds the node to the graph.
   *
   * @param node The node to add.
   */
  public synchronized void addNode(Node node)
  {
    boolean added = false;
    synchronized(getClass())
    {
      if(!nodeList.contains(node))
      {
        node.setGraph(this);
        nodeList.add(node);
        added = true;
      }
    }
    if(added)
      triggerNodeAddedEvent(node);
  }

  /**
   * Adds an edge to the graph.
   *
   * @param edge The edge to add.
   */
  public synchronized void addEdge(Edge edge)
  {
    boolean added = false;
    synchronized(getClass())
    {
      if(!edgeList.contains(edge) && contains(edge.getFrom()) && contains(edge.getTo()))
      {
        edge.setGraph(this);
        edgeList.add(edge);
        edge.getTo().addEdgeP(edge);
        edge.getFrom().addEdgeP(edge);
        added = true;
      }
    }
    if(added)
      triggerEdgeAddedEvent(edge);
  }

  /**
   * Remove a node from the graph. This will also remove all edges to and from
   * the node through a call to removeEdge().
   *
   * @param node The node to remove.
   */
  public synchronized void removeNode(Node node)
  {
    System.out.println("Graph.removeNode");
    if(node != null && nodeList.contains(node))
    {
      node.removeP();
      nodeList.remove(node);
      triggerNodeRemovedEvent(node);
    }
  }

  /**
   * Remove an edge from the graph.
   *
   * @param edge The edge to remove.
   */
  public synchronized void removeEdge(Edge edge)
  {
    if(edge != null && edgeList.contains(edge))
    {
      edge.removeP();
      edgeList.remove(edge);
      triggerEdgeRemovedEvent(edge);
    }
  }


  /**
   * The number of edges in this graph.
   *
   * @return The number of edges.
   */
  public synchronized int edgeSize()
  {
    return edgeList.size();
  }

  /**
   * The number of nodes in this graph.
   *
   * @return The number of nodes.
   */
  public synchronized int nodeSize()
  {
    return nodeList.size();
  }

  /**
   * Gives an iterator for the nodes in this graph.
   *
   * (access through iterator may not be thread-safe?)
   */
  public synchronized Iterator getNodeIterator()
  {
      return nodeList.iterator();
  }

  /**
   * Returns a an array of the nodes of the graph.
   */
  public synchronized Node[] getNodes()
  {
    return (Node[]) nodeList.toArray(new Node[0]);
  }

  /**
   * Returns an array of the edges of the graph.
   */
  public synchronized Edge[] getEdges()
  {
    return (Edge[]) edgeList.toArray(new Edge[0]);
  }

  /**
   * Gives an iterator for the edges from a specific node in the graph.
   *
   * (access through iterator may not be thread-safe?)
   */
  public synchronized Iterator getEdgeIterator(Node node)
  {
      return node.getEdgeIterator();
  }

  /**
   * Gives an iterator for all the edges in the graph.
   *
   * (access through iterator may not be thread-safe?)
   */
  public synchronized Iterator getEdgeIterator()
  {
      return edgeList.iterator();
  }

  /**
   * Returns true if the node supplied is present in this graph.
   */
  public synchronized boolean contains(Node node)
  {
    return nodeList.contains(node);
  }

  /**
   * Returns true if the edge suppleid is present in this graph.
   */
  public synchronized boolean contains(Edge edge)
  {
    return edgeList.contains(edge);
  }

  /**
   * Adds a change listener to this model. Any time a entity
   * or edge is added, removed or changed in any way via the graph, or the
   * set of selected entities are changed, this listener is notified.
   *
   * @param  listener  new listener to add notification list
   * @see #removeChangeListener
   */
  public synchronized void addGraphChangeListener(GraphChangeListener listener )
  {
    if(changeListeners == null)
      changeListeners = new ArrayList();
    if(!changeListeners.contains(listener))
      changeListeners.add(listener);
  } // addChangeListener

  /**
   * Remove a change listener from the listener list.
   *
   * @param  listener  listener to remove from notification list
   * @see #addChangeListener
   */
  public synchronized void removeGraphChangeListener( GraphChangeListener listener )
  {
    if(changeListeners != null)
      changeListeners.remove( listener );
  } // removeChangeListener

  protected void triggerNodeRemovedEvent(Node n)
  {
    GraphChangeEvent e = new GraphChangeEvent(n, GraphChangeEvent.NODE_REMOVED);
    if(changeListeners != null)
      for(Iterator it = changeListeners.iterator();it.hasNext();)
      {
        ((GraphChangeListener) it.next()).nodeRemoved(e);
      }
  }

  protected void triggerNodeAddedEvent(Node n)
  {
    GraphChangeEvent e = new GraphChangeEvent(n, GraphChangeEvent.NODE_ADDED);
    if(changeListeners != null)
      for (Iterator it = changeListeners.iterator(); it.hasNext(); )
      {
        ( (GraphChangeListener) it.next()).nodeAdded(e);
      }
  }

  protected void triggerEdgeRemovedEvent(Edge edge)
  {
    GraphChangeEvent e = new GraphChangeEvent(edge, GraphChangeEvent.EDGE_REMOVED);
    if(changeListeners != null)
      for (Iterator it = changeListeners.iterator(); it.hasNext(); )
      {
        ( (GraphChangeListener) it.next()).edgeRemoved(e);
      }
  }

  protected void triggerEdgeAddedEvent(Edge edge)
  {
    GraphChangeEvent e = new GraphChangeEvent(edge, GraphChangeEvent.EDGE_ADDED);
    if(changeListeners != null)
      for (Iterator it = changeListeners.iterator(); it.hasNext(); )
      {
        ( (GraphChangeListener) it.next()).edgeAdded(e);
      }
  }

  protected void triggerEdgeChangeEvent(GraphChangeEvent e)
  {
    if(changeListeners != null)
      for (Iterator it = changeListeners.iterator(); it.hasNext(); )
      {
        ( (GraphChangeListener) it.next()).edgeChanged(e);
      }
  }

  protected void triggerNodeChangeEvent(GraphChangeEvent e)
  {
    if(changeListeners != null)
      for (Iterator it = changeListeners.iterator(); it.hasNext(); )
      {
        ( (GraphChangeListener) it.next()).nodeChanged(e);
      }
  }

  void triggerNodeMovedEvent(GraphChangeEvent e)
  {
    if(changeListeners != null)
      for (Iterator it = changeListeners.iterator(); it.hasNext(); )
      {
        ( (GraphChangeListener) it.next()).nodeMoved(e);
      }
  }


  /**
   * Returns an exact copy of this graph. This will also copy the node and
   * edge objects in the graph maintaing the structure while making the copies
   * independent of eachother.
   */
  public Object clone()
  {
    Iterator nodeIterator, edgeIterator;
    Graph newGraph = new Graph();
    nodeIterator = getNodeIterator();
    Hashtable nodeMap = new Hashtable();
    while(nodeIterator.hasNext())
    {
        Node node = (Node)nodeIterator.next();
        Node newNode = (Node) node.clone();
        nodeMap.put(node, newNode);
        newGraph.addNode(newNode);
    }
    edgeIterator = newGraph.getEdgeIterator();
    while(edgeIterator.hasNext())
    {
       Edge edge = (Edge)edgeIterator.next();
       Edge newEdge = (Edge) edge.clone();
       newEdge.from = (Node) nodeMap.get(edge.getFrom());
       newEdge.to = (Node) nodeMap.get(edge.getTo());
       newGraph.addEdge(newEdge);
    }
    return newGraph;
  }

  /**
   * Creates a string representation of the graph.
   */
  public String toString()
  {
    String ret = "Graph - nodelist:\n";
    Iterator it = getNodeIterator();
    while(it.hasNext())
      ret +="\t"+it.next()+"\n";
    ret +="\nedgelist:\n";
    it = getEdgeIterator();
    while(it.hasNext())
      ret += "\t"+it.next();
    return ret;
  }

}
