package jcreek.util.graph;

import java.awt.Point;
import java.lang.Cloneable;
import java.io.Serializable;
import java.util.Iterator;
import java.util.Vector;

/**
 * This class represents a node with a 2D integer position. The node may also
 * be a member of a graph.
 *
 * @author Frode Srmo
 * @version 1.0
 *
 * @see Graph
 */

public abstract class Node implements Serializable, Cloneable
{
  private Vector edgeList = new Vector();
  protected Graph graph;
  private int x = 0;
  private int y = 0;
  private transient Point oldPos;
  private transient boolean eventReportSuspended = false;

  /**
   * Creates a new node. It will by default not be a member of any graph.
   */
  public Node()
  {
  }

  /**
   * Creates a new node with the given initial position.
   */
  public Node(int x, int y)
  {
    this.x = x;
    this.y = y;
  }

  /**
   * Returns the position of this node.
   */
  public Point getPosition()
  {
     Point position = new Point(x, y);
     return position;
  }

  /**
   * Returns the position of this node.
   */
  public void setPosition(int newX, int newY)
  {
     x = newX;
     y = newY;
     if(graph != null && !eventReportSuspended)
       graph.triggerNodeMovedEvent(new GraphChangeEvent(this, GraphChangeEvent.NODE_MOVED));
  }

  /**
   * Sets the position of this node.
   */
  public void setPosition(Point p)
  {
     setPosition(p.x, p.y);
  }

  /**
   * Moves the Node by dx, dy.
   */
  public void translate(int dx, int dy)
  {
    setPosition(x+dx, y+dy);
  }

  void setGraph(Graph g)
  {
    if(graph == null)
      this.graph = g;
  }

  /**
   * Returns the Graph this Node is a part of. Null is returned if it is not
   * part of any graphs.
   */
  public Graph getGraph()
  {
    return graph;
  }

  /**
   * Adds an edge to this node. This will fail if the node is not attached
   * to a graph.
   */
  public void addEdge(Edge edge)
  {
    if(graph != null)
      graph.addEdge(edge);
  }

  /**
   * Called from Graph to tell this node to add an edge to its list of edges from
   * or to this node. Note that this method should always be a result of calling
   * the addEdge() method, but edges may also be added from the Graph object
   * directly.
   */
  protected void addEdgeP(Edge e)
  {
    edgeList.add(e);
  }

  /**
   * Removes an edge from the node.
   */
  public void removeEdge(Edge edge)
  {
    if(edgeList.contains(edge))
      graph.removeEdge(edge);
  }

  /**
   * Called from Graph to tell this node to remove an edge to its list of edges from
   * or to this node. Note that this method should always be a result of calling
   * the removeEdge() method, but edges may also be added from the Graph object
   * directly.
   */
  protected void removeEdgeP(Edge e)
  {
    edgeList.remove(e);
  }

  /**
   * Tells this node to remove itself from its current graph.
   */
  public void remove()
  {
    if(graph != null)
      graph.removeNode(this);
  }

  /**
   * Called by the Graph object to notify the node that it has been removed
   * from the graph.
   */
  public void removeP()
  {
    while(!edgeList.isEmpty())
      ((Edge) edgeList.get(0)).remove();

    graph = null;
    edgeList = new Vector();
  }

  /**
   * Returns an iterator for the list of edges from/to this node.
   */
  public Iterator getEdgeIterator()
  {
    return edgeList.iterator();
  }

  /**
   * Gives the number of edges from/to this node.
   */
  public int edgeSize()
  {
    return edgeList.size();
  }

  /**
   * Suspends the reporting of "move" events for this node. When turned
   * off, any movement made in the meantime is reported in an event.
   */
  public void suspendEventReport(boolean suspend)
  {
    if(suspend != eventReportSuspended)
    {
      eventReportSuspended = suspend;
      if(suspend)
        oldPos = getPosition();
      else
      {
        if(oldPos != null && !oldPos.equals(getPosition()))
        {
          graph.triggerNodeMovedEvent(new GraphChangeEvent(this, GraphChangeEvent.NODE_MOVED));
        }
      }
    }
  }

  /**
   * Returns a clone of this node.
   */
  public Object clone()
  {
    try
    {
      return super.clone();
    }
    catch(CloneNotSupportedException e)
    {
      // Should never happen since this class implements clonable.
      return null;
    }
  }
}