package jcreek.util.graph.layout;

/**
 * <p>Title: SpringEmbedderLayout</p>
 * <p>Description: Implements the Spring Embedder algorithm
 * (T. Fruchterman and E. Reingold. Graph drawing by force-directed placement.
 * Softw. - Pract. Exp., 21(11):1129-1164, 1991.) to lay out a graph. </p>
 * @author Frode Srmo
 * @version 1.0
 */

import jcreek.util.graph.*;
//import jcreek.gui.conceptmap.*;
//import jcreek.gui.conceptmap.tool.*;
import java.util.*;

import javax.swing.*;
import jcreek.representation.*; import jcreek.representation.cbr.*;
import java.awt.*;

public class SpringEmbedderLayout implements Layout
{
  protected Rectangle bounds;
  protected Graph graph;
  protected double optimalDistanceConstant = 1;
  protected double initialTemp = 5;
  protected int iterationPause = 0;

  public SpringEmbedderLayout()
  {
  }

  public void setIterationPause(int delay)
  {
    iterationPause = delay;
  }

  public void setInitialTemperature(double temp)
  {
    initialTemp = temp;
  }

  public void layout(Graph graph, Rectangle bounds)
  {
    layout(graph, graph.getNodes(), bounds, 100);
  }

  public void layout(Graph graph, Rectangle bounds, int iterations)
  {
    layout(graph, graph.getNodes(), bounds, iterations);
  }

  public void layout(Graph graph, Node[] nodes, Rectangle bounds)
  {
    layout(graph, nodes, bounds, 100);
  }

  public void randomizePositions(Node[] nodes, Rectangle bounds)
  {
    for(int i=0;i<nodes.length;i++)
      nodes[i].setPosition(randomPoint(bounds));
  }

  public synchronized void setBounds(Rectangle bounds)
  {
    this.bounds = bounds;
  }

  public synchronized Rectangle getBounds()
  {
    return bounds;
  }

  private static Point randomPoint(Rectangle bounds)
  {
    return new Point((int) Math.round(bounds.x+Math.random()*bounds.width), (int) Math.round(bounds.y+Math.random()*bounds.height));
  }

  public void layout(Graph graph, Node[] nodes, Rectangle bounds, int iterations)
  {
    if(nodes.length == 0)
      return;
    long boundsArea = bounds.width*bounds.height;
    double optimalDistance = optimalDistanceConstant * Math.sqrt(boundsArea/graph.nodeSize());

    Hashtable xDispTable = new Hashtable(graph.nodeSize());
    Hashtable yDispTable = new Hashtable(graph.nodeSize());

    double temperature = initialTemp;

    for(int i=0;i<iterations;i++)
    {
      // Calculate the repulsive forces.
      for(Iterator it=graph.getNodeIterator(); it.hasNext();)
      {
        Node node = (Node) it.next();
        double xDisp= 0;
        double yDisp = 0;
        for(Iterator inner=graph.getNodeIterator(); inner.hasNext();)
        {
          Node compareNode = (Node) inner.next();
          if(compareNode != node)
          {
            double xDif = compareNode.getPosition().x-node.getPosition().x;
            double yDif = compareNode.getPosition().y-node.getPosition().y;
            double totalDif = Math.sqrt(xDif*xDif+yDif*yDif);

            xDisp += xDif / totalDif * repulsiveForce(optimalDistance, totalDif);
            yDisp += yDif / totalDif * repulsiveForce(optimalDistance, totalDif);
          }
        }
//        System.out.println("Initial disp for "+((VisualEntity) node).creekLEntity+" is "+xDisp+", "+yDisp);
        xDispTable.put(node, new Double(xDisp));
        yDispTable.put(node, new Double(yDisp));
      }

      // Calculate attractive forces
      for(Iterator it=graph.getEdgeIterator(); it.hasNext();)
      {
        Edge edge = (Edge) it.next();
        Node head = edge.getFrom();
        Node tail = edge.getTo();
        if(!head.equals(tail))
        {
          double xDif = head.getPosition().x-tail.getPosition().x;
          double yDif = head.getPosition().y-tail.getPosition().y;
          double totalDif = Math.sqrt(xDif*xDif+yDif*yDif);

          double headXDisp = ((Double) xDispTable.get(head)).doubleValue();
          double headYDisp = ((Double) yDispTable.get(head)).doubleValue();

//          System.out.println("Disp for "+((VisualEntity) head).getEntity()+" [Relation "+((VisualRelation) edge).getRelation()+"]: "+headXDisp+", "+headYDisp);

          headXDisp -= xDif / totalDif * attractiveForce(optimalDistance, totalDif);
          headYDisp -= yDif / totalDif * attractiveForce(optimalDistance, totalDif);

          xDispTable.put(head, new Double(headXDisp));
          yDispTable.put(head, new Double(headYDisp));

          double tailXDisp = ((Double) xDispTable.get(tail)).doubleValue();
          double tailYDisp = ((Double) yDispTable.get(tail)).doubleValue();

          tailXDisp += xDif / totalDif * attractiveForce(optimalDistance, totalDif);
          tailYDisp += yDif / totalDif * attractiveForce(optimalDistance, totalDif);

          xDispTable.put(tail, new Double(tailXDisp));
          yDispTable.put(tail, new Double(tailYDisp));
        }
      }

      // Limit the maximum movement to temperature and prevent from beeing
      // places outside boundary.

      for(int j=0;j<nodes.length;j++)
      {
          Node node = nodes[j];
          double xPos = node.getPosition().x;
          double yPos = node.getPosition().y;
          double xDisp = ((Double) xDispTable.get(node)).doubleValue();
          double yDisp = ((Double) yDispTable.get(node)).doubleValue();
//          System.out.println(((VisualEntity) node).getEntity()+" pos "+xPos+","+yPos+" disp "+xDisp+","+yDisp);
          double dispNorm = Math.sqrt(xDisp*xDisp+yDisp*yDisp);

          xPos += xDisp / dispNorm * Math.min(dispNorm, temperature);
          yPos += yDisp / dispNorm * Math.min(dispNorm, temperature);

          xPos = Math.min(bounds.x+bounds.width, Math.max(bounds.x, xPos));
          yPos = Math.min(bounds.y+bounds.height, Math.max(bounds.y, yPos));

          node.setPosition((int) xPos, (int) yPos);
      }

      // Gradually decrease temperature
      temperature = initialTemp * 1.0 - (iterations-i)/iterations;
//      view.setGraph(graph);
      if(iterationPause > 0)
      {
        try
        {
          Thread.sleep(iterationPause);
        }
        catch(Exception e) { e.printStackTrace();}
      }

    }
  }

  public double attractiveForce(double optimalDistance, double actualDistance)
  {
    return actualDistance*actualDistance/optimalDistance;
  }

  public double repulsiveForce(double optimalDistance, double actualDistance)
  {
    return -(optimalDistance*optimalDistance)/(actualDistance > 0?actualDistance:1);
  }

/*
  public static void main(String argv[])
  {
    JFrame frame;

    Dimension area = new Dimension(600,600);

    KnowledgeModel km = new LocalKnowledgeModel();
    ConceptMapModel model = new ConceptMapModel(km);
    activate(model, area, km.getEntity("Relation"), 2);
    ConceptMapView cm;

//    frame = new JFrame("Hei");
//    cm = new ConceptMapView(model);
//    ds.attach(graph);
//    ConceptMap cm = new ConceptMap(graph);
//    cm.setScale(0.2);
//    ds.setGraph(graph.getGraph());
//    dt.attach();

//    frame.getContentPane().add(new JScrollPane(cm));
//    frame.setSize(200,200);
//    cm.setOffsets(0.0, 0.0, true);
//    frame.show();

    cm = new ConceptMapView(model);
//    ds.attach(graph);
//    ConceptMap cm = new ConceptMap(graph);
    cm.setScale(1.0);
//    ds.setGraph(graph.getGraph());
     DragTool dt = new DragTool();
     dt.attach(cm);
//    dt.attach();
     frame = new JFrame("Main");
    frame.getContentPane().add(new JScrollPane(cm));
    frame.setSize(800,800);
//    cm.setOffsets(0.0, 0.0, true);
    frame.show();
//    try


//    {
//      Thread.sleep(2000);
//    }
//   catch(Exception e) { }


    SpringEmbedderLayout se = new SpringEmbedderLayout();
    cm.suspendBoundingBoxUpdate(true);
    se.setIterationPause(3);
    se.setInitialTemperature(5);
    se.layout(model, cm.getBoundingBox(),1000);
    cm.suspendBoundingBoxUpdate(false);
  }

  private static VisualEntity activate(ConceptMapModel graph, Dimension area, Entity ent, int level)
  {
    VisualEntity fromVE;
    fromVE = graph.getVisualEntity(ent);
    if(fromVE == null)
    {
//      fromVE = new VisualEntity(ent, randomPoint(area));
      fromVE = graph.addEntity(ent, randomPoint(area));
    }

    if(level < 1)
      return fromVE;

    Relation[] rels = ent.getLocalRelations("Relation");
    for(int i=0;i<rels.length;i++)
    {
      VisualEntity valueVE = activate(graph, area, rels[i].getValue(), level-1);
      VisualRelation vr = graph.getVisualRelation(rels[i]);
//      System.err.println("Activating relation from "+fromVE.getEntity()+" to "+valueVE.getEntity());
      if(vr == null)
      {
        vr = new VisualRelation(fromVE, valueVE, rels[i]);
        graph.addEdge(vr);
      }
    }

    return fromVE;
  }
*/
  private static Point randomPoint(Dimension area)
  {
    return new Point((int) Math.round(Math.random()*area.width), (int) Math.round(Math.random()*area.height));
  }
}
