package jcreek.util.selection;

/**
 * <p>Title: SelectionSet</p>
 * <p>Description: A kind of set (as defined in the Java Collections API),
 * that in addition to implementing the Set interface provides support for
 * listeners to be notified about any changes to the set.
 * selection set.
 *
 * The implementation is based on the java.util.HashSet class, but adds code
 * to catch and fire events.
 *
 * Like the HashSet class, this class is not thread-safe by itself.
 *
 * @author Frode Srmo
 * @version 1.0
 */
import java.util.*;

public class SelectionSet extends HashSet
{
  private ArrayList selectionListeners = new ArrayList();

  /**
   * Adds an object to the set, returns true if the object did not already
   * exist in the set, false if it did.
   */
  public boolean add(Object ob)
  {
    boolean result = super.add(ob);

    if(result)
    {
      SelectionChangeEvent event = new SelectionChangeEvent(SelectionChangeEvent.SELECTION_CHANGE_ADD, ob, this);
      for(Iterator listen = selectionListeners.iterator();listen.hasNext();)
      {
        ((SelectionChangeListener) listen.next()).selectionAdd(event);
      }
    }
    return result;
  }

  /**
   * Removes an object from the set. Returns true if the object existed in the
   * set, false if not.
   */
  public boolean remove(Object ob)
  {
    boolean result = super.remove(ob);

    if(result)
    {
      SelectionChangeEvent event = new SelectionChangeEvent(SelectionChangeEvent.SELECTION_CHANGE_REMOVE, ob, this);
      for(Iterator listen = selectionListeners.iterator();listen.hasNext();)
      {
        ((SelectionChangeListener) listen.next()).selectionRemove(event);
      }
    }
    return result;
  }

  /**
   * Clears the contents of the set and sets it to empty.
   */
  public void clear()
  {
    super.clear();

    SelectionChangeEvent event = new SelectionChangeEvent(SelectionChangeEvent.SELECTION_CHANGE_CLEAR, null, this);
    for(Iterator listen = selectionListeners.iterator();listen.hasNext();)
    {
      ((SelectionChangeListener) listen.next()).selectionClear(event);
    }

  }

  public void addSelectionChangeListener(SelectionChangeListener listener)
  {
    selectionListeners.add(listener);
  }

  public void removeSelectionChangeListener(SelectionChangeListener listener)
  {
    selectionListeners.remove(listener);
  }
}

