package weka.classifiers.lazy;

import jcreek.representation.*; import jcreek.representation.cbr.*;
import weka.core.*;
import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.bayes.NaiveBayes;

import java.io.*;
import java.util.Hashtable;

/**
 * <p>Title: WekaToCreekConverter</p>
 * <p>Description: Converts a Weka model to a Creek Knowledgemodel.</p>
 * @author FrodeSrmo
 * @version 1.0
 */

public class WekaToCreekConverter
{
  private Hashtable valueMap = new Hashtable();
  private Instances instances;
  private KnowledgeModel km;

  public WekaToCreekConverter(Instances instances) throws UnsupportedClassTypeException
  {
    this.instances=instances;
    this.km = convert();
  }

  /**
   * Returns the KnowledgeModel generated from the Weka model.
   */
  public KnowledgeModel getKnowledgeModel()
  {
    return km;
  }

  /**
   * Returns the corresponding Creek entity for a Weka attribute.
   */
  public Entity getAttribute(Attribute attr)
  {
    return km.getEntity(attr.name());
  }

  /**
   * Returns the corresponding Creek entity for a Weka attribute value
   */
  public Entity getAttributeValue(Attribute attr, int value)
  {
    String key = attr.name()+"="+attr.value(value);
    if(valueMap.containsKey(key))
      return (Entity) valueMap.get(key);
    return null;
  }

  private KnowledgeModel convert() throws UnsupportedClassTypeException
  {

    if(instances.classIndex() < 0)
    {
      Attribute sol = instances.attribute("Class");
      if(sol != null)
        instances.setClass(sol);
      else
        instances.setClassIndex(instances.numAttributes()-1);
    }

    if (!instances.classAttribute().isNominal()) {
        throw new UnsupportedClassTypeException("Class attribute has to be nominal for Creek!");
    }

    km = new LocalKnowledgeModel();
    valueMap = new Hashtable();
    km.getInferenceCache().suspendFlush(true);

    try
    {
      // Creating attributes
      for(int i=0;i<instances.numAttributes();i++)
      {
        Attribute attribute = instances.attribute(i);
        Entity attributeEntity = new Entity(km, attribute.name(), attribute.name());
        attributeEntity.setEntityObject(new Integer(i));
        if(attribute.isNominal())
        {
//        System.err.println("Creating "+attributeEntity.getName());
          attributeEntity.addRelation("instance of", km.getEntity("Symbol"));
          for(int j=0;j<attribute.numValues();j++)
          {
//            String valueName = attribute.value(j);
//            if(km.hasConcept(valueName))
            String valueName = attribute.name()+"="+attribute.value(j);
            Entity valueEntity = new Entity(km, valueName, "");
            valueMap.put(attribute.name()+"="+attribute.value(j), valueEntity);
//            valueEntity.setEntityData(new Integer(j));
            valueEntity.addRelation("instance of", attributeEntity);
//          System.err.println("  -> "+valueEntity.getName());
          }
        }
        else if(attribute.isNumeric())
          attributeEntity.addRelation("instance of", km.getEntity("Number"));

        if(instances.classIndex() != i)
        {
//          System.err.println("Creating relation 'has "+attribute.name()+"'");
          new RelationType(km,
                                  (RelationType) km.getRelationType("has finding"),
                                  "has "+attribute.name(),
                                  "relation of "+attributeEntity.getName(),
                                  attributeEntity.getName()+" of",
                                  null,
                                  0.5).addRelation("has instance", attributeEntity);
        }
      }

      for(int i=0;i<instances.numInstances();i++)
      {
        Instance wekaInstance = instances.instance(i);
        Case creekCase = new Case(km, "Case#"+i, "Case#");
//        creekCase.setEntityData(new Integer(i));

        buildCase(wekaInstance, creekCase);

        creekCase.setStatus(Case.SOLVEDCASE);
//        creekCase.calculateDefaultRelevance();
      }

      buildModel();

    }
    catch(NameAlreadyExistException e)
    {
      e.printStackTrace();
    }
//    catch(NoSuchRelationException e)
//    {
//      e.printStackTrace();
//    }
    catch(NoSuchRelationTypeException e)
    {
      e.printStackTrace();
    }
/*    catch(java.io.IOException e)
    {
      e.printStackTrace();
    }
    catch(ClassNotFoundException e)
    {
      e.printStackTrace();
    } */

    km.getInferenceCache().suspendFlush(false);

    return km;
  }

  private void buildCase(Instance wekaInstance, Case creekCase) throws NoSuchRelationTypeException
  {
//    System.err.println("Creating case "+creekCase.getName());
    creekCase.addRelation("instance of", creekCase.getKnowledgeModel().getEntity("Case"));
    for(int j=0;j<wekaInstance.numAttributes();j++)
    {
      Attribute attr = wekaInstance.attribute(j);

      String relationType;
      if(wekaInstance.classAttribute().equals(attr))
        relationType = "has solution";
      else
        relationType = "has "+attr.name();

//      System.err.println("Attribute: "+attr.name()+" has value "+attr.value((int) wekaInstance.value(j))+" relationtype is '"+relationType+"' attribute value is "+getAttributeValue(attr, (int) wekaInstance.value(j)));

      if(!wekaInstance.isMissing(j))
      {
        if(attr.isNominal())
        {
          creekCase.addRelation(relationType, getAttributeValue(attr, (int) wekaInstance.value(j)));
        }
        else if(attr.isNumeric())
        {
          Entity numeric = new jcreek.representation.NumberEntity(km, new Double(wekaInstance.value(j)), getAttribute(wekaInstance.attribute(j)));
          creekCase.addRelation(relationType, numeric);
        }
      }
    }
  }

  private void buildModel()
  {
    NaiveBayes model = new NaiveBayes();
    try
    {
      model.buildClassifier(instances);

      Attribute classAttr = instances.classAttribute();
      for(int k=0;k<classAttr.numValues();k++)
      {
        Entity solution = getAttributeValue(classAttr, k);
        for(int i=0;i<instances.numAttributes();i++)
        {
          if(instances.classIndex() != i)
          {
            Attribute attr = instances.attribute(i);
            double total = 0.0;
            for(int j=0;j<attr.numValues();j++)
              total += model.getProbability(k, i, (double) j);
            for(int j=0;j<attr.numValues();j++)
            {
              Entity value = getAttributeValue(attr, j);
//              System.err.println("Adding relation from "+value+" to "+solution);
              value.addRelation("implies", solution, model.getProbability(k, i, (double) j)/total);
            }
          }
        }
      }

    }
    catch(Exception e)
    {
      e.printStackTrace();
    }

  }

  public static void main(String argv[])
  {
    if(argv.length < 2)
      System.out.println("Arguments: <Weka .arff file> <Creek .km file>");

    try
    {
      BufferedReader arffReader = new BufferedReader(new FileReader(argv[0]));
      Instances wekaModel = new Instances(arffReader);
      WekaToCreekConverter wtc = new WekaToCreekConverter(wekaModel);
      KnowledgeModel km = wtc.getKnowledgeModel();
      ((LocalKnowledgeModel) km).saveAs(argv[1]);
    }
    catch(IOException e)
    {
      e.printStackTrace();
    }
    catch(UnsupportedClassTypeException e)
    {
      System.out.println(e.getMessage());
    }
  }
}
